#!/bin/sh
#============================================================================
# This software is distributed under the terms of the Livermore Public  
# License version 1.0.   
#	      
#       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
#                              
# Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
#
# Everyone is permitted to copy and distribute this license document.  The 
# intent of this license is to establish freedom to share and change the 
# software regulated by this license under the open source model.
#
# This license applies to any software containing a notice placed by the 
# copyright holder saying that it may be distributed under the terms of the 
# Livermore Public  License version 1.0. Such software is herein referred to as
# the Software. This license covers modification and distribution of the 
# Software, use of third-party application programs based on the Software, and
# development of free software which uses the Software.
#
# Granted Rights
#
# 1. You are granted the non-exclusive rights set forth in this license
#    provided you agree to and comply with any and all conditions in this
#    license. Whole or partial distribution of the Software, or software items
#    that link with the Software, in any form signifies acceptance of this 
#    license.
#
# 2. You may copy and distribute the Software in unmodified form provided that
#    the entire package, including, but not restricted to, copyright, trademark
#    notices and disclaimers, as released by the initial developer of the
#    Software, is distributed.
#
# 3. You may make modifications to the Software and distribute your
#    modifications, in a form that is separate from the Software, such as
#    patches. The following restrictions apply to modifications:
#
#    a. Modifications must not alter or remove any copyright notices in the
#       Software.
#
#    b. When modifications to the Software are released under this license, a
#       non-exclusive royalty-free right is granted to the initial developer of
#       the Software to distribute your modification in future versions of the
#       Software provided such versions remain available under these terms in
#       addition to any other license(s) of the initial developer.
#
# 4. You may distribute machine-executable forms of the Software or 
#    machine-executable forms of modified versions of the Software, provided
#    that you meet these restrictions:
#
#    a. You must include this license document in the distribution.
#
#    b. You must ensure that all recipients of the machine-executable forms are
#       also able to receive the complete machine-readable source code to the
#       distributed Software, including all modifications, without any charge 
#       beyond the costs of data transfer, and place prominent notices in the
#       distribution explaining this.
#
#    c. You must ensure that all modifications included in the machine-
#       -executable forms are available under the terms of this license.
#
# 5. You may use the original or modified versions of the Software to compile,
#    link and run application programs legally developed by you or by others. 
#      
# 6. You may develop application programs, reusable components and other 
#    software items that link with the original or modified versions of the
#    Software. These items, when distributed, are subject to the following
#    requirements:
#                   
#    a. You must ensure that all recipients of machine-executable forms of
#       these items are also able to receive and use the complete 
#       machine-readable source code to the items without any charge beyond
#       the costs of data transfer.
#           
#    b. You must explicitly license all recipients of your items to use and 
#       re-distribute original and modified versions of the items in both 
#       machine-executable and source code forms. The recipients must  be able
#       to do so without any charges whatsoever, and they must be able to 
#       re-distribute to anyone they choose.
# 
#    c. If the items are not available to the general public, and the initial
#       developer of the Software requests a copy of the items, then you must
#       supply one.
#	 
# Limitations of Liability
#
# In no event shall the initial developers or copyright holders be liable for 
# any damages whatsoever, including - but not restricted to - lost revenue or 
# profits or other direct, indirect, special, incidental or consequential 
# damages, even if they have been advised of the possibility of such damages, 
# except to the extent invariable law, if any, provides otherwise.
#
# THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
# EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
# This license is governed by the Laws of the State of Texas and any disputes
# shall be decided by mediation.
#
# T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
# Avenue Software, Inc. 
#=============================================================================
topdir="`pwd`"
if [ "`basename $topdir`" != 100 ]; then
  echo "Execute this script in the 100 directory."
  exit 1
fi

OS="`uname`"

nofr=0
while [ $# -gt 0 ]; do
  case $1 in
    -nofastresolve)  nofr=1
                     shift
                     ;;
    *)               echo "Usage: $0 [ -nofastresolve ]"
                     exit 1
                     ;;
  esac
done

# Build libdes
echo "Builing libdes ..."
cd $topdir/src/libdes
make

# Build libsrc
echo "Builing libsrc ..."
cd $topdir/src/libsrc
make

# Build gwuser
echo "Builing gwuser ..."
cd $topdir/src/util/gwuser
make

# Build sd
echo "Builing sd ..."
cd $topdir/src/sd
make

# Build dss/agent
echo "Builing dss/agent ..."
cd $topdir/src/dss/agent
make

# Configure socks
echo "Configuring socks ..."
cd $topdir/src/socks
rm config.status config.cache
./configure --with-pidfile=/etc/sockd.conf -disable-libwrap

# Configure apache
echo "Configuring apache ..."
cd $topdir/src/apache
rm config.status
./configure --prefix=/etc/apache --sysconfdir=/etc/apache
# Fix up Configuration file if necessary
echo "Configuring apache/src ..."
cd src
if [ "$OS" = "Linux" ]; then
  grep "^EXTRA_LDFLAGS" Configuration | grep -e "-ldb" >/dev/null
  if [ $? -ne 0 ]; then
    cp -p Configuration Configuration.bak
    sed '/^EXTRA_LDFLAGS/s/=/=-ldb /' Configuration > Configuration.new
    mv Configuration.new Configuration
  fi
else
  cp -p Configuration Configuration.bak
  sed '/^EXTRA_LDFLAGS/s/-ldb //' Configuration > Configuration.new
  mv Configuration.new Configuration
fi
./Configure

# Setup fastresolve only if -nofastresolve is not present
if [ $nofr = 0 ]; then
  #Configure and make adns
  echo "Configuring adns ..."
  cd $topdir/src/fastresolve/adns
  rm config.*
  ./configure
  echo "Building adns ..."
  make

  #Configure and make db
  echo "Configuring db ..."
  cd $topdir/src/fastresolve/db/dist
  rm config.status config.cache config.log
  if [ "$OS" = "Linux" ]; then
    # Linux is simple
    ./configure --enable-cxx
  else
    # Handle AIX and Solaris
    if  [ "$OS" = "AIX" ]; then
      CPLUSPLUS=xlC
    else
      CPLUSPLUS=CC
    fi
    type g++ | grep "not found" > /dev/null
    if [ $? -ne 0 ]; then
      # gnu compiler is there, see if libstdc++ is there
      if [ -f /usr/local/lib/libstdc++.so ] ||
			[ -f /usr/local/lib/libstdc++.a ]; then
        LD_LIBRARY_PATH=/usr/local/lib:$LD_LIBRARY_PATH CCC=g++ CC=gcc ./configure --enable-cxx
      else
        CCC=$CPLUSPLUS ./configure --enable-cxx
      fi
    else
      CCC=$CPLUSPLUS ./configure --enable-cxx
    fi
  fi
  echo "Building db ..."
  make

  #Configure and make perl
  if [ "$OS" != "Linux" ]; then
    echo "Configuring perl ..."
    cd $topdir/src/fastresolve/perl
    rm Policy.sh config.sh
    ./Configure -des
    if [ "$OS" = "AIX" ]; then
      make perl.exp
    fi
    echo "Building perl ..."
    make

  #Configure and make zlib
    echo "Configuring zlib ..."
    cd $topdir/src/fastresolve/zlib
    ./configure
    echo "Building zlib ..."
    make
  fi

  #Configure fastresolve
  echo "Configuring fastresolve ..."
  cd $topdir/src/fastresolve/fastresolve
  rm config.status config.cache config.log
  OSNAME=$OS
  if [ "$OSNAME" = "SunOS" ]; then
    type CC 2>/dev/null >/dev/null
    if [ $? -eq 0 ]; then
      OSNAME="Solaris"
    else
      OSNAME="Solaris with gcc and Gnu C++"
    fi
  fi
  # Get line number from the notes to use as the config command
  LINE=`grep -n "On ${OSNAME}:$" LSLI.notes | cut -d: -f1`
  LINE=`expr $LINE + 1`
  sed -n "${LINE}p" LSLI.notes > lsli.sh
  sh lsli.sh
  rm lsli.sh
  cd $topdir/src/fastresolve/fastresolve/dns-terror
  cp -p Makefile Makefile.bak
  sed "/^CXXFLAGS/s/-L\.\.\//-L\.\.\/\.\.\//g" Makefile > Makefile.tmp
  sed "/^CFLAGS/s/-L\.\.\//-L\.\.\/\.\.\//g" Makefile.tmp > Makefile
  rm Makefile.tmp
fi
