/*============================================================================
 * filename:		hash.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=============================================================================
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1997-2000.
 * All right reserved.
 *
 *
 * Date written:	February 18, 1998
 * last update:		February   8, 2000
 *
 *
 *============================================================================*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "mod_proxy.h"
#include "http_core.h"
#include "http_log.h"

#define BITMASK (BITS << pc->nshift)
#ifdef DEPTH
static int Depth[BITS];
#endif

#define Hash(Data) ((Data & BITMASK) >> pc->nshift)

/*============================================================================
 * Finds the hash data in the hash table.
 * If found, returns a pointer to the data
 * If not found, returns NULL
 *===========================================================================*/
static Node *FindNode (T Data, proxy_server_conf *pc) {
    Node *p;
    p = pc->HashTable[Hash(Data)];
    while (p && !CompEQ(p->Data, Data))
        p = p->Next;
    return p;
}

/*============================================================================
 * Input:
 *	Data  -	ip address
 *	site  -	pointer to the site name.
 *	dir   -	pointer to the directory string.
 *	finfo -	optional pointer to redirect path
 *	shared-	pointer to current shared memory area
 *	pc    -	pointer to struct proxy_server_conf
 *
 * Output:
 *	returns NULL	if unsuccessful
 *		non-NULL if successful and
 * 		and dir is set to point to the dirextory member that was
 *		malloc'ed
 *===========================================================================*/
Node *InsertNode(T Data, char *site, char *dir, struct f_info *finfo,
					char **shared, proxy_server_conf *pc) {
    Node *p, *p0;
    struct dir_ent *newdir;
    HashTableIndex bucket;

    /*====================================================================
     * Will insert at least the directory string entry so we build the entry
     * now.
     *===================================================================*/
    newdir = (struct dir_ent *)*shared;
    *shared += sizeof(struct dir_ent) + (dir ? strlen(dir)+1:0) +
							strlen(site) + 1;
    /* Align on word boundary */
    *shared = (char *)((((int)*shared + 3) >> 2) << 2);

    /*====================================================================
     * Set the pointers in the new dir_ent for dir and site
     * We have allocated enough space above for dir_ent, the dir string and
     * the site string.
     * The site pointer is now set to just after dir_ent
     * The dir pointer is set to just after the site string.
     *===================================================================*/
    newdir->site = (char *)newdir + sizeof(struct dir_ent);
    strcpy(newdir->site, site);
    if (dir) {
	newdir->dir = newdir->site + strlen(site) + 1;
	strcpy(newdir->dir, dir);
    }
    else
	newdir->dir = NULL;

    newdir->finfo = finfo;

    /*====================================================================
     * If IP already in hash table, insert only the new directory.
     *===================================================================*/
    if ((p = FindNode(Data, pc)) != NULL) {
	newdir->Next = p->Dirs;
	p->Dirs = newdir;
	return(p);
    }

    /* insert node at beginning of list */
    bucket = Hash(Data);
    p = (Node *)*shared;
    *shared += sizeof(Node);
    /* Align on word boundary */
    *shared = (char *)((((int)*shared + 3) >> 2) << 2);

    p0 = pc->HashTable[bucket];
    pc->HashTable[bucket] = p;
#ifdef DEPTH
    Depth[bucket]++;
#endif
    p->Next = p0;
    p->Data = Data;
    newdir->Next = p->Dirs;
    p->Dirs = newdir;
    return p;
}

/*
void DeleteNode(T Data) {
    Node *p0, *p;
    struct dir_ent *d,*nextdir;
    HashTableIndex bucket;

*/
    /* find node */
/*
    p0 = 0;
    bucket = Hash(Data);
    p = HashTable[bucket];
    while (p && !CompEQ(p->Data, Data)) {

        p0 = p;
        p = p->Next;
    }
    if (!p) return;
*/

    /* p designates node to delete, remove it from list */
/*
    if (p0)
*/
        /* not first node, p0 points to previous node */
/*
        p0->Next = p->Next;
    else
*/
        /* first node on chain */
/*
        HashTable[bucket] = p->Next;

    for (d=p->Dirs; d; d=nextdir) {
	nextdir = d->Next;
	free(d);
    }
    free (p);
}
*/

/*============================================================================
 * Check to see if a IP/path combination is blocked
 *
 * Input:
 *		ip - unsigned long IP address
 *		path - char * path name
 *
 * Output:
 *	returns	: 0 if not blocked
 *		  <0 if blocked, there is no remapped http
 *		  >0 if blocked, the returned value points to the
 *		  remapped http string
 *
 * A request is blocked if all five of these conditions are true:
 * 1. The clients IP address is not on the priviledged clients list
 * 2. The site's IP address matches one on the block list
 * 3. No time range was specified for this IP address on the block list
 * 4. The matched block list entry has a NULL dir field			OR
 * 	The matched block list entry has a dir field = "*"		OR
 *	The matched block list entry has a dir field = the path in the URL
 * 5. The site in the request URL is an IP address rather than a name	OR
	The matched block list entry has a site field = "*"		OR
 *	The matched block list entry has a site field = the one in the URL
 *===========================================================================*/
char *IsDirBlocked(unsigned long ip, char *site, char *path,
					proxy_server_conf *pc, request_rec *r)
{
    int i;
    unsigned long blocked_ip;
    char *blocked_dir;
    Node *node;
    struct in_addr *privlist;
    
    /*=======================================================================
     * If a list of priviledged clients were specified, let thru clients
     * on the list
     *======================================================================*/
    if (pc->privlist) {
	privlist = (struct in_addr *)pc->privlist->elts;
	/*===================================================================
	 * If the client IP is on the priviledged clients list, let it thru
	 *==================================================================*/
	for (i = 0; i < pc->privlist->nelts; i++) {
	    if (r->connection->remote_addr.sin_addr.s_addr ==
			    				privlist[i].s_addr)
	    return(NULL);
	}
    }

    if ((node =FindNode(ip, pc))) {
	for (blocked_ip=node->Data; node; blocked_ip = NEXTNODE(node)) {
	    if (ip == blocked_ip) {
		struct dir_ent *d = node->Dirs;
		time_t cur = time((time_t *)NULL);
		struct tm *local = localtime(&cur);
		/*===========================================================
		 * If it falls outside the block time range, do not block.
		 * This mess says....
		 * Do not block if
		 *   The first time range (tr1) is present	AND
		 *   current time falls outside tr1		AND
		 *   either
		 *     second time range (tr2) not present OR
		 *     curent time falls outside tr2
		 *==========================================================*/
		if ((d->finfo->bhour1 != 0 || d->finfo->bmin1 != 0 ||
				d->finfo->ehour1 != 0 || d->finfo->emin1 != 0)&&
			(
			 (	local->tm_hour < d->finfo->bhour1	||
				(local->tm_hour == d->finfo->bhour1 &&
				 local->tm_min < d->finfo->bmin1)	||
				local->tm_hour > d->finfo->ehour1	||
				(local->tm_hour == d->finfo->ehour1 &&
				 local->tm_min > d->finfo->emin1)	 )    &&
			 (	(!d->finfo->bhour2 && !d->finfo->bmin2 &&
				 !d->finfo->ehour2 && !d->finfo->emin2)	||
				(local->tm_hour < d->finfo->bhour2	||
				 (local->tm_hour == d->finfo->bhour2 &&
				  local->tm_min < d->finfo->bmin2)	||
				 local->tm_hour > d->finfo->ehour2	||
				 (local->tm_hour == d->finfo->ehour2 &&
				  local->tm_min > d->finfo->emin2))	 ))    )
		    continue;
		for (blocked_dir = d->dir; d; blocked_dir = NEXTDIR(d)) {
		    if ((!blocked_dir ||
			 blocked_dir[0] == '*' ||
			 !strncmp(path,blocked_dir,strlen(blocked_dir))) &&
			(strspn(site, "0123456789.") == strlen(site) ||
			 *(d->site) == '*' ||
			 !strncmp(d->site, site, strlen(d->site)))) {
			ap_log_error(NULL, 0, APLOG_NOTICE|APLOG_NOERRNO,
					r->server,
					"IP Address %s and path (/%s) is "
					"blocked by file \"%s\".  "
					"Requesting Host %s",
					site, path, d->finfo->filename,
					ap_get_remote_host(r->connection,
						r->per_dir_config,REMOTE_NAME)
					);
			if (d->finfo->remap)
			    return(d->finfo->remap);
			else {
			    return((char *)-1);
			}
		    }
		}
		return(NULL);
	    }
	}
    }
    return(NULL);
}


#ifdef DEPTH
int HashDepth()
{
    int depth=0;
    int i;

    for (i=0; i<BITS; i++) {
	if (Depth[i] > depth)
	    depth = Depth[i];
    }
    return (depth);
}
#endif
