/*============================================================================
 * filename:		loadhash.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=============================================================================
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1997-2000.  
 * All right reserved.
 *
 *
 * Date written:	February 18, 1998
 * last update:		February   8, 2000
 *
 *
 *============================================================================*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <syslog.h>
#include "mod_proxy.h"
#include "http_log.h"

#ifndef linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif

/*============================================================================
 * Parse a char string in the following formats into 24 hour time in hour
 * and minute
 * nn		n
 * nna		na	nnA	nA
 * nnam		nam	nnAM	nAM	nnAm	nAm	nnaM	naM
 * nnp		np	nnP	nP
 * nnpm		npm	nnPM	nPM	nnPm	nPm	nnpM	npM
 * nn:nn	n:nn
 * nn:nna	n:nna	nn:nnA	n:nnA
 * nn:nnam	n:nnam	nn:nnAM	n:nnAM	nn:nnAm	n:nnAm	nn:nnaM	n:nnaM
 * nn:nnp	n:nnp	nn:nnP	n:nnP
 * nn:nnpm	n:nnpm	nn:nnPM	n:nnPM	nn:nnPm	n:nnPm	nn:nnpM	n:nnpM
 * 
 * where n is [0-9]
 * 
 * Returns	NULL if successful, hour/minute will contain the parsed time
 * 		ptr to error message if failed
 *===========================================================================*/
char *parse_time(char *timep, short *hour, short *minute, cmd_parms *parms) {
    int i;
    int len;
    int tmpint;
    short ampm;
    char *cp;
    char *msgp;
    char tmp[8];
    
    /*=======================================================================
     * Look for illegal time values.
     *======================================================================*/
    i = strspn(timep, "0123456789:");
    len = strlen(timep) - i;		/* Length of AM or PM string	*/
    if (i == 0 || i > 5 || len > 2 ||
	(len >= 1 && timep[i] != 'p' && timep[i] != 'a' &&
			  timep[i] != 'P' && timep[i] != 'A') ||
	(len == 2 && timep[i+1] != 'm' && timep[i+1] != 'M')	) {
	msgp = ap_pstrcat(parms->temp_pool, "LoadHash: Invalid time "
			"argument \"", timep, "\" specified.", NULL);
	ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, parms->server, msgp);
	return(msgp);
    }
    strcpy(tmp, timep);
    cp = strtok(tmp, ":aApP");	/* Above if's already ensure this != NULL */
    tmpint = atoi(cp);

    /*=======================================================================
     * If a time is followed by 'pm', add 12 to the hour - unless if the
     * hour is 12 (12pm), then leave it at 12.
     * 12 am becomes 0 hour.  12 with no am or pm will be 12 noon.
     *======================================================================*/
    if (timep[i] == 'p' || timep[i] == 'P')
	if (tmpint == 12)
	    ampm = 0;
        else
	    ampm = 12;
    else {
	if (timep[i] == 'a' || timep[i] == 'A') {
	    if (tmpint == 12)
		tmpint = 0;
	}
	ampm = 0;
    }
    tmpint += ampm;

    if (tmpint < 0 || tmpint > 23 || (tmpint == 0 && ampm != 0)) {
	msgp = ap_pstrcat(parms->temp_pool, "LoadHash: Invalid time "
			"argument \"", timep, "\" specified.", NULL);
	ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, parms->server, msgp);
	return(msgp);
    }

    *hour = tmpint;
    
    *minute = 0;
    cp = strtok(NULL, "aApPmM");
    if (cp == NULL)
	return(NULL);
    /*=======================================================================
     * Get the minute
     *======================================================================*/
    tmpint = atoi(cp);
    if (tmpint < 0 || tmpint > 59) {
	msgp = ap_pstrcat(parms->temp_pool, "LoadHash: Invalid time "
			"argument \"", timep, "\" specified.", NULL);
	ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, parms->server, msgp);
	return(msgp);
    }
    *minute = tmpint;
    return(NULL);
}

/*============================================================================
 * Read a file and insert the IP address and paths into the hashtable
 *
 * Input:
 *		ifile - char * path name of the input file.
 *			ifile has to be in the form:
 *			server_name ip_address path
 *		optarg1 - char * optional block time or redirect path
 *		optarg2 - char * optional redirect path
 *		parms - ptr to struct proxy_server_conf
 *
 * Output:
 *	returns	: NULL if Successful
 *		  ptr to error msg if Unsuccessful (As of now, LoadHash() calls
 *			ap_log_error() directly and puts out a message and
 *			always returns NULL so processing coan continue after
 *			a list file error.
 *===========================================================================*/
char *LoadHash(char *ifile, char *optarg1, char *optarg2, cmd_parms *parms,
					proxy_server_conf *pc)
{
    int n=0;
    int fd=0;
    int nentries=0;
    int shm_size;
    char *cp;
    char *shared;
    unsigned long ip;
    char *server,*path;
    char tokensep[]=" \t\n\r";
    char *msgp;
    char *remap;
    char line[4096];
    char *permremap = NULL;
    char *permifile = NULL;
    short bhour1 = 0;
    short bmin1 = 0;
    short ehour1 = 0;
    short emin1 = 0;
    short bhour2 = 0;
    short bmin2 = 0;
    short ehour2 = 0;
    short emin2 = 0;
    struct f_info *permfinfo = NULL;
    FILE *infile;
    struct stat statbuf;
#if defined(_AIX) || defined(__sun) || defined(linux)
#define MAPOPEN open
#define MAPFLAGS O_RDWR
#define MAPPED_FILE "/dev/zero"
#else
#if defined(__hppa)
#define MAPOPEN open
#define MAPFLAGS O_RDWR|O_CREAT
#define MAPPED_FILE "/etc/firewall/_mmapfile"
#else
#if defined(__Lynx__)
#define MAPOPEN shm_open
#define MAPFLAGS O_RDWR|O_CREAT
#define MAPPED_FILE "/etc/firewall/_mmapfile"
#endif
#endif
#endif

    /*=======================================================================
     * Process the optional arguments
     *======================================================================*/
    if (optarg1 != NULL) {
	char timechars[] = "0123456789:-apm,";
	
	/*===================================================================
	 * Determine if a time argument was specified.
	 *==================================================================*/
	if ((*optarg1 >= '0' && *optarg1 <= '9') &&
			strspn(optarg1, timechars) == strlen(optarg1)) {
	    char tmp[32];	/* Max len nn:nnAM-nn:nnPM,nn:nnAM-nn:nnPm  */
	    char *tmp1, *tmp2;
	    char *cp;
	    char *t1;
	    char *t2;
	    
	    if (strlen(optarg1) > sizeof(tmp) - 1) {
		msgp = ap_pstrcat(parms->temp_pool, "LoadHash: ProxyBlockList "
				"time argument \"", optarg1, "\" too long.",
				NULL);
		ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO,
				parms->server, msgp);
		return(msgp);
	    }
	    strcpy(tmp, optarg1);
	    tmp1 = strtok(tmp, ",");
	    tmp2 = strtok(NULL, ",");

	    /*===============================================================
	     * Get the first time range
	     *==============================================================*/
	    t1 = strtok(tmp1, "-");
	    if ((t2 = strtok(NULL, "-")) == NULL) {
		msgp = ap_pstrcat(parms->temp_pool, "LoadHash: Invalid time "
			"range \"", optarg1, "\" specified.", NULL);
		ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO,
				parms->server, msgp);
		return(msgp);
	    }
	    if ((msgp = parse_time(t1, &bhour1, &bmin1, parms)) != NULL)
		return(msgp);
	    if ((msgp = parse_time(t2, &ehour1, &emin1, parms)) != NULL)
		return(msgp);
	    if (bhour1 > ehour1 || (bhour1 == ehour1 && bmin1 > emin1)) {
		msgp = ap_pstrcat(parms->temp_pool, "LoadHash: Invalid time "
			"range \"", optarg1, "\" specified.", NULL);
		ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO,
				parms->server, msgp);
		return(msgp);
	    }

	    /*===============================================================
	     * Get the second time range if it is there
	     *==============================================================*/
	    if (tmp2 != NULL ) {
		t1 = strtok(tmp2, "-");
		if ((t2 = strtok(NULL, "-")) == NULL) {
		    msgp = ap_pstrcat(parms->temp_pool, "LoadHash: Invalid "
			"time range \"", optarg1, "\" specified.", NULL);
		    ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO,
				parms->server, msgp);
		    return(msgp);
		}
		if ((msgp = parse_time(t1, &bhour2, &bmin2, parms)) != NULL)
		    return(msgp);
		if ((msgp = parse_time(t2, &ehour2, &emin2, parms)) != NULL)
		    return(msgp);
		if (bhour2 > ehour2 || (bhour2 == ehour2 && bmin2 > emin2)) {
		    msgp = ap_pstrcat(parms->temp_pool, "LoadHash: Invalid "
			"time range \"", optarg1, "\" specified.", NULL);
		    ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO,
				parms->server, msgp);
		    return(msgp);
		}
	    }
	    remap = optarg2;
	}
	else
	    remap = optarg1;
    }
    else
	remap = NULL;

    /*=======================================================================
     * Open the list file
     *======================================================================*/
    if ((infile = fopen(ifile,"r")) == NULL) {
	msgp = ap_pstrcat(parms->temp_pool, "LoadHash: Error opening blocklist "
			"file ", ifile, ".", NULL);
	ap_log_error(APLOG_MARK, APLOG_ERR, parms->server, msgp);
	return(msgp);
    }

    /*=======================================================================
     * Read the list file once to determine how many lines (how many blocked
     * sites) there are.  We don't really care about the contents here, just
     * how many newlines.
     *======================================================================*/
    while (fgets(line,sizeof(line),infile))
	nentries++;

    if (!feof(infile)) {
	msgp = ap_pstrcat(parms->temp_pool, "LoadHash: Error reading file ",
			ifile, ".", NULL);
	ap_log_error(APLOG_MARK, APLOG_ERR, parms->server, msgp);
	return(msgp);
    }

    /*=======================================================================
     * Reposition file pointer to beginning of file
     *======================================================================*/
    if (fseek(infile, 0, SEEK_SET) < 0) {
	msgp = ap_pstrcat(parms->temp_pool, "LoadHash: Error resetting file ",
			"pointer for file ",
			ifile, ".", NULL);
	ap_log_error(APLOG_MARK, APLOG_ERR, parms->server, msgp);
	return(msgp);
    }

    /*=======================================================================
     * stat the input blocklist file to get its size.
     * The memory area size will be set to the size of the file plus
     * nentries times the size of the structures.
     *======================================================================*/
    if (stat(ifile,&statbuf) < 0) {
	msgp = ap_pstrcat(parms->temp_pool, "LoadHash: Error stat-ing blocklist "
			"file ", ifile, ".", NULL);
	ap_log_error(APLOG_MARK, APLOG_ERR, parms->server, msgp);
	return(msgp);
    }

    if (statbuf.st_size <= 0) {
	msgp = ap_pstrcat(parms->temp_pool, "LoadHash: File ", ifile, " has "
			"zero length.", NULL);
	ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, parms->server, msgp);
	return(msgp);
    }

    /*=======================================================================
     * Compute shared memory size
     * This size is really bigger than what we need but it is simpler.
     * The worst case scenario is that each line entry in the list file is
     * allocated one Node structure, one dir_ent structure, one buffer for the
     * directory path and one buffer for the site name - plus there is one
     * buffer for a remap path for each file and a struct for info on
     * associated with the block file.
     * The statbuf.st_size (size of the list file) is more than the worst case
     * of one directory path buffer and one site name buffer for each entry.
     * The nentries * (sizeof(Node) + sizeof(struct dir_ent)) takes care of
     * the 2 structures.
     *======================================================================*/
    shm_size = statbuf.st_size + sizeof(struct f_info) + strlen(ifile) + 1 +
			nentries * (sizeof(Node) + sizeof(struct dir_ent));

#if !defined(__hppa)
    if ((fd = MAPOPEN(MAPPED_FILE, MAPFLAGS, 0600)) < 0) {
	msgp = ap_pstrcat(parms->temp_pool, "LoadHash: Error opening temporary "
			"map file.", NULL);
	ap_log_error(APLOG_MARK, APLOG_ERR, parms->server, msgp);
	return(msgp);
    }
#endif
#if defined(__Lynx__)
    /*=======================================================================
     * Extend the mapped area to the maximum size
     *======================================================================*/
    if (ftruncate(fd, shm_size) < 0) {
	msgp = ap_pstrcat(parms->temp_pool, "LoadHash: Error extending temporary "
			"map file.", NULL);
	ap_log_error(APLOG_MARK, APLOG_ERR, parms->server, msgp);
	return(msgp);
    }
#endif

#if defined(__hppa)
    if ((shared =  mmap(0, shm_size, PROT_READ | PROT_WRITE,
				MAP_ANONYMOUS|MAP_PRIVATE, -1, 0))
							== (caddr_t) -1) {
#else
    if ((shared =  mmap(0, shm_size, PROT_READ | PROT_WRITE,
				MAP_PRIVATE, fd, 0)) == (caddr_t) -1) {
#endif
	msgp = ap_pstrcat(parms->temp_pool, "LoadHash: Error mapping temporary "
			"map file.  ", NULL);
	ap_log_error(APLOG_MARK, APLOG_ERR, parms->server, msgp);
	return(msgp);
    }
    close(fd);
#if defined(__hppa) || defined(__Lynx__)
    unlink(MAPPED_FILE);
#endif

    if (remap) {
	permremap = shared;
	strcpy(permremap, remap);
	shared += strlen(permremap) + 1;
	/* Align on word boundary */
	shared = (char *)((((int)shared + 3) >> 2) << 2);
    }
    /*=======================================================================
     * Fill in the struct f_info with information associated with the
     * block file.
     *======================================================================*/
    permfinfo = (struct f_info *)shared;
    shared += ((sizeof(struct f_info) + 3) >> 2) << 2;
    permfinfo->remap = permremap;
    permfinfo->bhour1 = bhour1;
    permfinfo->bmin1 = bmin1;
    permfinfo->ehour1 = ehour1;
    permfinfo->emin1 = emin1;
    permfinfo->bhour2 = bhour2;
    permfinfo->bmin2 = bmin2;
    permfinfo->ehour2 = ehour2;
    permfinfo->emin2 = emin2;

    /*=======================================================================
     * Put basename of the blocklist file into the info structure.
     *======================================================================*/
    permifile = (char *)shared;
    if ((cp = strrchr(ifile, '/')) == NULL)
	cp = ifile;
    else
	cp++;		/* Point to 1st char after the last '/'	*/
    strcpy(permifile, cp);
    permfinfo->filename = permifile;
    shared += ((strlen(cp) + 3) >> 2) << 2;
    
    if (remap) {
	struct f_info **fp;
	/*===================================================================
	 * Create a list of block file info structures.
	 *==================================================================*/
	if (pc->blockfinfo == NULL) {
	    pc->blockfinfo = ap_make_array(parms->pool, 1,
			    		sizeof(struct f_info *));
	}
	fp = (struct f_info **)ap_push_array(pc->blockfinfo);
	*fp = permfinfo;
    }

    while (fgets(line,sizeof(line),infile)) {
	n++;
	server = strtok(line,tokensep);
	if (!server) {
	    char tmp[16];
	    sprintf(tmp, "%d", n);
	    msgp = ap_pstrcat(parms->temp_pool, "LoadHash: Error parsing line ",
			 tmp, " of ", ifile, " - no server.", NULL);
	    ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, parms->server,
			msgp);
	    continue;
	}
	cp = strtok(NULL,tokensep);
	if (cp == NULL || (ip = inet_addr(cp)) == 0 || ip == -1) {
	    char tmp[16];
	    sprintf(tmp, "%d", n);
	    msgp = ap_pstrcat(parms->temp_pool, "Error parsing line ", tmp,
			" of ", ifile, " - no IP.", NULL);
	    ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, parms->server,
			msgp);
	    continue;
	}
	path = strtok(NULL,tokensep);

	if (path)
	    ap_unescape_url(path);
	
	InsertNode(ip,server,path,permfinfo,&shared,pc);
    }
    if (!feof(infile)) {
	msgp = ap_pstrcat(parms->temp_pool, "LoadHash: Error reading file ",
			ifile, ".  ", sys_errlist[errno], NULL);
	ap_log_error(APLOG_MARK, APLOG_ERR, parms->server, msgp);
	return(msgp);
    }
    return(NULL);
}
