/*============================================================================
 * filename:		../aproxy/apconf.c					
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1997-2000.
 * All rights reserved.
 *		
 * Date written:	May        6, 1999
 * Last modified:	February   8, 2000 - added LPL
 *	                         
 * function:                       
 *	The apconf function reads the aproxy configuration file.
 *
 *  
 *
 * file format:             
 *
 * Manditory commands
 *  
 *	gproxpath 	= pathname	-      
 *
 * Optional Commands
 *     
 * 	timeout 	= nnn 		- number of seconds
 *	trace		= yes|no	- turn tracing on or off, default is no
 *	debug		= n		- turn debugging on to level n
 *	nprocs		= nnn nnn nnn	- number of processes to start,
 *					  min number of spare idle processes,
 *					  max number of spare idle processes,
 *	maxprocs	= nnn		- max number of aproxy processes
 *	maxuse		= nnn		- max number of requests a child
 *					  process handles before terminating
 *	checkprocs	= nnn		- interval in seconds between wakeups
 *					  to check on number of child processes
 *
 * Access Rules 
 *           
 *	permit from src_host to local_host local_port redirect dest_host \
 *				dest_port [ooba name ooba_port]  \
 *				[using local_address] [ignorerst sleep_time]
 *	permit from src_host to local_host local_port \
 *				redirect (dest_host1,dest_host2,...) \
 *				dest_port [ooba name ooba_port] \
 *				[using local_address] [ignorerst sleep_time]
 *
 *	deny   src_host src_port [shell cmd]
 *
 *============================================================================*/
#include	<sys/types.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<syslog.h>
#include	<errno.h>
#include	<string.h>
#include	<sys/time.h>
#include	<netdb.h>               /* defines servent structure      */
#include	<netinet/in.h>
#include	"firewall.h" 
#include	"gwuser.h"
#include	"aproxy.h"  
               
#ifndef linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif

extern long int timezone;

static NS_WGCONF ns_wgconf;		/* non-shared portion of WGCONF */

/*============================================================================
 * function prototypes
 *===========================================================================*/  
char *getshm(int size, char *mfile);
int permit_cmd(int op, WGCONF **ptr, int *size, int line, char * proxycfgfile);
int ooba(char *destIP,  char * username, int portid,
         struct WGCONF *ptr);
/*=============================================================================
 *
 * inputs:	ptr to source host IP address 
 *		ptr to local host IP address
 *		local host port id 
 *		ptr to destination host IP address
 *		ptr to destination port
 * 
 * outputs:	returns 0 if successful
 *		returns -1 if failed
 *===========================================================================*/
WGCONF *getwgconf() 
{ 
   char 	proxycfgfile[] = APROXYCFGFILE;
   char		record[MAXRECLN];
   char		*shmptr;
   WGCONF	*ptr;
   FILE 	*cfgfp; 
   char 	*cp; 
   char		*cp2;  
   char		*cp3;   
   char		*token1;
   char		*token2; 
   char		*token3; 
   char		tokensep[] = "\t,:;= \n";
   int		hostfound  = 0;  
   int		lhostfound = 0;
   int		line	   = 0; 
   int		proxpathfound = 0;
   int		rtncd;
   int		sport;                  /* source port from config file */
   int		dport;			/* dest port fromconfig file    */
   int		size = 0;
   struct servent	*sp; 		/* ptr to service entry 	*/  

   /*=======================================================================
    * malloc for the base WGCONF structure
    *======================================================================*/
   size = sizeof(WGCONF);
   if ((ptr = malloc(size)) == NULL) 
   {
      syslog(LOG_ERR,"Error: Could not malloc.  %s",sys_errlist[errno]);
      return(NULL);
   }

   /*================================================
    * Initialize the structure
    *===============================================*/
   memset((char *)ptr, 0, sizeof(WGCONF));
   ptr->ns		= &ns_wgconf;		/* Points to non-share area */
   ptr->ns->localport	= -1;
   ptr->ns->destport	= -1;
   ptr->traceflg	= 0;
   ptr->debugflg	= 0;
   ptr->checkprocs	= 5;
   ptr->nprocs		= 20;
   ptr->minspare	= -1;
   ptr->maxspare	= -1;
   ptr->maxprocs	= -1;
   ptr->maxuse		= 50;
   ptr->bufsize		= 4096;
   strcpy(ptr->gproxpath,"/home/hermes");
   /*================================================
    * open proxy conf file 
    *===============================================*/
   if (!(cfgfp = fopen(proxycfgfile, "r")))
   {
      syslog(LOG_ERR,"Error: open failed for %s: %m ",proxycfgfile);
      return(NULL);    			/* open failed signal error */
   } 
   cp = (char *) &record; 
   /*================================================= 
    * Read the config file and look for match 
    *=================================================*/        
   while( fgets(record, MAXRECLN, cfgfp) != NULL)
   {    
      line++;   
      /* skip over comments and null lines */
      if( (*cp == '#') || (*cp == '\n') ) goto getnextline; 
      if( (token1 = strtok(record,tokensep)) == NULL) goto getnextline; 
      /*=========================================================	
       * If timeout record then get timeout value and validate 
       *========================================================*/	
      if( strcmp(token1,"timeout") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for timeout on line"
			   " %d of %s.",
 			   line,proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL); 
         }
	 if( !isalldigits(token2) )
	 { 
            syslog(LOG_ERR,"Error: non numeric value '%s' specified for "
			   "timeout on line %d of %s.",
 			   token2, line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL);       
         } 
         ptr->timeout = atoi(token2); 	/* save timeout value	*/
         goto getnextline;  
      }  
      /*=========================================================	
       * If DNS record then get value and validate 
       *========================================================*/	
      if( strcmp(token1,"DNS") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for DNS on line %d of %s.",
 			   line,proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL); 
         }
	 if( strcmp(token2,"yes") == 0 )
	 { 
	    syslog(LOG_ERR, "Error: DNS = yes option is no longer supported.");
         }  
         else
         if( strcmp(token2,"no") != 0 )
         {  
            syslog(LOG_ERR,"Error: invalid value '%s' specified for DNS on "
			   "line %d of %s. Should be yes or no.",
 	  	   	   token2, line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL);  
         }
         ptr->DNSflg = 0;         
         goto getnextline;  
      }   
      /*=========================================================	
       * If trace record then get trace value and validate 
       *========================================================*/	
      if( strcmp(token1,"trace") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for trace on line %d "
			   "of %s.",
 			   line,proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL); 
         }
	 if( strcmp(token2,"yes") == 0 )
	 { 
            ptr->traceflg = 1;         
         }  
         else
         if( strcmp(token2,"no") == 0 )
	 { 
            ptr->traceflg = 0;         
         }
         else
         {  
            syslog(LOG_ERR,"Error: invalid value '%s' specified for trace on "
			   "line %d of %s. Should be yes or no.",
 	  	   	   token2, line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL);  
         }
         goto getnextline;  
      }  
      /*=========================================================	
       * If debug record then get debug value and validate 
       *========================================================*/	
      if( strcmp(token1,"debug") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for debug on line %d "
			   "of %s.",
 			   line,proxycfgfile);
	    fclose(cfgfp);
	    return(NULL);
         }
	 if( !isalldigits(token2) )
	 { 
            syslog(LOG_ERR,"Error: non numeric value '%s' specified for "
			   "debug on line %d of %s.",
 			   token2, line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL);       
         } 
         ptr->debugflg = atoi(token2);

         goto getnextline;
      }
      /*=========================================================	
       * If proxpath record then get path name and validate 
       * do not allow chroot to: "/" or "/etc" or "/usr" 
       *========================================================*/	
      if( strcmp(token1,"gproxpath") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for gproxpath on line"
			   " %d of %s.",
 			   line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL); 
         }
	 if( strlen(token2) > MAXPATHLN )
	 { 
            syslog(LOG_ERR,"Error: gproxpath name '%s' on line %d of %s"
			   " is longer than %d.",
 			   token2, line, proxycfgfile,MAXPATHLN); 
	    fclose(cfgfp);
	    return(NULL);       
         }
         if( ( strcmp(token2,"/")    == 0 )     ||
             ( strncmp(token2,"/dev",4) == 0 )  || 
             ( strncmp(token2,"/etc",4) == 0 )  || 
	     ( strncmp(token2,"/lib",4) == 0 )  ||
             ( strncmp(token2,"/sbin",5) == 0 ) || 
	     ( strncmp(token2,"/unix",5) == 0 ) ||
	     ( strncmp(token2,"/usr",4) == 0 )  ||   
	     ( strncmp(token2,"/var",4) == 0 )   ) 
         {
            syslog(LOG_ERR,"Error: gproxpath name '%s' on line %d of %s is "
			   "not allowed.",
 			   token2, line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL);    
         }
         strcpy(ptr->gproxpath,token2); 	/* copy gproxpath      	*/
         proxpathfound++;
         goto getnextline;  
      }
      /*==================================================================
       * if nprocs then get the number and optional parms
       * nprocs nnn [minspare [maxspare]]
       *==================================================================*/
      if(strcmp(token1,"nprocs") == 0)
      {   
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for nprocs on line"
			   " %d of %s.",
 			   line,proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL); 
         }
	 if( !isalldigits(token2) )
	 { 
            syslog(LOG_ERR,"Error: non numeric value '%s' specified for "
			   "nprocs on line %d of %s.",
 			   token2, line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL);       
         } 
         ptr->nprocs = atoi(token2); 	/* save number of processes	*/
         if (ptr->nprocs == 0) 
         {
            syslog(LOG_ERR,"Error: nprocs must be greater than 0 on line %d "
			   "of %s.",
 			   line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL);       
         }

         /*===============================================================
          * process optional paremeter minspare if specified
          *===============================================================*/
         if( (token2 = strtok(NULL,tokensep)) == NULL)
            goto getnextline;

	 if( !isalldigits(token2) )
	 { 
            syslog(LOG_ERR,"Error: non numeric value '%s' specified for "
			   "minspare on line %d of %s.",
 			   token2, line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL);       
         } 
         ptr->minspare = atoi(token2); 	/* save min number of spares	*/
         if (ptr->minspare == 0)
         {
            syslog(LOG_ERR,"Error: minspare must be greater than 0 on line %d "
			   "of %s.",
 			   line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL);       
         }

         /*===============================================================
          * process optional paremeter maxspare if specified
          *===============================================================*/
         if( (token2 = strtok(NULL,tokensep)) == NULL)
            goto getnextline;

	 if( !isalldigits(token2) )
	 { 
            syslog(LOG_ERR,"Error: non numeric value '%s' specified for "
			   "maxspare on line %d of %s.",
 			   token2, line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL);       
         } 
         ptr->maxspare = atoi(token2); 	/* save max number of spares	*/
         if (ptr->maxspare < ptr->minspare)
         {
            syslog(LOG_ERR,"Error: maxspare must be greater than minspare (%d) "
				"on line %d of %s.",
 				ptr->minspare, line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL);       
         }

         goto getnextline;
      }

      /*==================================================================
       * if maxuse then get the number
       *==================================================================*/
      if(strcmp(token1,"maxuse") == 0)
      {   
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for maxuse on line"
			   " %d of %s.",
 			   line,proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL); 
         }
	 if( !isalldigits(token2) )
	 { 
            syslog(LOG_ERR,"Error: non numeric value '%s' specified for "
			   "maxuse on line %d of %s.",
 			   token2, line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL);       
         } 
         ptr->maxuse = atoi(token2); 	/* save max number of uses	*/
         if (ptr->maxuse == 0) 
         {
            syslog(LOG_ERR,"Error: maxuse must be greater than 0 on line %d "
			   "of %s.",
 			   line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL);       
         }
         goto getnextline;
      }

      /*==================================================================
       * if maxprocs then get the number
       *==================================================================*/
      if(strcmp(token1,"maxprocs") == 0)
      {   
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for maxprocs on line"
			   " %d of %s.",
 			   line,proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL); 
         }
	 if( !isalldigits(token2) )
	 { 
            syslog(LOG_ERR,"Error: non numeric value '%s' specified for "
			   "maxprocs on line %d of %s.",
 			   token2, line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL);       
         } 
         ptr->maxprocs = atoi(token2); 	/* save max number of processes	*/
         if (ptr->maxprocs == 0) 
         {
            syslog(LOG_ERR,"Error: maxprocs must be greater than 0 on line %d "
			   "of %s.",
 			   line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL);       
         }
         goto getnextline;
      }

      /*==================================================================
       * if checkprocs then get the number
       *==================================================================*/
      if(strcmp(token1,"checkprocs") == 0)
      {   
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for checkprocs on line"
			   " %d of %s.",
 			   line,proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL); 
         }
	 if( !isalldigits(token2) )
	 { 
            syslog(LOG_ERR,"Error: non numeric value '%s' specified for "
			   "checkprocs on line %d of %s.",
 			   token2, line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL);       
         } 
         ptr->checkprocs = atoi(token2); 	/* save check interval	*/
         if (ptr->checkprocs == 0) 
         {
            syslog(LOG_ERR,"Error: checkprocs must be greater than 0 on line "
			 	"%d of %s.",
 				line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL);       
         }
         goto getnextline;
      }

      /*==================================================================
       * if bufsize then get the number
       *==================================================================*/
      if(strcmp(token1,"bufsize") == 0)
      {   
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for bufsize on line"
			   " %d of %s.",
 			   line,proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL); 
         }
	 if( !isalldigits(token2) )
	 { 
            syslog(LOG_ERR,"Error: non numeric value '%s' specified for "
			   "bufsize on line %d of %s.",
 			   token2, line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL);       
         } 
         ptr->bufsize = atoi(token2); 	/* save check interval	*/
         if (ptr->bufsize == 0 || ptr->bufsize > 65535) 
         {
            syslog(LOG_ERR,"Error: checkprocs must be greater than 0 and less "
				"than 65535 on line %d of %s.",
 				line, proxycfgfile); 
	    fclose(cfgfp);
	    return(NULL);       
         }
         goto getnextline;
      }

      /*==================================================================
       * if permit/deny rule then record it
       *==================================================================*/
      if(strcmp(token1,"permit") == 0 || strcmp(token1,"deny") == 0)
      {   
         /*===============================================================
          * realloc for more space if necessary
          *==============================================================*/
         if (ptr->nrules > 0) 
         {
            size += sizeof(RULE);
            if ((ptr = realloc(ptr, size)) == NULL) 
            {
               syslog(LOG_ERR,"Error: Could not realloc.  %s",
							sys_errlist[errno]);
               return(NULL);
            }
         }
         rtncd = permit_cmd(strcmp(token1,"deny") ? PERMIT : DENY,
						&ptr, &size,
						line, proxycfgfile);
         if ( rtncd != 0)          
         {  
            fclose(cfgfp);
            return(NULL);
         }
         goto getnextline;
      }
      else
      {
         syslog(LOG_ERR,"Error: invalid record on line %d of %s, unknown "
		        "keyword %s. Service denied.",
		        line, proxycfgfile, token1);
	 fclose(cfgfp);
         return(NULL);
      }
      getnextline:;
   }

   /*=======================================================================
    * Set default nprocs, minspare, maxspare, maxprocs if necessary
    *=======================================================================*/
   if (ptr->minspare == -1)
      ptr->minspare = ptr->nprocs;

   if (ptr->maxspare == -1)
      ptr->maxspare = ptr->minspare + 5;

   if (ptr->maxprocs == -1)
      ptr->maxprocs = ptr->maxspare + 5;

   if (ptr->maxprocs < ptr->nprocs)
      ptr->maxprocs = ptr->nprocs + 5;

   /*=======================================================================
    * malloc an io buffer for use in the infinite loop
    *=======================================================================*/
   if ((ptr->bufptr = malloc(ptr->bufsize)) == NULL)
   {
      syslog(LOG_ERR, "Error: Could not malloc io buffer.  %m.");
      fclose(cfgfp);
      return(NULL);
   }

   /*=======================================================================
    *
    *=======================================================================*/
   if( proxpathfound != 1)
   {
      syslog(LOG_ERR,"Error: No gproxpath record found in %s.", proxycfgfile);
   }
   fclose(cfgfp);
   /*=======================================================================
    * Now get a chunk of shared memory to hold this structure
    *=======================================================================*/
   if ((shmptr = getshm(size, SHRCONF)) == NULL)
      return(NULL);

   memcpy(shmptr, ptr, size);
   free(ptr);
   return((WGCONF *)shmptr);   /* no match found */    
}

/*============================================================================
 * Process the permit command.
 * Call the new function if using the new style of permit statement. 
 *    old:   permit src_IP src_port to dest_IP dest_port
 *    new:   permit src_IP to local_IP local_port redirect dest_IP dest_port 
 *
 * Return:                
 *	   -1 on error
 *          0 if match not found
 *          1 if permitted
 *	   Also ptrp may be changed to point to a new realloc'ed area
 *===========================================================================*/
int permit_cmd(int op, WGCONF **ptrp, int *sizep, int line, char * proxycfgfile)
{  
   char		*cp2;
   char		*cp3;   
   struct GWUSER gwu; 
   struct GWUSER *gwuptr  = &gwu; 
   int		portid	  = 0; 
   int		sport;                  /* source port from config file */  
   struct servent	*sp; 		/* ptr to service entry 	*/  
   char		*token2; 
   char		tokensep[] = "\t,:;= \n";
   int		newsyntax=0;
   WGCONF	*ptr = *ptrp;
   int		nrr = 0;		/* num of addn round robin rules*/
   char		rrstring[MAXRECLN];

   /*===============================================================  
    * Store the op - PERMIT or DENY
    *===============================================================*/  
   (ptr->rulep)[ptr->nrules].op = op;

   /*===============================================================  
    * Set defaults
    *===============================================================*/  
   (ptr->rulep)[ptr->nrules].roundrobin = 0;	/* assume no rrobin */
   (ptr->rulep)[ptr->nrules].lastused.tv_sec = (time_t)0;
   (ptr->rulep)[ptr->nrules].lastused.tv_usec = 0;
   strcpy((ptr->rulep)[ptr->nrules].localaddr2, "");
   strcpy((ptr->rulep)[ptr->nrules].exitname, "");
   (ptr->rulep)[ptr->nrules].ign_rst = 0;

   /*===============================================================  
    * Set default ooba parms in RULE
    *===============================================================*/  
   memset(ptr->rulep[ptr->nrules].oobauser, 0,
			sizeof(ptr->rulep[ptr->nrules].oobauser));
   ptr->rulep[ptr->nrules].oobaport = -1;
   
   /*===============================================================  
    * Next token can be "from" or an ip address
    *===============================================================*/  
   if( (token2 = strtok(NULL,tokensep)) == NULL) /* get from  */
   { 
      syslog(LOG_ERR,"Error: permit statement on line %d of %s is missing "
		     "parameters.",
 		     line,proxycfgfile);  
      return(-1);
   }

   /*===============================================================  
    * Check for new permit stmt syntax. 
    *===============================================================*/  
   if( strcmp(token2,"from") == 0)
   {
      newsyntax = 1;
      /*===============================================================  
       * Next token is from ip address
       *===============================================================*/  
      if( (token2 = strtok(NULL,tokensep)) == NULL) /* get source host  */
      {
         syslog(LOG_ERR,"Error: permit/deny statement on line %d of %s is "
		     "missing parameters.",
 		     line,proxycfgfile);  
         return(-1);
      }
   }

   /*===============================================================  
    * Store from address
    *===============================================================*/  
   if (strlen(token2) > sizeof((ptr->rulep)[ptr->nrules].fromaddr) - 1)
   {
      syslog(LOG_ERR,"Error: From address too long on line %d of %s",
			line,proxycfgfile);
      return(-1);
   }
   strcpy((ptr->rulep)[ptr->nrules].fromaddr, token2);

   /*===============================================================  
    * Next token is "to" in new format
    *               to_addr in old format
    *===============================================================*/  
   if( (token2 = strtok(NULL,tokensep)) == NULL) /* get to  */
   {
      syslog(LOG_ERR,"Error: permit statement on line %d of %s is missing "
		     "parameters.",
 		     line,proxycfgfile);  
      return(-1);
   }

   /*===============================================================  
    * If newsyntax, process the additional parameters
    *===============================================================*/  
   if (newsyntax) {

      if( strcmp(token2,"to") != 0)
      {
         syslog(LOG_ERR,"Error: permit/deny statement on line %d of %s is "
			"missing the \"to\" keyword",
			line,proxycfgfile);  
         return(-1);
      }
      /*===============================================================  
       * Next token is either to local address or dest address
       *===============================================================*/  
      if( (token2 = strtok(NULL,tokensep)) == NULL) /* get local addr  */
      { 
         syslog(LOG_ERR,"Error: permit statement on line %d of %s is missing "
		     "parameters.",
 		     line,proxycfgfile);  
         return(-1);
      }

      /*===============================================================  
       * Process local address and redirect if PERMIT
       *===============================================================*/  
      if (op == PERMIT) {
         char *rrptr;

         /*===============================================================  
          * Store to local address
          *===============================================================*/  
         if (strlen(token2) > sizeof((ptr->rulep)[ptr->nrules].toaddr) - 1)
         {
            syslog(LOG_ERR,"Error: To address too long on line %d of %s",
			line,proxycfgfile);
            return(-1);
         }
         strcpy((ptr->rulep)[ptr->nrules].toaddr, token2);

         /*===============================================================  
          * Next token is local port
          *===============================================================*/  
         if( (token2 = strtok(NULL,tokensep)) == NULL) /* get local port  */
         { 
            syslog(LOG_ERR,"Error: permit statement on line %d of %s is missing "
		     "parameters.",
 		     line,proxycfgfile);  
            return(-1);
         }

         /*===============================================================  
          * Store to local port
          *===============================================================*/  
         if (isalldigits(token2))
            (ptr->rulep)[ptr->nrules].toport = atoi(token2);
         else
         {  
            if((sp = getservbyname(token2,"tcp")) != (struct servent *) 0)
                  (ptr->rulep)[ptr->nrules].toport = ntohs(sp->s_port);
            else 
            {
               syslog(LOG_NOTICE,"Error: unknown service '%s' can not "
			"decode as port, on line %d of %s",
			token2,line,proxycfgfile);  
               return(-1);
            }
         }

         /*===============================================================  
          * Next token is "redirect"
          *===============================================================*/  
         if( (token2 = strtok(NULL,tokensep)) == NULL) /* get redirect  */
         { 
            syslog(LOG_ERR,"Error: permit statement on line %d of %s is missing "
		     "parameters.",
 		     line,proxycfgfile);  
            return(-1);
         }

         if( strcmp(token2,"redirect") != 0)
         {
            syslog(LOG_ERR,"Error: permit/deny statement on line %d of %s is "
			"missing the \"redirect\" keyword",
			line,proxycfgfile);  
            return(-1);
         }

         /*===============================================================  
          * See if round robin is requested.  If it is, we save the string
          * between the '(' and ')' delimeters for processing later.
          * If round robin not requested, the dest ip address is processed
          * now.
          *===============================================================*/  
         rrptr = strchr(token2 + strlen(token2) + 1, '(');
         if (rrptr != NULL) 
	 {
            char *cp;
            if (strchr(rrptr, ')') == NULL) {
               syslog(LOG_ERR, "Error: Unbalanced '(' and ')' "
				"after 'redirect' on line %d of %s",
				line, proxycfgfile);
               return(-1);
            }
            if ((cp = strtok(rrptr, "()")) == NULL) 
	    {
               syslog(LOG_ERR, "Error: Syntax error after 'redirect' "
				"on line %d of %s",
				line, proxycfgfile);
               return(-1);
            }
            strcpy(rrstring, cp);
            (ptr->rulep)[ptr->nrules].roundrobin = 1;
         }

         else {
            /*===============================================================  
             * Next token is dest ip address
             *===============================================================*/
            if( (token2 = strtok(NULL,tokensep)) == NULL) /* get dest addr  */
            { 
               syslog(LOG_ERR,"Error: permit statement on line %d of %s is "
		     "missing parameters.",
 		     line,proxycfgfile);  
               return(-1);
            }
         }
      }
   }

   if ((ptr->rulep)[ptr->nrules].roundrobin == 0) {
      /*===============================================================  
       * Store dest ip address if not processing round robin
       *===============================================================*/  
      if (strlen(token2) > sizeof((ptr->rulep)[ptr->nrules].destaddr) - 1)
      {
         syslog(LOG_ERR,"Error: Dest address too long on line %d of %s",
			line,proxycfgfile);
         return(-1);
      }
      strcpy((ptr->rulep)[ptr->nrules].destaddr, token2);
   }

   /*===============================================================  
    * Next token is dest port
    *===============================================================*/  
   if( (token2 = strtok(NULL,tokensep)) == NULL) /* get dest port*/
   { 
      syslog(LOG_ERR,"Error: permit statement on line %d of %s is missing "
			"parameters.",
			line,proxycfgfile);  
      return(-1);
   }

   /*===============================================================  
    * Store dest port
    *===============================================================*/  
   if (isalldigits(token2))
      (ptr->rulep)[ptr->nrules].destport = atoi(token2);
   else
   {  
      if((sp = getservbyname(token2,"tcp")) != (struct servent *) 0)
            (ptr->rulep)[ptr->nrules].destport = ntohs(sp->s_port);
      else 
      {
         syslog(LOG_NOTICE,"Error: unknown service '%s' can not "
			"decode as port, on line %d of %s",
			token2,line,proxycfgfile);  
         return(-1);
      }
   }

   /*===============================================================  
    * Look for optional keywords
    *===============================================================*/  
   while ( (token2 = strtok(NULL,tokensep)) != NULL) {
      if (*token2 == '#')
         continue;
      /*===============================================================  
       * Look for optional USING keyword
       *===============================================================*/  
      if( strcasecmp(token2,"USING") == 0) {
         if ((token2 = strtok(NULL, tokensep)) == NULL) {
            syslog(LOG_ERR, "Error: IP address must follow \"using\" keyword "
			"on line %d of %s file.",
			line,proxycfgfile);
            return(-1);
         }
         /*===============================================================  
          * Store to local address2
          *===============================================================*/  
         if (strlen(token2) > sizeof((ptr->rulep)[ptr->nrules].localaddr2) - 1)
         {
            syslog(LOG_ERR,"Error: Local address too long on line %d of %s",
			line,proxycfgfile);
            return(-1);
         }
         strcpy((ptr->rulep)[ptr->nrules].localaddr2, token2);
         continue;
      }
      /*==================================================================
       * if ignorerst then get the sleep value
       * Note that the default sleep value only works if IGNORERST is the
       * last keyword on the line.
       *==================================================================*/
      if(strcasecmp(token2,"ignorerst") == 0) {
         if ((token2 = strtok(NULL,tokensep)) == NULL) {
            (ptr->rulep)[ptr->nrules].ign_rst = 2; /* defalut sleep 2 seconds */
         }
         else if (!isalldigits(token2) ||
		((ptr->rulep)[ptr->nrules].ign_rst = atoi(token2)) <= 0) {
            syslog(LOG_ERR,"Error: Invalid value (%s) for IGNORERST "
                                "on line %d of %s",
                                token2, line, proxycfgfile);
            return(-1);
         }
         continue;
      }

      /*===============================================================  
       * Look for optional USEREXIT keyword
       *===============================================================*/  
      if( strcasecmp(token2,"USEREXIT") == 0) {
         if ((token2 = strtok(NULL, tokensep)) == NULL) {
            syslog(LOG_ERR, "Error: Missing value following \"userexit\" "
			"keyword on line %d of %s file.");
            return(-1);
         }
         /*===============================================================  
          * Store to exitname
          *===============================================================*/ 
         if (strlen(token2) > sizeof((ptr->rulep)[ptr->nrules].exitname) - 1) {
            syslog(LOG_ERR,"Error: exitname too long on line %d of %s",
			line,proxycfgfile);
            return(-1);
         }
         strcpy((ptr->rulep)[ptr->nrules].exitname, token2);
         continue;
      }

      /*===============================================================  
       * Look for optional OOBA keyword
       *===============================================================*/  
      if( strcasecmp(token2,"OOBA") == 0) {

         /*===============================================================  
          * ooba username
          *===============================================================*/  
         if( (token2 = strtok(NULL,tokensep)) == NULL) {
            syslog(LOG_NOTICE,"Error, expected user name after OOBA"
			" on line %d of %s", line,proxycfgfile);
            return(-1);
         }

         /*===============================================================  
          * Store ooba username
          *===============================================================*/  
         if (strlen(token2) > UNMSIZ - 1) 
	 {
            syslog(LOG_ERR, "Error: OOBA username %.16s longer than %d "
			"characters on line %d of %s",
			token2, UNMSIZ, line, proxycfgfile);
            return(-1);
         }
         strcpy(ptr->rulep[ptr->nrules].oobauser, token2);

         /*===============================================================  
          * ooba port
          *===============================================================*/  
         if( (token2 = strtok(NULL,tokensep)) == NULL) {
            syslog(LOG_NOTICE,"Error, expected port number after 'OOBA"
			" username' on line %d of %s",
			line,proxycfgfile);
            return(-1);
         }

         if(isalldigits(token2)) {
            portid = atoi(token2);
            if( portid > 65536) {
               syslog(LOG_NOTICE,"Error: OOBA portnumber '%d' > 65536."
			" See line %d of %s",
			portid, line,proxycfgfile);
               return(-1);
            }
         }
         else {
            syslog(LOG_NOTICE,"Error: OOBA portnumber '%s' is not all "
			"digits. See line %d of %s.",
			token2, line, proxycfgfile);
            return(-1);
         }

         /*===============================================================  
          * Store ooba port
          *===============================================================*/  
         ptr->rulep[ptr->nrules].oobaport = portid;
         continue;
      }
      /*===============================================================  
       * Invalid parameter
       *===============================================================*/  
      syslog(LOG_NOTICE,"Error, Invalid parameter \"%s\" found "
			"on line %d of %s file.",
			token2,line,proxycfgfile);
      return(-1);
   }

   /*===============================================================  
    * Use defaults for old syntax lines
    *===============================================================*/  
   if (!newsyntax) {
      strcpy((ptr->rulep)[ptr->nrules].toaddr, "*");
      (ptr->rulep)[ptr->nrules].toport = (ptr->rulep)[ptr->nrules].destport;
   }

   /*===============================================================  
    * Use destaddr and destport as toaddr and toport if DENY
    *===============================================================*/  
   if (op == DENY) {
      strcpy((ptr->rulep)[ptr->nrules].toaddr,
			(ptr->rulep)[ptr->nrules].destaddr);
      (ptr->rulep)[ptr->nrules].toport = (ptr->rulep)[ptr->nrules].destport;
   }

   /*===============================================================  
    * Now process round robin dest ip addresses if specified
    *===============================================================*/  
   if (ptr->rulep[ptr->nrules].roundrobin) {
      int fstrule = ptr->nrules;
      token2 = strtok(rrstring, " ;:\t,");
      /*===============================================================  
       * Store 1st round robin dest ip address
       *===============================================================*/  
      if (strlen(token2) > sizeof((ptr->rulep)[ptr->nrules].destaddr) - 1)
      {
         syslog(LOG_ERR,"Error: Dest address too long on line %d of %s",
			line,proxycfgfile);
         return(-1);
      }
      strcpy((ptr->rulep)[ptr->nrules].destaddr, token2);

      while ((token2 = strtok(NULL, " \t,")) != NULL) {
         /*===============================================================
          * malloc more space for each subsequent round robin address
          *===============================================================*/
         *sizep += sizeof(RULE);
         if ((*ptrp = realloc(*ptrp, *sizep)) == NULL) {
            syslog(LOG_ERR,"Error: Could not realloc.  %s", sys_errlist[errno]);
            return(-1);
         }
         ptr = *ptrp;		/* Update the local pointer	*/
         (ptr->nrules)++;
         /*===============================================================  
          * Clone the rule
          *===============================================================*/  
         memcpy(&(ptr->rulep)[ptr->nrules], &(ptr->rulep)[fstrule],
							sizeof(RULE));
         (ptr->rulep)[ptr->nrules].roundrobin = 0;
         /*===============================================================  
          * Store a round robin dest ip address
          *===============================================================*/  
         if (strlen(token2) > sizeof((ptr->rulep)[ptr->nrules].destaddr) - 1) {
            syslog(LOG_ERR,"Error: Dest address too long on line %d of %s",
			line,proxycfgfile);
            return(-1);
         }
         strcpy((ptr->rulep)[ptr->nrules].destaddr, token2);
         /*===============================================================  
          * Update round robin address count
          *===============================================================*/  
         (ptr->rulep)[fstrule].roundrobin++;
      }
   }

   (ptr->nrules)++;
   return(0);
}    
