/*============================================================================
 * filename:		../webgate/webgwconf.c	../aproxy/apgetprox.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *										
 * written by:		Jim Livermore						
 *
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000. All rights reserved.
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.
 * All rights reserved.
 *		
 * Date written:	February  28, 1995
 * Last modified:	February   8, 2000
 *	                         
 * function:
 *                       
 *	The proxyconf program reads the proxy configuration file.
 *  
 *
 * file format:             
 *
 * Manditory commands
 *  
 *	gproxpath 	= pathname	-      
 *
 * Optional Commands
 *     
 *      httplog         = filename      - HTTP log file name 
 *					  default is webgate.log
 * 	timeout 	= nnn 		- number of seconds
 *	trace		= yes|no	- turn tracing on or off, default is no
 *
 * Access Rules 
 *           
 * There are two forms of permit statements.  The new permit statement was
 * added to support multiple WWW Servers
 * 
 * Old permit: 
 *
 * 	permit src_host src_port to dest_host dest_port [ooba name ooba_port] 
 *
 * New permit:
 *
 *	permit from src_host to local_host local_port redirect dest_host \
 *                                  dest_port [ooba name ooba_port] [VLOG]
 *
 *	deny   src_host src_port [shell cmd]
 *
 *============================================================================*/
#include	<sys/types.h>
#include	<stdio.h>
#include	<syslog.h>
#include	<string.h>
#include	<sys/time.h>
#include	<netdb.h>               /* defines servent structure      */
#include	<netinet/in.h>
#include	"firewall.h" 
#include	"gwuser.h"
#ifdef APROXY
#include	"aproxy.h"
#else
#include	"webgate.h"  
#endif
               
/*============================================================================
 * function prototypes
 *===========================================================================*/  
int ooba(char *destIP,  char * username, int portid,
         struct WGCONF *wproxptr);
/*=============================================================================
 *
 * inputs: 	local host port id 
 *		ptr to WGCONF struct
 * 
 * outputs:	return  1 if access is permitted
 *			side effect: wproxptr->ns->rule_index is set here if
 *					a permit rule is selected
 *		return -1 if access is denied
 *===========================================================================*/
int getprox(int srcport, WGCONF *wproxptr)
{   
   int		i;
   char		username[UNMSIZ+1];
   char 	*cp; 
   char		*cp2;  
   char		*cp3;   
   char		*token1;
   char		*token2; 
   char		*token3; 
   char		tokensep[] = "\t,:;= \n";
   int		hostfound  = 0;  
   int		lhostfound = 0;
   int		line	   = 0; 
   int		proxpathfound = 0;
   int		rtncd;
   int		sport;                  /* source port from config file */
   int		dport;			/* dest port fromconfig file    */
   struct servent	*sp; 		/* ptr to service entry 	*/  

   wproxptr->ns->rule_index = -1;
   for (i = 0; strlen(wproxptr->rulep[i].fromaddr) > 0; i++) 
   {
#define ipstr	wproxptr->rulep[i].fromaddr
#define tostr	wproxptr->rulep[i].toaddr
      /*==============================================================
       * If src addr, local addr and local port all match this rule,
       * use it.  If rule is a deny rule, wildcarded local addresses
       * can be matched too.
       *=============================================================*/
      if( ((strchr(ipstr,'*') &&
		strncmp(ipstr, wproxptr->ns->srcIP,strlen(ipstr) - 1) == 0) ||
	   strcmp(ipstr,wproxptr->ns->srcIP) == 0) &&
          (strcmp(tostr,wproxptr->ns->localIP) == 0 ||
	   (wproxptr->rulep[i].op == DENY && strchr(tostr,'*') &&
	    strncmp(tostr,wproxptr->ns->localIP,strlen(tostr)-1) == 0)) &&
          srcport == wproxptr->rulep[i].toport) 
      {

         /*===========================================================
          * Found match, if it is a deny rule, return -1
          *==========================================================*/
         if (wproxptr->rulep[i].op == DENY) 
         {
            syslog(LOG_ALERT,"Security Alert: request for service '%d' "
			" from host '%s' explicitly denied.",
			srcport, wproxptr->ns->srcIP);
            return(-1);
         }

         /*===========================================================
          * Found permit match, return index of selected permit rule
          *==========================================================*/
         wproxptr->ns->rule_index = i;

         strcpy(username, wproxptr->rulep[i].oobauser);
         /*===========================================================
          * Handle OOBA if specified in rule
          *==========================================================*/
         if (strlen(wproxptr->rulep[i].oobauser) > 0) 
	 {
            if (ooba(wproxptr->ns->srcIP, username,
			wproxptr->rulep[i].oobaport, wproxptr) != 0)
               return(-1);
         }
         return(1);
      }
   }

   /*=======================================================================
    * Not rule matched this connection, deny by default.
    *======================================================================*/
   syslog(LOG_ALERT,"Security Alert: service '%d' from host '%s' denied. "
		"Local IP '%s' Service not permitted for host.",
		srcport, wproxptr->ns->srcIP, wproxptr->ns->localIP);

   return(0);   /* no match found */    
} 
