/*============================================================================
 * filename:		../aproxy/aproxy.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *										
 * written by:		Jim Livermore
 *
 * (C) COPYRIGHT R. James Livermore 2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 2000.  All rights
 * reserved.    
 *
 *		
 * Date written:        May        6, 1999 - adapted from webgate
 *			October   12, 1999 - added Linux support 
 * Last modified:	February   8, 2000 - added LPL
 *			March     28, 2000 - changed connect and disconnect
 *					     messages to LOG_NOTICE
 *	                         
 * function:                       
 *
 *============================================================================*/
#include	<fcntl.h>
#include	<stdio.h>
#include	<string.h>
#include	<syslog.h>
#include	<sys/errno.h> 
#include	<sys/ioctl.h>  
#include	<sys/poll.h>  
#ifdef __sun
#include	<sys/sockio.h>		/* socket I/O controls 		*/
#include	<sys/stat.h>
#endif
#include	<sys/types.h>
#include	<sys/time.h>  

#ifndef __hppa
#include 	<sys/select.h>		/* to support fd_set 		  */ 
#endif 
#ifdef linux
#include	<sys/stat.h>
#endif
#include	<sys/signal.h>
#include	<sys/socket.h> 
#include	<time.h>   
#include	<unistd.h>
#include	<netdb.h>               /* defines servent structure      */
#include	<netinet/in.h> 
#include	<arpa/inet.h>
#include	"firewall.h"   
#include	"aproxy.h"
#include	"user_exits.h"
/*===========================================================================
 * external variables for user exits
 *==========================================================================*/
extern	U_EXIT	u_exits[];
/*===========================================================================
 * external variables
 *==========================================================================*/
static	int		extrfd; 
static	int		portid = -1; 
static	time_t		starttime;
static	time_t		endtime; 
	char		urgsigmsg[132] = "\0";
static  long		clktck = 0;
static	struct timeval	timo;
	struct timeval	*tp = (struct timeval *) 0;
extern	int	errno;

#ifndef linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif

extern int      daylight;
extern long int	timezone; 
extern char* tzname[]; 

#define	NON_CONNECT_IDLE	5 * 60;	/* Don't use a rule for 5 minutes
						if connection failed	*/
 
/*============================================================================
 * function prototypes
 *===========================================================================*/
int acceptrule(int srcport, WGCONF *wproxptr);
int  connectserver(char *serv,int portnum, int priv, char *localIP); 
void debugbind(unsigned short debugport); 
int getprox(int srcport, WGCONF *wproxptr); 
void hexprint(int fd, int bytes, char * cptr);  
int  peername(int fd, char *lname,char *sname,int z);   
int trace_init(struct tracehdr *th, struct WGCONF *wgconfptr, char *tbuf);
int trace_it(struct tracehdr *th, char cs, int br, char *obuf, char *tbuf);
static	void	trap_sigurg();
static	int	shouldexit = 0;
static	WGCONF	*mywgconfp;

void sigusr1(int sig) 
{
   int i;

   shouldexit = 1;

    /*=======================================================================
     * Close all the listening sockets now so a new aproxy can be started
     * right away.  They are no longer needed here anyway since this process
     * will exit after this connection is terminated.
     *======================================================================*/
   for (i = 0; i < mywgconfp->nfds; i++)
      close(mywgconfp->fdp[i].fd);
   return;
}

/*============================================================================
 * alarm handler - just return
 *===========================================================================*/
void sigalrm(int sig) {
    close(0);
    exit(0);
}

/*============================================================================
 * main function
 *===========================================================================*/
child_main(WGCONF *wgconfp, int slot)
{  
   int		PeerRtn;
   int		rc;
   int		fd;
   int		maxfd;
   int		i, j;
   int		argc;
   int		error_no;
   char		*arg[10];
   char		*servnam; 		/* service name		*/ 
   struct servent	*sp;
   long		tempport;  
 
   unsigned short	debugport = 8080; 	/* default debug port	*/ 
   /* may change later */
   int		IPsecure;			/* 0 if src IP secure	*/ 
   struct 	sockaddr_in	la;
   int		siz;    
   struct sockaddr	clientaddr;
   int			clilen;
   struct sigaction	sigact;
   struct sigaction	osigact;

   tzset();

   DEBUG(3, wgconfp, (DBLOG, "Child starting up"));

   /*=======================================================================
    * Save a copy of the wgconfp pointer to static area so the signal
    * handler for SIGUSR1 can get to it.
    *======================================================================*/
   mywgconfp = wgconfp;

    /*=================================================================
     * Change the root directory to directory specified in aproxy.conf
     * file.  Must be root user to do this.
     *===============================================================*/
   chdir("/");   		/* change current directory  */
   if( ( chdir(wgconfp->gproxpath) || chroot(wgconfp->gproxpath) ) ) 
   {
      syslog(LOG_ERR,"Error: cannot chroot to %s: %m",
				wgconfp->gproxpath);
      exit(1);
   }
   chdir("/");


   /*=======================================================================
    * Infinite loop to wait for and then handle connections
    *======================================================================*/
   for ( ; ; ) 
   {
      int flags;

      maxfd = 0;
      IPsecure = -1;
   
      signal(SIGUSR1, SIG_DFL);
      /*======================================================================
       * Block until a connection comes in
       *=====================================================================*/

      clilen = sizeof(clientaddr);
      memset((char *)&clientaddr, 0, clilen);

      DEBUG(3, mywgconfp, (DBLOG, "Trying to acquire lock"));

      /*======================================================================
       * Lock before poll()
       *=====================================================================*/
      if (my_lock_wait(mywgconfp->lockfd) < 0)
      {
         exit(1);
      }

      if (poll(mywgconfp->fdp, mywgconfp->nfds, -1) < 0) {
         syslog(LOG_ERR,"Error: poll() failed.  %s.  Connection cannot "
			"complete, process exiting", sys_errlist[errno]);
         exit(1);
      }

      for (i = 0; i < mywgconfp->nfds; i++)
      {
         if (mywgconfp->fdp[i].revents & POLLIN)
            break;
      }
      if (i >= mywgconfp->nfds) {
         syslog(LOG_ERR,"Error: Unexpected error, poll() unblocked but no "
		"file descriptor ready.  Connection cannot complete");
         /*===================================================================
          * Find any returned errors and report them
          *==================================================================*/
         for (i = 0; i < mywgconfp->nfds; i++)
            if ((mywgconfp->fdp[i].revents & ~POLLIN) != 0) 
            {
               syslog(LOG_ERR, "Error: fd %d returned revent 0x%08x",
				mywgconfp->fdp[i].fd,mywgconfp->fdp[i].revents);
            }
         /*===================================================================
          * Unlock before continuing
          *==================================================================*/
         if (my_lock_release(mywgconfp->lockfd) < 0) {
            exit(1);
         }
         continue;	/* Can't continue with this connection, next pls */
      }

      fd = accept(mywgconfp->fdp[i].fd, (struct sockaddr*)&clientaddr,
							(socklen_t *) &clilen);
      error_no = errno;

      flags = fcntl(fd, F_GETFL, 0);
      fcntl(fd, F_SETFL, flags & ~O_NONBLOCK);

      /*======================================================================
       * Reset all the return-events.
       *=====================================================================*/
      for (j = 0; j < mywgconfp->nfds; j++)
         mywgconfp->fdp[j].revents = 0;

      sigact.sa_handler = sigusr1;
      sigemptyset(&sigact.sa_mask);
#if !defined(__hppa)
      sigact.sa_flags = SA_RESTART;
#else
      sigact.sa_flags = 0;
#endif
      rc = sigaction(SIGUSR1, &sigact, &osigact);

      /*======================================================================
       * Set child slot to reflect this new activity
       * If maxuse is reached for this child, set shouldexit so it exits after
       * this connection.
       *=====================================================================*/
      if (++(mywgconfp->child[slot].nconn) >= mywgconfp->maxuse)
         shouldexit = 1; 
      mywgconfp->child[slot].status = CHILD_BUSY;

      /*======================================================================
       * Unlock after accept()
       *=====================================================================*/
      if (my_lock_release(mywgconfp->lockfd) < 0) {
         exit(1);
      }
      if (rc < 0) {
         syslog(LOG_ERR,"Error: sigaction failed");
         mywgconfp->child[slot].status = CHILD_IDLE;
         if (fd >= 0)
            close(fd);
         continue;
      }

      /*======================================================================
       * Ignore some errors returned from accept()
       *=====================================================================*/
      if (fd < 0) {
         if (error_no == ECONNABORTED || error_no == EWOULDBLOCK ||
		error_no == EPROTO || error_no == EINTR) {
            DEBUG(3, mywgconfp, (DBLOG, "Ignored errno %d from accept()",
					error_no));
            mywgconfp->child[slot].status = CHILD_IDLE;
            continue;
         }
         syslog(LOG_ERR,"Error: accept() failed.  %s.  Connection cannot "
			"complete, process exiting",sys_errlist[error_no]);
         exit(1);
      }

      DEBUG(2, mywgconfp, (DBLOG, "accept() successful, status set to %d, "
				"nconn = %d.", mywgconfp->child[slot].status,
				mywgconfp->child[slot].nconn));
      if (shouldexit)
         DEBUG(2, mywgconfp, (DBLOG, "Child will exit after this connection."));

      /*======================================================================
       * Make the incoming socket fd 0
       *=====================================================================*/
      if (fd != 0) {
         close(0);
         dup(fd);
         close(fd);
      } 
      /* Initialize clock ticks per second */
      if( ( clktck = sysconf(_SC_CLK_TCK)) < 0)
      {
         syslog(LOG_ERR,"Error: sysconf(_SC_CLK_TIC) function failed.");
         clktck = 0;		/* set to zero to stop cpu timing.	*/
      }
      time(&starttime);  
      /*======================================================================
       * Added trap for urgent signal
       *=====================================================================*/
      signal(SIGURG,trap_sigurg); 
      /* set out of band signalling on fd 0	*/ 
#ifndef __hppa   
      if( fcntl(0,F_SETOWN,getpid() ) < 0)
      {
         syslog(LOG_ERR,"Error: fcntl(0,FSETOWN) failed %m");
      }
#endif   
      /*=========================================================
       * get remote hostname and IP address in dot decimal format 
       *========================================================*/
      PeerRtn = PeerData(0,mywgconfp->ns->srcHost,mywgconfp->ns->srcIP,
                                sizeof(mywgconfp->ns->srcHost),DNSOFF);
      if(PeerRtn)
      {
         syslog(LOG_ERR,"Can not get remote host: %m.  Connection cannot "
			"complete");
         close(0);
         mywgconfp->child[slot].status = CHILD_IDLE;
         continue;	/* Can't continue with this connection, next pls */
      }  
      /*====================================================================
       * Check for IP address spoofing. If source host IP address is secure and
       * the local IP address is not then packet came to the wrong port! 
       * Get local IP address of socket.
       *====================================================================*/
      siz = sizeof(la);
      if(getsockname(0,(struct sockaddr *)&la,(socklen_t *)&siz) < 0) 
      {
         syslog(LOG_ERR,"Error: getsockname failed: %s.  Connection cannot "
			"complete",sys_errlist[errno]);
         close(0);
         mywgconfp->child[slot].status = CHILD_IDLE;
         continue;	/* Can't continue with this connection, next pls */
      }
      /*================================================================== 
       * convert local IP address to ASCII string in dot decimal format 
       *=================================================================*/  
      portid = ntohs(la.sin_port);
      strncpy(mywgconfp->ns->localIP,inet_ntoa(la.sin_addr),IPSIZ-1);
      mywgconfp->ns->localIP[IPSIZ-1] = '\0'; 
      /* ipsecure returns 0 if secure else -1	*/
      IPsecure = isIPlisted(mywgconfp->ns->srcIP, mywgconfp->snp);
      /*=====================================================================
       * Check if host with a secure source IP address is connected via a secure
       * port. If not issue a Security Alert and drop the connection.   
       *====================================================================*/ 
      if( IPsecure && !isIPlisted(mywgconfp->ns->localIP, mywgconfp->spp) ) 
      {    
         syslog(LOG_ALERT,"Security Alert: attempt to use unsecured port %s "
		"using secure source IP address %s aka %s.",
		mywgconfp->ns->localIP,mywgconfp->ns->srcIP,
		mywgconfp->ns->srcHost);  
         close(0);
         mywgconfp->child[slot].status = CHILD_IDLE;
         continue;	/* Can't continue with this connection, next pls */
      }	
      /*=======================================================================
       * Initialize time out value.
       *=====================================================================*/
      timo.tv_usec = 0; 
      timo.tv_sec  = 60;		/* set default timeout to 60 sec */
      tp = &timo;
      /*=======================================================================
       * Compare the requested service with the rules in the aproxy.conf file.
       * If the rule is accepted getprox returns a 1.  
       *======================================================================*/
      if( getprox(portid, mywgconfp) < 1) 
      { 
         close(0);	
         mywgconfp->child[slot].status = CHILD_IDLE;
         continue;	/* Can't continue with this connection, next pls */
      } 
      timo.tv_sec = mywgconfp->timeout;
      mywgconfp->ns->localport = portid; 
      strcpy(mywgconfp->ns->destIP,
		mywgconfp->rulep[mywgconfp->ns->rule_index].destaddr);
      mywgconfp->ns->destport =
		mywgconfp->rulep[mywgconfp->ns->rule_index].destport;

      /*=======================================================================
       * Call user exit if necessary
       *======================================================================*/
      if (strlen(mywgconfp->rulep[mywgconfp->ns->rule_index].exitname) > 0) 
      {
         int i;
         int breakout;
         breakout = 0;
         for (i = 0; u_exits[i].match_str != NULL; i++) {
            if (strcmp(u_exits[i].match_str, 
		mywgconfp->rulep[mywgconfp->ns->rule_index].exitname) == 0) {
               if (u_exits[i].Auth != NULL) {
                  UE_PERMIT ue_permit;
                  UE_AUTH_ARGS ue_auth_args;

                  ue_auth_args.permitinfo = &ue_permit;
                  ue_auth_args.clientfd = 0;
                  ue_permit.srcIP = inet_addr(mywgconfp->ns->srcIP);
                  ue_permit.localIP = inet_addr(mywgconfp->ns->localIP);
                  ue_permit.localport = mywgconfp->ns->localport;
                  ue_permit.destIP = inet_addr(mywgconfp->ns->destIP);
                  ue_permit.destport = mywgconfp->ns->destport;
                  if ((u_exits[i].Auth)(&ue_auth_args) < 0) {
                     syslog(LOG_NOTICE, "User provided Auth exit rejected "
				"connection.");
                     close(0);
                     mywgconfp->child[slot].status = CHILD_IDLE;
                     breakout = 1;
                     break;
                  }
               }
               break;
            }
         }
         if (breakout)
             continue;
         if (u_exits[i].match_str == NULL) {
            syslog(LOG_ERR, "Error: Exit name \"%s\" not found, cannot "
			"continue.",
			mywgconfp->rulep[mywgconfp->ns->rule_index].exitname);
            close(0);
            mywgconfp->child[slot].status = CHILD_IDLE;
            continue;
         }
      }

      acceptrule(portid, mywgconfp);

      /*=======================================================================
       * Ignore SIGURG until after the next accept()
       *======================================================================*/
      signal(SIGURG, SIG_IGN);

      if (shouldexit) {
         DEBUG(3, mywgconfp, (DBLOG, "Child normal exit, nconn = %d.",
				mywgconfp->child[slot].nconn));
         exit(0);
      }

      DEBUG(3, mywgconfp, (DBLOG, "Setting status back to IDLE"));
      mywgconfp->child[slot].status = CHILD_IDLE;

   }
}

#define ORIG_RULE	wproxptr->rulep[wproxptr->ns->rule_index]
#define CURRENT_RULE	wproxptr->rulep[indx]
#define TMP_RULE	wproxptr->rulep[i]
/*===========================================================================
 * functione:	leastused
 * inputs:
 *		ptr to WGCONF struct
 * outputs:
 *		returns index of the rule to be used
 *==========================================================================*/
int leastused(WGCONF *wproxptr) 
{
   int i;
   time_t idx_earliest = wproxptr->ns->rule_index;
   if (ORIG_RULE.roundrobin > 1) {
      for (i = wproxptr->ns->rule_index + 1;
		i < ORIG_RULE.roundrobin+wproxptr->ns->rule_index; i++) {
         if (TMP_RULE.lastused.tv_sec <
			wproxptr->rulep[idx_earliest].lastused.tv_sec ||
	     (TMP_RULE.lastused.tv_sec ==
			wproxptr->rulep[idx_earliest].lastused.tv_sec &&
	      TMP_RULE.lastused.tv_usec <
			wproxptr->rulep[idx_earliest].lastused.tv_usec))
            idx_earliest = i;
      }
   }
   return(idx_earliest);
}

/*===========================================================================
 * function:	acceptrule 
 * inputs:
 * 		source port number
 *		ptr to WGCONF record
 *		number of argumnets
 *		argument list: 
 *		   to 	dest_host
 *		   port dest_port 
 * outputs:
 *         
 *		On successful conclusion close sockets and exit(0)  
 *		On error close sockets and exit(1)
 *
 * returns:	0 if request was accepted and handled successfully
 *		-1 if request denied or failed
 *==========================================================================*/ 
int acceptrule(int srcport, WGCONF *wproxptr)
{ 
   struct servent	*sp;
   fd_set		redy;
   fd_set		iced; 
   int 		duration;
   int		hours;
   int		min;
   int		sec;
   int		ibytes 	 = 0;		/* input bytes			*/
   int		obytes	 = 0;           /* output bytes			*/
   int		i;
   int		br;			/* bytes read			*/
   int		bw;			/* bytes written		*/
   int		x;  
   int		rc;			/* return code for select	*/
   int		count;
   int		indx;			/* Index of the rule to be used	*/
   int		exitidx = -1;		/* Index of the exits to be used*/
   int		srvind	 = -1;
   int		serfd;
   int		privport = 0; 
   float	cpu;
   char		*buf; 
   char		logbuf[1024];
   char         tracebuf[512];
   struct trace  	*tracebufadr;  
   struct tracehdr 	traceh;          
   UE_PERMIT	ue_permit;
#ifdef linux
   struct timeval tt;
#endif

   /*===================================================================
    * Initialize counters in the Non-shared area
    *====================================================================*/ 
   wproxptr->ns->totalreads = 0; 
   if(wproxptr->ns->localport == -1) 
   {   
      syslog(LOG_ERR,"Error: not sure what port to connect to"); 
      close(0);
      return(-1);
   }

   signal(SIGALRM, sigalrm);

   /*=====================================================================
    * If round robin requested, find the matched rule that has not been used
    * the longest.  If connection failed, we retry untill all the matched
    * round robin rules are exhausted.  Each time we failed, we
    * set the lastused member of the rule to NON_CONNECT_IDLE seconds +
    * current time so it won't be used again for NON_CONNECT_IDLE seconds.
    *=====================================================================*/
   count = ORIG_RULE.roundrobin;
   if (ORIG_RULE.roundrobin > 1) {
#ifndef linux
      int zero = 0;
#else
      struct timezone zero;
#endif
      char *usingIP;
      do {
         if (my_lock_wait(wproxptr->rrfd) < 0)
            exit(1);
         indx = leastused(wproxptr);
         gettimeofday(&CURRENT_RULE.lastused, &zero);/* Marked as just used*/
         if (my_lock_release(wproxptr->rrfd) < 0)
            exit(1);

         if (strlen(CURRENT_RULE.localaddr2) > 0)
            usingIP = CURRENT_RULE.localaddr2;
         else
            usingIP = NULL;
         
         /*===================================================================
          * Retry on RST every ign_rst seconds.  Try until the lesser of
          * timeout or 5 minutes is up.
          *==================================================================*/
         alarm((wproxptr->timeout > 5 * 60) ? 5 * 60 : wproxptr->timeout);
         while ( (serfd = connectserver(CURRENT_RULE.destaddr,
			CURRENT_RULE.destport,privport, usingIP) ) == -4 &&
			(errno == ECONNRESET || errno == ECONNREFUSED)) {
            alarm(0);
            if (CURRENT_RULE.ign_rst) {
               char tmp[10];
               fd_set rds;
               struct timeval ts;
               /*============================================================
                * Make sure the client side has not closed the connection
                *===========================================================*/
               FD_ZERO(&rds);
               FD_SET(0, &rds);
               ts.tv_sec = 0;
               ts.tv_usec = 0;
               if (select(1, &rds, (fd_set *)0, (fd_set *)0, &ts) > 0 &&
			recv(0, tmp, 1, MSG_PEEK) == 0) {
                  syslog(LOG_NOTICE, "Remote client %s disconnected "
				"before connection to server %s is "
				"established.",
				CURRENT_RULE.fromaddr, CURRENT_RULE.destaddr);
                  return(-1);
               }
               /*============================================================
                * Use select() as sleep() since we don't want to mix alarm()
                * with sleep().
                *===========================================================*/
               ts.tv_sec = CURRENT_RULE.ign_rst;
               ts.tv_usec = 0;
               select(0, NULL, NULL, NULL, &ts);
            }
            else
               break;
         }
         alarm(0);
         if (serfd < 0) {
            syslog(LOG_ERR,"Error: cannot connect to server %s port %d: %m.",
		         CURRENT_RULE.destaddr,CURRENT_RULE.destport); 
            if (my_lock_wait(wproxptr->rrfd) < 0)
               exit(1);
            gettimeofday(&CURRENT_RULE.lastused, &zero);
            CURRENT_RULE.lastused.tv_sec += NON_CONNECT_IDLE;
            if (my_lock_release(wproxptr->rrfd) < 0)
               exit(1);
         }
         else
            break;
      } while (--count > 0);
   }
   else {
      char *usingIP;
      indx = wproxptr->ns->rule_index;
      if (strlen(CURRENT_RULE.localaddr2) > 0)
         usingIP = CURRENT_RULE.localaddr2;
      else
         usingIP = NULL;
      /*===================================================================
       * Retry on RST every ign_rst seconds.  Try until the lesser of
       * timeout or 5 minutes is up.
       *==================================================================*/
      alarm((wproxptr->timeout > 5 * 60) ? 5 * 60 : wproxptr->timeout);
      while ( (serfd = connectserver(ORIG_RULE.destaddr,
			ORIG_RULE.destport,privport, usingIP) ) == -4 &&
			(errno == ECONNRESET || errno == ECONNREFUSED)) {
         alarm(0);
         if (ORIG_RULE.ign_rst) {
            char tmp[10];
            fd_set rds;
            struct timeval ts;
            /*===============================================================
             * Make sure the client side has not closed the connection
             *==============================================================*/
            FD_ZERO(&rds);
            FD_SET(0, &rds);
            ts.tv_sec = 0;
            ts.tv_usec = 0;
            if (select(1, &rds, (fd_set *)0, (fd_set *)0, &ts) > 0 &&
		recv(0, tmp, 1, MSG_PEEK) == 0) {
               syslog(LOG_NOTICE, "Remote client %s disconnected "
				"before connection to server %s is "
				"established.",
				CURRENT_RULE.fromaddr, CURRENT_RULE.destaddr);
               return(-1);
            }
            /*============================================================
             * Use select() as sleep() since we don't want to mix alarm()
             * with sleep().
             *===========================================================*/
            ts.tv_sec = ORIG_RULE.ign_rst;
            ts.tv_usec = 0;
            select(0, NULL, NULL, NULL, &ts);
         }
         else
            break;
      }
      alarm(0);
      if (serfd < 0)
         syslog(LOG_ERR,"Error: cannot connect to server %s port %d: %m.",
		         ORIG_RULE.destaddr,ORIG_RULE.destport); 
   }

   /*=====================================================================
    * If every address in the round robin has been tried and still no
    * success, return failure.
    *=====================================================================*/
   if (serfd < 0) {
      close(0);
      return(-1);
   }

   extrfd = serfd;
   /*=====================================================================
    * Log successful connection.
    *=====================================================================*/
   syslog(LOG_NOTICE,"connect src_host = %s (%s) localIP = %s local_port = %d "
		"dest_host = %s dest_port = %d",
		wproxptr->ns->srcHost, wproxptr->ns->srcIP,
		wproxptr->ns->localIP, wproxptr->ns->localport,
		CURRENT_RULE.destaddr,CURRENT_RULE.destport); 
   /*=====================================================================
    * If the trace flag is on then initialize the trace buffers
    *=====================================================================*/
   if( wproxptr->traceflg )
   {
      if( trace_init(&traceh,wproxptr,tracebuf) < 0) 
      {  
         syslog(LOG_ERR,"Error: trace_init returned error, function is not "
			"active. strerror(errno)");
         wproxptr->traceflg = 0; 
      }
   }

   /*=========================================================================
    * Determine the index for the user exits
    *========================================================================*/
   if (strlen(ORIG_RULE.exitname) > 0) 
   {
      int i;
      for (i = 0; u_exits[i].match_str != NULL; i++) {
         if (strcmp(u_exits[i].match_str, ORIG_RULE.exitname) == 0) {
            exitidx = i;
            break;
         }
      }
      if (exitidx < 0) {
         syslog(LOG_ERR, "Error: Exit name \"%s\" not found, cannot continue.",
			ORIG_RULE.exitname);
         return(-1);
      }
      ue_permit.srcIP = inet_addr(wproxptr->ns->srcIP);
      ue_permit.localIP = inet_addr(wproxptr->ns->localIP);
      ue_permit.localport = wproxptr->ns->localport;
      ue_permit.destIP = inet_addr(wproxptr->ns->destIP);
      ue_permit.destport = wproxptr->ns->destport;
   }

#ifdef linux
   memcpy(&tt, tp, sizeof(tt));
#endif
   FD_ZERO(&iced);   		/* Clear all bits in fdset iced	*/
   FD_SET(0,&iced);             /* Turn on bit for fd 0 	*/
   FD_SET(serfd,&iced);    	/* turn on bit for fd serfd	*/
   /*=========================================================================
    * Near Infinite Loop
    *========================================================================*/
   while(1) 
   {
      char *outp;
      int outsz;

#ifdef linux
      memcpy(tp, &tt,  sizeof(*tp));
#endif
      /*================================================================= 
       * Intialize fd_set pointed to by redy and select the next request.
       *================================================================*/
      (void)bcopy(&iced,&redy,sizeof(fd_set));
#ifdef __hppa
      while((rc = select(serfd + 2,(int *)&redy,(int *)0,(int *)0,tp)) < 0 &&
			errno == EINTR);
#else
      while ((rc = select(serfd + 2,&redy,(fd_set *)0,(fd_set *)0,tp)) < 0 &&
			errno == EINTR);
#endif
      if( urgsigmsg[0] != '\0')
      {
         if( wproxptr->traceflg)
            trace_it(&traceh, 'U',strlen(urgsigmsg),urgsigmsg,tracebuf);
         urgsigmsg[0] = '\0';   
         obytes++; 
         continue;
      }
      if( rc == 0) 
      {
         syslog(LOG_NOTICE,"connection timeout: %d sec",timo.tv_sec); 
         if( wproxptr->traceflg)
            trace_it(&traceh, 'T',0," timeout.",tracebuf);
	 break;
      }
      if( rc < 0) 
      {
         syslog(LOG_ERR,"Error: select returned an error: %d sec.  %m",rc); 
         if( wproxptr->traceflg)
            trace_it(&traceh, 'E',0,"  select error.",tracebuf);
	 break;
      }
      /*==================================================================
       * If request is from src host read src data and write to server.
       * Find "User-Agent:" and insert client IP address before ")".
       *================================================================*/
      if(FD_ISSET(0,&redy)) 
      {
         br = read(0,wproxptr->bufptr,wproxptr->bufsize);
         if( br < 1)
         {   
	    /*=================================================================
             * If EOF do a half close and pass it on to server and stop    
             * listening to client.
             *===============================================================*/
            if( br == 0 ) 
            {  
	       if( wproxptr->traceflg)
                  trace_it(&traceh, 'C', br, " EOF",tracebuf);  
               if( shutdown(serfd,1) < 0) /* send EOF to server */ 
               {
                  syslog(LOG_ERR,"Error: half close to server failed errno %d:"
                                 " %s",
			 errno, sys_errlist[errno]);
                  break;   
               } 
               FD_CLR(0,&iced); 	/* Stop listening to client	*/ 
               goto waitfornext; 
            }
            /*=================================================================
             * If br < 0 we have a read error.  Ignore errno = ECONNRESET as it
	     * is caused when browser hits stop button. Log the error and quit. 
             *================================================================*/ 
	    if( errno != ECONNRESET)
	       syslog(LOG_ERR,"Error: while reading from client host(%s). "
			   "errno %d: %s",
			   wproxptr->ns->srcIP,errno, sys_errlist[errno]);
            if( wproxptr->traceflg)
               trace_it(&traceh, 'C', br, " READ ERROR",tracebuf);  

            break;     
         } 
         wproxptr->bufptr[br] = '\0'; 
         wproxptr->ns->totalreads++; 
         
         if( wproxptr->traceflg)
            trace_it(&traceh, 'C', br, wproxptr->bufptr,tracebuf); 

         outp = wproxptr->bufptr;
         outsz = br;
         /*====================================================================
          * Call user exit if necessary
          *===================================================================*/
         if (exitidx >= 0) {
            int rc;
            if (u_exits[exitidx].CtoSBuf != NULL) {
               UE_BUF_ARGS ue_buf_args;

               ue_buf_args.permitinfo = &ue_permit;
               ue_buf_args.inptr = outp;
               ue_buf_args.insize = outsz;
               ue_buf_args.outptr = NULL;
               ue_buf_args.outsize = 0;
               if ((rc = (u_exits[exitidx].CtoSBuf)(&ue_buf_args)) < 0) {
                  syslog(LOG_NOTICE, "User provided CtoSBuf exit "
				"returned \"%d\".  Terminating connection.",
				rc);
                  break;
               }
               else {
                  if (ue_buf_args.outptr != NULL) {
                     outp = ue_buf_args.outptr;
                     outsz = ue_buf_args.outsize;
                  }
               }
            }
         }

         if((bw = write(serfd,outp,outsz)) != outsz) 
         {  
	    syslog(LOG_ERR,"Error: write to server returned %d expected %d.  %m",
				bw,outsz);
	    break; 
         }
	 obytes += bw;  
      } 
      /*==================================================================
       * If request is from server read server data and write to src host
       *=================================================================*/
      if(FD_ISSET(serfd,&redy)) 
      {
         if((br = read(serfd,wproxptr->bufptr,wproxptr->bufsize)) < 1 )
         {
            if( br == 0)
            {  
               if( wproxptr->traceflg)
                  trace_it(&traceh, 'S', br, " EOF",tracebuf);
               break;		/* EOF so quit	*/ 
            }
            syslog(LOG_ERR,"Error: while reading from server host.  %m"); 
            if( wproxptr->traceflg)
               trace_it(&traceh, 'S', br, " READ ERROR",tracebuf);
	    break; 
         } 
         wproxptr->ns->totalreads++;   
         if( wproxptr->traceflg)
            trace_it(&traceh, 'S', br, wproxptr->bufptr,tracebuf);

         outp = wproxptr->bufptr;
         outsz = br;
         /*====================================================================
          * Call user exit if necessary
          *===================================================================*/
         if (exitidx >= 0) {
            int rc;
            if (u_exits[exitidx].StoCBuf != NULL) {
               UE_BUF_ARGS ue_buf_args;

               ue_buf_args.permitinfo = &ue_permit;
               ue_buf_args.inptr = outp;
               ue_buf_args.insize = outsz;
               ue_buf_args.outptr = NULL;
               ue_buf_args.outsize = 0;
               if ((rc = (u_exits[exitidx].StoCBuf)(&ue_buf_args)) < 0) {
                  syslog(LOG_NOTICE, "User provided StoCBuf exit "
				"returned \"%d\".  Terminating connection.",
				rc);
                  break;
               }
               else {
                  if (ue_buf_args.outptr != NULL) {
                     outp = ue_buf_args.outptr;
                     outsz = ue_buf_args.outsize;
                  }
               }
            }
         }
	 if((bw = write(0,outp,outsz)) != outsz) 
         { 
	    syslog(LOG_ERR,"Error: write to source returned %d expected %d.  %m",
				bw,outsz);
	    break;  
         }
	 ibytes += bw;
      }   
waitfornext:;
   }
   /*=================================================================  
    * gather and log statistics     
    *=================================================================*/ 
   time(&endtime); 

   duration = endtime - starttime;      
   hours    = duration/3600;
   duration = duration -(hours*3600);
   min	    = duration/60;
   sec 	    = duration - (min*60); 

   syslog(LOG_NOTICE,"disconnect src_host = %s (%s) src_port = %d, dest_host "
		   "= %s dest_port = %d  input = %d bytes output = %d bytes, " 
		   "Blocks = %d, "
		   "duration = %3.2d:%2.2d:%2.2d",
		   wproxptr->ns->srcHost, wproxptr->ns->srcIP, srcport,
		   CURRENT_RULE.destaddr, CURRENT_RULE.destport, ibytes, obytes,
		   wproxptr->ns->totalreads,
                   hours, min, sec);
   /*========================================================================
    * Close input and output sockets and exit.  If tracing is active then
    * write the trace records.
    *=======================================================================*/
   close(0);                   			/* added 6/02/95 rjl	*/
   close(serfd); 
   if( wproxptr->traceflg)
      close(traceh.fd);
   return(0);
}        
/*==================================================================
 * function: 	isalldigits
 * input:	a pointer to a character string 
 * output:	returns 1 if all the characters in the string are digits 
 *==================================================================*/  
isalldigits(char *s)  
{
   while(isdigit(*s))
      s++;
   return(*s == '\0');
} 

/*==========================================================================
 *  function:   write a hex formatted dump to the file fd. 
 *  input:	file descriptor
 *		number of byttes to dump
 *		pointer to start of data to be dumped. 	
 *==========================================================================*/
void hexprint(int fd, int bytes, char * cptr)    
{     
   char buf[(WGTRACESIZE+1)*2];  
   int blks = 0;  
   int disp = 0;           
   int i; 
   int j = 0;     
   int k = 14;
   int l;   
   int left; 
   int numc;
   int offset = 60; 

   unsigned char c; 
   char *cptr1;    
   char *cptr2;
   cptr1 = cptr;		 	/* initialize to beginning */
   cptr2 = cptr; 
   buf[0] = '\n';

   memset(&buf[1],' ',sizeof(buf)-2);
   buf[sizeof(buf)-1] = '\0'; 
   left = bytes;   
   /*=========================================================
    *  print hex dump of buffer followed by ascii text dump 
    *========================================================*/ 
   for(i = 0; i < bytes; i++)
   {   
      j++;                
      c = *cptr1++; 
      if(k == 14)
      {
         sprintf(&buf[1],"%8.8x     ",disp);
         disp = disp + 16; 
      }
      sprintf(&buf[k],"%2.2x",c); 
      
      k = k + 2; 
      /* every fourth set of hex numbers add a space */
      if( j == 4 )
      {
         j = 0;
         sprintf(&buf[k]," ");  
         k++;
         blks++;

      } 
      buf[offset+18] = '\0';
      if( blks == 4  ||  i == (bytes-1) )
      {  
         buf[strlen(buf)] = ' ';    
         buf[offset]    = '|';  
         buf[offset+17] = '|';       
         l = 0;
         while( left &&  l < 16 )
         {                       
            l++;
            left--;                   
            if( *cptr2 < 32 || *cptr2 > 126)
               buf[offset+l] = '.';
            else
               buf[offset+l] = *cptr2;
            cptr2++;   
            k = 14; 
            /*buf[offset+1+l] = '\0';*/
         }

         /*fprintf(fd,buf);*/
         write(fd,buf,strlen(buf));
         memset(&buf[10],' ',40);    
         memset(&buf[offset+1],' ',16);
         blks = 0;
      } 
   } 
   write(fd,"\n",1);   
   return;

}
/*============================================================================
 * function: trace_init 
 *	Initialize trace header, open trace file and write header     
 * inputs:
 *	address of trace header      
 *      address of WGCONF
 * outputs:  
 *	returns  0 if ok  
 *		-1 on error  
 *===========================================================================*/
int trace_init(struct tracehdr *th, struct WGCONF *wgconfptr, char *tbuf)
{
   int	i;
   char		*ctimeptr;
   long		now = 0;   
   char tracefile[256];
   struct trace *tracebufadr; 
   struct trace *trptr;
   trptr = tracebufadr;
   th->recno = 0;
   time(&now);
   ctimeptr = ctime(&now);
   strcpy(th->title,"Aproxy V 1.0 Trace Output:  ");
   strcat(th->title,ctimeptr);
   strcpy(th->srcHost,wgconfptr->ns->srcHost);
   strcpy(th->srcIP,wgconfptr->ns->srcIP);
   strcpy(th->localIP,wgconfptr->ns->localIP);
   th->localport = wgconfptr->ns->localport;
   strcpy(th->destIP,wgconfptr->ns->destIP);
   th->destport = wgconfptr->ns->destport; 
   th->pid      = getpid();
   strcpy(tracefile,"trace.aproxy"); 
   strcat(tracefile,".");
   strcat(tracefile,th->srcIP);
   if((th->fd = open(tracefile, O_WRONLY | O_CREAT | O_APPEND,
                                S_IRUSR | S_IWUSR | S_IRGRP)) < 0)
   {
      syslog(LOG_ERR,"Error: open failed for tracefile %s. strerror(%s)",
                         tracefile, strerror(errno));
      return(-1);    
   }  
   /*===================================
    * print trace header      
    *===================================*/
   sprintf(tbuf,"\n%s\n" 
	       "Client:     IP %15.15s hostname '%s'\n"
               "local:      IP %15.15s port     '%d'\n"
               "Server:     IP %15.15s port     '%d'\n"
               "PID:               %d\n\n",
	       th->title,th->srcIP,th->srcHost,
	       th->localIP,th->localport,   
               th->destIP,th->destport,th->pid); 
   write(th->fd,tbuf,strlen(tbuf));


   return(0);
} 
/*============================================================================
 * function: trace_it 
 *
 * inputs:
 *	address of trace header
 *	source (C,S,...)         
 *      number of bytes
 *      buffer address
 * outputs:  
 *	returns 0	OK 
 *		1       error 
 *===========================================================================*/
int trace_it(struct tracehdr *th, char sr, int br, char *buf, char *tbuf)
{     
   int		bytes = 0;       
   char		*ctimeptr;
   long		now = 0;   
   time(&now);
   ctimeptr = ctime(&now);  
   th->recno++; 
   /* Limit the maximum number of bytes printed */
   if( br < WGTRACESIZE)
      bytes     = br; 
   else
      bytes = WGTRACESIZE; 
   if( br < 2)
      sprintf(tbuf,"\n%19.19s pid %6d record %6d  %5d bytes received from %c,"
                   "  %s \n",
	           ctimeptr, th->pid, th->recno, br, sr, buf);   
   else
      sprintf(tbuf,"\n%19.19s pid %6d record %6d  %5d bytes received from %c \n",
	           ctimeptr, th->pid, th->recno, br, sr);   
   write(th->fd,tbuf,strlen(tbuf));
   hexprint(th->fd, bytes, buf);  
   return(0);
} 

/*===========================================================================
 * debugbind 
 * input:	debug port number
 *==========================================================================*/
void debugbind(unsigned short debugport)
{
   struct	sockaddr_in	mya;
   int	x;
   int	nread;

   if((x = socket(AF_INET,SOCK_STREAM,0)) < 0) 
   {
      perror("socket");
      exit(1);
   }
   memset(&mya, 0, sizeof(mya));
   mya.sin_family = AF_INET;
   mya.sin_port = htons(debugport);
   if(bind(x,(struct sockaddr *)&mya,sizeof(mya))) 
   {
      perror("bind");
      exit(1);
   }
   if(listen(x,1) < 0) 
   {
      perror("listen");
      exit(1);
   }
   if((nread = accept(x,0,0)) < 0) 
   {
      perror("accept");
      exit(1);
   }
   close(0);
   dup(nread);
   close(1);
   dup(nread);
}
/*===========================================================================
 * Trap Urgent Signals
 *==========================================================================*/
static	void trap_sigurg(int sig)
{
   int	atmark;
   char	buf[512];
   char	c;

   signal(SIGURG,trap_sigurg);
   while(1) 
   {
      if(ioctl(0,SIOCATMARK,(char *)&atmark) < 0)
         return;
      if(atmark)
         break;
      if (read(0,buf,sizeof(buf)) <= 0)
         return;
   }
   if(extrfd != -1) 
   {
      if(recv(0,(char *)&c,sizeof(c),MSG_OOB) < 0)
         return;
      send(extrfd,(char *)&c,sizeof(c),MSG_OOB); 
      sprintf(urgsigmsg,"Urgent signal trapped and  '%c' sent.",c);
      syslog(LOG_NOTICE,urgsigmsg);
   }
}
