/*============================================================================
 * filename:		../webgate/ooba.c	../aproxy/ooba.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *										
 * written by:		Jim Livermore
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.
 *	All rights reserved.
 *		
 * Date written:	July      31, 1995
 *			October    2, 1996 - added ooba2 support
 * Last modified:	February   8, 2000 - added LPL
 *	                         
 * function:                       
 *	The ooba function performs Out Of Band Authentication for the 
 *      webgate/aproxy proxy server. Ooba connects to the source host, creates
 *	an eight digit random number and sends it to the ooba server on the
 *	source host. The random number is formatted as an eight digit character
 *	string. Ooba then reads the reply and sends to snkverify for 
 *	verification. The OOB connection is closed and control is returned to 
 *	getprox.    
 *  
 *	
 *
 *============================================================================*/
#include	<sys/types.h>
#include	<stdio.h>
#include	<syslog.h>
#include	<string.h>
#include	<sys/time.h> 
#ifndef __hppa
#include 	<sys/select.h>		/* to support fd_set 		  */ 
#endif 
#include	<netdb.h>               /* defines servent structure      */
#include	<netinet/in.h>
#include	"firewall.h" 
#include	"gwuser.h" 
#ifdef APROXY
#include	"aproxy.h"
#else
#include	"webgate.h"    
#endif
/*===============================================================
 * external variables
 *==============================================================*/ 
extern struct 	timeval *tp;        /* pointer to timeout value	*/
extern char	urgsigmsg[];              
/*=====================================================
 * function prototypes
 *===================================================*/ 

/*=============================================================================
 *
 * inputs:	ptr to source host IP address
 *		source host port id 
 *		ptr to destination host IP address
 *		ptr to destination port
 * 
 * outputs:	return  0 if access is permitted
 *		return -1 if access is denied   
 *
 * Called by getprox. 
 *===========================================================================*/
int ooba(char *destIP,  char * username, int portid,
         struct WGCONF *wproxptr)
{   
#ifndef	BSIZ
#define	BSIZ	2048
#endif     
   struct GWUSER gwu; 
   struct GWUSER *gwuptr  = &gwu; 
   char		buf[BSIZ];
   char		challenge[9]; 
   char 	*cp; 
   int		destfd;			/* destination host FD		*/  
   unsigned int	seed; 
   int		ichall;
   int		msglen;  
   struct timeval now;
   int		privport 	= 0;
   int		rc; 
   struct servent	*sp; 		/* ptr to service entry 	*/
   /* added to support select statement for timeouts	*/   
   fd_set	redy;
   fd_set	iced;
   if( *username == '&')
   {
      return( ooba2(destIP,  username, portid,wproxptr) );
   }
   if( (destfd = connectserver(destIP,portid,privport,NULL) ) < 0 ) 
   {
      syslog(LOG_WARNING,"Security Alert: cannot connect to OOBA server %s "
                         "port %d: %m.",
		         destIP,portid);  
      return(-1);
   } 
   if( get_gwuser(username, gwuptr) < 0)
   {
      syslog(LOG_NOTICE,"Error:  Unable to establish OOBA connection. "
			"Get failed for gw user '%s'. Check preceeding "
 			"%s error message.",username, PROCNAME);   
      return(-1);
   }  
   challenge[8] = '\0';      		/* set end of string	*/
   /*================================================================= 
    * Generate random number to be used as challenge for ooba server.
    * Use gettimeofday which has microsecond resolution.
    *================================================================*/   
   gettimeofday(&now,NULL);
   seed = now.tv_sec/97 + now.tv_usec/7;	/* prevent overflow	 */
   srand48((int) seed);			/* seed the random # generator  */
   					/* Limit max size to 98,808,865 */
   ichall = (lrand48() & 0x03FFFFFF) + 31700001; 
   sprintf(challenge,"%d",ichall);
   strcpy(buf,challenge);
   msglen = strlen(buf)+1;   

   FD_ZERO(&iced);   			/* Clear all bits in fdset iced	*/
   FD_SET(destfd,&iced);             	/* Turn on bit for destfd  	*/
   tp->tv_usec = 0; 
   tp->tv_sec = wproxptr->timeout;
   
   /*=================================================================
    * Send challenge to ooba server  
    *================================================================*/
   if(write(destfd,buf,msglen) != msglen)
   {   
      syslog(LOG_ERR,"Error: unable to write OOBA challenge to %s port %d.  %m.",
			destIP,portid);
      close(destfd);
      return(-1);
   }
   /*=================================================================
    * get response 
    * Intialize fd_set pointed to by redy and select the next request.		
    *================================================================*/
   (void)bcopy(&iced,&redy,sizeof(fd_set));
#ifdef __hppa
   rc = select(destfd + 2,(int *)&redy,(int *)0,(int *)0,tp);
#else
   rc = select(destfd + 2,&redy,(fd_set *)0,(fd_set *)0,tp);
#endif
   if( urgsigmsg[0] != '\0')
   {  
      syslog(LOG_WARNING,"Security Alert: received urgent signal from client's "
			 "OOBA server: %s port %d.  %m.",
			 destIP, portid);
      close(destfd);
      return(-1);
   }
   if( rc == 0) 
   {
      syslog(LOG_WARNING,"Security Alert: OOBA connection timeout after %d sec."
    			 " host = (%s) port = %d",
             tp->tv_sec, destIP, portid); 
/*
      if( wproxptr->traceflg)
         trace_it(&traceh, tracebufadr,'T',0," timeout.");
*/    
      close(destfd);
      return(-1);  
   }
   if( rc < 0) 
   {
      syslog(LOG_ERR,"Error: OOBA select returned an error: rc = %d. "
 		     "host = (%s)",
              	     rc, destIP); 
/*
      if( wproxptr->traceflg)
         trace_it(&traceh, tracebufadr,'E',0,"  select error.");
*/   
      close(destfd);
      return(-1);
   }
   buf[0] = '\0';
   if( read(destfd,(unsigned char *)buf,BSIZ - 1) < 0)  
   {
      syslog(LOG_WARNING,"Security Alert: unable to read response to challenge "
			 "from (%s) port %d.  %m.", destIP,portid); 
      close(destfd);
      return(-1); 
   } 
   
   rc = snkverify(challenge, buf, gwuptr);
   /*===================================================================
    * Send ooba server message indicating success or failure. 
    *==================================================================*/ 
   if( rc == 0)
   {
      strcpy(buf,"001 Connection permitted.");
   }                                        
   else
   { 
      
      syslog(LOG_WARNING,"Security Alert: OOBA connection denied. host (%s) "
			 "port %d challenge = %s response = %s",
	     destIP, portid, challenge, buf); 
      strcpy(buf,"000 Connection denied."); 
   }
   msglen = strlen(buf)+1;
   if(write(destfd,buf,msglen) != msglen) 
   {
      syslog(LOG_ERR,"Error: unable to write OOBA Rtn Code message to %s port"
		     " %d.  %m.",
		     destIP,portid); 
   }
   close(destfd);
   return(rc);
}
/*=======================================================================
 * OOBA Version 2  
 *
 * inputs:
 *	ptr to destination IP address
 *	ptr to user name
 *	port ID for challenge
 *	ptr to WGCONF structure   
 *
 * logic:
 *	Connect to ooba server
 *	Give the remote user three tries to respond to the challenge 
 *		1. send request user name 
 *		2. read name 
 *		3. lookup up user record
 *		4. send challenge
 *		5. read response
 *		6. verify response
 *		7. send message indicating success or failure
 *
 * output:
 *  	rc = 0 of user is authenticated. 
 *=======================================================================*/   
int ooba2(char *destIP,  char * username, int portid,
         struct WGCONF *wproxptr)
{   
#ifndef	BSIZ
#define	BSIZ	2048
#endif     
   struct GWUSER gwu; 
   struct GWUSER *gwuptr  = &gwu; 
   struct OOBA2 oobablock;
   struct OOBA2	*obptr  = &oobablock;
   char		challenge[9]; 
   char 	*cp; 
   int		destfd;			/* destination host FD		*/  
   time_t	epoch; 
   int		ichall;
   int		msglen; 
   struct timeval now;
   int		privport 	= 0;
   int		rc;  
   unsigned int seed;
   struct servent	*sp; 		/* ptr to service entry 	*/   
   int          tries = 3;
   /* added to support select statement for timeouts	*/   
   fd_set	redy;
   fd_set	iced;
#ifdef linux
   struct timeval	tt;
#endif

   if( (destfd = connectserver(destIP,portid,privport,NULL) ) < 0 ) 
   {
      syslog(LOG_WARNING,"Security Alert: cannot connect to OOBA server %s "
			 "port %d: %m.",
		         destIP,portid);  
      return(-1);
   } 
#ifdef linux
   memcpy(&tt, tp, sizeof(tt));
#endif
   FD_ZERO(&iced);   		/* Clear all bits in fdset iced	*/
   FD_SET(0,&iced);             /* Turn on bit for fd 0 	*/
   FD_SET(destfd,&iced);    	/* turn on bit for fd serfd	*/
   while ( tries )        
   {
      tries--;
      /*=================================================================
       * Send user name prompt  
       *================================================================*/
      obptr->sync = 253;			/* re-SYNC		*/
      obptr->command = OOBAPROMPT;  
      strcpy(obptr->msg,"username: ");
      msglen = strlen(obptr->msg) + 3;
      if(write(destfd,(char *)obptr,msglen) != msglen)
      {   
         syslog(LOG_ERR,"Error: unable to write username prompt to %s port %d.  %m.",
			destIP,portid);
         close(destfd);
         return(-1);
      } 
      /*=================================================================
       * get user name 
       * Intialize fd_set pointed to by redy and select the next request.		
       *================================================================*/
#ifdef linux
      memcpy(tp, &tt,  sizeof(*tp));
#endif
      (void)bcopy(&iced,&redy,sizeof(fd_set));
#ifdef __hppa
      rc = select(destfd + 2,(int *)&redy,(int *)0,(int *)0,tp);
#else
      rc = select(destfd + 2,&redy,(fd_set *)0,(fd_set *)0,tp);
#endif
      if( urgsigmsg[0] != '\0')
      {  
         syslog(LOG_WARNING,"Security Alert: received urgent signal from"
			    " client's OOBA server: %s port %d.  %m.",
			    destIP, portid);
         close(destfd);
         return(-1);
      }
      if( rc == 0) 
      {
         syslog(LOG_WARNING,"Security Alert: OOBA connection timeout waiting"
			    " for name, %d sec. host = (%s) port = %d",
                	    tp->tv_sec, destIP, portid); 
/*
      if( wproxptr->traceflg)
         trace_it(&traceh, tracebufadr,'T',0," timeout.");
*/    
         close(destfd);
         return(-1);  
      }
      if( rc < 0) 
      {
         syslog(LOG_ERR,"Error: OOBA select returned an error: rc = %d."
			" host = (%s)",
              		rc, destIP); 
/*
      if( wproxptr->traceflg)
         trace_it(&traceh, tracebufadr,'E',0,"  select error.");
*/   
         close(destfd);
         return(-1);
      }
      if( read(destfd,(char *)obptr,sizeof(struct OOBA2)) < 0)  
      {
         syslog(LOG_WARNING,"Security Alert: unable to read username "
			    "from (%s) port %d.  %m.", destIP,portid); 
         close(destfd);
         return(-1); 
      }      
      strncpy(username,obptr->msg,17);
      if( (rc = get_gwuser(username, gwuptr)) < 0)
      {      
         syslog(LOG_NOTICE,"Error:  Unable to establish OOBA connection. "
			"Get failed for gw user '%s'. Check preceeding "
 			"genproxy error message.",username);   
         
         goto endofloop;
      }      
      if( rc > 0) 
      {  
         switch( rc )
         {  
            case 1:
            	syslog(LOG_WARNING,"Security Alert:  OOBA rejected. User '%s' "
				"record locked.", username); 
            case 2:   
            	syslog(LOG_WARNING,"Security Alert:  OOBA rejected user '%s'. "
				"Current date outside of effective start and"
				"stop dates.", username);  
            case 3:
            	syslog(LOG_WARNING,"Security Alert:  OOBA rejected user '%s'. "
				   "Current time is not within the allowed "
				   "times.", username);  
         }
         rc = -1;
         goto endofloop;
      }      

      /*================================================================= 
       * Generate random number to be used as challenge for ooba server.
       * Use gettimeofday which has microsecond resolution.
       *================================================================*/   
      gettimeofday(&now,NULL);
      seed = now.tv_sec/97 + now.tv_usec/7;	/* prevent overflow	 */
      srand48((int) seed);		/* seed the random # generator */
      ichall = (lrand48() & 0x03FFFFFF) + 31700001; 
      sprintf(challenge,"%d",ichall);  
      obptr->sync = 1;
      obptr->command = OOBAPROMPT;  
      strcpy(obptr->msg,challenge);
      msglen = strlen(obptr->msg) + 3;


      FD_ZERO(&iced);  			/* Clear all bits in fdset iced	*/
      FD_SET(destfd,&iced);            	/* Turn on bit for destfd  	*/
      tp->tv_usec = 0; 
      tp->tv_sec = wproxptr->timeout;
   
      /*=================================================================
       * Send challenge to ooba server  
       *================================================================*/
      if(write(destfd,(char *) obptr,msglen) != msglen)
      {   
         syslog(LOG_ERR,"Error: unable to write OOBA challenge to %s port %d."
			"  %m.",
			destIP,portid);
         close(destfd);
         return(-1);
      }
      /*=================================================================
       * get response 
       * Intialize fd_set pointed to by redy and select the next request.
       *================================================================*/
#ifdef linux
      memcpy(tp, &tt,  sizeof(*tp));
#endif
      (void)bcopy(&iced,&redy,sizeof(fd_set));
#ifdef __hppa
      rc = select(destfd + 2,(int *)&redy,(int *)0,(int *)0,tp);
#else
      rc = select(destfd + 2,&redy,(fd_set *)0,(fd_set *)0,tp);
#endif
      if( urgsigmsg[0] != '\0')
      {  
         syslog(LOG_WARNING,"Security Alert: received urgent signal from "
			    "client's OOBA server: %s port %d.  %m.",
	   		    destIP, portid);
         close(destfd);
         return(-1);
      }
      if( rc == 0) 
      {
         syslog(LOG_WARNING,"Security Alert: OOBA connection timeout waiting"
			    " for response, %d sec. host = (%s) port = %d",
             		    tp->tv_sec, destIP, portid); 
/*
      if( wproxptr->traceflg)
         trace_it(&traceh, tracebufadr,'T',0," timeout.");
*/    
         close(destfd);
         return(-1);  
      }
      if( rc < 0) 
      {
         syslog(LOG_ERR,"Error: OOBA select returned an error: rc = %d."
			" host = (%s)",
                 	rc, destIP); 
/*
      if( wproxptr->traceflg)
         trace_it(&traceh, tracebufadr,'E',0,"  select error.");
*/   
         close(destfd);
         return(-1);
      }
      if( read(destfd,(char *)obptr,sizeof(struct OOBA2)) < 0)  
      {
         syslog(LOG_WARNING,"Security Alert: unable to read response to"
			    " challenge from (%s) port %d.  %m.",
			    destIP,portid); 
         close(destfd);
         return(-1); 
      } 
   
      rc = snkverify(challenge, obptr->msg, gwuptr);
      /*===================================================================
       * Send ooba server message indicating success or failure. 
       *==================================================================*/ 
      if( rc == 0)
      {
         strcpy(obptr->msg,"Connection permitted.");  
      }                                        
      else
      {    
         syslog(LOG_WARNING,"Security Alert: OOBA connection denied. "
			    "source host %s (%s) port %d user = %s"
			    " challenge = %s response = '%s'",
	        	wproxptr->ns->srcHost, wproxptr->ns->srcIP, portid,
			username,
                	challenge, obptr->msg); 
         strcpy(obptr->msg,"Connection denied."); 
      } 
      obptr->sync = 3;
      obptr->command = OOBASHOW; 
      msglen = strlen(obptr->msg) + 3;
      if(write(destfd,(char *) obptr,msglen) != msglen) 
      {
         syslog(LOG_ERR,"Error: unable to write OOBA Rtn Code message to"
			" %s port %d.  %m.",
			destIP,portid); 
      }
      if( rc == 0 ) 
         break;
      else     
         wait(1);

  
endofloop:;
   } /* end of tries */ 
      close(destfd);
      return(rc);  
}
