/*============================================================================
 * filename:		../webgate/secfile.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *										
 * written by:		Tommy Chan
 *
 * (C) COPYRIGHT R. James Livermore 1997-2000.  All rights reserved.
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1997-2000.
 * All rights reserved. 
 *           
 *		
 * Date written:	December  20, 1997   
 *			September  8, 1999 - added Linux support
 * Last modified:	February   8, 2000 - added LPL
 *  
 * function:
 *============================================================================*/
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/ipc.h>
#include <netinet/in.h>
#include <sys/mman.h>
#include "firewall.h"
#ifdef APROXY
#include <aproxy.h>
#else
#include "webgate.h"
#endif

#ifndef linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif

/*=======================================================================
 * This is defined here so it can override the ipsecure in libsrc/firewall.a.
 * main.c for webgate/aproxy calls peername() (in firewall.a) which in turn
 * calls ipsecure().  This ipsecure() defined here will force the firewall.a
 * ipsecure() to NOT be called.
 *======================================================================*/
int ipsecure(char *IP)
{
    return(isIPlisted(IP, SECURENETFILE));
}

/*=======================================================================
 * Input:
 *	size	The size of the area to map in shm
 *	mfile	Name of file to be mapped or NULL for "anonymous" map
 * Returns:
 *	ptr	to the mapped region if successful
 *	NULL	if failed
 *======================================================================*/
char *getshm(int size, char *mfile) {
    int fd;
    char *cp;
    char *fn;
#ifdef linux
#define TMPFILE	FW_CFG_DIR "/.shr.tmp"
#else
#define TMPFILE "/dev/zero"
#endif
    
    if (mfile == NULL)
	fn = TMPFILE;
    else
	fn = mfile;

    /*===================================================================
     * Open the mapped file
     *==================================================================*/
    if ((fd = open(fn, O_RDWR|O_CREAT, 0600)) < 0) {
	syslog(LOG_ERR,"Error: open of file \"%s\" failed.  %s",
					fn, sys_errlist[errno]);
	return(NULL);
    }
    
    /*=======================================================================
     * Set the area size if mfile is not NULL or if on Linux
     *======================================================================*/
#ifndef linux
    if (mfile != NULL) {
#endif
	if (lseek(fd, size, SEEK_SET) < 0) {
	    syslog(LOG_ERR, "Error: Cannot seek to offset %d in file %s.  %m",
			    size, fn);
	    return(NULL);
	}
	if (write(fd, "", 1) != 1) {
	    syslog(LOG_ERR, "Error: Cannot write to offset %d in file %s.  %m",
			    size, fn);
	    return(NULL);
	}
#ifndef linux
    }
#endif

    /*=======================================================================
     * Map the area
     *======================================================================*/
    if ((cp = mmap(NULL, size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0))
		    						== (void *)-1) {
	syslog(LOG_ERR,"Error: mmap failed for file \"%s\", size %d.  %m",
					fn, size);
	close(fd);
	return(NULL);
    }
    
    close(fd);
#ifdef linux
    /*=======================================================================
     * On linux, unlink the tmp file
     *======================================================================*/
    if (mfile == NULL)
	if (unlink(fn) < 0) {
	    syslog(LOG_ERR, "Error: Cannot unlink %s.  %m", fn);
	}
#endif
    
    return(cp);
}

/*=======================================================================
 * Reads the file fn and writes the ip address entries to the mapped area
 * pointed to by ptr.
 *
 * Returns:
 *		number of entries in the nentriesp argument
 *		ptr to shared memory area holding if successful
 *		NULL if failed
 *======================================================================*/
IPSTR *read_secfile(char *fn, int *nentriesp) {
    int line = 0;
    int secport;
    int size = 0;
    FILE *fp;
    char buf[BUFSIZ];
    char *shmptr;
    IPSTR *ipstr = NULL;

    *nentriesp = 0;
    if (strcmp(fn, SECUREPORTFILE) == 0)
	secport = 1;
    else
	secport = 0;

    /*=======================================================================
     * Open the input file
     *======================================================================*/
    if ((fp = fopen(fn, "r")) == NULL) {
	syslog(LOG_ERR,"Error: Could not open file %s.  %s",
				fn, sys_errlist[errno]);
	return(NULL);
    }

    size = 0;

    /*=======================================================================
     * Read the input file
     *======================================================================*/
    while (fgets(buf, sizeof(buf), fp) != NULL) {
	int len;
	char *token;

	line++;
	token = strtok(buf, " \r\n\t");

	if (*token == '#')		/* skip comment line	*/
	    continue;

	len = strlen(token);
	if (len == 0)			/* skip blank lines	*/
	    continue;

	if (len < 2) {
	    syslog(LOG_ERR,"Error: IP address '%s' on line %d of file %s is < "
			"2 characters.",
			token, line, fn);
	    fclose(fp);
	    return(NULL);
	}

	if (len > 16) {
	    syslog(LOG_ERR,"Error: IP address '%.21s' on line %d of file %s "
			"is > 15 characters.",
			token, line, fn);
	    fclose(fp);
	    return(NULL);
	}

	/*===============================================================
	 * If reading secureports file, any '*' is invalid
	 *==============================================================*/
	if (secport && strchr(token,'*')) {
	    syslog(LOG_ERR,"Error: %s file cannot have entries with '*'",fn);
	    return(NULL);
	}

	/*===============================================================
	 * Now we know we have a good entry, malloc (realloc) space for it
	 * malloc 2 * size initially for the first record and a null record
	 * that signifies the end
	 *==============================================================*/
	size += sizeof(IPSTR);
	if (size == sizeof(IPSTR)) {
	    size += size;
	    if ((ipstr = malloc(size)) == NULL) {
		syslog(LOG_ERR,"Error: Could not malloc space.  %s",
						sys_errlist[errno]);
		return(NULL);
	    }
	}
	else {
	    if ((ipstr = realloc(ipstr, size)) == NULL) {
		syslog(LOG_ERR,"Error: Could not realloc.  %s",
						sys_errlist[errno]);
		return(NULL);
	    }
	}

	/*===============================================================
	 * Write to the malloc area
	 *==============================================================*/
	strcpy(ipstr[(*nentriesp)++], buf);
	strcpy(ipstr[(*nentriesp)], "");
    }

    /*=======================================================================
     * Make sure it wasn't an error that stopped the read
     *======================================================================*/
    if (ferror(fp)) {
	syslog(LOG_ERR,"Error: reading record %d in file %s.  %s.",
			line, fn, sys_errlist[errno]);
	fclose(fp);
	return(NULL);
    }

    fclose(fp);

    /*=======================================================================
     * Now we know the size, get a chunk of shared memory to hold the
     * entries and copy the entries over
     *======================================================================*/
    if ((shmptr = getshm(size, NULL)) == NULL)
	return(NULL);

    memcpy(shmptr, ipstr, size);

    free(ipstr);
    return((IPSTR *)shmptr);
}

/*=======================================================================
 * Input:
 *	IP	String representation of the IP address to check
 *	list	pointer to a list of IP entries from the securenets or
 *		secureports file
 * Returns:
 *	1	The IP is on the list
 *	0	The IP is not on the list
 *======================================================================*/
int isIPlisted(char *IP, IPSTR *list)
{
    int i;

    for (i = 0; strlen(list[i]) > 0; i++) {
	if (strcmp(IP, list[i]) == 0)
	    return(1);
    }
    return(0);
}
