/*
 * $Id: des.c,v 1.3 1996/04/29 19:43:56 rwt Exp rwt $
 * copyright 1996 (c) Digital Pathways, Inc.  All Rights Reserved
 *
 * This source code is provided under license from Digital Pathways,
 * Inc.  It is provided "AS-IS", without warranty of any kind, either
 * express or implied. It may be incorporated wholly or modified as
 * needed, subject to the following restrictions.
 *
 * 1) This notice, including the Digital Pathways, Inc. notice of
 * copyright must be included in any source code or derivative works.
 * If this source code is modified by the licensee, a notice of
 * modification must be included as well.
 *
 * 2) This Program, and any derivative works, may not be distributed
 * is source form, on any media, without the express written consent
 * of Digital Pathways, Inc.  There is no restriction on object or
 * executable code.
 */

static char rcsid[] = "$Id: des.c,v 1.3 1996/04/29 19:43:56 rwt Exp rwt $";

/* defines */

#define BITMASK0	0x80
#define BITMASK1	0x40
#define BITMASK2	0x20
#define BITMASK3	0x10
#define BITMASK4	0x08
#define BITMASK5	0x04
#define BITMASK6	0x02
#define BITMASK7	0x01

static short bitmask[] = {
	BITMASK0, BITMASK1, BITMASK2, BITMASK3,
	BITMASK4, BITMASK5, BITMASK6, BITMASK7
};

/*
 *	DES tables
 *
 *	The following tables are used by the DES algorithm.  Depending on
 *	your compiler, you may need to place these tables in a sepaate
 *	machine language file.
 *	
 */

static char round_shift[] = {
	1, 1, 2, 2, 2, 2, 2, 2,
	1, 2, 2, 2, 2, 2, 2, 1
};

/*
 *
 *	permutation tables:
 *
 *	Each byte is a bit index relative to the source bytes.  Note tables
 *	do not correspspond to NBS spec, because:
 *
 *	a) first bit (leftmost) in NBS = 1, processor index = 0
 *	b) some arrays are not packed (< 8 bits) but the indexing scheme
 *	assumes packed.
 *
 */

/*
 * Permuted Choice One
 * This is the initial bitwise permutation done on the key and is 
 * only done once.
 */

static char pch1[] = {
	56, 48, 40, 32, 24, 16, 8, 0,
	57, 49, 41, 33, 25, 17, 9, 1,
	58, 50, 42, 34, 26, 18, 10, 2,
	59, 51, 43, 35,
	62, 54, 46, 38, 30, 22, 14, 6,
	61, 53, 45, 37, 29, 21, 13, 5,
	60, 52, 44, 36, 28, 20, 12, 4,
	                27, 19, 11, 3
};

/*
 * Permuted Choice Two 
 * This is the bitwise compression permutation done to reduce the
 * rotated 56 bit key to 48 bits.  Since the left half and right half
 * of the key are permuted independently, the table can be split into
 * separate tables.
 */

static long left_pch2[] = {
	0x00004000, 0x00000800, 0x00020000, 0x00000010, 
	0x08000000, 0x00800000, 0x02000000, 0x00000001, 
	0x00002000, 0x00400000, 0x00000080, 0x00040000, 
	0x00000020, 0x00000200, 0x00010000, 0x01000000, 
	0x00000004, 0x00100000, 0x00001000, 0x00200000, 
	0x00000002, 0x00000100, 0x00008000, 0x04000000, 
};

static long right_pch2[] = {
	0x00008000, 0x00000010, 0x02000000, 0x00080000, 
	0x00000200, 0x00000002, 0x04000000, 0x00010000, 
	0x00000020, 0x00000800, 0x00800000, 0x00000100, 
	0x00001000, 0x00000080, 0x00020000, 0x00000001, 
	0x00400000, 0x00000008, 0x00000400, 0x00004000, 
	0x00000040, 0x00100000, 0x08000000, 0x01000000, 
};

/*
 * Initial Permutation
 * The initial permutation of the data.
 */

static char ip[] = {
	57, 49, 41, 33, 25, 17, 9, 1,
	59, 51, 43, 35, 27, 19, 11, 3,
	61, 53, 45, 37, 29, 21, 13, 5,
	63, 55, 47, 39, 31, 23, 15, 7,
	56, 48, 40, 32, 24, 16, 8, 0,
	58, 50, 42, 34, 26, 18, 10, 2,
	60, 52, 44, 36, 28, 20, 12, 4,
	62, 54, 46, 38, 30, 22, 14, 6
};

/*
 *	Inverse initial permutation (final perm)
 *
 *	note: table accounts for fact that left and right have been
 *	switched before permutation
 */

static char iip[] = {
	7, 39, 15, 47, 23, 55, 31, 63, 
	6, 38, 14, 46, 22, 54, 30, 62, 
	5, 37, 13, 45, 21, 53, 29, 61, 
	4, 36, 12, 44, 20, 52, 28, 60, 
	3, 35, 11, 43, 19, 51, 27, 59, 
	2, 34, 10, 42, 18, 50, 26, 58, 
	1, 33, 9, 41, 17, 49, 25, 57, 
	0, 32, 8, 40, 16, 48, 24, 56
};

/*
 *	Tables for S_box contractions
 * p-box permutation is built into these tables
 */

static unsigned long sbox_1[] = {
	0x00808200L, 0x00000000L, 0x00008000L, 0x00808202L, 
	0x00808002L, 0x00008202L, 0x00000002L, 0x00008000L, 
	0x00000200L, 0x00808200L, 0x00808202L, 0x00000200L, 
	0x00800202L, 0x00808002L, 0x00800000L, 0x00000002L, 
	0x00000202L, 0x00800200L, 0x00800200L, 0x00008200L, 
	0x00008200L, 0x00808000L, 0x00808000L, 0x00800202L, 
	0x00008002L, 0x00800002L, 0x00800002L, 0x00008002L, 
	0x00000000L, 0x00000202L, 0x00008202L, 0x00800000L, 
	0x00008000L, 0x00808202L, 0x00000002L, 0x00808000L, 
	0x00808200L, 0x00800000L, 0x00800000L, 0x00000200L, 
	0x00808002L, 0x00008000L, 0x00008200L, 0x00800002L, 
	0x00000200L, 0x00000002L, 0x00800202L, 0x00008202L, 
	0x00808202L, 0x00008002L, 0x00808000L, 0x00800202L, 
	0x00800002L, 0x00000202L, 0x00008202L, 0x00808200L, 
	0x00000202L, 0x00800200L, 0x00800200L, 0x00000000L, 
	0x00008002L, 0x00008200L, 0x00000000L, 0x00808002L, 
};

static unsigned long sbox_2[] = {
	0x40084010L, 0x40004000L, 0x00004000L, 0x00084010L, 
	0x00080000L, 0x00000010L, 0x40080010L, 0x40004010L, 
	0x40000010L, 0x40084010L, 0x40084000L, 0x40000000L, 
	0x40004000L, 0x00080000L, 0x00000010L, 0x40080010L, 
	0x00084000L, 0x00080010L, 0x40004010L, 0x00000000L, 
	0x40000000L, 0x00004000L, 0x00084010L, 0x40080000L, 
	0x00080010L, 0x40000010L, 0x00000000L, 0x00084000L, 
	0x00004010L, 0x40084000L, 0x40080000L, 0x00004010L, 
	0x00000000L, 0x00084010L, 0x40080010L, 0x00080000L, 
	0x40004010L, 0x40080000L, 0x40084000L, 0x00004000L, 
	0x40080000L, 0x40004000L, 0x00000010L, 0x40084010L, 
	0x00084010L, 0x00000010L, 0x00004000L, 0x40000000L, 
	0x00004010L, 0x40084000L, 0x00080000L, 0x40000010L, 
	0x00080010L, 0x40004010L, 0x40000010L, 0x00080010L, 
	0x00084000L, 0x00000000L, 0x40004000L, 0x00004010L, 
	0x40000000L, 0x40080010L, 0x40084010L, 0x00084000L, 
};

static unsigned long sbox_3[] = {
	0x00000104L, 0x04010100L, 0x00000000L, 0x04010004L, 
	0x04000100L, 0x00000000L, 0x00010104L, 0x04000100L, 
	0x00010004L, 0x04000004L, 0x04000004L, 0x00010000L, 
	0x04010104L, 0x00010004L, 0x04010000L, 0x00000104L, 
	0x04000000L, 0x00000004L, 0x04010100L, 0x00000100L, 
	0x00010100L, 0x04010000L, 0x04010004L, 0x00010104L, 
	0x04000104L, 0x00010100L, 0x00010000L, 0x04000104L, 
	0x00000004L, 0x04010104L, 0x00000100L, 0x04000000L, 
	0x04010100L, 0x04000000L, 0x00010004L, 0x00000104L, 
	0x00010000L, 0x04010100L, 0x04000100L, 0x00000000L, 
	0x00000100L, 0x00010004L, 0x04010104L, 0x04000100L, 
	0x04000004L, 0x00000100L, 0x00000000L, 0x04010004L, 
	0x04000104L, 0x00010000L, 0x04000000L, 0x04010104L, 
	0x00000004L, 0x00010104L, 0x00010100L, 0x04000004L, 
	0x04010000L, 0x04000104L, 0x00000104L, 0x04010000L, 
	0x00010104L, 0x00000004L, 0x04010004L, 0x00010100L, 
};

static unsigned long sbox_4[] = {
	0x80401000L, 0x80001040L, 0x80001040L, 0x00000040L, 
	0x00401040L, 0x80400040L, 0x80400000L, 0x80001000L, 
	0x00000000L, 0x00401000L, 0x00401000L, 0x80401040L, 
	0x80000040L, 0x00000000L, 0x00400040L, 0x80400000L, 
	0x80000000L, 0x00001000L, 0x00400000L, 0x80401000L, 
	0x00000040L, 0x00400000L, 0x80001000L, 0x00001040L, 
	0x80400040L, 0x80000000L, 0x00001040L, 0x00400040L, 
	0x00001000L, 0x00401040L, 0x80401040L, 0x80000040L, 
	0x00400040L, 0x80400000L, 0x00401000L, 0x80401040L, 
	0x80000040L, 0x00000000L, 0x00000000L, 0x00401000L, 
	0x00001040L, 0x00400040L, 0x80400040L, 0x80000000L, 
	0x80401000L, 0x80001040L, 0x80001040L, 0x00000040L, 
	0x80401040L, 0x80000040L, 0x80000000L, 0x00001000L, 
	0x80400000L, 0x80001000L, 0x00401040L, 0x80400040L, 
	0x80001000L, 0x00001040L, 0x00400000L, 0x80401000L, 
	0x00000040L, 0x00400000L, 0x00001000L, 0x00401040L, 
};

static unsigned long sbox_5[] = {
	0x00000080L, 0x01040080L, 0x01040000L, 0x21000080L, 
	0x00040000L, 0x00000080L, 0x20000000L, 0x01040000L, 
	0x20040080L, 0x00040000L, 0x01000080L, 0x20040080L, 
	0x21000080L, 0x21040000L, 0x00040080L, 0x20000000L, 
	0x01000000L, 0x20040000L, 0x20040000L, 0x00000000L, 
	0x20000080L, 0x21040080L, 0x21040080L, 0x01000080L, 
	0x21040000L, 0x20000080L, 0x00000000L, 0x21000000L, 
	0x01040080L, 0x01000000L, 0x21000000L, 0x00040080L, 
	0x00040000L, 0x21000080L, 0x00000080L, 0x01000000L, 
	0x20000000L, 0x01040000L, 0x21000080L, 0x20040080L, 
	0x01000080L, 0x20000000L, 0x21040000L, 0x01040080L, 
	0x20040080L, 0x00000080L, 0x01000000L, 0x21040000L, 
	0x21040080L, 0x00040080L, 0x21000000L, 0x21040080L, 
	0x01040000L, 0x00000000L, 0x20040000L, 0x21000000L, 
	0x00040080L, 0x01000080L, 0x20000080L, 0x00040000L, 
	0x00000000L, 0x20040000L, 0x01040080L, 0x20000080L, 
};

static unsigned long sbox_6[] = {
	0x10000008L, 0x10200000L, 0x00002000L, 0x10202008L, 
	0x10200000L, 0x00000008L, 0x10202008L, 0x00200000L, 
	0x10002000L, 0x00202008L, 0x00200000L, 0x10000008L, 
	0x00200008L, 0x10002000L, 0x10000000L, 0x00002008L, 
	0x00000000L, 0x00200008L, 0x10002008L, 0x00002000L, 
	0x00202000L, 0x10002008L, 0x00000008L, 0x10200008L, 
	0x10200008L, 0x00000000L, 0x00202008L, 0x10202000L, 
	0x00002008L, 0x00202000L, 0x10202000L, 0x10000000L, 
	0x10002000L, 0x00000008L, 0x10200008L, 0x00202000L, 
	0x10202008L, 0x00200000L, 0x00002008L, 0x10000008L, 
	0x00200000L, 0x10002000L, 0x10000000L, 0x00002008L, 
	0x10000008L, 0x10202008L, 0x00202000L, 0x10200000L, 
	0x00202008L, 0x10202000L, 0x00000000L, 0x10200008L, 
	0x00000008L, 0x00002000L, 0x10200000L, 0x00202008L, 
	0x00002000L, 0x00200008L, 0x10002008L, 0x00000000L, 
	0x10202000L, 0x10000000L, 0x00200008L, 0x10002008L, 
};

static unsigned long sbox_7[] = {
	0x00100000L, 0x02100001L, 0x02000401L, 0x00000000L, 
	0x00000400L, 0x02000401L, 0x00100401L, 0x02100400L, 
	0x02100401L, 0x00100000L, 0x00000000L, 0x02000001L, 
	0x00000001L, 0x02000000L, 0x02100001L, 0x00000401L, 
	0x02000400L, 0x00100401L, 0x00100001L, 0x02000400L, 
	0x02000001L, 0x02100000L, 0x02100400L, 0x00100001L, 
	0x02100000L, 0x00000400L, 0x00000401L, 0x02100401L, 
	0x00100400L, 0x00000001L, 0x02000000L, 0x00100400L, 
	0x02000000L, 0x00100400L, 0x00100000L, 0x02000401L, 
	0x02000401L, 0x02100001L, 0x02100001L, 0x00000001L, 
	0x00100001L, 0x02000000L, 0x02000400L, 0x00100000L, 
	0x02100400L, 0x00000401L, 0x00100401L, 0x02100400L, 
	0x00000401L, 0x02000001L, 0x02100401L, 0x02100000L, 
	0x00100400L, 0x00000000L, 0x00000001L, 0x02100401L, 
	0x00000000L, 0x00100401L, 0x02100000L, 0x00000400L, 
	0x02000001L, 0x02000400L, 0x00000400L, 0x00100001L, 
};

static unsigned long sbox_8[] = {
	0x08000820L, 0x00000800L, 0x00020000L, 0x08020820L, 
	0x08000000L, 0x08000820L, 0x00000020L, 0x08000000L, 
	0x00020020L, 0x08020000L, 0x08020820L, 0x00020800L, 
	0x08020800L, 0x00020820L, 0x00000800L, 0x00000020L, 
	0x08020000L, 0x08000020L, 0x08000800L, 0x00000820L, 
	0x00020800L, 0x00020020L, 0x08020020L, 0x08020800L, 
	0x00000820L, 0x00000000L, 0x00000000L, 0x08020020L, 
	0x08000020L, 0x08000800L, 0x00020820L, 0x00020000L, 
	0x00020820L, 0x00020000L, 0x08020800L, 0x00000800L, 
	0x00000020L, 0x08020020L, 0x00000800L, 0x00020820L, 
	0x08000800L, 0x00000020L, 0x08000020L, 0x08020000L, 
	0x08020020L, 0x08000000L, 0x00020000L, 0x08000820L, 
	0x00000000L, 0x08020820L, 0x00020020L, 0x08000020L, 
	0x08020000L, 0x08000800L, 0x08000820L, 0x00000000L, 
	0x08020820L, 0x00020800L, 0x00020800L, 0x00000820L, 
	0x00000820L, 0x00020020L, 0x08000000L, 0x08020800L
};

/*
 * pbits(target, source, table, count)
 *
 * permute the source bits into the target bits, using the specified
 * table.  Do this for count source bytes
 *
 * char *target;	output,
 * char *source;	input, bit addr from table, 0 = msb of zero byte
 * char *table;	decimal bit addr from source for next target byte
 * short count; length of the target, in bytes
 */

static void
pbits(char *target, char *source, char *table, short count)
{
	short i, j, mask;

	for (j = 0; j < count; j++) {
		*target = 0;
		for (mask = BITMASK0; mask; mask >>= 1) {
			i = *table++;
			if (source[i >> 3] & bitmask[i & 7])
				*target |= mask;
		}
		target++;
	}
}

/*
 *
 *	desfunction(target, source, flag, pkey)
 *
 *	if the flag is 0, encrypt the 8 characters of data pointed to by
 *	pointer.  If != 0, then decrypt the data.	in either case, use the
 *	permuted DES key stored in pkey
 *
 */

static void
desfunction(char *target, char *src, short decflag, unsigned long *pkey)
{
	short i;
	unsigned long left, right;
	unsigned long temp, temp1;
	unsigned char sd2[8];
	unsigned long *keyp;

	/*
	 * Do the initial permutation on the block.  Note that everything is
	 * declared unsigned to prevent sign extension problems.
	 */
	pbits((char *)sd2, src, ip, 8);

	/*
	 * Split the permuted cleartext into 32 bit left and right halves.  We
	 * use unsigned longs for the calculations.
	 */
	left = sd2[0] << 24 | sd2[1] << 16 | sd2[2] << 8 | sd2[3];
	right = sd2[4] << 24 | sd2[5] << 16 | sd2[6] << 8 | sd2[7];

	/*
	 * Initialize the per round key pointer to the end of the array if
	 * it is a decryption or to the start of the array if encryption.
	 */
	if (decflag)
		keyp = &pkey[30];
	else
		keyp = pkey;

	for (i = 0; i < 16; i++) {
		/*
		 * ebox expansion and sbox contraction are done at the same time.
		 */
		temp1 = (right >> 3 | right << 29) ^ keyp[0];
		temp = sbox_1[temp1 >> 24 & 0x3f];
		temp |= sbox_3[temp1 >> 16 & 0x3f];
		temp |= sbox_5[temp1 >> 8 & 0x3f];
		temp |= sbox_7[temp1 & 0x3f];

		temp1 = (right << 1 | right >> 31) ^ keyp[1];
		temp |= sbox_2[temp1 >> 24 & 0x3f];
		temp |= sbox_4[temp1 >> 16 & 0x3f];
		temp |= sbox_6[temp1 >> 8 & 0x3f];
		temp |= sbox_8[temp1 & 0x3f];

		/*
		 * old rd becomes new ld, old ld ^ text becomes new rd
		 */
		temp ^= left;
		left = right;
		right = temp;

		/*
		 * Move the per round key pointer to the next set of keys
		 */
		if (decflag)
			keyp -= 2;
		else
			keyp += 2;
	}

	sd2[0] = (unsigned char)(left >> 24 & 0xff);
	sd2[1] = (unsigned char)(left >> 16 & 0xff);
	sd2[2] = (unsigned char)(left >> 8 & 0xff);
	sd2[3] = (unsigned char)(left & 0xff);
	sd2[4] = (unsigned char)(right >> 24 & 0xff);
	sd2[5] = (unsigned char)(right >> 16 & 0xff);
	sd2[6] = (unsigned char)(right >> 8 & 0xff);
	sd2[7] = (unsigned char)(right & 0xff);

	pbits(target, (char *)sd2, iip, 8);
}

/*
 *	da_permute_key(key, pkey)
 *
 *	Key is a pointer to an 8 character array containing the 64 bit DES
 *	encryption key. Permute key and store in pkey
 *
 */

void
da_permute_key(char *key, unsigned long *key_p)
{
	unsigned char pkey[7];
	short i, j;
	short shift;
	unsigned long lkey, rkey;
	unsigned long ltmp, rtmp;

	/*
	 * Do the initial permutation on the key.  Note that everything is
	 * declared unsigned to prevent sign extension problems.
	 */
	pbits((char *)pkey, key, pch1, 7);

	/*
	 * Split the key into 28 bit left and right halves.  We use
	 * unsigned longs for the calculations.
	 */
	lkey = pkey[0];
	lkey = lkey << 8 | pkey[1];
	lkey = lkey << 8 | pkey[2];
	lkey = lkey << 4 | pkey[3] >> 4;

	rkey = pkey[3] & 0xf;
	rkey = rkey << 8 | pkey[4];
	rkey = rkey << 8 | pkey[5];
	rkey = rkey << 8 | pkey[6];

	/*
	 * Now rotate the key for each round.  
	 */
	for (i = 0; i < 16; i++) {
		shift = round_shift[i];
		lkey = (lkey << shift & 0xfffffff) | lkey >> (28 - shift);
		rkey = (rkey << shift & 0xfffffff) | rkey >> (28 - shift);

		/*
		 * The key permutation is done inline to save time.
		 */
		ltmp = rtmp = 0L;
		for (j = 0; j < 24; j++) {
			ltmp *= 2;
			if (lkey & left_pch2[j])
				ltmp |= 1;
			rtmp *= 2;
			if (rkey & right_pch2[j])
				rtmp |= 1;
		}

		*key_p++ = (ltmp << 6 & 0x3f000000) | (ltmp << 10 & 0x3f0000) |
			(rtmp >> 10 & 0x3f00) | (rtmp >> 6 & 0x3f);
		*key_p++ = (ltmp << 12 & 0x3f000000) | (ltmp << 16 & 0x3f0000) |
			(rtmp >> 4 & 0x3f00) | (rtmp & 0x3f);
	}
}

/*
 * da_des_encrypt()
 *
 * Given a pointer to an 8 char array, encrypt the array using the DES
 * key which has been previously permuted (by da_permute_key()) into pkey.
 * Return the encrypted data in the array ciphertext.
 * Note: Ciphertext and cleartext may be the same.  In this case, the 
 * cleartext will be overwritten by ciphertext.
 */

void
da_des_encrypt(char *ciphertext, char *cleartext, unsigned long *pkey)
{
	desfunction(ciphertext, cleartext, 0, pkey);
}

/*
 * da_des_decrypt()
 *
 * Given a pointer to an 8 char array of ciphertext, decrypt it using the
 * permuted DES key stored in pkey.  Return the decrypted data in the
 * array cleartext.
 * Note: Ciphertext and cleartext may be the same.  In this case, the 
 * ciphertext will be overwritten by cleartext.
 */

void
da_des_decrypt(char *cleartext, char *ciphertext, unsigned long *pkey)
{
	desfunction(cleartext, ciphertext, 1, pkey);
}

/*
 * da_str_encrypt()
 *
 * Given a pointer to an len byte array, encrypt the array using the
 * DES key which has been previously permuted (by da_permute_key()) into
 * pkey.  Return the encrypted data in the array ciphertext.
 * NOTE: Because of the nature of DES, len MUST be a multiple of eight.
 * While this is not enforced by error, any remainder bytes will be left
 * unencrypted.
 * Note: Ciphertext and cleartext may be the same.  In this case, the 
 * cleartext will be overwritten by ciphertext.
 */

void
da_str_encrypt(char *ciphertext, char *cleartext, int len, unsigned long *pkey)
{
	len /= 8;
	while (len--) {
		desfunction(ciphertext, cleartext, 0, pkey);
		ciphertext += 8;
		cleartext += 8;
	}
}

/*
 * da_str_decrypt()
 *
 *	original array with the plaintext.
 * Given a pointer to an len byte array, decrypt the array using the
 * DES key which has been previously permuted (by da_permute_key()) into
 * pkey.  Return the encrypted data in the array ciphertext.
 * NOTE: Because of the nature of DES, len MUST be a multiple of eight.
 * While this is not enforced by error, any remainder bytes will be
 * ignored.
 * Note: Ciphertext and cleartext may be the same.  In this case, the 
 * ciphertext will be overwritten by cleartext.
 */

void
da_str_decrypt(char *cleartext, char *ciphertext, int len, unsigned long *pkey)
{
	len /= 8;
	while (len--) {
		desfunction(cleartext, ciphertext, 1, pkey);
		ciphertext += 8;
		cleartext += 8;
	}
}
