/*
 * $Id: dsagent.h,v 1.3 1996/04/29 19:44:55 rwt Exp rwt $
 * copyright 1996 (c) Digital Pathways, Inc.  All Rights Reserved
 *
 * This source code is provided under license from Digital Pathways,
 * Inc.  It is provided "AS-IS", without warranty of any kind, either
 * express or implied. It may be incorporated wholly or modified as
 * needed, subject to the following restrictions.
 *
 * 1) This notice, including the Digital Pathways, Inc. notice of
 * copyright must be included in any source code or derivative works.
 * If this source code is modified by the licensee, a notice of
 * modification must be included as well.
 *
 * 2) This Program, and any derivative works, may not be distributed
 * is source form, on any media, without the express written consent
 * of Digital Pathways, Inc.  There is no restriction on object or
 * executable code.
 */

#define AGENT_DESC	"DPI agent library"	/* Ascii description of agent */
#define AGENT_VER	"v1.00"		/* Ascii agent version number */
#define AUTH_LEVEL	2		/* Desired Agent/Server Authentication Level */
	/* 0 - Generic ID (does not require AgentID) Server may impose Limits */
	/* 1 - ID only (Does not require DES.C or AgentKey in configStr) */
	/* 2 - ID+dualSNK (requires DES.C, may have export problems) */
	/* 3 - ID+dualSNK+encryptedLink (requires DES.C) */

/* these values must match equivalents in Security Server - do not change */

#define USER_ID_LEN			10
#define AGENT_ID_LEN		18
#define AGENT_KEY_LEN		8
#define AGENT_DESC_LEN		34
#define AGENT_VER_LEN		34
#define SERVER_ADDR_LEN 	42

/*
 * commandFromAgent
 *
 * These codes represent events/requests from the Agent
 */
#define SAW_DCD 		0x40	/* Agent reports start of user session */
#define CHECK_ID		0x41	/* Agent reports start of user session */
								/* with pre-known user ID */
#define USER_INPUT		0x42	/* Agent reports User Input */
#define END_CALL		0x43	/* Agent reports end of User Session */
#define AGENT_ID		0x44	/* Agent authentication ID */
#define AGENT_RESPONSE	0x45	/* Agent Authentication Response */

/*
 * commandFromServer
 *
 * These codes represent primary commands from the Server to the Agent
 * Some commands (such as GET_INPUT) are used in multiple states of the
 * security dialog
 */
#define NO_COMMAND			0x50	/* server requests no action */
#define GET_INPUT			0x51	/* server wants User Session input */
#define SERVER_CHALLENGE	0x52	/* server challenges agent */
#define SERVER_APPROVAL 	0x53	/* Server approves agent */
#define SERVER_REJECTION	0x54	/* server rejects agent */
#define AGENT_ALIVE			0x55	/* Server requests Agent Information */
#define AGENT_STATUS		0x56	/* Server requests Agent channel summary */
#define SESSION_STATUS		0x57	/* Server requests Agent channel Detail */
#define SESSION_CLEAR		0x58	/* Server requests Agent to kill session */
#define PROBE_READY			0x59	/* Server has built probe file (one of N) */
#define PROBE_FAILED		0x5A	/* Requested Probe failed */
/* also ACCESS_APPROVED
 *  and ACCESS_DENIED
 */

/*
 * responseFromServer
 *
 * note: entries must match order of server text table
 * This table is used by the server to report each state of the security dialog
 * It also identifies specific prompts to be sent to the end-user.
 */

#define AGENT_NOT_LOGGED_ON			0
#define ASK_FOR_AGENT_RESP			1
#define GIVE_SERVER_APPROVAL		2

#define ASK_FOR_ID					3
#define ASK_FOR_SNK_RESP			4
#define UNKNOWN_USER				5
#define INVALID_SNK					6
#define ACCESS_APPROVED 			7
#define ACCESS_DENIED				8
#define SERVER_BUSY					9
#define UNKNOWN_AGENT_PACKET		10
#define GOT_DCD_ON_ACTIVE_CHANNEL	11
#define GOT_CHECK_ID_AT_BAD_TIME	12
#define GOT_UNEXPECTED_INPUT		13
#define ASK_FOR_PW					14

#define INVALID_ID_PW				15
#define FORCE_PW_ASK_FOR_PW1_CHANGE	16
#define ASK_FOR_PW1_CHANGE			17
#define ASK_FOR_PW2_CHANGE			18
#define INVALID_PW_CHANGE			19
#define BAD_TIMECLASS_DENIED		20
#define BAD_DATECLASS_DENIED		21
#define OLD_PW_ENTERED				22
#define NEW_PW_BAD_LENGTH			23
#define NEW_PW_NON_NUMERIC			24
#define ASK_FOR_OLD_PASSWORD		25
#define BAD_VIOL_COUNT_DENIED		26
#define USR_ALREADY_ACTIVE_DENIED	27

#define MAX_PROMPTS					28	/* one more than last entry in table */

/* Note: Server may provide additional codes here */

/*
 * Note:
 *		The Dialog packet definition was replaced with a series of
 * lengths and byte offsets.  The assumption that the dialog packet
 * could be represented by a universal structure on a wide variety
 * of machine architectures was chucked.  There are additional 
 * security implications of utilizing a single buffer for state and
 * dialog that were troubling.  The structure definition is left for
 * reference, but is not used.
 */

#define PACKET_SIZE	372	/* Size of the dialog packet in bytes */
/*
 * Byte offsets into the dialog packet
 */
#define CALLSLOT		0	/* DSS tag */
#define SESSION_ID		2	/* Agent local session identifier */
#define TID				4	/* DSS assigned User login session id */
#define AGENT_CMD		8	/* Agent cmd to DSS */
#define DSS_CMD			10	/* DSS cmd to agent */
#define DSS_RESPONSE	12	/* DSS response to agent query */
#define LINE_CTRL		14	/* DSS echo option request */
#define CALL_COMPLETE	16	/* Dialog complete, end of dialog indicator */
#define CHALLENGE		18	/* 10 byte challenge field */

/*
 * The following fields share the same space in the Prompt field
 */
#define PROMPT_FIELD		28	/* Prompt to user from DSS */
#define INFO_FIELD			28	/* User info on completion of authent. */
#define OPTION_FIELD		28	/* agent options set by DSS */
#define USER_PROMPT_LEN 	128	/* Length of this field */

/*
 * Byte offsets into the info field - the original structure definition
 * is included for reference.
 *
 *struct informationStr {
 *	short	maxChans;		// number of channels
 *	short	reportEndCalls;	// TRUE if Host can report END_CALL events
 *	short	primaryServer;	// TRUE if Agent feels Server is primary, not backup
 *	short	future[18];		// reserved
 *	char	hostDesc[AGENT_DESC_LEN+2];		// Ascii desciption of agent
 *	char	hostVersion[AGENT_VER_LEN+2];	// Ascii version info of agent
 *};
 */

#define IF_MAXCHANS		0	/* maximum number of simultaneous users */
#define IF_ENDCALLS		2 	/* TRUE if Host can report END_CALL events */
#define IF_PRIMARY		4	/* TRUE if Server is primary, not backup */
#define IF_AGENTDESC	24	/* description of agent */
#define IF_AGENTVERS	IF_AGENTDESC + AGENT_DESC_LEN	/* version of agent */

/*
 * Byte offsets into the options field - the original structure definition
 * is included for reference.
 *
 *#define MAX_OPTIONS 32
 *
 *struct serverOptionsStr {
 *	short serverType;			// Security Server type
 *	short serverVersion;		// Security server version
 *	short opt[MAX_OPTIONS];
 *};
 */

#define OF_SVR_TYPE		0	/* Security Server type */
#define OF_SVR_VERS		2 	/* Security Server version */
#define OF_OPTIONS		4	/* Options list */

/*
 * The following fields share the same space in the User Input field
 */
#define INPUT_FIELD			156	/* User input to DSS */
#define AUTH_FIELD			156	/* Agent authentication field */
#define INPUT_LEN			80	/* Length of this field */
/*
 * user record, returned on successful ACCESS_APPROVED in the input
 * field.
 *
 * offset 0  = char userID[10]
 * offset 10 = char info1[10]
 * offset 20 = char info2[10]
 * offset 30 = char phone1[20]
 * offset 50 = phone2[20]
 * offset 70 = 0x00 terminator
 */

/*
 * The link field has NO useful function, except that it is, for
 * Historical Purposes, always set to 2.  On big endian machines, this
 * may be incorrectly set to 0x2000000, because it is not used and hence
 * not debugged.  This is the ONE invariant identifier in this packet.
 */
#define LINK_FIELD		INPUT_FIELD + INPUT_LEN
#define LINK_VALUE		2

/*
 * Byte offsets into the authentication field - the original structure
 * definition is included for reference.
 *
 *struct authenticationStr {
 *	short state;		// 0-ID, 1-Challenge, 2-Response, 3-Approval, 4-Rejected
 *	short level;		// Authentication level Agent is asking for
 *	unsigned long challenge;	// Binary SNK challenge
 *	unsigned long response; 	// Binary SNK response
 *	unsigned long sessionKey[2];	// encryption key for level 3 authent'd session
 *	char agentID[AGENT_ID_LEN+2];
 *	short future[16];		// reserved for future
 *};
 */

#define DP_AF_STATE		0	/* Phase of the authentication process */
#define DP_AF_LEVEL		2 	/* Authentication level requested by agent */
#define DP_AF_CHALLENGE	4	/* Binary SNK challenge */
#define DP_AF_RESPONSE		8	/* Binary SNK response */
#define DP_AF_KEY			12	/* level 3 encryption key */
#define DP_AF_AGENTID		DP_AF_KEY + 8		/* Agent ID */

/*
 * Internal representation of the Agent-Server dialog
 */

typedef struct dialog {
	short callslot;			/* Server tag identifier */
	long tid;				/* Transaction ID assigned by the server */
	short session_id;		/* Session ID */
	short cmd_from_agent;	/* Command to server from agent */
	short cmd_from_server;	/* Command to agent from server */
	short server_response;	/* Response code from server */
	short line_flags;		/* passed through to user input function */
	short call_complete;	/* marks end of transaction dialog*/
	char *prompt;			/* pointer to the prompt field */
	char *input;			/* pointer to the user input field */
	char packet[372];		/* raw 372 byte packet */
} dialog_t;

/*
 * macro to convert from an input buffer pointer to a dialog
 */
#define itod(p)		(dialog_t *)((char *)p - \
						&((dialog_t *)NULL)->packet[INPUT_FIELD])

/*
 * This is the original dialog packet definition.  It is included for
 * reference only, and should not be used.  The 372 byte packet passed
 * between is defined by the byte packed, little endian representation
 * of this structure.
 *
 * struct dialogPacketStr {
 *	short callSlot; // slot of callTable in use for this dialog (-1 if unknown)
 *	short channelID;	// channel # 0 to MAX_CHAN-1
 *	long transactionNumber;	// provided by Server
 *	short commandFromAgent; // from Agent:
 *			// SAW_DCD	Agent sees new call
 *			// CHECK_ID	Agent already has ID
 *			// USER_INPUT	Agent providing requested line from user
 *	short commandFromServer;	// From Server
 *			// GET_INPUT	please send prompt and then collect line from user
 *			// NO_COMMAND  end of packet exchange, no action needed
 *			// or one of the response codes
 *	short responseFromServer;	// From Server 
 *			// ASK_FOR_ID 
 *			// ASK_FOR_SNK_RESP
 *			// UNKNOWN_USER
 *			// ACCESS_APPROVED
 *			// ACCESS_DENIED
 *			// SERVER_BUSY
 *			// UNKNOWN_AGENT_PACKET
 *			// GOT_DCD_ON_ACTIVE_CHANNEL
 *			// GOT_CHECK_ID_AT_BAD_TIME
 *			// GOT_UNEXPECTED_INPUT
 *	short echo;	* when collecting user input
 *			// 1 = echo as typed
 *			// 0 = no echo at all
 *			// -1 = echo asterisks
 *	short callComplete;		// TRUE if call record may be released now
 *	char challenge[10];		// SNK challenge + null
 *	char prompt[USER_PROMPT_LEN];	// prompt to send to user
 *	char input[INPUT_LEN];		// data to collect from user
 *	void *link; 		// link to next buffer in list
 *	short daReq;		// Last Agent Request
 *	short padding;		// to regain DWORD alignment
 *	short future[64];	// reserve for future
 *};
 */

/*
 * user session structure
 * This is malloc'ed and used on a per authentication basis.
 */

typedef struct session {
	short session_id;		/* agent assigned session identifier */
	short callslot;			/* Server authentication slot */
	long transaction_id;	/* Server transaction number */
	struct in_addr in_addr;	/* Internet Address of remote host */
	struct dialog_io callback;	/* callback function table */
	char userid[USER_ID_LEN]; 	/* user id string */
	char *user;					/* user id string for returns */
	struct session *next;		/* link list */
} session_t;

/*
 * Authentication Structure is used to pass information between
 * Agent and Security Server during initial connection.
 * (host may ignore)
 */

#define DA_AS_ID			0	/* Authentication State - Agent ID */
#define DA_AS_CHALLENGE 	1	/* Authentication State - Server Challenge */
#define DA_AS_RESPONSE		2	/* Authentication State - Agent Response */
#define DA_AS_APPROVAL		3	/* Authentication State - Server Approval */
#define DA_AS_REJECTED		4	/* Authentication State - Server Rejected Agent */

#define DA_AL_GENERIC		0	/* authentication level - Generic ID */
#define DA_AL_ID			1	/* authentication level - Agent ID */
#define DA_AL_SNK			2	/* authentication level - Agent ID + dual binary SNK */
#define DA_AL_ENC			3	/* authentication level - Agent ID + dual binary SNK + session encryption */

typedef struct auth {
	short state;	/* 0-ID, 1-Challenge, 2-Response, 3-Approval, 4-Rejected */
	short level;	/* Authentication level Agent is asking for */
	unsigned long challenge;	/* Binary SNK challenge */
	unsigned long response; 	/* Binary SNK response */
	unsigned long session_key[2];	/* encryption key for L3 auth session */
} auth_t;

/*
 * Option structure is loaded by Server and micromanages Agent
 * (host may ignore)
 */

#define MAX_OPTIONS 32

typedef struct options {
	short serverType;				/* Security Server type */
	short serverVersion;			/* Security server version */
	short opt[MAX_OPTIONS];
} options_t;

/*
 * Configuration (common for all channels under control of agent
 */

typedef struct agent_cfg {
	int server_fd;	/* File desc. of the server socket */
	struct da_config da_cfg;	/* copy of the user configuration data */
	struct sockaddr sa;			/* Address of remote host */
	struct options options;
	int active_sessions;		/* Count of active sessions */
	struct session *sessions;	/* List of active sessions */
} cfg_t;

/*
 * Generic definitions
 */
#ifndef ERROR
#define ERROR		-1
#endif

#ifndef SUCCESS
#define SUCCESS		0
#endif
