/*
 * $Id: dslib.h,v 1.3 1996/04/29 19:44:42 rwt Exp rwt $
 * copyright 1996 (c) Digital Pathways, Inc.  All Rights Reserved
 *
 * This source code is provided under license from Digital Pathways,
 * Inc.  It is provided "AS-IS", without warranty of any kind, either
 * express or implied. It may be incorporated wholly or modified as
 * needed, subject to the following restrictions.
 *
 * 1) This notice, including the Digital Pathways, Inc. notice of
 * copyright must be included in any source code or derivative works.
 * If this source code is modified by the licensee, a notice of
 * modification must be included as well.
 *
 * 2) This Program, and any derivative works, may not be distributed
 * is source form, on any media, without the express written consent
 * of Digital Pathways, Inc.  There is no restriction on object or
 * executable code.
 */

#define ID_LEN		16
#define KEY_LEN		8

extern int da_errno;

/*
 * Error codes returned in da_errno
 */
#define DA_ALLOC_FAILED				0x7001
#define DA_SESSION_LIMIT			0x7002
#define DA_NO_CONFIGURE				0x7003
#define DA_INVALID_PARAMETER		0x7004
#define DA_TRANSMISSION_ERR			0x7020
#define DA_RECEIVE_ERR				0x7021
#define DA_SERVER_TIMEOUT			0x7022
#define DA_LOST_USER_CONNECTION		0x7040

/*
 * Configuration (common for all channels under control of agent
 */

#define PRIMARY_SERVER		1	/* Indicates agent considers DSS a primary */
#define REPORT_END_OF_CALL	2	/* Indicates agent can report end of call */

/*
 * Line control values
 */

#define ECHO_NONE	0
#define ECHO_NORMAL	1
#define ECHO_STARS	-1

typedef struct da_config {
	unsigned int flags;		/* bit 0		1 if sockaddr is Primary server */
							/* bit 1		End of Call Supported */
							/* bits 2-31	Reserved */
	int max_sessions;		/* Max. numbers of user sessions */
	int timeout;			/* Max. seconds to wait on an I/O request */
	char agent_id[ID_LEN];		/* ascii ID with which to log on to Server */
	unsigned long pkeys[32]; 	/* permuted keys for each of the 16 rounds */
} da_config_t;

typedef struct dialog_io {
	int (*read)(int, char *, int, int, int);
	int (*write)(int, char *, int);
	int (*notify)(int, int, long, void *);
	int (*error)(int, int);
} dialog_io_t;

/*
 * Function Prototypes
 */

int da_wait_server(fd_set *);
int da_user_input(char *, int);
int da_process_packet(void);
int da_close_session(int);
int da_open_session(char *, struct dialog_io *);
int da_end_call(char *, long);
int da_authenticate(char *, struct dialog_io *, long *, void *);
void da_disconnect(void);
int da_connect(struct sockaddr *, struct da_config *);

void da_permute_key(char *, unsigned long *);
void da_des_encrypt(char *, char *, unsigned long *);
void da_des_decrypt(char *, char *, unsigned long *);
void da_str_encrypt(char *, char *, int, unsigned long *);
void da_str_decrypt(char *, char *, int, unsigned long *);
