/*============================================================================
 * file:	../dss/dssapi/dssapi.c
 * 
 * copyright 1996 (c) Digital Pathways, Inc.  All Rights Reserved 
 *
 * copyright 1997-2000 (c) Freemont Avenue Software, Inc.  All Rights Reserved
 *
 * This source code is provided under license from Digital Pathways,
 * Inc.  It is provided "AS-IS", without warranty of any kind, either
 * express or implied. It may be incorporated wholly or modified as
 * needed, subject to the following restrictions.
 *
 * 1) This notice, including the Digital Pathways, Inc. notice of
 * copyright must be included in any source code or derivative works.
 * If this source code is modified by the licensee, a notice of
 * modification must be included as well.
 *
 * 2) This Program, and any derivative works, may not be distributed
 * is source form, on any media, without the express written consent
 * of Digital Pathways, Inc.  There is no restriction on object or
 * executable code.  
 *
 *----------------------------------------------------------------------------
 * Modifications made by RJL at Freemont Avenue Software.
 *
 * last modified:       November 20, 1996 - added include for fd_set
 *                      January   3, 1997 - modified myread and mywrite to
 *                                          work with tnproxy.
 *
 *
 *
 *===========================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <string.h> 
#include <syslog.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>
#include <sys/time.h>
#include <stdarg.h>	
#include <fcntl.h>
#include <sys/stat.h>  
#ifndef __hppa
#include 	<sys/select.h>		/* to support fd_set 	RJL	  */
#endif 
#include "dslib.h" 
#include "firewall.h"
#include "gwuser.h"

static char rcsid[] = "$Id$";

char	*key;
char	*agent_id;
char	*dss_addr;
int	dss_port    = 2626;
int 	dss_timeout =   60;

#define RLSIZE		128

void cfg_key(void *, char *);
void cfg_str(void *, char *);
void cfg_int(void *, char *);

struct parm {
	char *string;
	void (*func)(void *, char *);
	void *target;
} parm[] = {
	{"agentkey", cfg_key, (void *)&key},
	{"agentid", cfg_str, (void *)&agent_id},
	{"dss_address", cfg_str, (void *)&dss_addr},
	{"dss_port", cfg_int, (void *)&dss_port},
	{"dss_timeout", cfg_int, (void *)&dss_timeout},
	{NULL, NULL, NULL},
};

char *filename = FW_CFG_DIR "/dssagent.conf";    
/*============================================================================
 * function prototypes
 *===========================================================================*/
void cfg_key(void *p, char *ascii_key);
void cfg_str(void *p, char *string);
void cfg_int(void *p, char *string);
int  dss_config(void); 
int  getline(int fd, unsigned char *buf, int siz); 
int  myread(int session_id, char *buf, int len, int state, int flags);
int  mywrite(int session_id, char *buf, int len);
int  dssapi(struct GWUSER *gwup);

/*============================================================================
 * Convert 8 sets of ascii numbers into integers
 *===========================================================================*/
void cfg_key(void *p, char *ascii_key)
{
   char **key = p;
   char *tp;
   int 	i;

   tp = strtok(ascii_key, " \t=,");
   *key = malloc(8);
   memset(*key, 0, 8);

   for (i = 0; i < 8 && tp; i++) 
   {
      (*key)[i] = (char)strtol(tp, NULL, 0);   /* convert string to integer */
      tp = strtok(NULL, " \t,");
   }
}
/*==========================================================================
 *
 *==========================================================================*/
void cfg_str(void *p, char *string)
{
   char **target = p;
   char *tp;

   if ((tp = strtok(string, " \t=\n"))) 
   {
      *target = malloc(strlen(tp) + 1);
      strcpy(*target, tp);
   }
}
/*==========================================================================
 *
 *==========================================================================*/
void cfg_int(void *p, char *string)
{
   int *target = p;
   char *tp;

   if ((tp = strtok(string, " \t=\n")))
      *target = atoi(tp);
}
/*==========================================================================
 * Process the DSS configuration file.   Look for keyword found in parm[] and
 * call assocaited function.
 *========================================================================*/
int dss_config(void)
{
   char *cp;
   struct parm *p;
   char line[RLSIZE];
   FILE *f;

   /*====================================================
    * Open the configuration file and check for errors
    *===================================================*/
   if( (f = fopen(filename, "r")) == NULL ) 
   {   
      if( errno == ENOENT)
      {
         return(errno);
      }
      syslog(LOG_ERR,"Error: dssagent could not open %s\n", filename);
      return(-1);
   }

   while (fgets(line, RLSIZE, f)) 
   {
      /*
       * first, drop and comment fields
       */
      cp = strchr(line, '#');
      if (cp)
         *cp = '\0';

      /*
       * Now look for a token
       */
      cp = strtok(line, " \t=");
      if (cp) 
      {
         for (p = parm; p->string; p++) 
         {
	    if (!strcmp(cp, p->string))
	       break;
	 }
	 if (p->string)
	    (*p->func)(p->target, strtok(NULL, "\n"));
      }
   }
   return(0);
}
/*=========================================================================== 
 *
 *==========================================================================*/
int myread(int session_id, char *buf, int len, int state, int flags)
{  
static char	nprompt[] = "Username: ";  
   char *cp;
   char newbuf[128];
   if(getline(0,(unsigned char *)buf,len)
          < 0 )
   {  
      syslog(LOG_ERR,"Error: while reading user name.");
      return(-3);    
   }  

   cp = strstr(buf,"PASS");
   if(cp)
   {
      strncpy(newbuf,buf+5,sizeof(newbuf)-1);
      strcpy(buf,newbuf);
   }
   return(0); 
}
/*=========================================================================== 
 *
 *=========================================================================*/ 
int mywrite(int session_id, char *buf, int len)
{  
   char *cp;
   struct termios t; 
   char newbuf[160];
   strcpy(newbuf,"331 ");
   cp = strstr(buf,"SNK");
   if(cp)
   {
      cp = strstr(cp,"Challenge");
      if(cp)
      {
         strcat(newbuf,cp);
         cp = strchr(newbuf,'\n'); 
         cp++;
         *cp = '\0';
         len = strlen(newbuf);
      }
      else
         strcpy(newbuf,buf);
   } 
   else
      strncpy(newbuf,buf,158);
   if((write(0,newbuf,len)) != len) 
   {  
      syslog(LOG_ERR,"Error: while writing '%s'.",buf);
      return(-2);
   }  
   return(len);
}                                                                           
/*========================================================================== 
 * Initialize callback with addresses of myread and mywrite routines. 
 * Pointers to notify and error are NULL.
 *=========================================================================*/
dialog_io_t callback = {myread, mywrite, NULL, NULL};
/*===========================================================================
 * dss_API  
 *
 * inputs: 
 *	none
 * outputs:
 *	 0 if user is authenticated. 
 *      -1 if user is disallowed  
 *       2 if dssagent.conf file does not exist
 *==========================================================================*/

int dssapi(struct GWUSER *gwup)
{
   da_config_t cfg;             	/* defined in in dslib.h	*/
   struct sockaddr_in sin;
   long tid;
   int 	fd;    
   int	rc;
   int	result;
   char *user; 
   /* Get configuration data and check for errors	*/
   if( (rc = dss_config()) == ENOENT)
   {
      return(rc+1000);  
   }
   if (!dss_addr || !agent_id || !key) 
   {
      syslog(LOG_ERR,"Error: dssagent configuration error.");
      return(-1);
   }
   /*==================================================== 
    * Initialize da_config_t and sockaddr_in structures	
    *===================================================*/
   memset(&cfg, 0, sizeof(da_config_t));
   memset(&sin, 0, sizeof(struct sockaddr_in));
   sin.sin_family      = AF_INET;
   sin.sin_addr.s_addr =  inet_addr(dss_addr);
   sin.sin_port        = htons(dss_port);
   cfg.flags           = PRIMARY_SERVER;
   cfg.max_sessions    = 2;
   cfg.timeout         = 10;
   strcpy(cfg.agent_id, agent_id);
   /* permute key -> cfg.pkeys	*/ 	
   da_permute_key(key, cfg.pkeys);
   /* Connect to DSS server	*/	
   fd = da_connect((struct sockaddr *)&sin, &cfg);
   if (fd >= 0) 
   {
      user = (char *) &gwup->username;
      result = da_authenticate(user, &callback, &tid, NULL);
      da_disconnect();  
     
      return(result);
   }
   else         
   {
      syslog(LOG_ERR,"Error: Failed to connect to DSS server, da_errno = %d, "
                     "errno = %d\n",
			da_errno, errno);  
      return(-1);
   } 
}
