/*============================================================================
 * filename:		../ftproxy/ftproxconf.c					
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  
 * All rights reserved.
 *
 *		
 * Date written:	April	  10, 1995
 *			July      10, 1996 - added support for anonymous ftp
 * Last modified:	February   8, 2000
 *	                         
 * function:                       
 *	The ftproxconf program reads the file transfer proxy configuration file
 *	to extract configuration information.
 *
 *  
 *
 * file format:           
 * 
 *   Manditory Commands
 *
 *	groupid	  	= groupname
 *	timeout 	= nnn 		- number of seconds 
 * 	tnproxpath 	= pathname
 *	userid  	= name		-                    
 *   
 *   Optional Commands
 *      
 *	anonymousftp    = x.x.x.x       - IP address of anonymous FTP server
 *	auth		= yes|no	- Is user authentication required for
 *					  protected clients?  
 *	buffersize      = nnnnn         - default is DBUFSIZ
 *	debug		= yes|no        - default is no debug 
 *	external	= yes|no	- default is no external ftp users 
 *  	
 *
 *   The following rules are ingnored, as they will be processed by another
 *   function.
 * 	
 * 	permit users = usrlist groups = grplist from src_addr to dest_addr  
 *	deny   users = usrlist groups = grplist from src_addr to dest_addr
 *
 *============================================================================*/
#include	<sys/types.h>
#include	<stdio.h>
#include	<syslog.h>
#include	<string.h>
#include	<sys/time.h> 
#include	<pwd.h>
#include	<grp.h> 
#include	<netdb.h>               /* defines servent structure      */
#include	<netinet/in.h>  
#include	"firewall.h"  
#include	"ftproxy.h" 
/*=============================================================================
 * input:	pointer to TNPROXY structure
 * outputs:	updated TNPROXY structure
 *              return  0 if OK
 *		return -1 if error
 *===========================================================================*/
int ftproxconf(struct FTPROXY *proxptr)
{   
   #define	MAXLEN 1024
   char 	ftproxcfgfile[] = FTPROXCFGFILE;
   char		record[MAXLEN];
   FILE 	*cfgfp; 
   char 	*cp; 
   char		*cp2;
   char		*token1;
   char		*token2; 
   char		*token3; 
   char		*token4;
   char		*token5; 
   char		*token6; 
   char		*token7;
   char		tokensep[]   = "\t,:;= \n";
   int		hostfound    = 0;
   int		i;
   int		line	     = 0;  
   int		userfound    = 0; 
   int		groupfound   = 0;
   int		pathfound    = 0;
   int		timeoutfound = 0;
   struct servent	*sp; 		/* ptr to service entry 	*/ 
   struct passwd	*pw; 
   struct group		*gp; 
   if( proxptr == NULL)
   {
      syslog(LOG_ERR,"Error: ftproxconf: NULL proxptr.");
      return(-1);    			
   }
   /*=====================================================
    * Initialize FTPROXY record	
    *====================================================*/ 
   memset(proxptr->username,0,sizeof(proxptr->username));
   proxptr->userid	= -1;
   memset(proxptr->groupname,0,sizeof(proxptr->groupname));
   proxptr->groupid	= -1; 
   proxptr->auth        = AUTHREQ;      /* user authorization required	*/ 
   proxptr->x4ward	= X4WARDNO;	/* default is NO xforwarding */
   proxptr->ftuserid  	= -1; 
   proxptr->ftgroupid 	= -1;
   proxptr->timeout   	= 60;   	/* 60 second default		*/
   memset(proxptr->sourceIP,0,sizeof(proxptr->sourceIP));
   memset(proxptr->destIP,0,sizeof(proxptr->destIP));
   memset(proxptr->anonymousftp,0,sizeof(proxptr->anonymousftp));
   proxptr->debug 	= 0;		/* default is no debug		*/ 
   proxptr->dbuffersize = DBUFSIZ;
   proxptr->external	= NOEXFTP;	/* default is no external ftp users*/
   /*===================================
    * open ftproxy conf file 
    *===================================*/
   if (!(cfgfp = fopen(ftproxcfgfile, "r")))
   {
      syslog(LOG_ERR,"Error: open failed for %s: %m",ftproxcfgfile);
      return(-1);    			/* open failed signal error */
   } 
   cp = (char *) &record;
   /*===================================================== 
    * Get commands from ftproxy config file. 
    *====================================================*/        
   while(   fgets(record, MAXLEN, cfgfp) != NULL) 
   {
      line++;   
      /* skip over comments and null lines */
      if( (*cp == '#') || (*cp == '\n') ) goto getnextline1; 
      if( (token1 = strtok(record,tokensep)) == NULL) goto getnextline1; 
      /*=========================================================	
       * If anonymousftp record then get IP address and validate 
       *========================================================*/	
      if( strcmp(token1,"anonymousftp") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for anonymousftp on line"
			   " %d of %s.",
 		line,ftproxcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
         if( strlen(token2) < 5 || strlen(token2) > 255)
         {
            syslog(LOG_ERR,"Error: invalid name or IP address '%s' specified "
		"for anonymousftp on line %d of %s.",
  		token2, line, ftproxcfgfile);
  	    fclose(cfgfp);
	    return(-1); 
         }
         strcpy(proxptr->anonymousftp,token2);
         goto getnextline1;  
      } 
      /*=========================================================	
       * If authorization record then get auth value and validate 
       *========================================================*/	
      if( strcmp(token1,"auth") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for auth on line %d " 
			   "of %s.",
 			   line,ftproxcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if((strcasecmp(token2,"yes") != 0) && (strcasecmp(token2,"no") != 0))
	 { 
            syslog(LOG_ERR,"Error: invalid auth value '%s' on line %d of %s."
			   " Must be 'yes' or 'no'.",
 		token2, line, ftproxcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         }
         if ( strcasecmp(token2,"no") == 0)
         {  
            proxptr->auth = NOAUTHREQ;
         } 
    
         goto getnextline1;  
      }
      /*=========================================================	
       * If buffersize record then get buffer value and validate 
       *========================================================*/	
      if( strcmp(token1,"buffersize") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for buffersize on line %d " 
			   "of %s.",
 			   line,ftproxcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if( !isalldigits(token2) )
	 { 
            syslog(LOG_ERR,"Error: non numeric value '%s' specified for " 
			   "buffersize on line %d of %s.",
 			   token2, line, ftproxcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         } 
	 proxptr->dbuffersize = atoi(token2);
	 if( proxptr->dbuffersize < MINDBUF || proxptr->dbuffersize > MAXDBUF)
	 { 
            syslog(LOG_ERR,"Error: invalid buffersize(%d), must be >= %d and"
	    		   " <= %d", 
			   proxptr->dbuffersize, MINDBUF, MAXDBUF);
	    fclose(cfgfp);
	    return(-1);       
         }
         if ( strcasecmp(token2,"no") == 0)
         {  
            proxptr->auth = NOAUTHREQ;
         } 
    
         goto getnextline1;  
      }
      /*=========================================================	
       * If debug record then get debug value and validate 
       *========================================================*/	
      if( strcmp(token1,"debug") == 0 )
      {   
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for debug on line %d of"
			   " %s.",
 			   line,ftproxcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         } 
         if((strcasecmp(token2,"yes") != 0) && (strcasecmp(token2,"no") != 0))
	 { 
            syslog(LOG_ERR,"Error: invalid debug value '%s' on line %d of %s."
			   " Must be 'yes' or 'no'.",
 			   token2, line, ftproxcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         }
         if ( strcasecmp(token2,"no") == 0)
         {  
            proxptr->debug = 0;
         }
         else
         {
	    proxptr->debug = 1; 
         }
         goto getnextline1;  
      } 
      /*=========================================================	
       * If external record then get external value and validate 
       *========================================================*/	
      if( strcmp(token1,"external") == 0 )
      {   
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for external on line %d"
			   " of %s.",
 			   line,ftproxcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         } 
         if((strcasecmp(token2,"yes") != 0) && (strcasecmp(token2,"no") != 0))
	 { 
            syslog(LOG_ERR,"Error: invalid external value '%s' on line %d "
			   "of %s. Must be 'yes' or 'no'.",
 			   token2, line, ftproxcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         }
         if ( strcasecmp(token2,"no") == 0)
         {  
            proxptr->external = NOEXFTP;
         }
         else
         {
	    proxptr->external = EXFTP; 
         }
         goto getnextline1;  
      }
      /*=========================================================	
       * If timeout record then get timeout value and validate 
       *========================================================*/	
      if( strcmp(token1,"timeout") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for timeout on line %d"
			   "of %s.",
	  		   line,ftproxcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if( !isalldigits(token2) )
	 { 
            syslog(LOG_ERR,"Error: non numeric value '%s' specified for " 
			   "timeout on line %d of %s.",
 			   token2, line, ftproxcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         } 
         proxptr->timeout = atoi(token2); 	/* save timeout value	*/ 
         timeoutfound++;
         goto getnextline1;  
      }
      /*=========================================================	
       * If ftproxpath record then get path name and validate 
       * do not allow chroot to: "/" or "/etc" or "/usr" 
       *========================================================*/	
      if( strcmp(token1,"ftproxpath") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for ftproxpath on line"
			   " %d of %s.",
 			   line, ftproxcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if( strlen(token2) > 255 )
	 { 
            syslog(LOG_ERR,"Error: ftproxpath name '%s' is too long on line"
			   " %d of %s.",
 			   token2, line, ftproxcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         }
         if( ( strcmp(token2,"/")    == 0 ) ||
             ( strcmp(token2,"/etc") == 0 ) ||
	     ( strcmp(token2,"/usr") == 0 )   ) 
         {
            syslog(LOG_ERR,"Error: ftproxpath name '%s' on line %d of %s is" 
			   "not allowed.",
 			   token2, line, ftproxcfgfile); 
	    fclose(cfgfp);
	    return(-1);    
         }
         strcpy(proxptr->ftproxpath,token2); 	/* copy ftproxpath     	*/
         pathfound++;
         goto getnextline1;  
      } 

      /*=========================================================	
       * If userid record then get name and validate 
       *========================================================*/	
      if( strcmp(token1,"userid") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: No value specified for userid on line %d"
			   " of %s.",
 			   line, ftproxcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if( strlen(token2) > 8 )
	 { 
            syslog(LOG_ERR,"Error: User name '%s' is too long on line %d"
			   " of %s.",
 			   token2, line, ftproxcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         }
        
	 if(isalldigits(token2))
	 {
	    proxptr->ftuserid = atoi(token2); 
         }
         else
         {
 	    if((pw = getpwnam(token2)) == (struct passwd *)0)
            {
               syslog(LOG_ERR,"Error: user name '%s' on line %d of %s is not"
			      " valid user.",
 	       		      token2, line, ftproxcfgfile); 
	       fclose(cfgfp);
	       return(-1);
            } 
	    proxptr->ftuserid = pw->pw_uid; 		/* save user id	*/ 
         }
         if( proxptr->ftuserid == 0)
         {
            syslog(LOG_ERR,"Error: user name '%s' on line %d of %s is rejected"
			   " because user id = %d.",
 			   token2, line, ftproxcfgfile, proxptr->ftuserid); 
	    fclose(cfgfp);
	    return(-1);    
         }
         userfound++;
         goto getnextline1;  
      } 
      /*=========================================================	
       * If groupid record then get name and validate 
       *========================================================*/	
      if( strcmp(token1,"groupid") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for groupid on line %d"
			   " of %s.",
 			   line, ftproxcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if( strlen(token2) > 8 )
	 { 
            syslog(LOG_ERR,"Error: Group name '%s' is too long on line %d of %s.",
 		token2, line, ftproxcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         }
         if(isalldigits(token2))
         {
	    proxptr->ftgroupid = atoi(token2); 
         }
         else
         {
	    if((gp = getgrnam(token2)) == (struct group *)0)
            {
	       syslog(LOG_ERR,"Error: group name '%s' on line %d of %s is not"
			      " valid group.",
 	       token2, line, ftproxcfgfile); 
	       fclose(cfgfp);
	       return(-1); 
	    }
	    proxptr->ftgroupid = gp->gr_gid;	/* copy groupid  */ 
         }
	 strcpy(proxptr->groupname,token2);
         groupfound++;
         goto getnextline1;  
      } 
      /*=========================================================	
       * See if it is an invalid record
       *========================================================*/	
      if (strcmp(token1,"deny") && strcmp(token1,"permit"))
      {
         syslog(LOG_ERR,"Error: Invalid record on line %d of %s",
		line, ftproxcfgfile);
         fclose(cfgfp);
         return(-1);
      }

      getnextline1:;
   }            
   if( timeoutfound && userfound && groupfound && pathfound )
   {
      return(0);
   }
   if( !timeoutfound ) 
   {
      syslog(LOG_ERR,"Error: timeout command not found in %s.", 
		     ftproxcfgfile); 
   }  
   if( !userfound ) 
   {
      syslog(LOG_ERR,"Error: userid command not found in %s.", 
		     ftproxcfgfile); 
   }
   if( !groupfound ) 
   {
      syslog(LOG_ERR,"Error: groupid command not found in %s.",
		     ftproxcfgfile); 
   }
   if( !pathfound ) 
   {
      syslog(LOG_ERR,"Error: ftproxpath command not found in %s.",
		     ftproxcfgfile); 
   }
   return(-1);
}
