/*============================================================================
 * filename:		../ftproxy/ftsecdyn.c					
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.   All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  All rights
 * reserved.   
 *		
 * Date written:	April     28, 1995
 * Last modified:	February   8, 2000 - added LPL
 *
 * function:
 *
 *      This function provides an interface for the Security Dynamics Card. 
 *  	It uses sd_pin, sd_check, and sd_next. This function is designed to
 *	work with the telnet proxy. Thus it uses read and write calls instead
 *	of gets and puts.     
 * 
 * input:	ptr to username
 *
 * return codes:
 *		
 *	 0	user passed authentication test
 *	-1	user failed authentication test: bad response
 *     	-2	user failed authentication test: program error
 *
 *============================================================================*/
                

/* everything (PIN, CARDCODE) is printed to the screen,
   and there are no timeouts; in actual usage such features would be 
   desirable  */
#include <stdio.h>
#include <errno.h>
#include <syslog.h>	
#include "sdi_athd.h"
#include "sdi_size.h"
#include "sdi_type.h"
#include "sdacmvls.h"
#include "sdconf.h" 
#include "firewall.h"
#include "ftproxy.h"
#include "gwuser.h"
union config_record configure;
#define CANCELLED	1   
#define APE 1
/*==========================================================================
 *
 *=========================================================================*/
int tnsecdyn(char *username, char *buf)
{
   struct SD_CLIENT sd_dat, *sd;
   int 	ret;
   int  msglen;
   int	n;
   char input[64];
   char	passcode[LENPRNST];
   char pinsize[20];
   char pintype[20];
   char reply[3];
   char	newpin[LENPRNST];
   char *c;

   memset(&sd_dat, 0, sizeof(sd_dat));   /* clear struct */
   sd = &sd_dat;
   creadcfg();			/*  accesses sdconf.rec  */
   if (sd_init(sd))
   {
      sprintf(buf,"Cannot connect to ACE server.");
      syslog(LOG_ERR,"Error: Cannot connect to ACE server.");
      msglen = strlen(buf); 
      if(write(0,buf,msglen) != msglen)
         syslog(LOG_ERR,"Error: Cannot write error msesage to user.");
      return(-2);  	/* authentication failed: program error. */
   }
   
   /*==============================================
    * request the one time passcode from the user 
    *=============================================*/
   sprintf(buf,"\nenter PASSCODE: "); 
   msglen = strlen(buf); 
   if(write(0,buf,msglen) != msglen)
      return(-2);
   if( getline((unsigned char *)buf,CBUFSIZ - 1,
		     buf,sizeof(buf),APE) < 0 ) 
   {
      syslog(LOG_ERR,"Unable to read one time password: %m.");
      return(-2); 
   }  
   strncpy(input,username,UNMSIZ);
   ret = sd_check(buf, input, sd);
   switch (ret)
   {
      case ACM_OK:

         return(0);     	/* user authenticated */
	 break;

      case ACM_ACCESS_DENIED:

         sprintf(buf,"You entered an invalid login name or password.\r\n");
         syslog(LOG_ERR,"Access denied to user %s",input);
         return(-1);            /* access denied */
	 break;

    case ACM_NEXT_CODE_REQUIRED:

         sprintf(buf,"\r\nNext Cardcode: ");
         msglen = strlen(buf); 
         if(write(0,buf,msglen) != msglen)
            return(-2);
	 if( getline((unsigned char *)buf,CBUFSIZ - 1,
		     buf,sizeof(buf),APE) < 0 ) 
         {
            syslog(LOG_ERR,"Unable to read next Card code.");
            return(-2); 
   	 } 
	 if(strlen(buf) == 0)
	 {
	    sprintf(buf,"\r\nNo next Card code provided."); 
            syslog(LOG_ERR,"No next Card code provided.");
            msglen = strlen(buf);
	    if(write(0,buf,msglen) != msglen)
               return(-2);
            return(-1);            /* access denied */
         }
         if( sd_next(buf, sd) == ACM_OK )
	 {
	    sprintf(buf,"Next passcode accepted. User %s authenticated.",
			input); 
            msglen = strlen(buf);
	    if(write(0,buf,msglen) != msglen)
            {
	       syslog(LOG_ERR,"Unable to write authentication response to"
			      " user");
               return(-2);
	    } 
            return(0);			/* user authenticated */
	 }
	 else
	 {
	    sprintf(buf,"secdyn: access denied, next Cardcode bad.\r\n");
            syslog(LOG_ERR,"Next Cardcode bad, access denied to user %s",input);
            msglen = strlen(buf);
	    if(write(0,buf,msglen) != msglen)
            {
	       syslog(LOG_ERR,"Unable to write Bad Card Code message to"
			      " user.");
               return(-2);
	    } 
            return(-1);            /* access denied */
	 }
	 break;

    case ACM_NEW_PIN_REQUIRED:

	 sprintf(buf,"New PIN required; do you wish to continue? (y/n) [n]: ");
         msglen = strlen(buf); 
         if(write(0,buf,msglen) != msglen)
         {
	    syslog(LOG_ERR,"Unable to write New PIN Req message to user.");
            return(-2);
	 }
	 if( getline((unsigned char *)buf,CBUFSIZ - 1,
		     buf,sizeof(buf),APE) < 0 ) 
         {
            syslog(LOG_ERR,"Unable to read New PIN code.");
            return(-2); 
   	 }  
	 if (strncmp(buf, "y", 1))
	 {
	    sd_pin("", CANCELLED, sd);
	    sprintf(buf,"Cancelled new PIN protocol \n");
            msglen = strlen(buf); 
            if(write(0,buf,msglen) != msglen)
            {
	       syslog(LOG_ERR,"Unable to write 'Canceled new PIN' message to"
			      " user.");
               return(-2);
	    }
	    return(0);
	 }
	 if (sd->alphanumeric)
	    strcpy(pintype, "characters");
	else 
	    strcpy(pintype, "digits");
	if (sd->min_pin_len == sd->max_pin_len)
	    sprintf(pinsize, "%d", sd->min_pin_len);
	else 
	    sprintf(pinsize, "%d to %d", sd->min_pin_len, sd->max_pin_len); 
        /*==========================
	 * User generated PIN 
         *==========================*/
	if (sd->user_selectable)
	{
	    sprintf(buf,"\r\nEnter your new PIN, containing %s %s,\n\
  or\r\n<Return> to generate a new PIN and display it on the screen\r\n",
	        pinsize, pintype);
	    msglen = strlen(buf); 
            if(write(0,buf,msglen) != msglen)
            {
	       syslog(LOG_ERR,"Unable to write 'Enter your new PIN' message"
			      " to user.");
               return(-2);
            }
            if( getline((unsigned char *)buf,CBUFSIZ - 1,
		     buf,sizeof(buf),APE) < 0 ) 
            {
               syslog(LOG_ERR,"Unable to read New PIN for SDI.");
               return(-2); 
   	    } 
	    /*c = fgets(newpin, sd->max_pin_len + 1, stdin);*/
	    if (!strncmp(buf, "\n", 1))
	    {
		strncpy(newpin, sd->system_pin, LENMAXPIN);
		sprintf(buf,"\r\n\nPIN:	%s \n", newpin);
                msglen = strlen(buf); 
                if(write(0,buf,msglen) != msglen)
                {
	           syslog(LOG_ERR,"Unable to write 'PIN: newpin' message to" 
				  " user.");
                   return(-2);
                }
	    }
	    else 
		strncpy(newpin, buf, LENMAXPIN);
	    if (sd_pin(newpin, 0, sd) == ACM_NEW_PIN_ACCEPTED)
		sprintf(buf,"New PIN accepted \r\n");
	    else
		sprintf(buf,"New PIN rejected \r\n"); 
            msglen = strlen(buf); 
            if(write(0,buf,msglen) != msglen)
            {
	       syslog(LOG_ERR,"Unable to write 'New PIN: accepted|rejected'"  
			      "	message to user.");
               return(-2);
            }
	} 
	/*====================================
	 * system-generated PIN 	
	 *====================================*/       
         else		
	 {
	    sprintf(buf,"\r\nPress <Return> to generate a new PIN and display"
			" it on the screen\r\n");
	    msglen = strlen(buf); 
            if(write(0,buf,msglen) != msglen)
            {
	       syslog(LOG_ERR,"Unable to write 'Press <Return> to gen new "
			      "PIN'  message to user.");
               return(-2);
            }
	    if( getline((unsigned char *)buf,CBUFSIZ - 1,
		     buf,sizeof(buf),APE) < 0 ) 
            {
               syslog(LOG_ERR,"Unable to read reply for New PIN for SDI.");
               return(-2); 
     	    } 
          
	    /*fgets(reply, 3, stdin);*/
	    if (!strncmp(buf, "\n", 1))
	    {
	       strncpy(newpin, sd->system_pin, LENMAXPIN);
	       sprintf(buf,"\r\n\nPIN:	%s \r\n", newpin); 
               msglen = strlen(buf);
	       if(write(0,buf,msglen) != msglen)
               {
	          syslog(LOG_ERR,"Unable to write 'PIN: newpin'  message to"
				 " user.");
                  return(-2);
               }
            }
            else
            {
	       sprintf(buf,"New PIN protocol aborted \n");
	       msglen = strlen(buf); 
	       if(write(0,buf,msglen) != msglen)
               {
	          syslog(LOG_ERR,"Unable to write 'Press <Return> to gen new "
				 "PIN'  message to user.");
                  return(-2);
               }
	       exit(0);
            } 
            /* accept new PIN	*/
	    if( sd_pin(newpin, 0, sd) == ACM_NEW_PIN_ACCEPTED )
		sprintf(buf,"New PIN accepted \n");
	    else
		sprintf(buf,"New PIN rejected \n"); 
            msglen = strlen(buf); 
	    if(write(0,buf,msglen) != msglen)
            {
	       syslog(LOG_ERR,"Unable to write 'New PIN accepted|rejected'"
			      "	message to user.");
               return(-2);
            }
	 }
	break;		/*  end of new PIN required  */
         
    default:

         sprintf(buf,"unknown return code %d \n", ret); 
         msglen = strlen(buf); 
	 if(write(0,buf,msglen) != msglen)
         {
	    syslog(LOG_ERR,"Unable to write 'unknown return code' message"
			   " to user.");
            return(-2);
         }
	break;
   }      	/* switch on ret from sd_check  */
   return(0);
}
