/*============================================================================
 * filename:	../fwmon/fwmon.c  
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=============================================================================
 *										
 * written by:		Jim Livermore
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  All rights
 * reserved.    
 *
 *
 * Date written:	September 24, 1995
 *			September 17, 1999 - added Linux support
 * Last modified:	February   8, 2000 - added LPL
 *			
 *
 * function:		Monitor TREX firewall system 	
 *       
 *
 *	fwmon monitors the system parameters for the firewall.  It also
 *      starts up the synmon daemon.
 *
 *
 *============================================================================*/

static  char	ident[] = "@(#) fwmon, V 1.0, 00/01/21"
 			  " (C) Freemont Avenue Software, Inc."
			  " 1996-2000";


#include	<stdio.h> 
#include	<string.h>
#include	<syslog.h>
#include	<sysexits.h>
#include	<ctype.h>
#include	<sys/param.h>
#include	<sys/types.h>
#include	<errno.h>
#include	<sys/stat.h> 
#include 	<sys/utsname.h>
#include	<sys/file.h>
#ifdef _AIX
#include	<sys/dir.h>
#endif
#include	<sys/signal.h> 
#include	<setjmp.h>
#include	<sys/wait.h>
#include	<dirent.h>
#include	<limits.h>
#include	"fwmon.h"
#include	"firewall.h" 
#ifdef _AIX
#include	<procinfo.h>
#else
#ifdef __sun
#define		_KMEMUSER 1
#include	<sys/proc.h>
#include	<kvm.h>
#include	<fcntl.h>
#else
#ifdef __hppa
#include	<sys/pstat.h>
#else
#ifdef linux
#include	<dirent.h>
#endif
#endif
#endif
#endif

#define FWMON
#define __MAINFILE
#include	"chkpoint.h"

/*===========================================================================
 * function prototypes
 *===========================================================================*/
int allblanks(char *s);
int chkaddr(char *s);
int ip4wardCHK(void);   
int ip4wardNOT(void); 
int ip_source_route_CHK(void);   
int ip_source_route_NOT(void); 
int incr_somaxconn(void); 
int pmconf(struct pmonconf *conf);
int synmon(struct synparms *parms);
int kill_old_fwmon(void);
int incr_somaxconn(void);
int addpid(char *file);
void procmon();
/*===========================================================================
 * External variables
 *==========================================================================*/
#ifndef linux
extern	char	*sys_errlist[];
#else
extern	const char	*const sys_errlist[];
#endif

int	debugflag  	= 0; 
static  jmp_buf env;
static	time_t ancient_restart_attempt = 0;
static	time_t old_restart_attempt = 0;
static	time_t last_restart_attempt = 0;

#define TOO_RAPID	60
/*====================================================================
 * signal handler for SIGUSR2
 *===================================================================*/
void restart(int sig) 
{
   time_t	current_restart_attempt;
   signal(SIGUSR2, SIG_IGN);
   time(&current_restart_attempt);
   /*=======================================================================
    * We don't want to keep restarting too rapidly so we look at the
    * intervals between the last 3 restarts.  If the intervals are all
    * less than an arbitrarily small interval, we don't restart.
    *======================================================================*/
   if (	old_restart_attempt != 0 &&
		last_restart_attempt != 0 &&
		old_restart_attempt-ancient_restart_attempt < TOO_RAPID &&
		last_restart_attempt-old_restart_attempt < TOO_RAPID &&
		current_restart_attempt-last_restart_attempt < TOO_RAPID) {
	syslog(LOG_ERR, "Error: A T.Rex monitor may be exiting too rapidly "
		"and will not be re-started automatically.");
	signal(SIGUSR2, restart);
	return;
   }
   ancient_restart_attempt = old_restart_attempt;
   old_restart_attempt = last_restart_attempt;
   last_restart_attempt = current_restart_attempt;
   kill_old_fwmon();
   signal(SIGUSR2, restart);
   longjmp(env,1);
}
/*============================================================================
 *
 *===========================================================================*/
void sigchld(int sig) 
{
   pid_t pid;
   int stat;
   while ((pid = waitpid(-1, &stat, WNOHANG)) > 0);
   signal(SIGCHLD, sigchld);
   return;
}

/*=============================================================================
 * input arguments:
 *			normally none.
 *			-d [level]     debug flag
 *				levels 1 to 3 are supported with 1 = default
 *============================================================================*/
main(int argc, char *argv[]) 
{  
   pid_t	pid;
   struct utsname utsdata; 
   struct utsname *utsptr = &utsdata;
   int	wakeup 	= 3600;			/* wakeup every hour */
   int        rc;
   int        i;

   INIT_CHKPOINTS();
   SETUP_CHKPOINT_HANDLERS();

   openlog("fwmon",LOG_PID|LOG_NDELAY,LOG_DAEMON); 

   /*=====================================================================
    * Terminate any old fwmon that are runnning
    *====================================================================*/
   kill_old_fwmon();

   /*=====================================================================
    * Ignore USR2 until all the monitors have been started.
    *====================================================================*/
   signal(SIGUSR2, SIG_IGN);

   /*========================================================================
    * We don't want zombies
    *=======================================================================*/
   signal(SIGCHLD, sigchld);

   /*================================
    * check for debug flags
    *================================*/
   if(argc > 1 )
   {
      if( !strcmp(argv[1],"-d")) 
      {
         debugflag = 1; 
         if( argc > 2)
         {
            if(!isalldigits(argv[2]))
            {
               fprintf(stderr,"\nError: invalid parameter %s. Must be an "
				"integer > 1 and < 4.\n", argv[2]);
	       return(-1);
            } 
            if( ((debugflag = atoi(argv[2])) > 3) || 
		((debugflag = atoi(argv[2])) < 1)  )
            {
               fprintf(stderr,"\nError: invalid input parameter '%s'.\n",
				argv[2]);
	       return(-1);
            }
         }
      }
      else
      {
         fprintf(stderr,"\nError: invalid input parameter '%s'.\n",
				argv[1]);
         return(-1);
      }
   }

   CHKPOINT();
   if( uname(utsptr) < 0 )
   {
      syslog(LOG_ERR,"Error: Uname call failed. %d.",
			errno);
      return(-1);  
   } 

 
   if( debugflag )
      fprintf(stderr,"\nfwmon starting: debug level = %d.\n\n",debugflag); 

   CHKPOINT();
   if(!debugflag)
      daemonize();
 
   CHKPOINT();
   /*======================================================================
    * This is where the signal handler returns to when it detects that a child
    * has died.
    *======================================================================*/
   if (setjmp(env))
      syslog(LOG_NOTICE,"fwmon Version 1.0 restarted - one of the T.Rex "
		"monitors exited unexpectedly. wakeup = %d sec. %s %s "
		"%s", wakeup, utsptr->sysname, utsptr->version, utsptr->release);
   else
      syslog(LOG_NOTICE,"fwmon Version 1.0 started. wakeup = %d sec. %s %s "
		"%s", wakeup, utsptr->sysname, utsptr->version, utsptr->release);

#ifdef _AIX
   CHKPOINT();
   /*======================================================================
    * Make sure we don't load the synmon kernel extension or run spoofmon
    * if we are not on AIX V4
    *=====================================================================*/
   if (!strcmp(utsdata.version,"4"))
   {
	
	CHKPOINT();
	/*===================================================================
	 * Fork to run a process (built on 4.1) that will load synmonkx
	 *==================================================================*/
	if ((pid = fork()) < 0)
	{
	    if (debugflag)
		fprintf(stderr,"\nfwmon could not fork process for "
			"%s.  %s.\n",SYNMONLD,sys_errlist[errno]);
	    syslog(LOG_ERR,"Error: fwmon could not fork process for "
			"%s.  %s.\n",SYNMONLD,sys_errlist[errno]);
	    exit(1);
	}

	CHKPOINT();
	if (pid == 0)
	{
	    CHKPOINT();
	    execl(SYNMONLD,SYNMONLD,(char *)0);
	    if (debugflag)
		fprintf(stderr,"fwmon could not exec %s.  %s.\n",
			SYNMONLD,sys_errlist[errno]);
	    syslog(LOG_ERR,"Error: fwmon could not exec %s.  %s.\n",
			SYNMONLD,sys_errlist[errno]);
	    exit(1);
	}
	CHKPOINT();
	wait(&i);
	
	CHKPOINT();
	/*==================================================================
	 * Exec synmon
	 *=================================================================*/
	if ((pid = fork()) < 0)
	{
	    syslog(LOG_ERR,"Error: fwmon could not fork process for "
			"%s.  %s.\n",SYNMON,sys_errlist[errno]);
	    exit(1);
	}
	CHKPOINT();
	if (pid == 0)
	{
	    unlink(SYNMONPID);
	    CHKPOINT();
	    execl(SYNMON,SYNMON,(char *)0);
	    syslog(LOG_ERR,"Error: fwmon could not exec %s.  %s.\n",
			SYNMON,sys_errlist[errno]);
	    exit(1);
	}
#endif /* _AIX */
#if !defined(__sun) 
	CHKPOINT();
	/*===================================================================
	 * Start spoofmon
	 *==================================================================*/
	if ((pid=fork()) < 0)
	{
	    if (debugflag)
		fprintf(stderr,"\nfwmon could not fork process for %s.  "
			"%s.\n",SPOOFMON,sys_errlist[errno]);
	    syslog(LOG_ERR,"Error: fwmon could not fork process for %s"
			"%s.\n",SPOOFMON,sys_errlist[errno]);
	    closelog();
	    exit(1);
	}

	CHKPOINT();
	if (pid == 0)
	{
	    CHKPOINT();
	    execl(SPOOFMON,SPOOFMON,(char *)0);

	    if (debugflag)
		fprintf(stderr,"\nfwmon could not exec %s, %s.",
			SPOOFMON,sys_errlist[errno]);
	    syslog(LOG_ERR,"Error: fwmon could not exec %s, %s.",
			SPOOFMON,sys_errlist[errno]);
	    closelog();
	    exit(1);
	}
#endif /* ndef __sun */

	CHKPOINT();
	/*===================================================================
	 * Start procmon if necessary
	 *==================================================================*/
	procmon();
	CHKPOINT();

#ifdef _AIX
   }
#endif /* _AIX */

   addpid(FWMONPIDS);

   /*======================================================================
    * Continuous loop:  Call the functions to be performed           
    *	1. check for IP forwarding and turn off
    *   2. check for source routing            
    *	3. increase so_maxconn to protect against SYN flood attacks
    *   
    *=====================================================================*/
   while(1) 
   {      
      if(debugflag)
         fprintf(stderr,"fwmon re-scanning system\n");
      CHKPOINT();
      if( ip4wardCHK() )  
      {
         CHKPOINT();
         ip4wardNOT();
         CHKPOINT();
         syslog(LOG_NOTICE,"IP forwarding set = 0.");  
      }
      CHKPOINT();
      if( ip_source_route_CHK() )  
      {
         CHKPOINT();
         ip_source_route_NOT();
         CHKPOINT();
         syslog(LOG_NOTICE,"IP source routing set = 0.");  
      }
/*      incr_somaxconn(); */

      
      syslog(LOG_NOTICE,"system ok, rechecking in %d seconds.",wakeup);
      /*====================================================
       * take a nap for wakeup seconds
       *====================================================*/
takenap:
      if(debugflag)
         fprintf(stderr,"Parent napping for %d seconds.\n",wakeup);
      CHKPOINT();
      signal(SIGUSR2, restart);
      sleep(wakeup);
      CHKPOINT();
   }
} 

/*===========================================================================
 * function: ip4wardCHK
 *
 * input:	
 *
 * outputs:
 *	return codes	 0 if ok 
 *			-1 on error  
 *	
 *==========================================================================*/
int ip4wardCHK(void)   
{                       
#define	MAXLEN 128
   FILE 	*tfp; 
   const char 	tmpip4[] = "/tmp/fwmon.tmpip4";
#ifdef _AIX
   const char 	netopt[] = "no -o ipforwarding > ";
#elif __hppa
   const char 	netopt[] = "/usr/contrib/bin/nettune ip_forwarding > ";
#elif __sun
   const char 	netopt[] = "ndd /dev/tcp ip_forwarding > ";
#elif linux
   const char 	netopt[] = "more /proc/sys/net/ipv4/ip_forward  > ";
#endif
   char		netoption[256];
   int		rc = 0;		/* return code 	*/ 
   char		record[MAXLEN];
   char		*token1;

   INIT_CHKPOINTS();
   /*===================================================================== 
    * Check to see if IP packet forwarding is enabled
    *=====================================================================*/ 

   strcpy(netoption,netopt);
   strcat(netoption,tmpip4);
   CHKPOINT();
   if( system(netoption) < 0 )
   {
      syslog(LOG_NOTICE,"Error: system(%s) failed. errno = %d.",
		netoption, errno);               
      rc = -1;
   }
   CHKPOINT();
   /* open temporary file with IP forwarding status */
   if (!(tfp = fopen(tmpip4, "r")))
   {
      syslog(LOG_ERR,"Error: open failed for %s: %m ",tmpip4);
      return(-1);    			/* open failed signal error */
   }
   CHKPOINT();
   /* scan IP4 file looking for ipforwarding option */    
   while( fgets(record, MAXLEN, tfp) != NULL)
   {
       CHKPOINT();
#ifdef _AIX
       token1 = strtok(record,"= \n\r");
       if( strcmp(token1,"ipforwarding") == 0)
       {
          token1 = strtok(NULL,"= \n\r");
          if ( strcmp(token1,"0") == 0)
             rc = 0;
          else 
             rc = -1; 
          break;
       }            
#else
       token1 = strtok(record,"\n\r");
          if (strcmp(token1,"0") == 0)
             rc = 0;
          else
             rc = -1;
#endif
   }    
   CHKPOINT();
   fclose(tfp);
   return(rc);
} /* end of ip4wardCHK function     */

/*===========================================================================
 * function: ip_source_route_CHK
 *
 * input:	
 *
 * outputs:
 *	return codes	 0 if ok 
 *			-1 on error  
 *	
 *==========================================================================*/
int ip_source_route_CHK(void)
{                       
   int		rc = 0;		/* return code 	*/ 
#ifndef __hppa	/* (waltc) Can add for HP-UX? (Token Ring only) */
#define	MAXLEN 128
   FILE 	*tfp; 
   const char 	tmpsrcr[] = "/tmp/fwmon.tmpsrcr";
#ifdef _AIX
   const char 	netopt[] = "no -o nonlocsrcroute > ";
#elif __sun
   const char 	netopt[] = "ndd /dev/ip ip_forward_src_routed > ";
#elif linux
   const char   netopt[] = ""; 		/* temporary klude for Linux */
#endif
   char		netoption[256];
   char		record[MAXLEN];
   char		*token1;

#ifdef linux
   return(0);
#endif

   INIT_CHKPOINTS();
   /*===================================================================== 
    * Check to see if IP source routing is enabled
    *=====================================================================*/ 

   strcpy(netoption,netopt);
   strcat(netoption,tmpsrcr);
   CHKPOINT();
   if( system(netoption) < 0 )
   {
      syslog(LOG_NOTICE,"Error: system(%s) failed. errno = %d.",
		netoption, errno);               
      rc = -1;
   }
    
   CHKPOINT();
   /* open temporary file with IP source routing status */
   if (!(tfp = fopen(tmpsrcr, "r")))
   {
      syslog(LOG_ERR,"Error: open failed for %s: %m ",tmpsrcr);
      return(-1);    			/* open failed signal error */
   }
   CHKPOINT();
   /* scan IP4 file looking for nonlocsrcroute option */    
   while( fgets(record, MAXLEN, tfp) != NULL)
   {
       CHKPOINT();
#ifdef _AIX
       token1 = strtok(record,"= \n\r");
       if( strcmp(token1,"nonlocsrcroute") == 0)
       {
          token1 = strtok(NULL,"= \n\r");
          if ( strcmp(token1,"0") == 0)
             rc = 0;
          else 
             rc = -1; 
          break;
       }            
#else
       token1 = strtok(record,"\n\r");
          if (strcmp(token1,"0") == 0)
             rc = 0;
          else
             rc = -1;
#endif
   }    
   fclose(tfp);
#endif
   CHKPOINT();
   return(rc);
} /* end of ip_source_route_CHK function     */

/*===========================================================================
 * function: ip4wardNOT
 *
 * input:	
 *
 * outputs:
 *	return codes	 0 if ok 
 *			-1 on error  
 *	
 *==========================================================================*/
int ip4wardNOT(void)   
{ 
#ifdef _AIX
   const char netopt[] = "no -o ipforwarding=0";
#elif __hppa
   const char netopt[] = "/usr/contrib/bin/nettune -s ip_forwarding 0";
#elif __sun
   const char netopt[] = "ndd -set /dev/tcp ip_forwarding 0";
#elif linux
   const char   netopt[] = "echo 0 > /proc/sys/net/ipv4/ip_forward";
#endif
   int	rc = 0;		/* return code 	*/

   INIT_CHKPOINTS();
   /*===================================================================== 
    * Turn off IP packet forwarding
    *=====================================================================*/ 

   if( system(netopt) < 0 )
   {
      syslog(LOG_NOTICE,"Error: system(%s) failed. errno = %d.",
		netopt, errno);               
      rc = -1;
   }    
   CHKPOINT();
   return(rc);
} /* end of ip4wardNOT function     */

/*===========================================================================
 * function: ip_source_route_NOT
 *
 * input:	
 *
 * outputs:
 *	return codes	 0 if ok 
 *			-1 on error  
 *	
 *==========================================================================*/
int ip_source_route_NOT(void)   
{ 
#ifdef _AIX
   const char netopt[] = "no -o nonlocsrcroute=0";
#elif __hppa
   FILE *fp;
#define	MAXLEN 128
   char record[MAXLEN];
   char	netoption[256];
   const char netopt[] = "lanscan -i >";
   const char tmplan[] = "/tmp/fwmon.lan_adapters";
#elif __sun
   const char netopt[] = "ndd -set /dev/ip ip_forward_src_routed 0";
#elif linux
   const char netopt[] = "";		/* temporary kludge for Linux */
#endif
   int	rc = 0;		/* return code 	*/

#ifdef linux
   return(0);
#endif

   INIT_CHKPOINTS();
   /*===================================================================== 
    * Turn off IP source routing
    *=====================================================================*/ 

#ifdef __hppa 
   strcpy(netoption,netopt);
   strcat(netoption,tmplan);
   CHKPOINT();
   if( system(netoption) < 0 )
   {
      syslog(LOG_NOTICE,"Error: system(%s) failed. errno = %d.",
		netoption, errno);               
      rc = -1;
   }
   else
   {
      CHKPOINT();
      if ((fp = fopen(tmplan,"r")) == NULL)
      {
         syslog(LOG_ERR,"Error: open failed for %s: %m ",tmplan);
         rc = -1; 
      }
      else
      {
         CHKPOINT();
         while(fgets(record,MAXLEN,fp) != NULL)
         {
            CHKPOINT();
            sprintf(netoption,"lanconfig %s -rif",record);
            if(system(netoption) < 0)
            {
               syslog(LOG_NOTICE,"Error: system(%s) failed. errno = %d.",
	   	      netoption, errno);               
               rc = -1;
               break;
            }
            CHKPOINT();
         }
         CHKPOINT();
         fclose(fp);
      }
      CHKPOINT();
   }
#else
   CHKPOINT();
   if( system(netopt) < 0 )
   {
      syslog(LOG_NOTICE,"Error: system(%s) failed. errno = %d.",
		netopt, errno);
      rc = rc - 1;
   }
#endif
   CHKPOINT();
   return(rc);
} /* end of ip_source_route_NOT function     */
/*===========================================================================
 * function: incr_somaxconn
 * 
 * This function increases the SO_MAXCONN variable to allow the system
 * to queue more partially completed connections, in order to resist 
 * SYN flood attacks.
 *  	  
 * input:	
 *
 * outputs:
 *	return codes	 0 if ok 
 *			-1 on error  
 *	
 *==========================================================================*/
int incr_somaxconn(void)   
{  

   int	rc = 0;		/* return code 	*/
#ifndef __hppa		/* (waltc) Can add for HP-UX? */
#ifdef _AIX
   const char netopt[] = "no -o somaxconn=1000";
#elif __sun
   const char netopt[] = "ndd -set /dev/tcp tcp_conn_req_max 1024";
#elif linux
   const char netopt[] = ""; 		/* temporary kludge for Linux */
#endif

#ifdef linux
   return(0);
#endif
   /*===================================================================== 
    * Increase somaxconn value
    *=====================================================================*/ 

   if( system(netopt) < 0 )
   {
      syslog(LOG_NOTICE,"Error: system(%s) failed. errno = %d.",
		netopt, errno);               
      rc = -1;
   }    
#endif
   return(rc);   
}
/*==========================================================================
 * function:	allblanks
 * input:	a pointer to a character string
 * output:	returns 1 if string is all blanks
 *	   	returns 0 is string is not all blanks 
 *=========================================================================*/
int allblanks(char *s)  
{
   int	ch;

   INIT_CHKPOINTS();
   /* loop through string */
   while((ch = *s++) != 0)
   {
      CHKPOINT();
      if(!isascii(ch) || !isspace(ch))
         return(0);
   }
   CHKPOINT();
   return(1);
}

/*=======================================================================
 * function: 	kill_old_fwmon
 * input:	the fwmon.pids file
 * output:	
 *
 * note:	When a system error is found, this function only logs the
 *		error in syslog and then continues.  It doesn't exit since
 *		it is not a big problem if we cannot kill the old processes.
 *======================================================================*/
int kill_old_fwmon()
{
   FILE	*pf;
   pid_t mypid;
   char	pmpid[LINE_MAX];

   INIT_CHKPOINTS();
   if (!(pf = fopen(FWMONPIDS,"r")))
   {
	/*====================================================================
	 * If the file does not exist, it is not an error.
	 * Log everything else to syslog.
	 *===================================================================*/
	if (errno != ENOENT)
	{
	    syslog(LOG_ERR,"Error: Failed opening %s for reads, errno=%d.\n",
			FWMONPIDS,errno);
	    return(-1);
	}
	return(0);
   }
   CHKPOINT();
   mypid = getpid();
   while (fgets(pmpid,LINE_MAX,pf) != NULL)
   {
	pid_t pid;
	CHKPOINT();
	strcpy(pmpid,strtok(pmpid," \t\n"));
	if ((pid = atoi(pmpid)) != mypid) {
	   CHKPOINT();
	   if (kill(pid,SIGKILL) && errno != ESRCH)
		syslog(LOG_NOTICE,"Could not kill pid %s, %s.\n",
			pmpid,sys_errlist[errno]);
	   else if (errno != ESRCH)
		syslog(LOG_INFO, "Sent KILL signal to pid %d.", pid);
	}
	CHKPOINT();
   }
   CHKPOINT();
   if (!feof(pf))
	syslog(LOG_ERR,"Error: Failed reading %s, errno=%d.\n",
		FWMONPIDS,errno);

   fclose(pf);

   if (unlink(FWMONPIDS))
	syslog(LOG_ERR,"Error: Failed to unlink %s, errno=%d.\n",
		FWMONPIDS,errno);

   CHKPOINT();
   sleep(1);	/* Give processes time to die	*/
   return(0);
}

/*=========================================================================
 * Startup procmon
 *========================================================================*/
void procmon()
{
   pid_t pid;

   INIT_CHKPOINTS();
   if ((pid = fork()) < 0)
   {
      if (debugflag)
		fprintf(stderr,"\nfwmon could not fork process for "
			"procmon.  %s.\n",sys_errlist[errno]);
      syslog(LOG_ERR,"Error: fwmon could not fork process for "
			"procmon.  %s.\n",sys_errlist[errno]);
      closelog();
      exit(1);
   }
   CHKPOINT();
   if (pid != 0)
      return;

   CHKPOINT();
   if (execl(PROCMON,PROCMON,(char *)0) != 0)
   {
      syslog(LOG_ERR,"Error: fwmon could not exec %s.  %s.\n",
			PROCMON,sys_errlist[errno]);
      exit(1);
   }
   CHKPOINT();
   exit(0);
}
