/*============================================================================
 * filename:		../fwmon/pmconf.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1996-2000.  All rights
 * reserved.
 *		
 * Date written:	October   14, 1996
 *			September 17, 1999 - added Linux support
 * Last modified:	February   8, 2000 - added LPL
 *	                         
 * function:                       
 *	The pmconf function reads the fwmon configuration file.
 *
 *  
 *
 * fwmon configuration file format: 
 *
 * keyword 		value 		description
 *
 * syn_chk_interval 	nnn 		number of seconds between checking the
 *					so_q0 queues for all ports
 * syn_qlen		qlen		length of so_q0 
 * syn_himark		max		number of entries in so_q0 for a port
 *					above which an alert will be issued
 *					and so_q0 cleanup will start
 * syn_lomark		keep		max number of entries to keep in so_q0
 *					during each cleanup iteration.
 * syn_verbose		y/n		verbose mode will log more info in
 *					syslog
 * syn_iplog		filename	name of file to log foreign ip's and
 *					local port numbers of all deleted
 *					so_q0 entries
 * syn_trusted		filename	name of file which contains ip/mask
 *					pair of trusted ip's whose so_q0
 *					entries should not be deleted during
 *					cleanup
 *                                                              
 *
 *============================================================================*/
#include	<sys/types.h>
#include	<sys/stat.h> 
#include	<errno.h> 
#include	<stdio.h>
#include	<stdlib.h>
#include	<syslog.h>
#include	<grp.h> 
#include	<pwd.h>
#include	<string.h>
#include	<sys/time.h>
#include	<netdb.h>               /* defines servent structure      */
#include	<netinet/in.h> 
#include	<limits.h>
#include	"firewall.h"
#include	"fwmon.h"

#ifndef linux
extern char	*sys_errlist[];
#else
extern const char	*const sys_errlist[];
#endif

static char trusted_file[LINE_MAX];

/*==================================================================
 * function: 	isalldigits
 * input:	a pointer to a character string 
 * output:	returns 1 if all the characters in the string are digits 
 *==================================================================*/  
int isalldigits(char *s)  
{
   while(isdigit(*s))
      s++;
   return(*s == '\0');
} 

/*=========================================================================
 * chkuintparm - given the name of the pmonconf record, this functions checks
 *		 to make sure there is a value specified for the record and
 *		 that the value specified is valid and not negative.
 *		 It assumes that it can call strtok wil a NULL parameter to
 *		 pick up the parameter value.
 * input: 	pointer to parameter name
 *		pointer to the pmonconf member to be updated
 *		line number where record was found
 * outputs: 	updated pmonconf record
 *		rc = 0 if ok 
 *		rc = -1 if error
 *		rc = 1 if value specified is zero
 *========================================================================*/
int chkuintparm(char *parm, unsigned short *intparm, int line)
{
   char *token1;
   unsigned long temp;
   char	tokensep[]= "\t,:;= \n";

   if( (token1 = strtok(NULL,tokensep)) == NULL)
   {
      syslog(LOG_ERR,"Error: No value specified for %s "
			"on line %d of %s.", 
			parm, line, PMONCONFFILE); 
      return(-1); 
   }
        
   if (*token1 == '-')
   {
      syslog(LOG_ERR,"Error: value for %s on line %d of %s "
			"cannot be negative.",
			parm, line, PMONCONFFILE);
      return(-1);
   }

   if(isalldigits(token1))
   {
      errno = 0;
      temp = strtoul(token1,NULL,10);
      if (temp == ULONG_MAX ||
	  (temp == 0 && errno != 0) ||
	  temp > 0xffff)
      {
         syslog(LOG_ERR,"Error: %s value out of range on line %d of %s.",
			parm, line, PMONCONFFILE);
         return(-1);
      }

      if (temp == 0 && strcasecmp(parm,"syn_chk_interval") != 0)
      {
         syslog(LOG_ERR,"Error: %s value cannot be zero on line %d of %s.",
			parm, line, PMONCONFFILE);
         return(-1);
      }
      *intparm = temp;
      return(0);
   }
   else
   {
      syslog(LOG_ERR,"Error: value for %s on line %d of %s "
			"is not valid.",
			parm, line, PMONCONFFILE);
      return(-1);
   } 
}

/*=========================================================================
 * input: 	pointer to pmonconf record
 * outputs: 	updated pmonconf record
 *		rc = 0 if ok 
 *		rc = -1 if error
 *		rc = 1 if fwmon.conf does not exist.  This situation
 *		     may be OK.  E.g. on hp, where spoofmon is supported and
 *		     synmon is not.  The only info that is interesting is
 *		     spoofmon = yes|no.  Since this info is optional, the
 *		     file should not be required.
 *========================================================================*/
int pmconf(struct pmonconf *pmonconf)  
{   
   char		record[LINE_MAX];
   FILE 	*fp; 
   char 	*cp; 
   char		*token1;
   char		*token2; 
   char		tokensep[] 	= "\t,:;= \n";
   int		qlenfound 	= 0;
   int		himarkfound 	= 0;
   int		lomarkfound 	= 0;
   int		intfound	= 0;
   int		rc; 
   int		line=0;
   int		curidx=0;
   struct synport	*curport;
   procmonparms	*newprocmon;

   if( pmonconf == NULL)
   {
      syslog(LOG_ERR,"Error: invalid pmonconf pointer passed");
      return(-1);
   }
   
   /*======================================================================  
    * Intialize defaults 
    *=====================================================================*/
   memset(pmonconf,0,sizeof(struct pmonconf));
   pmonconf->synparms.chk_interval = -1;
   pmonconf->spoofmon = 1;
   pmonconf->procmon_intv1 = 3600;	/* default procmon interval to 1 hr */
   pmonconf->procmon_intv2 = pmonconf->procmon_intv1;
   pmonconf->procmon_email = "root";	/* default address to root on fw */
  
   if (!(fp = fopen(PMONCONFFILE, "r")))
   {  
      if (errno == ENOENT)
         return(1);
      syslog(LOG_ERR,"Error: open failed for %s: %m",PMONCONFFILE);
      return(-1);    			/* open failed signal error */
   }

   cp = (char *) &record;

   /*===================================================================== 
    * Read the config file and look for match 
    *====================================================================*/
   while( fgets(record, sizeof(record), fp) != NULL)
   { 
      line++;   

      /*================================================================== 
       * skip over comments and null lines
       *=================================================================*/
      token1 = strtok(record,tokensep);
      if ((token1 == NULL) || (*token1 == '#'))
         continue;

      /*================================================================== 
       * If spoofmon then get yes or no setting
       *=================================================================*/
      if ( strcmp(token1,"spoofmon") == 0 )
      {
         token1 = strtok(NULL,tokensep);
         if (strcasecmp(token1,"no") == 0)
            pmonconf->spoofmon = 0;
         else if (strcasecmp(token1,"yes") != 0)
         {
             syslog(LOG_ERR,"Error: Invalid value specified for spoofmon on "
				"line %d of %s.",line,PMONCONFFILE);
             fclose(fp);
             return(-1);
         }
         continue;
      }

       /*==================================================================
       * If procmon_interval then get the procmon interval
       *=================================================================*/
      if ( strcmp(token1,"procmon_interval") == 0 )
      {
         if ((token1 = strtok(NULL,tokensep)) == NULL)
         {
             syslog(LOG_ERR,"Error: Missing value for procmon_interval on "
				"line %d of %s.",line,PMONCONFFILE);
             fclose(fp);
             return(-1);
         }
         if (!isalldigits(token1))
         {
             syslog(LOG_ERR,"Error: Invalid value '%s' for procmon "
				"check interval on "
				"line %d of %s.",token1,line,PMONCONFFILE);
             fclose(fp);
             return(-1);
         }
         pmonconf->procmon_intv1 = atoi(token1);
         /*=============================================================== 
          * Optional alert interval
          *==============================================================*/
         if ((token1 = strtok(NULL,tokensep)) != NULL)
         {
            if (!isalldigits(token1))
            {
                syslog(LOG_ERR,"Error: Invalid value '%s' for "
				"procmon alert interval on "
				"line %d of %s.",token1,line,PMONCONFFILE);
                fclose(fp);
                return(-1);
            }
            pmonconf->procmon_intv2 = atoi(token1);
         }
         else
            pmonconf->procmon_intv2 = pmonconf->procmon_intv1;
         continue;
      }

      /*================================================================== 
       * If procmon_email then get the procmon email address
       *=================================================================*/
      if ( strcmp(token1,"procmon_email") == 0 )
      {
         if ((token1 = strtok(NULL,tokensep)) == NULL)
         {
             syslog(LOG_ERR,"Error: Missing value for procmon_email on "
				"line %d of %s.",line,PMONCONFFILE);
             fclose(fp);
             return(-1);
         }
         if ((pmonconf->procmon_email = malloc(strlen(token1) + 1)) == NULL)
         {
             syslog(LOG_ERR,"Error: malloc for procmon email address failed.");
             fclose(fp);
             return(-1);
         }
         strcpy(pmonconf->procmon_email,token1);
         continue;
      }

      /*================================================================== 
       * If procmon then get the process name
       *=================================================================*/
      if ( strcmp(token1,"procmon") == 0 )
      {
         if ((token1 = strtok(NULL,tokensep)) == NULL)
         {
             syslog(LOG_ERR,"Error: Missing value for procmon on "
				"line %d of %s.",line,PMONCONFFILE);
             fclose(fp);
             return(-1);
         }
         newprocmon = malloc(sizeof(procmonparms));
         if (newprocmon == NULL)
         {
            syslog(LOG_ERR,"Error: malloc for procmon parameter failed.");
            fclose(fp);
            return(-1);
         }
         newprocmon->next = NULL;
         if (pmonconf->procs == NULL)
            pmonconf->procs = newprocmon;
         else
         {
             procmonparms	**lastproc;
             /*==========================================================
              * Forward to the last entry
              *=========================================================*/
             for (lastproc=&(pmonconf->procs); *lastproc;
			lastproc=&(*lastproc)->next)
             ;

             *lastproc = newprocmon;
         }

         newprocmon->procname = malloc(strlen(token1) + 1);
         if (newprocmon->procname == NULL)
         {
            syslog(LOG_ERR,"Error: malloc for procmon parameter failed.");
            fclose(fp);
            return(-1);
         }
         strcpy(newprocmon->procname,token1);

         /*==============================================================
          * Look for the optional n parameter
          *=============================================================*/
         if ((token1 = strtok(NULL,tokensep)) != NULL)
         {
            if (!isalldigits(token1))
            {
                syslog(LOG_ERR,"Error: Invalid n1 value '%s' on "
				"line %d of %s.",token1,line,PMONCONFFILE);
                fclose(fp);
                return(-1);
            }
            newprocmon->expected_n = atoi(token1);
            /*===========================================================
             * Look for the optional hi_n parameter
             *==========================================================*/
            if ((token1 = strtok(NULL,tokensep)) != NULL)
            {
               if (!isalldigits(token1))
               {
                   syslog(LOG_ERR,"Error: Invalid n2 value '%s' on "
				"line %d of %s.",token1,line,PMONCONFFILE);
                   fclose(fp);
                   return(-1);
               }
               newprocmon->expected_hi_n = atoi(token1);
               if (newprocmon->expected_hi_n < newprocmon->expected_n)
               {
                  syslog(LOG_ERR,"Error: The first number (%d) cannot be "
				"greater than the second number (%d) on a "
				"procmon statement. - Line %d of %s",
				newprocmon->expected_n,
				newprocmon->expected_hi_n,
				line,PMONCONFFILE);
                  fclose(fp);
                  return(-1);
               }
            }
            else
               newprocmon->expected_hi_n = newprocmon->expected_n;
         }
         else
         {
            newprocmon->expected_n = 1;
            newprocmon->expected_hi_n = newprocmon->expected_n;
         }
         
         continue;
      }
     
      /*=========================================================	
       * If syn_chk_interval record then get interval
       *========================================================*/	
      if( strcmp(token1,"syn_chk_interval") == 0 )
      {
         if (rc = chkuintparm(token1,&pmonconf->synparms.chk_interval,line))
         {
            fclose(fp);
            return(rc);
         }

         if( pmonconf->synparms.chk_interval == 0)
            syslog(LOG_NOTICE,"synmon disabled because syn_chk_interval value "
			"on line %d of %s is zero.",line,PMONCONFFILE);
         intfound++;
         continue;
      } 

      /*=========================================================	
       * If qlen record then get qlen
       *========================================================*/	
      if( strcmp(token1,"syn_qlen") == 0 )
      {
         if(rc = chkuintparm(token1,&pmonconf->synparms.qlen,line))
         {
            fclose(fp);
            return(rc);
         }

         qlenfound++;
         continue;
      } 

      /*=========================================================	
       * If himark record then get himark
       *========================================================*/	
      if( strcmp(token1,"syn_himark") == 0 )
      {
         if(rc = chkuintparm(token1,&pmonconf->synparms.himark,line))
         {
            fclose(fp);
            return(rc);
         }
         himarkfound++;
         continue;
      }

      /*=========================================================	
       * If lomark record then get himark
       *========================================================*/	
      if( strcmp(token1,"syn_lomark") == 0 )
      {
         if(rc = chkuintparm(token1,&pmonconf->synparms.lomark,line))
         {
            fclose(fp);
            return(rc);
         }
         lomarkfound++;
         continue;
      }

       /*=========================================================	
       * If syn_verbose record then get verbose value and validate 
       *========================================================*/	
      if( strcmp(token1,"syn_verbose") == 0 )
      {
         if( (token1 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for verbose on line %d "
				"of %s.", line, PMONCONFFILE); 
	    fclose(fp);
	    return(-1); 
         }
	 if( strcmp(token1,"yes")  == 0 ) 
         {
            pmonconf->synparms.verbose = 1;    
	 }
         else if( strcmp(token1,"no")  == 0 )  
         {
            pmonconf->synparms.verbose = 0;
	 }
         else
	 { 
            syslog(LOG_ERR,"Error: invalid value for verbose '%s' specified on "
				"line %d of %s.",
 				token1, line, PMONCONFFILE); 
	    fclose(fp);
	    return(-1);       
         } 
         continue;
      } 

      /*=========================================================	
       * If syn_iplog record then get filename
       *========================================================*/	
      if( strcmp(token1,"syn_iplog") == 0 )
      {
         if( (token1 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for syn_iplog on line %d "
				"of %s.", line, PMONCONFFILE); 
	    fclose(fp);
	    return(-1); 
         }
         pmonconf->synparms.iplog = fopen(token1,"a");
         if (!pmonconf->synparms.iplog)
            syslog(LOG_ERR,"Error: Cannot open %s for output (line %d of %s)."
				"  %s.  Continuing with output disabled.",
				token1,line,PMONCONFFILE,sys_errlist[errno]);
         continue;
      }  
      /*=========================================================	
       * If syn_trusted record then get filename
       *========================================================*/	
      if( strcmp(token1,"syn_trusted") == 0 )
      {
         if( (token1 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for syn_trusted on line "
				"%d of %s.", line, PMONCONFFILE); 
	    fclose(fp);
	    return(-1); 
         }
         strncpy(trusted_file,token1,sizeof(trusted_file));
         pmonconf->synparms.trusted = trusted_file;
         continue;
      } 
      /*=========================================================	
       * If syn_port record then get port specific parms
       *========================================================*/	
      if( strcmp(token1,"syn_port") == 0 )
      {
         int		pqlenfound 	= 0;
         int		phimarkfound 	= 0;
         int		plomarkfound 	= 0;

         /*=========================================================	
          * malloc 2 structs in the 1st round because there is always
          * an empty one with port = 0 to indicate the end.
          *========================================================*/	
         if (pmonconf->synparms.plist == NULL)
         {
            pmonconf->synparms.plist = malloc(2*sizeof(struct synport));
            pmonconf->synparms.plistsz = 2*sizeof(struct synport);
         }
         else
         {
            curidx++;
            pmonconf->synparms.plist =
		realloc(pmonconf->synparms.plist,
			(curidx+2)*sizeof(struct synport));
            pmonconf->synparms.plistsz = (curidx+2)*sizeof(struct synport);
         }

         curport = pmonconf->synparms.plist + curidx;
         memset(curport,0,sizeof(struct synport));

         if(rc = chkuintparm(token1,&curport->port,line))
         {
            free(pmonconf->synparms.plist);
	    fclose(fp);
	    return(-1); 
         }

         /*=========================================================	
          * Process sub-parameters for syn_port
          *========================================================*/	
         while (token1 = strtok(NULL,tokensep))
         {

            /*=========================================================	
             * If qlen parm then get qlen
             *========================================================*/	
            if( strcmp(token1,"qlen") == 0 )
            {
               if(rc = chkuintparm("port.qlen",&curport->qlen,line))
               {
                  free(pmonconf->synparms.plist);
                  fclose(fp);
                  return(rc);
               }

               pqlenfound++;
               continue;
            } 

            /*=========================================================	
             * If himark parm then get himark
             *========================================================*/	
            if( strcmp(token1,"himark") == 0 )
            {
               if(rc = chkuintparm("port.himark",&curport->himark,line))
               {
                  free(pmonconf->synparms.plist);
                  fclose(fp);
                  return(rc);
               }
               phimarkfound++;
               continue;
            }

            /*=========================================================	
             * If lomark parm then get himark
             *========================================================*/	
            if( strcmp(token1,"lomark") == 0 )
            {
               if(rc = chkuintparm("port.lomark",&curport->lomark,line))
               {
                  free(pmonconf->synparms.plist);
                  fclose(fp);
                  return(rc);
               }
               plomarkfound++;
               continue;
            }

            /*=========================================================	
             * Invalid syn_port sub-parameter
             *========================================================*/	
            else
            {
               syslog(LOG_ERR,"Error: Invalid syn_port sub-parameter %s on "
				"%d of %s",token1,line,PMONCONFFILE);
               free(pmonconf->synparms.plist);
               fclose(fp);
               return(-1);
            }
         }

         /*=========================================================	
          * Make sure syn_port record is complete
          *========================================================*/	
         if (!pqlenfound || !phimarkfound || !plomarkfound)
         {
            syslog(LOG_ERR,"Error: Line %d of %s - syn_port record must "
				"include complete specification of qlen, "
				"himark and lomark.",line,PMONCONFFILE);
            free(pmonconf->synparms.plist);
            fclose(fp);
            return(-1);
         }

         /*=========================================================	
          * Make sure syn_port sub-parameter values make sense
          *========================================================*/	
         if (curport->himark > (3 * curport->qlen / 2 + 1))
         {
            syslog(LOG_ERR,"Error: himark value (%d) is greater than 3/2 * "
				"the qlen value (%d) + 1 on line %d of %s.",
				curport->himark,curport->qlen,line,
				PMONCONFFILE);
            free(pmonconf->synparms.plist);
            fclose(fp);
            return(-1);
         }

         if (curport->lomark >= curport->qlen)
         {
            syslog(LOG_ERR,"Error: lomark value (%d) must be less than qlen "
				"value (%d) on line %d of %s.",
				curport->lomark,curport->qlen,line,
				PMONCONFFILE);
            free(pmonconf->synparms.plist);
            fclose(fp);
            return(-1);
         }

         continue;
      }  

      /*==================================================================
       * Unknown keyword!
       *=================================================================*/
      syslog(LOG_ERR,"Error: invalid record on line %d of %s, unknown keyword "
			"'%s'.", line, PMONCONFFILE, token1);
      fclose(fp);
      return(-1);
       
   } 

   /*================================================================= 
    * Mark the end of the synport list
    *================================================================*/
   if (pmonconf->synparms.plist)
   {
      curport++;
      curport->port = 0;
   }

#ifdef _AIX
   /*================================================================= 
    * now check to make certain all variables were supplied:
    *  syn_chk_interval, syn_qlen, syn_himark, syn_lomark
    *================================================================*/
   
   if( intfound == 0)
   {
      syslog(LOG_ERR,"Error: No syn_chk_interval record found in %s.",
		     PMONCONFFILE);
      fclose(fp);
      return(-1);
   }

   /*================================================================= 
    * If syn_chk_interval was specified and is non-zero, then sysnmon
    *  is to be activated so we check to make sure the other synmon
    * parameters are OK.
    *================================================================*/
   if (pmonconf->synparms.chk_interval != 0)
   {
      if( qlenfound == 0 )
      {
         syslog(LOG_ERR,"Error: No syn_qlen record found in %s.",
		     PMONCONFFILE);
         fclose(fp);
         return(-1);
      }
      if( himarkfound == 0 )
      {
         syslog(LOG_ERR,"Error: No syn_himark record found in %s.",
		     PMONCONFFILE);
         fclose(fp);
         return(-1);
      } 
      if( lomarkfound == 0 )
      {
         syslog(LOG_ERR,"Error: No syn_lomark record found in %s.",
		     PMONCONFFILE);
         fclose(fp);
         return(-1);
      } 

      /*=========================================================	
       * Make sure parameter values make sense
       *========================================================*/	
      if (pmonconf->synparms.himark > (3 * pmonconf->synparms.qlen / 2 + 1))
      {
         syslog(LOG_ERR,"Error: syn_himark value (%d) is greater than 3/2 * "
			"the syn_qlen value (%d) + 1 in %s.",
			pmonconf->synparms.himark,
			pmonconf->synparms.qlen,PMONCONFFILE);
         fclose(fp);
         return(-1);
      }

      if (pmonconf->synparms.lomark >= pmonconf->synparms.qlen)
      {
         syslog(LOG_ERR,"Error: syn_lomark value (%d) must be less than "
			"syn_qlen value (%d) in %s.",
			pmonconf->synparms.lomark,
			pmonconf->synparms.qlen,PMONCONFFILE);
         fclose(fp);
         return(-1);
      }
   }
#endif

   fclose(fp);
   return(0);  			/* everything is ok */

} 
