/*============================================================================
 * filename:	../fwmon/procmon.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=============================================================================
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1997-2000.  All rights
 * reserved.
 *
 *
 * Date written:	June      10, 1997
 *			September 17, 1999 - added Linux support
 * Last modified:	February   8, 2000 - added LPL
 *			
 *
 * function:		Monitor named processes on a TREX firewall system 	
 *
 *
 *	procmon monitors named process for the firewall.
 *
 *============================================================================*/

static  char	ident[] = "@(#) procmon, V 1.0, 00/01/21"
 			  " (C) Freemont Avenue Software, Inc."
			  " 1997-2000";


#include	<stdio.h>
#include	<string.h>
#include	<syslog.h>
#include	<ctype.h>
#include	<sys/param.h>
#include	<sys/types.h>
#include	<errno.h>
#include	<limits.h>
#include	<sys/signal.h>
#include	"fwmon.h"
#include	"firewall.h"
#ifdef _AIX41
#include	<procinfo.h>
#else
#ifdef __sun
#define		_KMEMUSER 1
#include	<sys/proc.h>
#include	<kvm.h>
#include	<fcntl.h>
#else
#ifdef __hppa
#include	<sys/pstat.h>
#else
#ifdef linux
#include	<dirent.h>
#endif
#endif
#endif
#endif
/*===========================================================================
 * External variables
 *==========================================================================*/
#ifndef linux
extern	char	*sys_errlist[];
#else
extern	const char	*const sys_errlist[];
#endif
static pid_t	pmonpid;

#define __MAINFILE
#include "chkpoint.h"

/*=======================================================================
 * exit handling
 *======================================================================*/
void pmexit() {
    kill(pmonpid, SIGUSR2);
    return;
}

/*=============================================================================
 * input arguments:
 *============================================================================*/
main(int argc, char *argv[])
{
   int rc;
   int alert;
   int alert_countdown = 0;
   char *tmpf="/tmp/.procmon";
   pid_t pid;
   struct pmonconf conf;
#ifdef _AIX
   struct procsinfo     monitor;
#else
#ifdef __sun
   kvm_t		*kd;
   struct proc		*monitor;
#else
#ifdef __hppa
   struct pst_status    monitor;
#else
#ifdef linux
    DIR                 *dp;
    struct dirent       *dirp;
    char                buf[BUFSIZ];
    #define PIDNO       dirp->d_name
#endif
#endif
#endif
#endif

   /*================================================================
    * Sleep 2 seconds to let all the fwmon/spoofmon/synmon startup
    * activities settle before monitoring the process table.  Without
    * the sleep(), procmon may incorrectly issue warnings on mismatched
    * counts on some fwmon processes during startup.
    *===============================================================*/
   sleep(2);

   SETUP_CHKPOINT_HANDLERS();
   INIT_CHKPOINTS();

   openlog("procmon",LOG_PID|LOG_NDELAY,LOG_DAEMON);

   if (pmconf(&conf) < 0)
      exit(1);

   if (!conf.procs)
      exit(0);

   CHKPOINT();
   pmonpid = getppid();
   atexit(pmexit);

   close(0);
   close(1);
   close(2);

   /*================================================================
    * Add procmon to the fwmon.pids file.
    *===============================================================*/
   addpid(FWMONPIDS);

   syslog(LOG_NOTICE, "procmon daemon started");

   CHKPOINT();
   /*================================================================
    * Wake up every conf->procmon_intv1 seconds and check on the
    * specified processes
    *===============================================================*/
   while (1)
   {
      FILE *tf;
      procmonparms *cur;
      char msg[BUFSIZ];
      CHKPOINT();
      /*=============================================================
       * Zero out the actual counts before checking
       *============================================================*/
      for (cur=conf.procs; cur; cur = cur->next)
         cur->actual_n = 0;

      CHKPOINT();
      /*=============================================================
       * Loop thru all the process table entries
       *============================================================*/
      pid = 0;
#ifdef _AIX
      while ((rc = getprocs(&monitor,sizeof(monitor),NULL,0,&pid,1)) == 1)
      {
         CHKPOINT();
         for (cur=conf.procs; cur; cur = cur->next)
         {
            if (strcmp(cur->procname,monitor.pi_comm) == 0)
            {
               cur->actual_n++;
               break;
            }
         }
      }
#else
#ifdef __sun
      kd = kvm_open(NULL,NULL,NULL,O_RDONLY,"fwmon:");
      if (kd == NULL) {
         syslog(LOG_ERR, "Error: kvm_open returned NULL, procmon exiting.  %m");
         exit(1);
      }
      while ((monitor = kvm_nextproc(kd)) != NULL)
      {
         int error;
         char *cmd;
         char *progname;
         char **arg;
         struct user *user;

         CHKPOINT();
         cmd = NULL;
         progname = NULL;
         /*===================================================================
          * kvm_getu() returns a NULL for a zombie.
          * Just to make sure we don't core dump here
          *==================================================================*/
         if ((user = kvm_getu(kd, monitor)) == NULL)
            continue;
         CHKPOINT();
         /*===================================================================
          * Get the command from the user area
          *==================================================================*/
         if ((error = kvm_getcmd(kd, monitor, user, &arg, NULL)) == -1) {
            if (user->u_comm[0] != '\0')
                cmd = user->u_comm;
         }
         else
            cmd = arg[0];

         CHKPOINT();

         if (cmd != NULL) {
            /*===============================================================
             * Get rid of path info
             *==============================================================*/
            if ((progname = strrchr(cmd, '/')) != NULL)
                progname++;
            else
                progname = cmd;

            for (cur=conf.procs; cur; cur = cur->next)
            {
               if (strcmp(cur->procname,progname) == 0)
               {
                  cur->actual_n++;
                  break;
               }
            }
         }

         CHKPOINT();
         if (error != -1)
            free(arg);
      }
      kvm_close(kd);
#else
#ifdef __hppa
      while ((rc = pstat_getproc(&monitor,sizeof(monitor),1,pid++)) == 1)
      {
         CHKPOINT();
         for (cur=conf.procs; cur; cur = cur->next)
         {
            if (strcmp(cur->procname,monitor.pst_ucomm) == 0)
            {
               cur->actual_n++;
               break;
            }
         }
      }
#else
#ifdef linux
    dp = opendir("/proc");
    if (dp == NULL) {
        syslog(LOG_ERR, "Error: Could not open /proc directory. %s", sys_errlist[errno]);
        return(-1);
    }
    while ((dirp = readdir(dp)) != NULL) {
        int ffd;
        char *cp;
        /*===============================================================
         * Build a pseudo filename of the form /proc/<pidno>/stat and
         * try to open the file
         * If we encounter any error in the open or the read, do not treat
         * them as errors.  I assume this is correct way to handle them
         * since /proc is a pseudo filesystem anyway and some of these
         * pseudo files may be "weird" in the eyes of open and read.
         *==============================================================*/
        CHKPOINT();
        strcpy(buf, "/proc/");
        strcat(buf, PIDNO);
        strcat(buf, "/stat");

        if ((ffd = open(buf, 0)) < 0)
            continue;
        if (read(ffd, buf, sizeof(buf)) <= 0)
            continue;
        close(ffd);

        cp = strtok(buf, " \t()");
        if (cp == NULL)
            continue;
        /*===============================================================
         * Get 2nd field which has proc name
         *==============================================================*/
        if ((cp = strtok(NULL, " \t()")) == NULL)
            continue;
        for (cur=conf.procs; cur; cur = cur->next)
        {
           if (strcmp(cur->procname, cp) == 0) {
              cur->actual_n++;
              break;
           }
        }
   }
#endif
#endif
#endif
#endif

      CHKPOINT();
      /*=============================================================
       * See if any of the specified processes did not occur expected_n
       * times in the process table.
       *============================================================*/
      alert = 0;
      if (unlink(tmpf) < 0 && errno != ENOENT)
         syslog(LOG_ERR,"Error: Cannot unlink %s, %m",tmpf);
      for (cur=conf.procs; cur; cur = cur->next)
      {
         if (cur->actual_n < cur->expected_n ||
		cur->actual_n > cur->expected_hi_n)
         {
            if (alert_countdown <= 0)
            {
               if ((tf = fopen(tmpf,"a")) == NULL)
                  syslog(LOG_ERR,"Error: failed to open %s, %m",tmpf);
               if (cur->expected_n == cur->expected_hi_n)
                  sprintf(msg,"Expected exactly %d occurences of process "
			"\"%s\" but found %d occurences in the process table.",
			cur->expected_n,cur->procname,cur->actual_n);
               else
                  sprintf(msg,"Expected between %d and %d occurences of "
			"process \"%s\" but found %d occurences in the "
			"process table.",
			cur->expected_n,cur->expected_hi_n,
			cur->procname,cur->actual_n);
               syslog(LOG_NOTICE,msg);
               if (tf)
               {
                  fprintf(tf,"%s\n",msg);
                  fclose(tf);
               }
            }
            alert = 1;
         }
      }

      CHKPOINT();
      /*=============================================================
       * If alert is set and alert_countdown <= 0, we just issued some
       * alerts - so we set the next alert timer to procmon_intv2.
       *============================================================*/
      if (alert && alert_countdown <= 0)
      {
         alert_countdown = conf.procmon_intv2;
#ifndef linux
         if (tf)
            sprintf(msg,"/usr/bin/mailx -s \"procmon issued alarm\" %s <%s",
		conf.procmon_email,tmpf);
         else
            sprintf(msg,"/usr/bin/mailx -s \"procmon issued alarm\" %s",
		conf.procmon_email);
#else
         if (tf)
            sprintf(msg,"/bin/mail -s \"procmon issued alarm\" %s <%s",
		conf.procmon_email,tmpf);
         else
            sprintf(msg,"/bin/mail -s \"procmon issued alarm\" %s",
		conf.procmon_email);
#endif
         if (system(msg))
            syslog(LOG_ERR,"Error: Attempt to mail procmon alert to %s failed",
			conf.procmon_email);
      }
      CHKPOINT();
      if (alert_countdown > 0)
         alert_countdown = alert_countdown - conf.procmon_intv1;

      CHKPOINT();
      sleep(conf.procmon_intv1);
   }
}
