/*============================================================================
 * filename:	../fwmon/synmon.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=============================================================================
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1996-2000.  All rights
 * reserved.    
 *
 * Date written:	October    15, 1996
 * Last modified:	February   8, 2000
 *
 * function:
 *		Monitors TREX firewall against syn flood attacks
 *
 *============================================================================*/
static  char	ident[] = "@(#) synmon, V 1.0, 00/01/21"
 			  " (C) Freemont Avenue Software, Inc."
			  " 1996-2000";
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stddef.h>
#include <limits.h>
#ifdef _AIX
#include <sys/machine.h>
#include <sys/sysmacros.h>
#include <sys/sysconfig.h>
#include <storclass.h>
#endif /* _AIX */
#include <syslog.h>
#include <signal.h>
#include <sys/file.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netinet/tcp_timer.h>
#include <netinet/ip_var.h>
#include <netinet/tcpip.h>
#include <netinet/tcp_var.h>
#define TCPSTATES
#include <netinet/tcp_fsm.h>
#include <errno.h>
#include <netdb.h>
#include <nlist.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <net/route.h>
#include <netinet/in_pcb.h>
#include <sys/socketvar.h>
#include <setjmp.h>
#ifdef DEBUG_TIME
#include <sys/time.h>
#endif
#include "fwmon.h"
#include "firewall.h"

int synchk(struct synparms *synparms, unsigned char *warned,
#ifndef _AIX41
		struct inpcb *tcbp,
#endif
		struct tip *tipsp, unsigned long ntips,
		struct tip *utipsp, unsigned long nutips);

extern char    *sys_errlist[];

#ifndef _AIX41
struct nlist nl[] = {
#define	N_TCB		0
    { "tcb" },
#define NELEM_NLIST	1
    "",
};

#define _PATH_KMEM	"/dev/kmem"

char	*kmemf = _PATH_KMEM;
int	kmem;
#endif

void addpid(char *file);

static  jmp_buf env;
static	struct tip *tips;
static	unsigned long ntips=0;
static	size_t tipsz=0;
static	pid_t pmonpid;
/*===================================================================
 * Do not trust any reserved private IP addresses.
 * They are:
 * 10.*
 * 172.16.* to 172.31.*
 * 192.168.*
 *==================================================================*/
static	struct tip utips[3]={	{0x0a000000,0xff000000},
				{0xac100000,0xfff00000},
				{0xc0a80000,0xffff0000} };
static	unsigned long nutips=sizeof(utips)/sizeof(*utips);

extern	off_t lseek();

/*=======================================================================
 * exit handling
 *======================================================================*/
void pmexit() {
    kill(pmonpid, SIGUSR2);
    return;
}

#ifndef _AIX41
/*====================================================================
 * Write to kernel memory
 *
 * Input:
 *	off_t k_off	- offset into kernel memory
 *	char *buf	- buffer to write from
 *	int size	- number of bytes to write
 *
 * Output:
 *	returns the number of bytes written
 *===================================================================*/
int k_write(k_off, buf, size)
	off_t 	k_off;
	char	*buf;
	int	size;
{
	lseek(kmem,k_off,L_SET);
	return write(kmem, buf, size);
}

/*====================================================================
 * Read from kernel memory
 *
 * Input:
 *	off_t k_off	- offset into kernel memory
 *	char *buf	- buffer to read into
 *	int size	- number of bytes to read
 *
 * Output:
 *	returns the number of bytes read
 *===================================================================*/
int k_read(k_off, buf, size)
	off_t 	k_off;
	char	*buf;
	int	size;
{
	lseek(kmem,k_off,L_SET);
	return read(kmem, buf, size);
}
#endif

/*====================================================================
 * Builds a table to hold the trusted addresses
 *
 * Input:
 *	char *fn		- name of file which contains trusted IPs
 *				  to build table from.
 *
 * Output:
 *	0 = successful
 *	-1 = unsuccessful
 *===================================================================*/
int bld_tips(char *fn)
{
#define	CHUNK_SIZE 1024
   FILE *infile;
   char buf[BUFSIZ];
   int line=0;

   /*=================================================================
    * Open the input file
    *================================================================*/
   infile = fopen(fn,"r");
   if (!infile)
   {
	syslog(LOG_ERR,"Error: Cannot open %s for input.  %s.  ",
			fn,sys_errlist[errno]);
	return(-1);
   }

   /*=================================================================
    * Loop to read input file and generate tips entries
    *================================================================*/
   while (fgets(buf, BUFSIZ, infile) != NULL)
   {
	char *token,*mtoken;
	int i=0;
	unsigned long mask=0;
	unsigned long octet[4]={0,0,0,0};

	line++;
	/*=============================================================
	 * The first * or white space in the string terminates the address
	 * This implies that * by itself will not be processed.  This is
	 * OK since if you trust * (everything), you may as well not run
	 * synmon.
	 *============================================================*/
	if (!(token = strtok(buf,"* \t\n")) || *token == '#')
	    continue;

	/*=============================================================
	 * If a mask is specified, mtoken will be set to point to it
	 *============================================================*/
	mtoken = strtok(NULL,"* \t\n");

	/*=============================================================
	 * Get the specified address
	 *============================================================*/
	if (token = strtok(token,"."))
	    do
	    {
		octet[i++] = strtoul(token,NULL,10);
	    }
	    while (token = strtok(NULL,"."));

	/*=============================================================
	 * See if a mask is specified
	 *============================================================*/
	if (mtoken)
	{
	    unsigned long moctet[4]={0,0,0,0};
	    int j=0;
	    /*=========================================================
	     * A mask is specified, use it.
	     *========================================================*/
	    if (mtoken = strtok(mtoken,"."))
		do
		{
		    moctet[j++] = strtoul(mtoken,NULL,10);
		}
		while (mtoken = strtok(NULL,"."));
	    else
	    {
		/*=========================================================
		 * A mask is specified, use it.
		 *========================================================*/
		syslog(LOG_ERR,"Error: Syntax error on line %d of %s.  Line "
			"ignored.",line,fn);
		continue;
	    }
	    mask = moctet[0]<<24 | moctet[1]<<16 | moctet[2]<<8 | moctet[3];
	}
	else
	{
	    /*=========================================================
	     * A mask is not specified, calculate it from the IP.
	     * i still holds the number of octet specified in the IP.
	     *========================================================*/
	    switch (i) {
		case 1:
			mask = 0xff000000;
			break;
		case 2:
			mask = 0xffff0000;
			break;
		case 3:
			mask = 0xffffff00;
			break;
		case 4:
			mask = 0xffffffff;
			break;
	    }
	}

	/*=================================================================
	 * Allocate a chunk of memory for the tips array, if necessary
	 *================================================================*/
	if (sizeof(*tips)*ntips >= tipsz)
	{
	    if (ntips == 0)
		tips = malloc(CHUNK_SIZE);
	    else
		tips = realloc(tips,tipsz+CHUNK_SIZE);
	    if (tips == NULL)
	    {
		syslog(LOG_ERR,"Error: malloc of tips failed, %s.",
				sys_errlist[errno]);
		return(-1);
	    }
	    else
		tipsz += CHUNK_SIZE;
	}

	/*=============================================================
	 * Put ip and mask into table
	 *============================================================*/
	(tips+ntips)->ip = (octet[0]<<24 | octet[1]<<16 |
				octet[2]<<8 | octet[3]) & mask;
	(tips+ntips)->mask = mask;
	ntips++;
   }
   fclose(infile);
   return(0);

}

/*====================================================================
 * signal handler for SIGUSR1
 *===================================================================*/
void refresh(int sig)
{
   signal(SIGUSR1,refresh);
   longjmp(env,1);
}

/*====================================================================
 * synmon
 *===================================================================*/
main(int argc, char *argv[])
{
#ifdef _AIX
   unsigned char warned[65536];		/* Each element represent one port
				   0 means a syn flood alert has not been
				   issued for the port. */
   struct pmonconf pmonconf;
   struct synparms *synparms;
   struct inpcb pcb;
   struct inpcb *tcbp;
   struct socket sock;
   int ch;
   int rc;
   short exp=1;
   short tmp;
   pid_t pid;
   char *iname,*oname;
   FILE *sfile;
   struct cfg_load cfg;
   char *re="";

   openlog("synmon",LOG_PID|LOG_NDELAY,LOG_DAEMON);

   /*=======================================================================
    * Dissociate from controlling terminal
    *======================================================================*/
   if (setpgrp() == -1)
	syslog(LOG_NOTICE,"Cannot change process group, errno = %d\n",errno);

   /*===============================================================
    * Setup signal and setjmp so we can accept a USR1 signal to refresh
    *==============================================================*/
   signal(SIGUSR1,refresh);
   if (setjmp(env))
   {
	struct pmonconf tpmonconf;
	/*===============================================================
 	 * This is where the USR1 signal handler returns to.
	 *==============================================================*/
	re = "re-";

	/*===============================================================
 	 * Re-read the synmon parameters 
	 * ==============================================================*/
	if (pmconf(&tpmonconf))
	    syslog(LOG_ERR,"Error: %s, synmon continuing with previous "
				"set of parameters.",PMONCONFFILE);
	else
	{
	    if (pmonconf.synparms.plist)
		free(pmonconf.synparms.plist);
	    pmonconf = tpmonconf;
	}
   }
   else
	/*===============================================================
 	 * Read in the synmon parameters for the first time
	 *==============================================================*/
	if (rc = pmconf(&pmonconf))
	{
	    if (rc == 1)
		syslog(LOG_ERR,"Error: %s does not exist.",PMONCONFFILE);
	    exit(1);
	}

   synparms = &pmonconf.synparms;

   if (synparms->chk_interval == 0)
	exit(0);

   pmonpid = getppid();
   atexit(pmexit);

   addpid(SYNMONPID);
   addpid(FWMONPIDS);

   if (synparms->verbose)
	syslog(LOG_NOTICE,"synmon %sstarted.  interval = %d"
			"  qlen = %d"
			"  himark = %d"
			"  lomark = %d",
			re,
			synparms->chk_interval,
			synparms->qlen,
			synparms->himark,
			synparms->lomark);

   /*===============================================================
    *  Build the table of trusted IPs
    *==============================================================*/
   bld_tips(SECURENETFILE);
   if (synparms->trusted)
	bld_tips(synparms->trusted);

#ifndef _AIX41
   /*===================================================================
    * Open /dev/kmem to read in kernel structures
    *==================================================================*/
   kmem = open(kmemf, O_RDWR);
   if (kmem < 0) {
	perror(kmemf);
	exit(1);
   }

   /*===================================================================
    * Read in kernel symbols
    *==================================================================*/
   knlist(nl, NELEM_NLIST, sizeof (struct nlist));

   if (!nl[N_TCB].n_value) {
	syslog(LOG_ERR,"Error: Could not get address for TCB.  Cannot "
				"continue.");
	exit(1);
   }
#endif

   /*===================================================================
    * Initialize the warned array so we don't repeat issuing an alert for
    * the same port
    *==================================================================*/
   memset(warned,0,sizeof(warned));

   /*===================================================================
    * Check the entire queue every chk_interval
    *==================================================================*/
   while (1)
   {
	int n;
#ifdef DEBUG_TIME
	struct timeval       tst1;
	struct timeval       tst2;
	struct timezone      tz={0,0};
	if (synparms->verbose)
			gettimeofday(&tst1,&tz);
#endif

	/*======================================================
	 * Call the kernel extension
	 *=====================================================*/
#ifdef _AIX41
	if ((n = synchk(synparms,warned,tips,ntips,utips,nutips)) < 0)
#else
	if ((n = synchk(synparms,warned,(struct inpcb *)nl[N_TCB].n_value,
				tips,ntips,utips,nutips)) < 0)
#endif
	{
	    syslog(LOG_ERR,"Error: synchk returned %d, errno=%d",n,errno);
	}

#ifdef DEBUG_TIME
	if (synparms->verbose)
	{
	    gettimeofday(&tst2,&tz);
	    if (tst2.tv_usec < tst1.tv_usec)
	    {
			    tst1.tv_usec = tst2.tv_usec+1000000 - tst1.tv_usec;
			    tst2.tv_sec--;
	    }
	    else
		tst1.tv_usec = tst2.tv_usec- tst1.tv_usec;
	    tst1.tv_sec = tst2.tv_sec - tst1.tv_sec;
	    syslog(LOG_NOTICE,"Cleanup loop time = %d.%06d",
			tst1.tv_sec,tst1.tv_usec);
	}
#endif

	/*===================================================================
	 * sleep until chk_interval is up.
	 *==================================================================*/
	sleep(synparms->chk_interval);
   
   }

   exit(0);

#endif /* _AIX */
}
