/*============================================================================
 * file:	fwpulse/active_if.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=======================================================================
 *
 * (C) COPYRIGHT Freemont Avenue Software, Inc. 1999-2000
 * All rights reserved.   
 *											    
 * Date written:	April      8, 1999
 * Last modified:	February   8, 2000
 *
 *
 *=========================================================================*/
#include	<errno.h>
#include	<sys/types.h>
#include	<sys/socket.h>
#include	<syslog.h>
#include	<net/if.h>
#include	"fwpulse.h"
#include	"get_ifnet.h"
#include	"getkmemdata.h"

#ifdef _AIX
/*============================================================================
 * Given an ip address, this function looks in the kernel data to see if the
 * interface associated with the specified ip address has been active since
 * the last call.
 * inputs:	ip address that is not responding
 *		cfginfo
 * 
 * output:
 *		returns 0 if interface is not active
 *		returns non-zero if interface has activities
 *		If there was an error getting the kernel structures that we
 *		we need, we have to return 0 as if the interface is down.
 *===========================================================================*/
int active_if(unsigned long ip, struct cfginfo *cfginfo) {
    struct ifnet	*ifnet = NULL;
    struct sockaddr_in	sock;
    struct ifaddr	ifa;
    struct ifaddr	*ifp;
    int			rc;
    int			found = 0;
    unsigned long	ibytes;
    unsigned long	obytes;
    unsigned long	ipackets;
    unsigned long	opackets;
    unsigned long	ierrors;
    unsigned long	oerrors;
    unsigned long	iqdrops;

    /*=======================================================================
     * Following the kernel structures
     *
     *         ifnet                  ifaddr              sockaddr
     *         +-----------+          +--------+          +-------+
     * ifnet-->|           |    ----->|ifa_addr|--------->|       |
     *         |if_addrlist|----+     |        |          |       |
     *         |           |          |        |          |       |
     *         |           |          |        |          |       |
     *    +----|if_next    |    +-----|ifa_next|          |       |
     *    |    +-----------+    |     +--------+          +-------+
     *    |                     |
     *    |    +-----------+    |     +--------+          +-------+
     *    +--->|           |    +---->|ifa_addr|--------->|       |
     *         |if_addrlist|---+      |        |          |       |
     *         |           |   |      |        |          |       |
     *         |           |   |      |        |          |       |
     *    +----|if_next    |   |      |        |          |       |
     *    |    +-----------+   |      +--------+          +-------+
     *    |                    |
     *    |                    |
     *    +--->      .         +----->     .                  .
     *               .                     .                  .
     *               .                     .                  .
     *======================================================================*/
    if (get_ifnet(&ifnet) == 0) {
	syslog(LOG_ERR, "Error: Unable to read kernel data ifnet.  %m.");
	return(0);
    }

    while (1) {
	for (ifp = ifnet->if_addrlist; ifp != NULL; ifp = ifa.ifa_next) {
	    if ((rc = getkmemdata((char *)&ifa, sizeof(ifa),
					(caddr_t)ifp) < sizeof(ifa))) {
		syslog(LOG_ERR, "Error: Unable to read kernel data ifaddr.  "
				"rc %d, expecting %d.  %m.",
				rc, sizeof(ifa));
		return(0);
	    }
	    if (ifa.ifa_addr == NULL) {
		syslog(LOG_ERR, "Error: Unexpected null pointer in ifa_addr.");
		return(0);
	    }
	    if ((rc = getkmemdata((char *)&sock, sizeof(sock),
				(caddr_t)ifa.ifa_addr)) != sizeof(sock)) {
		syslog(LOG_ERR, "Error: unable to read kernel data sockaddr.  "
				"rc %d, expecting %d.  %m.",
				rc, sizeof(sock));
		return(0);
	    }
	    if (ip == sock.sin_addr.s_addr) {
		found = 1;
		break;
	    }
	}
	if (found)
	    break;

	if (ifnet->if_next == NULL)
	    break;

	if ((rc = getkmemdata((char *)ifnet, sizeof(struct ifnet),
			(caddr_t)ifnet->if_next)) != sizeof(struct ifnet)) {
	    syslog(LOG_ERR, "Error: unable to read kernel data if_next.  "
				"rc %d, expecting %d.  %m.",
				rc, sizeof(sock));
	    return(0);
	}
    }
    if (!found) {
	syslog(LOG_ERR, "Error: Could not find ip address %s that was marked "
			"as down by remote fwpulse.  fwpulsed will assume "
			"that it is down.", inet_ntoa(ip));
	return(0);
    }

    /*=======================================================================
     * We have found the suspect ip address in the kernel structures.  Now
     * get the statistics from the interface and see if there are activities.
     *======================================================================*/
    ipackets = ifnet->if_ipackets;
    opackets = ifnet->if_opackets;
    ibytes = ifnet->if_ibytes;
    obytes = ifnet->if_obytes;
    ierrors = ifnet->if_ierrors;
    oerrors = ifnet->if_oerrors;
    iqdrops = ifnet->if_iqdrops;

    /*=======================================================================
     * Sleep for 1/3 threshold and get statistics again
     *======================================================================*/
    sleep((cfginfo->threshold)/3);
    
    /*=======================================================================
     * Use the back pointer in ifaddr to get the interfacce ifnet struct
     *======================================================================*/
    if (ifa.ifa_ifp == NULL) {
	syslog(LOG_ERR, "Error: Unexpected null pointer in ifa_ifp.");
	return(0);
    }
    if ((rc = getkmemdata((char *)ifnet, sizeof(struct ifnet),
			(caddr_t)ifa.ifa_ifp)) != sizeof(struct ifnet)) {
	syslog(LOG_ERR, "Error: Unable to read kernel data ifnet using "
			"ifa_ifp.  rc %d, expecting %d.  %m",
			rc, sizeof(struct ifnet));
	return(0);
    }
    syslog(LOG_INFO, "Interface summary for ip %s: ibytes = %u, obytes = "
		"%u, ipackets = %u, opackets = %u, ierrors = %u, "
		"oerrors = %u, iqdrops = %u.",
		inet_ntoa(ip),
		(unsigned long)ifnet->if_ibytes - ibytes,
		(unsigned long)ifnet->if_obytes - obytes,
		(unsigned long)ifnet->if_ipackets - ipackets,
		(unsigned long)ifnet->if_opackets - opackets,
		(unsigned long)ifnet->if_ierrors - ierrors,
		(unsigned long)ifnet->if_oerrors - oerrors,
		(unsigned long)ifnet->if_iqdrops - iqdrops);

    /*=======================================================================
     * If any of the io statistics has changed, conclude there was activities.
     *======================================================================*/
    if (	ibytes != ifnet->if_ibytes ||
		obytes != ifnet->if_obytes ||
		ipackets != ifnet->if_ipackets ||
		opackets != ifnet->if_opackets	) {
	return(1);
    }

    return(0);
}
#else /* not AIX */
int active_if(unsigned long ip, struct cfginfo *cfginfo) {
    return(0);
}
#endif /* AIX */
