/*============================================================================
 * file:	fwpulse/arp_broadcast.c 
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=======================================================================
 *  
 * function:	This function sends out a gratuitous ARP WHO-HAS request for
 *		each IP address on a machine.  This
 *		should prompt remote machines to update their ARP cache to
 *		reflect IP/ethernet address pairs that may have just been
 *		changed.
 * 
 * (C) COPYRIGHT Freemont Avenue Software, Inc. 1999-2000 All rights reserved.
 *                                                                      
 * Date written:	May       16, 1999
 * Last modified:	February   8, 2000 - added LPL
 *                                                                      
 *=========================================================================*/
#if !defined(_AIX) && !defined(linux)
int main() {
}
#else
#include	<stdio.h>   
#include	<syslog.h>   
#include	<sys/ioctl.h>
#include	<errno.h>
#include	<stdlib.h>
#include	<sys/types.h> 
#include	<unistd.h>
#include	<sys/socket.h>
#include	<sys/socketvar.h>
#include	<netinet/in.h> 
#include	<net/if.h> 
#include	<netinet/if_ether.h> 
#include	<net/if_arp.h>
#ifdef _AIX
#include	<net/af.h> 
#include	<net/if_dl.h> 
#include	<sys/ndd_var.h> 

extern char *sys_errlist[];

#define MAXIF	16	/* Max 16 interfaces assumed	*/
#define MACSIZE	6	/* Size of a MAC address	*/

typedef struct {
    unsigned char dst[MACSIZE];
    unsigned char src[MACSIZE];
    unsigned short ethertype;
    unsigned char  data[1500];
} xmit;
#endif

/*=========================================================================
 * This program sends an ARP-WHOHAS request for
 * each IP address defined on the machine.
 * input:
 *
 *========================================================================*/
int main()
{  
   struct arphdr	*hdrp;
   int			fd;
   int			outfd;
   long			i,x; 
   int			n,len,lastlen;
#ifdef _AIX
   int			nif = 0;
   char			ifnames[MAXIF][IFNAMSIZ];
   char			macs[MAXIF][MACSIZE];
   unsigned long	brdcastaddrs[MAXIF];
   xmit			xbuf;
   char			*macp;
   struct sockaddr_dl	*sdl;
   struct sockaddr_ndd_8022 sa;
#endif
   struct arphdr	*ahdr;
   struct ifconf	ifc;
   struct ifreq		*ifb,ifr,*ifrp,*ifend,*ifnext;


   /*=======================================================================
    * Get a list of all the local IP addresses
    *======================================================================*/
   fd = socket(AF_INET, SOCK_STREAM, 0);
#ifdef SIOCGIFNUM
   if (ioctl(fd, SIOCGIFNUM, &n) < 0)
   {
	fprintf(stderr,"Error: SIOCGIFNUM error, %s\n",sys_errlist[errno]);
	close(fd);
	return(1);
   }
   ifc.ifc_len = n * sizeof(struct ifreq);
   if ((ifb = malloc(ifc.ifc_len)) == NULL)
   {
	fprintf(stderr,"Error: malloc() failed for ifreq.  %s\n",
				sys_errlist[errno]);
	close(fd);
	return(1);
   }
   ifc.ifc_buf = (caddr_t)ifb;
   if (ioctl(fd,SIOCGIFCONF,(char *)&ifc) < 0)
   {
	fprintf(stderr,"Error: SIOCGIFCONF error, %s\n",sys_errlist[errno]);
	close(fd);
	return(1);
   }
#else	/* SIOCGIFNUM	*/
   lastlen = 0;
   len = 100 * sizeof(struct ifreq);
   for ( ; ; )
   {
	if ((ifb = malloc(len)) == NULL)
	{
	    fprintf(stderr,"Error: malloc() failed for ifreq.  %s\n",
				sys_errlist[errno]);
	    close(fd);
	    return(1);
	}
	ifc.ifc_len = len;
	ifc.ifc_buf = (caddr_t)ifb;
	if (ioctl(fd,SIOCGIFCONF,(char *)&ifc) < 0)
	{
	    if (errno != EINVAL || lastlen != 0)
	    {
		fprintf(stderr,"Error: SIOCGIFCONF error, %s\n",
						sys_errlist[errno]);
		close(fd);
		return(1);
	    }
	}
	else
	{
	    if (ifc.ifc_len == lastlen)
		break;    
	    lastlen = ifc.ifc_len;
	}
	len += 10 * sizeof(struct ifreq);
	free(ifb);
   }
#endif

   /*====================================================================
    * Go thru all the address entries
    *===================================================================*/
   ifend = (struct ifreq *)((char *)ifb + ifc.ifc_len);
   for (ifrp=ifb; ifrp<ifend; ifrp=ifnext)
   {
	struct sockaddr *sia = &(ifrp->ifr_addr);
#if !defined(_AIX)	/* must be linux	*/
	ifnext = ifrp + 1;
#else /* !defined _AIX	*/
	n = ifrp->ifr_addr.sa_len + sizeof(ifrp->ifr_name);
	if (n < sizeof(*ifrp))
	    ifnext = ifrp + 1;
	else
	    ifnext = (struct ifreq *)((char *)ifrp + n);
#endif
	strncpy(ifr.ifr_name, ifrp->ifr_name, sizeof(ifr.ifr_name));
	if (ioctl(fd,SIOCGIFFLAGS,(char *)&ifr) < 0) {
	    fprintf(stderr,"Error: SIOCGIFFLAGS error, errno = %d\n",errno);
	    return(1);
	}
	if (ifr.ifr_flags & IFF_LOOPBACK)
	    continue;

	if (((struct sockaddr_in *)sia)->sin_family != AF_INET)
	    continue;
	fprintf(stderr,"ifname = \"%s\", addr \"%s\"\n",
		ifrp->ifr_name,
		inet_ntoa(((struct sockaddr_in *)&ifrp->ifr_addr)->sin_addr));
	if (arpwhohas(ifrp->ifr_name,
		((struct sockaddr_in *)&ifrp->ifr_addr)->sin_addr) != 0) {
	    fprintf(stderr, "Failed to send arpwhohas broadcast, look in "
			    "syslog for more information.\n");
	    return(1);
	}
   }

   close(fd);
   return(0);
}
#endif /* !defined(_AIX) && !defined(linux)	*/
