/*============================================================================
 * file:	fwpulse/arpwhohas.c 
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=======================================================================
 *  
 * function:	Given an interface name and an ip address (assumed to be
 *		defined on the interface), this function sends out a gratuitous
 *		ARP WHO-HAS request for that IP address on the interface.  This
 *		should prompt remote machines to update their ARP cache to
 *		reflect IP/ethernet address pairs that may have just been
 *		changed.
 *
 *		Note: This function uses the AIX specific AF_NDD interface so it
 *			won't work on another OS.  Sparc does the arp-whohas
 *			automatically after an IP address is newly defined so
 *			it is not necessary to send an arp-whohas ourselves
 *			on sparc.
 * 
 * (C) COPYRIGHT Freemont Avenue Software, Inc. 1999-2000 All rights reserved.
 *                                                                      
 * Date written:	April     27, 1999
 * Last modified:	February   8, 2000 - added LPL
 *                                                                      
 *=========================================================================*/
#if defined(_AIX) || defined (linux)

#include	<syslog.h>   
#include	<sys/ioctl.h>
#include	<errno.h>
#include	<stdlib.h>
#include	<sys/types.h> 
#include	<unistd.h>
#include	<sys/socket.h>
#include	<sys/socketvar.h>
#include	<netinet/in.h> 

#ifdef _AIX
#include	<net/if.h> 
#include	<net/if_arp.h>
#include	<netinet/if_ether.h> 
#include	<net/af.h> 
#include	<sys/ndd_var.h> 
#include	<net/if_dl.h> 
extern char *sys_errlist[];
#else
#include	<linux/if.h>
#include	<linux/if_arp.h>
#endif

/*=========================================================================
 * Given an interface name and an IP address (assumed to be defined on the
 * interface), this function sends an ARP-WHOHAS request for the specified
 *  IP address so that neighboring machines can use the hint to update their
 * arp cache which may not have the correct information anymore after an
 * autotakeover.
 * input:
 *	ifname	name of the interface (e.g. en0)
 *	ip	ipaddress for which an arp who-has needs to be sent
 *
 * returns	0 if successful
 *		1 if failed
 *========================================================================*/
int arpwhohas(char *ifname, unsigned long ip)
{

#ifdef _AIX
#define MIN_PACKET_SIZE	60	/* Has to be at least 60 for AF_NDD to work */

    typedef struct {
	unsigned char dst[6];
	unsigned char src[6];
	unsigned short ethertype;
	unsigned char  data[100];
    } xmit;
   struct arphdr	*hdrp;
   int			fd;
   int			outfd;
   long			i,x; 
   int			n,len,lastlen;
   char			mac_addr[6];
   xmit			xbuf;
   unsigned long	toaddr;
   struct arphdr	*ahdr;
   char			*macp;
   struct ifconf	ifc;
   struct ifreq		*ifb,ifr,*ifrp,*ifend,*ifnext;
   struct sockaddr_dl	*sdl;
   struct sockaddr_ndd_8022 sa;


   /*=======================================================================
    * Get the network configuration
    *======================================================================*/
   fd = socket(AF_INET,SOCK_STREAM,0);
#ifdef SIOCGIFNUM
   if (ioctl(fd, SIOCGIFNUM, &n) < 0)
   {
	syslog(LOG_ERR,"Error: SIOCGIFNUM error, %s",sys_errlist[errno]);
	close(fd);
	return(1);
   }
   ifc.ifc_len = n * sizeof(struct ifreq);
   if ((ifb = malloc(ifc.ifc_len)) == NULL)
   {
	syslog(LOG_ERR,"Error: malloc() failed for ifreq.  %s",
				sys_errlist[errno]);
	close(fd);
	return(1);
   }
   ifc.ifc_buf = (caddr_t)ifb;
   if (ioctl(fd,SIOCGIFCONF,(char *)&ifc) < 0)
   {
	syslog(LOG_ERR,"Error: SIOCGIFCONF error, %s",sys_errlist[errno]);
	close(fd);
	return(1);
   }
#else
   lastlen = 0;
   len = 100 * sizeof(struct ifreq);
   for ( ; ; )
   {
	if ((ifb = malloc(len)) == NULL)
	{
	    syslog(LOG_ERR,"Error: malloc() failed for ifreq.  %s",
				sys_errlist[errno]);
	    close(fd);
	    return(1);
	}
	ifc.ifc_len = len;
	ifc.ifc_buf = (caddr_t)ifb;
	if (ioctl(fd,SIOCGIFCONF,(char *)&ifc) < 0)
	{
	    if (errno != EINVAL || lastlen != 0)
	    {
		syslog(LOG_ERR,"Error: SIOCGIFCONF error, %s",
						sys_errlist[errno]);
		close(fd);
		return(1);
	    }
	}
	else
	{
	    if (ifc.ifc_len == lastlen)
		break;    
	    lastlen = ifc.ifc_len;
	}
	len += 10 * sizeof(struct ifreq);
	free(ifb);
   }
#endif

   /*====================================================================
    * Go thru all the address entries once to get the link addresses so
    * we can find the hardware address and the broadcast address.
    *===================================================================*/
   ifend = (struct ifreq *)((char *)ifb + ifc.ifc_len);
   for (ifrp=ifb; ifrp<ifend; ifrp=ifnext)
   {
	struct sockaddr *sia = &(ifrp->ifr_addr);
#if !defined(_AIX) && !defined(__Lynx__)
	ifnext = ifrp + 1;
#else
	n = ifrp->ifr_addr.sa_len + sizeof(ifrp->ifr_name);
	if (n < sizeof(*ifrp))
	    ifnext = ifrp + 1;
	else
	    ifnext = (struct ifreq *)((char *)ifrp + n);
#endif

	/*===============================================================
	 * Look for the specified interface
	 *==============================================================*/
	if (strcmp(ifrp->ifr_name, ifname) != 0)
	    continue;

	/*===============================================================
	 * Look for sin_family AF_LINK.
	 *==============================================================*/
	if (((struct sockaddr_in *)sia)->sin_family != AF_LINK)
	    continue;

	/*===============================================================
	 * Save the hardware address of the interface
	 *==============================================================*/
	sdl = (struct sockaddr_dl *)sia;
	memcpy(mac_addr, LLADDR(sdl), sizeof(mac_addr));

	/*====================================================================
	 * Does interface support broadcast?
	 *===================================================================*/
	strncpy(ifr.ifr_name, ifrp->ifr_name, sizeof(ifr.ifr_name));
			
	if (ioctl(fd,SIOCGIFFLAGS,(char *)&ifr) < 0)
	{
	    syslog(LOG_ERR,"Error: SIOCGIFFLAGS error, %s\n",
				sys_errlist[errno]);
	    close(fd);
	    return(1);
	}
	if ((ifr.ifr_flags & IFF_BROADCAST)) {
	    /*===========================================================
	     * Get broadcast address
	     *==========================================================*/
	    strncpy(ifr.ifr_name, ifrp->ifr_name, sizeof(ifr.ifr_name));
	    if (ioctl(fd,SIOCGIFBRDADDR,(char *)&ifr) < 0)
	    {
		syslog(LOG_ERR,"Error: SIOCGIFBRDADDR error, %s\n",
				sys_errlist[errno]);
		close(fd);
		return(1);
	    }

	    toaddr = ((struct sockaddr_in *)&(ifr.ifr_addr))->sin_addr.s_addr;
	}
	else
	    toaddr = ip;

	break;
   }

   free(ifb);

    /*===============================================================
     * Build the raw packet
     *==============================================================*/
    memset(&xbuf, 0, sizeof(xbuf));
    memset(xbuf.dst, 0xff, sizeof(xbuf.dst));
    memcpy(xbuf.src, &mac_addr, sizeof(xbuf.src));
    xbuf.ethertype = ETHERTYPE_ARP;
    ahdr = (struct arphdr *)&xbuf.data;
    ahdr->ar_hrd = ARPHRD_ETHER;
    ahdr->ar_pro = ETHERTYPE_IP;
    ahdr->ar_hln = ETHER_ADDRLEN;
    ahdr->ar_pln = sizeof(struct in_addr);
    ahdr->ar_op  = ARPOP_REQUEST;
    macp = (char *)ahdr + sizeof(struct arphdr);
    memcpy(macp, mac_addr, sizeof(mac_addr));
    memcpy(macp + sizeof(mac_addr), &ip, sizeof(struct in_addr));
    memcpy(macp + sizeof(mac_addr)+sizeof(mac_addr)+sizeof(struct in_addr),
		&ip, sizeof(struct in_addr));
    
    /*===============================================================
     * Open the AF_NDD interface
     *==============================================================*/
    if ((outfd = socket(AF_NDD, SOCK_DGRAM, 0)) < 0) {
	syslog(LOG_ERR, "AF_NDD socket create failed.  %s.",
						sys_errlist[errno]);
	close(fd);
	return(1);
    }

    memset(&sa, 0, sizeof(sa));
    sa.sndd_8022_family = AF_NDD;
    sa.sndd_8022_len = sizeof(sa);
    sa.sndd_8022_filtertype = NS_ETHERTYPE;
    sa.sndd_8022_ethertype = ETHERTYPE_ARP;
    sa.sndd_8022_filterlen = sizeof(struct ns_8022);
    strcpy(sa.sndd_8022_nddname, ifname);

    if (connect(outfd, (struct sockaddr *)&sa, sizeof(sa))) {
	syslog(LOG_ERR, "connect to ndd failed.  %s.", sys_errlist[errno]);
	close(fd);
	close(outfd);
	return(1);
    }

    /*===============================================================
     * Write the raw arp-whohas packet
     *==============================================================*/
    if (write(outfd, &xbuf, MIN_PACKET_SIZE) < 0) {
	syslog(LOG_ERR, "AF_NDD send failed.  %s.", sys_errlist[errno]);
	close(fd);
	close(outfd);
	return(1);
    }

   close(outfd);
   close(fd);
   return(0);
}

#else /* _AIX */
#ifdef linux
    typedef unsigned long in_addr_t;
    extern const char *const sys_errlist[];
    int fd;
    int len;
    char *cp;
    char buf[100];
    struct ifreq ifr;
    struct arphdr *ahdr;
    struct sockaddr_ll myaddr;
    struct sockaddr_ll toaddr;

    /*=======================================================================
     * Get index for this interface
     *======================================================================*/
    if ((fd = socket(PF_PACKET, SOCK_DGRAM, 0)) < 0) {
	syslog(LOG_ERR, "Error: Unable to create socket. %m.");
	return(1);
    }
    memset(&ifr, 0, sizeof(ifr));
    strcpy(ifr.ifr_name, ifname);
    if (ioctl(fd, SIOCGIFINDEX, &ifr) < 0) {
	syslog(LOG_ERR, "Error: Unable to find interface %s.  %m.", ifname);
	return(1);
    }
    
    /*=======================================================================
     * Bind the interface
     *======================================================================*/
    memset(&myaddr, 0, sizeof(myaddr));
    myaddr.sll_family = AF_PACKET;
    myaddr.sll_ifindex = ifr.ifr_ifindex;
    myaddr.sll_protocol = htons(ETH_P_ARP);
    if (bind(fd, (struct sockaddr*)&myaddr, sizeof(myaddr)) == -1) {
	syslog(LOG_ERR, "Error: bind to interface %s failed.  %m", ifname);
	return(1);
    }
    
    /*=======================================================================
     * Get the socket info
     *======================================================================*/
    len = sizeof(myaddr);
    if (getsockname(fd, (struct sockaddr *)&myaddr, &len) < 0) {
	 syslog(LOG_ERR, "Error: Unable to get socket info.  %m");
	 return(1);
    }
    
    memcpy(&toaddr, &myaddr, sizeof(toaddr));
    memset(toaddr.sll_addr, 0xff, toaddr.sll_halen);
    /*=======================================================================
     * Setup arp header
     *======================================================================*/
    ahdr = (struct arphdr *)buf;
    ahdr->ar_hrd = htons(myaddr.sll_hatype);
    ahdr->ar_pro = htons(ETH_P_IP);
    ahdr->ar_hln = myaddr.sll_halen;
    ahdr->ar_pln = 4;
    ahdr->ar_op = htons(ARPOP_REQUEST);
    
    cp = buf + sizeof(struct arphdr);
    memcpy(cp, myaddr.sll_addr, ahdr->ar_hln);
    cp += ahdr->ar_hln;
    memcpy(cp, &ip, sizeof(ip));
    cp += sizeof(ip);
    memcpy(cp, myaddr.sll_addr, ahdr->ar_hln);
    cp += ahdr->ar_hln;
    memcpy(cp, &ip, sizeof(ip));
    
    /*=======================================================================
     * Send request
     *======================================================================*/
    if (sendto(fd, buf, cp + sizeof(ip) - buf, 0, (struct sockaddr*)&toaddr,
				sizeof(toaddr)) < 0) {
	syslog(LOG_ERR, "Error: Unable to send ARPWHOHAS, ip = %s.  %m",
				inet_ntoa(ip));
	return(1);
    }
    
    return(0);
}

#endif /* linux */
#endif /* _AIX */
#endif /* _AIX  || linux */
