/*============================================================================
 * file:	fwpulse/autotakeover.c  
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=======================================================================
 *  
 * function:	Automatically perform the takeover procedure when a failure
 *		is detected on the remote firewall.
 * 
 * (C) COPYRIGHT Freemont Avenue Software, Inc. 1999-2000
 * All rights reserved.   
 *                                                                      
 * Date written:	December   9, 1998     
 * Last modified:	February   8, 2000 - added LPL
 *                                                                      
 *=========================================================================*/
#include	<ctype.h>  
#include	<stdio.h>  
#include	<stdlib.h>  
#include	<string.h>  
#include	<syslog.h>  
#include	<limits.h>  
#include	<errno.h>  
#include	<sys/wait.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<sys/socket.h>
#include	<sys/ioctl.h>
#include	<net/if.h>
#include	<net/if_arp.h>
#include	<netinet/in.h>
#include	"des.h"
#include	"fwpulse.h"

#ifndef linux
extern char	*sys_errlist[];
#else
extern const char	*const sys_errlist[];
#endif

int arpwhohas(char *ifname, unsigned long ip);
#ifdef DES
void pingloop(struct cfginfo *cfginfo, int *active, int *inactive,
		des_key_schedule ksched);
#else
void pingloop(struct cfginfo *cfginfo, int *active, int *inactive);
#endif

/*==========================================================================
 * Note that this function does not return, it only exits.
 *
 * Output:
 *	exit 0 if successful
 *	exit 1 if failed and function cannot proceed with takeover
 *=========================================================================*/
#ifdef DES
int autotakeover(int inactive, int active, struct cfginfo *cfginfo,
			des_key_schedule ksched)
#else
int autotakeover(int inactive, int active, struct cfginfo *cfginfo)
#endif
{ 
   int		fd;  
   int		rc;
   int		i;
#ifndef linux
   int		zero = 0;
#else
   struct timezone zero;
#endif
   int		attempts = 0;
   char		msg[BUFSIZ];
   char		buf[BUFSIZ];
   FILE		*revfile;
   FILE		*scriptfile;
   struct timeval	tv;

   sprintf(msg, "Checking to make sure remote firewall is down before "
		"auto-takeover.");
   syslog(LOG_NOTICE, msg);

   /*=======================================================================
    * If everything fails against us, we will try MAX_ATTEMPTS times and quit.
    *======================================================================*/
   while (active && ++attempts <= MAX_ATTEMPTS) {
	commpkt	pkt;
	commpkt	resp;
#ifdef DES
	int j;
	char *cp;
	char ebuf[EBUFSIZ];
#endif
	/*=================================================================
	 * Try to connect to the target system's private port
	 *================================================================*/
	for (i = 0; i < cfginfo->nif; i++) {
	    if (cfginfo->net[i].local || cfginfo->net[i].alias ||
			cfginfo->net[i].notakeover)
		continue;
	    alarm(cfginfo->threshold);
	    if ((fd = connectserver(cfginfo->net[i].ip,
				cfginfo->private_port, 0, NULL)) >= 0) {
		alarm(0);
		break;
	    }
	    alarm(0);
	}

	if (fd < 0)
	    break;
	/*===================================================================
	 * Find an interface that is suspected to be down
	 *==================================================================*/
	for (j = 0; j < cfginfo->nif; j++) {
	    if (cfginfo->net[j].down_suspect) {
		if (j == i) {
		    /*===================================================
		     * We have a conflict here, we just connected to this
		     * interface but it was suspected to be down.
		     *==================================================*/
		    syslog(LOG_NOTICE, "Remote ip address (%s) that was marked "
				"as not responsive is now responding, "
				"autotakeover aborting.", cfginfo->net[j].ip);
		    return(REJECTED);
		}
		break;
	    }
	}
	/*===================================================================
	 * Should not happen, but just in case
	 *==================================================================*/
	if (j >= cfginfo->nif) {
	    syslog(LOG_ERR, "Error: Internal error, autotakeover is requested "
			"but no interface is marked as down");
#ifdef __sun
#define INADDR_NONE     0xffffffff
#endif
	    pkt.down_ip = INADDR_NONE;
	}
	else
	    pkt.down_ip = inet_addr(cfginfo->net[j].ip);

	pkt.dup_ip = pkt.down_ip;
	/*===================================================================
	 * We were successful in connecting to one interface on the
	 * target system.  We can request it to shutdown all its interfaces
	 *==================================================================*/
	pkt.code = htonl(SHUTDOWN);
	pkt.dup_code = pkt.code;
	gettimeofday(&tv, &zero);
	pkt.current_time.tv_sec = htonl(tv.tv_sec);
	pkt.current_time.tv_usec = htonl(tv.tv_usec);
	pkt.dup_time.tv_sec = pkt.current_time.tv_sec;
	pkt.dup_time.tv_usec = pkt.current_time.tv_usec;

#ifdef DES
	/*===============================================================
	 * Encrypt
	 *==============================================================*/
	cp = (char *)&pkt;
	for (j = 0; j < EBUFSIZ; j += 8)
	    des_ecb_encrypt((des_cblock *)(cp + j), (des_cblock *)(ebuf + j),
			ksched, DES_ENCRYPT);
#endif

	alarm(cfginfo->threshold);
#ifdef DES
	if (write(fd, ebuf, sizeof(ebuf)) != sizeof(pkt)) {
#else
	if (write(fd, &pkt, sizeof(pkt)) != sizeof(pkt)) {
#endif
	    alarm(0);
	    errno = 0;
	    close(fd);
	    if (tv.tv_sec - cfginfo->net[i].lastlogged >= LONGWAIT) {
		sprintf(msg, "Error: Unable to send SHUTDOWN request to "
				"remote firewall %s.  %s.  Attempt will be "
				"made again in %d seconds.",
				cfginfo->net[i].ip, sys_errlist[errno],
				cfginfo->threshold);
		syslog(LOG_ERR, msg);
		sprintf(buf, "/usr/bin/mailx -s \"%s\" %s </dev/null",
			msg, cfginfo->email);
		if (rc = system(buf))
		    syslog(LOG_ERR,"Error: Attempt to mail attention "
			"notification to %s failed.  rc = 0x%08x.",
			cfginfo->email, rc);
		cfginfo->net[i].lastlogged = tv.tv_sec;
	    }
	}
	else {
	    /*===========================================================
	     * Use 2 times threshold here since the remote fwpulsed will
	     * be issuing pings to its routers and that may take time before
	     * it can respond.
	     *==========================================================*/
	    alarm(cfginfo->threshold * 2);
#ifdef DES
	    if (read(fd, ebuf, sizeof(ebuf)) <= 0) {
#else
	    if (read(fd, &resp, sizeof(resp)) <= 0) {
#endif
		/*=======================================================
		 * Note that a read failure or timeout here does not give
		 * us a positive indication whether the remote is down and
		 * we will have to probe again anyway.  No warning need to
		 * be issued here.
		 *======================================================*/
		alarm(0);
		errno = 0;
		close(fd);
	    }
	    else {
		alarm(0);
#ifdef DES
		/*=======================================================
		 * Decrypt
		 *======================================================*/
		cp = (char *)&resp;
		for (j = 0; j < EBUFSIZ; j += 8)
		    des_ecb_encrypt((des_cblock *)(ebuf + j),
					(des_cblock *)(cp + j),
					ksched, DES_DECRYPT);
#endif
		if (ntohl(resp.code) != OK) {
		    errno = 0;
		    close(fd);
		    if (tv.tv_sec - cfginfo->net[i].lastlogged >= LONGWAIT) {
			sprintf(msg, "Error: Return code from remote firewall "
				"%s indicates SHUTDOWN request was rejected.  "
				"Aborting automatic takeover for now.",
				cfginfo->net[i].ip);
			syslog(LOG_ERR, msg);
			sprintf(buf, "/usr/bin/mailx -s \"%s\" %s </dev/null",
				msg, cfginfo->email);
			if (rc = system(buf))
			    syslog(LOG_ERR,"Error: Attempt to mail attention "
				"notification to %s failed.  rc = 0x%08x.",
				cfginfo->email, rc);
			cfginfo->net[i].lastlogged = tv.tv_sec;
			return(ntohl(resp.code));
	            }
		}
	    }
	}

	close(fd);

	/*===============================================================
	 * Now do the pingloop again to make sure all interfaces are down
	 *==============================================================*/
#ifdef DES
	pingloop(cfginfo, &active, &inactive, ksched);
#else
	pingloop(cfginfo, &active, &inactive);
#endif

	if (active)
	    sleep(cfginfo->threshold);
   }

   /*=======================================================================
    * If attempted at least MAX_ATTEMPTS times, then conclude that there's
    * not much else we can do.
    * Note the index i (in cfginfo->net[i].ip) here looks incorrect but it's
    * really OK.  i is set in the for loop within the while loop above.
    * The only time this is true is when the break statement is executed
    * in the for loop after a successful connection to the remote firewall.
    * It subsequently failed to get the remote system to shutdown after
    * MAX_ATTEMPTS tries.  i should still be pointing to a valid member.
    *======================================================================*/
   if (attempts > MAX_ATTEMPTS) {
	sprintf(msg, "Alert: Attempted %d times to shutdown "
			"remote firewall %s with no success, autotakeover "
			"aborting, immediate attention is required.",
			MAX_ATTEMPTS, cfginfo->net[i].ip);
	syslog(LOG_ALERT, msg);
	sprintf(buf, "/usr/bin/mailx -s \"%s\" %s </dev/null",
		msg, cfginfo->email);
	if (rc = system(buf))
	    syslog(LOG_ERR,"Error: Attempt to mail attention "
			"notification to %s failed.  rc = 0x%08x.",
			cfginfo->email, rc);
	exit(1);
   }

   /*=======================================================================
    * Notify that we are starting the takeover
    *======================================================================*/
   sprintf(msg, "Attention: Begin attempt to auto-takeover remote firewall.");
   syslog(LOG_NOTICE, msg);
   sprintf(buf, "/usr/bin/mailx -s \"%s\" %s </dev/null", msg, cfginfo->email);
   if (rc = system(buf))
	syslog(LOG_ERR,"Error: Attempt to mail attention notification to %s "
		"failed.  rc = 0x%08x.",
		cfginfo->email, rc);

   /*=======================================================================
    * Invoke the pre_takeover script if it exists
    *======================================================================*/
   if ((scriptfile = fopen(PRETAKEOVER, "r")) != NULL)
   {
	int rc;
	fclose(scriptfile);
	rc = system(PRETAKEOVER);
	if (WEXITSTATUS(rc) != 0)
	{
	    syslog(LOG_ERR, "Error: autotakeover: Non-zero return code "
			"(0x%08x) from %s, autotakeover not proceeding.",
			rc, PRETAKEOVER);
	    exit(1);
	}
   }

   /*=======================================================================
    * Begin take over
    *======================================================================*/
   if (strlen(cfginfo->hostname))
   {
#ifdef linux
	sprintf(buf,"/bin/hostname %s",cfginfo->hostname);
#else
	sprintf(buf,"/usr/bin/hostname %s",cfginfo->hostname);
#endif
	if (rc = system(buf))
	{   
	    syslog(LOG_ERR,"Error: autotakeover: hostname failed to set new "
			"hostname to %s.\n",
			cfginfo->hostname);
	}
   }

   /*=======================================================================
    * Create the reverse file
    *======================================================================*/
   if ((revfile = fopen(REVTAKEOVER, "w")) == NULL)
   {
	syslog(LOG_ERR, "Error: autotakeover: cannot create file \"%s\" (%s), "
			"autotakeover continues.  Manual procedures will "
			"have to be used later to reverse the takeover "
			"effects.", REVTAKEOVER, sys_errlist[errno]);
   }

   /*=======================================================================
    * Generate code to invoke a pre-reverse_takeover script if it exists
    *======================================================================*/
#ifdef linux
   if (revfile && fprintf(revfile, "#!/bin/sh\n"
#else
   if (revfile && fprintf(revfile, "#!/bin/ksh\n"
#endif
					"if [ -f %s ]; then\n"
					"  %s\n"
					"  let rc=$?\n"
					"  if [ $rc -ne 0 ]; then\n"
					"    echo \"Non-zero return code from "
						"%s, reverse takeover "
						"stopped\"\n"
					"    exit $rc\n"
					"  fi\n"
					"fi\n",
					PREREVTAKEOVER,
					PREREVTAKEOVER,
					PREREVTAKEOVER) < 0)
   {
	syslog(LOG_ERR, "Error: autotakeover: write to \"%s\" failed "
			"(%s), autotakeover continues.  Manual procedures will "
			"have to be used later to reverse the takeover "
			"effects.", REVTAKEOVER, sys_errlist[errno]);
	fclose(revfile);
	revfile = NULL;
	unlink(REVTAKEOVER);
   }

   /*=======================================================================
    * Now takeover the IP addresses
    *======================================================================*/
   for (i = 0; i < cfginfo->nif; i++)
   {
	char netmask[24];

	/*===============================================================
	 * Do not takeover local addresses
	 *==============================================================*/
	if (cfginfo->net[i].local || cfginfo->net[i].notakeover)
	    continue;

	if (strlen(cfginfo->net[i].netmask) != 0)
	{
	    strcpy(netmask,"netmask ");
	    strcat(netmask,cfginfo->net[i].netmask);
	}
	else
	    strcpy(netmask,"");

#ifdef _AIX
	sprintf(buf,"/usr/sbin/ifconfig %s inet %s %s alias up",
		cfginfo->net[i].netif,cfginfo->net[i].ip,netmask);
#endif
#ifdef __sun
	sprintf(buf,"/usr/sbin/ifconfig %s inet %s %s up",
		cfginfo->net[i].netif,cfginfo->net[i].ip,netmask);
#endif
#ifdef __hppa
	sprintf(buf,"/usr/contrib/bin/ifalias %s add %s",
		cfginfo->net[i].netif,cfginfo->net[i].ip,netmask);
#endif
#ifdef linux
	sprintf(buf,"/sbin/ifconfig %s inet %s %s up",
		cfginfo->net[i].netif,cfginfo->net[i].ip,netmask);
#endif
	syslog(LOG_NOTICE, "autotakeover: Issuing the following takeover "
			"command: \"%s\"", buf);
	if (rc = system(buf))
	{  
           syslog(LOG_ERR,"Error: autotakeover: ifconfig failed, IP address %s "
			"cannot be "
			"activated on interface %s.", cfginfo->net[i].ip,
			cfginfo->net[i].netif);
	}
	else
	{
	   int fd;
	   struct arpreq ar;
	   struct sockaddr_in *sin;
	   syslog(LOG_NOTICE, "autotakeover successful");
	
#if defined(_AIX) || defined(linux)
	   /*===========================================================
	    * Send a gratuitous ARP WHO_HAS request.  This is only needed
	    * on AIX and Linux since Solaris does it automatiacally.
	    *==========================================================*/
 	   arpwhohas(cfginfo->net[i].netif, inet_addr(cfginfo->net[i].ip));
#endif
	
	/*===================================================================
	 * Write an entry to the reverse_takeover script
	 *==================================================================*/
#ifdef _AIX
	   if (revfile && fprintf(revfile, "/usr/sbin/ifconfig %s %s "
			"-alias\n", cfginfo->net[i].netif, cfginfo->net[i].ip)
						< 0)
#endif
#ifdef __sun
	   if (revfile && fprintf(revfile, "/usr/sbin/ifconfig %s %s "
			"down\n", cfginfo->net[i].netif, cfginfo->net[i].ip)
						< 0)
#endif
#ifdef __hppa
	   if (revfile &&
			((cfginfo->net[i].alias &&
				fprintf(revfile, "/usr/contrib/bin/ifalias %s "
					"delete  %s\n", cfginfo->net[i].netif,
					cfginfo->net[i].ip) < 0)	||
			 (!cfginfo->net[i].alias &&
				fprintf(revfile, "/etc/ifconfig %s %s down\n",
					cfginfo->net[i].netif,
					cfginfo->net[i].ip) < 0)))
#endif
#ifdef linux
	   if (revfile && fprintf(revfile, "/sbin/ifconfig %s %s "
			"down\n", cfginfo->net[i].netif, cfginfo->net[i].ip)
						< 0)
#endif
	   {
		syslog(LOG_ERR, "Error: autotakeover: write to \"%s\" failed "
			"(%s), autotakeover continues.  Manual procedures will "
			"have to be used later to reverse the takeover "
			"effects.", REVTAKEOVER, sys_errlist[errno]);
		fclose(revfile);
		revfile = NULL;
		unlink(REVTAKEOVER);
	   }
	}
   }

   /*=======================================================================
    * Generate code to invoke a post-reverse_takeover script if it exists
    * Note that the reverse_takeover script also removes itself.
    *======================================================================*/
   if (revfile && fprintf(revfile, "#!/bin/ksh\n"
					"if [ -f %s ]; then\n"
					"  %s\n"
					"  let rc=$\n"
					"  if [ $rc -ne 0 ]; then\n"
					"    echo \"Non-zero return code from "
						"%s after reverse takeover "
						"was performed\"\n"
					"    exit $rc\n"
					"  fi\n"
					"fi\n"
					"rm -f %s\n"
					"exit 0\n",
					POSTREVTAKEOVER,
					POSTREVTAKEOVER,
					POSTREVTAKEOVER,
					REVTAKEOVER) < 0)
   {
	syslog(LOG_ERR, "Error: autotakeover: write to \"%s\" failed "
			"(%s), autotakeover continues.  Manual procedures will "
			"have to be used later to reverse the takeover "
			"effects.", REVTAKEOVER, sys_errlist[errno]);
	fclose(revfile);
	revfile = NULL;
	unlink(REVTAKEOVER);
   }

   if (revfile) {
	fclose(revfile);
	if (chmod(REVTAKEOVER, S_IRUSR|S_IWUSR|S_IXUSR) < 0) {
	    syslog(LOG_ERR, "Error: autotakeover: unable to change file "
			"access permissions for %s.  %s.", REVTAKEOVER,
			sys_errlist[errno]);
	}
   }

   /*=======================================================================
    * Invoke the post_takeover script if it exists
    *======================================================================*/
   if ((scriptfile = fopen(POSTTAKEOVER, "r")) != NULL)
   {
	int rc;
	fclose(scriptfile);
	rc = system(POSTTAKEOVER);
	if (WEXITSTATUS(rc) != 0)
	{
	    syslog(LOG_ERR, "Error: Non-zero return code (0x%08x) from %s "
			"after takeover was performed.",
			rc, POSTTAKEOVER);
	    exit(1);
	}
   }

   exit(0);
} 
