/*============================================================================
 * file:	fwpulse/backupfw.c  
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=======================================================================
 *  
 * function:	Function to parse the fwpulse.conf file and return the values
 *		in the structure passed as a parameter.
 *		The function reads the entire fwpulse.conf file before
 *		returning.
 *
 * Returns 0 if successful
 * Returns 1 if unsuccessful
 * 
 * (C) COPYRIGHT Freemont Avenue Software, Inc. 1997-2000
 * All rights reserved.    
 *								      
 * Date written:	February  12, 1996
 * Last modified:	February   8, 2000 - added LPL
 *								      
 *=========================================================================*/
#include	<ctype.h>  
#include	<stdio.h>  
#include	<limits.h>
#include	<sys/types.h>
#ifndef MSDOS
#include	<syslog.h>
#include        <netinet/in.h>
#else
#include        <winsock.h>
#endif
#include	<errno.h>
#include	<stdlib.h>
#include	<string.h>
#include	"des.h"
#include	"fwpulse.h"

#ifndef linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif

/*==========================================================================
 * Depending on the program used, use either printf or syslog to report an
 * error.
 *=========================================================================*/
void report4(char *msg,char *progname, char *parm, int line,char *conffn)
{
   /*=======================================================================
    * If progname is fwpulse or fwpulsed , use syslog to report an error,
    * otherwise, use printf
    *======================================================================*/
   if (strstr(progname,"fwpulse"))
	syslog(LOG_ERR,msg,progname,parm,line,conffn);
   printf(msg,progname,parm,line,conffn);
   printf("\n");
   return;
}

/*==========================================================================
 * Depending on the program used, use either printf or syslog to report an
 * error.
 *=========================================================================*/
void report3(char *msg,char *progname, int line,char *conffn)
{
   /*=======================================================================
    * If progname is fwpulse or fwpulsed , use syslog to report an error,
    * otherwise, use printf
    *======================================================================*/
   if (strstr(progname,"fwpulse"))
	syslog(LOG_ERR,msg,progname,line,conffn);
   printf(msg,progname,line,conffn);
   printf("\n");
   return;
}
/*==========================================================================
 * Depending on the program used, use either printf or syslog to report an
 * error.
 *=========================================================================*/
#ifdef linux
void report2(char *msg,char *progname, char const *conffn)
#else
void report2(char *msg,char *progname, char *conffn)
#endif
{
   /*=======================================================================
    * If progname is fwpulse or fwpulsed , use syslog to report an error,
    * otherwise, use printf
    *======================================================================*/
   if (strstr(progname,"fwpulse"))
	syslog(LOG_ERR,msg,progname,conffn);
   printf(msg,progname,conffn);
   printf("\n");
   return;
}
/*==========================================================================
 * Function to read in the fwpulse.conf file.
 * Arguments:
 *	progname	Name of the program - this is to be used in error
 *			and informational messages.
 *	cfginfo		The ptr to the config info structure is returned
 *			in this indirect ptr.
 *	localip		Points to the array of local IP addresses that was
 *			returned from an earlier call to getlocal().
 *	nlips		Number of local IP addresses in the localip[ array.
 * Return codes:
 *	0	Successful
 *	1	Unsuccessful
 *=========================================================================*/
int fwpulseconf(char *progname, struct cfginfo **cfginfo,
			in_addr_t *localip, long nlips)
{  
   static char cprght[]
	  = "(C) COPYRIGHT Freemont Avenue Software, Inc. 1996-1999";
   int		fd;  
   int		line=0;
   FILE		*cfp;
   char		*token;
   char	 tokensep[] = "\t,;= \n";
   char		conffn[] = FW_CFG_DIR "/fwpulse.conf";
   char		linebuf[BUFSIZ];
   int		rc=0;
   int		nnet=0,email=0;
   int		i;
   char		*ep;
   size_t	cfg_size;
   unsigned long octbyte;

   /*=======================================================================
    * Open the config file
    *======================================================================*/
   if ( (cfp = fopen(conffn,"r") ) == NULL) 
   {
      report2("%s: Unable to open config file.  %s.\n",
		progname, sys_errlist[errno]);
      exit(1);
   }

   /*=======================================================================
    * Read the config file once frist to determine how much space to malloc.
    *======================================================================*/
   while (fgets(linebuf,sizeof(linebuf),cfp) != NULL) {
	if( (token = strtok(linebuf,tokensep)) == NULL)
	   continue;
	if (!strcmp(token,"network") || !strcmp(token,"permit"))
	   line++;
   }

   /*=======================================================================
    * Re-position the file pointer so we can read from the top again
    *======================================================================*/
   if (fseek(cfp, 0, SEEK_SET) < 0) {
	report2("Error: %s: Unable to reposition file pointer for the config "
			"file.  %s.", progname, sys_errlist[errno]);
	exit(1);
   }

   /*=======================================================================
    * malloc the space for the cfginfo structure
    *======================================================================*/
   cfg_size = sizeof(struct cfginfo) + (sizeof(struct hostid) * (line - 1));
   if ((*cfginfo = malloc(cfg_size)) == NULL) {
	report2("Error: %s: Unable to malloc space for internal structures.  "
			"%s.", progname, sys_errlist[errno]);
	exit(1);
   }

   /*=======================================================================
    * Initialize the cfginfo structure
    *======================================================================*/
   memset(*cfginfo,0,cfg_size);
   (*cfginfo)->nif = line;
   (*cfginfo)->wakeup=default_wakeup;
   (*cfginfo)->autotakeover = 0;
   (*cfginfo)->verbosemail = 0;
   (*cfginfo)->threshold = 3;
   (*cfginfo)->startup_delay = 5 * 60;

   line = 0;

   /*==========================================================================
    * /etc/firewall/fwpulse.conf file format
    * hostname=<hostname>
    * network ip_address=<ip_address> netmask=<netmask>
    *		{router=<ip_address> | alias}
    *		net_if=<net_if> [critical_port=<port_number>]
    * email=<email_address>
    * wakeup=<secs>
    * autotakeover
    * threshold=<secs>
    * private_port <port_number> <encryption_key_specified_as_8_octal_numbers>
    ==========================================================================*/

   while (fgets(linebuf,sizeof(linebuf),cfp) != NULL)
   {
	line++;
	/*==================================================================
	 * Skip over comments and null lines
	 *=================================================================*/
	if( (*linebuf == '#') || (*linebuf == '\n') )
	   continue;
	if( (token = strtok(linebuf,tokensep)) == NULL)
	   continue;
	/*==================================================================
	 * Determine what statement it is and process it
	 *=================================================================*/
	if (!strcmp(token,"hostname"))
	{
	   /*===============================================================
	    * "hostname" statement
	    *==============================================================*/
	   if ((token = strtok(NULL,tokensep)) == NULL)
	   {
		report3("Error: %s: Empty \"hostname\" parameter on line %d "
			"of %s.", progname,line,conffn);
		return(1);
	   }
	   else
	   {
		if (strlen(token) > sizeof((*cfginfo)->hostname) - 1)
		{
		   report3("Error: %s: hostname parameter longer than 80",
			progname,line,conffn);
		   return(1);
		}
		else
		   strcpy((*cfginfo)->hostname,token);
	   }

	}
	else if (!strcmp(token,"permit"))
	{
	   /*===============================================================
	    * "permit" statement
	    * permit a.b.c.d
	    * permit from a.b.c.d
	    *==============================================================*/
	   in_addr_t tmpip;
	   memset(&((*cfginfo)->net[nnet]), 0, sizeof(hostid));

	   if ((token=strtok(NULL,tokensep)) == NULL ||
	   		(strcmp(token, "from") == 0 &&
				(token=strtok(NULL,tokensep)) == NULL))
	   {
		report3("Error: %s: Missing parameter in \"permit\" statement "
				"on line %d of %s.",progname,line,conffn);
		return(1);
	   }
	   if (strlen(token) > sizeof((*cfginfo)->net[0].ip) - 1)
	   {
		report3("Error: %s: \"permit\" parameter longer "
				"than 15", progname,line,conffn);
		return(1);
	   }
	   tmpip = inet_addr(token);
	   if (tmpip == INADDR_ANY || tmpip == 0xffffffff)
	   {
		report3("Error: %s: Invalid \"permit\" "
				"parameter on line %d of %s.",
				progname,line,conffn);
		return(1);
	   }
	   strcpy((*cfginfo)->net[nnet].ip,token);
	   /*===============================================================
	    * Determine this address is local or not
	    *==============================================================*/
	   (*cfginfo)->net[nnet].local = 0;
	   for (i = 0; i < nlips; i++)
	   {
		if (tmpip == localip[i]) {
		    (*cfginfo)->net[nnet].local = 1;
		    break;
		}
	   }
	   (*cfginfo)->net[nnet].notakeover = 1;
	   (*cfginfo)->net[nnet].alias = 1;

	   nnet++;
	}
	else if (!strcmp(token,"network"))
	{
	   in_addr_t tmpip;

	   memset(&((*cfginfo)->net[nnet]), 0, sizeof(hostid));
	   /*===============================================================
	    * Process the rest of the line
	    *==============================================================*/
	   while ((token=strtok(NULL,tokensep)) != NULL)
	   {
		if (!strcmp(token,"ip_address"))
		{
		   if ((token=strtok(NULL,tokensep)) == NULL)
	   	   {
			report3("Error: %s: Empty \"ip_address\" parameter on "
				"line %d of %s.",progname,line,conffn);
			return(1);
		   }
		   else
		   {
			if (strlen(token) > sizeof((*cfginfo)->net[0].ip) - 1)
			{
			   report3("Error: %s: \"ip_address\" parameter longer "
				"than 15", progname,line,conffn);
			   return(1);
			}
			tmpip = inet_addr(token);
			if (tmpip == INADDR_ANY || tmpip == 0xffffffff)
			{
			   report3("Error: %s: Invalid \"ip_address\" "
				"parameter on line %d of %s.",
				progname,line,conffn);
			   return(1);
			}
			strcpy((*cfginfo)->net[nnet].ip,token);
		   }
		}
		else if (!strcmp(token,"netmask"))
		{
#ifdef __hppa
		   report3("Error: %s: \"netmask\" is not supported on this "
				"operating system.",progname,line,conffn);
		   return(1);
#else
		   if ((token=strtok(NULL,tokensep)) == NULL)
	   	   {
			report3("Error: %s: empty \"netmask\" parameter on "
				"line %d of %s.",progname,line,conffn);
			return(1);
		   }
		   else
		   {
			if (strlen(token) >
				sizeof((*cfginfo)->net[0].netmask) - 1)
			{
			   report3("Error: %s: \"netmask\" parameter longer "
				"than 15", progname,line,conffn);
			   return(1);
			}
			tmpip = inet_addr(token);
			if (tmpip == INADDR_ANY || tmpip == 0xffffffff)
			{
			   report3("Error: %s: Invalid \"netmask\" "
				"parameter on line %d of %s.",
				progname,line,conffn);
			   return(1);
			}
			strcpy((*cfginfo)->net[nnet].netmask,token);
		   }
#endif
		}
		else if (!strcmp(token,"net_if"))
		{
		   if ((token=strtok(NULL,tokensep)) == NULL)
	   	   {
			report3("Error: %s: empty \"net_if\" parameter on line "
				"%d of %s.",progname,line,conffn);
			return(1);
	   	   }
		   else
		   {
			if (strlen(token) > sizeof((*cfginfo)->net[0].netif)-1)
			{
			   report3("Error: %s: net_if parameter longer than 15",
				progname,line,conffn);
			   return(1);
			}
			else
			   strcpy((*cfginfo)->net[nnet].netif,token);
		   }
		}
		else if (!strcmp(token,"router"))
		{
		   if ((token=strtok(NULL,tokensep)) == NULL)
	   	   {
			report3("%s: Empty \"router\" parameter on line "
				"%d of %s.",progname,line,conffn);
			return(1);
		   }
		   else
		   {
			if (strlen(token) > sizeof((*cfginfo)->net[0].router)-1)
			{
			   report3("Error: %s: \"router\" parameter longer "
				"than 15", progname,line,conffn);
			   return(1);
			}
			tmpip = inet_addr(token);
			if (tmpip == INADDR_ANY || tmpip == 0xffffffff)
			{
			   report3("Error: %s: Invalid \"router\" parameter on "
				"line %d of %s.",progname,line,conffn);
			   return(1);
			}
			strcpy((*cfginfo)->net[nnet].router,token);
		   }
		}
		else if (!strcmp(token,"critical_port"))
		{
		   if ((token=strtok(NULL,tokensep)) == NULL)
	   	   {
			report3("Error: %s: empty \"critical_port\" parameter "
				"on line %d of %s.",progname,line,conffn);
			return(1);
	   	   }
		   if ((*cfginfo)->net[nnet].alias != 0) {
			report3("Error: %s: The alias parameter and the "
					"critical_port parameter cannot "
					"be specified on the same \"network\" "
					"statement - line %d of %s.",
					progname,line,conffn);
			return(1);
		   }
		   (*cfginfo)->net[nnet].toport = atoi(token);
		   if ((*cfginfo)->net[nnet].toport <= 0 ||
			(*cfginfo)->net[nnet].toport > 65535) {
			report3("Error: %s: \"critical_port\" parameter on "
				"line %d of %s must "
				"be be greater than 0 and less than 65535.",
				progname,line,conffn);
			return(1);
		   }
		}
		else if (!strcmp(token,"alias"))
		{
		   if ((*cfginfo)->net[nnet].toport != 0) {
			report3("Error: %s: The alias parameter and the "
					"critical_port parameter cannot "
					"be specified on the same \"network\" "
					"statement - line %d of %s.",
					progname,line,conffn);
			return(1);
		   }
		   (*cfginfo)->net[nnet].alias = 1;
		}
		else
		{
		   report4("Error: %s: Invalid parameter \"%s\" in \"network\" "
				"statement on line %d of %s.",
				progname,token,line,conffn);
		   return(1);
		}
	   }
	   /*===============================================================
	    * if any of the required network parameters is not set, issue a
	    * failure message
	    *==============================================================*/
	   if (!strcmp((*cfginfo)->net[nnet].ip,"")	||
		(!strcmp((*cfginfo)->net[nnet].router,"") &&
		 !(*cfginfo)->net[nnet].alias) ||
		!strcmp((*cfginfo)->net[nnet].netif,""))
	   {
		report3("Error: %s: Incomplete \"network\" statement on line "
			"%d of %s.",progname,line,conffn);
		return(1);
	   }
	   /*===============================================================
	    * Determine this address is local or not
	    *==============================================================*/
	   tmpip = inet_addr((*cfginfo)->net[nnet].ip);
	   (*cfginfo)->net[nnet].local = 0;
	   for (i = 0; i < nlips; i++)
	   {
		if (tmpip == localip[i]) {
		    (*cfginfo)->net[nnet].local = 1;
		    break;
		}
	   }

	   nnet++;
	}
	else if (!strcmp(token,"email"))
	{
	   /*===============================================================
	    * "email" statement
	    *==============================================================*/
	   email=1;
	   if ((token=strtok(NULL,tokensep)) == NULL)
	   {
		report3("Error: %s: Empty \"email\" parameter on line %d of %s.",
			progname,line,conffn);
		return(1);
	   }
	   else
	   {
		if (strlen(token) > sizeof((*cfginfo)->email) - 1)
		{
		   report3("Error: %s: email parameter longer than 80",
			progname,line,conffn);
		   return(1);
		}
		else
		   strcpy((*cfginfo)->email,token);
	   }
	}
	else if (!strcmp(token,"wakeup"))
	{
	   /*===============================================================
	    * "wakeup" statement
	    *==============================================================*/
	   if ((token=strtok(NULL,tokensep)) == NULL)
	   {
		report3("Error: %s: Empty \"wakeup\" parameter on line %d of "
					"%s.",
			progname,line,conffn);
		return(1);
	   }
	   (*cfginfo)->wakeup=atoi(token);
	}
	else if (!strcmp(token,"threshold"))
	{
	   /*===============================================================
	    * "threshold" statement
	    * Number of seconds to wait for a response before concluding that
	    * an interface is down.
	    *==============================================================*/
	   if ((token=strtok(NULL,tokensep)) == NULL)
	   {
		report3("Error: %s: Empty \"threshold\" parameter on line %d "
					"of %s.",
			progname,line,conffn);
		return(1);
	   }
	   (*cfginfo)->threshold=atoi(token);
	   if ((*cfginfo)->threshold <= 0) {
		report3("Error: %s: Invalid \"threshold\" parameter on line %d "
					"of %s.",
			progname,line,conffn);
		return(1);
	   }
	}
	else if (!strcmp(token,"startup_delay"))
	{
	   /*===============================================================
	    * "startup_delay" statement
	    * Number of seconds during startup during which fwpulse can
	    * ignore non-responses.
	    *==============================================================*/
	   if ((token=strtok(NULL,tokensep)) == NULL)
	   {
		report3("Error: %s: Empty \"startup_delay\" parameter on line "
			"%d of %s.",
			progname,line,conffn);
		return(1);
	   }
	   (*cfginfo)->startup_delay=atoi(token);
	   if ((*cfginfo)->startup_delay <= 0) {
		report3("Error: %s: Invalid \"startup_delay\" parameter on "
			"line %d of %s.",
			progname,line,conffn);
		return(1);
	   }
	}
	else if (!strcmp(token,"verbosemail"))
	{
	   /*===============================================================
	    * "verbosemail" statement
	    *==============================================================*/
	   (*cfginfo)->verbosemail = 1;
	}
	else if (!strcmp(token,"autotakeover"))
	{
	   /*===============================================================
	    * "autotakeover" statement
	    *==============================================================*/
	   (*cfginfo)->autotakeover = 1;
	}
	else if (!strcmp(token,"private_port"))
	{
	   /*===============================================================
	    * "private_port" statement
	    * private_port <port_number> nnn nnn nnn nnn nnn nnn nnn nnn
	    *==============================================================*/
	   int foundlocal = 0;
	   in_addr_t tmpip;
	   /*===============================================================
	    * Get the private port number
	    *==============================================================*/
	   if ((token=strtok(NULL,tokensep)) == NULL)
	   {
		report3("Error: %s: No private port specified with "
			"\"private_port\" statement on line %d of %s.",
			progname,line,conffn);
		return(1);
	   }
	   if (((*cfginfo)->private_port = atoi(token)) <= 0 ||
			(*cfginfo)->private_port > 65535)
	   {
		report4("Error: %s: Invalid private port \"%s\" on line "
			"%d of %s.  Must be less than 65536.",
			progname,token,line,conffn);
		return(1);
	   }
#ifdef DES
	   /*===============================================================
	    * Get the private key
	    *==============================================================*/
	   for (i = 0; i < sizeof(des_cblock); i++) {
		if ((token=strtok(NULL,tokensep)) == NULL) {
		    report3("Error: %s: Imcomplete private key specified "
				"on line %d of %s.",
				progname, line, conffn);
		    return(1);
		}
		if ((octbyte = (int) strtoul(token, &ep, 8)) == 0 ||
			octbyte == ULONG_MAX ||
			octbyte > 0xff) {
		    report4("Error: %s: Invalid private key value (%s) on "
				"line %d of %s.",
				progname,token,line,conffn);
		    return(1);
		}
		(*cfginfo)->private_key[i] = octbyte;
	   }
#endif
	}
	else
	{
	   report3("Error: %s: Invalid statement on line %d of %s.",
			progname,line,conffn);
	   return(1);
	}
   }

   if ((*cfginfo)->private_port == 0)
   {
	report2("Error: %s: \"private_port\" statement not specified in %s.",
		progname,conffn);
	rc = 1;
   }

   /*========================================================================
    * If we got here and the file is not at EOF, we got a read error
    *=======================================================================*/
   if (!feof(cfp))
   {
	report3("%s: Read error (errno=%d) encountered in %s.",
			progname,errno,conffn);
	rc=1;
   }
	
   /*=======================================================================
    * If the required info is not in the fwpulse.conf file, issue a message
    *======================================================================*/
   if (nnet < 2)
   {
	report2("Error: %s: less than 2 network statement found in %s. Must "
		"have at 2 to configure a functional firewall.",
		progname,conffn);
	rc=1;
   }
   if (!email)
   {
	report2("Error: %s: missing an email statement in %s.",progname,conffn);
	rc=1;
   }
   fclose(cfp);
   return(rc); 
}
