/*============================================================================
 * file:	fwpulse/fwpulse.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=======================================================================
 *
 * function:	Pings the firewall to determine if the
 *		the firewall is down and the sysadmin needs to be warned
 *		about bringing up a backup firewall.
 *
 * (C) COPYRIGHT Freemont Avenue Software, Inc. 1997-2000
 * All rights reserved.
 *
 * Date written:	February   8, 1996
 * Last modified:	February   8, 2000 - added LPL
 *
 *=========================================================================*/
#include	<ctype.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<limits.h>
#include	<errno.h>
#include	<sys/types.h>
#include	<syslog.h>
#include	<signal.h>
#include	<sys/wait.h>
#include	<sys/stat.h>
#include	<sys/socket.h>
#ifndef __hppa
#include	<sys/select.h>
#endif
#include	<netinet/in.h>
#include	<net/if.h>
#include	"des.h"
#include	"fwpulse.h"

int fwpulseconf(char *, struct cfginfo **cfginfo, in_addr_t *localip,
				int nlips);
#ifdef DES
int autotakeover(int inactive, int active, struct cfginfo *cfginfo,
                        des_key_schedule ksched);
#else
int autotakeover(int inactive, int active, struct cfginfo *cfginfo);
#endif
#ifndef linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif
static struct timeval	startup;
static int	absolutely_no_takeover = 0;	/* Overrides all decisions for
						   takeover.  E.g. when cannot
						   ping router, there's no
						   point in a takeover	*/

/*=======================================================================
 * Signal handler for alarms.  All we need to do is return.
 *======================================================================*/
void sigalrm(int sig) {
    psignal(SIGALRM, sigalrm);
    return;
}

/*=======================================================================
 * Signal handler for TERM.  Log the event and return
 *======================================================================*/
void sigterm(int sig) {
    syslog(LOG_NOTICE, "TERM signal received.  fwpulse exiting.");
    exit(1);
}

/*=======================================================================
 * Check fd one more time to see if the last connect finally responded.
 * This is called if an initial connect to the remote private port failed
 * with EINTR (alarm went off) and after pingloop (almost) decided that it
 * needs to do a takeover.  It is used as a re-double check to see if a
 * takeover can be avoided.

 * Returns:
 *	0	if fd is still not connected
 *	1	if fd is connected
 *======================================================================*/
int fd_connected(int fd) {
    int rc;
    int err;
    size_t len;
    fd_set rdset;
    struct timeval tv;

    tv.tv_sec = 0;
    tv.tv_usec = 0;
    FD_ZERO(&rdset);
    FD_SET(fd, &rdset);

    /*=======================================================================
     * If select returned error, fd is NOT connected.
     *======================================================================*/
    if ((rc = select(fd+1, &rdset, (fd_set *)0, (fd_set *)0, &tv)) < 0) {
	syslog(LOG_ERR, "Error: select failed.  %s.", sys_errlist[errno]);
	return(0);
    }

    /*=======================================================================
     * If select returned with nothing ready, fd is NOT connected.
     *======================================================================*/
    if (rc == 0)
	return(0);

    /*=======================================================================
     * If socket has a pending error, fd is NOT connected.
     *======================================================================*/
    len = sizeof(err);
    rc = getsockopt(fd, SOL_SOCKET, SO_ERROR, (char *)&err, (socklen_t *)&len);
    if (rc != 0 || err != 0)
	return(0);

    return(1);
}

/*=======================================================================
 * Input:
 *	cfginfo		structure containing config info from fwpulseconf()
 *	active		ptr to integer to return number of active interfaces
 *	inactive	ptr to integer to return number of inactive interfaces
 *	ksched		des key schedule for encryption
 *======================================================================*/
#ifdef DES
void pingloop(struct cfginfo *cfginfo, int *active, int *inactive,
		des_key_schedule ksched)
#else
void pingloop(struct cfginfo *cfginfo, int *active, int *inactive)
#endif
{
    int i;
    int rc;
    int error_no;
    struct timeval tv;
    struct sockaddr_in sock;

    (*inactive) = 0;
    (*active) = 0;
    absolutely_no_takeover = 0;
	
    /*===================================================================
     * try to connect to the network ip addresses that are not local
     *==================================================================*/
    psignal(SIGALRM, sigalrm);
    for (i = 0; i < cfginfo->nif; i++)
    {
	int fd, critfd;
	int	trylater;
	time_t curtime;
	char msg[BUFSIZ];
	char cmd[BUFSIZ];
#ifdef DES
	int j;
	char * cp;
	char ebuf[EBUFSIZ];
#endif
	commpkt pkt;
	commpkt resp;

	gettimeofday(&tv, NULL);
	trylater = 0;
	cfginfo->net[i].down_suspect = 0;

	/*===============================================================
	 * Only query remote addresses that are not aliases
	 *==============================================================*/
	if (cfginfo->net[i].local || cfginfo->net[i].alias ||
			cfginfo->net[i].notakeover)
	    continue;

	if ((fd = socket(AF_INET,SOCK_STREAM,0)) < 0) {
	    syslog(LOG_ERR, "Error: Unable to create socket.  %s.",
			sys_errlist[errno]);
	    exit(1);
	}
	memset(&sock, 0, sizeof(sock));
	sock.sin_family = AF_INET;
	sock.sin_port = htons(cfginfo->private_port);
	sock.sin_addr.s_addr = inet_addr(cfginfo->net[i].ip);

	/*===============================================================
	 * Set alarm so we don't wait longer than the threshold
	 *==============================================================*/
	alarm(cfginfo->threshold);
	if (connect(fd, (struct sockaddr *)&sock, sizeof(sock)) < 0) {
	    register struct timeval tmplastlogged;

	    if (errno == EINTR)
		trylater = 1;
	    error_no = errno;
	    tmplastlogged.tv_sec = (time_t)-1;
	    tmplastlogged.tv_usec = 0;
	    alarm(0);

	    /*===========================================================
	     * Only log the event if we had not logged an event on this
	     * IP address in the last LONGWAIT seconds.
	     *==========================================================*/
	    if (tv.tv_sec - cfginfo->net[i].lastlogged >= LONGWAIT) {
		sprintf(msg,"Attention: fwpulsed at firewall IP address %s "
			"not responding.  %s.",
			cfginfo->net[i].ip, sys_errlist[error_no]);
		syslog(LOG_ERR, msg);
		if (cfginfo->verbosemail) {
		    sprintf(cmd, "/usr/bin/mailx -s \"%s\" %s </dev/null",
			msg, cfginfo->email);
		    if (rc = system(cmd))
			syslog(LOG_ERR,"Error: Attempt to mail attention "
				"notification to %s failed.  rc = 0x%08x.",
				cfginfo->email, rc);
		}
		tmplastlogged.tv_sec = cfginfo->net[i].lastlogged;
		cfginfo->net[i].lastlogged = tv.tv_sec;
	    }
	    /*===========================================================
	     * Before we conclude that the remote IP address is down,
	     * let's ping the router and if we can't ping the router,
	     * may be the local interface, not the remote, is down.  If we
	     * can ping the router, then conclude there is something wrong
	     * with the remote.
	     *==========================================================*/
#ifdef _AIX
	    sprintf(cmd,"/usr/sbin/ping -c 1 %s >/dev/null 2>&1",
				cfginfo->net[i].router);
#endif
#ifdef __hppa
	    sprintf(cmd,"/usr/sbin/ping %s -n 1 | grep ' 0%% packet loss' "
			">/dev/null 2>&1",
				cfginfo->net[i].router);
#endif
#ifdef __sun
	    sprintf(cmd,"/usr/sbin/ping %s >/dev/null 2>&1",
				cfginfo->net[i].router);
#endif
#ifdef linux
	    sprintf(cmd,"/bin/ping -c 1 %s >/dev/null 2>&1",
				cfginfo->net[i].router);
#endif

	    alarm(cfginfo->threshold);
	    if ((rc = system(cmd))) {
		error_no = errno;
		alarm(0);
		/*===================================================
		 * Ping did not run successfully.  We don't want to
		 * increment the inactive count in this case or attempt
		 * to autotakeover; otherwise, we may end up taking over
		 * from the remote machine while it still has the same
		 * IP addresses active.  There's no point to takeover
		 * anyway if the local machine cannot contact the router.
		 *
		 * If tmplastlogged not -1, then it reported a non-connect
		 * error above, we want to report here also.
		 *==================================================*/
		if (tmplastlogged.tv_sec != (time_t)-1) {
		    char tmp[BUFSIZ];
		    gethostname(tmp, sizeof(tmp) - 80);
		    sprintf(msg,"Attention: Unable to ping router %s using "
			"command \"%s\" rc = 0x%08x, errno = %d (%s) - "
			"perhaps problem is with the local machine (%s).  "
			"Remote firewall %s is assumed to be fine and no "
			"automatic failover will be performed.",
			cfginfo->net[i].router,
			cmd, rc, error_no, sys_errlist[error_no],
			tmp, cfginfo->net[i].ip);
		    syslog(LOG_ERR, msg);
		    if (cfginfo->verbosemail) {
		        sprintf(cmd, "/usr/bin/mailx -s \"%s\" %s </dev/null",
				msg, cfginfo->email);
			if (rc = system(cmd))
				syslog(LOG_ERR,"Error: Attempt to mail "
				"attention notification to %s failed.  "
				"rc = 0x%08x.",
				cfginfo->email, rc);
		    }
		}
		close(fd);
		absolutely_no_takeover = 1;
		continue;
	    }
	    alarm(0);
	    /*===========================================================
	     * We can ping the router so we can conclude that the problem
	     * is likely on the remote firewall.
	     * Before we REALLY conclude that this IP address is down, see
	     * if the config file has specified a critical_port for us to
	     * double check.
	     *==========================================================*/
	    if (cfginfo->net[i].toport) {
		/*=======================================================
		 * Try to connect to the critical port.
		 *======================================================*/
		if ((critfd = socket(AF_INET,SOCK_STREAM,0)) < 0) {
		    syslog(LOG_ERR, "Error: Unable to create socket.  %s.",
				sys_errlist[errno]);
		    exit(1);
		}

		sock.sin_port = htons(cfginfo->net[i].toport);

		alarm(cfginfo->threshold);
		if (connect(critfd, (struct sockaddr *)&sock, sizeof(sock))
								< 0) {
		    error_no = errno;
		    alarm(0);
		    close(critfd);
		    /*===================================================
		     * If tmplastlogged not -1, then it reported a non-connect
		     * error above, we want to report here also.
		     *==================================================*/
		    if (tmplastlogged.tv_sec != (time_t)-1) {
			sprintf(msg,"Attention: Critical port %d at firewall "
				"IP address %s not responding.  %s.",
				cfginfo->net[i].toport, cfginfo->net[i].ip,
				sys_errlist[error_no]);
			syslog(LOG_ERR, msg);
			if (cfginfo->verbosemail) {
			    sprintf(cmd, "/usr/bin/mailx -s \"%s\" %s "
				"</dev/null",
				msg, cfginfo->email);
			    if (rc = system(cmd))
				syslog(LOG_ERR,"Error: Attempt to mail "
					"attention notification to %s failed.  "
					"rc = 0x%08x.",
					cfginfo->email, rc);
			}
		    }
		    /*===================================================
		     * Check one more time to see if the initial connect()
		     * call finally got thru.
		     *==================================================*/
		    if (trylater && fd_connected(fd)) {
			if (tmplastlogged.tv_sec != (time_t)-1) {
			    sprintf(msg, "Received delayed response from "
				"fwpulsed at firewall %s.  fwpulse is assuming "
				"previous non-response messages were caused "
				"by a busy network.  No further fwpulse "
				"action will be taken for now.",
				cfginfo->net[i].ip);
			    syslog(LOG_ERR, msg);
			    if (cfginfo->verbosemail) {
				sprintf(cmd, "/usr/bin/mailx -s \"%s\" %s "
					"</dev/null", msg, cfginfo->email);
				if (rc = system(cmd))
				    syslog(LOG_ERR,"Error: Attempt to mail "
					"attention notification to %s failed.  "
					"rc = 0x%08x.",
					cfginfo->email, rc);
			    }
			}
		    }
		    else {
			if (trylater && tmplastlogged.tv_sec != (time_t)-1)
			    syslog(LOG_NOTICE, "Rechecked connection request "
					"to firewall at %s but still not "
					"able to connect.", cfginfo->net[i].ip);
			close(fd);
			(*inactive)++;
			cfginfo->net[i].down_suspect = 1;
			continue;
		    }
		}
		/*===========================================================
		 * The critical port responded.  This should lead us to
		 * conclude that we don't want to auto-takeover now.
		 * We only want to log the following message if we just
		 * reported an fwpulsed non-response above.
		 *==========================================================*/
		else if (tmplastlogged.tv_sec != (time_t)-1) {
		    alarm(0);
		    close(critfd);
		    sprintf(msg, "Critical port %d at firewall %s is "
				"responding - no takeover activity is deemed "
				"necessary at this time.  However, the "
				"previously logged fwpulsed non-response "
				"message may still point to a situation at the "
				"firewall that needs to be investigated.",
				cfginfo->net[i].toport, cfginfo->net[i].ip);
		    syslog(LOG_ERR, msg);
		    if (cfginfo->verbosemail) {
			sprintf(cmd, "/usr/bin/mailx -s \"%s\" %s </dev/null",
				msg, cfginfo->email);
			if (rc = system(cmd))
				syslog(LOG_ERR,"Error: Attempt to mail "
					"attention notification to %s failed.  "
					"rc = 0x%08x.",
					cfginfo->email, rc);
		    }
		}
		close(fd);
		alarm(0);
		(*active)++;
		continue;
	    }
	    else {
		/*===================================================
		 * Check one more time to see if the initial connect()
		 * call finally got thru.
		 *==================================================*/
		if (trylater && fd_connected(fd)) {
		    if (tmplastlogged.tv_sec != (time_t)-1) {
			sprintf(msg, "Received delayed response from fwpulsed "
				"at firewall %s.  fwpulse is assuming "
				"previous non-response messages were caused "
				"by a busy network.  No further fwpulse "
				"action will be taken for now.",
				cfginfo->net[i].ip);
			syslog(LOG_ERR, msg);
			if (cfginfo->verbosemail) {
			    sprintf(cmd, "/usr/bin/mailx -s \"%s\" %s "
					"</dev/null", msg, cfginfo->email);
			    if (rc = system(cmd))
				syslog(LOG_ERR,"Error: Attempt to mail "
					"attention notification to %s failed.  "
					"rc = 0x%08x.",
					cfginfo->email, rc);
			}
		    }
		}
		else {
		    if (trylater && tmplastlogged.tv_sec != (time_t)-1)
			syslog(LOG_NOTICE, "Rechecked connection request "
					"to firewall at %s but still not "
					"able to connect.", cfginfo->net[i].ip);
		    close(fd);
		    (*inactive)++;
		    cfginfo->net[i].down_suspect = 1;
		    continue;
		}
	    }
	}

	/*===============================================================
	 * Disable alarm
	 *==============================================================*/
	alarm(0);

	/*===============================================================
	 * As long as we can connect, we consider that interface up.
	 * If the NOP request failed for any reason, we will report it
	 * in syslog but still count it as up.
	 *==============================================================*/
	(*active)++;
	cfginfo->net[i].lastlogged = (time_t)0;

	/*===============================================================
	 * If not sending NOP requests, we are done for this iteration.
	 *==============================================================*/
	if (cfginfo->net[i].sent_nop) {
	    close(fd);
	    continue;
	}

	/*===============================================================
	 * Send NOP packet
	 *==============================================================*/
	gettimeofday(&tv, NULL);
	pkt.current_time.tv_sec = htonl(tv.tv_sec);
	pkt.current_time.tv_usec = htonl(tv.tv_usec);
	pkt.down_ip = lrand48();
	pkt.dup_ip = pkt.down_ip;
	pkt.dup_time.tv_sec = pkt.current_time.tv_sec;
	pkt.dup_time.tv_usec = pkt.current_time.tv_usec;
	pkt.code = htonl(NOP);
	pkt.dup_code = pkt.code;

#ifdef DES
	/*===============================================================
	 * Encrypt
	 *==============================================================*/
	cp = (char *)&pkt;
	for (j = 0; j < EBUFSIZ; j += 8)
	    des_ecb_encrypt((des_cblock *)(cp + j), (des_cblock *)(ebuf + j),
			ksched, DES_ENCRYPT);
	if (write(fd, ebuf, sizeof(ebuf)) < 0) {
#else
	if (write(fd, &pkt, sizeof(pkt)) < 0) {
#endif
	    error_no = errno;
	    if (tv.tv_sec - cfginfo->net[i].lastlogged >= LONGWAIT) {
		sprintf(msg, "Error: Unable to write NOP request to remote "
			"firewall %s.  %s.  Attention is required.",
			cfginfo->net[i].ip, sys_errlist[error_no]);
		syslog(LOG_ERR, msg);
		if (cfginfo->verbosemail) {
		    sprintf(cmd, "/usr/bin/mailx -s \"%s\" %s </dev/null",
			msg, cfginfo->email);
		    if (rc = system(cmd))
			syslog(LOG_ERR,"Error: Attempt to mail attention "
				"notification to %s failed.  rc = 0x%08x",
				cfginfo->email, rc);
		}
		cfginfo->net[i].lastlogged = tv.tv_sec;
	    }
	    continue;
	}

	/*===============================================================
	 * Get response packet
	 *==============================================================*/
	alarm(cfginfo->threshold);
#ifdef DES
	if ((rc = read(fd, ebuf, sizeof(ebuf))) != sizeof(ebuf)) {
#else
	if ((rc = read(fd, &resp, sizeof(resp))) != sizeof(resp)) {
#endif
	    error_no = errno;
	    alarm(0);
	    if (tv.tv_sec - cfginfo->net[i].lastlogged >= LONGWAIT) {
		sprintf(msg, "Error: Unable to read response from remote "
			"firewall %s, rc = 0x%08x.  %s.  Attention is "
			"required.",
			cfginfo->net[i].ip, rc, sys_errlist[error_no]);
		syslog(LOG_ERR, msg);
		if (cfginfo->verbosemail) {
		    sprintf(cmd, "/usr/bin/mailx -s \"%s\" %s </dev/null",
			msg, cfginfo->email);
		    if (rc = system(cmd))
			syslog(LOG_ERR,"Error: Attempt to mail attention "
				"notification to %s failed.  rc = 0x%08x.",
				cfginfo->email, rc);
		}
		cfginfo->net[i].lastlogged = tv.tv_sec;
	    }
	    continue;
	}
	alarm(0);

#ifdef DES
	/*===============================================================
	 * Decrypt
	 *==============================================================*/
	cp = (char *)&resp;
	for (j = 0; j < EBUFSIZ; j += 8)
	    des_ecb_encrypt((des_cblock *)(ebuf + j), (des_cblock *)(cp + j),
			ksched, DES_DECRYPT);
#endif
	if (ntohl(resp.code) != OK) {
	    if (tv.tv_sec - cfginfo->net[i].lastlogged >= LONGWAIT) {
		sprintf(msg, "Error: NOP request resulted in failure return "
			"code from remote firewall %s.  Attention is required.",
			cfginfo->net[i].ip);
		syslog(LOG_ERR, msg);
		if (cfginfo->verbosemail) {
		    sprintf(cmd, "/usr/bin/mailx -s \"%s\" %s </dev/null",
			msg, cfginfo->email);
		    if (rc = system(cmd))
			syslog(LOG_ERR,"Error: Attempt to mail attention "
				"notification to %s failed.  rc = 0x%08x.",
				cfginfo->email, rc);
		}
		cfginfo->net[i].lastlogged = tv.tv_sec;
	    }
	    continue;
	}

	/*===============================================================
	 * If we have logged an error message for this interface but got
	 * here this time, we want to log an entry that the interface is
	 * now OK.
	 *==============================================================*/
	if (cfginfo->net[i].lastlogged != (time_t)0) {
	    syslog(LOG_NOTICE, "Remote firewall %s is now responding normally.",
			cfginfo->net[i].ip);
	    /*===========================================================
	     * If startup != 0, this is the first success we got.   From
	     * now on, we report all failures.   By setting all lastlogged
	     * to zero, any failures will be reported immediately.
	     *==========================================================*/
	    if (startup.tv_sec != (time_t)0) {
		int j;
		for (j = 0; j < cfginfo->nif; j++)
		    cfginfo->net[j].lastlogged = (time_t)0;
	    }
	}
	close(fd);
	cfginfo->net[i].sent_nop = 1;
    }
    /*==================================================================
     * If there was success and no failure, reset startup time so
     * autotakeover will happen.
     *=================================================================*/
    if (*active != 0 && *inactive == 0)
	startup.tv_sec = (time_t)0;
    return;
}

/*==========================================================================
 *
 *=========================================================================*/
main(int ac,char **av)
{
   static  char	ident[] = "@(#) fwpulse, V 1.0, 00/01/21"
 			  " (C) Freemont Avenue Software, Inc."
			  " 1995-2000";
   int		inactive=0;
   int		active=0;
   int		rc;
   int		i;
   int		no_rpt_loops=0;
   pid_t	pid;
   char		tmpf[]="/tmp/.fwpulse";
   in_addr_t	*localip = NULL;
   long		nlips;
   struct cfginfo *cfginfo;
   FILE		*rtf;
   FILE		*pf;
#ifdef DES
    unsigned char	key[8];
   des_key_schedule	ksched;
#endif

   openlog("fwpulse",LOG_PID|LOG_NDELAY,LOG_DAEMON);

   signal(SIGCHLD, SIG_IGN);

   gettimeofday(&startup, NULL);
   srand48((int)startup.tv_sec);

   /*=======================================================================
    * If a reverse_takeover file exists, execute it, remove it before
    * continuing.
    *======================================================================*/
   if ((rtf = fopen(REVTAKEOVER, "r")) != NULL) {
	int rc;
	fclose(rtf);
	rc = system(REVTAKEOVER);
	if (WEXITSTATUS(rc) != 0) {
	    syslog(LOG_ERR, "Error: %s returned non-zero return code 0x%08x.  "
			"fwpulse aborting.", REVTAKEOVER, rc);
	    exit(1);
	}
	unlink(REVTAKEOVER);
   }

   close(0);
   close(1);
   close(2);

   setsid();
   chdir("/");

   if ((pid = fork()) < 0)
   {
	printf("%s cannot fork\n",av[0]);
	syslog(LOG_ERR,"Error: %s cannot fork. Fwpulse not active.",av[0]);
	exit(1);
   }
   if (pid != 0)
	exit(0);

    /*========================================================================
     * Write a pid file so the daemon can be signalled easily
     * If the open or the write failed, we continue without a pid file.
     *=======================================================================*/
    if (!(pf = fopen(PIDFILE,"w"))) {
	syslog(LOG_ERR,"Error: Failed opening %s for writes.  %s\n",
			PIDFILE,sys_errlist[errno]);
    }

    if (fprintf(pf,"%d\n",getpid()) < 0) {
	syslog(LOG_ERR,"Error: Failed writing to %s.  %s\n",PIDFILE,
			sys_errlist[errno]);
    }

    fclose(pf);

    chmod(PIDFILE,S_IRUSR|S_IWUSR|S_IRGRP);

    psignal(SIGTERM, sigterm);

   /*=======================================================================
    * Get the list of all the local ip addresses
    *======================================================================*/
   if (getlocal(&localip, NULL, &nlips) != 0)
	return(1);

   /*=======================================================================
    * Parse fwpulse.conf file.  If return code is non-zero, return immediately
    *======================================================================*/
   if ((rc=fwpulseconf(av[0],&cfginfo,localip,nlips)) != 0)
	return(rc);

   syslog(LOG_NOTICE, "fwpulse Version 1.0 started.  Awaiting first "
			"successful communication sequence with remote...");

   /*=======================================================================
    * Do not want to report a non-response when starting up - the remote
    * firewall may be starting up too.  Set the lastlogged to
    * LONGWAIT - (LONGWAIT - startup_delay) minutes ago so it won't consider
    * a failure until after startup_delay minutes.
    *======================================================================*/
   startup.tv_sec = startup.tv_sec - (LONGWAIT - cfginfo->startup_delay);
   for (i = 0; i < cfginfo->nif; i++)
	cfginfo->net[i].lastlogged = startup.tv_sec;

#ifdef DES
   swirl(cfginfo->private_key, key);
   des_set_key((des_cblock *)key, ksched);
#endif

   /*===========================================
    * Loop forever, sleeping between each loop
    *==========================================*/
   while (1)
   {
#ifdef DES
	pingloop(cfginfo, &active, &inactive, ksched);
#else
	pingloop(cfginfo, &active, &inactive);
#endif
	
	/*===================================================================
	 * autotakeover if necessary
	 *==================================================================*/
	if(inactive > 0)
	{
	    if (absolutely_no_takeover) {
		syslog(LOG_ERR, "Attention: A condition for automatic takeover "
				"was detected but the local machine could not "
				"ping one of its routers and decided to abort "
				"the automatic takeover attempt.");
	    }
	    else {
		if (time(NULL) - startup.tv_sec > LONGWAIT &&
				cfginfo->autotakeover)
		{
		    /*=======================================================
		     * Autotakeover
		     * If autotakeover() returns a non-zero return code, bypass
		     * the sleep and check again.
		     *======================================================*/
#ifdef DES
		    if (autotakeover(inactive, active, cfginfo, ksched))
#else
		    if (autotakeover(inactive, active, cfginfo))
#endif
			continue;
		}
	    }
	}

	sleep(cfginfo->wakeup);

   }

}
