/*============================================================================
 * file:	fwpulse/fwpulse.h
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=======================================================================
 * 
 * (C) COPYRIGHT Freemont Avenue Software, Inc. 1997-1999
 * All rights reserved.
 *
 * Date written:	February   8, 1996
 * Last modified:	February   8, 2000
 *
 *=========================================================================*/
#include <sys/time.h>
#include "firewall.h"

#define REVTAKEOVER	FW_CFG_DIR "/reverse_takeover"
#define PRETAKEOVER	FW_CFG_DIR "/pre_takeover"
#define POSTTAKEOVER	FW_CFG_DIR "/post_takeover"
#define PREREVTAKEOVER	FW_CFG_DIR "/pre_rev_takeover"
#define POSTREVTAKEOVER	FW_CFG_DIR "/post_rev_takeover"
#define FWPULSEHIST	FW_CFG_DIR "/fwpulse.hist"
#define PIDFILE		FW_CFG_DIR "/fwpulse.pid"

#define	LONGWAIT	30*60	/* 30 minutes wait */
#define default_wakeup	300
#define	MAX_ATTEMPTS	6

#define isalldigits(a) (strspn(a,"0123456789") == strlen(a))

struct hostid {
	char	ip[16];
	char	netmask[16];
	char	netif[16];
	char	router[16];
	char	local;
	char	alias;
	char	notakeover;
	time_t	lastlogged;
	unsigned short toport;
	int	sent_nop;
	int	down_suspect;
};

typedef struct hostid hostid;

struct cfginfo {
	char	hostname[81];
	char	email[81];
	int	wakeup;
	int	threshold;
	int	startup_delay;
	int	autotakeover;
	int	verbosemail;
      	unsigned short	private_port;
	unsigned char	private_key[8];
	int	nif;
	hostid	net[1];
};


/*=======================================================================
 * Since we use ECB to encrypt, it may be possible for someone to replay
 * part of the request (one 8-byte block).  We duplicate the info in the
 * structure and have each piece at different alignment so that we can
 * make sure that all pieces in one request belong together.
 * With this declaration:
 * code and current_time.tv_sec are in one des_cblock,
 * current_time.tv_usec and dup_time.tv_sec are in on des_cblock and
 * dup_tv.usec and code are in the third des_cblock.
 *======================================================================*/
struct commpkt {
	int	code;		/* operation or return code	*/
	struct timeval	current_time;
	unsigned long	down_ip;/* ip that's suspected down	*/
	unsigned long	dup_ip;	/* ip that's suspected down	*/
	struct timeval	dup_time;
	int	dup_code;	/* operation or return code	*/
};
typedef struct commpkt commpkt;

/*=======================================================================
 * Size of an encrypted commpkt.  It's the size of commpkt rounded up to
 * the next multiple of 8.
 *======================================================================*/
#define EBUFSIZ	((sizeof(commpkt)+7)/8)*8

/*=======================================================================
 * Operation codes for the code member in commpkt
 *======================================================================*/
#define	NOP		0
#define	SHUTDOWN	1
/*=======================================================================
 * Return codes for the code member in commpkt
 *======================================================================*/
#define	OK		0
#define REJECTED        1

/*=======================================================================
 * Scramble the private_key to get the encryption key
 *======================================================================*/
#define	swirl(pk, rk)	{	register int i; \
				i = pk[3] + pk[5]; \
				rk[6] = i - pk[1] + 2; \
				i = (pk[3] << 1)  + pk[4]; \
				rk[4] = (i - 7) & 0xff; \
				rk[3] = pk[2] ^ pk[5] ^ pk[0] - 3; \
				rk[1] = ((pk[5] & pk[1]) & 0x5a) + 1; \
				rk[7] = (pk[2] >> (pk[0] & 0x03)) ^ 0xcc; \
				i = pk[1] + pk[4] + pk[3] - pk[2] + pk[0]; \
				rk[0] = (i >> 1) & 0xfe; \
				rk[2] = 0x80 | pk[4] | (pk[6] & 0x49); \
				rk[5] = pk[7] + pk[2] - pk[6] + pk[1] + 11; \
			}

#define psignal(sig, handler)	{	struct sigaction act; \
					act.sa_handler = handler; \
					sigemptyset(&act.sa_mask); \
					act.sa_flags = 0; \
					sigaction(sig, &act, NULL); \
				}
