/*============================================================================
 * file:	fwpulse/fwpulsed.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=======================================================================
 *  
 * function:	Runs under inetd and listens for requests from a remote
 *		fwpulse and process SHUTDOWN requests if necessary.
 * 
 * (C) COPYRIGHT Freemont Avenue Software, Inc. 1999-2000
 * All rights reserved.   
 *
 * Date written:	December  11, 1998
 * Last modified:	February   8, 2000 - added LPL
 *
 *=========================================================================*/
#include	<ctype.h>  
#include	<stdio.h>  
#include	<stdlib.h>  
#include	<unistd.h>
#include	<syslog.h>
#include	<signal.h>
#include	<sys/stat.h>
#include	<sys/socket.h>
#include	<netinet/in.h>
#include	<net/if.h>
#include	<limits.h>
#include	<errno.h>
#include	<fcntl.h>
#include	<sys/types.h>
#include	"des.h"
#include	"firewall.h"
#include	"fwpulse.h"

static  char	ident[] = "@(#) fwpulsed, V 1.0, 00/01/21"
 			  " (C) Freemont Avenue Software, Inc."
			  " 1999-2000";
#ifdef DES
#define	respond(value)	{	gettimeofday(&tv, NULL); \
				resp.current_time.tv_sec = htonl(tv.tv_sec); \
				resp.current_time.tv_usec = htonl(tv.tv_usec); \
				resp.code = htonl(value); \
				resp.down_ip = 0; \
				resp.dup_ip = 0; \
				resp.dup_code = resp.code; \
				resp.dup_time.tv_sec = \
					resp.current_time.tv_sec; \
				resp.dup_time.tv_usec = \
					resp.current_time.tv_usec; \
				cp = (char *)&resp; \
				for (j = 0; j < EBUFSIZ; j += 8) \
				    des_ecb_encrypt((des_cblock *)(cp + j), \
						(des_cblock *)(ebuf + j), \
						ksched, DES_ENCRYPT); \
				if (write(0, ebuf, sizeof(ebuf)) < 0) { \
					syslog(LOG_ERR, "Error: Unable to " \
						"send response to remote " \
						"host.  %s.", \
						sys_errlist[errno]); \
						lockf(histfd, F_ULOCK, 0); \
						close(histfd); \
				} \
			}
#else
#define	respond(value)	{	gettimeofday(&tv, NULL); \
				resp.current_time.tv_sec = htonl(tv.tv_sec); \
				resp.current_time.tv_usec = htonl(tv.tv_usec); \
				resp.code = htonl(value); \
				resp.down_ip = 0; \
				resp.dup_ip = 0; \
				resp.dup_code = resp.code; \
				resp.dup_time.tv_sec = \
					resp.current_time.tv_sec; \
				resp.dup_time.tv_usec = \
					resp.current_time.tv_usec; \
				if (write(0, &resp, sizeof(resp)) < 0) { \
					syslog(LOG_ERR, "Error: Unable to " \
						"send response to remote " \
						"host.  %s.", \
						sys_errlist[errno]); \
						lockf(histfd, F_ULOCK, 0); \
						close(histfd); \
				} \
			}
#endif

#ifndef linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif
				
/*=======================================================================
 * Signal handler for alarm signals.  No need to do anything, just return.
 *======================================================================*/
void sigalrm(int sig) {
    psignal(SIGALRM, sigalrm);
    return;
}

/*======================================================================
 * pbind
 * input:	port number
 * output:	fd for the bound socket
 *======================================================================*/
int pbind(unsigned short debugport)
{
    struct	 sockaddr_in     mya;
    int  x;
    int  nread;
    int on = 1;

    if((x = socket(AF_INET,SOCK_STREAM,0)) < 0)
    {
	perror("socket");
	return(-1);
    }
    memset(&mya, 0, sizeof(mya));
    mya.sin_family = AF_INET;
    mya.sin_port = htons(debugport);

    if (setsockopt(x, SOL_SOCKET, SO_REUSEADDR, (const char *)&on, sizeof(on)) < 0)
	syslog(LOG_ERR, "Error: Unable to set socket option reuse.  %s.",
			sys_errlist[errno]);

    if(bind(x,(struct sockaddr *)&mya,sizeof(mya)))
    {
	perror("bind");
	return(-1);
    }
    if(listen(x,50) < 0)
    {
	perror("listen");
	return(-1);
    }
    return(x);
}

#define BINDDEBUG
/*============================================================================
 * main fwpulsed function 
 * inputs:   argument count & argument list
 * 
 * Debug option:   fwpulsed fwpulsed -D 7700
 *===========================================================================*/
main(int ac, char *argv[])
{   
    long		tempport;
    unsigned short	debugport = 9000;	/* default debug port	*/
    struct sockaddr_in	paddr;
    char		buf[BUFSIZ];
    char		fromip[BUFSIZ];
    int			valid_addr = 0;
    int			i;
    int			rc;
    int			not_inetd = 0;
    size_t		len;
    int			fd;
    int			histfd;
    time_t		lasttime = (time_t)0;
    in_addr_t		*localip = NULL;
    char		*portname[IFNAMSIZ];
    long		nlips;
    struct cfginfo	*cfginfo;
    commpkt		pkt;
    commpkt		lastpkt;
    commpkt		resp;
    struct timeval	tv;
    struct sockaddr_in	addr;
#ifdef DES
    unsigned char	key[8];
    des_key_schedule	ksched;
    int			j;
    char		*cp;
    unsigned char	ebuf[EBUFSIZ];
#endif

    openlog("fwpulsed",LOG_PID|LOG_NDELAY,LOG_DAEMON);
    psignal(SIGALRM, sigalrm);

    /*=======================================================================
     * Get the list of all the local ip addresses
     *======================================================================*/
    if (getlocal(&localip, NULL, &nlips) != 0)
	exit(1);

    /*=======================================================================
     * Parse fwpulse.conf file.  If return code is non-zero, return immediately
     *======================================================================*/
    if ((rc=fwpulseconf("fwpulsed",&cfginfo,localip,nlips)) != 0)
	exit(rc);

    /*===================================================================
     *	See if fwpulsed was started by inetd
     *=================================================================*/
    not_inetd = getpeername(0, (struct sockaddr *)&addr, (socklen_t *)&len);

    /*===================================================================
     *	If Debug is specified this it will bind the port and listen
     *	directly (circumventing inetd) so you can run the program under
     *	a debugger
     *=================================================================*/
    if (not_inetd) {
	pid_t pid;

	if( ac > 1)
	{
	    if( strcmp(argv[1],"-D") == 0)
	    {
		/* check for debug port number  */
		if( ac > 2 )
		{
		    if( (strlen(argv[2]) < 6)  && isalldigits(argv[2]) )
		    {
			if( (tempport = atoi(argv[2])) > 64000)
			{
			    printf("\nfwpulsed: Error: portnumber > "
				"64000. Please try again.\n");
			    return(-1);
			}
			debugport = tempport;
		    }
		    else
		    {
			printf("\nfwpulsed: Error: invalid portnumber "
				"(%s), must be "
				"a positive integer less than 64000.\n	    "
				"Please try again.\n",
			    argv[2]);
			return(-1);
		    }
		}
		printf("\n fwpulsed: Debug mode specified. Use port %d.\n",
			debugport);
		fd = pbind(debugport);
	    }
	}

	/*===================================================================
	 * Bind to the port in config file or the default port
	 *==================================================================*/
	else {
	    fd = pbind(cfginfo->private_port);
	}

	if (fd < 0)
	    return(-1);

	/*===================================================================
	 * No zombies
	 *==================================================================*/
	psignal(SIGCHLD, SIG_IGN);
	/*===================================================================
	 * Run as a daemon if not started by inetd
	 *==================================================================*/
	if ((pid = fork()) < 0) {
	    perror("fork failed");
	    syslog(LOG_ERR,"Error: fork failed.  %s",sys_errlist[errno]);
	    return(-1);
	}

	if (pid != 0)
	    exit(0);

	/*===================================================================
	 *	Dissociate from the controlling terminal
	 *=================================================================*/
	setsid();
	chdir("/");
    }

    /*===================================================================
     * If not started by inetd, have an infinite loop in parent to accept
     * new connections
     *=================================================================*/
    if (not_inetd) {
	syslog(LOG_NOTICE, "fwpulsed V 1.0 started.");
	while ( 1 ) {
	    pid_t pid;
	    int nread;

	    close(0);
	    if((nread = accept(fd,0,0)) < 0) {
		int en = errno;
		perror("accept");
		syslog(LOG_ERR,"accept failed.  %s.",sys_errlist[errno]);
		if (en == EPROTO || en == EINTR
#ifdef ECONNABORTED
			|| en == ECONNABORTED
#endif
#ifdef ECONNRESET
			|| en == ECONNRESET
#endif
#ifdef ETIMEDOUT
			|| en == ETIMEDOUT
#endif
#ifdef EHOSTUNREACH
			|| en == EHOSTUNREACH
#endif
#ifdef ENETUNREACH
			|| en == ENETUNREACH
#endif
		)
		    continue;
		exit(1);
	    }
	    if ((pid = fork()) < 0) {
		perror("fork failed");
		syslog(LOG_ERR,"Error: fork failed.  %s",sys_errlist[errno]);
		exit(1);
	    }
	    if (nread) {
		dup(nread);
		close(nread);
	    }
	    if (pid != 0)
		continue;

	    /*===========================================================
	     * Child closes listening socket
	     *==========================================================*/
	    close(fd);
	    break;
	}
    }

#ifdef DES
    swirl(cfginfo->private_key, key);
    des_set_key((des_cblock *)key, ksched);
#endif

    /*===================================================================
     * Get the address of the remote machine
     *=================================================================*/
    len = sizeof(paddr);
    if (getpeername(0, (struct sockaddr *)&paddr, (socklen_t *)&len) < 0) {
	syslog(LOG_ERR, "Error: getpeername failed.  %s.  Unable to continue.",
			sys_errlist[errno]);
	close(0);
	exit(1);
    }

    /*===================================================================
     * Make sure the remote address is one of those in fwpulse.conf
     *=================================================================*/
    strcpy(fromip, (char *)inet_ntoa(paddr.sin_addr));
    for (i = 0; i < cfginfo->nif; i++) {
	if ((!cfginfo->net[i].local || cfginfo->net[i].notakeover) &&
		strcmp(cfginfo->net[i].ip, fromip) == 0) {
	    valid_addr = 1;
	    break;
	}
    }

    if (!valid_addr) {
	syslog(LOG_ALERT, "Security Alert: Unauthorized attempt to connect "
			"to fwpulsed from address %s denied.", fromip);
	close(0);
	exit(1);
    }

    syslog(LOG_INFO, "Permitted connection from %s.", fromip);

    /*===================================================================
     * Read the request from the remote machine
     *=================================================================*/
    alarm(60 * 60);		/* Make sure recv() doesn't block for > 1 hr */
#ifdef DES
    if ((rc = recv(0, (char *)ebuf, sizeof(ebuf), 0)) <= 0) {
#else
    if ((rc = recv(0, &pkt, sizeof(pkt), 0)) <= 0) {
#endif
	if (rc < 0)
	    syslog(LOG_ERR, "Error: recv failed rc = %d.  %s.  Unable to "
			"continue.",
			rc, sys_errlist[errno]);
	close(0);
	exit(1);
    }

    /*===================================================================
     * Make sure request is correct length
     *=================================================================*/
#ifdef DES
    if (rc != sizeof(ebuf)) {
#else
    if (rc != sizeof(pkt)) {
#endif
	syslog(LOG_ERR, "Error: Expecting request packet to be %d bytes, "
			"received %d bytes.", sizeof(pkt), rc);
	close(0);
	exit(1);
    }

#ifdef DES
    /*===================================================================
     * Decrypt the request
     *=================================================================*/
    cp = (char *)&pkt;
    for (j = 0; j < EBUFSIZ; j += 8)
	    des_ecb_encrypt((des_cblock *)(ebuf + j), (des_cblock *)(cp + j),
			ksched, DES_DECRYPT);
#endif

    /*===================================================================
     * Open the history file
     *=================================================================*/
    if ((histfd = open(FWPULSEHIST, O_RDWR|O_CREAT, S_IRUSR|S_IWUSR)) < 0) {
	syslog(LOG_ERR, "Error: Unable to open fwpulse history file %s.  %s.",
			FWPULSEHIST, sys_errlist[errno]);
	close(0);
	exit(1);
    }

    /*===================================================================
     * Read the history file
     *=================================================================*/
#ifdef DES
    if ((len = read(histfd, ebuf, sizeof(ebuf))) < 0) {
#else
    if ((len = read(histfd, &lastpkt, sizeof(lastpkt))) < 0) {
#endif
	syslog(LOG_ERR, "Error: Unable to read fwpulse history file %s.  %s.",
			FWPULSEHIST, sys_errlist[errno]);
	close(histfd);
	close(0);
	exit(1);
    }

    /*===================================================================
     * Reset the history file pointer
     *=================================================================*/
    if (lseek(histfd, 0, SEEK_SET) < 0) {
	syslog(LOG_ERR, "Error: Unable to reset the fwpulse history file %s.  "
			"%s.",
			FWPULSEHIST, sys_errlist[errno]);
	close(histfd);
	close(0);
	exit(1);
    }

    /*===================================================================
     * If no history file exists, use the current time minus 5 minutes.
     * The 5 minutes is to provide for any possible date discrepencies
     * that may exist between the 2 machines.  If they are more than 5
     * minutes apart, then they will have to be synchronized better or
     * else fwpulsed will not proceed.
     *==================================================================*/
    if (len == 0) {
	gettimeofday(&tv, NULL);
	lastpkt.current_time.tv_sec = htonl(tv.tv_sec - 5 * 60);
	lastpkt.current_time.tv_usec = 0;
    }
#ifdef DES
    else if (len != sizeof(ebuf)) {
	syslog(LOG_ERR, "Error: Expecting %d bytes in history file %s but "
			"read %d bytes,  proceeding without history "
			"information.",
			sizeof(ebuf), FWPULSEHIST, len);
#else
    else if (len != sizeof(lastpkt)) {
	syslog(LOG_ERR, "Error: Expecting %d bytes in history file %s but "
			"read %d bytes,  proceeding without history "
			"information.",
			sizeof(lastpkt), FWPULSEHIST, len);
#endif
	gettimeofday(&tv, NULL);
	lastpkt.current_time.tv_sec = htonl(tv.tv_sec - 5 * 60);
	lastpkt.current_time.tv_usec = 0;
    }
#ifdef DES
    else  {
	/*===============================================================
	 * Decrypt the history record
	 *=============================================================*/
	cp = (char *)&lastpkt;
	for (j = 0; j < EBUFSIZ; j += 8)
	    des_ecb_encrypt((des_cblock *)(ebuf + j), (des_cblock *)(cp + j),
			ksched, DES_DECRYPT);
    }
#endif

    /*===================================================================
     * Check integrity of the request
     *=================================================================*/
    if (pkt.code != pkt.dup_code ||
	pkt.current_time.tv_sec != pkt.dup_time.tv_sec ||
	pkt.current_time.tv_usec != pkt.dup_time.tv_usec) {
	syslog(LOG_ERR, "Error: Request appears to be corrupted.  "
			"Hex dump: %08x %08x %08x %08x %08x %08x.",
			pkt.code, pkt.current_time.tv_sec,
			pkt.current_time.tv_usec, pkt.dup_time.tv_sec,
			pkt.dup_time.tv_usec, pkt.dup_code);
	close(histfd);
	close(0);
	exit(1);
    }

    /*===================================================================
     * Make sure the request is not a replay
     *=================================================================*/
    if (((unsigned int)ntohl(pkt.current_time.tv_sec) <
			(unsigned int)ntohl(lastpkt.current_time.tv_sec)) ||
		((unsigned int)ntohl(pkt.current_time.tv_sec) ==
			(unsigned int)ntohl(lastpkt.current_time.tv_sec) &&
		 (unsigned int)ntohl(pkt.current_time.tv_usec) <=
			(unsigned int)ntohl(lastpkt.current_time.tv_usec))) {
	time_t tmptime1, tmptime2;
	char tmp[BUFSIZ];
	tmptime1 = ntohl(lastpkt.current_time.tv_sec);
	strcpy(tmp, ctime(&tmptime1));
	tmptime2 = ntohl(pkt.current_time.tv_sec);
	syslog(LOG_ALERT, "Security Alert: Request to fwpulsed does not "
				"contain a current timestamp, possibly a "
				"replay attack.  Last timestamp at %s "
				"(+%d usec), "
				"new timestamp at %s (+%d usec).  Hex: old("
				"0x%08x, 0x%08x), new(0x%08x, 0x%08x)",
				tmp, ntohl(lastpkt.current_time.tv_usec),
				ctime(&tmptime2),
				ntohl(pkt.current_time.tv_usec),
				ntohl(lastpkt.current_time.tv_sec),
				ntohl(lastpkt.current_time.tv_usec),
				ntohl(pkt.current_time.tv_sec),
				ntohl(pkt.current_time.tv_usec));
	close(histfd);
	close(0);
	exit(1);
    }
    /*===================================================================
     * Make sure a valid request before we update history file
     *=================================================================*/
    if (ntohl(pkt.code) != NOP && ntohl(pkt.code) != SHUTDOWN) {
	syslog(LOG_ALERT, "Security Alert: Request value (%d) to fwpulsed "
				"is not valid.", ntohl(pkt.code));
	close(histfd);
	close(0);
	exit(1);
    }

#ifdef DES
    /*=======================================================================
     * Encrypt history record
     *======================================================================*/
    cp = (char *)&pkt;
    for (j = 0; j < EBUFSIZ; j += 8)
	des_ecb_encrypt((des_cblock *)(cp + j), (des_cblock *)(ebuf + j),
		ksched, DES_ENCRYPT);
#endif

    /*===================================================================
     * Lock the history file
     *=================================================================*/
    alarm(cfginfo->threshold);
    if (lockf(histfd, F_LOCK, 0) < 0) {
	alarm(0);
	if (errno == EINTR)
	    syslog(LOG_ERR, "Error: Threshold of %d seconds reached and "
				" fwpulsed was unable to acquire a "
				"lock on %s, attempting to write to the file "
				"anyway.", cfginfo->threshold, FWPULSEHIST);
	else
	    syslog(LOG_ERR, "Error: Cannot lock file %s.  %s.  fwpulsed "
				"attempting to write to the file anyway.",
				FWPULSEHIST, sys_errlist[errno]);
    }
    alarm(0);

    /*===================================================================
     * Write the new timestamp
     *=================================================================*/
#ifdef DES
    if (write(histfd, ebuf, sizeof(ebuf)) < 0) {
#else
    if (write(histfd, &pkt, sizeof(pkt)) < 0) {
#endif
	syslog(LOG_ERR, "Error: Unable to write to fwpulse history file %s.  "
			"%s.", FWPULSEHIST, sys_errlist[errno]);
	lockf(histfd, F_ULOCK, 0);
	close(histfd);
	close(0);
	exit(1);
    }
    /*===================================================================
     * Unlock the history file
     *=================================================================*/
    if (lockf(histfd, F_ULOCK, 0) < 0) {
	syslog(LOG_ERR, "Error: Cannot unlock file %s.  %s.  fwpulsed "
				"attempting to continue anyway.",
				FWPULSEHIST, sys_errlist[errno]);
    }
    close(histfd);
    /*===================================================================
     * See what the request is for
     *=================================================================*/
    if (ntohl(pkt.code) == NOP) {
	time_t tmptime = ntohl(pkt.current_time.tv_sec);;
	respond(OK);
	syslog(LOG_INFO, "Reponded OK to NOP request dated %s (+%d usec).  "
			"Exiting.",
			ctime(&tmptime), ntohl(pkt.current_time.tv_usec));
	close(0);
	exit(0);
    }
    else if (ntohl(pkt.code) == SHUTDOWN) {
	time_t tmptime = ntohl(pkt.current_time.tv_sec);
	syslog(LOG_NOTICE, "Checking network status ...");
	/*===============================================================
	 * Check to see if there are activities on the ip address that
	 * is supposed to be down.
	 *==============================================================*/
	if (active_if(pkt.down_ip, cfginfo) != 0) {
	    respond(REJECTED);
	    syslog(LOG_ERR, "Error: Received SHUTDOWN request from %s but "				"there are still activities on ip address %s which "
			" was marked as down.  Request rejected.",
			fromip, inet_ntoa(pkt.down_ip));
	    exit(1);
	}
	/*===============================================================
	 * Look at each local, non-alias network statement and see if we
	 * can ping the router.  If we can ping all the routers, then 
	 * conclude that the local machine and its network connections
	 * are OK and we should reject the SHUTDOWN request.
	 *==============================================================*/
	for (i = 0; i < cfginfo->nif; i++) {
	    char cmd[BUFSIZ];
	    if (!cfginfo->net[i].local || cfginfo->net[i].alias ||
			cfginfo->net[i].notakeover)
		continue;
#ifdef _AIX
	    sprintf(cmd,"/usr/sbin/ping -c 1 %s >/dev/null 2>&1",
				cfginfo->net[i].router);
#endif
#ifdef __hppa
	    sprintf(cmd,"/usr/sbin/ping %s -n 1 | grep ' 0%% packet loss' "
			">/dev/null 2>&1",
				cfginfo->net[i].router);
#endif
#ifdef __sun
	    sprintf(cmd,"/usr/sbin/ping %s >/dev/null 2>&1",
				cfginfo->net[i].router);
#endif
#ifdef linux
	    sprintf(cmd,"/sbin/ping -c 1 %s >/dev/null 2>&1",
				cfginfo->net[i].router);
#endif

	    alarm(cfginfo->threshold);
	    if (system(cmd)) {
		alarm(0);
		break;
	    }
	    alarm(0);
	}
	/*===============================================================
	 * If all the pings were successful, reject the SHUTDOWN request.
	 *==============================================================*/
	if (i >= cfginfo->nif) {
	    respond(REJECTED);
	    syslog(LOG_ERR, "Error: Received SHUTDOWN request from %s but "
				"network interfaces appear to be fine.  "
				"Request rejected.", fromip);
	    exit(1);
	}
	/*===============================================================
	 * kill fwpulse so it doesn't keep monitoring a dead interface
	 *==============================================================*/
	sprintf(buf, "kill -TERM `/usr/bin/cat %s`", PIDFILE);
	system(buf);

	/*===============================================================
	 * Responds OK now (before shutting down so the remote fwpulse
	 * knows that the SHUTDOWN command is accepted.
	 *==============================================================*/
	respond(OK);

	syslog(LOG_NOTICE, "Processing SHUTDOWN request from %s ...", fromip);
	for (i = 0; i < cfginfo->nif; i++) {
	    if (!(cfginfo->net[i].local) || cfginfo->net[i].notakeover)
		continue;
#ifndef linux
	    sprintf(buf, "/usr/sbin/ifconfig %s down", cfginfo->net[i].netif);
#else
	    sprintf(buf, "/sbin/ifconfig %s down", cfginfo->net[i].netif);
#endif
	    syslog(LOG_NOTICE, "Issuing command: \"%s\"", buf);
	    if (system(buf)) {
		syslog(LOG_ERR, "Error: Unable to shutdown net interface %s.",
				cfginfo->net[i].netif);
	    }
	}
	tmptime = ntohl(pkt.current_time.tv_sec);
	syslog(LOG_NOTICE, "SHUTDOWN request dated %s (+%d usec) processed.  "
			"Exiting.",
			ctime(&tmptime), ntohl(pkt.current_time.tv_usec));
	close(0);
	exit(0);
    }
}
