/*============================================================================
 * file:	fwpulse/takeover.c  
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=======================================================================
 *  
 * function:	Reads the config info from the /etc/firewall/fwpulse.conf file
 *		and setup the network interfaces accordingly.
 * 
 * (C) COPYRIGHT Freemont Avenue Software, Inc. 1997-2000
 * All rights reserved.   
 *                                                                      
 * Date written:	February   9, 1996     
 * Last modified:	February   8, 2000 - added LPL
 *                                                                      
 *=========================================================================*/
#include	<ctype.h>  
#include	<stdio.h>  
#include	<stdlib.h>  
#include	<string.h>  
#include	<syslog.h>  
#include	<limits.h>  
#include	<sys/types.h>
#include	<sys/socket.h>
#include	<netinet/in.h>
#include	<net/if.h>
#include	"fwpulse.h"

int fwpulseconf(char *, struct cfginfo**, in_addr_t *, long);
int arpwhohas(char *ifname, unsigned long ip);

/*==========================================================================
 *
 *=========================================================================*/
main(int ac,char **av) 
{ 
   static  char	ident[] = "@(#) takeover, V 1.0, 00/01/21"
 			  " (C) Freemont Avenue Software, Inc."
			  " 1995-2000";
   int		fd;  
   int		rc;
   int		i;
   char		buf[LINE_MAX];
   char		response;
   struct cfginfo *cfginfo;
   char		*portname[IFNAMSIZ];
   in_addr_t	*localip;
   long		nlips;

   /*=======================================================================
    * Get the list of all the local ip addresses
    *======================================================================*/
   if (getlocal(&localip, &portname, &nlips) != 0)
	return(1);

   if ((rc=fwpulseconf(av[0],&cfginfo,localip,nlips)) != 0)
	return(rc);

   /*=========================================================================
    * Prompt the user to make sure the primary network interfaces are inactive
    * before proceeding.
    *========================================================================*/
   if (strlen(cfginfo->hostname))
   {
	printf("\nThis program is about to setup this machine with the\n"
		"following hostname.  Enter \"Y\" to proceed or \"N\""
		"to bypass.\n");
	printf("%s\n",cfginfo->hostname);
	fgets(buf, LINE_MAX, stdin);
        buf[1] = '\0';
	if (strcasecmp(buf,"y") == 0)
	{
	    sprintf(buf,"/usr/bin/hostname %s",cfginfo->hostname);
	    if (system(buf))
	    {   
		syslog(LOG_ERR,"Error: takeover: hostname failed to set new "
			"hostname to %s.\n",
		cfginfo->hostname);
		printf("takeover: hostname failed to set new hostname "
			"to %s.\n",cfginfo->hostname);
		return(1);
	    }
	    printf("Hostname takeover successful.\n");
	}
   }
   printf("\nThe IP addresses specified in the " FW_CFG_DIR "/fwpulse.conf\n"
		"file will be presented below one at a time.\n");
   printf("Please check each IP address and enter a \"Y\" to its\n");
   printf("right if you want to takeover that address or enter an \"N\"\n");
   printf("to its right if you do not wish to takeover that address.\n");
   if (strlen(cfginfo->hostname))
	printf("\nHostname:	%s\n",cfginfo->hostname);
   printf("\nNetwork Interface     IP Address       takeover(Y/N)?\n");
   printf("-----------------     ---------------  --------------\n");

   printf("\n");
   for (i = 0; i < cfginfo->nif; i++)
   {
	char netmask[24];
	if (cfginfo->net[i].local || cfginfo->net[i].notakeover)
	    continue;
	printf("%-16s      %-16s       ",
		cfginfo->net[i].netif,cfginfo->net[i].ip);
	fgets(buf, LINE_MAX, stdin);
        buf[1] = '\0';
	if (strcasecmp(buf,"y") != 0)
	    continue;
	if (strlen(cfginfo->net[i].netmask) != 0)
	{
	    strcpy(netmask,"netmask ");
	    strcat(netmask,cfginfo->net[i].netmask);
	}
	else
	    strcpy(netmask,"");

#ifdef _AIX
	sprintf(buf,"/usr/sbin/ifconfig %s inet %s %s alias up",
		cfginfo->net[i].netif,cfginfo->net[i].ip,netmask);
#endif
#ifdef __sun
	sprintf(buf,"/usr/sbin/ifconfig %s inet %s %s up",
		cfginfo->net[i].netif,cfginfo->net[i].ip,netmask);
#endif
#ifdef __hppa
	sprintf(buf,"/usr/contrib/bin/ifalias %s add %s",
		cfginfo->net[i].netif,cfginfo->net[i].ip,netmask);
#endif
#ifdef __sun
	sprintf(buf,"/sbin/ifconfig %s inet %s %s up",
		cfginfo->net[i].netif,cfginfo->net[i].ip,netmask);
#endif
	printf("takeover: Issuing the following command:\n  %s\n",buf);
	if (system(buf))
	{  
           syslog(LOG_ERR,"Error: takeover: ifconfig failed, IP address %s "
			"cannot be "
			"activated on interface %s.\n", cfginfo->net[i].ip,
			cfginfo->net[i].netif);
	   printf("takeover: ifconfig failed, IP address %s cannot be\n"
		"          activated on interface %s.\n", cfginfo->net[i].ip,
		cfginfo->net[i].netif);
	   return(1);
	}
	else
	{
	   syslog(LOG_NOTICE,buf);
   	   printf("takeover: ifconfig successful, IP address %s\n"
		"          activated on interface %s.\n",cfginfo->net[i].ip,
		cfginfo->net[i].netif);
	
#if defined(_AIX) || defined(linux)
	   /*===========================================================
	    * Send a gratuitous ARP WHO_HAS request.  This is only needed
	    * on AIX and Linux since Solaris does it automatiacally.
	    *==========================================================*/
 	   arpwhohas(cfginfo->net[i].netif, inet_addr(cfginfo->net[i].ip));
#endif
	}
   }

   return(0);
} 
