/*============================================================================
 * filename:		Base.java
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
//============================================================================
//
// R. James Livermore grants Freemont Avenue Software, Inc. a nonexclusive
// license to use, modify, and market this program using their tradename
// T.Rex.
//
//
// (C) COPYRIGHT R. James Livermore 1998-2000.  All rights reserved.
//
// (C) COPYRIGHT Freemont Avenue Software, Incorporated 1998-2000.
// All rights reserved.
//
// Date written:	June      24, 1998
// Last modified:	February   8, 2000
//
//===========================================================================*/
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import jHelp.*;

public class Base extends Panel implements FocusListener,ActionListener {
	Label paneltitle;
	ScrollPane sp = new ScrollPane();
	BorderLayout bl = new BorderLayout();
	Button okbutton = new Button("OK");
	Button canbutton = new Button("Cancel");
	Button helpbutton = new Button("Help");
	Base parent_panel;
	Panel body = new Panel();
	Panel buttons = new Panel();
	String panel_prompt = "";
	String panel_prompt_override = null;
	String focus_prompt = null;
	boolean complete = true;
	String tmpfn;
	MyDialog dldialog;
	String dldisp;	// Whether to continue or cancel download of a
			// file that's checked out.
	int component;	// The component identifier (meaningless in some cases)
	int lineno = 0;	// line number of the config file line being read
	String lowname;

	// The fw variable is very important in that it enables all
	// instances of Base() to access methods and variables that are
	// owned by a display frame.  Multiple display frames can be active
	// simultaneously.  Each display frame has a PFrame() instance and
	// a fwwin() instance associated with it.  The fw variable
	// in each Base() instance references the fwwin() instance
	// associated with its display frame.
	// Note that if a Base() instance is the first one (top) in a window,
	// the fw variable is invalid during constrution of the Base()
	// instance.  So DO NOT (or at least be real careful) use the
	// fw variable until after the constructor is done and the
	// setRootwin() method has been called to set the fw variable.
	protected fwwin fw;	// This contains the root parent of
					// of each instance of Base

	// This constructor is called only for the very first panel in each
	// window
	Base(fwwin fwtop, String title, String prompt) {
	    fw = fwtop;
	    parent_panel = this;
	    newbase(title, prompt);

	}

	// This is the constructor that will be used most of the time, i.e.
	// after the first panel in each panel
	Base(Base parent, String title, String prompt) {
	    parent_panel = parent;
	    fw = parent_panel.fw;
	    newbase(title,prompt);
	}

	private void newbase(String title, String prompt) {
	    paneltitle = new Label(title,Label.CENTER);

	    panel_prompt = prompt;

	    setLayout(bl);

	    // Use Times Roman font
	    setFont(new Font("TimesRoman",Font.PLAIN,fwmain.fontsize));

	    // A title centered at the top
	    paneltitle.setFont(new Font("TimesRoman",Font.BOLD,14));
	    add("North",paneltitle);

	    // Put the scroll pane in the center
	    sp.getVAdjustable().setUnitIncrement(10);
	    sp.getHAdjustable().setUnitIncrement(10);
	    add("Center",sp);

	    // Add the OK, Cancel and help buttons at the bottom
	    addButtons();
	}

	/*=================================================================
	 * Always override this member to display the body for a proxy
	 *================================================================*/
	void addBody() {
	    sp.add(new Button("Dummy"));
	}

	/*=================================================================
	 * Override this member if an OK button is not to be displayed
	 * Default is to display an OK button
	 *================================================================*/
	void addOK() {
	    buttons.add(new Label(""));
	    buttons.add(okbutton);
	}

	/*=================================================================
	 * Override this member if a help button is not to be displayed
	 * Default is to display a Help button
	 *================================================================*/
	void addhelp() {
	    buttons.add(new Label(""));
	    buttons.add(helpbutton);
	}

	/*=================================================================
	 * Add buttons at the bottom of the panel
	 *================================================================*/
	void addButtons() {

	    buttons.setLayout(new GridLayout(1,7));
	    addOK();
	    buttons.add(new Label(""));
	    buttons.add(canbutton);
	    addhelp();
	    buttons.add(new Label(""));

	    // Set the command name so they will compare even internationalized
	    okbutton.setActionCommand("OK");
	    canbutton.setActionCommand("CANCEL");
	    helpbutton.setActionCommand("HELP");

	    // Register the action listeners for the buttons
	    okbutton.addActionListener(this);
	    canbutton.addActionListener(this);
	    helpbutton.addActionListener(this);

	    // Put the buttons at the bottom of the Base panel
	    add("South",buttons);
	}

	/*=================================================================
	 * Always override this member to handle the OK button
	 * Default handling is to print a message to the tty
	 *================================================================*/
	public void handleOK() {
	    System.out.println("OK button clicked");
	}

	/*=================================================================
	 * Always override this member to handle the Cancel button
	 * Default handling is to print a message to the tty
	 *================================================================*/
	public void handleCan() {
	    fw.newpanel(parent_panel);
	}

	/*=================================================================
	 * Always override this member to handle the help button
	 * Default handling is to print a message to the tty
	 *================================================================*/
	public void handleHelp() {
	    String JAR;
	    boolean dohi = false;
	    Dimension d = getPreferredSize();
	    String htmlfile = toString();

	    if (System.getProperty("file.separator").equals("\\"))
		JAR = new String(fwmain.dosdir + "trexdocs.jar");
	    else
		JAR = new String("/usr/local/etc/trexdocs.jar");
	    htmlfile = htmlfile.substring(0,htmlfile.indexOf('['));
	    if (htmlfile.equals("tnproxy") || htmlfile.equals("aproxy") ||
			htmlfile.equals("gwuser"))
		htmlfile = lowname;
	    htmlfile = htmlfile.concat(".html");
	    if (fwmain.hf == null) {
		fwmain.hf = new htmlHelpFrame(fw.getFrame(), JAR,
							d.width, d.height);
		dohi = true;
	    }
	    fwmain.hf.addFile(fw.getFrame(), htmlfile);
	    fwmain.hf.show(htmlfile);
	    fwmain.hf.setVisible(true);

	    if (dohi) {
		new htmlHelpIndexFrame("Index", "html.index",
							"html.wdx",
							fw.getFrame(),
							fwmain.hf,
							JAR);
	    }
	}

	/*=================================================================
	 * Returns the parent of this panel
	 *================================================================*/
	public Base getParentPanel() {
	    return parent_panel;
	}

	/*=================================================================
	 * Returns the completion status of this instance
	 *================================================================*/
	public boolean isComplete() {
	    return complete;
	}

	/*=================================================================
	 * Returns the prompt string for this panel
	 * If an override is present, return it instead
	 *================================================================*/
	public String getPrompt() {
	    if (panel_prompt_override != null) {
		String tmp = panel_prompt_override;
		panel_prompt_override = null;
		return tmp;
	    }
	    else
		return panel_prompt;
	}

	/*=================================================================
	 * Sets the fw variable of this instance
	 *================================================================*/
	public void setRootwin(fwwin p) {
	    fw = p;
	}

	/*=================================================================
	 * Override getPreferredSize() so that we can control the size
	 *================================================================*/
	public Dimension getPreferredSize() {
	    Dimension d = getSize();
	    if (d.height < 500)
		d.height = 500;
	    if (d.width < 500)
		d.width  = 500;
	    return d;
	}

	/*=================================================================
	 * This overrides the default insets
	 *================================================================*/
	public Insets getInsets() {
	    return new Insets(10,10,10,10);
	}

	/*=================================================================
	 * Each time a component gets the focus, let the server know we
	 * are still alive.  If focus_prompt is non-null, it means someone
	 * wanted the string in focus_prompt displayed on the msgline while
	 * focus is moved.  E.g., when an error is found in a TextField,
	 * the error message is put into focus_prompt and then requestFocus()
	 * is called to transfer the focus to the field in error.
	 *================================================================*/
	public void focusGained(FocusEvent e) {
	    pktinfo pktrc;
	    if (focus_prompt == null)
		fw.msgline.setText("");
	    else {
		fw.msgline.setText(focus_prompt);
		focus_prompt = null;
	    }
	    pktrc = fw.c.sendpkt(fw.c.TIMER,fw.c.KEEPALIVE,"",1);
	    if (pktrc.rc != cglue.OK) {
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefd();
	    }
	    return;
	}

	/*==================================================================
	 * This is defined so that we implement the full KeyListener interface
	 * otherwise, the compiler complains
	 *=================================================================*/
	public void focusLost(FocusEvent e) {
	}

	/*=================================================================
	 * Override this method if the panel contains action buttons
	 * Also, don't forget to addActionListener() for each button
	 * Default here is no action
	 *================================================================*/
	public void menubuttons(String s) {
	    return;
	}


	/*=================================================================
	 * Download config file from the server
	 *================================================================*/
	public pktinfo download() {
	    pktinfo pktrc;

	    pktrc = fw.c.getconf(component);
	    if (pktrc.rc < 0) {
		complete = false;
		if (pktrc.rc != fw.c.NONFATAL) {
		    fw.c.closefd();
		    fw.newpanel(fw.getTop());
		}
		fw.msgline.setText(pktrc.errmsg);
		return(pktrc);
	    }

	    // Inform the user that a local save file was used instead of a 
	    // download
	    if (pktrc.rc == fw.c.LOCALFILE) {
		panel_prompt_override = new String("Panel shows changes " +
			"saved locally from last use, cancel and re-enter " +
			"if this is not desirable");
		pktrc.rc = fw.c.OK;
		return(pktrc);
	    }

	    if (pktrc.rc == fw.c.CONTINUE) {
		// Put up a Dialog to inform that someone else may be editing
		// the same config file and ask if we should continue
		Button filecont = new Button("Continue");
		Button filecan = new Button("Cancel");
		dldisp = new String("");
		Panel p = new Panel();

		// Put the "Continue" and "Cancel" buttons in the panel p
		p.setLayout(new GridLayout(1,5));
		p.add(new Label(""));
		filecont.setActionCommand("FILECONT");
		filecont.addActionListener(this);
		p.add(filecont);
		p.add(new Label(""));
		filecan.setActionCommand("FILECAN");
		filecan.addActionListener(this);
		p.add(filecan);
		p.add(new Label(""));

		dldialog = new MyDialog(fw.getFrame(),
				paneltitle.getText(),true,
				new Label(pktrc.errmsg),
				new Label("Press the appropriate button to " +
						"continue or cancel"), p);
		dldialog.setVisible(true);

		// Just returned from the event handler, continue based on
		// whether the "Continue" button or the "Cancel" button was
		// pressed
		File tmpfile = new File(pktrc.str);
		tmpfile.delete();
		if (dldisp.equals("FILECAN")) {
		    complete = false;
		    fw.c.sendpkt(component,fw.c.CANCEL,"",1);
		    pktrc.rc = fw.c.CONTINUE;
		    return(pktrc);
		}
		else {
		    return(download());
		}
	    }

	    tmpfn = pktrc.str;	// this may be redundant in some cases
	    return(pktrc);
	}

	/*=================================================================
	 * Upload config file to the server
	 *================================================================*/
	public pktinfo upload() {
	    pktinfo pktrc;

	    pktrc = fw.c.putconf(component,tmpfn);
	    if (pktrc.rc < 0) {
		fw.msgline.setText(pktrc.errmsg);
		if (pktrc.rc != fw.c.NONFATAL) {
		    fw.c.closefd();
		    fw.newpanel(fw.getTop());
		}
	    }
	    return(pktrc);
	}

	/*===================================================================
	 * Queries the server to find out what components are present there.
	 * Returns a 64 bit number which represents the components
	 *==================================================================*/
	long getcomps() {
	    pktinfo pktrc;
	    String comp;
	    int number;
	    long retcode = 0;

	    pktrc = fw.c.sendpkt(cglue.ALL, cglue.GETLIST, "", 1);
	    if (pktrc.rc < 0) {
		fw.msgline.setText(pktrc.errmsg);
		return((long)pktrc.rc);
	    }

	    // Get the response
	    pktrc = fw.c.recvpkt(1);
	    if (pktrc.rc < 0) {
		fw.msgline.setText(pktrc.errmsg);
		return((long)pktrc.rc);
	    }

	    StringTokenizer st = new StringTokenizer(pktrc.str, " ");

	    while (true) {
		try {
		    comp = st.nextToken();
		} catch (NoSuchElementException e) {
		    break;
		}

		try {
		    number = Integer.parseInt(comp);
		} catch (NumberFormatException e) {
		    fw.msgline.setText("Component number format error in " +
				"GETLIST response");
		    return((long)cglue.NONFATAL);
		}
		retcode |= 1 << (number) - 1;
	    }
	    return(retcode);
	}

	/*=================================================================
	 * Override this method if additional action are to be processed
	 * Return true if method processed an action (so actionPerformed()
	 *	will return immediately)
	 * Return false if method did not process an action  (so
	 *	actionPerformed() will continue its processing
	 *================================================================*/
	boolean moreAction(String s) {
	    return(false);
	}

	/*=================================================================
	 * Event handler that routes the button actions to the appropriate
	 * member
	 *================================================================*/
	public void actionPerformed(ActionEvent e) {
	    String s = e.getActionCommand();

	    this.fw.msgline.setText("");
	    if (fw.c.fd >= 0) {
		pktinfo pktrc;
		pktrc = fw.c.sendpkt(fw.c.TIMER,fw.c.KEEPALIVE,
							"",1);
		if (pktrc.rc != cglue.OK) {
		    fw.msgline.setText(pktrc.errmsg);
		    fw.c.closefd();
		    fw.newpanel(fw.getTop());
		    return;
		}
	    }

	    // First see if special action is wanted
	    if (moreAction(s))
		return;

	    if ("OK".equals(s)) {
		handleOK();
		return;
	    }

	    if ("CANCEL".equals(s)) {
		handleCan();
		return;
	    }

	    if ("HELP".equals(s)) {
		handleHelp();
		return;
	    }
	    if ("FILECONT".equals(s) || "FILECAN".equals(s)) {
		dldisp = s;
		dldialog.dispose();
		return;
	    }
	    menubuttons(s);
	}
}
