/*============================================================================
 * filename:		aproxy.java
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
//============================================================================
//
// R. James Livermore grants Freemont Avenue Software, Inc. a nonexclusive
// license to use, modify, and market this program using their tradename
// T.Rex.
//
//
// (C) COPYRIGHT R. James Livermore 1998-2000.  All rights reserved.
//
// (C) COPYRIGHT Freemont Avenue Software, Incorporated 1998-2000.
// All rights reserved.
//
// Date written:	June      24, 1998
// Last modified:	February   8, 2000
//
//===========================================================================*/
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

public class aproxy extends Base {
    TextField gproxpath_f;
    TextField timeout_f;
    TextField nprocs_f;
    TextField minspare_f;
    TextField maxspare_f;
    TextField maxprocs_f;
    TextField checkprocs_f;
    TextField maxuse_f;
    TextField logformat_f;
    Choice dns_f;
    TextField httplog_f;
    Choice trace_f;
    Rules rules_f;
    String gproxpath;
    String timeout;
    String nprocs;
    String minspare;
    String maxspare;
    String maxprocs;
    String checkprocs;
    String maxuse;
    String logformat;
    boolean dns;
    String httplog;
    boolean trace;
    String rules;
    pktinfo pktrc;
    boolean gen;		// true = aproxy, false = webgate
    public final static boolean GEN = true;
    public final static boolean WEB = false;
    String capname;

    aproxy(Base p, boolean gen) {
	super(p,"APROXY","");
	this.gen = gen;
	addBody();
    }

    void addBody() {
	int rc;

	// Initialize for aproxy or webgate based on the gen parameter
	if (gen == GEN) {
	    component = fw.c.APROXY;
	    capname = new String("APROXY");
	    lowname = new String("aproxy");
	}
	else {
	    component = fw.c.WEBGATE;
	    capname = new String("WEBGATE");
	    lowname = new String("webgate");
	    paneltitle.setText(capname);
	}

	// Initialize the values for this panel
	gproxpath = new String("/home/hermes");
	timeout = new String("300");
	nprocs = new String("20");
	minspare = new String("5");
	maxspare = new String("20");
	maxprocs = new String("25");
	maxuse = new String("500");
	checkprocs = new String("5");
	logformat = new String("CLFExt");
	dns = false;
	httplog = new String("http.access.log");
	trace = false;

	GridBagLayout gbl = new GridBagLayout();
	GridBagConstraints gbc = new GridBagConstraints();
	body.setLayout(gbl);

	// Get the config file over here to a temp file
	pktrc = download();
	tmpfn = pktrc.str;

	if (pktrc.rc != fw.c.OK) {
	    complete = false;
	    return;
	}

	// Construct the rules box now so it can collect permit/deny lines
	// while we read the config file
	if (gen)
	    rules_f = new aproxyrules(fw);
	else
	    rules_f = new webgaterules(fw);

	// Read in the temp file
	if ((rc = readconf()) == fw.c.FAILED) {
	    fw.c.closefd();
	    fw.newpanel(fw.getTop());
	}

	if (rc < 0)
	    return;

	gbc.anchor = GridBagConstraints.WEST;
	gbc.gridwidth = GridBagConstraints.REMAINDER;
	Button edit_b = new Button("Edit");
	edit_b.addActionListener(this);
	gbl.setConstraints(edit_b,gbc);
	body.add(edit_b);

	gbc.gridwidth = GridBagConstraints.RELATIVE;
	Label gproxpath_l = new Label("Directory which " + lowname +
		" should run under");
	gbl.setConstraints(gproxpath_l,gbc);
	body.add(gproxpath_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	gproxpath_f = new TextField(gproxpath,20);
	gbl.setConstraints(gproxpath_f,gbc);
	body.add(gproxpath_f);
	gproxpath_f.addFocusListener(this);

	gbc.gridwidth = GridBagConstraints.RELATIVE;
	Label timeout_l = new Label("Maximum idle time in seconds before " +
		"disconnect");
	gbl.setConstraints(timeout_l,gbc);
	body.add(timeout_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	timeout_f = new TextField(timeout,4);
	gbl.setConstraints(timeout_f,gbc);
	body.add(timeout_f);
	timeout_f.addFocusListener(this);

	gbc.gridwidth = GridBagConstraints.RELATIVE;
	Label nprocs_l = new Label("Number of copies of " + lowname +
			" to run");
	gbl.setConstraints(nprocs_l,gbc);
	body.add(nprocs_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	nprocs_f = new TextField(nprocs,4);
	gbl.setConstraints(nprocs_f,gbc);
	body.add(nprocs_f);
	nprocs_f.addFocusListener(this);

	gbc.gridwidth = GridBagConstraints.RELATIVE;
	Label minspare_l = new Label("Minimum number of spare idle " +
				lowname + " processes");
	gbl.setConstraints(minspare_l,gbc);
	body.add(minspare_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	minspare_f = new TextField(minspare,4);
	gbl.setConstraints(minspare_f,gbc);
	body.add(minspare_f);
	minspare_f.addFocusListener(this);

	gbc.gridwidth = GridBagConstraints.RELATIVE;
	Label maxspare_l = new Label("Maximum number of spare idle " +
				lowname + " processes");
	gbl.setConstraints(maxspare_l,gbc);
	body.add(maxspare_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	maxspare_f = new TextField(maxspare,4);
	gbl.setConstraints(maxspare_f,gbc);
	body.add(maxspare_f);
	maxspare_f.addFocusListener(this);

	gbc.gridwidth = GridBagConstraints.RELATIVE;
	Label maxprocs_l = new Label("Maximum number of " + lowname +
				" processes allowed");
	gbl.setConstraints(maxprocs_l,gbc);
	body.add(maxprocs_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	maxprocs_f = new TextField(maxprocs,4);
	gbl.setConstraints(maxprocs_f,gbc);
	body.add(maxprocs_f);
	maxprocs_f.addFocusListener(this);

	gbc.gridwidth = GridBagConstraints.RELATIVE;
	Label maxuse_l = new Label("Number of times to use a process " +
				"before replacing it");
	gbl.setConstraints(maxuse_l,gbc);
	body.add(maxuse_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	maxuse_f = new TextField(maxuse,4);
	gbl.setConstraints(maxuse_f,gbc);
	body.add(maxuse_f);
	maxuse_f.addFocusListener(this);

	gbc.gridwidth = GridBagConstraints.RELATIVE;
	Label checkprocs_l = new Label("Interval in seconds between " +
				"checking and adjusting " + lowname +
				" processes");
	gbl.setConstraints(checkprocs_l,gbc);
	body.add(checkprocs_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	checkprocs_f = new TextField(checkprocs,4);
	gbl.setConstraints(checkprocs_f,gbc);
	body.add(checkprocs_f);
	checkprocs_f.addFocusListener(this);

	if (gen == WEB) {
	    gbc.gridwidth = GridBagConstraints.RELATIVE;
	    Label httplog_l = new Label("HTTP log file name");
	    gbl.setConstraints(httplog_l,gbc);
	    body.add(httplog_l);

	    gbc.gridwidth = GridBagConstraints.REMAINDER;
	    httplog_f = new TextField(httplog,16);
	    gbl.setConstraints(httplog_f,gbc);
	    body.add(httplog_f);
	    httplog_f.addFocusListener(this);

	    gbc.gridwidth = GridBagConstraints.RELATIVE;
	    Label logformat_l = new Label("HTTP log format");
	    gbl.setConstraints(logformat_l,gbc);
	    body.add(logformat_l);

	    gbc.gridwidth = GridBagConstraints.REMAINDER;
	    logformat_f = new TextField(logformat,20);
	    gbl.setConstraints(logformat_f,gbc);
	    body.add(logformat_f);
	    logformat_f.addFocusListener(this);
	}

	gbc.gridwidth = GridBagConstraints.RELATIVE;
	Label trace_l = new Label("Turn trace on");
	gbl.setConstraints(trace_l,gbc);
	body.add(trace_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	trace_f = new Choice();
	trace_f.addItem("Yes");
	trace_f.addItem("No");
	if (trace)
	    trace_f.select("Yes");
	else
	    trace_f.select("No");
	gbl.setConstraints(trace_f,gbc);
	body.add(trace_f);
	trace_f.addFocusListener(this);

	gbl.setConstraints(rules_f,gbc);
	body.add(rules_f);

	sp.add(body);
    }

    /*=======================================================================
     * Reads the temp file in and interpret it
     *
     * Returns 0 if successful
     *		-1 if failed
     *		NONFATAL if connection need not be closed
     *
     *======================================================================*/
    int readconf() {
	int fd;
	int found_nprocs = 0;
	int found_minspare = 0;
	int found_maxspare = 0;
	int found_maxprocs = 0;

	if (tmpfn.equals("")) {
	    fw.msgline.setText("Internal error: " + lowname + " config " +
			"file has not been downloaded yet");
	    return(fw.c.NONFATAL);
	}
	File tmpfile = new File(tmpfn);
	if (!tmpfile.exists() || !tmpfile.isFile()) {
	    fw.msgline.setText("Internal error: the downloaded " +
			lowname + " conf file cannot be found");
	    return(fw.c.NONFATAL);
	}


	pktrc = fw.c.opencbcfile(tmpfn, 0);
	if (pktrc.rc < 0) {
	    fw.msgline.setText(pktrc.errmsg);
	    return(pktrc.rc);
	}

	fd = pktrc.rc;

	while ( true ) {
	    String tmp;
	    String statement;

	    pktrc = fw.c.cbcreadLine(fd);
	    if (pktrc.rc != cglue.OK) {
		if (pktrc.rc == cglue.CLOSED)
		    break;
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		return(pktrc.rc);
	    }

	    // Exit loop at EOF
	    if (pktrc.str == null)
		break;

	    lineno++;

	    // Get rid of comment lines (first character is #) and empty lines
	    if (pktrc.str.trim().length() == 0 ||
			pktrc.str.trim().substring(0,1).equals("#"))
		continue;

	    StringTokenizer st = new StringTokenizer(pktrc.str.trim()," \t=:");

	    // Get the statement type
	    try {
		statement = st.nextToken().toLowerCase();
	    } catch (NoSuchElementException e) {
		continue;
	    }

	    // Get the next token into tmp - ignore error with no token found
	    // We do this here so we don't have to do a "try" for every
	    // statement type
	    try {
		tmp = st.nextToken();
	    } catch (NoSuchElementException e) {
		complete = false;
		fw.c.closefile(fd);
		tmpfile.delete();
		fw.newpanel(new edit(this,parent_panel,"EDIT " + capname));
		fw.msgline.setText("Highlighted line contains an " +
			"invalid statement, please correct and retry");
		return(fw.c.NONFATAL);
	    }

	    if (statement.equals("gproxpath")) {
		gproxpath = tmp;
		continue;
	    }

	    if (statement.equals("timeout")) {
		timeout = tmp;
		continue;
	    }

	    if (statement.equals("trace")) {
		if (tmp.toLowerCase().equals("yes"))
		    trace = true;
		continue;
	    }

	    // Process nprocs
	    if (statement.equals("nprocs")) {
		found_nprocs = lineno;
		nprocs = tmp;
		try {
		    tmp = st.nextToken();
		} catch (NoSuchElementException e) {
		    minspare = nprocs;
		    int i;
		    try {
			i = Integer.valueOf(nprocs).intValue();
			i += 5;
			maxspare = Integer.toString(i);
		    } catch (NumberFormatException f) {
			complete = false;
			fw.c.closefile(fd);
			fw.newpanel(new edit(this,parent_panel,
					"EDIT " + capname));
			fw.msgline.setText("nprocs on the highlighted "+
				"line must be numeric");
			return(fw.c.NONFATAL);
		    }
		    continue;
		}
		found_minspare = lineno;
		minspare = tmp;
		try {
		    tmp = st.nextToken();
		} catch (NoSuchElementException e) {
		    int i;
		    try {
			i = Integer.valueOf(nprocs).intValue();
			i += 5;
			maxspare = Integer.toString(i);
		    } catch (NumberFormatException f) {
			complete = false;
			fw.c.closefile(fd);
			fw.newpanel(new edit(this,parent_panel,
					"EDIT " + capname));
			fw.msgline.setText("nprocs on the highlighted "+
				"line must be numeric");
			return(fw.c.NONFATAL);
		    }
		    continue;
		}
		found_maxspare = lineno;
		maxspare = tmp;
		continue;
	    }

	    // Process maxprocs
	    if (statement.equals("maxprocs")) {
		found_maxprocs = lineno;
		maxprocs = tmp;
		continue;
	    }

	    // Process checkprocs
	    if (statement.equals("checkprocs")) {
		checkprocs = tmp;
		continue;
	    }

	    // Process maxuse
	    if (statement.equals("maxuse")) {
		maxuse = tmp;
		continue;
	    }

	    if (gen == WEB) {
		// Process httplog
		if (statement.equals("httplog")) {
		    httplog = tmp;
		    continue;
		}

		// Process logformat
		if (statement.equals("logformat")) {
		    logformat = tmp;
		    continue;
		}
	    }

	    if (statement.equals("permit") || statement.equals("deny")) {
		rules_f.body.addItem(pktrc.str);
		continue;
	    }

	    // Invalid statement
	    complete = false;
	    fw.c.closefile(fd);
	    fw.newpanel(new edit(this,parent_panel,"EDIT " + capname));
	    fw.msgline.setText("Highlighted line contains an invalid " +
		"statement, please correct and retry");
	    return(fw.c.NONFATAL);
	}

	int nprocs_n;
	int minspare_n;
	int maxspare_n;
	int maxprocs_n;
	try {
	    nprocs_n = Integer.valueOf(nprocs).intValue();
	} catch (NumberFormatException e) {
	    complete = false;
	    fw.c.closefile(fd);
	    lineno = found_nprocs;
	    fw.newpanel(new edit(this,parent_panel,"EDIT " + capname));
	    fw.msgline.setText("nprocs on the highlighted line " +
				"must be numeric");
	    return(fw.c.NONFATAL);
	}
	try {
	    minspare_n = Integer.valueOf(minspare).intValue();
	} catch (NumberFormatException e) {
	    complete = false;
	    fw.c.closefile(fd);
	    lineno = found_minspare;
	    fw.newpanel(new edit(this,parent_panel,"EDIT " + capname));
	    fw.msgline.setText("minspare on the highlighted line " +
				"must be numeric");
	    return(fw.c.NONFATAL);
	}
	try {
	    maxspare_n = Integer.valueOf(maxspare).intValue();
	} catch (NumberFormatException e) {
	    complete = false;
	    fw.c.closefile(fd);
	    lineno = found_maxspare;
	    fw.newpanel(new edit(this,parent_panel,"EDIT " + capname));
	    fw.msgline.setText("maxspare on the highlighted line " +
				"must be numeric");
	    return(fw.c.NONFATAL);
	}
	try {
	    maxprocs_n = Integer.valueOf(maxprocs).intValue();
	} catch (NumberFormatException e) {
	    complete = false;
	    fw.c.closefile(fd);
	    lineno = found_maxprocs;
	    fw.newpanel(new edit(this,parent_panel,"EDIT " + capname));
	    fw.msgline.setText("maxprocs on the highlighted line " +
				"must be numeric");
	    return(fw.c.NONFATAL);
	}
	// Make sure nprocs, minspare, maxspare, maxprocs all contain
	// values that are OK relative to each other.
	// First take care to make sure minspare must be less than or
	// equal to maxspare.
	if (minspare_n > maxspare_n) {
	    // If both minspare and maxspare are specified, issue error
	    if (found_minspare != 0 && found_maxspare != 0) {
		complete = false;
		fw.c.closefile(fd);
		lineno = found_maxspare;
		fw.newpanel(new edit(this,parent_panel,"EDIT " + capname));
		fw.msgline.setText("maxspare on the highlighted line " +
				"must not be less than minspare");
		return(fw.c.NONFATAL);
	    }
	    // If maxspare is not specified, default to minspare + 5
	    if (found_maxspare == 0) {
		maxspare_n = minspare_n + 5;
		maxspare = Integer.toString(minspare_n + 5);
	    }
	    // If maxspare specified but not minsapre, default minspare
	    // to maxspare - 5, if maxspare - 5 < 0, default to maxspare.
	    else if (found_minspare == 0) {
		minspare_n = maxspare_n - 5;
		minspare = Integer.toString(maxspare_n - 5);
		if (minspare_n < 0)
		    minspare = maxspare;
	    }
	}
	// Next, make sure maxprocs is greater than or equal to nprocs
	if (maxprocs_n < nprocs_n) {
	    // If both nprocs and maxprocs are specified, issue error
	    if (found_nprocs != 0 && found_maxprocs != 0) {
		complete = false;
		fw.c.closefile(fd);
		lineno = found_maxprocs;
		fw.newpanel(new edit(this,parent_panel,"EDIT " + capname));
		fw.msgline.setText("maxprocs on the highlighted line " +
				"must not be greater than nprocs on line " +
				found_nprocs);
		return(fw.c.NONFATAL);
	    }
	    // If maxprocs is not specified, default to nprocs + 5
	    if (found_maxprocs == 0) {
		    maxprocs_n = nprocs_n + 5;
		    maxprocs = Integer.toString(nprocs_n + 5);
	    }
	    // If maxprocs specified and but not nprocs, default nprocs
	    // to 5, if maxprocs - 5 < 0, default to maxprocs.
	    else if (found_nprocs == 0) {
		nprocs_n = 20;
		nprocs = new String("20");
		if (nprocs_n > maxprocs_n)
		    nprocs = maxprocs;
	    }
	}
	// Next, make sure maxprocs is greater than or equal to maxspare
	if (maxprocs_n < maxspare_n) {
	    // If both maxspare and maxprocs are specified, issue error
	    if (found_maxspare != 0 && found_maxprocs != 0) {
		complete = false;
		fw.c.closefile(fd);
		lineno = found_maxprocs;
		fw.newpanel(new edit(this,parent_panel,"EDIT " + capname));
		fw.msgline.setText("maxprocs on the highlighted line " +
				"must not less than maxspare on line " +
				found_maxspare);
		return(fw.c.NONFATAL);
	    }
	    // If maxspare is specified, default maxprocs to maxspare + 5
	    if (found_maxspare != 0) {
		maxprocs_n = maxspare_n + 5;
		maxprocs = Integer.toString(maxspare_n + 5);
	    }
	    // If maxspare not specified but maxprocs is, default maxspare
	    // to maxprocs - 5;
	    else if (found_maxprocs != 0) {
		maxspare_n = maxprocs_n - 5;
		maxspare = Integer.toString(maxprocs_n - 5);
		// If the new default for maxspare 
		if (maxspare_n < minspare_n) {
		    if (found_minspare != 0 && maxprocs_n < minspare_n) {
			complete = false;
			fw.c.closefile(fd);
			lineno = found_maxprocs;
			fw.newpanel(new edit(this,parent_panel,"EDIT " +
					capname));
			fw.msgline.setText("maxprocs on the highlighted " +
				"line must not less than minspare on line " +
				found_minspare);
			return(fw.c.NONFATAL);
		    }
		    maxspare = minspare;
		}
	    }
	    else {
		maxprocs_n = maxspare_n + 5;
		maxprocs = Integer.toString(maxspare_n + 5);
	    }
	}

	// Add an end marker so the user can insert after the last real rule
	rules_f.body.addItem("[end marker]");

	// Delete the temp file since we no longer need it
	fw.c.closefile(fd);
	tmpfile.delete();
	return (0);
    }

    public void handleOK() {

	if (saveconf() < 0)
	    return;

	pktrc = upload();
	if (pktrc.rc < 0)
	    return;

	fw.msgline.setText("New " + lowname + " config file " +
		"successfully created");
	fw.newpanel(parent_panel);
    }

    public int saveconf() {
	int fd;
	File outfile;
	String tmp;

	outfile = new File(tmpfn);

	pktrc = fw.c.opencbcfile(tmpfn, 0);
	if (pktrc.rc < 0) {
	    fw.msgline.setText(pktrc.errmsg);
	    return(pktrc.rc);
	}

	fd = pktrc.rc;

	// gproxpath statement
	pktrc = fw.c.cbcwriteLine(fd, "gproxpath = " +
					gproxpath_f.getText() + "\n");
	if (pktrc.rc != cglue.OK) {
	    fw.msgline.setText(pktrc.errmsg);
	    fw.c.closefile(fd);
	    return(-1);
	}

	// timeout statement
	pktrc = fw.c.cbcwriteLine(fd, "timeout = " +
					timeout_f.getText() + "\n");
	if (pktrc.rc != cglue.OK) {
	    fw.msgline.setText(pktrc.errmsg);
	    fw.c.closefile(fd);
	    return(-1);
	}

	// trace statement
	if (trace_f.getSelectedItem().toLowerCase().equals("yes"))
	    tmp = new String("yes");
	else
	    tmp = new String("no");
	pktrc = fw.c.cbcwriteLine(fd, "trace = " + tmp + "\n");
	if (pktrc.rc != cglue.OK) {
	    fw.msgline.setText(pktrc.errmsg);
	    fw.c.closefile(fd);
	    return(-1);
	}

	// nprocs statement
	if (nprocs_f.getText().length() > 0) {
	    String s1, s2, s3;
	    s1 = minspare_f.getText();
	    s2 = maxspare_f.getText();
	    if (s1.length() == 0 && s2.length() > 0) {
		fw.msgline.setText("maxspare value must be " +
				"accompanied by non-blank minspare value");
		fw.c.closefile(fd);
		    return(-1);
	    }
	    if (s1.length() > 0) {
		s3 = new String(" " + s1);
		if (s2.length() > 0)
		    s3 = s3.concat(" " + s2);
	    }
	    else
		s3 = new String();
	    pktrc = fw.c.cbcwriteLine(fd, "nprocs = " +
					nprocs_f.getText() + s3 + "\n");
	    if (pktrc.rc != cglue.OK) {
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		return(-1);
	    }
	}
	else {
	    if (minspare_f.getText().length() > 0 ||
			maxspare_f.getText().length() > 0) {
		fw.msgline.setText("minspare and maxspare values " +
			"must be accompanied by non-blank nprocs value");
		fw.c.closefile(fd);
		return(-1);
	    }
	}

	// maxprocs statement
	if (maxprocs_f.getText().length() > 0) {
	    pktrc = fw.c.cbcwriteLine(fd, "maxprocs = " +
					maxprocs_f.getText() + "\n");
	    if (pktrc.rc != cglue.OK) {
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		return(-1);
	    }
	}

	// checkprocs statement
	if (checkprocs_f.getText().length() > 0) {
	    pktrc = fw.c.cbcwriteLine(fd, "checkprocs = " +
					checkprocs_f.getText() + "\n");
	    if (pktrc.rc != cglue.OK) {
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		return(-1);
	    }
	}

	// maxuse statement
	if (maxuse_f.getText().length() > 0) {
	    pktrc = fw.c.cbcwriteLine(fd, "maxuse = " +
					maxuse_f.getText() + "\n");
	    if (pktrc.rc != cglue.OK) {
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		return(-1);
	    }
	}

	if (gen == WEB) {
	    // httplog statement
	    if (httplog_f.getText().length() > 0) {
		pktrc = fw.c.cbcwriteLine(fd, "httplog = " +
					httplog_f.getText() + "\n");
		if (pktrc.rc != cglue.OK) {
		    fw.msgline.setText(pktrc.errmsg);
		    fw.c.closefile(fd);
		    return(-1);
		}
	    }

	    // logformat statement
	    if (logformat_f.getText().length() > 0) {
		pktrc = fw.c.cbcwriteLine(fd, "logformat = " +
					logformat_f.getText() + "\n");
		if (pktrc.rc != cglue.OK) {
		    fw.msgline.setText(pktrc.errmsg);
		    fw.c.closefile(fd);
		    return(-1);
		}
	    }
	}

	//permit/deny rules
	// Note that we only process n - 1 lines, the last one is the
	// end marker
	for (int i = 0; i < rules_f.body.getItemCount() - 1; i++) {
	    pktrc = fw.c.cbcwriteLine(fd, rules_f.body.getItem(i) + "\n");
	    if (pktrc.rc != cglue.OK) {
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		return(-1);
	    }
	}
	fw.c.closefile(fd);

	return(0);
    }

    // Add processing code for the edit button
    protected boolean moreAction(String s) {
	if ("Edit".equals(s)) {
	    lineno = 0;		// Don't want edit to highlight any lines
	    saveconf();
	    fw.newpanel(new edit(this,parent_panel,"EDIT " + capname));
	    return(true);
	}
	return(false);
    }
}
