/*============================================================================
 * filename:		fwpulserules.java
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
//============================================================================
//
// R. James Livermore grants Freemont Avenue Software, Inc. a nonexclusive
// license to use, modify, and market this program using their tradename
// T.Rex.
//
//
// (C) COPYRIGHT R. James Livermore 1998-2000.  All rights reserved.
//
// (C) COPYRIGHT Freemont Avenue Software, Incorporated 1998-2000.
// All rights reserved.
//
// Date written:	June      24, 1998
// Last modified:	February   8, 2000
//
//===========================================================================*/
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

public class fwpulserules extends Rules {
    TextField	ipaddr_f;
    String	ipaddr;
    TextField	netmask_f;
    String	netmask;
    TextField	netif_f;
    String	netif;
    TextField	router_f;
    String	router;
    Choice	alias_f;
    boolean	alias;
    TextField	cport_f;
    String	cport;

    fwpulserules(fwwin p) {
	super(p);
	body.addActionListener(this);
    }

    public void addEditline() {
	GridLayout gl = new GridLayout(2,3);
	editline.setLayout(gl);

	editline.add(new Label("IP address"));
	editline.add(new Label("Network interface"));
	editline.add(new Label("Router address"));
	editline.add(new Label("Netmask"));
	editline.add(new Label("Critical Port"));
	editline.add(new Label("Alias"));

	ipaddr_f = new TextField();
	netif_f = new TextField();
	router_f = new TextField();
	netmask_f = new TextField();
	cport_f = new TextField();
	alias_f = new Choice();
	alias_f.addItem("Yes");
	alias_f.addItem("No");
	alias_f.select("No");	// Default to No

	editline.add(ipaddr_f);
	editline.add(netif_f);
	editline.add(router_f);
	editline.add(netmask_f);
	editline.add(cport_f);
	editline.add(alias_f);

	ipaddr_f.addFocusListener(this);
	router_f.addFocusListener(this);
	netif_f.addFocusListener(this);
	netmask_f.addFocusListener(this);
	cport_f.addFocusListener(this);
	alias_f.addFocusListener(this);

	add("North",editline);
    }

    /*======================================================================
     * Read all the input fields on the editline and insert a new permit/deny
     * line at the selected position in the body object
     * returns the new permit/deny line read from editline
     * returns null if failed
     *=====================================================================*/
    String readEditline() {
	String newrule;
	String tmp;
	String pd;

	newrule = new String("network ip_address=");

	tmp = ipaddr_f.getText();
	if (tmp.length() == 0) {
	    focus_prompt = new String("Must specify IP address");
	    ipaddr_f.requestFocus();
	    return(null);
	}
	newrule = newrule.concat(tmp);

	newrule = newrule.concat(" net_if=");
	tmp = netif_f.getText();
	if (tmp.length() == 0) {
	    focus_prompt = new String("Must specify network interface");
	    netif_f.requestFocus();
	    return(null);
	}
	newrule = newrule.concat(tmp);

	newrule = newrule.concat(" router=");
	tmp = router_f.getText();
	if (tmp.length() == 0) {
	    focus_prompt = new String("Must specify router address");
	    router_f.requestFocus();
	    return(null);
	}
	newrule = newrule.concat(tmp);

	tmp = netmask_f.getText();
	if (tmp.length() > 0) {
	    newrule = newrule.concat(" netmask=");
	    newrule = newrule.concat(tmp);
	}

	tmp = cport_f.getText();
	if (tmp.length() > 0) {
	    try {
		Integer x = Integer.valueOf(tmp);
	    }
	    catch (NumberFormatException e) {
		focus_prompt = new String("Critical Port must be a valid " +
					"number");
		cport_f.requestFocus();
		return(null);
	    }
	    newrule = newrule.concat(" critical_port=");
	    newrule = newrule.concat(tmp);
	}

	tmp = alias_f.getSelectedItem();
	if (tmp.equals("Yes")) {
	    newrule = newrule.concat(" alias");
	}

	return(newrule);
    }

    /*======================================================================
     * Parse a network line
     * returns 0 if successful (parsed values in the instance variables)
     * returns -1 if failed, caller should not use values in instance variables
     *=====================================================================*/
    public int readRule(String line) {
	String tmp1;
	StringTokenizer st = new StringTokenizer(line.trim()," \t=:");

	ipaddr = new String();
	ipaddr_f.setText(ipaddr);
	netmask = new String();
	netmask_f.setText(netmask);
	netif = new String();
	netif_f.setText(netif);
	cport = new String();
	cport_f.setText(cport);
	router = new String();
	router_f.setText(router);
	alias = false;

	// Get the statement type
	try {
	    tmp1 = st.nextToken().toLowerCase();
	} catch (NoSuchElementException e) {
	    fw.msgline.setText("Line is empty");
	    return(-1);
	}

	if (!tmp1.equalsIgnoreCase("network")) {
	    fw.msgline.setText("Statement type not \"network\"");
	    return(-1);
	}

	while (true) {
	    // Get the next token
	    try {
		tmp1 = st.nextToken().toLowerCase();
	    } catch (NoSuchElementException e) {
		break;
	    }

	    if (tmp1.equals("ip_address")) {
		try {
		    tmp1 = st.nextToken().toLowerCase();
		} catch (NoSuchElementException e) {
		    fw.msgline.setText("Missing value after " +
			"\"ip_address\"");
		    return(-1);
		}
		ipaddr = tmp1;
	    }

	    if (tmp1.equals("net_if")) {
		try {
		    tmp1 = st.nextToken().toLowerCase();
		} catch (NoSuchElementException e) {
		    fw.msgline.setText("Missing value after " +
			"\"net_if\"");
		    return(-1);
		}
		netif = tmp1;
	    }

	    if (tmp1.equals("router")) {
		try {
		    tmp1 = st.nextToken().toLowerCase();
		} catch (NoSuchElementException e) {
		    fw.msgline.setText("Missing value after " +
			"\"router\"");
		    return(-1);
		}
		router = tmp1;
	    }

	    if (tmp1.equals("netmask")) {
		try {
		    tmp1 = st.nextToken().toLowerCase();
		} catch (NoSuchElementException e) {
		    fw.msgline.setText("Missing value after " +
			"\"netmask\"");
		    return(-1);
		}
		netmask = tmp1;
	    }

	    if (tmp1.equals("critical_port")) {
		Integer x;
		try {
		    tmp1 = st.nextToken().toLowerCase();
		} catch (NoSuchElementException e) {
		    fw.msgline.setText("Missing value after " +
			"\"critical_port\"");
		    return(-1);
		}
		try {
		    x = Integer.valueOf(tmp1);
		} catch (NumberFormatException e) {
		    fw.msgline.setText("Value after \"critical_port\" " +
			"must be a valid number");
		    return(-1);
		}
		if (x.intValue() <= 0 || x.intValue() > 65535) {
		    fw.msgline.setText("Value after \"critical_port\" " +
			"must be between 0 and 65535");
		    return(-1);
		}
		cport = tmp1;
	    }

	    if (tmp1.equals("alias")) {
		alias = true;
	    }
	}

	if (ipaddr == null || ipaddr.length() == 0) {
	    fw.msgline.setText("Missing required \"ip_address\" value");
	    return(-1);
	}
	ipaddr_f.setText(ipaddr);

	if (netif == null || netif.length() == 0) {
	    fw.msgline.setText("Missing required \"net_if\" value");
	    return(-1);
	}
	netif_f.setText(netif);

	/*===============================================================
	 * If not alias, must specify router address
	 *==============================================================*/
	if ( !alias && (router == null || router.length() == 0)) {
	    fw.msgline.setText("Missing required \"router\" value");
	    return(-1);
	}
	router_f.setText(router);

	netmask_f.setText(netmask);
	cport_f.setText(cport);
	if (alias)
	    alias_f.select("Yes");
	else
	    alias_f.select("No");

	return (0);
    }
}
