/*============================================================================
 * filename:		listarchives.java
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
//============================================================================
//
// R. James Livermore grants Freemont Avenue Software, Inc. a nonexclusive
// license to use, modify, and market this program using their tradename
// T.Rex.
//
//
// (C) COPYRIGHT R. James Livermore 1998-2000.  All rights reserved.
//
// (C) COPYRIGHT Freemont Avenue Software, Incorporated 1998-2000.
// All rights reserved.
//
// Date written:	June      24, 1998
// Last modified:	February   8, 2000
//
//===========================================================================*/
import java.io.*;
import java.awt.*;
import java.awt.event.*;

public class listarchives extends Base {
    TextField filename_f;
    Dir dir;
    int i;
    int ver;
    int flag = 0;
    Button[] listb;
    Button[] deleteb;
    Button[] renameb;

    listarchives(Base p) {
	super(p,"T.Rex Archives List","Click on the desired action");
	addBody();
	okbutton.setLabel("Continue");
	canbutton.setLabel("Return");
    }

    void addBody() {
	String fn;
	String tfn;

	GridBagLayout gbl = new GridBagLayout();
	GridBagConstraints gbc = new GridBagConstraints();
	body.setLayout(gbl);

	gbc.anchor = GridBagConstraints.WEST;
	gbc.insets = new Insets(10,10,10,10);

	// Open the archive directory
	try {
	    dir = new Dir(fw.c.getvar("HOME").str +
		System.getProperty("file.separator") + "fw_archive");
	}
	    catch (IllegalArgumentException e) {
	    fw.msgline.setText("Archive directory error: " +
							e.getMessage());
	}

	listb = new Button[dir.count];
	deleteb = new Button[dir.count];
	renameb = new Button[dir.count];

	gbc.gridwidth = 1;
	Label t1 = new Label("Archive Name");
	gbl.setConstraints(t1,gbc);
	body.add(t1);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	Label t2 = new Label("Last Written");
	gbl.setConstraints(t2,gbc);
	body.add(t2);

	// Display each archive file in the Panel
	for (i = 0; i < dir.count; i++) {

	    gbc.gridwidth = 1;
	    Label l1 = new Label(dir.filenames[i]);
	    gbl.setConstraints(l1,gbc);
	    body.add(l1);

	    gbc.gridwidth = 1;
	    Label l2 = new Label(dir.moddates[i]);
	    gbl.setConstraints(l2,gbc);
	    body.add(l2);

	    gbc.gridwidth = 1;
	    listb[i] = new Button("List Contents");
	    listb[i].addActionListener(this);
	    listb[i].setActionCommand("list" + i);
	    gbl.setConstraints(listb[i],gbc);
	    body.add(listb[i]);

	    gbc.gridwidth = GridBagConstraints.RELATIVE;
	    deleteb[i] = new Button("Delete");
	    deleteb[i].addActionListener(this);
	    deleteb[i].setActionCommand("delete" + i);
	    gbl.setConstraints(deleteb[i],gbc);
	    body.add(deleteb[i]);

	    gbc.gridwidth = GridBagConstraints.REMAINDER;
	    renameb[i] = new Button("Rename");
	    renameb[i].addActionListener(this);
	    renameb[i].setActionCommand("rename" + i);
	    gbl.setConstraints(renameb[i],gbc);
	    body.add(renameb[i]);
	}

	sp.add(body);
    }

    /*======================================================================
     * Override since we don't want an OK button here
     *=====================================================================*/
    void addOK() {
    }

    // Add processing code for the edit/browse buttons
    protected boolean moreAction(String s) {
	int i;

	for (i = 0; i < dir.count; i++) {
	    if (s.equals("list" + i)) {
		fw.newpanel(new listtar(this,dir,dir.filenames[i]));
		return(true);
	    }
	    if (s.equals("delete" + i)) {
		Button cont = new Button("OK");
		Button can = new Button("Cancel");

		Panel p = new Panel();
		p.setLayout(new GridLayout(1,5));
		p.add(new Label(""));
		cont.setActionCommand("FILECONT");
		cont.addActionListener(this);
		p.add(cont);
		p.add(new Label(""));
		can.setActionCommand("FILECAN");
		can.addActionListener(this);
		p.add(can);
		p.add(new Label(""));
		
		dldialog = new MyDialog(fw.getFrame(),
				"Delete Archive",
				true,
				new Label("Delete this archive file?"),
				new Label(dir.filenames[i]),
				p);;
		dldialog.setVisible(true);

		if (dldisp.equals("FILECONT")) {
		    File df = new File(dir.toString() +
				System.getProperty("file.separator") +
				dir.filenames[i]);
		    df.delete();

		    // Refresh this panel so the deleted file no longer shows
		    fw.newpanel(new listarchives(parent_panel));
		    fw.msgline.setText("Archive file " + dir.filenames[i] +
				" deleted");
		}
		else
		    fw.msgline.setText("Archive file " + dir.filenames[i] +
				" not deleted");
		return(true);
	    }
	    if (s.equals("rename" + i)) {
		TextField newname_f = new TextField();
		Button cont = new Button("OK");
		Button can = new Button("Cancel");

		Panel p = new Panel();
		p.setLayout(new GridLayout(1,5));
		p.add(new Label(""));
		cont.setActionCommand("FILECONT");
		cont.addActionListener(this);
		p.add(cont);
		p.add(new Label(""));
		can.setActionCommand("FILECAN");
		can.addActionListener(this);
		p.add(can);
		p.add(new Label(""));
		
		dldialog = new MyDialog(fw.getFrame(),
				"Delete Archive",
				true,
				new Label("Enter new name for archive file " +
					dir.filenames[i]),
				newname_f,
				p);
		dldialog.setVisible(true);

		if (dldisp.equals("FILECONT")) {
		    String newname = newname_f.getText();
		    if (newname.length() == 0) {
			fw.msgline.setText("Must specify new name for " +
				"rename");
			return(true);
		    }
		    File df = new File(dir.toString() +
				System.getProperty("file.separator") +
				dir.filenames[i]);

		    // If a "/" ("\" on Windows) exists in filename, use it
		    // asis; otherwise, prepend the archive directory
		    if (newname.indexOf(System.getProperty("file.separator"))
					< 0)
			newname = new String(dir.toString() +
					System.getProperty("file.separator") +
					newname);
		    df.renameTo(new File(newname));

		    // Refresh this panel so the deleted file no longer shows
		    fw.newpanel(new listarchives(parent_panel));
		    fw.msgline.setText("Archive file " + dir.filenames[i] +
				" renamed to " + newname_f.getText());
		}
		else
		    fw.msgline.setText("Archive file " + dir.filenames[i] +
				" not renamed");
		return(true);
	    }
	}

	return(false);
    }
}
