/*============================================================================
 * filename:		monitor.java
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
//============================================================================
//
// R. James Livermore grants Freemont Avenue Software, Inc. a nonexclusive
// license to use, modify, and market this program using their tradename
// T.Rex.
//
//
// (C) COPYRIGHT R. James Livermore 1998-2000.  All rights reserved.
//
// (C) COPYRIGHT Freemont Avenue Software, Incorporated 1998-2000.
// All rights reserved.
//
// Date written:	June      24, 1998
// Last modified:	February   8, 2000
//
//===========================================================================*/
import java.awt.*;
import java.awt.event.*;

public class monitor extends Base {
    int nlines;
    volatile Threadme activeThread;

    monitor(Base b, int component, String title, int nlines) {
	super(b, "","");
	paneltitle.setText(title);
	this.component = component;
	this.nlines = nlines;
	addBody();
	okbutton.setLabel("Done");
	canbutton.setLabel("Pause");
    }

    void addBody() {

	int i;

	body.setLayout(new BorderLayout());
	sp.add("Center",body);
	activeThread = new Threadme();
    }

    class Threadme extends Thread {
	int cnt = 0;
	int i;
	int j;
	pktinfo pktrc;

	public void run() {
	    body.setLayout(new BorderLayout());
	    body.setFont(new Font("Courier",Font.PLAIN,fwmain.fontsize));

	    Thread thisThread = Thread.currentThread();

	    // The text area is nlines + 1 because the last line is a blank
	    // line
	    TextArea t = new TextArea(nlines+1,0);
	    t.setEditable(false);
	    t.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
	    String s = new String();
	    body.add(t);

	    while (!interrupted() && (activeThread == thisThread)) {
		// Read something from the socket
		pktrc = fw.c.recvpkt(2);
		if (pktrc.rc == fw.c.CONTINUE) {
		    yield();
		    continue;
		}

		// Server responds with a CLOSED rc when it receives a
		// CANCEL request.
		if (pktrc.rc == fw.c.CLOSED) {
		    fw.c.flush();
		    fw.newpanel(parent_panel);
		    activeThread = null;
		    break;
		}
		if (pktrc.rc < 0) {
		    // Only treat it as an error if it was not interrupted
		    // and errno not EINTR
		    if (!interrupted() && fw.c.Errno != fw.c.Eintr) {
			fw.c.flush();
			fw.newpanel(parent_panel);
			fw.msgline.setText(pktrc.errmsg);
			activeThread = null;
		    }
		    continue;
		}

		// Once the max lines are filled, remove some text from the
		// beginning each time so we don't take up all memory.
		// Except when nlines = 0, in which case, the caller wants
		// to keep all lines in the buffer
		if (nlines > 0)
		    if (cnt == nlines)
			s = s.substring(s.indexOf('\n') + 1);
		    else
			cnt++;

		// Add the new line to the TextArea
		s = s.concat(pktrc.str + "\n");
		t.setText(s);

		body.validate();

		// Set the cursor with a sufficiently large number to make
		// sure it is positioned at the end.
		t.setCaretPosition(s.length());
	    }
	}
    }

    public void handleCan() {
	pktinfo pktrc;
	// Ignore this button if activeThread is null.  This is to protect
	// ourselves from getting a NullPointerException when the user
	// presses "Done" and then press this button.
	if (activeThread == null)
	    return;

	if (canbutton.getLabel().equals("Pause")) {
	    pktrc = fw.c.sendpkt(component, fw.c.PAUSE, "",1);
	    if (pktrc.rc < 0)
		return;
	    canbutton.setLabel("Resume");
	}
	else {
	    pktrc = fw.c.sendpkt(component, fw.c.RESUME, "",1);
	    if (pktrc.rc < 0)
		return;
	    canbutton.setLabel("Pause");
	}
    }

    public void handleOK() {
	// The Done button is disabled while the screen is paused
	if (canbutton.getLabel().equals("Resume"))
	    return;
	// If activeThread is null, don't send another request.  The user
	// must have pressed "Done" more than once and so all subsequent
	// after the first can be ignored.
	if (activeThread == null)
	    return;

	activeThread = null;
	fw.c.sendpkt(component,fw.c.CANCEL,"",1);
    }
}
