/*============================================================================
 * filename:		padminrules1.java
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
//============================================================================
//
// R. James Livermore grants Freemont Avenue Software, Inc. a nonexclusive
// license to use, modify, and market this program using their tradename
// T.Rex.
//
//
// (C) COPYRIGHT R. James Livermore 1998-2000.  All rights reserved.
//
// (C) COPYRIGHT Freemont Avenue Software, Incorporated 1998-2000.
// All rights reserved.
//
// Date written:	June      24, 1998
// Last modified:	February   8, 2000
//
//===========================================================================*/
import java.util.*;
import java.awt.*;
import java.io.*;

public class padminrules1 extends Rules {
    int		pd;		// Holds index of the selected List object
    String	fromaddr;
    String	frommask;

    Choice	pd_f;
    TextField	fromaddr_f;
    TextField	frommask_f;

    padminrules1(fwwin p) {
	super(p);
	body.addActionListener(this);
    }

    public void addEditline() {
//	GridLayout gl = new GridLayout(10,3);
//	editline.setLayout(gl);

	GridBagLayout gbl = new GridBagLayout();
	GridBagConstraints gbc = new GridBagConstraints();
	editline.setLayout(gbl);
	gbc.anchor = GridBagConstraints.WEST;

	pd_f = new Choice();
	fromaddr_f = new TextField();
	frommask_f = new TextField();

	pd_f.addItem("permit");
	pd_f.addItem("deny");
	pd_f.select(1);

	gbc.fill = GridBagConstraints.HORIZONTAL;
//	gbc.gridwidth = GridBagConstraints.RELATIVE;
	gbc.gridwidth = 1;
	Label pd_l = new Label("Permit/Deny");
	gbl.setConstraints(pd_l,gbc);
	editline.add(pd_l);

	gbc.gridwidth = GridBagConstraints.RELATIVE;
	Label fromaddr_l = new Label("From Address");
	gbl.setConstraints(fromaddr_l,gbc);
	editline.add(fromaddr_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	Label frommask_l = new Label("From Address Mask");
	gbl.setConstraints(frommask_l,gbc);
	editline.add(frommask_l);

//	gbc.gridwidth = GridBagConstraints.RELATIVE;
	gbc.gridwidth = 1;
	gbl.setConstraints(pd_f,gbc);
	editline.add(pd_f);

	gbc.gridwidth = GridBagConstraints.RELATIVE;
	gbl.setConstraints(fromaddr_f,gbc);
	editline.add(fromaddr_f);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	gbl.setConstraints(frommask_f,gbc);
	editline.add(frommask_f);

	pd_f.addFocusListener(this);
	fromaddr_f.addFocusListener(this);
	frommask_f.addFocusListener(this);

	add("North",editline);
    }

    /*======================================================================
     * Read all the input fields on the editline and insert a new permit/deny
     * line at the selected position in the body object
     * returns the new permit/deny line read from editline
     * returns null if failed
     *=====================================================================*/
    String readEditline() {
	String newrule;
	String tmp;
	int tmpint;

	newrule = pd_f.getSelectedItem();

	// From IP
	tmp = fromaddr_f.getText();
	if (tmp.length() > 0) {
	    newrule = newrule.concat(" ");
	    newrule = newrule.concat(tmp);
	}
	else
	    return(newrule);

	// From Mask
	tmp = frommask_f.getText();
	if (tmp.length() > 0) {
	    newrule = newrule.concat(" ");
	    newrule = newrule.concat(tmp);
	}

	return(newrule);
    }

    /*======================================================================
     * Parse a permit line
     * returns 0 if successful (parsed values in the instance variables)
     * returns -1 if failed, caller should not use values in instance variables
     *=====================================================================*/
    public int readRule(String line) {
	int i;
	String tmp1,tmp2;
	StringTokenizer st = new StringTokenizer(line.trim()," \t");

	fromaddr = new String();
	fromaddr_f.setText(fromaddr);
	frommask = new String();
	frommask_f.setText(frommask);

	// Get the statement type
	try {
	    tmp1 = st.nextToken().toLowerCase();
	} catch (NoSuchElementException e) {
	    fw.msgline.setText("Encountered an empty line");
	    return(-1);
	}

	// pd is set to the index of the selected List object
	if (tmp1.equals("permit"))
	    pd = 0;
	else if (tmp1.equals("deny"))
	    pd = 1;
	else {
	    fw.msgline.setText("Invalid statement type " + tmp1);
	    return(-1);
	}

	pd_f.select(pd);

	// Get the next token which is fromaddr
	try {
	    fromaddr = st.nextToken();
	} catch (NoSuchElementException e) {
	    return(0);
	}
	fromaddr_f.setText(fromaddr);

	// Get the next token which is frommask
	try {
	    frommask = st.nextToken();
	} catch (NoSuchElementException e) {
	    return(0);
	}
	frommask_f.setText(frommask);

	return (0);
    }
}
