/*============================================================================
 * filename:		sockd.java
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
//============================================================================
//
// R. James Livermore grants Freemont Avenue Software, Inc. a nonexclusive
// license to use, modify, and market this program using their tradename
// T.Rex.
//
//
// (C) COPYRIGHT R. James Livermore 1998-2000.  All rights reserved.
//
// (C) COPYRIGHT Freemont Avenue Software, Incorporated 1998-2000.
// All rights reserved.
//
// Date written:	June      24, 1998
// Last modified:	February   8, 2000
//
//===========================================================================*/
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

public class sockd extends Base {
    final int MAXIFENT = 100;
    TextField logoutput_f;
    Checkbox syslog_cb;
    Checkbox stderr_cb;
    Checkbox stdout_cb;
    Choice external_c;
    TextField method_f;
    TextField priv_f;
    TextField notpriv_f;
    TextField conntimeout_f;
    TextField iotimeout_f;
    Choice srchost_c;
    sockdrules clntrules_f;
    sockdrules svrrules_f;
    String logoutput;
    boolean syslog = false;
    boolean stderr = false;
    boolean stdout = false;
    String external;
    String method;
    String priv;
    String notpriv;
    String conntimeout;
    String iotimeout;
    pktinfo pktrc;
    String capname;
    int nifent = 0;
    Checkbox intip_cb[] = new Checkbox[MAXIFENT];
    TextField intport_f[] = new TextField[MAXIFENT];
    String iflist[] = new String[MAXIFENT];

    sockd(Base p) {
	super(p,"SOCKD","");
	addBody();
    }

    void addBody() {
	int i;
	int rc;
	component = fw.c.SOCKD;
	capname = new String("SOCKD");
	lowname = new String("sockd");

	// Initialize the values for this panel
	logoutput = new String();
	external = new String();
	method = new String("none");
	priv = new String("nobody");
	notpriv = new String("nobody");
	conntimeout = new String("30");
	iotimeout = new String("0");

	external_c = new Choice();

	GridBagLayout gbl = new GridBagLayout();
	GridBagConstraints gbc = new GridBagConstraints();
	body.setLayout(gbl);

	// Get the config file over here to a temp file
	pktrc = download();
	tmpfn = pktrc.str;

	if (pktrc.rc != fw.c.OK) {
	    complete = false;
	    return;
	}

	// Get the network interface addresses from the firewall
	pktrc = fw.c.sendpkt(cglue.APACHE, cglue.GETIFLIST, "", 1);
	if (pktrc.rc < 0) {
	    complete = false;
	    fw.msgline.setText(pktrc.errmsg);
	    if (pktrc.rc != cglue.NONFATAL) {
		fw.c.closefd();
		fw.newpanel(fw.getTop());
	    }
	    return;
	}

	// Get the response
	pktrc = fw.c.recvpkt(1);
	if (pktrc.rc < 0) {
	    complete = false;
	    fw.msgline.setText(pktrc.errmsg);
	    if (pktrc.rc != cglue.NONFATAL) {
		fw.c.closefd();
		fw.newpanel(fw.getTop());
	    }
	    return;
	}

	// Loop through each entry and set the String array
	String ifs = pktrc.str;
	while (ifs.length() > 0 && nifent < MAXIFENT) {
	    String ip;
	    int end = ifs.indexOf(' ');
	    if (end < 0)
		ip = ifs;
	    else
		ip = ifs.substring(0, end);

	    // Sanity check to make sure ip is not too long
	    if (ip.length() > 15) {
		fw.msgline.setText("Interanl error: ip address too long.");
		fw.c.closefd();
		fw.newpanel(fw.getTop());
		return;
	    }
	    iflist[nifent] = ip;
	    intip_cb[nifent] = new Checkbox(ip);
	    intport_f[nifent] = new TextField("1080");
	    external_c.addItem(ip);
	    nifent++;
	    if (end < 0)
		break;
	    ifs = ifs.substring(end + 1);
	}

	// Construct the rules boxes now so they can client and server
	// permit/deny lines while we read the config file
	clntrules_f = new sockdrules(fw, sockdrules.CLIENT);
	svrrules_f = new sockdrules(fw, sockdrules.SERVER);

	srchost_c = new Choice();
	srchost_c.addItem("ALL");
	srchost_c.addItem("nounknown");

	// Read in the temp file
	if ((rc = readconf()) == fw.c.FAILED) {
	    fw.c.closefd();
	    fw.newpanel(fw.getTop());
	}

	if (rc < 0)
	    return;

	gbc.anchor = GridBagConstraints.WEST;
	gbc.gridwidth = GridBagConstraints.REMAINDER;
	Button edit_b = new Button("Edit");
	edit_b.addActionListener(this);
	gbl.setConstraints(edit_b,gbc);
	body.add(edit_b);

	gbc.gridwidth = 1;
	Label logoutput_l = new Label("Log output to");
	gbl.setConstraints(logoutput_l,gbc);
	body.add(logoutput_l);

	Panel logoutput_p = new Panel();
	logoutput_p.setLayout(gbl);

	gbc.gridwidth = 1;
	Label b1_l = new Label("");
	gbl.setConstraints(b1_l,gbc);
	logoutput_p.add(b1_l);

	gbc.gridwidth = 1;
	Label b2_l = new Label("");
	gbl.setConstraints(b2_l,gbc);
	logoutput_p.add(b2_l);

	gbc.gridwidth = 1;
	Label b3_l = new Label("");
	gbl.setConstraints(b3_l,gbc);
	logoutput_p.add(b3_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	Label fn_l = new Label("Optional Filename");
	gbl.setConstraints(fn_l,gbc);
	fn_l.setFont(new Font("TimesRoman", Font.PLAIN, fwmain.fontsize-2));
	logoutput_p.add(fn_l);

	gbc.gridwidth = 1;
	syslog_cb = new Checkbox("syslog");
	syslog_cb.setState(syslog);
	gbl.setConstraints(syslog_cb,gbc);
	logoutput_p.add(syslog_cb);
	syslog_cb.addFocusListener(this);

	gbc.gridwidth = 1;
	stderr_cb = new Checkbox("stderr");
	stderr_cb.setState(stderr);
	gbl.setConstraints(stderr_cb,gbc);
	logoutput_p.add(stderr_cb);
	stderr_cb.addFocusListener(this);

	gbc.gridwidth = 1;
	stdout_cb = new Checkbox("stdout");
	stdout_cb.setState(stdout);
	gbl.setConstraints(stdout_cb,gbc);
	logoutput_p.add(stdout_cb);
	stdout_cb.addFocusListener(this);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	logoutput_f = new TextField(logoutput,20);
	gbl.setConstraints(logoutput_f,gbc);
	logoutput_p.add(logoutput_f);
	logoutput_f.addFocusListener(this);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	gbl.setConstraints(logoutput_p,gbc);
	body.add(logoutput_p);

	gbc.gridwidth = 1;
	Label external_l = new Label("External IP address");
	gbl.setConstraints(external_l,gbc);
	body.add(external_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	gbl.setConstraints(external_c,gbc);
	body.add(external_c);
	external_c.addFocusListener(this);

	// Setup the internal IP addresses as Checkboxes.
	gbc.gridwidth = 1;
	Label internal1_l = new Label("Listen to Internal");
	gbl.setConstraints(internal1_l,gbc);
	body.add(internal1_l);

	gbc.gridwidth = 1;
	Label internal2_l = new Label("IP Addresses");
	gbl.setConstraints(internal2_l,gbc);
	body.add(internal2_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	Label internal3_l = new Label("Ports");
	gbl.setConstraints(internal3_l,gbc);
	body.add(internal3_l);

	for (i = 0; i < nifent; i++) {
	    gbc.gridwidth = 1;
	    Label tmp_l = new Label("");
	    gbl.setConstraints(tmp_l,gbc);
	    body.add(tmp_l);

	    gbc.gridwidth = 1;
	    gbl.setConstraints(intip_cb[i],gbc);
	    body.add(intip_cb[i]);
	    intip_cb[i].addFocusListener(this);

	    gbc.gridwidth = GridBagConstraints.REMAINDER;
	    gbl.setConstraints(intport_f[i],gbc);
	    body.add(intport_f[i]);
	    intport_f[i].addFocusListener(this);
	}

	gbc.gridwidth = 1;
	Label priv_l = new Label("Privileged username");
	gbl.setConstraints(priv_l,gbc);
	body.add(priv_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	priv_f = new TextField(priv, 12);
	gbl.setConstraints(priv_f,gbc);
	body.add(priv_f);
	priv_f.addFocusListener(this);

	gbc.gridwidth = 1;
	Label notpriv_l = new Label("Non-Privileged username");
	gbl.setConstraints(notpriv_l,gbc);
	body.add(notpriv_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	notpriv_f = new TextField(notpriv, 12);
	gbl.setConstraints(notpriv_f,gbc);
	body.add(notpriv_f);
	notpriv_f.addFocusListener(this);

	gbc.gridwidth = 1;
	Label conntimeout_l = new Label("Connection request timeout");
	gbl.setConstraints(conntimeout_l,gbc);
	body.add(conntimeout_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	conntimeout_f = new TextField(conntimeout, 5);
	gbl.setConstraints(conntimeout_f,gbc);
	body.add(conntimeout_f);
	conntimeout_f.addFocusListener(this);

	gbc.gridwidth = 1;
	Label iotimeout_l = new Label("Idle timeout");
	gbl.setConstraints(iotimeout_l,gbc);
	body.add(iotimeout_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	iotimeout_f = new TextField(iotimeout, 5);
	gbl.setConstraints(iotimeout_f,gbc);
	body.add(iotimeout_f);
	iotimeout_f.addFocusListener(this);

	gbc.gridwidth = 1;
	Label srchost_l = new Label("Accepted source hosts");
	gbl.setConstraints(srchost_l,gbc);
	body.add(srchost_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	gbl.setConstraints(srchost_c,gbc);
	body.add(srchost_c);
	srchost_c.addFocusListener(this);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	Label blank1_l  = new Label("");
	gbl.setConstraints(blank1_l,gbc);
	body.add(blank1_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	Label clntrules_l = new Label("Client Rules");
	gbl.setConstraints(clntrules_l,gbc);
	body.add(clntrules_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	gbl.setConstraints(clntrules_f,gbc);
	body.add(clntrules_f);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	Label blank2_l  = new Label("");
	gbl.setConstraints(blank2_l,gbc);
	body.add(blank2_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	Label svrrules_l = new Label("Socks Rules");
	gbl.setConstraints(svrrules_l,gbc);
	body.add(svrrules_l);

	clntrules_f.addFocusListener(this);
	gbl.setConstraints(svrrules_f,gbc);
	body.add(svrrules_f);
	svrrules_f.addFocusListener(this);

	sp.add(body);
    }

    /*=======================================================================
     * Reads the temp file in and interpret it
     *
     * Returns 0 if successful
     *		-1 if failed
     *		NONFATAL if connection need not be closed
     *
     *======================================================================*/
    int readconf() {
	int i;
	int fd;
	if (tmpfn.equals("")) {
	    fw.msgline.setText("Internal error: " + lowname + " config " +
			"file has not been downloaded yet");
	    return(fw.c.NONFATAL);
	}
	File tmpfile = new File(tmpfn);
	if (!tmpfile.exists() || !tmpfile.isFile()) {
	    fw.msgline.setText("Internal error: the downloaded " +
			lowname + " conf file cannot be found");
	    return(fw.c.NONFATAL);
	}

	pktrc = fw.c.opencbcfile(tmpfn, 0);
	if (pktrc.rc < 0) {
	    fw.msgline.setText(pktrc.errmsg);
	    return(pktrc.rc);
	}

	fd = pktrc.rc;

	while ( true ) {
	    String tmp;
	    String statement;

	    pktrc = fw.c.cbcreadLine(fd);
	    if (pktrc.rc != cglue.OK) {
		if (pktrc.rc == cglue.CLOSED)
		    break;
		complete = false;
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		return(pktrc.rc);
	    }

	    lineno++;

	    // Get rid of empty lines
	    if (pktrc.str.trim().length() == 0)
		continue;

	    StringTokenizer st = new StringTokenizer(pktrc.str.trim()," \t=:");

	    // Get the statement type
	    try {
		statement = st.nextToken();
	    } catch (NoSuchElementException e) {
		continue;
	    }

	    // Get rid of comment lines (first character is #)
	    if (pktrc.str.trim().substring(0,1).equals("#"))
		continue;

	    // Get the next token into tmp - ignore error with no token found
	    // We do this here so we don't have to do a "try" for every
	    // statement type
	    try {
		tmp = st.nextToken();
	    } catch (NoSuchElementException e) {
		complete = false;
		fw.c.closefile(fd);
		tmpfile.delete();
		fw.newpanel(new edit(this,parent_panel,"EDIT " + capname));
		fw.msgline.setText("Highlighted line contains an " +
			"invalid statement, please correct and retry");
		return(fw.c.NONFATAL);
	    }

	    if (statement.equals("logoutput")) {
		while (tmp != null) {
		    if (tmp.equalsIgnoreCase("syslog"))
			syslog = true;
		    else if (tmp.equalsIgnoreCase("stderr"))
			stderr = true;
		    else if (tmp.equalsIgnoreCase("stdout"))
			stdout = true;
		    else
			logoutput = tmp;
		    try {
			tmp = st.nextToken();
		    } catch (NoSuchElementException e) {
			break;
		    }
		}
		continue;
	    }

	    if (statement.equals("internal")) {
		if (tmp == null) {
		    fw.c.closefile(fd);
		    fw.newpanel(new edit(this,parent_panel,"EDIT " + capname));
		    fw.msgline.setText("Highlighted line is missing " +
				"arguments, please correct and retry");
		    return(fw.c.NONFATAL);
		}
		for (i = 0; i < nifent; i++) {
		    if (tmp.equals(iflist[i])) {
			intip_cb[i].setState(true);
			break;
		    }
		}
		// If the IP on an internal: line does not match one in the
		// iflist, ignore it.
		if (i >= nifent)
		    continue;
		// If no port present, let it default
		try {
		    tmp = st.nextToken();
		} catch (NoSuchElementException e) {
		    continue;
		}
		if (!tmp.equalsIgnoreCase("port")) {
		    fw.c.closefile(fd);
		    fw.newpanel(new edit(this,parent_panel,"EDIT " + capname));
		    fw.msgline.setText("Highlighted line contains an invalid " +
				"keyword, please correct and retry");
		    return(fw.c.NONFATAL);
		}
		try {
		    tmp = st.nextToken();
		} catch (NoSuchElementException e) {
		    fw.c.closefile(fd);
		    fw.newpanel(new edit(this,parent_panel,"EDIT " + capname));
		    fw.msgline.setText("Highlighted line is missing " +
				"arguments, please correct and retry");
		    return(fw.c.NONFATAL);
		}
		intport_f[i] = new TextField(tmp);
		continue;
	    }

	    // If the external: IP address is not on the list, it is
	    // ignored.
	    if (statement.equals("external")) {
		for (i = 0; i < nifent; i++) {
		    if (iflist[i].equals(tmp)) {
			external_c.select(i);
			break;
		    }
		}
		continue;
	    }

	    if (statement.equals("user.privileged")) {
		priv = tmp;
		continue;
	    }

	    if (statement.equals("user.notprivileged")) {
		notpriv = tmp;
		continue;
	    }

	    if (statement.equals("connecttimeout")) {
		conntimeout = tmp;
		continue;
	    }

	    if (statement.equals("iotimeout")) {
		iotimeout = tmp;
		continue;
	    }

	    // Ignore method statements
	    if (statement.equals("method")) {
		continue;
	    }

	    if (statement.equals("srchost")) {
		if (tmp.equalsIgnoreCase("nounknown"))
		    srchost_c.select(1);
		else {
		    fw.c.closefile(fd);
		    fw.newpanel(new edit(this,parent_panel,"EDIT " +
				capname));
		    fw.msgline.setText("Highlighted line contains " +
				"invalid keyword, please correct and retry");
		    return(fw.c.NONFATAL);
		}
		continue;
	    }

	    if (statement.equals("client") ||
		statement.equals("pass") || statement.equals("block")) {
		String merge = new String();
		while (true) {
		    if (!pktrc.str.trim().substring(0,1).equals("#")) {
			merge = merge.concat(pktrc.str);
			// '}' is assumed to be the end of the client statement
			if (pktrc.str.indexOf('}') != -1)
			    break;
		    }
		    pktrc = fw.c.cbcreadLine(fd);
		    lineno++;
		    if (pktrc.rc != cglue.OK) {
			if (pktrc.rc == cglue.CLOSED) {
			    fw.c.closefile(fd);
			    fw.newpanel(new edit(this,parent_panel,"EDIT " +
				capname));
			    fw.msgline.setText("Highlighted line " +
				"is incomplete," + pktrc.errmsg +
				" please correct and retry");
			    return(fw.c.NONFATAL);
			}
			complete = false;
			fw.msgline.setText(pktrc.errmsg);
			fw.c.closefile(fd);
			return(pktrc.rc);
		    }
		}
		if (statement.equals("client"))
		    clntrules_f.body.addItem(merge);
		else
		    svrrules_f.body.addItem(merge);
		continue;
	    }
	}

	// Add an end marker so the user can insert after the last real rule
	clntrules_f.body.addItem("[end marker]");
	svrrules_f.body.addItem("[end marker]");

	// Delete the temp file since we no longer need it
	fw.c.closefile(fd);
	tmpfile.delete();
	return (0);
    }

    public void handleOK() {

	if (saveconf() < 0)
	    return;

	pktrc = upload();
	if (pktrc.rc < 0)
	    return;

	fw.msgline.setText("New " + lowname + " config file " +
		"successfully created");
	fw.newpanel(parent_panel);
    }

    /*=======================================================================
     * Given a string, this method splits it into lines (separated by '\n')
     * and sends each line individually to cbcwriteLine().
     * If comp is a String, this function makes sure each line starts with
     * that string.  If comp is null, no comparison is made.
     * Returns	0 if succcessful
     *		-1 if failed
     *		-3 if A line does not begin with the String comp
     *======================================================================*/
    private int split_line(int fd, String s, String comp) {
	String line;
	StringTokenizer st = new StringTokenizer(s,"\n");

	while (true) {
	    // Get one statement
	    try {
		line = st.nextToken();
	    } catch (NoSuchElementException e) {
		break;
	    }
	    if (comp != null) {
		if (!line.trim().toLowerCase().startsWith(comp.toLowerCase())) {
		    return(fw.c.NONFATAL);
		}
	    }
	    pktrc = fw.c.cbcwriteLine(fd, line + "\n");
	    if (pktrc.rc != cglue.OK) {
		fw.msgline.setText(pktrc.errmsg);
		return(-1);
	    }
	}
	return(0);
    }

    public int saveconf() {
	int i;
	int fd;
	File infile;
	File outfile;
	String unknown_statements = new String();
	String tmp;
	String tmpoutfn;
	StringTokenizer st;
	String statement;
	int end_of_dirname;

	// Re-get the config file over here to a temp file
	pktrc = fw.c.getconf(component);
	tmpfn = pktrc.str;

	// Form the temp output filename
	end_of_dirname = tmpfn.lastIndexOf('/');
	if (end_of_dirname < 0)
	    end_of_dirname = tmpfn.lastIndexOf('\\');
	if (end_of_dirname < 0)
	    tmpoutfn = new String("_padmsoc.pd");
	else {
	    tmpoutfn = tmpfn.substring(0,end_of_dirname+1);
	    tmpoutfn = tmpoutfn.concat("_padmsoc.pd");
	}
	outfile = new File(tmpoutfn);

	infile = new File(tmpfn);

	if (pktrc.rc == fw.c.OK) {
	    if (!infile.exists() || !infile.isFile()) {
		fw.msgline.setText("Temp sockd.conf file cannot be found");
		return(-1);
	    }

	    pktrc = fw.c.opencbcfile(tmpfn, 0);
	    if (pktrc.rc < 0) {
		fw.msgline.setText(pktrc.errmsg);
		return(pktrc.rc);
	    }

	    fd = pktrc.rc;

	    // Process each line in the input file, get rid of all comment lines
	    // and all the statements that are 'managed' by the gui.  I.e.
	    // write out all the statements not known to the gui.
	    while ( true ) {
		pktrc = fw.c.cbcreadLine(fd);
		if (pktrc.rc != cglue.OK) {
		    if (pktrc.rc == cglue.CLOSED)
			break;
		    fw.msgline.setText(pktrc.errmsg);
		    fw.c.closefile(fd);
		    return(pktrc.rc);
		}

		// Exit loop at EOF
		if (pktrc.str == null)
		    break;

		lineno++;

		// Get rid of comment lines (first character is #) and empty
		// line
		if (pktrc.str.trim().length() == 0 ||
			pktrc.str.trim().substring(0,1).equals("#"))
		    continue;

		st = new StringTokenizer(pktrc.str.trim()," \t=!<>");

		// Get the statement type
		try {
		    statement = st.nextToken().toLowerCase();
		} catch (NoSuchElementException e) {
		    continue;
		}

		// Get rid of commands and sub-commands
		if (statement.equals("logoutput:")	||
		    statement.equals("internal:")	||
		    statement.equals("external:")	||
		    statement.equals("user.privileged:")	||
		    statement.equals("user.notprivileged:")	||
		    statement.equals("connecttimeout:")	||
		    statement.equals("iotimeout:")	||
		    statement.equals("srchost:")	||
		    statement.equals("client")	||
		    statement.equals("pass")	||
		    statement.equals("block")	||
		    statement.equals("from:")	||
		    statement.equals("to:")	||
		    statement.equals("port")	||
		    statement.equals("log:")	||
		    statement.equals("command:")	||
		    statement.equals("protocol:")	||
		    statement.equals("proxyprotocol:")	||
		    statement.equals("}"))
		    continue;

		// OK, this is not a comment line and the statement type is not
		// recognized, save it.
		unknown_statements = unknown_statements.concat(pktrc.str+"\n");
	    }

	    // Close/delete the input file and open the output file
	    fw.c.closefile(fd);

	    infile.delete();
	}

	pktrc = fw.c.opencbcfile(tmpoutfn, 0);
	if (pktrc.rc < 0) {
	    fw.msgline.setText(pktrc.errmsg);
	    return(pktrc.rc);
	}

	fd = pktrc.rc;

	// logoutput statement
	tmp = logoutput_f.getText();
	if (tmp.length() > 0 || syslog_cb.getState() || stderr_cb.getState() ||
			stdout_cb.getState()) {
	    String lop = new String("logoutput: ");
	    if (syslog_cb.getState())
		lop = lop.concat("syslog ");
	    if (stderr_cb.getState())
		lop = lop.concat("stderr ");
	    if (stdout_cb.getState())
		lop = lop.concat("stdout ");
	    if (tmp.length() > 0)
		lop = lop.concat(tmp);
	    pktrc = fw.c.cbcwriteLine(fd, lop + "\n");
	    if (pktrc.rc != cglue.OK) {
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		return(-1);
	    }
	}

	// external statement
	pktrc = fw.c.cbcwriteLine(fd, "external: " +
				external_c.getSelectedItem() + "\n");
	if (pktrc.rc != cglue.OK) {
	    fw.msgline.setText(pktrc.errmsg);
	    fw.c.closefile(fd);
	    return(-1);
	}

	// internal statement
	for (i = 0; i < nifent; i++) {
	    String ip = intip_cb[i].getLabel();
	    String port;
	    if (intip_cb[i].getState()) {
		if (intport_f[i].getText().length() == 0)
		    port = new String("");
		else
		    port = new String(" port = " + intport_f[i].getText());
		pktrc = fw.c.cbcwriteLine(fd, "internal: " + ip + port + "\n");
		if (pktrc.rc != cglue.OK) {
		    fw.msgline.setText(pktrc.errmsg);
		    fw.c.closefile(fd);
		    return(-1);
		}
	    }
	}

	//  user.privileged statement
	tmp = priv_f.getText();
	if (tmp.length() > 0) {
	    pktrc = fw.c.cbcwriteLine(fd, "user.privileged: " + tmp + "\n");
	    if (pktrc.rc != cglue.OK) {
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		return(-1);
	    }
	}
	else {
	    fw.msgline.setText("Privileged username cannot be left blank");
	    fw.c.closefile(fd);
	    return(-1);
	}

	//  user.notprivileged statement
	tmp = notpriv_f.getText();
	if (tmp.length() > 0) {
	    pktrc = fw.c.cbcwriteLine(fd, "user.notprivileged: " + tmp + "\n");
	    if (pktrc.rc != cglue.OK) {
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		return(-1);
	    }
	}
	else {
	    fw.msgline.setText("Non-Privileged username cannot be left blank");
	    fw.c.closefile(fd);
	    return(-1);
	}

	//  connecttimeout statement
	tmp = conntimeout_f.getText();
	if (tmp.length() > 0) {
	    pktrc = fw.c.cbcwriteLine(fd, "connecttimeout: " + tmp + "\n");
	    if (pktrc.rc != cglue.OK) {
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		return(-1);
	    }
	}

	//  iotimeout statement
	tmp = conntimeout_f.getText();
	if (tmp.length() > 0) {
	    pktrc = fw.c.cbcwriteLine(fd, "iotimeout: " + tmp + "\n");
	    if (pktrc.rc != cglue.OK) {
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		return(-1);
	    }
	}

	//  srchost statement
	int tmpint = srchost_c.getSelectedIndex();
	if (tmpint > 0) {
	    pktrc = fw.c.cbcwriteLine(fd, "srchost: " +
				srchost_c.getSelectedItem() + "\n");
	    if (pktrc.rc != cglue.OK) {
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		return(-1);
	    }
	}

	// client rules
	// Note that we only process n - 1 lines, the last one is the
	// end marker
	for (i = 0; i < clntrules_f.body.getItemCount() - 1; i++) {
	    pktrc = fw.c.cbcwriteLine(fd, clntrules_f.body.getItem(i) + "\n");
	    if (pktrc.rc != cglue.OK) {
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		return(-1);
	    }
	}

	// server rules
	// Note that we only process n - 1 lines, the last one is the
	// end marker
	for (i = 0; i < svrrules_f.body.getItemCount() - 1; i++) {
	    int idx;
	    String line;
	    String line2half;
	    // Kludge to force a new line on "command:", "protocol:" and
	    // "proxyprotocol:"
	    line = svrrules_f.body.getItem(i);
	    if ((idx = line.toLowerCase().indexOf("command:")) != -1) {
		line2half = line.substring(idx);
		line = line.substring(0, idx);
		line = line.concat("\n");
		line = line.concat(line2half);
	    }
	    if ((idx = line.toLowerCase().indexOf("protocol:")) != -1) {
		line2half = line.substring(idx);
		line = line.substring(0, idx);
		line = line.concat("\n");
		line = line.concat(line2half);
	    }
	    if ((idx = line.toLowerCase().indexOf("proxyprotocol:")) != -1) {
		line2half = line.substring(idx);
		line = line.substring(0, idx);
		line = line.concat("\n");
		line = line.concat(line2half);
	    }
	    pktrc = fw.c.cbcwriteLine(fd, line + "\n");
	    if (pktrc.rc != cglue.OK) {
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		return(-1);
	    }
	}

	// All the unknown statements
	if (unknown_statements.length() > 0) {
	    if (split_line(fd, unknown_statements, null) < 0) {
		fw.c.closefile(fd);
		return(-1);
	    }
	}
	
	fw.c.closefile(fd);

	// Rename the outfile to tmpfn
	outfile.renameTo(infile);

	return(0);
    }

    // Add processing code for the edit button
    protected boolean moreAction(String s) {
	if ("Edit".equals(s)) {
	    lineno = 0;		// Don't want edit to highlight any lines
	    saveconf();
	    fw.newpanel(new edit(this,parent_panel,"EDIT " + capname));
	    return(true);
	}
	return(false);
    }
}
