/*============================================================================
 * filename:		tnproxy.java
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
//============================================================================
//
// R. James Livermore grants Freemont Avenue Software, Inc. a nonexclusive
// license to use, modify, and market this program using their tradename
// T.Rex.
//
//
// (C) COPYRIGHT R. James Livermore 1998-2000.  All rights reserved.
//
// (C) COPYRIGHT Freemont Avenue Software, Incorporated 1998-2000.
// All rights reserved.
//
// Date written:	June      24, 1998
// Last modified:	February   8, 2000
//
//===========================================================================*/
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

public class tnproxy extends Base {
    TextField groupid_f;
    TextField tnproxpath_f;
    TextField userid_f;
    TextField timeout_f;
    Choice auth_f;
    Choice xforward_f;
    Choice external_f;
    TextField anon_f;
    tnproxyrules rules_f;
    String groupid;
    String tnproxpath;
    String userid;
    String timeout;
    boolean auth;
    boolean xforward;
    boolean external;
    String anon;
    String rules;
    pktinfo pktrc;
    boolean tn;		// true = tnproxy, false = ftproxy
    public static boolean TN = true;
    public static boolean FTP = false;
    String capname;
    String proxpath;

    tnproxy(Base p, boolean tn) {
	super(p,"TNPROXY","");
	this.tn = tn;
	addBody();
    }

    void addBody() {
	int rc;

	// Initialize for tnproxy or ftproxy based on the tn parameter
	if (tn) {
	    component = fw.c.TNPROXY;
	    capname = new String("TNPROXY");
	    lowname = new String("tnproxy");
	    proxpath = new String("tnproxpath");
	}
	else {
	    component = fw.c.FTPROXY;
	    capname = new String("FTPROXY");
	    lowname = new String("ftproxy");
	    proxpath = new String("ftproxpath");
	    paneltitle.setText(capname);
	}

	// Initialize the values for this panel
	groupid = new String("staff");
	tnproxpath = new String("/home/hermes");
	userid = new String("hermes");
	timeout = new String("300");
	auth = true;
	xforward = false;
	external = false;
	anon = new String("");

	GridBagLayout gbl = new GridBagLayout();
	GridBagConstraints gbc = new GridBagConstraints();
	body.setLayout(gbl);

	// Get the config file over here to a temp file
	pktrc = download();
	tmpfn = pktrc.str;

	if (pktrc.rc != fw.c.OK) {
	    complete = false;
	    return;
	}

	// Construct the rules box now so it can collect permit/deny lines
	// while we read the config file
	rules_f = new tnproxyrules(fw);

	// Read in the temp file
	if ((rc = readconf()) == fw.c.FAILED) {
	    fw.c.closefd();
	    fw.newpanel(fw.getTop());
	}

	if (rc < 0)
	    return;

	gbc.anchor = GridBagConstraints.WEST;
	gbc.gridwidth = GridBagConstraints.REMAINDER;
	Button edit_b = new Button("Edit");
	edit_b.addActionListener(this);
	gbl.setConstraints(edit_b,gbc);
	body.add(edit_b);

	gbc.gridwidth = GridBagConstraints.RELATIVE;
	Label groupid_l = new Label("Group name which " + lowname +
		"should run under");
	gbl.setConstraints(groupid_l,gbc);
	body.add(groupid_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	groupid_f = new TextField(groupid,8);
	gbl.setConstraints(groupid_f,gbc);
	body.add(groupid_f);
	groupid_f.addFocusListener(this);

	gbc.gridwidth = GridBagConstraints.RELATIVE;
	Label tnproxpath_l = new Label("Directory which " + lowname +
		"should run under");
	gbl.setConstraints(tnproxpath_l,gbc);
	body.add(tnproxpath_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	tnproxpath_f = new TextField(tnproxpath,20);
	gbl.setConstraints(tnproxpath_f,gbc);
	body.add(tnproxpath_f);
	tnproxpath_f.addFocusListener(this);

	gbc.gridwidth = GridBagConstraints.RELATIVE;
	Label userid_l = new Label("User name which " + lowname + " should " +
		"run under");
	gbl.setConstraints(userid_l,gbc);
	body.add(userid_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	userid_f = new TextField(userid,8);
	gbl.setConstraints(userid_f,gbc);
	body.add(userid_f);
	userid_f.addFocusListener(this);

	gbc.gridwidth = GridBagConstraints.RELATIVE;
	Label timeout_l = new Label("Maximum idle time in seconds before " +
		"disconnect");
	gbl.setConstraints(timeout_l,gbc);
	body.add(timeout_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	timeout_f = new TextField(timeout,4);
	gbl.setConstraints(timeout_f,gbc);
	body.add(timeout_f);
	timeout_f.addFocusListener(this);

	gbc.gridwidth = GridBagConstraints.RELATIVE;
	Label auth_l = new Label("Require authentication for users in " +
		"protected networks");
	gbl.setConstraints(auth_l,gbc);
	body.add(auth_l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	auth_f = new Choice();
	auth_f.addItem("Yes");
	auth_f.addItem("No");
	if (auth)
	    auth_f.select("Yes");
	else
	    auth_f.select("No");
	gbl.setConstraints(auth_f,gbc);
	body.add(auth_f);
	auth_f.addFocusListener(this);

	if (tn) {
	    // Only process xforward if tnproxy
	    gbc.gridwidth = GridBagConstraints.RELATIVE;
	    Label xforward_l = new Label("Allow XWindows forwarding");
	    gbl.setConstraints(xforward_l,gbc);
	    body.add(xforward_l);

	    gbc.gridwidth = GridBagConstraints.REMAINDER;
	    xforward_f = new Choice();
	    xforward_f.addItem("Yes");
	    xforward_f.addItem("No");
	    if (xforward)
		xforward_f.select("Yes");
	    else
		xforward_f.select("No");
	    gbl.setConstraints(xforward_f,gbc);
	    body.add(xforward_f);
	    xforward_f.addFocusListener(this);
	}
	else {
	    // Only process external and anonymousftp if ftproxy
	    gbc.gridwidth = GridBagConstraints.RELATIVE;
	    Label external_l = new Label("Allow FTP from outside network");
	    gbl.setConstraints(external_l,gbc);
	    body.add(external_l);

	    gbc.gridwidth = GridBagConstraints.REMAINDER;
	    external_f = new Choice();
	    external_f.addItem("Yes");
	    external_f.addItem("No");
	    if (external)
		external_f.select("Yes");
	    else
		external_f.select("No");
	    gbl.setConstraints(external_f,gbc);
	    body.add(external_f);
	    external_f.addFocusListener(this);

	    gbc.gridwidth = GridBagConstraints.RELATIVE;
	    Label anon_l = new Label("Anonymous FTP server address");
	    gbl.setConstraints(anon_l,gbc);
	    body.add(anon_l);

	    gbc.gridwidth = GridBagConstraints.REMAINDER;
	    anon_f = new TextField(anon,16);
	    gbl.setConstraints(anon_f,gbc);
	    body.add(anon_f);
	    anon_f.addFocusListener(this);
	}

	gbl.setConstraints(rules_f,gbc);
	body.add(rules_f);

	sp.add(body);
    }

    /*=======================================================================
     * Reads the temp file in and interpret it
     *
     * Returns 0 if successful
     *		-1 if failed
     *		NONFATAL if connection need not be closed
     *
     *======================================================================*/
    int readconf() {
	int fd;
	if (tmpfn.equals("")) {
	    fw.msgline.setText("Internal error: " + lowname + " config " +
			"file has not been downloaded yet");
	    return(fw.c.NONFATAL);
	}
	File tmpfile = new File(tmpfn);
	if (!tmpfile.exists() || !tmpfile.isFile()) {
	    fw.msgline.setText("Internal error: the downloaded " +
			lowname + " conf file cannot be found");
	    return(fw.c.NONFATAL);
	}

	pktrc = fw.c.opencbcfile(tmpfn, 0);
	if (pktrc.rc < 0) {
	    fw.msgline.setText(pktrc.errmsg);
	    return(pktrc.rc);
	}

	fd = pktrc.rc;

	while ( true ) {
	    String tmp;
	    String statement;

	    pktrc = fw.c.cbcreadLine(fd);
	    if (pktrc.rc != cglue.OK) {
		if (pktrc.rc == cglue.CLOSED)
		    break;
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		return(pktrc.rc);
	    }

	    // Exit loop at EOF
	    if (pktrc.str == null)
		break;

	    lineno++;

	    // Get rid of comment lines (first character is #)
	    if (pktrc.str.trim().length() == 0 ||
			pktrc.str.trim().substring(0,1).equals("#"))
		continue;

	    StringTokenizer st = new StringTokenizer(pktrc.str.trim()," \t=:");

	    // Get the statement type
	    try {
		statement = st.nextToken().toLowerCase();
	    } catch (NoSuchElementException e) {
		continue;
	    }

	    // Get the next token into tmp - ignore error with no token found
	    // We do this here so we don't have to do a "try" for every
	    // statement type
	    try {
		tmp = st.nextToken();
	    } catch (NoSuchElementException e) {
		complete = false;
		fw.c.closefile(fd);
		tmpfile.delete();
		fw.newpanel(new edit(this,parent_panel,"EDIT " + capname));
		fw.msgline.setText("Highlighted line contains an " +
			"invalid statement, please correct and retry");
		return(fw.c.NONFATAL);
	    }

	    if (statement.equals("groupid")) {
		groupid = tmp;
		continue;
	    }

	    if (statement.equals(proxpath)) {
		tnproxpath = tmp;
		continue;
	    }

	    if (statement.equals("userid")) {
		userid = tmp;
		continue;
	    }

	    if (statement.equals("timeout")) {
		timeout = tmp;
		continue;
	    }

	    if (statement.equals("auth")) {
		if (tmp.toLowerCase().equals("no"))
		    auth = false;
		continue;
	    }

	    if (tn) {
		// Process xforward if tnproxy
		if (statement.equals("xforward")) {
		    if (tmp.toLowerCase().equals("yes"))
			xforward = true;
		    continue;
		}
	    }
	    else {
		// Process external and anonymousftp if ftproxy
		if (statement.equals("external")) {
		    if (tmp.toLowerCase().equals("yes"))
			external = true;
		    continue;
		}

		if (statement.equals("anonymousftp")) {
		    anon = tmp;
		continue;
		}
	    }

	    if (statement.equals("permit") || statement.equals("deny")) {
		rules_f.body.addItem(pktrc.str);
		continue;
	    }

	    // Invalid statement
	    complete = false;
	    fw.c.closefile(fd);
	    tmpfile.delete();
	    fw.newpanel(new edit(this,parent_panel,"EDIT " + capname));
	    fw.msgline.setText("Highlighted line contains an invalid " +
		"statement, please correct and retry");
	    return(fw.c.NONFATAL);
	}

	// Add an end marker so the user can insert after the last real rule
	rules_f.body.addItem("[end marker]");

	// Delete the temp file since we no longer need it
	fw.c.closefile(fd);
	tmpfile.delete();
	return (0);
    }

    public void handleOK() {

	if (saveconf() < 0)
	    return;

	pktrc = upload();
	if (pktrc.rc < 0)
	    return;

	fw.msgline.setText("New " + lowname + " config file " +
		"successfully created");
	fw.newpanel(parent_panel);
    }

    public int saveconf() {
	int fd;
	String tmp;

	pktrc = fw.c.opencbcfile(tmpfn, 0);
	if (pktrc.rc < 0) {
	    fw.msgline.setText(pktrc.errmsg);
	    return(pktrc.rc);
	}

	fd = pktrc.rc;

	// groupid statement
	pktrc = fw.c.cbcwriteLine(fd,"groupid = " + groupid_f.getText() +
				"\n");
	if (pktrc.rc != cglue.OK) {
	    fw.msgline.setText(pktrc.errmsg);
	    fw.c.closefile(fd);
	    return(-1);
	}

	// tnproxpath statement
	pktrc = fw.c.cbcwriteLine(fd,proxpath + " = " +
				tnproxpath_f.getText() + "\n");
	if (pktrc.rc != cglue.OK) {
	    fw.msgline.setText(pktrc.errmsg);
	    fw.c.closefile(fd);
	    return(-1);
	}

	// userid statement
	pktrc = fw.c.cbcwriteLine(fd,"userid = " + userid_f.getText() +
				"\n");
	if (pktrc.rc != cglue.OK) {
	    fw.msgline.setText(pktrc.errmsg);
	    fw.c.closefile(fd);
	    return(-1);
	}

	// timeout statement
	pktrc = fw.c.cbcwriteLine(fd,"timeout = " + timeout_f.getText() +
				"\n");
	if (pktrc.rc != cglue.OK) {
	    fw.msgline.setText(pktrc.errmsg);
	    fw.c.closefile(fd);
	    return(-1);
	}

	// auth statement
	if (auth_f.getSelectedItem().toLowerCase().equals("no"))
	    tmp = new String("no");
	else
	    tmp = new String("yes");
	pktrc = fw.c.cbcwriteLine(fd,"auth = " + tmp + "\n");
	if (pktrc.rc != cglue.OK) {
	    fw.msgline.setText(pktrc.errmsg);
	    fw.c.closefile(fd);
	    return(-1);
	}

	if (tn) {
	    // tnproxy xforward statement
	    if (xforward_f.getSelectedItem().toLowerCase().equals("yes"))
		tmp = new String("yes");
	    else
		tmp = new String("no");
	    pktrc = fw.c.cbcwriteLine(fd,"xforward = " + tmp + "\n");
	    if (pktrc.rc != cglue.OK) {
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		return(-1);
	    }
	}
	else {
	    // ftproxy external statement
	    if (external_f.getSelectedItem().toLowerCase().equals("yes"))
		tmp = new String("yes");
	    else
		tmp = new String("no");
	    pktrc = fw.c.cbcwriteLine(fd,"external = " + tmp + "\n");
	    if (pktrc.rc != cglue.OK) {
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		return(-1);
	    }

	    // ftproxy anonymous statement
	    if (anon_f.getText().length() > 0) {
		pktrc = fw.c.cbcwriteLine(fd,"anonymousftp = " +
				anon_f.getText()+"\n");
		if (pktrc.rc != cglue.OK) {
		    fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		    return(-1);
		}
	    }
	}

	//permit/deny rules
	// Note that we only process n - 1 lines, the last one is the
	// end marker
	for (int i = 0; i < rules_f.body.getItemCount() - 1; i++) {
	    pktrc = fw.c.cbcwriteLine(fd,rules_f.body.getItem(i) + "\n");
	    if (pktrc.rc != cglue.OK) {
		fw.msgline.setText(pktrc.errmsg);
		fw.c.closefile(fd);
		return(-1);
	    }
	}
	fw.c.closefile(fd);

	return(0);
    }

    // Add processing code for the edit button
    protected boolean moreAction(String s) {
	if ("Edit".equals(s)) {
	    lineno = 0;		// Don't want edit to highlight any lines
	    saveconf();
	    fw.newpanel(new edit(this,parent_panel,"EDIT " + capname));
	    return(true);
	}
	return(false);
    }
}
