/*============================================================================
 * file:	include/des_user.h
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=======================================================================
 * 
 * (C) COPYRIGHT Freemont Avenue Software, Inc. 1996-2000.
 *                                                                      
 * Author:	M. Everett Hinckley
 * History:	                         
 * Date written:	January   13, 1998
 * Last modified:	February   8, 2000
 *
 * function:	header file for des_user.c
 *		describes status modes, macros, and function prototypes
 *		for tnkey() -o encrypted file manipulation
 * 		decrypts DES key from tnkey() encrypted file and 
 *		acts as an interactive CRYPTO Card.  
 *                                                                     
 *=========================================================================*/

#include <stdio.h>
#include <string.h> 
#include <time.h>
#include "tnkey.h"
#include "des.h"

#ifndef _DES_USER_
#define _DES_USER_

/* des_user modes */
  /* 1 byte of data w/ high 4 bits indicating user mode (debug/normal operation)
     and low 4 bits for ready mode (ready/not ready) */
#ifndef DU_NORM_NOT_READY
#define DU_NORM_NOT_READY	0x00
#endif
#ifndef DU_NORM_READY
#define DU_NORM_READY		0x0F
#endif
#ifndef DU_DEBUG_NOT_READY
#define DU_DEBUG_NOT_READY	0xF0
#endif
#ifndef DU_DEBUG_READY
#define DU_DEBUG_READY		0xFF
#endif

/* des_user macros */
  /* return true if in debug mode */
#ifndef DU_DEBUG
#define DU_DEBUG(stat)		((unsigned char)(((stat)>>4)&0x0F) == 0x0F)
#endif
  /* return true if in ready mode */
#ifndef DU_READY
#define DU_READY(stat)		((unsigned char)((stat)&0x0F) == 0x0F)
#endif
  /* return value of stat w/ changed ready mode */
#ifndef DU_CH_READY
#define DU_CH_READY(stat)		((unsigned char)((stat)^0x0F))
#endif
  /* return value of stat w/ changed user mode */
#ifndef DU_CH_USER
#define DU_CH_USER(stat)		((unsigned char)((stat)^0xF0))
#endif
  /* return value of stat in ready mode w/ same user mode */
#ifndef DU_MAKE_READY
#define DU_MAKE_READY(stat)	((unsigned char)((stat)|0x0F))
#endif   
  /* return value of stat in not ready mode w/ same user mode */
#ifndef DU_MAKE_NOT_READY
#define DU_MAKE_NOT_READY(stat)	((unsigned char)((stat)&0xF0))
#endif

/* cdes_user function return codes */
#ifndef DU_OK
#define DU_OK			0
#endif
#ifndef DU_FAIL
#define DU_FAIL			-1
#endif

/* des_user default timeout period in seconds */
#ifndef DU_TIMEOUT
#define DU_TIMEOUT		60
#endif

/* des_user error message string length */
#ifndef DU_ERROR_LENGTH
#define DU_ERROR_LENGTH		250
#endif

/* global variables */
time_t duCurrentTime, 			/* time of current authentication */
       duLastTime;			/* times of last authentication */
char duError[DU_ERROR_LENGTH];		/* des_user error message string */
int duStatus;				/* current mode status */

/* function prototypes */
void du_init(void);
void du_setschedule(des_key_schedule  * keysched, unsigned long key1, 
		    unsigned long key2);
int du_validate(char *filename, unsigned char * name, unsigned char * key, 
                char * passwd);
int du_respond(char * challenge, unsigned char * key, char * response);
int du_timeout(long timeOut);

#endif
