/*============================================================================
 * filename:		include/fwmon.h
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * (C) COPYRIGHT Freemont Avenue Software, Inc. 1996-2000.
 * All rights reserved.
 *		
 * Date written:	September 26, 1996
 * Last modified:	February   8, 2000
 *		
 *	                         
 * function:
 *
 *  
 *
 *===========================================================================*/

/*===========================================================================
 * structure to hold parameters used to repel SYN flood attacks
 * Some of the members in this structure (qlen, himark and lomark) can be
 * specified as port specific parameters also.  The values in here (as opposed
 * to the port specific ones specified in struct synport) are to be applied
 * to all ports except when overriden by port specific parameters.
 *==========================================================================*/
struct synparms {
   FILE *iplog;			/* File pointer of ip log file */
   char *trusted;		/* Filename of trusted ip's */
   unsigned short chk_interval;	/* #secs between check to SYN_RCVD state queue
				   to repel SYN attacks
				   Set to zero to disable syn check */
   unsigned short qlen;		/* Set SYN_RCVD queue to this length */
   unsigned short himark;	/* #sockets in SYN_RCVD queue above which
				   cleanup of the queue should start */
   unsigned short lomark;	/* #sockets to keep in SYN_RCVD queue during
				   cleanup.  Assume 10% of syn_qlen to be a
				   good number */
   int	verbose;		/* determines whether to perform verbose log */
   int	plistsz;		/* #bytes in plist array */
   struct synport *plist;	/* points to list of port specific parameters
				   NULL if no port specific parameters */
};

#define DFLT_QLEN	10
#define DFLT_HIMARK	14
#define DFLT_LOMARK	3

/*===========================================================================
 * structure to hold port specific parameters used to repel SYN flood attacks
 * If synparms.plist is non-NULL, it points to an array of struct synports.
 * A 0 in synports.port indicates the end of the list.
 *==========================================================================*/
struct synport {
   unsigned short port;
   unsigned short qlen;
   unsigned short himark;
   unsigned short lomark;
};

/*=====================================================================
 * Structure for trusted and untrusted ip addresses for synmon
 *====================================================================*/
struct tip {
   unsigned long ip;
   unsigned long mask;
};

/*=====================================================================
 * Structure for procmon monitored processes
 *====================================================================*/
typedef struct procmonparms procmonparms;
struct procmonparms {
   char *procname;
   int  expected_n;
   int  expected_hi_n;
   int	actual_n;
   procmonparms *next;
};

struct pmonconf
{
   struct synparms synparms;
   procmonparms	*procs;		/* pointer to '|' separated list of
				   processes to monitor */
   int		procmon_intv1;	/* Interval in seconds for procmon to check */
   int		procmon_intv2;	/* Interval in seconds for procmon to alert
				   default to procmon_intv1 */
   char		*procmon_email;	/* address to email procmon alerts */
   u_char	spoofmon;	/* 0=do not run spoofmon
				   1=run spoofmon */
};

/*=====================================================================
 * Return codes from our kernel extension
 *====================================================================*/
#define	SYN_CANT_PIN	-1	/* Cannot pin */
#define	SYN_CANT_XMT	-2	/* Cannot xmalloc for tips */
#define SYN_CANT_XMU	-3	/* Cannot xmalloc for utips */
#define SYN_CANT_CPT	-4	/* Cannot copyin tips */
#define SYN_CANT_CPU	-5	/* Cannot copyin utips */
#define SYN_CANT_CPS	-6	/* Cannot copyin synparms */
#define SYN_CANT_PTI	-7	/* Cannot pin tips */
#define SYN_CANT_PUT	-8	/* Cannot pin utips */
#define SYN_CANT_XMP	-9	/* Cannot xmalloc portparm */
#define SYN_CANT_CPP	-10	/* Cannot copyin portparm */
