/*============================================================================
 * filename:		include/gwuser.h					
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * written by:		R. James Livermore 
 *			1602 Mossy Stone
 *			Houston, TX 77077
 *
 * R. James Livermore grants Freemont Avenue Software, Inc. a nonexclusive
 * license to use, modify, and market this program using their tradename
 * T.Rex.	
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000. 
 * All rights reserved. 
 *		
 * Date written:	January   13, 1995
 * Last modified:	February   8, 2000
 *	                         
 * function:
 *
 *  The gwuser structure contains information used to control user access     
 *  to gateway functions.
 *
 *===========================================================================*/
#ifdef MSDOS
#include	<winsock.h>
#else
#include	<sys/time.h>
#ifdef _AIX
   #include	"ndbm.h"  
#else
   #ifdef linux
      #include <db1/ndbm.h> 
      #include <string.h>
   #else
      #include 	<ndbm.h> 
   #endif
#endif
#endif


/*=====================================
 * Old database record 
 *====================================*/

struct GWUSER_OLD 
{ 
   short reclen;			/* record length                   */
   char	 ver;    			/* version number 		   */
   char  rel;                           /* release number 		   */
   char  mod;                           /* modification level 		   */
   char  username[UNMSIZ];		/* user name                       */
   char  groupname[GRPSIZ];		/* group name                      */
   char  password[PWDSIZ];		/* user password                   */
					/*---------------------------------*/
   char  tokentype;                     /* type of security token          */
#define  NONE		0x00		/* no security token               */
#define  CRYPTO 	0x01		/* CRYPTOCard                      */
#define  SNK		0x02		/* SecureNet Key                   */
#define  SDI		0x03		/* Security Dynamics               */
/*#define  PASSWD		0x04*/	/* Standard password		   */
					/*---------------------------------*/
   char  userkey[9];	                /* SNK key                         */
   char  in_shell[64]; 			/* shell for protected user        */
   char  ext_shell[64];                 /* shell for unprotected user      */
					/*---------------------------------*/
   char  ftpaccess;			/* ftp access priviledges          */
#define  NO_FTP	  	0x00		/* no ftp access allowed           */
#define	 PRO_FTP  	0x01		/* can ftp from protected host	   */
#define  UNPRO_FTP	0x02		/* can ftp from unprotected host   */
#define  PRO_FTPSA	0x04		/* protected ftp with Strong Auth  */
					/*---------------------------------*/
   char  tnaccess;			/* telnet access                   */
#define  NO_TN		0x00		/* no telnet access allowed        */
#define  PRO_TN		0x01		/* telnet from protected host      */
#define  PRO_TNSA	0x02		/* protected telnet with StrongAuth*/
#define  UNPRO_TN	0x04		/* telnet from unprotected host    */
#define  ADM_TN  	0x08		/* can login to firewall           */
					/*---------------------------------*/
   time_t exptime;	                /* exp date in epoch format 	   */
};

/* Open NDBM database mode */

#define VERIFYADMIN	0x01		/* verify admin (database exists)  */
#define EXIST		0x02		/* database exists                 */
#define CREATE		0x04            /* create database                 */
#define NO_LOCK		0x08            /* don't lock database (read only) */

/*======================================
 * User record
 *=====================================*/

#define REALSIZ 	33		/* Real name max length (+1)       */
#define PHONESIZ 	33		/* Phone number max length (+1)    */

struct GWUSER 
{
   char type;				/* record type (USER)              */
   char random[2];			/* random bytes for encryption     */
#define USER		0x01
#define GROUP		0x02
#define INFO		0x03
   char  username[UNMSIZ];		/* user ID                         */
   char  realname[REALSIZ];		/* real name                       */
   char  phonenum[PHONESIZ];		/* phone number                    */
   char  groupname[GRPSIZ];		/* group name                      */
# define MAXPWDS	12	/* max no of current & previous passwords  */
   char  password[MAXPWDS][PWDSIZ];	/* user password                   */
   time_t passworddate;			/* password create/mod date (epoch)*/
   char  tokentype;			/* type of security token          */
   char  userkey[9];	                /* token key                       */
   char  tnaccess;			/* telnet access                   */
   int   ftpaccess;			/* ftp access                      */   
#define GET_FTP		0x08		/* ftp gets from secure network    */
#define PUT_FTP		0x10		/* ftp puts to secure network      */
#define DEL_FTP		0x20		/* ftp deletes on secure network   */
   time_t startdate;			/* start date (epoch)              */
   time_t stopdate;	                /* expire date (epoch)             */
   char  accesstimes[21];		/* allowed access times (hours)    */
   time_t recorddate;	                /* record create/mod date (epoch)  */ 
   int	no_login_attempts;		/* number of unsuccessful logins   */	
   char status;				/* account status                  */
#define LOCKED		0x00
#define UNLOCKED	0x01
};

/*======================================
 * Group record 
 *=====================================*/   
struct GWGROUP 
{
   char type;				/* record type (GROUP)             */
   char random[2];			/* random bytes for encryption     */
   char groupname[GRPSIZ];		/* group ID                        */
   char realname[REALSIZ];		/* real name                       */
   char tokentype;			/* type of security token          */
   char tnaccess;			/* telnet access                   */
   int  ftpaccess;			/* ftp access                      */   
   char accesstimes[21];		/* allowed access times (hours)    */
   time_t recorddate;	                /* record create/mod date (epoch)  */
};

/* Summary information record */

#define GWINFO_KEY "GWINFO"

#define GWINFO_CREATE		0
#define GWINFO_MODIFY		1
#define GWINFO_VERIFY		2
#define MODIFYDATE		0x00
#define NUMUSERS		0x01
#define NUMGROUPS		0x02
/*======================================
 *
 *======================================*/ 
struct GWINFO 
{
   char type;				/* record type (INFO)              */
   char random[2];			/* random bytes for encryption     */
   unsigned char version;		/* database version                */
#define GWVERSION		1	/* initial NDBM version            */
   long total_users;			/* number of user records          */
   long total_groups;			/* number of group records         */
   time_t createdate;			/* creation date (epoch)           */
   time_t modifydate;                   /* modification date (epoch)       */
};

/* Obscured DES key for record encryption and decryption */

#define GWDESKEY "&t*yr6sc*3s91d^#"

#define GWIVEC   (des_cblock *)"86afd129"

/* Stored data length: first byte for record type is unencrypted, */
/* remaining data is written in multiples of eight bytes          */

#define encrypted_sizeof(struct) (((sizeof(struct) - 1) / 8 + 1) * 8 + 1)
#define decrypted_sizeof(struct) (encrypted_sizeof(struct))

/* Encrypted, decrypted data array typedefs */

typedef char GWUSER_ENCRYPT[encrypted_sizeof(struct GWUSER)];
typedef char GWUSER_DECRYPT[decrypted_sizeof(struct GWUSER)];

typedef char GWGROUP_ENCRYPT[encrypted_sizeof(struct GWGROUP)];
typedef char GWGROUP_DECRYPT[decrypted_sizeof(struct GWGROUP)];

typedef char GWINFO_ENCRYPT[encrypted_sizeof(struct GWINFO)];
typedef char GWINFO_DECRYPT[decrypted_sizeof(struct GWINFO)];

/*=============================== 
 * Configuration file structure 
 *==============================*/

struct GWCONF 
{
   char groupname[GRPSIZ];	/* user default                            */
   char tokentype;		/* group default                           */
   char tnaccess;		/* group default                           */
   int  ftpaccess;		/* group default                           */
   char accesstimes[21];	/* group default                           */
   int password_min_len;	/* password min length                     */
   int password_min_alpha;	/* password min alpha chars                */
   int password_min_other;	/* password min other chars                */
   int password_max_age;	/* password max change interval (days)     */
   int no_unique_passwords;	/* no unique passwords before reuse        */
   int no_login_attempts;	/* no failed login attempts before locked  */
};

typedef struct GWCONF GWCONF;

/*=============================== 
 * Administrator structure 
 *==============================*/  
struct GWADMIN 
{
  char user[UNMSIZ];			/* current administrator name      */
  char group[GRPSIZ];			/* current adminstrator group      */
};

/*==============================
 * User options structure 
 *=============================*/

#ifndef bool
#  define bool char
#endif

struct GWOPTIONS 
{
   char program[8];			/* "gwuser" or "gwgroup"           */
   int action;				/* command line leading argument   */
#define GWHELP			0
#define GWADD			1
#define GWADD_HELP		2
#define GWMOD			3
#define GWMOD_HELP		4
#define GWDEL			5
#define GWDEL_HELP		6
#define GWLIST			7
#define GWLIST_HELP		8
#define GWQUERY			9
#define GWQUERY_HELP		10
#define GWCONVERT		11
#define GWCONVERT_HELP		12
#define MAX_USERS		64	/* Max users on command line       */
#define MAX_GROUPS		64	/* Max groups on command line      */
   char users[MAX_USERS][UNMSIZ];	/* User names to list              */
   char groups[MAX_GROUPS][GRPSIZ];	/* Group names to list             */
   int nusers;				/* Number of user names            */
   int ngroups;				/* Number of group names           */
   bool allinfo;			/* List all information            */
   bool user;				/* User name specified             */
   bool real;				/* Real name specified             */
   bool phone;				/* Phone number specified          */
   bool group;				/* Group membership specified      */
   bool pass;				/* Password specified              */
   bool token;				/* Token specified                 */
   bool key;				/* Token key specified             */
   bool telnet;				/* Telnet access specified         */
   bool ftp;				/* FTP access specified            */
   bool start;				/* Start date specified            */
   bool stop;				/* Stop date specified             */
   bool times;				/* Time access specified           */
   bool stat;				/* Status specified                */
};

typedef struct GWOPTIONS GWOPTIONS;

/*=============================================================================
 * Function prototypes
 *===========================================================================*/

#ifdef __cplusplus
extern "C"
{
#endif
int addgroup(void);
int adduser(void);
int authadmin(void);
int check_password(char *, char [MAXPWDS][PWDSIZ], struct GWCONF *, char *);
int close_db();
int convert_db(void);
int create_db(void);
void decrypt_record(char *, char *, int);
int delgroup(void);
int deluser(void);
void des_encryption(char *, char *, int, char);
void encrypt_record(char *, char *, int);
int get_date(int, char **, char *, time_t *);
int get_ftp(int argc, char **, int *); 
int get_gwgroup(char *groupnmae, struct GWGROUP *); 
int get_gwuser(char *, struct GWUSER *); 
int get_gwuserconf(struct GWCONF *);
int get_key(int, char **, char *);
int get_password(int, char **, char *);
int get_string(int, char **, char *, char *, int);
int get_telnet(int, char **, char *); 
int get_times(int argc, char **argv, char *access);
int get_token(int, char **, char *);
int glob2reg(char *, char *, int);
int groupargs(int, char **);
void grouphelp(int, FILE *);
int gwuserconf(void);
int info_db(char, char, long, long);
int listgroup(void);
int listuser(void);
int lockuser(char *);
int modgroup(void);
int moduser(void);
int open_db(int);
int pattern_match(char *, char *);
int query_db(void);
void randomkey(char *);
int set_nologins(char *, int, int);
int set_password(char *, char *);
int userargs(int, char **);
void userhelp(int, FILE *);
#ifdef __cplusplus
}
#endif
