/*============================================================================
 * filename:		include/smwrap.h
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *										
 * written by:		Jim Livermore						
 *			Freemont Avenue Software, Incorporated
 *
 * (C) COPYRIGHT Freemont Avenue Software, Inc. 1995-2000.  All rights
 * reserved.
 *		
 * Date written:	March      6, 1995
 * Last modified:	February   8, 2000
 *			February  29, 2000 - updated function prototypes
 *	                         
 * function:
 *
 *  Define constants used by smwrap, the smwrap structure, the mail blocker
 *  structure, the trace structure and function prototypes. 
 *  
 *
 *===========================================================================*/
#define MAXADDRLN     256	/* maximum e-mail address length	*/
#define MAXSMTPLEN    990	/* maximum SMTP command length          */
#define MAXHEADERSZ 32768	/* maximum total header size - default  */
#define MINMAXHDR    2048 	/* min value allowed for maxheadersz    */
#define MAXLINELN    2048	/* maximum line length			*/
#define MAXMIMEARGS    20	/* maximum MIME arguments		*/
#define	MAXMIMEHDR  32768	/* maximum MIME Header size		*/
#define MAXNAME	      256	/* maximum name length			*/
#define MAXSMARG      512	/* maximum number of sendmail arguments */
#define MSGOK		0	/* message OK				*/
#define UNEXPEXIT	1	/* unexpected peer departure		*/
#define MSGTOOBIG       2	/* message exceeds maxbytes		*/
#define HDRTOOBIG	3	/* message Header exceeds maxheadersz   */
#define HDRADRBAD	4	/* dangerous header address		*/
#define FPRTERR		5	/* fprintf error                        */
struct SMWRAP 
{
   unsigned long userid;	/* user id from passwd file     	*/
   unsigned long groupid;	/* group id                     	*/ 
   int	maxbytes;		/* maximum message size			*/
   int	maxheadersz;		/* maximum total size of headers        */
   int	maxreceipts;		/* max receipts per message     	*/
   int 	timeout;		/* smwrap timeout in seconds           	*/
   int	wakeup;			/* smwrapd sleep time between scans	*/
   int	maxpids;		/* maximum process ids (children)	*/
   int	traceflg;		/* trace flag				*/
   int  linecnt;		/* message line count			*/  
   char spoolpath[MAXPATHLN+1];	/* smwrap spool directory   		*/   
   char smwpdpath[MAXPATHLN+1]; /* smwrapd executable pathname		*/   
   char smtppath[MAXPATHLN+1];	/* smtp (sendmail) executablepath	*/
   char undelivpath[MAXPATHLN+1];/* undeliverable mail path 		*/   
   char aliasflag;		/* 0->use aliases, NOALIAS, ALIASREQ	*/
   char bogushelo;		/* bogus helo  yes=1, no=0  RFC 1123	*/    
   char translate;		/* translate outbound mail headers      */   
   char postoffice[128];	/* hostname of internal post office     */
}; 
#define NOALIAS		1       /* do not use aliases, use P.O.		*/ 
#define ALIASREQ	2       /* alias record required for outbound	*/ 
#define TRANSLATEON     1 
#define TRANSLATEOFF    0

/*============================================================================
 * mail block structure.  When a pointer is NULL, it is assumed to be a
 * wildcard.
 *===========================================================================*/
typedef struct mailblk mailblk;
struct	mailblk
{
   int	status;			/* permit/deny/block/forward/sequester	*/
#define MB_PERMIT	0
#define MB_BLOCK	1	/* default is deny - block		*/
#define MB_FORWARD	2
#define MB_SEQUESTER	4
   char *fromuser;		/* sender user name			*/
   char *fromhost;		/* sender host name			*/
   char *touser;		/* recipient user name			*/
   char *tohost;		/* recipient host name			*/
   char *forw;			/* user@host forward address		*/
   mailblk *next;		/* link to next				*/
};
/*============================================================================
 * Smwrap trace entry and trace header
 *===========================================================================*/  
struct	trace
{
   int		entryno; 
   char		sr;		/* S = send, R = Receive */
   char		entry[96];
}; 
struct  tracehdr
{
   int		recno;
   int		tindx;
   char		title[80];
#define	MAXTRACE 	32 
}; 
/*===========================================================================
 * function prototypes
 *===========================================================================*/
int authmail(char *from, char *to, struct mailblk *topoflist, char * forward);
extern void burring(int sig);
int bldbody(FILE *in, FILE *out,struct SMWRAP *smwpptr,int IPsecure,
            	char *extFrom,char *lochostname, char *fromname, 
		struct tracehdr *tracehdrptr, struct trace  *tracebufadr); 
void debugbind(unsigned short debugport);  
int externalrelay(char *name);
int getsmwp(struct SMWRAP *smwpptr);  
extern int getswocrlf(char* buf, int siz, FILE *fp);
int filesize(int fd); 
int internalrcpt(char *name, char *aka);
static FILE *smwrapopen(char rnamebuf[]); 
int trace_dump(struct tracehdr *th, struct trace *tracebuf,
               struct SMWRAP *smwpptr );
struct trace *trace_init(struct tracehdr *th); 
int trace_it(struct tracehdr *th, struct trace tracea[],char sr, char *obuf);
int tstpname(int fd, char *PeerHost, char *PeerIP, int z); 
int lockun_fd(int f);
int lock_fd(int f); 		
int mkstemp(char * );		/* added for C++ */ 
int transcc(char *buf, char *lochostname); 
int transend(char *name, char *aka, char *lochostname, char translate);
int validrcpt(char *name, char *aka);  
