/*============================================================================
 * filename:		include/chk_spoof.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*====================================================================== 
 *
 * written by:		R. James Livermore
 *
 * (C) COPYRIGHT R. James Livermore 1996-2000.  All rights reserved.
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1996-2000.
 * All rights reserved.
 *
 * Date written:	May       13, 1996
 * Last modified:	February   8, 2000 - added LPL
 *	
 * functions:
 *		chk_spoof()
 *		Check for IP address spoofing.
 *		For telnet and ftp, if the client address looks like a
 *		protected IP address, chk_spoof is called to make sure
 *		the address is not spoofed.
 * 
 *=====================================================================*/  
static	char	cprght[] = 
  	"(C) COPYRIGHT Freemont Avenue Software, Inc. 1996-2000 chk_spoof,"
  	" V 1.0 00/02/14"; 

#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <signal.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#ifdef __hppa
#  include <sys/param.h>
#  include <sys/pstat.h>
#endif
#ifdef _AIX
#  include <sys/utsname.h>
#endif
#include "firewall.h"
#include "spoofmon.h"

int islocal(in_addr_t ip);
int getlocal(in_addr_t local_ip[],char portname[MAXPORTS][IFNAMSIZ],long *nlips);

/*=========================================================================*
 * Function: int chk_spoof(char *pname,struct in_addr inside_ipaddr,
 *				int inside_port,
 *				struct in_addr clnt_ipaddr,int clnt_port,
 *				int IPsecure)
 *
 *		pname:	character string representing name of the calling
 *			program (e.g. "tnproxy").  This is used only in
 *			displaying error messages.
 *		inside_ipaddr:	struct in_addr representing the
 *				firewall protected IP address.
 *		inside_port:	port number on the firewall side of socket.
 *		clnt_ipaddr:	struct in_addr representing the
 *				IP address of the client.
 *		clnt_port:	Port number on the client side of the socket.
 *		IPsecure:	indicates whether the source address is in
 *				the protected net or not.
 *
 * Outputs:
 *	return 0 if OK.  Note that it only means that the src addr/port and
 *		dst addr/port combination has not been logged
 *		(by the fwmon/spoofmon daemon) in the shared memory
 *		segment as an attempted	spoofed address attack.  This is why
 *		chk_spoof is called twice by the proxies.
 *      exit 1 if a system error (e.g. read error, error opening pipe)
 *		occurred.
 *=========================================================================*/

int chk_spoof(char *pname,struct in_addr inside_ipaddr,int inside_port,
		struct in_addr clnt_ipaddr,int clnt_port,int IPsecure)
{
#ifndef __sun
   FILE		*pipe_in;
   FILE		*pf;
   int		pipefds[2];
   int		i;
   int		rc;
   int		shmid=0;
   char		pids[LINE_MAX];
   char		tmport[16];
   char		*USERALERT= "IP spoofing activities detected, please contact "
   			    "the network administrator immediately.\n";
   struct conn_entry	*shp,*shtop;
#ifdef __hppa
   struct pst_status	monitor;
   int		pid;
#else
   pid_t	pid;
#endif
#ifdef _AIX
   struct utsname utsname;

   /*=======================================================================
    * If AIX 3.2, just return OK,  there is no tcpdump
    *======================================================================*/
   if (uname(&utsname) < 0)
   {
	syslog(LOG_ERR,"Error: Uname call failed in chk_spoof. errno=%d.\n",
			errno);
	return(1);
   }
   if (strcmp(utsname.version,"4"))
	return(0);
#endif

   sleep(1);

   /*=======================================================================
    * Get the shared memory segment.
    *======================================================================*/
   if ((shmid = shmget(SHMKEY,SHMSIZE,0)) < 0)
   {
	syslog(LOG_ERR,"Error: shmget failed, errno = %d, make sure fwmon"
			" is running.\n",errno);
	say(0,"Connections are disabled, please contact network administrator.\n");
	closelog();
	exit(1);
   }
   if ((shtop = (struct conn_entry *)shmat(shmid,(char *)0,0)) == (struct conn_entry *)-1)
   {
	syslog(LOG_ERR,"Error: shmat failed, errno = %d\n",errno);
	closelog();
	exit(1);
   }

   /*==================================================================*
    * if SHMTOP is a zero offset, it is an indication that spoofmon has
    * been disabled by the fwmon -n option.
    *==================================================================*/
   if (SHMTOP == 0)
   {
	shmdt((char *)shtop);
	return(0);
   }

   /*==================================================================*
    * Make sure spoofmon is running
    *==================================================================*/
   if (!(pf = fopen(FWMONPIDS,"r")))
   {
	syslog(LOG_ERR,"Error: chk_spoof failed to open %s, errno=%d.\n",
		FWMONPIDS,errno);
	closelog();
	exit(1);
   }
   while (fgets(pids,LINE_MAX,pf) != NULL)
   {
      pid = atoi(strtok(pids," \t\n"));
#ifdef __hppa
      rc = pstat_getproc(&monitor,sizeof(monitor),0,pid);
      if (rc < 0 ||
	    monitor.pst_pid != pid 	  ||
	    monitor.pst_stat == PS_ZOMBIE ||
	    monitor.pst_stat == PS_STOP )
#else
      if (kill(pid,SIGCONT))
#endif
      {
         syslog(LOG_ERR,"Error: spoofmon not running, %s connection disabled"
	    		" until spoofmon is activated.\n",pname);
	 say(0,"Connections are disabled, please contact network administrator.\n");
	 fclose(pf);
	 closelog();
	 exit(1);
      }
   }

   if (!feof(pf))
	syslog(LOG_ERR,"Error: spoofmon failed reading %s, errno=%d.\n",
		FWMONPIDS,errno);
   fclose(pf);

   /*==================================================================*
    * Check the shared memory segment for a match
    *==================================================================*/
   for (shp=(struct conn_entry *)((char *)shtop + SHMTOP);
					shp < shtop+SHMNENT; shp++)
   {
      /*=================================================================
       * If the source ipaddr, port and destination ipaddr, port
       * combination match an entry in the shared memory segment,
       * then this is a connection to the inside port of the firewall,
       * perhaps one with a spoofed address.
       *================================================================*/
      if( shp->src   == clnt_ipaddr.s_addr             &&
	  (shp->sport == clnt_port || shp->sport == 0) &&
	  shp->dst   == inside_ipaddr.s_addr           &&
	  (shp->dport == inside_port || shp->dport == 0))
      {
         in_addr_t	local_ip[MAXPORTS];
	 long	nlips=0;

	 strcpy(tmport,inet_ntoa(clnt_ipaddr));

	 if (islocal(clnt_ipaddr.s_addr) > 0)
	 {
	    syslog(LOG_ALERT,"Security Alert: Detected spoofing of a local"
			     " firewall address.  Src (address=%s,port=%d)"
			     " dst (address=%s,port=%d).  %s exiting.\n",
			     tmport,clnt_port,inet_ntoa(inside_ipaddr),
			     inside_port,pname);
	    closelog();
	    exit(1);
         }

	 if (IPsecure == SECURED)
	 {
	    syslog(LOG_ALERT,"Security Alert: Address spoofing detected. "
	    		     "Src (address=%s,port=%d) dst (address=%s,"
			     "port=%d).  %s exiting.\n",
			     tmport,clnt_port,inet_ntoa(inside_ipaddr),
			     inside_port,pname);
	    say(0,USERALERT);
	 }
	 else
	    syslog(LOG_ALERT,"Security Alert: Attempt to connect to a secured"
	    		     " port from an unprotected network.  Src "
			     "(address=%s,port=%d) dst (address=%s,port=%d)."
			     "  %s exiting.\n",
			     tmport,clnt_port,inet_ntoa(inside_ipaddr),
			     inside_port,pname);
	 closelog();
	 exit(1);
      }
   }
   shmdt((char *)shtop);
   return(0);
#endif /* __sun */
}
