/*============================================================================
 * filename:            libsrc/getlocal.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*======================================================================
 *
 * written by:		R. James Livermore
 *
 * (C) COPYRIGHT R. James Livermore 1996-2000.  All rights reserved.
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1996-2000.
 * All rights reserved.
 *
 *
 * Date written:        June	  18, 1996
 * Last modified:	February   8, 2000 - added LPL
 *
 * functions:
 *		collects a list of all the local IP addresses on the system.
 *
 *=====================================================================*/
#include	<syslog.h>   
#include	<sys/ioctl.h>
#include	<errno.h>
#include	<stdlib.h>
#include	<sys/types.h> 
#ifndef __hppa
#include 	<sys/select.h>		/* to support fd_set 		  */
#endif 
#ifdef __sun
#include	<sys/sockio.h>
#endif
#include	<unistd.h>
#include	<sys/socket.h>
#ifndef __sun
#include	<sys/socketvar.h>
#endif
#include	<netinet/in.h> 
#include	<net/if.h> 
#ifndef	linux
#include	<net/af.h> 
#endif
#include	"spoofmon.h"

#ifndef	linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif
#ifdef IFF_LOOPBACK
#define ISLOOPBACK(p) ((p)->ifr_flags & IFF_LOOPBACK)
#else
#define ISLOOPBACK(p) (strcmp((p)->ifr_name, "lo0") == 0)
#endif

/*=========================================================================
 * Get a list of local IP address.
 * input:
 *	local_ip	array of in_addr_t where ip addresses are returned
 *	portname	ptr to ptr to strings where names of the interfaces are
 *			returned
 *			if NULL, no interface names will be returned and the
 *			"interface not up" message will not be issued.
 *			portname=NULL is intended to be used by islocal()
 *	nlips		number of local ip's returned.
 * returns	0 if successful
 *		1 if failed
 *========================================================================*/
int getlocal(in_addr_t **local_ip,char **portname,long *nlips) 
{  
   int			fd;
   long			i,pid,x; 
   int			n,len,lastlen;
   struct ifconf	ifc;
   struct ifreq		*ifb,ifr,*ifrp,*ifend,*ifnext;

   *local_ip = NULL;
   *nlips = 0;

   /*=======================================================================
    * Get a list of all the local IP address that are UP and are not
    * loopbacks.
    * First get the network config on the system.
    *======================================================================*/
   fd = socket(AF_INET,SOCK_STREAM,0);
#ifdef SIOCGIFNUM
   if (ioctl(fd, SIOCGIFNUM, &n) < 0)
   {
	syslog(LOG_ERR,"Error: SIOCGIFNUM error, %s",sys_errlist[errno]);
	closelog();
	return(1);
   }
   ifc.ifc_len = n * sizeof(struct ifreq);
   if ((ifb = malloc(ifc.ifc_len)) == NULL)
   {
	syslog(LOG_ERR,"Error: malloc() failed for ifreq.  %s",
				sys_errlist[errno]);
	closelog();
	return(1);
   }
   ifc.ifc_buf = (caddr_t)ifb;
   if (ioctl(fd,SIOCGIFCONF,(char *)&ifc) < 0)
   {
	syslog(LOG_ERR,"Error: SIOCGIFCONF error, %s",sys_errlist[errno]);
	closelog();
	return(1);
   }
#else
   lastlen = 0;
   len = 100 * sizeof(struct ifreq);
   for ( ; ; )
   {
	if ((ifb = malloc(len)) == NULL)
	{
	    syslog(LOG_ERR,"Error: malloc() failed for ifreq.  %s",
				sys_errlist[errno]);
	    closelog();
	    return(1);
	}
	ifc.ifc_len = len;
	ifc.ifc_buf = (caddr_t)ifb;
	if (ioctl(fd,SIOCGIFCONF,(char *)&ifc) < 0)
	{
	    if (errno != EINVAL || lastlen != 0)
	    {
		syslog(LOG_ERR,"Error: SIOCGIFCONF error, %s",
						sys_errlist[errno]);
		closelog();
		return(1);
	    }
	}
	else
	{
	    if (ifc.ifc_len == lastlen)
		break;    
	    lastlen = ifc.ifc_len;
	}
	len += 10 * sizeof(struct ifreq);
	free(ifb);
   }
#endif

   ifend = (struct ifreq *)((char *)ifb + ifc.ifc_len);
   for (ifrp=ifb; ifrp<ifend; ifrp=ifnext)
   {
	struct sockaddr *sia = &(ifrp->ifr_addr);
#if !defined(_AIX) && !defined(__Lynx__)
	ifnext = ifrp + 1;
#else
	n = ifrp->ifr_addr.sa_len + sizeof(ifrp->ifr_name);
	if (n < sizeof(*ifrp))
	    ifnext = ifrp + 1;
	else
	    ifnext = (struct ifreq *)((char *)ifrp + n);
#endif
	/*===============================================================
	 * At least on AIX, SIOCGIFCONF returns the link level addresses
	 * also.  We need to filter it out here so we don't add a bogus
	 * IP address to the local list.  BTW, Link level addresses are
	 * returned as sin_family AF_LINK.
	 *==============================================================*/
	if (((struct sockaddr_in *)sia)->sin_family != AF_INET)
	    continue;

	/*====================================================================
	 * Is interface UP and not LOOPBACK?
	 * Need a template to preserve address info that is used below to
	 * locate the next entry.  (Otherwise, SIOCGIFFLAGS stomps over it
	 * because the requests are returned in a union.)
	 *===================================================================*/
	strncpy(ifr.ifr_name, ifrp->ifr_name, sizeof(ifr.ifr_name));
	if (ioctl(fd,SIOCGIFFLAGS,(char *)&ifr) < 0)
	{
	    syslog(LOG_ERR,"Error: SIOCGIFFLAGS error, errno = %d\n",errno);
	    closelog();
	    return(1);
	}
	if ((ifr.ifr_flags & IFF_UP) == 0)
	{
	    if (portname)
		syslog(LOG_NOTICE,"Network interface %s is not up.\n",
			ifr.ifr_name);
	    continue;
	}
	if (ISLOOPBACK(&ifr))
	    continue;

	/*===================================================================
	 * If not on the list yet, add it.
	 *==================================================================*/
	i=0;
	while (i<*nlips && *local_ip &&
		(*local_ip)[i] != ((struct sockaddr_in *)sia)->sin_addr.s_addr)
	    i++;
	if (i>=*nlips)
	{
	    if (portname)
	    {
		if (*nlips == 0)
		    *portname = malloc(IFNAMSIZ);
		else
		    *portname = realloc(*portname, (*nlips + 1) * IFNAMSIZ);
		if (*portname == NULL)
		{
		    syslog(LOG_ERR,"Error: malloc/realloc failed for "
				"portname.  %s", sys_errlist[errno]);
		    closelog();
		    return(1);
		}
		strcpy(*portname + (IFNAMSIZ * *nlips),ifrp->ifr_name);
	    }

	    if (*nlips == 0)
		*local_ip = malloc(sizeof(in_addr_t));
	    else
		*local_ip = realloc(*local_ip,(*nlips + 1) * sizeof(in_addr_t));
	    if (*local_ip == NULL)
	    {
		syslog(LOG_ERR,"Error: malloc/realloc failed for local_ip.  %s",
				sys_errlist[errno]);
		closelog();
		return(1);
	    }

	    (*local_ip)[(*nlips)++] =
				((struct sockaddr_in *)sia)->sin_addr.s_addr;
	}
   }

   return(0);

}


/*=======================================================================
 * Checks to see if an IP address is local on the system.
 *
 * Returns 0 if ip is not local and 1 if ip is local.
 * Returns -1 if it encounters an error getting a list of local addresses.
 *======================================================================*/
int islocal(in_addr_t ip)
{
   in_addr_t	*local_ip;
   long		nlips,i;

   /*=======================================================================
    * Get a list of local IP addresses
    *======================================================================*/
   if (getlocal(&local_ip,NULL,&nlips)) 
   {
	free(local_ip);
	return(-1);
   }

   for (i=0; i<nlips; i++)
   {
	if (ip == local_ip[i]) 
	{
	    free(local_ip);
	    return(1);
	}
   }
   free(local_ip);
   return(0);
}
