/*============================================================================
 * filename:		libsrc/ipsecure.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000. 
 *  All rights reserved. 
 *
 * Date written:	March     27, 1995
 * Last modified:	February   8, 2000 - added LPL
 *
 * function:
 *		The ipsecure function determines  
 * 
 * input:	A pointer to a character string containing IP address
 *		of user's host.
 * output:	Returns an integer:
 * 			 1 the address is secure, or private
 *			 0 the address is NOT secure or private. 
 *			-1 on error condition
 *
 * Logic:	Open securenet file. If the file isn't present then write 	        
 *              error message and consider all addresses to be unsecured.
 *  
 * Warning this function is not to be used by tnproxy since it issues messages.
 *
 *===========================================================================*/
#include	<sys/errno.h> 
#include	<stdio.h> 
#include	<string.h>
#include	<syslog.h> 
#include	"firewall.h"  

#ifndef	linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif 
int nperiods(char *cp);
/*============================================================================
 * Check to see if IP address is on a secure net
 *==========================================================================*/
int ipsecure( char *IP)
{
   return(netcheck(IP,SECURENETFILE));
}  
/*============================================================================
 * Check to see if IP address is on a private net
 *==========================================================================*/
int ipprivate( char *IP)
{
   return(netcheck(IP,PRIVATENETFILE));
}

int netcheck(char *IP, char *secnetfile)
{  
   char		buffer[MAXRECLN+1]; 
   char		*token;
   extern int 	errno;
   int		IPlen;
   int		netlen;
   FILE 	*snfp;  
   int		recnum = 0;
   buffer[MAXRECLN] = '\0';
   /*=================================================================              
    * validate pointer to IP address.                     
    * string length must be greater than 6 (x.x.x.x) and less than 16
    *================================================================*/ 
   IPlen = strlen(IP);
   if( IPlen < 7 )
   {
      syslog(LOG_ERR,"Error: invalid IP address '%s'. Length (%d) is < 7.",
		     IP,IPlen);
      return(-1);
   }
   if( IPlen > 15)
   {
      syslog(LOG_ERR,"Error: invalid IP address '%s'. Length (%d) > 15",
			IP,IPlen);
      return(-1);
   }  
   /*==================================
    * open secure net file 
    *==================================*/
   if (!(snfp = fopen(secnetfile, "r")))
   {
      /*================================================= 
       * privatenets file is optional, so if it doesn't
       * exist, don't report it as an error.
       *================================================*/  
      if (errno != ENOENT || strcmp(secnetfile,PRIVATENETFILE))
         syslog(LOG_ERR,"Error: open failed for %s: %s",
			secnetfile,sys_errlist[errno]);
      return(NOTSECURED);    	/* open failed all addresses are unsecured */
   } 
   /*================================================= 
    * Read the secure network file and look for match 
    *================================================*/  
   while( fgets(buffer, MAXRECLN, snfp) != NULL)
   {   
      recnum++; 
      /* Drop off CRLF, TAB, and skip blank lines */
      if( (token = strtok(buffer," \r\t\n")) == NULL)
         continue;		 
      if( *token == '#')   			/* skip comment lines  	    */	
         continue;
      netlen = strlen(token);			/* skip blank lines	    */
      if( netlen == 0 )
         continue;		
      if( netlen < 2 )
      {
         syslog(LOG_ERR,"Error: IP address '%s' on line %d of file %s is < 2"
                        " characters.",
		        token, recnum, secnetfile);
         fclose(snfp); 
         return(-1);
      } 
      else if( netlen > 15 )
      {
         syslog(LOG_ERR,"Error: ipsecure: IP address '%.21s' on line %d of"
                        " file %s is > 15 characters.",
		        token, recnum, secnetfile); 
         fclose(snfp);
         return(-1);
      } 
      else
      {  
         if( token[netlen-1] == '*' )
         {                    
            netlen--;
            token[netlen] = '\0';
            if(memcmp(IP, token, netlen) == 0)  
            {
               fclose(snfp);
               return(SECURED);
            }
         } 
         else if( token[netlen-1] == '.'  )
         {                    
            if(memcmp(IP, token, netlen) == 0)  
            {
               fclose(snfp);
               return(SECURED);
            }
         }    
         else if( nperiods(token) == 3)
         {  
            if( strcmp(IP, token) == 0 )  
            {
               fclose(snfp);
               return(SECURED); 
            }  
         }
         else
         {  
            syslog(LOG_ERR,"Error: invalid address '%s' specified on line %d"
                           " of file %s.  Address is truncated and does not"
                           " end with '.' or'*'.",
                           token, recnum, secnetfile);

            fclose(snfp);
            return(-1);           
         } 
      }
   }
   if( ferror(snfp) )
   {
      syslog(LOG_ERR,"Error reading record %d in file %s.  %s.",
			recnum, secnetfile, strerror(errno));
      fclose(snfp);
      return(-1);   
   }
   fclose(snfp);   
   return(NOTSECURED); 				/* IP address is unsecured */
}
/*===========================================================================
 * function: count the number of periods '.' in the character string.
 *
 *==========================================================================*/
int nperiods(char *cp)
{       
   char *cp1;
   char *cp2;
   int	i;
   cp1 = cp;
   for( i = 0; i < 4;i++)
   {
     if( (cp2 = strchr(cp1,'.')) == NULL)
        break; 
     cp1 = cp2+1; 
   }
   return(i);   

}
