/*============================================================================
 * filename:		libsrc/peerinfo.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *										
 * written by:		Jim Livermore
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  
 * All rights reserved.   
 *  
 * (C) COPYRIGHT Freemont Avenue Software Laboratories, Inc. 1995-2000. 
 * All Rights Reserved.
 *		
 * Date written:	February   8, 1995
 * Last modified:	February   8, 2000 - added LPL
 *	                         
 * function:
 *	Get the following information regarding the peer host and store it in 
 *	the FTPROXY structure.  IP address, host name, whether it is a secure 
 *	host or not, and wheter it is a private host or not. 
 *
 *	If the host is NOT secured then perform redumentary spoof checking. 
 *	First get the name from the address, then get the address from the name
 * 	and cross-check the two IP addresses.  If the IP addresses do not match then
 *	set the host name to "unknown".  Set all names to lower case.
 *
 *
 * inputs:
 *  	socket file descriptor
 *	ptr to FTPROXY structure. 
 *      
 * outputs:  
 *	Returns a 1 if an error occurs 
 *	copies peer socket from remote end to sname.  
 *============================================================================*/


#include	<sys/types.h>
#include	<sys/socket.h>
#include	<netinet/in.h>
#include	<netdb.h>  		/* hostent struct	*/
#include	<syslog.h>
#include	<ctype.h> 
#include	<sys/errno.h> 
#include	"firewall.h"  
#include	"ftproxy.h"		/* structure common to telnet & ftproxy */
#include	"tnproxy.h"
extern	int	errno;
static	char	unknown[] = "unknown";  
extern	char	*sys_errlist[];
/*================================================================================ 
 * function prototypes 
 *   The Pgethostbyname and Pgethostbyaddr functions use the Internal DNS server
 *   pointed to by /etc/firewall/resolv.inside.conf. 
 *===============================================================================*/
struct hostent *Pgethostbyname(const char *name);
struct hostent * Pgethostbyaddr(const char *addr, int len, int type); 
 
int tnpeerinfo(int fd, struct TNPROXY * proxyptr)
{
 return( peerinfo( fd, (struct FTPROXY *) proxyptr)); 
}
/*==========================================================================
 *inputs:	socket file descriptor
 *		ptr to area to store remote host name
 *		ptr to area to store remote host IP addresss in dot dec fmt
 *		size of area
 *outputs:	stores the following in the proxy control block:
 *		src IP address in dot decimal format, src host name
 *		srcIPsecure status  
 *      	srcIPprivate status
 *		
 *		return  0 if successful                 
 *	        return  1 on error
 *=========================================================================*/ 
int peerinfo(int fd, struct FTPROXY * proxyptr)
{
   struct sockaddr_in	a;
   struct in_addr 	*p_addr;
   struct hostent 	*p;
   int			y;
   int			eq = 0; 
   int			IPsize;
   IPsize = sizeof(proxyptr->sourceIP);
   y = sizeof(a);   
   /*============================================
    * get the address of the peer socket 
    *===========================================*/
   if(getpeername(fd,(struct sockaddr *)&a,(socklen_t *)&y) < 0) 
   { 
      syslog(LOG_ERR,"getpeername failed for fd %d. %s",fd, sys_errlist[errno]);
      return(1);
   }
   /*====================================================================== 
    * Convert remote host IP address to ASCII string in dot decimal format
    * and store in sname. 
    *====================================================================*/  
   strncpy(proxyptr->sourceIP,inet_ntoa(a.sin_addr),IPsize);
   proxyptr->sourceIP[IPsize - 1] = '\0';          
   proxyptr->srcIPsecure = ipsecure(proxyptr->sourceIP);
   /*====================================================================== 
    * If src is from a private net, set the flag in FTPROXY struct
    *====================================================================*/  
   if (ipprivate(proxyptr->sourceIP))
      proxyptr->privateconnect = SRCIP_IS_PRIVATE;
   /*===================================================================== 
    * Get the hostent structure for the remote host.
    * If /etc/resolv.conf exists then use DNS else use /etc/hosts file.
    * If the host is secured or private then use internal DNS else use
    * external DNS.
    *====================================================================*/ 
   if( proxyptr->srcIPsecure == SECURED ||
       proxyptr->privateconnect == SRCIP_IS_PRIVATE)
   {
      p = Pgethostbyaddr((const char *)&a.sin_addr,sizeof(a.sin_addr),AF_INET);
   }
   else
   {
      p = gethostbyaddr((const char *)&a.sin_addr,sizeof(a.sin_addr),AF_INET);
   }

 

   if(p != (struct hostent *)0) 
   {
      char	*x;
      /*==========================================
       * copy host name from hostent structure 
       *=========================================*/
      strncpy(proxyptr->srcname,p->h_name,IPsize);
      proxyptr->srcname[IPsize - 1] = '\0';
      /*===============================================================
       * If the host is secured or private, skip spoof check else use
       * external DNS.
       *================================================================*/ 
      if( proxyptr->srcIPsecure == SECURED ||
          proxyptr->privateconnect == SRCIP_IS_PRIVATE)
      {  
         goto chkIPoptions;  
      }
      else
      {
         p = gethostbyname(proxyptr->srcname);  
      }
      if(p == (struct hostent *)0) 
      {
         syslog(LOG_NOTICE,"%s (%s) host name lookup failed",
                proxyptr->srcname,proxyptr->sourceIP);
	 goto badguy;
      }
      /*==========================================================================
       * Check to see if IP addresses match.
       *=========================================================================*/ 
     while((p_addr = (struct in_addr *)*p->h_addr_list++) != (struct in_addr *)0) 
      {
         if(bcmp(p_addr,(char *)&a.sin_addr,p->h_length) == 0) 
         {
	    eq = 1;
	    break;
	 }
      }
      if(!eq) 
      {
         bcopy(p->h_addr,&a.sin_addr,p->h_length);
	 syslog(LOG_NOTICE,"Security Alert: possible spoof %s (%s) != %s name lookup mismatch",
   		proxyptr->srcname,proxyptr->sourceIP,inet_ntoa(a.sin_addr));
	 goto badguy;
      }
      /*============================================================================
       * If IP options are record Security Alert and nullify options.
       *============================================================================*/
chkIPoptions:
#ifdef IP_OPTIONS
      {
         unsigned char	 optbuf[1024 / 3];
	 unsigned char	 *cp;
	 char		 lbuf[1024];
	 char		 *lp;
	 int		 optsize;
	 int		 ipproto;
	 struct protoent *ip;

	 if((ip = getprotobyname("ip")) != (struct protoent *)0)
	    ipproto = ip->p_proto;
	 else
	    ipproto = IPPROTO_IP;
	 optsize = sizeof(optbuf);
	 if(getsockopt(0,ipproto,IP_OPTIONS,(char *)optbuf,
			(socklen_t *)&optsize) == 0 && optsize != 0) 
         {
	    lp = lbuf;
	    for(cp = optbuf; optsize > 0; cp++, optsize--, lp += 3)
	       sprintf(lp," %2.2x",*cp);
	    syslog(LOG_NOTICE,"Security Alert: Connection received using IP options (ignored):%s",lbuf);
	    if(setsockopt(0,ipproto,IP_OPTIONS,(char *)0,optsize) != 0) 
            {
	       syslog(LOG_NOTICE,"setsockopt IP_OPTIONS NULL: %m");
	       goto badguy;
	    }
         }
      }
#endif
      /* convert all uppercase characters to lower case */
      for(x = proxyptr->srcname; *x != '\0'; x++)
         if(isupper(*x))
	    *x = tolower(*x);
      return(0);
   }
badguy:
   strcpy(proxyptr->srcname,unknown);
   return(0);
}
