/*============================================================================
 * filename:		../libsrc/smwpconf.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  
 * All rights reserved.   
 *		
 * Date written:	March      6, 1995
 * Last modified:	February   8, 2000 - added LPL
 *	                         
 * function:                       
 *	The smwpconf program reads the smwrap configuration file.
 *
 *  
 *
 * smwrap configuration file format: 
 *
 * keyword 	value 		description
 * 
 * aliasreq			alias required for outbound & inbound mail
 * bogushelo    yes|no		should bogus helo (invalid hostname) be
 *				permitted  - see RFC 1123 section 5.2.5
 * timeout 	nnn 		number of seconds before smwrap timeout 
 * groupid      name		groupid that smwrap will run under
 * userid	name		userid that smwrap will run under
 * spoolpath	pathname	spool directory name used by smwrap & smwrapd
 * maxbytes	number		maximum number of bytes in a message
 * maxheadersz  nnnnn		maximum total header size
 * maxreceipts	number		maximum number of recipts for a single message
 * noaliases	po = x.x.x.x	direct in-bound mail to post office server 
 * trace	yes/no/dyn	turn trace on, off or dynamic - default = off
 *				dynamic - print trace only on error condition 
 * translate    on|off          translation of outbound email addr in headers
 *
 * Mail Blocking rules 
 *     
 * deny from src_addr to rcpt_addr [block|forward|sequester]
 *
 *
 * permit from src_addr to rcpt_addr
 *				Intended as override to deny statements.
 *
 *                                                              
 * keywords used by the smwrapd program
 *
 * keyword      value 		description
 *                      
 * maxchildren  nnn		maximum number of child process to spawn
 * smwpdpath	pathname	pathname of the smwrapd executable code
 * smtppath 	pathname	pathname of the SMTP (sendmail) executable
 * undelivpath  pathname	pathname for undeliverable mail
 * wakeup	nnn		number of seconds smwrapd sleeps between scans                                         
 *
 *============================================================================*/
#include	<sys/types.h>
#include	<sys/stat.h> 
#include	<stdio.h>
#include	<stdlib.h>
#include	<syslog.h>
#include	<grp.h> 
#include	<pwd.h>
#include	<string.h>
#include	<sys/time.h>
#include	<netdb.h>               /* defines servent structure      */
#include	<netinet/in.h> 
#include	"firewall.h"
#include	"smwrap.h"

mailblk		*mailgate = NULL;
/*=============================================================================
 * input: 	pointer to SMWRAP record
 * outputs: 	updated SMWRAP record
 *		rc = 0 if ok 
 *		rc = -1 if error
 *===========================================================================*/
extern int getsmwp(struct SMWRAP *smwpptr)  
{   
   #define	MAXLEN 1024
   char 	smwrapcfgfile[256] = SMWRAPCFGFILE;
   char		record[MAXLEN];
   FILE 	*cfgfp; 
   char 	*cp; 
   char		*cp2;
   char		*token1;
   char		*token2; 
   char		tokensep[] 	= "\t,:;= \n";
   int		userfound 	= 0;
   int		groupfound 	= 0;
   int		maxbfound 	= 0;
   int		maxrfound 	= 0;
   int		spoolfound 	= 0;
   int		dpathfound	= 0;
   int		smtppathfound   = 0;
   int		undelivpathfound = 0;
   int		timfound	= 0;
   int		line	  	= 0; 
   int		i; 
   struct passwd	*pw;
   struct group		*gp;  
   struct	stat	sbuf;
   mailblk	*lastmg	= NULL;

   /* If SMWRAP pointer is NUll return with error */
   if( smwpptr == NULL)
   {
      syslog(LOG_ERR,"Error: invalid SMWRAP pointer passed");
      return(-1);
   }
   
   /*========================================================================  
    * Intialize SMWRAP defaults and open proxy config file. 
    *=======================================================================*/ 
   memset(smwpptr,0,sizeof(struct SMWRAP));  
   smwpptr->maxpids   =  8; 		/* default max pid value            */ 
   smwpptr->wakeup    = 60;		/* default wakeup value for smwrapd */
   smwpptr->translate = TRANSLATEON;    /* translate outbound email HDRS    */
   smwpptr->maxheadersz = MAXHEADERSZ;  /* set total max header size        */
   if (!(cfgfp = fopen(smwrapcfgfile, "r")))
   {  
      syslog(LOG_ERR,"Error: open failed for %s: %m",smwrapcfgfile);
      return(-1);    			/* open failed signal error */
   }
   cp = (char *) &record;
   /*================================================= 
    * Read the config file and look for match 
    *=================================================*/        
   while( fgets(record, MAXLEN, cfgfp) != NULL)
   { 
      line++;   
      /* skip over comments and null lines */
      if( (*cp == '#') || (*cp == '\n') ) goto getnextline; 
      token1 = strtok(record,tokensep);

      if (token1 == NULL)
         continue;
     
      /*=========================================================	
       * If userid record then get name and validate 
       *========================================================*/	
      if( token1 && strcmp(token1,"userid") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: No value specified for userid on line %d"
                           " of %s.",
 		line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if( strlen(token2) > 8 )
	 { 
            syslog(LOG_ERR,"Error: User name '%s' is too long on line %d of %s.",
 		token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         }   
	 if(isalldigits(token2))
	 {
	    smwpptr->userid = atoi(token2); 
         }
         else
         {
 	    if((pw = getpwnam(token2)) == (struct passwd *)0)
            {
               syslog(LOG_ERR,"Error: user name '%s' on line %d of %s is not "
                              "valid user.",
 	       token2, line, smwrapcfgfile); 
	       fclose(cfgfp);
	       return(-1);
            } 
	    smwpptr->userid = pw->pw_uid; 		/* save user id	*/ 
         }
         if( smwpptr->userid == 0)
         {
            syslog(LOG_ERR,"Error: user name '%s' on line %d of %s is rejected"
                           " because user id = %d.",
 		   token2, line, smwrapcfgfile, smwpptr->userid); 
	    fclose(cfgfp);
	    return(-1);    
         }
         userfound++;
         goto getnextline;  
      } 
      /*=========================================================	
       * If groupid record then get name and validate 
       *========================================================*/	
      if( strcmp(token1,"groupid") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for groupid on line %d "
                           "of %s.",
 		   line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if( strlen(token2) > 8 )
	 { 
            syslog(LOG_ERR,"Error: Group name '%s' is too long on line %d of %s.",
 		token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         }
         if(isalldigits(token2))
         {
	    smwpptr->groupid = atoi(token2); 
         }
         else
         {
	    if((gp = getgrnam(token2)) == (struct group *)0)
            {
	       syslog(LOG_ERR,"Error: group name '%s' on line %d of %s is not"
                              " valid group.",
 	                      token2, line, smwrapcfgfile); 
	       fclose(cfgfp);
	       return(-1); 
	    }
	    smwpptr->groupid = gp->gr_gid;	/* copy groupid  */ 
         }
         groupfound++;
         goto getnextline;  
      } 
      /*=========================================================	
       * If maxbytes record then get maxbytes value and vaildate 
       *========================================================*/	
      if( strcmp(token1,"maxbytes") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for maxbytes on line %d "
			   "of %s.",
 			   line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if( !isalldigits(token2) )
	 { 
            syslog(LOG_ERR,"Error: non numeric value '%s' specified for "
			   "maxbytes on line %d of %s.",
 			   token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         } 
         smwpptr->maxbytes = atoi(token2); 	/* save maxbytes value	*/ 
         if( smwpptr->maxbytes < 1)
         {
            syslog(LOG_ERR,"Error: maxbytes = '%s' on line %d of %s. Unable "
			   "to deliver mail. Exiting.",
 			   token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
         if( smwpptr->maxbytes > 10000000)
         {
            syslog(LOG_NOTICE,"Caution: maxbytes = '%s' on line %d of %s "
			   "exceeds 10000000. Execution continues.",
 			   token2, line, smwrapcfgfile); 
         }
         maxbfound++;
         goto getnextline;
      }
      /*=====================================================================	
       * If maxheadersz record then get maxheadersz value and vaildate 
       *====================================================================*/	
      if( strcmp(token1,"maxheadersz") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for maxheadersz on line"
			   " %d of %s.",
 			   line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if( !isalldigits(token2) )
	 { 
            syslog(LOG_ERR,"Error: non numeric value '%s' specified for "
			   "maxheadersz on line %d of %s.",
 			   token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         } 
         smwpptr->maxheadersz = atoi(token2); 	/* save maxbytes value	*/ 
         if( smwpptr->maxheadersz < MINMAXHDR)
         {
            syslog(LOG_ERR,"Error: maxheadersz = '%s' on line %d of %s iss less"
			   "than %d.   Exiting.",
 			   token2, line, smwrapcfgfile,MINMAXHDR); 
	    fclose(cfgfp);
	    return(-1); 
         }
         if( smwpptr->maxheadersz > 4*MAXHEADERSZ)
         {
            syslog(LOG_NOTICE,"Caution: maxheadersz = '%s' on line %d of %s "
			   "exceeds %d. Execution continues.",
 			   token2, line, smwrapcfgfile,4*MAXHEADERSZ); 
         }
         maxbfound++;
         goto getnextline;
      }
      
      
      /*===============================================================	
       * If maxreceipts record then get maxrecipts value and validate  
       * RFC 821 specifes max recipients that must be bufferd is 100
       *==============================================================*/	
      if( strcmp(token1,"maxreceipts") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for maxreceipts on line"
			   " %d of %s.",
 			   line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if( !isalldigits(token2) )
	 { 
            syslog(LOG_ERR,"Error: non numeric value '%s' specified for "
			   "maxreceipts on line %d of %s. Exiting.",
 			   token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         } 
         smwpptr->maxreceipts = atoi(token2); 	/* save maxreipts value	*/ 
         if( smwpptr->maxreceipts < 1)
         {
            syslog(LOG_ERR,"Error: maxreceipts = '%s' on line %d of %s. "
			   "Unable to deliver mail. Exiting.",
 			   token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         } 
         if( smwpptr->maxreceipts > 100)
         {
            syslog(LOG_ERR,"Notice: maxreceipts '%s' on line %d of %s exceeds"
			   " RFC 821 req of 100.  Continuing.",
 			   token2, line, smwrapcfgfile); 
         }
         maxrfound++; 
         token1 = 0;
         goto getnextline;
      }                                                                   
      /*=========================================================	
       * If aliasesreq record then set aliasflag
       *========================================================*/	
      if( strcmp(token1,"aliasreq") == 0 )
      {
         if( smwpptr->aliasflag && NOALIAS )
         {
            syslog(LOG_ERR,"Error: aliasreq keyword on line  %d of %s is "
                           "mutually exclusive with noaliases keyword.",
 		line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }  
         smwpptr->aliasflag = ALIASREQ;  
         goto getnextline;
      }
      /*=========================================================	
       * If noaliases record then get IP value and validate 
       *========================================================*/	
      if( strcmp(token1,"noaliases") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: expected 'PO = hostname' to follow "
                           "noaliases on line %d of %s. Found nothing.",
 		line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
         if( strcasecmp(token2,"PO") != 0)
         {
            syslog(LOG_ERR,"Error: Expected 'PO' to follow noaliases "
                           "on line %d of %s.",
 		line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no address after keyword PO "
                           "on line %d of %s.",
 		line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }   
         if( smwpptr->aliasflag && ALIASREQ)
         {
            syslog(LOG_ERR,"Error: noaliases keyword on line  %d of %s is "
                           "mutually exclusive with aliasreq keyword.",
 		line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }  
         smwpptr->aliasflag = NOALIAS;
         strncpy(smwpptr->postoffice,token2,sizeof(smwpptr->postoffice)-1);
         goto getnextline;
      }
       /*=========================================================	
       * If bogushelo record then set bogushelo flag
       *========================================================*/	
      if( strcmp(token1,"bogushelo") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for bogushelo on line"
	    		   " %d of %s.",
 			   line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if( strcmp(token2,"yes")  == 0 ) 
         {
            smwpptr->bogushelo = 1;    
	 }
         else if( strcmp(token2,"no")  == 0 )  
         {
            smwpptr->bogushelo = 0;
	 }
         else
	 { 
            syslog(LOG_ERR,"Error: invalid value for bogushelo '%s' specified"
			   " on line %d of %s.",
 			   token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         }  
         goto getnextline;
      }
      /*=========================================================	
       * If timeout record then get timeout value and validate 
       *========================================================*/	
      if( strcmp(token1,"timeout") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for timeout on line %d "
			   "of %s.",
 			   line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if( !isalldigits(token2) )
	 { 
            syslog(LOG_ERR,"Error: non numeric value '%s' specified for "
			   "timeout on line %d of %s.",
 			   token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         } 
         smwpptr->timeout = atoi(token2); 	/* save timeout value	*/ 
         timfound++;
         goto getnextline;  
      }  
       /*=========================================================	
       * If trace record then get trace value and validate 
       *========================================================*/	
      if( strcmp(token1,"trace") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for trace on line %d of"
			   " %s.",
 			   line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if( strcmp(token2,"yes")  == 0 ) 
         {
            smwpptr->traceflg = 2;    
	 }
         else if( strcmp(token2,"dyn") == 0 )
         {
            smwpptr->traceflg = 1;
	 } 
         else if( strcmp(token2,"no")  == 0 )  
         {
            smwpptr->traceflg = 0;
	 }
         else
	 { 
            syslog(LOG_ERR,"Error: invalid value for trace '%s' specified on "
			   "line %d of %s.",
 			   token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         } 
         goto getnextline;  
      } 
       /*=========================================================	
       * If trace record then get trace value and validate 
       *========================================================*/	
      if( strcmp(token1,"translate") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for translate on line %d"
         		   "of  %s.",
 			   line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if( strcmp(token2,"on")  == 0 ) 
         {
            smwpptr->translate = TRANSLATEON;    
	 }
         else if( strcmp(token2,"off") == 0 )
         {
            smwpptr->translate = TRANSLATEOFF;
	 }
         else
	 { 
            syslog(LOG_ERR,"Error: invalid value for translate '%s' specified"
          		   " on line %d of %s.",
 			   token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         }             
          goto getnextline;  
      } 
      /*=========================================================	
       * If wakeup record then get wakeup value and validate 
       *========================================================*/	
      if( strcmp(token1,"wakeup") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for wakeup on line %d "
			   "of %s.",
 			   line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if( !isalldigits(token2) )
	 { 
            syslog(LOG_ERR,"Error: non numeric value '%s' specified for wakeup"
                           " on line %d of %s.",
 		token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         } 
         smwpptr->wakeup = atoi(token2); 	/* save wakeup value	*/ 
         goto getnextline;  
      }  
      /*=========================================================	
       * If spoolpath record then get path name and validate 
       * do not allow chroot to: "/" or "/etc" or "/usr" 
       *========================================================*/	
      if( strcmp(token1,"spoolpath") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for spoolpath on line %d "
                           "of %s.",
 		line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         } 
         if( *token2 != '/' )
         {
            syslog(LOG_ERR,"Error: spoolpath name '%s' on line %d of %s does "
			   "not start with a '/'.",
 		           token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);               
         }
	 if( strlen(token2) > 200 )
	 { 
            syslog(LOG_ERR,"Error: spoolpath name '%s' on line %d of %s is "
                           "longer than 200.",
 		           token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         }
         if( ( strcmp(token2,"/")    == 0 )     ||
             ( strncmp(token2,"/dev",4) == 0 )  || 
             ( strncmp(token2,"/etc",4) == 0 )  || 
	     ( strncmp(token2,"/lib",4) == 0 )  ||
             ( strncmp(token2,"/sbin",5) == 0 ) || 
	     ( strncmp(token2,"/unix",5) == 0 ) ||
	     ( strncmp(token2,"/usr",4) == 0 )  ||   
	     ( strncmp(token2,"/var",4) == 0 )   ) 
         {
            syslog(LOG_ERR,"Error: spoolpath name '%s' on line %d of %s is not"
			   " allowed.",
 			   token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);    
         }
         strcpy(smwpptr->spoolpath,token2); 	/* copy spoolpath      	*/
         spoolfound++;
         goto getnextline;  
      } 
      /*=========================================================	
       * If permit/deny record then get from/to addresses
       *========================================================*/	
       if( strcmp(token1,"permit") == 0 ||
		strcmp(token1,"deny") == 0 )
      {
          char *at;
          /*=====================================================
           * malloc a new structure and append it at the end of
           * the chain
           *====================================================*/
          if (lastmg) 
	  {
              lastmg->next = malloc(sizeof(mailblk));
              lastmg = lastmg->next;
          }
          else 
          {
              lastmg = malloc(sizeof(mailblk));
              mailgate = lastmg;
          }
          if (lastmg == NULL) 
          {
             syslog(LOG_ERR,"Error: malloc failed for mailblk "
				"structure.");
             fclose(cfgfp);
             return(-1);
          }
          memset(lastmg,'\0',sizeof(mailblk));
          if (strcmp(token1,"permit") == 0)
             lastmg->status = MB_PERMIT;
          else
             lastmg->status = MB_BLOCK;	/* default all deny's to block */
          /*=====================================================
           * Get the from value
           *====================================================*/
          if ((token2 = strtok(NULL,tokensep)) == NULL ||
			strcasecmp(token2,"from") != 0) 
          {
             syslog(LOG_ERR,"Error: permit/deny not followed by \"from\" on "
			"line %d of %s.",line,smwrapcfgfile);
             fclose(cfgfp);
             return(-1);
          }
          if ((token2 = strtok(NULL,tokensep)) == NULL) 
          {
             syslog(LOG_ERR,"Error: Missing from address on line %d of %s.",
			line,smwrapcfgfile);
             fclose(cfgfp);
             return(-1);
          }
          /*=====================================================
           * Malloc for the from address
           *====================================================*/
          if ((lastmg->fromuser = malloc(strlen(token2)+1)) == NULL) 
	  {
             syslog(LOG_ERR,"Error: malloc failed for mailblk from address.");
             fclose(cfgfp);
             return(-1);
          }
          strcpy(lastmg->fromuser,token2);
          /*=====================================================
           * If no '@' found, we are done with the from address
           * If '@' found, then we need to setup fromhost
           *====================================================*/
          if ((at = strchr(lastmg->fromuser,'@')) != NULL) 
          {
             if (at == lastmg->fromuser) 
             {
                /*==================================================
                 * If at == token2, then we have a from address
                 * specified as "@hostname" without a username
                 *=================================================*/
                lastmg->fromhost = lastmg->fromuser + 1;
                lastmg->fromuser = NULL;
             }
             else 
             {
                /*==================================================
                 * We have user@hostname.  We terminate the string
                 * at the '@' sign and fromhost will be next to it.
                 *=================================================*/
                *at = '\0';
                lastmg->fromhost = ++at;
             }
          }
          /*=====================================================
           * If the fromuser or fromhost consist of exactly one '*'
           * then null the field to indicate it is a wildcard
           *====================================================*/
          if (lastmg->fromuser && !strcmp(lastmg->fromuser,"*"))
             lastmg->fromuser = NULL;
          if (lastmg->fromhost && !strcmp(lastmg->fromhost,"*"))
             lastmg->fromhost = NULL;
          /*=====================================================
           * Get the to value
           *====================================================*/
          if ((token2 = strtok(NULL,tokensep)) == NULL ||
			strcasecmp(token2,"to") != 0) 
          {
             syslog(LOG_ERR,"Error: Missing \"to\" on "
			"line %d of %s.",line,smwrapcfgfile);
             fclose(cfgfp);
             return(-1);
          }
          if ((token2 = strtok(NULL,tokensep)) == NULL) 
          {
             syslog(LOG_ERR,"Error: Missing to address on line %d of %s.",
			line,smwrapcfgfile);
             fclose(cfgfp);
             return(-1);
          }
          /*=====================================================
           * Malloc for the to address
           *====================================================*/
          if ((lastmg->touser = malloc(strlen(token2)+1)) == NULL) 
          {
             syslog(LOG_ERR,"Error: malloc error allocating to address.");
             fclose(cfgfp);
             return(-1);
          }
          strcpy(lastmg->touser,token2);
          /*=====================================================
           * If no '@' found, we are done with the to address
           * If '@' found, then we need to setup tohost
           *====================================================*/
          if ((at = strchr(lastmg->touser,'@')) != NULL) 
          {
             if (at == lastmg->touser) 
             {
                /*==================================================
                 * If at == token2, then we have a to address
                 * specified as "@hostname" without a username
                 *=================================================*/
                lastmg->tohost = lastmg->touser + 1;
                lastmg->touser = NULL;
             }
             else 
             {
                /*==================================================
                 * We have user@hostname.  We terminate the string
                 * at the '@' sign and tohost will be next to it.
                 *=================================================*/
                *at = '\0';
                lastmg->tohost = ++at;
             }
          }
          /*=====================================================
           * If the touser or tohost consist of exactly one '*'
           * then null the field to indicate it is a wildcard
           *====================================================*/
          if (lastmg->touser && !strcmp(lastmg->touser,"*"))
             lastmg->touser = NULL;
          if (lastmg->tohost && !strcmp(lastmg->tohost,"*"))
             lastmg->tohost = NULL;
          /*=====================================================
           * Process optional forward/sequester keywords
           *====================================================*/
          if ((token2 = strtok(NULL,tokensep)) != NULL) 
          {
             if (lastmg->status == MB_PERMIT) 
             {
                syslog(LOG_ERR,"Error: Argument \"%s\" is not valid on a "
			"\"permit\" statement on line %d of %s.",
			token2,line,smwrapcfgfile);
                fclose(cfgfp);
                return(-1);
             }
             if (strcasecmp(token2,"forward") == 0) 
                {
                lastmg->status = MB_FORWARD;
                if ((token2 = strtok(NULL,tokensep)) == NULL) 
                {
                   syslog(LOG_ERR,"Error: \"forward\" keyword must be followed "
				"by a forward-to address on line %d of %s.",
				line,smwrapcfgfile);
                   fclose(cfgfp);
                   return(-1);
                }
                if ((lastmg->forw = malloc(strlen(token2)+1)) == NULL) 
                {
                   syslog(LOG_ERR,"Error: malloc failed for mailblk forward "
				" address.");
                   fclose(cfgfp);
                   return(-1);
                }
                strcpy(lastmg->forw,token2);
             }
             else if (strcasecmp(token2,"sequester") == 0) 
             {
                lastmg->status = MB_SEQUESTER;
             }
             else if (strcasecmp(token2,"block") == 0) 
             {
                lastmg->status = MB_BLOCK;
             }
             else 
             {
                syslog(LOG_ERR,"Error: Invalid argument \"%s\" on line %d of "
			"%s.",token2,line,smwrapcfgfile);
                fclose(cfgfp);
                return(-1);
             }
         }
         goto getnextline;
      }
      /*=========================================================	
       * If maxchildren record then get number and validate 
       *========================================================*/	
      if( strcmp(token1,"maxchildren") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for maxchildren on line"
			   " %d of %s.",
 			   line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if( !isalldigits(token2) )
	 { 
            syslog(LOG_ERR,"Error: non numeric value '%s' specified for "
			   "maxchildren on line %d of %s.",
 			   token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         } 
         smwpptr->maxpids = atoi(token2); 	/* save maxchildren value */ 
         if( smwpptr->maxpids < 1)
         {
            syslog(LOG_ERR,"Error: maxchildren = '%s' on line %d of %s is too"
			   " small. Unable to deliver mail. Exiting.",
 			   token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
         if(smwpptr->maxpids > 12)
         {
            syslog(LOG_ERR,"Warning maxchildren =  '%d' is > 12. See line %d "
			   "of %s.",
 			   smwpptr->maxpids, line, smwrapcfgfile); 
         } 
         token1 = 0;               /* reset token ptr */
         goto getnextline;
      } 
      /*=========================================================	
       * If smwpdpath record then get path name and validate 
       *========================================================*/	
      if( strcmp(token1,"smwpdpath") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for smwpdpath on line "
			   "%d of %s.",
 			   line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if( strlen(token2) > 255 )
	 { 
            syslog(LOG_ERR,"Error: smwpdpath name '%s' is too long on line %d "
			   "of %s.",
 			   token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         } 
         /* check for file existence */ 
         if( stat(token2,&sbuf) )
         {
            syslog(LOG_ERR,"Error: smwpdpath name '%s' on line %d of %s not "
			   "found.",
 			   token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);  
         }
         strcpy(smwpptr->smwpdpath,token2); 	/* copy smwpd path     	*/
         dpathfound++;
         goto getnextline;  
      }  
      /*=========================================================	
       * If smtppath record then get path name and validate 
       *========================================================*/	
      if( strcmp(token1,"smtppath") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: no value specified for smtppath on line %d "
			   "of %s.",
 			   line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if( strlen(token2) > 255 )
	 { 
            syslog(LOG_ERR,"Error: smtppath name '%s' is too long on line %d "
			   "of %s.",
 			   token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         }
         /* check for file existence */ 
         if( stat(token2,&sbuf) )
         {
            syslog(LOG_ERR,"Error: smtppath name '%s' on line %d of %s not "
			   "found.",
 			   token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);  
         }
         strcpy(smwpptr->smtppath,token2); 	/* copy smtp path     	*/
         smtppathfound++;
         goto getnextline;  
      } 
      /*=========================================================	
       * If undelivpath record then get path name and validate 
       *========================================================*/	
      if( strcmp(token1,"undelivpath") == 0 )
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL)
         {
            syslog(LOG_ERR,"Error: No value specified for undelivpath on line"
			   " %d of %s.",
 			   line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1); 
         }
	 if( strlen(token2) > 200 )
	 { 
            syslog(LOG_ERR,"Error: undelivpath name '%s' on line %d of %s is"
			   " longer than 200.",
 			   token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);       
         }
          /* check for directory existence */ 
         if( stat(token2,&sbuf) )
         {
            syslog(LOG_ERR,"Error: undelivpath name '%s' on line %d of %s not"
			   " found.",
 			   token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);  
         } 
         strcpy(smwpptr->undelivpath,token2); 	/* copy undeliv path   	*/  
         if( strncmp(smwpptr->spoolpath,smwpptr->undelivpath,
                     strlen(smwpptr->spoolpath)) )
         {
            syslog(LOG_ERR,"Error: undelivpath name '%s' on line %d of %s is "
                           "not a sub directory of spoolpath",
 			   token2, line, smwrapcfgfile); 
	    fclose(cfgfp);
	    return(-1);      
         }
         undelivpathfound++;
         goto getnextline;  
      } 

      /*==================================================================
       * Unknown keyword!
       *=================================================================*/
      syslog(LOG_ERR,"Error: invalid record on line %d of %s, unknown keyword"
		     " '%s'.",
		     line, smwrapcfgfile, token1);
      fclose(cfgfp);
      return(-1);
       
      getnextline:;   
   } 
   /*================================================================= 
    * now check to make certain all variables were supplied:
    *  userfound, groufound, maxbfound, maxrfound, dirfound, timfound
    *================================================================*/
   
   if( userfound != 1)
   {
      syslog(LOG_ERR,"Error: No userid record found in %s.",
		     smwrapcfgfile);
      fclose(cfgfp);
      return(-1);
   }
   if( groupfound != 1)
   {
      syslog(LOG_ERR,"Error: No groupid record found in %s.",
		     smwrapcfgfile);
      fclose(cfgfp);
      return(-1);
   }
   if( maxbfound != 1)
   {
      syslog(LOG_ERR,"Error: No maxbytes record found in %s.",
		     smwrapcfgfile);
      fclose(cfgfp);
      return(-1);
   } 
   if( maxrfound != 1)
   {
      syslog(LOG_ERR,"Error: No maxreceipts record found in %s.",
		     smwrapcfgfile);
      fclose(cfgfp);
      return(-1);
   } 
   if( timfound != 1)
   {
      syslog(LOG_ERR,"Error: No timeout record found in %s.",
		     smwrapcfgfile);
      fclose(cfgfp);
      return(-1);
   } 
     
   if( spoolfound != 1)
   {
      syslog(LOG_ERR,"Error: No spoolpath record found in %s.",
		     smwrapcfgfile);
      fclose(cfgfp);
      return(-1);
   } 
   
   if( dpathfound != 1)
   {
      syslog(LOG_ERR,"Error: No smwpdpath record found in %s.",
		     smwrapcfgfile);
      fclose(cfgfp);
      return(-1);
   } 
   if( smtppathfound != 1)
   {
      syslog(LOG_ERR,"Error: No smtppath record found in %s.",
		     smwrapcfgfile);
      fclose(cfgfp);
      return(-1);          
   }   
   if( undelivpathfound != 1)
   {
      syslog(LOG_ERR,"Error: No undelivpath record found in %s.",
		     smwrapcfgfile);
      fclose(cfgfp);
      return(-1);
   }
   fclose(cfgfp);
   return(0);  			/* everything is ok */

} 
