/*============================================================================
 * file: 	netlib/inet_addr.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *                              
 * modified by:		R. James Livermore
 *
 * Last modified:	February   8, 2000 - added LPL
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  All rights
 * reserved.  
 *
 * This program is a modification of the BSD inet_addr.c program. 
 * This version looks at the /etc/firewall/resolv.inside.conf file to find the
 * Internal DNS.     
 * 
 *============================================================================
 *
 * Copyright (c) 1983, 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *=======================================================================*/
static char	cprght[] = 
 "(C) COPYRIGHT Freemont Avenue Software, Inc. 1995-2000 inet_addr, V 1.0 00/01/21";
#if defined(LIBC_SCCS) && !defined(lint)
static char sccsid[] = "@(#)inet_addr.c	5.11 (Berkeley) 12/9/91";
#endif /* LIBC_SCCS and not lint */

#include <sys/param.h> 
#include <sys/types.h>		/* added by Jim Livermore		*/
#include <netinet/in.h>
#ifdef __sun
#define INADDR_NONE	((u_long)0xffffffff)	/* -1 return */
#endif
#include <ctype.h>
/* function prototypes */
unsigned long  inet_aton( register const char *cp, struct in_addr *addr); 
unsigned long inet_addr(register const char *);

/*========================================================================
 * Ascii internet address interpretation routine.
 * The value returned is in network order.
 *========================================================================*/
u_long
inet_addr(cp)
	register const char *cp;
{
   struct in_addr val;

   if (inet_aton(cp, &val))
      return (val.s_addr);
   return (INADDR_NONE);
}

/*============================================================================= 
 * Check whether "cp" is a valid ascii representation
 * of an Internet address and convert to a binary address.
 * Returns 1 if the address is valid, 0 if not.
 * This replaces inet_addr, the return value from which
 * cannot distinguish between failure and a local broadcast address.
 *============================================================================*/
unsigned long  inet_aton( register const char *cp, struct in_addr *addr) 
{
   register u_long val, base, n;
   register char c;
   u_long parts[4], *pp = parts;

   for (;;) 
   {
		/*
		 * Collect number up to ``.''.
		 * Values are specified as for C:
		 * 0x=hex, 0=octal, other=decimal.
		 */
		val = 0; base = 10;
		if (*cp == '0') {
			if (*++cp == 'x' || *cp == 'X')
				base = 16, cp++;
			else
				base = 8;
		}
		while ((c = *cp) != '\0') {
			if (isascii(c) && isdigit(c)) {
				val = (val * base) + (c - '0');
				cp++;
				continue;
			}
			if (base == 16 && isascii(c) && isxdigit(c)) {
				val = (val << 4) + 
					(c + 10 - (islower(c) ? 'a' : 'A'));
				cp++;
				continue;
			}
			break;
		}
		if (*cp == '.') {
			/*
			 * Internet format:
			 *	a.b.c.d
			 *	a.b.c	(with c treated as 16-bits)
			 *	a.b	(with b treated as 24 bits)
			 */
			if (pp >= parts + 3 || val > 0xff)
				return (0);
			*pp++ = val, cp++;
		} else
			break;
	}
	/*
	 * Check for trailing characters.
	 */
	if (*cp && (!isascii(*cp) || !isspace(*cp)))
		return (0);
	/*
	 * Concoct the address according to
	 * the number of parts specified.
	 */
	n = pp - parts + 1;
	switch (n) {

	case 1:				/* a -- 32 bits */
		break;

	case 2:				/* a.b -- 8.24 bits */
		if (val > 0xffffff)
			return (0);
		val |= parts[0] << 24;
		break;

	case 3:				/* a.b.c -- 8.8.16 bits */
		if (val > 0xffff)
			return (0);
		val |= (parts[0] << 24) | (parts[1] << 16);
		break;

	case 4:				/* a.b.c.d -- 8.8.8.8 bits */
		if (val > 0xff)
			return (0);
		val |= (parts[0] << 24) | (parts[1] << 16) | (parts[2] << 8);
		break;
	}
	if (addr)
		addr->s_addr = htonl(val);
	return (1);
}
