/*============================================================================
 * filename:		netlib/presolv.h
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*===========================================================================
 *
 * written by:		R. James Livermore
 *
 * Last modified:	February   8, 2000 - added LPL
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  All rights
 * reserved.  
 *
 * This program is a modification of the BSD resolv.h header file.  
 * This version uses the /etc/firewall/resolv.inside.conf file to find the
 * Internal DNS.   
 *
 *
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 *
 * (c) Copyright 1990, 1991, 1992 OPEN SOFTWARE FOUNDATION, INC. 
 * ALL RIGHTS RESERVED 
 *===========================================================================*/

#ifndef _H_RESOLV
#define _H_RESOLV

/*=============================================================================
 * Resolver configuration file.
 * Normally not present, but may contain the address of the
 * inital name server(s) to query and the domain search list.
 * "/etc/resolv" has been replaced with "/etc/firewall/resolv.inside.conf"
 *===========================================================================*/

#ifndef _PATH_RESCONF
#define _PATH_RESCONF        "/etc/firewall/resolv.inside.conf"
#endif

/*=============================================================================
 * Global defines and variables for resolver stub.
 *===========================================================================*/
#define	MAXNS		3  		/* max # name servers we'll track    */
#define	MAXDFLSRCH	3  		/* # default domain levels to try    */
#define	MAXDNSRCH	6		/* max # domains in search path      */
#define	LOCALDOMAINPARTS 2		/* min levels in name that is "local"*/

#define	RES_TIMEOUT	5		/* min. seconds between retries      */

struct state {
	int	retrans;	 	/* retransmition time interval 	     */
	int	retry;			/* number of times to retransmit     */
	long	options;		/* option flags - see below.         */
	int	nscount;		/* number of name servers            */
	struct	sockaddr_in nsaddr_list[MAXNS];	/* address of name server    */
#define	nsaddr	nsaddr_list[0]		/* for backward compatibility        */
	u_short	id;			/* current packet id                 */
	char	defdname[MAXDNAME];	/* default domain                    */
	char	*dnsrch[MAXDNSRCH+1];	/* components of domain to search    */
};

/*
 * Resolver options
 */
#define RES_INIT	0x0001		/* address initialized 		*/
#define RES_DEBUG	0x0002		/* print debug messages 	*/
#define RES_AAONLY	0x0004		/* authoritative answers only 	*/
#define RES_USEVC	0x0008		/* use virtual circuit 		*/
#define RES_PRIMARY	0x0010		/* query primary server only 	*/
#define RES_IGNTC	0x0020		/* ignore trucation errors 	*/
#define RES_RECURSE	0x0040		/* recursion desired 		*/
#define RES_DEFNAMES	0x0080		/* use default domain name 	*/
#define RES_STAYOPEN	0x0100		/* Keep TCP socket open 	*/
#define RES_DNSRCH	0x0200		/* search up local domain tree  */

#define RES_DEFAULT	(RES_RECURSE | RES_DEFNAMES | RES_DNSRCH)

extern struct state _res;

extern char *p_cdname(), *p_rr(), *p_type(), *p_class(), *p_time();

#ifdef _NO_PROTO
#ifdef _THREAD_SAFE
extern int res_init_r();
extern int res_mkquery_r();
extern int res_query_r();
extern int res_search_r();
extern int res_querydomain_r();
extern int res_send_r();
#endif /* _THREAD_SAFE */

#else /* _NO _PROTO */

#ifdef	_THREAD_SAFE
extern int res_init_r(struct state *st_ptr);
extern int res_mkquery_r(int op, char *dname, int class, int type, char *data, int datalen, struct rrec *newrr, char *buf, int buflen, struct state *st_ptr);
extern int res_query_r(char *name, int class, int type, u_char *answer, int anslen, struct state *st_ptr);
extern int res_search_r(char *name, int class, int type, u_char *answer, int anslen, struct state *st_ptr);
extern int res_querydomain_r(char *name, char *domain, int class, int type, u_char *answer, int anslen, struct state *st_ptr);
extern int res_send_r(char *buf,int buflen, char *answer, int anslen, int *socket_ptr, struct state *st_ptr);
#endif /* _THREAD_SAFE */

#endif /* _NO_PROTO */

#ifdef _THREAD_SAFE
extern int	hostalias_r(char *name, char *abuf, int len);
#else
extern char	*hostalias(char *name);
#endif /* _THREAD_SAFE */

#endif /* _H_RESOLV */
