/*============================================================================
 * filename:		oobasrv/oobasrv.c					
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *										
 * written by:		Jim Livermore
 * 
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.
 *
 * (C) COPYRIGHT Freemont Avenue Software Laboratories, Inc. 1995-2000.
 * All rights reserved. 
 *  
 * Date written:	August	   1, 1995
 * Last modified:	February   8, 2000 - added LPL
 *					  
 *	                         
 * function:                       
 *	
 *	The Out Of Band Authentication Server runs under the control of  
 *	the inetd. oobasrv receives control when a genproxy server on a remote 
 *	host needs to authenticate this host before providing the proxy service.
 *	
 *	
 *  
 *	Debug option
 *                     
 * 	Normally oobasrv will be invoked by inetd. However it can be invoked 
 *	at the command line with the Debug option. To use the debug option
 * 	type in the program name, -D and an optional port number
 *	used for the test. If the optional port is not specified the 
 *	program will use port number 7777. 
 *
 *		oobasrv -D 7777
 *       
 *
 *============================================================================*/

#include	<stdio.h>
#include	<syslog.h>
#include	<sys/errno.h> 
#include	<string.h>
#include	<sys/types.h>
#include	<sys/time.h>  
#include	<sys/times.h>           /* cpu times			  */
#ifndef __hppa
#include 	<sys/select.h>		/* to support fd_set 		  */ 
#endif
#include	<sys/socket.h>
#include	<netdb.h>               /* defines servent structure      */
#include	<netinet/in.h> 

#include	"des.h"
/*===========================================================================
 * external variables
 *==========================================================================*/
static	int		portid = -1; 

extern	int	errno;
#ifndef linux
extern	char	*sys_errlist[];
#else
extern	const	char	*const	sys_errlist[];
#endif

/*============================================================================
 * function prototypes
 *===========================================================================*/ 
int  bldresponse(char *challenge,char *key,char *response);   
void debugbind(unsigned short debugport);
int  isalldigits(char *s);   
int  oobahosts(char *IP);
int  oobakey(char *key);
int  peername(int fd, char *lname,char *sname,int z);
/*============================================================================
 * main function
 *===========================================================================*/
main(int ac,char *av[])
{ 
   static char ident[] =
              "(C) COPYRIGHT Freemont Avenue Software, Inc. 1995-2000. "
	      "oobasrv V 1.0, 00/01/21"; 
   long			tempport;
   unsigned short	debugport = 7777; 	/* default debug port	*/ 

   char		lipaddr[20]; 		/* local IP address	*/ 
   char	timeout[8]	= "\0\0\0\0\0\0\0\0";	/* timeout in seconds 	*/ 
   struct sockaddr_in	la; 
   char			inbuf[128];
   char			outbuf[128];
   int			br 	= 0;            /* bytes read		*/
   int			bw	= 0;		/* bytes written	*/ 
   char			key[32]; 
   char			rhostname[256];
   char			rhostIP[24];

   int			siz;  
   char			*token1;
   fd_set		redy;
   fd_set		iced; 
   /*======================================================================
    * Open the syslog and issue start message.
    *=====================================================================*/
   openlog("oobasrv",LOG_PID|LOG_NDELAY,LOG_DAEMON); 
   syslog(LOG_NOTICE,"OOBA V 1.0 started.");
   /*=====================================================================
    *	If Debug is specified this it will bind the port and listen  
    *	directly (circumventing inetd) so you can run the program under
    *	a debugger
    *=====================================================================*/
   if( ac > 1)
   {
      if( strcmp(av[1],"-D") == 0)
      {  
	 /* check for debug port number  */
         if( ac > 2 )
         {
	    if( (strlen(av[2]) < 6)  && isalldigits(av[2]) )
            {
               if( (tempport = atoi(av[2])) > 64000)
               {
                  printf("\noobasrv: portnumber > 64000. Please try again.\n"); 
		  close(0);
	          exit(1);
               }
               debugport = tempport;
            }
            else
            {
               printf("\noobasrv: invalid portnumber (%s), must be a positive "
			"integer less than 64000."
		       "\n          Please try again.\n", av[2]); 
               close(0);
               exit(1); 
            }   
         }
         printf("\n oobasrv: Debug mode specified. Use port %d.\n",
                 debugport);
         debugbind(debugport); 
      }
   }    
   /*=========================================================
    * get remote hostname and IP address in dot decimal format 
    *========================================================*/
   if(peername(0,rhostname,rhostIP,sizeof(rhostname))) 
   {
      syslog(LOG_NOTICE,"Can not get remote host: %m");  
      close(0);
      exit(1);
   } 
   /*========================================================================
    * Check to see if this is a host we are willing to talk to.
    *========================================================================*/
   if( oobahosts(rhostIP) != 1)
   {     
      syslog(LOG_ALERT,"Security Alert: host %s (%s) tried to connect to"
      		       " OOBASRV",
                       rhostname, rhostIP); 
      close(0);
      exit(1); 
   } 
   /*=========================================================================
    * Get key used to generate response.
    *========================================================================*/
   if( oobakey(key) != 0)
   {          
      close(0);
      exit(1); 
   } 
   /*=====================================================================
    * Log successful connection.
    *=====================================================================*/
   syslog(LOG_NOTICE,"connect src_host = %s (%s) ",
		  rhostname, rhostIP);
   /*================================================
    * read challenge from genproxy
    *===============================================*/
   if((br = read(0,inbuf,sizeof(inbuf))) < 1 ) 
   {  
      if( br == 0 )
      {
         syslog(LOG_NOTICE,"Error: unexpected EOF.");
      }
      /* if x < 0 we have a read error	*/ 
      syslog(LOG_NOTICE,"Error: while reading from source host.");   
      close(0);
      exit(1);
   }  
   token1 = strtok(inbuf,"\n\r\t ");
   if( strlen(token1) < 8 )
   { 
      syslog(LOG_NOTICE,"Error: challenge '%s' is less than 8 bytes.",
		        inbuf);  
      strcpy(outbuf,"Good Bye.");
      write(0,outbuf,strlen(outbuf)); 
      close(0);
      exit(1);
   } 
   if( isalldigits(token1) != 1 )
   { 
      syslog(LOG_NOTICE,"Error: challenge '%s' is not all digits.",
		        token1);  
      strcpy(outbuf,"Good Bye.");
      write(0,outbuf,strlen(outbuf)); 
      close(0);
      exit(1);
   }
   /*================================================
    * create response by encrypting challenge
    *===============================================*/
   bldresponse(token1,key,outbuf);  
   /*================================================
    * write response back to genproxy
    *===============================================*/
   if((bw = write(0,outbuf,br)) != br) 
   {  
      syslog(LOG_NOTICE,"Error: write to server returned %d expected %d.",
				bw,br);
   } 
   /*===================================================
    * read message to see if response was accepted.
    *==================================================*/ 
   *inbuf = '\0';
   if((br = read(0,inbuf,sizeof(inbuf))) < 1 ) 
   {  
      if( br == 0 )
      {
         syslog(LOG_NOTICE,
	 "Error: unexpected EOF trying to read acceptance message from %s (%s)",
	        rhostIP, rhostname);
      }
      /* if x < 0 we have a read error	*/ 
      syslog(LOG_NOTICE,"Error: reading acceptance message from %s (%s).",
		rhostIP, rhostname);   
      close(0);
      exit(1);
   }
    syslog(LOG_NOTICE,"OOBA acceptance message from %s (%s): '%s'.",
		rhostIP, rhostname,inbuf);     
   close(0);
   exit(0);
}
/*==================================================================
 * function: 	bldresponse
 * inputs:	pointer to a challenge
 *		pointer to key  "nnn nnn nnn nnn nnn nnn nnn nnn"
 *				where nnn are octal digits
 *		pointer to response buffer 
 * outputs:	 
 *		updated response buffer if ok
 *	returns  0 if ok
 *		-1 on error 
 *==================================================================*/  
int bldresponse(char *challenge,char *key,char *response)  
{
   des_key_schedule	keysched;
   des_cblock		kblock;
   char			buf[12];
   char			cbuf[12]; 
   int			i;
   int			j;
   unsigned long	kval;
   int			temp;
   char			*token1;  

   token1 = strtok(key," ");
   /*=================================================================
    * Convert each three digit octal number into a character and store
    * it in the kblock. 
    *=================================================================*/
   for( j = 0; j < 8; j++)  
   {  
      
      sscanf(token1,"%o",&temp);
      kblock[j] = temp; 
      token1 = strtok(NULL," ");      	/* get next 3 digit octal number */

   }
   strncpy(buf,challenge,9);
   buf[9] = '\0';  
   des_set_key(&kblock,keysched);
   /*======================================================================== 
    * Excrypt the challenge using the key schedule  	
    *=======================================================================*/
   des_ecb_encrypt((des_cblock *) buf, (des_cblock *) cbuf,keysched,DES_ENCRYPT);
      /* Extract the bits out of the ciphertext into a long */
   for(i=0; i<4; i++)
      for(j = 0; j < 8; j++)
         kval = (kval << 1) | ((cbuf[i] >> (7 - j)) & 1);
   /*=================================================================
    * convert long into a hexadecimal string  
    *=================================================================*/
   sprintf(response,"%08x",kval);
   
   return(0);
} 
/*==================================================================
 * function: 	isalldigits
 * input:	a pointer to a character string 
 * output:	returns 1 if all the characters in the string are digits 
 *==================================================================*/  
int isalldigits(char *s)  
{
   while(isdigit(*s))
      s++;
   return(*s == '\0');
} 
/*===========================================================================
 * debugbind 
 * input:	debug port number
 *==========================================================================*/
void debugbind(unsigned short debugport)
{
   struct	sockaddr_in	mya;
   int	x;
   int	nread;

   if((x = socket(AF_INET,SOCK_STREAM,0)) < 0) 
   {
      perror("socket");
      exit(1);
   }
   mya.sin_family = AF_INET;
   bzero(&mya.sin_addr,sizeof(mya.sin_addr));
   mya.sin_port = htons(debugport);
   if(bind(x,(struct sockaddr *)&mya,sizeof(mya))) 
   {
      perror("bind");
      exit(1);
   }
   if(listen(x,1) < 0) 
   {
      perror("listen");
      exit(1);
   }
   if((nread = accept(x,0,0)) < 0) 
   {
      perror("accept");
      exit(1);
   }
   close(0);
   dup(nread);
   close(1);
   dup(nread);
}

