/*============================================================================
 * filename:		padmin/backup.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1998-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1998-2000.  
 * All rights reserved.   
 *
 * Date written:	June	  24, 1998
 * Last modified:	February   8, 2000 - added LPL
 *
 *===========================================================================*/
#include <stdio.h>
#include <errno.h>
#include <syslog.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifndef __hppa
#include <sys/select.h>
#endif
#include "firewall.h"
#include "gwuser.h"
#include "des.h"
#include "cs.h"
#include "enc.h"
#include "cbc_io.h"
#include "tnkey.h"

#ifndef linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif
extern RESPPKT error_resp;
extern des_key_schedule ckeysched;
extern des_key_schedule dkeysched;
extern PACONF *padmconf;
extern  struct sockaddr_in addr;                /* Remote host's address */
extern struct GWADMIN admin;

#define TMPDIR FW_CFG_DIR "/tmp"
#define TREXLIST FW_CFG_DIR "/trex.bulist"
#define APACHELIST FW_CFG_DIR "/apache.bulist"
#define SYSTEMLIST FW_CFG_DIR "/system.bulist"
#define CUSTOMLIST FW_CFG_DIR "/custom.bulist"

int getdata(int fd, char *buf, int len, des_key_schedule keysched);
int putdata(int fd, char *buf, int len, des_key_schedule keysched);

int backup(int fd, int msglen) {
    int rc;
    int ifd;
    int nread;
    int total;
    char *fn;
    char buf[BUFSIZ];
    int flag = 0;
    struct stat stat;
    SCRESP resp;
    CBCBLK cbcblk;

    /*=======================================================================
     * Make sure the msg fits into our buffer
     *======================================================================*/
    if (msglen > sizeof(buf)) {
	strcpy(error_resp.msg,"Flag too long");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    if (msglen == 0)
	flag = TYPICAL_BU | CUSTOM_BU;
    else {
	/*==================================================================
	 * Read in the request to get the flag
	 *=================================================================*/
	if ((rc = getdata(fd, buf, msglen, ckeysched)) < 0) {
	    strcpy(error_resp.msg,"padmin backup cannot read flag value");
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    return(NONFATAL);
	}

	if (rc == 0) {
	    syslog(LOG_NOTICE,"Connection timed out.  Remote host %s",
			inet_ntoa(addr.sin_addr));
	    return(CLOSED);
	}
	flag = atoi(buf);
	if (flag == 0) {
	    strcpy(error_resp.msg,"Backup flag value of 0 is invalid");
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    return(NONFATAL);
	}
    }

    /*=======================================================================
     * Build the tar command
     *======================================================================*/
    strcpy(buf, "/bin/tar -cf ");
    fn = tempnam(TMPDIR, "_padm");
    strcat(buf, fn);
    strcat(buf, " -C / ");
    strcat(buf, "`cat");
    if (flag & TYPICAL_BU) {
	strcat(buf, " ");
	strcat(buf, TREXLIST);
	strcat(buf, " ");
	strcat(buf, APACHELIST);
	strcat(buf, " ");
	strcat(buf, SYSTEMLIST);
    }
    if (flag & CUSTOM_BU) {
	strcat(buf, " ");
	strcat(buf, CUSTOMLIST);
    }
    strcat(buf, " | grep -v \"^#\"` 2>/dev/null");

    /*=======================================================================
     * Exec tar
     *======================================================================*/
    system(buf);

    /*=======================================================================
     * Open the tar file
     *======================================================================*/
    if ((ifd = open(fn, O_RDONLY, 0)) < 0) {
	sprintf(error_resp.msg, "Open of backup file %s failed.  %s",
			fn, sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	unlink(fn);
	return(NONFATAL);
    }

    if (fstat(ifd, &stat) < 0) {
	sprintf(error_resp.msg, "Cannot stat local backup file %s.  %s",
			fn, sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close(ifd);
	unlink(fn);
	return(NONFATAL);
    }

    resp.rc = OK;
    resp.msglen = stat.st_size;
    order_bytes_in_resp(&resp);

    if ((rc = putdata(fd, (char *)&resp, sizeof(resp), ckeysched)) < 0) {
	sprintf(error_resp.msg,"putdata() failed.  %s",
			sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close(ifd);
	unlink(fn);
	return(rc);
    }

    /*=======================================================================
     * Initialize the ivec for binary encryption/decryption
     *======================================================================*/
    memset((char *)&cbcblk, 0, sizeof(CBCBLK));
    memcpy(&cbcblk.ivec, IVEC, sizeof(des_cblock));
    cbcblk.flag = CBC_BIN;

    /*=======================================================================
     * Read and send the file
     *======================================================================*/
    for (total=0; total < stat.st_size; total += nread) {
	if ((nread = read(ifd, buf, sizeof(buf))) < 0) {
	    sprintf(error_resp.msg,"Read error reading %s.  %s",
			fn, sys_errlist[errno]);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    close(ifd);
	    unlink(fn);
	    return(NONFATAL);
	}

	if (nread == 0) {
	    sprintf(error_resp.msg,"File %s size conflict - stat() "
			"returns %d bytes but file contains %d bytes",
			fn, stat.st_size, total);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    close(ifd);
	    unlink(fn);
	    return(NONFATAL);
	}

/*
	if ((rc = putdata(fd, buf, nread, ckeysched)) < nread) {
*/
	if ((rc = cbc_write(fd, buf, nread, ckeysched, &cbcblk)) < nread) {
	    sprintf(error_resp.msg,"Error writing to client.  %s",
			sys_errlist[errno]);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    close(ifd);
	    unlink(fn);
	    return(FAILED);
	}
    }
    close(ifd);
    unlink(fn);
}

int restore(int fd, int msglen) {
    int rc;
    int ofd;
    int nread;
    int total;
    char *fn;
    char buf[BUFSIZ];
    char args[BUFSIZ];
    char cmd[BUFSIZ];
    int flag = 0;
    struct stat stat;
    CSREQ	req;
    SCRESP resp;
    CBCBLK cbcblk;

    /*=======================================================================
     * Make sure the msg fits into our buffer
     *======================================================================*/
    if (msglen > sizeof(buf)) {
	strcpy(error_resp.msg,"Args too long");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    if (msglen != 0) {
	/*==================================================================
	 * Read in the request to get the args
	 *=================================================================*/
	if ((rc = getdata(fd, args, msglen, ckeysched)) < 0) {
	    strcpy(error_resp.msg,"padmin restore cannot read args");
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    return(NONFATAL);
	}

	if (rc == 0) {
	    syslog(LOG_NOTICE,"Connection timed out.  Remote host %s",
			inet_ntoa(addr.sin_addr));
	    return(CLOSED);
	}
    }
    else
	strcpy(buf,"");

    /*=======================================================================
     * Send an OK so the client can start uploading the tar file
     *======================================================================*/
    resp.rc = OK;
    resp.msglen = 0;
    order_bytes_in_resp(&resp);

    if ((rc = putdata(fd, (char *)&resp, sizeof(resp), ckeysched)) < 0) {
	sprintf(error_resp.msg,"putdata() failed.  %s",
			sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(rc);
    }

    /*======================================================================
     * Get the second restore request from the client (this time with msglen
     * containing the length of the file)
     *=====================================================================*/
    if ((rc = getdata(fd, (char *)&req, sizeof(req), ckeysched)) < 0) {
	sprintf(error_resp.msg,"getdata() for second putconf request "
			"failed.  %s",sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(rc);
    }
    order_bytes_in_req(&req);

    if (req.msglen <= 0) {
	strcpy(error_resp.msg,"Upload file length in second restore request "
			"cannot be zero or negative");
	return(-1);
    }

    fn = tempnam(TMPDIR, "_padm");
    /*======================================================================
     * Open the local file to be written
     *=====================================================================*/
    if ((ofd = open(fn, O_WRONLY|O_CREAT, 0600)) < 0) {
	sprintf(error_resp.msg,"Cannot open %s.  %s",fn,sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    /*=======================================================================
     * Initialize the ivec for encryption/decryption
     *======================================================================*/
    memset((char *)&cbcblk, 0, sizeof(CBCBLK));
    memcpy(&cbcblk.ivec, IVEC, sizeof(des_cblock));
    cbcblk.flag = CBC_BIN;

    /*=======================================================================
     * Get the contents of the file over
     *======================================================================*/
    for (total=0; total < req.msglen; total += nread) {
	int len;
	char *p;

	len = req.msglen - total > sizeof(buf) ?
					sizeof(buf) : req.msglen - total;

	if ((nread = cbc_read(fd, buf, len, ckeysched, &cbcblk)) < 0) {
	    sprintf(error_resp.msg,"Failed reading %s.  %s",
					fn,sys_errlist[errno]);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    close(ofd);
	    unlink(fn);
	    return(NONFATAL);
	}

	if (nread == 0) {
	    sprintf(error_resp.msg,"Connection was closed");
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    close(ofd);
	    unlink(fn);
	    return(NONFATAL);
	}

	if ((rc = write(ofd,(unsigned char *)buf,nread)) != nread) {
	    sprintf(error_resp.msg,"Error writing to local temp file %s.  %s",
					fn,sys_errlist[errno]);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    close(ofd);
	    unlink(fn);
	    return(NONFATAL);
	}

    }
    close(ofd);

    /*=======================================================================
     * Build the tar command to restore files
     *======================================================================*/
    sprintf(cmd, "tar -xf %s", fn);
    strcat(cmd, args);

    /*=======================================================================
     * Issue the tar command
     *======================================================================*/
    getcwd(buf, sizeof(buf));
    chdir("/");
    system(cmd);
    chdir(buf);

    /*=======================================================================
     * Send an OK
     *======================================================================*/
    resp.rc = OK;
    resp.msglen = 0;
    order_bytes_in_resp(&resp);

    if ((rc = putdata(fd, (char *)&resp, sizeof(resp), ckeysched)) < 0) {
	sprintf(error_resp.msg,"putdata() failed.  %s",
			sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(rc);
    }

    return(0);
}
