/*============================================================================
 * filename:		padmin/padata.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1998-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1998-2000.  
 * All rights reserved.   
 *
 * Date written:	June	  24, 1998
 * Last modified:	February   8, 2000 - addd LPL
 *
 *===========================================================================*/
#include <stdio.h>
#include <syslog.h>
#include <errno.h>
#include <sys/time.h>
#if !defined(__Lynx__) && !defined(__hppa)
#include <sys/select.h>
#endif
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "des.h"
#include "cs.h"
#include "enc.h"

extern des_key_schedule ckeysched;
extern des_key_schedule dkeysched;

#ifndef linux
extern  char    *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif

extern	PACONF *padmconf;
extern	struct sockaddr_in addr;		/* Remote host's address*/
extern int cleft_over;				/* in libsrc/enc_io.c	*/
extern int sleft_over;				/* used to workaround	*/
						/* readahead problem	*/
						/* with select/enc_read	*/
/*=======================================================================
 * This function gets up to len bytes of data from fd.
 *
 * Input:
 *	fd	file descriptor of the input socket
 *	msg	address of the buffer where results are returned
 *	len	number of bytes to get
 *	keysched encryption key to use
 * Output:
 *	msg	filled with the expected data if successful
 *	rc 0	EOF
 *	rc n	n bytes read
 *	rc -1	failed
 *======================================================================*/
int getsomedata(int fd, char *buf, int len, des_key_schedule keysched) {
    int rc;
    struct timeval timeout;
    fd_set rdy;

    FD_ZERO(&rdy);
    FD_SET(fd,&rdy);

    timeout.tv_sec = padmconf->timeout;
    timeout.tv_usec = 0;

    /*=======================================================================
     * Before select, need to check to make sure enc_read() does not have
     * enough bytes in its read ahead buffer to satisfy this request.  If
     * we don't check first, select() may block until timeout while the
     * needed bytes are sitting in enc_read() buffer ready to be used.
     *======================================================================*/
    if ((fd == 0 && cleft_over < len) ||
	(fd != 0 && sleft_over < len)) {
#ifdef __hppa
	if((rc = select(fd+1,(int *)&rdy,0,0,&timeout)) < 0) {
#else
	if((rc = select(fd+1,&rdy,(fd_set *)0,(fd_set *)0,&timeout)) < 0) {
#endif
	    shutdown(fd,2);
	    syslog(LOG_ERR,"Error: select failed. %s.  Remote host %s",
    		sys_errlist[errno],inet_ntoa(addr.sin_addr));
	    return(FAILED);
	}

	if (rc == 0) {
	    shutdown(fd,2);
	    syslog(LOG_NOTICE,"Connection timed out.  Remote host %s",
		inet_ntoa(addr.sin_addr));
	    return(0);
	}
    }

    if ((rc = enc_read(fd, buf, len, keysched)) < 0) {
	shutdown(fd,2);
	syslog(LOG_ERR,"Error: Read failed.  %s.  "
		"Remote host %s",
    		sys_errlist[errno],inet_ntoa(addr.sin_addr));
	return(FAILED);
    }

    /*===============================================================
     * Set rc for EOF or short read
     *==============================================================*/
    if (rc == 0)
    	return(0);

    /*===============================================================
     * Success!
     *==============================================================*/
    return(rc);
}
/*=======================================================================
 * This function gets exactly len bytes of data from fd.
 *
 * Input:
 *	fd	file descriptor of the input socket
 *	msg	address of the buffer where results are returned
 *	len	number of bytes to get
 *	keysched encryption key to use
 * Output:
 *	msg	filled with the expected data if successful
 *	rc 0	EOF
 *	rc n	n bytes read
 *	rc -1	failed
 *======================================================================*/
int getdata(int fd, char *buf, int len, des_key_schedule keysched) {
    int rc;
    int n;

    for (n=0; n < len; n += rc) {
	if  ((rc = getsomedata(fd, &buf[n], len-n, keysched)) <= 0)
	    return(rc);
    }
    return(n);
}

/*=======================================================================
 * Note that this function may be called prior to or after authentication
 * so this function does not directly report errors back to the client.
 *
 * Input:
 *	fd	file descriptor of the output socket
 *	msg	address of the buffer where data is to be written
 *	len	number of bytes to put
 *	keysched encryption key to use
 * Output:
 *	rc n	n bytes written
 *	rc -1	failed
 *======================================================================*/
int putdata(int fd, char *buf, int len, des_key_schedule keysched) {
    int rc;

    if ((rc = enc_write(fd, buf, len, keysched)) < 0) {
	shutdown(fd,2);
	syslog(LOG_ERR,"Error: Write failed.  %s.  "
		"Remote host %s",
    		sys_errlist[errno],inet_ntoa(addr.sin_addr));
	return(FAILED);
    }
    
    if (rc < len) {
	syslog(LOG_ERR,"Error: protocol error, attempted to write %d bytes, "
		"%d bytes written  Remote host %s",
		len, rc, inet_ntoa(addr.sin_addr));
	return(FAILED);
    }

    /*===============================================================
     * Success!
     *==============================================================*/
    return(rc);
}

/*=======================================================================
 * This function should be called ONLY from the inner loop in padminsvr
 * to get a request from the client.  It will poll both the regular 
 * data socket as well as out-of-band data.  If the request came OOB,
 * then this function makes sure it is a CALLBACK request, otherwise, it
 * logs an alert and returns an error.
 *
 * Input:
 *	fd	file descriptor of the output socket
 *	req	address of the req buffer where data is to be read into
 * Output:
 *rc 0	connection closed
 *	rc n	n bytes read
 *	rc -1	failed
 *======================================================================*/
int getreq(int fd, CSREQ *reqp) {
    int rc;
    int nread;
    int nleft;
    int justreadoob = 0;
    char urg[4];
    struct timeval timeout;
    fd_set rdy,xcpt;

    FD_ZERO(&rdy);
    FD_ZERO(&xcpt);

    nread = 0;
    nleft = sizeof(CSREQ);
    while (nread < sizeof(CSREQ)) {
	FD_SET(fd,&rdy);
	if (justreadoob == 0)
	    FD_SET(fd,&xcpt);

	timeout.tv_sec = padmconf->timeout;
	timeout.tv_usec = 0;

#ifdef __hppa
	if((rc = select(fd+1,(int *)&rdy,0,(int *)&xcpt,&timeout)) < 0) {
#else
	if((rc = select(fd+1,&rdy,(fd_set *)0,(fd_set *)&xcpt,&timeout)) < 0) {
#endif
	    syslog(LOG_ERR,"Error: select failed. %s.  Remote host %s",
    		sys_errlist[errno],inet_ntoa(addr.sin_addr));
	    return(FAILED);
	}

	if (rc == 0) {
	    shutdown(fd, 2);
	    syslog(LOG_NOTICE,"Connection timed out.  Remote host %s",
		inet_ntoa(addr.sin_addr));
	    return(0);
	}

	if (FD_ISSET(fd,&xcpt)) {
	    if ((rc = recv(fd, urg, 1, MSG_OOB)) < 0) {
		syslog(LOG_ERR,"Error: Recv failed.  %s.  Remote host %s",
    			sys_errlist[errno],inet_ntoa(addr.sin_addr));
		return(FAILED);
	    }
	    if (rc == 0) {
		syslog(LOG_NOTICE,"Connection closed.  Remote host %s",
			inet_ntoa(addr.sin_addr));
		return(0);
	    }
	    justreadoob = 1;
	    FD_CLR(fd,&xcpt);
	}
#ifdef __sun
	/* Solaris needs this to work around its behavior which apparently
	   only return from select() once for regular data when OOB is also
	   present */
	if (justreadoob || FD_ISSET(fd,&rdy)) {
#else
	if (FD_ISSET(fd,&rdy)) {
#endif
	    if (justreadoob) {
		if ((rc = enc_read(fd, (char *)reqp + nread, nleft,
							dkeysched)) < 0) {
		    syslog(LOG_ERR,"Error: Read failed.  %s.  Remote host %s",
    			sys_errlist[errno],inet_ntoa(addr.sin_addr));
		    return(FAILED);
		}
		if (rc == 0) {
		    syslog(LOG_NOTICE,"Connection closed.  Remote host %s",
			inet_ntoa(addr.sin_addr));
		    return(0);
		}

		if (ntohl(reqp->cmd) != CALLBACK ||
			ntohl(reqp->component) != CLOGIN) {
		    syslog(LOG_ALERT,"Security Alert: OOB data from remote "
				"host %s is not a CALLBACK request",
				inet_ntoa(addr.sin_addr));
		    return(FAILED);
		}
	    }
	    else {
		if ((rc = enc_read(fd, (char *)reqp + nread, nleft,
							ckeysched)) < 0) {
		    syslog(LOG_ERR,"Error: Read failed.  %s.  Remote host %s",
    			sys_errlist[errno],inet_ntoa(addr.sin_addr));
		    return(FAILED);
		}
		if (rc == 0) {
		    syslog(LOG_NOTICE,"Connection closed.  Remote host %s",
			inet_ntoa(addr.sin_addr));
		    return(0);
		}
	    }
	    justreadoob = 0;
	    nread += rc;
	    nleft -= rc;
	}
    }
    order_bytes_in_req(reqp);

    return(nread);
}

/*=======================================================================
 * This function flushes the socket to get rid of extra bytes, usually
 * after an error is detected in a request.
 *
 * Input:
 *	fd	file descriptor of the input socket
 *	len	number of bytes to get rid of
 *	keysched key scedulke for decryption
 * Output:
 *	OK	 if successful
 *	FAILED	if failed
 *======================================================================*/
int flushreq(int fd, int len, des_key_schedule keysched) {
    char buf[BUFSIZ];
    int cnt;
    int readlen = 0;

    for (cnt = 0; cnt < len; cnt += readlen) {
	if ((readlen = enc_read(fd, buf, sizeof(buf), keysched)) < 0)
	    return(FAILED);
	if (readlen == 0)
	    return(CLOSED);
    }
    return(OK);
}
