/*============================================================================
 * filename:		padmin/padminconf.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1998-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1998-2000.  
 * All rights reserved.   
 *
 * Date written:	June	  24, 1998
 * Last modified:	February   8, 2000 - added LPL
 *
 *===========================================================================*/
#include <stdio.h>
#include <errno.h>
#include <syslog.h>
#include <string.h>
#include <malloc.h>
#include "cs.h"

/*===========================================================================
 * Input:
 * Output:
 *	padmconf points to new PACONF structure
 *	0	OK, PACONF struct filled
 *	1	No /etc/firewall/padmin.conf found
 *	-1	Failed
 *==========================================================================*/
#ifndef linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif

extern PACONF *padmconf;

#define PACONFFILE FW_CFG_DIR "/padmin.conf"

int padminconf(rootuser **rupp, int *nrootusers) {
    char linebuf[BUFSIZ];
    int line=0;
    int longport;
    FILE *infile;
    char *cp;
    rootuser *rup;

    if ((padmconf = (PACONF *)malloc(sizeof(PACONF))) == NULL) {
	syslog(LOG_ERR,"Error: Out of memory, %s",sys_errlist[errno]);
	return(-1);
    }

    if (!(infile = fopen(PACONFFILE,"r"))) {
	if (errno == ENOENT || errno == ENOTDIR) {
	    padmconf->port = DEFAULT_PADMIN_PORT;
	    padmconf->timeout = 600;
	    return(1);
	}
	syslog(LOG_ERR,"Error: Open failed for %s.  %s",PACONFFILE,
				sys_errlist[errno]);
	free(padmconf);
	padmconf = NULL;
	return(-1);
    }

    memset(padmconf,0,sizeof(PACONF));
    padmconf->timeout = 600;	/* Default to 10 minutes */

    while( fgets(linebuf, BUFSIZ, infile) != NULL) {
	line++;
	/*===================================================================
	 * Skip comment and null lines
	 *==================================================================*/
	if ((cp = strtok(linebuf," \t")) == NULL || *cp == '#')
	    continue;

	/*===================================================================
	 * Get the port number
	 *==================================================================*/
	if (!strcasecmp(cp,"port")) {
	    if ((cp = strtok(NULL," \t=:")) == NULL) {
		syslog(LOG_ERR,"Error: Port value not specified on line %d of "
			"%s",line,PACONFFILE);
		return(-1);
	    }

	    if ((longport = atoi(cp)) <= 0 || longport > 64000) {
		syslog(LOG_ERR,"Error: Port value must be an integer between "
			"1 and 64000 on line %d of %s.",line,PACONFFILE);
		return(-1);
	    }
	    padmconf->port = longport;
	    continue;
	}

	/*===================================================================
	 * Get the timeout value
	 *==================================================================*/
	if (!strcasecmp(cp,"timeout")) {
	    if ((cp = strtok(NULL," \t=:")) == NULL) {
		syslog(LOG_ERR,"Error: Timeout value not specified on line %d "
			"of %s",line,PACONFFILE);
		return(-1);
	    }
	    if ((padmconf->timeout = atoi(cp)) < 0) {
		syslog(LOG_ERR,"Error: Timeout value cannot be less than 0 "
			"on line %d of %s",line,PACONFFILE);
		return(-1);
	    }
	    continue;
	}

	/*===================================================================
	 * Process permit/deny lines
	 *==================================================================*/
	if (!strcasecmp(cp,"permit") || !strcasecmp(cp,"deny")) {
	    /*===============================================================
	     * Expand the IPS array if more than 1 entry is present
	     *==============================================================*/
	    if (padmconf->nips) {
		if ((padmconf = (PACONF *)realloc(padmconf,
			sizeof(PACONF)+((padmconf->nips+1)*sizeof(IPS))))									== NULL){
		    syslog(LOG_ERR,"Error: realloc - out of memory, %s",
			sys_errlist[errno]);
		    return(-1);
		}
	    }
	    /*===============================================================
	     * Set the permit/deny status for this entry
	     *==============================================================*/
	    if (strcasecmp(cp,"deny"))
		padmconf->hosts[padmconf->nips].op = PERMIT;
	    else
		padmconf->hosts[padmconf->nips].op = DENY;
	    /*===============================================================
	     * Validate the IP address
	     *==============================================================*/
	    if ((cp = strtok(NULL," \t:=")) == NULL) {
		syslog(LOG_ERR,"Error: Missing IP address after %s on line %d "
			"of %s",
			padmconf->hosts[padmconf->nips].op?
					"\"deny\"":"\"permit\"",
			line,PACONFFILE);
		return(-1);
	    }
	    if ((padmconf->hosts[padmconf->nips].ip = inet_addr(cp))
							== INADDR_NONE) {
		syslog(LOG_ERR,"Error: Invalid IP address on line %d of %s",
			line,PACONFFILE);
		return(-1);
	    }
	    /*===============================================================
	     * Get and validate the optional mask if present
	     * If no mask is specified, use INADDR_NONE (0xffffffff) so
	     * the default is exact match.
	     *==============================================================*/
	    if ((cp = strtok(NULL," \t:=")) == NULL) {
		padmconf->hosts[padmconf->nips].mask = INADDR_NONE;
		padmconf->nips++;
		continue;
	    }
	    if ((padmconf->hosts[padmconf->nips].mask = inet_addr(cp))
							== INADDR_NONE
			&& strcmp(cp,"255.255.255.255")) {
		syslog(LOG_ERR,"Error: Invalid IP mask on line %d of %s",
			line,PACONFFILE);
		return(-1);
	    }
	    padmconf->nips++;
	    continue;
	}

	/*===================================================================
	 * Get the rootuser info
	 *==================================================================*/
	if (!strcasecmp(cp,"rootuser")) {
	    /*===============================================================
	     * Allocat enough room for the rup array
	     *==============================================================*/
	    if (*nrootusers) {
		if ((rup = (rootuser *)realloc(rup,
				sizeof(rootuser)*(*nrootusers+1))) == NULL) {
		    syslog(LOG_ERR,"Error: realloc - out of memory, %s",
			sys_errlist[errno]);
		    return(-1);
		}
	    }
	    else {
		if ((rup = (rootuser *)malloc(sizeof(rootuser))) == NULL) {
		    syslog(LOG_ERR,"Error: malloc - out of memory, %s",
			sys_errlist[errno]);
		    return(-1);
		}
	    }

	    memset((char *)&rup[*nrootusers],0,sizeof(rootuser));
	    /*===============================================================
	     * Get the username
	     *==============================================================*/
	    if ((cp = strtok(NULL," \t:=\n")) == NULL) {
		    syslog(LOG_ERR,"Error: Missing username for rootuser on "
				"line %d of %s",line,PACONFFILE);
		    return(-1);
	    }
	    strcpy((char *)rup[*nrootusers].username,cp);
	    rup[*nrootusers].mask = 0xffffffff;

	    /*===============================================================
	     * Get the optional ip address
	     * If '*', treat it the same as when no ip addr is specified,
	     * except that there may be an optional mask following '*'
	     *==============================================================*/
	    if ((cp = strtok(NULL," \t:=\n")) != NULL) {
		if (strcmp(cp,"*") != 0 &&
		    (rup[*nrootusers].addr = inet_addr(cp)) == INADDR_NONE) {
		    syslog(LOG_ERR,"Error: Invalid IP address on line %d of %s",
			line,PACONFFILE);
		    return(-1);
		}
		/*===========================================================
		 * Get the optional ip mask
		 *==========================================================*/
		if ((cp = strtok(NULL," \t:=\n")) != NULL) {
		    if ((rup[*nrootusers].mask=inet_addr(cp)) == INADDR_NONE) {
			syslog(LOG_ERR,"Error: Invalid IP mask on line %d "
				"of %s",line,PACONFFILE);
			return(-1);
		    }
		}
	    }
	    (*nrootusers)++;
	    continue;
	}

	syslog(LOG_ERR, "Error: Invalid statement \"%s\" on line %d of %s",
			cp, line, PACONFFILE);
	return(-1);
	
    }

    fclose(infile);
    *rupp = rup;
    return(0);
}
