/*============================================================================
 * filename:		padmin/padminsvr.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1998-2000.  All rights reserved.
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1998-2000.
 * All rights reserved.   
 *
 * Date written:	June	  24, 1998
 * Last modified:	February   8, 2000 - added LPL
 *
 *===========================================================================*/
static  char	ident[] = "@(#) padminsvr, V 1.0, 00/01/21"
 			  " (C) Freemont Avenue Software, Inc."
			  " 1998-2000";
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <signal.h>
#include <fcntl.h>
#include <setjmp.h>
#include <limits.h>
#include <sys/stat.h>
#if !defined(__hppa) && !defined(__Lynx__)
#include <sys/select.h>
#endif
#include <sys/time.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/shm.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "des.h"
#include "cs.h"
#include "gwuser.h"

#define ISALLDIGITS(a)	(strspn(a,"0123456789") == strlen(a))

#define SETERR(a)	{strcpy(error_resp.msg,a); \
			 error_resp.hdr.msglen = strlen(error_resp.msg) + 1;}
#define DENIED	"Permission denied"
#define LOGFILENAME	FW_CFG_DIR "/padmin.binlog"

int padminconf(rootuser **rup, int *nrootusers);
int getdata(int fd, char *buf, int len, des_key_schedule keysched);
int putdata(int fd, char *buf, int len, des_key_schedule keysched);
int getreq(int fd, char *buf);
int portbind(unsigned short port);
int confile_to_client(int fd, int component, char *filelog, char *username);
int confile_from_client(int fd, int component,char *filelog);
int call_client(int fd, int component);
int proxy(int fd, int msglen);

#ifndef linux
extern  char    *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif

static local_connect=0;			/* src/dest both 127.0.0.1	*/
static  jmp_buf env;

RESPPKT error_resp;
struct GWADMIN admin;			/* user & groupname of login	*/
extern struct GWOPTIONS options;
struct sockaddr_in addr;		/* Remote host's address	*/
char *bufptr;				/* ptr to buffer for IO		*/
int buflen=0;				/* Lenght of buffer bufptr	*/
int csecrets=0;				/* encryption indicator		*/
des_key_schedule ckeysched;
des_key_schedule dkeysched;
PACONF *padmconf;

/*=====================================================================
 * Handles SIGCHLD so we don't leave zombies
 *====================================================================*/
void sig_chld(int sig) {
    int stat;

    while (waitpid(-1, &stat, WNOHANG) > 0);
    signal(SIGCHLD, sig_chld);
    return;
}

void order_bytes_in_req(CSREQ *reqp) {
    reqp->component = ntohl(reqp->component);
    reqp->cmd = ntohl(reqp->cmd);
    reqp->msglen = ntohl(reqp->msglen);
}

void order_bytes_in_resp(SCRESP *resp) {
    resp->rc = htonl(resp->rc);
    resp->msglen = htonl(resp->msglen);
}

/*=======================================================================
 * Check the peer's ip address against those specified in the conf structure
 * to determine if a connection should be permitted or denied.  If after
 * all the specified addresses have been checked and no match was found,
 * then deny the connection ... unless both the peer's ip and the socket's
 * local ip are 127.0.0.1 (this special connection is allowed unless
 * specifically denied in the conf file, in which case a match would have
 * been found earlier).
 *
 * Note that errors encountered here are all prior to an established trust
 * so we don't report errors back to the client.
 *
 * Input:
 *	fd	file descriptor of the input socket
 * Output:
 *	rc 0	trust this host
 *	rc -1	don't trust this host
 *======================================================================*/
int authhost(int fd) {
    int i;
    char remote[16];
    int len=sizeof(addr);

    local_connect = 0;

    /*==================================================================
     * Default to permission denied until authenticated
     *=================================================================*/
    SETERR(DENIED);

    if (getpeername(fd, (struct sockaddr *)&addr, &len) < 0) {
	perror("getpeername");
	syslog(LOG_ERR,"Error: getpeername failed.  %s",sys_errlist[errno]);
	return(-1);
    }

    /*==================================================================
     * Look thru the specified ip addresses in the config to see if any
     * matches the peer's ip address.  The first match will end the search
     *=================================================================*/
    for (i=0; i<padmconf->nips; i++) {
	if ((addr.sin_addr.s_addr & padmconf->hosts[i].mask) == 
		(padmconf->hosts[i].ip & padmconf->hosts[i].mask)) {
	    if (padmconf->hosts[i].op == PERMIT)
		return(0);
	    else {
		syslog(LOG_ALERT,"Security Alert: Denied (by rule) remote "
			"administration login request from host %s.",
			inet_ntoa(addr.sin_addr));
		return(-1);
	    }
	}
    }

    /*==================================================================
     * No match found.  Deny unless both source and destination of
     * socket have ip 127.0.0.1.
     *=================================================================*/
    if (addr.sin_addr.s_addr != htonl(0x7f000001)) {
	syslog(LOG_ALERT,"Security Alert: Denied remote administration "
		"login request from host %s.",
		inet_ntoa(addr.sin_addr));
	return(-1);
    }

    strcpy(remote,inet_ntoa(addr.sin_addr));
    len = sizeof(addr);
    if (getsockname(fd, (struct sockaddr *)&addr, &len) < 0) {
	perror("getsockname");
	syslog(LOG_ERR,"Error: getsockname failed.  %s",sys_errlist[errno]);
	return(-1);
    }
    if (addr.sin_addr.s_addr != htonl(0x7f000001)) {
	syslog(LOG_ALERT,"Security Alert: Denied remote administration "
		"login request from host %s to ip %s.",
		remote,inet_ntoa(addr.sin_addr));
	return(-1);
    }
    else
	local_connect=1;
    SETERR("");
    return(0);

}

/*====================================================================
 * signal handler for SIGUSR1
 *===================================================================*/
void refresh(int sig) {
   signal(SIGUSR1,refresh);
   longjmp(env,1);
}

main(int ac, char *argv[]) {
    unsigned short debugport;
    int fd;
    int len;
    int tempport;
    int not_inetd=0;
    int fl;
    char *filelog;
    char buf[BUFSIZ];
    pid_t pid;
    FILE *pf;
    int nrootusers = 0;
    rootuser *rup = NULL;

    error_resp.hdr.rc = FAILED;
    openlog("padmin",LOG_PID|LOG_NDELAY,LOG_DAEMON);

    /*========================================================================
     * Handle SIGCHLD so we don't get zombies
     *=======================================================================*/
    signal(SIGCHLD, sig_chld);

    /*===================================================================
     *	Dissociate from the controlling terminal
     *=================================================================*/
    setsid();
    chdir("/");

    /*===================================================================
     *	See if padmin was started by inetd
     *=================================================================*/
    not_inetd = getpeername(0,&addr,&len);
    syslog(LOG_NOTICE,"padmin daemon started");

    /*===================================================================
     *	Read the conf file
     *=================================================================*/
    if (padminconf(&rup, &nrootusers) < 0)
	exit(1);

    /*=======================================================================
     * Define program name and get the gwuser conf file info
     *======================================================================*/
    strcpy(options.program,"padmin");

    if (gwuserconf() < 0) {
	char buf[BUFSIZ];
	sprintf(buf, "Encountered problem accessing gwuser.conf");
	fprintf(stderr,buf);
	syslog(LOG_ERR,"Error: %s",buf);
	exit(-1);
    }

    /*===================================================================
     *	The listening port is assigned in this order of precedence:
     *	1. if inted was the initiator, then port is already determined
     *	2. if -D is specified, the port number sepcified with it is used
     *  3. the "port" line in /etc/firewall/paconf.conf is used
     *  4. DEFAULT_PADMIN_PORT is used (see cs.h)
     *=================================================================*/
    /*===================================================================
     *	If Debug is specified this will bind the port and listen
     *	directly (circumventing inetd) so you can run the program under
     *	a debugger
     *=================================================================*/
    if (not_inetd) {

	if( ac > 2 ) {
	    if( strcmp(argv[2],"-D") == 0) {
		/* check for debug port number  */
		if( ac > 3 ) {
		    if( (strlen(argv[3]) < 6)  && ISALLDIGITS(argv[3]) ) {
			if ( (tempport = atoi(argv[3])) > 64000) {
			    printf("\n%s: portnumber > 64000. Please try "
				"again.\n",argv[0]);
			    return(-1);
			}
		    }
		    else {
			printf("\n%s: invalid portnumber (%s), must be a "
			"positive integer less than 64000.\n          Please "
			"try again.\n", argv[0], argv[3]);
			return(-1);
		    }
		}
		debugport = tempport;
		printf("\n%s: Debug mode specified. Use port %d.\n",
			argv[0], debugport);
		fd = portbind(debugport);
	    }
	}

	/*===================================================================
	 * Bind to the port in config file or the default port
	 *==================================================================*/
	else {
	    if (padmconf->port)
		fd = portbind(padmconf->port);
	    else
		fd = portbind(DEFAULT_PADMIN_PORT);
	}

	if (fd < 0)
	    return(-1);

	/*===================================================================
	 * Run as a daemon if not started by inetd
	 *==================================================================*/
	if ((pid = fork()) < 0) {
	    perror("fork failed");
	    syslog(LOG_ERR,"Error: fork failed.  %s",sys_errlist[errno]);
	    return(-1);
	}

	if (pid != 0)
	    exit(0);
    }

    /*========================================================================
     * Write a pid file so the server can be started/stopped easily
     *=======================================================================*/
    if (!(pf = fopen(PIDFILE,"w"))) {
	syslog(LOG_ERR,"Error: Failed opening %s for writes.  %s\n",
			PIDFILE,sys_errlist[errno]);
	return(-1);
    }

    if (fprintf(pf,"%d\n",getpid()) < 0) {
	syslog(LOG_ERR,"Error: Failed writing to %s.  %s\n",PIDFILE,
			sys_errlist[errno]);
	return(-1);
    }

    fclose(pf);

    chmod(PIDFILE,S_IRUSR|S_IWUSR|S_IRGRP);

    /*====================================================================
     * Map the log file to keep track of config file downloads
     * If open failed because the file does not exist, create the file
     * and then write enough zeroes in it so mmap will work (on AIX 4.2).
     *
     * Note:
     * On Lynx, a filesytem file is not used because its mmap does not 
     * support mapping of filesystem objects.  A Posix shared memory object
     * is used instead.  This means that the filelog information is not kept
     * permanent (on a filesystem) but is only kept for the duration of one
     * invocation of padmin.
     *===================================================================*/
#if !defined(__Lynx__)
    if ((fl = open(LOGFILENAME,O_RDWR,0)) < 0  && errno == ENOENT) {
	if ((fl = open(LOGFILENAME,O_RDWR|O_CREAT,0600)) < 0) {
	    sprintf(buf,"Failed opening %s",LOGFILENAME);
	    perror(buf);
	    syslog(LOG_ERR,"Error: %s.  %s",buf,sys_errlist[errno]);
	    exit(1);
	}
	memset(buf,0,(NCOMP+1)*UNMSIZ);
	if (write(fl,buf,(NCOMP+1)*UNMSIZ) < 0) {
	    sprintf(buf,"Failed writing %s",LOGFILENAME);
	    perror(buf);
	    syslog(LOG_ERR,"Error: %s.  %s",buf,sys_errlist[errno]);
	    exit(1);
	}
    }
#else
    if ((fl = shm_open(LOGFILENAME,O_RDWR|O_CREAT,0600)) < 0) {
	sprintf(buf,"Failed opening shared memory for %s",LOGFILENAME);
	perror(buf);
	syslog(LOG_ERR,"Error: %s.  %s",buf,sys_errlist[errno]);
	exit(1);
    }
    if (ftruncate(fl, (NCOMP+1)*UNMSIZ) < 0) {
	sprintf(buf,"Failed opening shared memory for %s",LOGFILENAME);
	perror(buf);
	syslog(LOG_ERR,"Error: %s.  %s",buf,sys_errlist[errno]);
	exit(1);
    }
#endif

    if ((filelog = mmap(0, (NCOMP+1)*UNMSIZ, PROT_READ|PROT_WRITE,
#if !defined(__sun) && !defined(__Lynx__)
	MAP_FILE|
#endif
	MAP_SHARED, fl, 0)) == (caddr_t)-1) {
	char tmp[256];
	sprintf(tmp,"Error mapping file %s",LOGFILENAME);
	perror(tmp);
	syslog(LOG_ERR,"Error: %s.  %s",tmp,sys_errlist[errno]);
	exit(1);
    }

    /*===============================================================
     * Setup signal and setjmp so we can accept a USR1 signal to refresh
     *==============================================================*/
    signal(SIGUSR1,refresh);
    if (setjmp(env)) {
	if (rup != NULL) {
	    free(rup);
	    rup = NULL;
	    nrootusers = 0;
	}
	if (padmconf != NULL) {
	    free(padmconf);
	    padmconf = NULL;
	}
	if (padminconf(&rup, &nrootusers) < 0)
	    exit(1);
	syslog(LOG_NOTICE,"padmin daemon refreshed");
    }

    /*====================================================================
     * Macros used in the while loop below
     *===================================================================*/
    #define terminate_connection(a,b,c) {  error_resp.hdr.rc = FAILED; \
				    error_resp.hdr.msglen = strlen(b) + 1; \
				    if (b != error_resp.msg) \
					strncpy(error_resp.msg,b, \
						error_resp.hdr.msglen); \
				    order_bytes_in_resp(&(error_resp.hdr)); \
				    putdata(0,(char *)&error_resp, \
					sizeof(SCRESP)+ \
					ntohl(error_resp.hdr.msglen), c); \
				    shutdown(0,2); \
				    error_resp.hdr.msglen = 0; \
				    return(a); }

    #define terminate_connection_nomsg(a) {  shutdown(0,2); \
				    error_resp.hdr.msglen = 0; \
				    return(a); }
    #define nonfatal_error(a,b) {  error_resp.hdr.rc = NONFATAL; \
				    error_resp.hdr.msglen = strlen(a) + 1; \
				    if (a != error_resp.msg) \
					strncpy(error_resp.msg,a, \
						error_resp.hdr.msglen); \
				    order_bytes_in_resp(&(error_resp.hdr)); \
				    putdata(0,(char *)&error_resp, \
					sizeof(SCRESP)+ \
					ntohl(error_resp.hdr.msglen), b); \
				    error_resp.hdr.msglen = 0; }

    /*===================================================================
     * Outer loop that listens for a login request
     *==================================================================*/
    while (1) {
	CSREQ req;
	struct GWUSER gwuser;
	int gwuserrc;
	int rc;
	char username[UNMSIZ];
	time_t epoch;
	struct {
	    SCRESP resp;
	    char passwd[9];
	} response;

	/*===================================================================
	 * Setup/reset default encryption
	 *==================================================================*/
	{
	    unsigned long tkey[2];
	    tkey[0] = htonl(KEY1);
	    tkey[1] = htonl(KEY2);
	    des_set_key((des_cblock *)tkey,dkeysched);
	}

	/*===============================================================
	 * If not started by inetd, do an accept() and then spawn a child
	 * to handle it
	 *==============================================================*/
	if (not_inetd) {
	    int nread;

	    close(0);
	    while((nread = accept(fd,0,0)) < 0 && errno == EINTR);
	    if (nread < 0) {
		perror("accept");
		syslog(LOG_ERR,"accept failed.  %s.",sys_errlist[errno]);
		return(-1);
	    }

	    dup(nread);
	    if ((pid = fork()) < 0) {
		perror("fork failed");
		syslog(LOG_ERR,"Error: fork failed.  %s",sys_errlist[errno]);
		return(-1);
	    }
	    if (pid != 0)
		continue;

	    /*===========================================================
	     * ignore SIGCHLD in the child processes so we don't have to
	     * worry anymore about interrupted system calls
	     *==========================================================*/
	    signal (SIGCHLD, SIG_IGN);

	    close(fd);
	}

	/*===============================================================
	 * Check to see if this connection is allowed
	 *==============================================================*/
	if (authhost(0))
	    terminate_connection(-1,DENIED,dkeysched);

	/*===============================================================
	 * Send an OK response
	 *==============================================================*/
	response.resp.rc = OK;
	response.resp.msglen = 0;
	order_bytes_in_resp(&(response.resp));
	if ((rc = putdata(0,(char *)&response, sizeof(SCRESP),
						dkeysched)) <= 0) {
	    terminate_connection(rc,"Cannot send response to client",dkeysched);
	}

	/*===============================================================
	 * Get the first request
	 *==============================================================*/
	if ((rc = getdata(0,(char *)&req,sizeof(req), dkeysched)) <= 0)
	    terminate_connection(rc,"Connection request failed",dkeysched);
	order_bytes_in_req(&req);

	/*===============================================================
	 * Stop it right here if request is not for login
	 *==============================================================*/
	if (req.component != CLOGIN || req.cmd != CHALLENGE) {
	    syslog(LOG_ERR,"Error: Request received is not for login "
			"but only login requests are valid at this time.  "
			"Remote host %s.",inet_ntoa(addr.sin_addr));
	    terminate_connection(-1,"Request not accepted",dkeysched);
	}

	/*===============================================================
	 * Make sure data does not overflow our buffer of UNMSIZ
	 *==============================================================*/
	if (req.msglen < 0 || req.msglen > UNMSIZ) {
	    syslog(LOG_ALERT,"Security Alert: Length of username (%d) "
			"is out of range of 1 and %d.  Remote host %s",
			req.msglen,UNMSIZ-1,inet_ntoa(addr.sin_addr));
	    terminate_connection(-1,DENIED,dkeysched);
	}

	/*===============================================================
	 * If this is not the special case where both source and destination
	 * are 127.0.0.1, perform strong user authentication
	 *==============================================================*/
	if (!local_connect) {

	    /*===============================================================
	     * Read in the username
	     *==============================================================*/
	    if (req.msglen > 0 &&
		(rc = getdata(0, username, req.msglen, dkeysched)) <= 0)
		terminate_connection(rc,"Request failed",dkeysched);

	    /*===============================================================
	     * Get the gwuser record
	     * Don't terminate the connection here even if get_gwuser returns
	     * a bad rc.  We want to challenge no matter what and then
	     * terminate.
	     *==============================================================*/
	    memset(&gwuser,0,sizeof(gwuser));
	    gwuserrc = get_gwuser(username,&gwuser);

	    /*================================================================
	     * If get_gwuser returned failure, generate a random number and
	     * make it look like the userkey
	     *===============================================================*/
	    if (gwuserrc != 0) {
		epoch = time(&epoch);
		srand48((int) epoch);
		sprintf(gwuser.userkey, "%d", lrand48());
		gwuser.tokentype = CRYPTO;
	    }

	    /*===============================================================
	     * Generate and send the challenge
	     *==============================================================*/
	    switch (gwuser.tokentype)  
	    {
		int i;
		int rc;
		int ichall;
		int msglen;
		struct {
		    SCRESP resp;
		    char rnum[9];
		} challenge;
		char userresp[9];

		case CRYPTO:	        /* CRYPTOCard support	  */
		case SNK:	     	/* SecureNet Key  Support */   
	    
		    epoch = time(&epoch); 
		    srand48((int) epoch);	/* seed the rand # generator */
		    ichall = (lrand48() & 0x03FFFFFF) + 32800000; 
		    sprintf(challenge.rnum,"%d",ichall);
		    challenge.resp.rc = CONTINUE;
		    challenge.resp.msglen = 9;
		    order_bytes_in_resp(&challenge.resp);
		    if((rc=putdata(0,(char *)&challenge,
			sizeof(challenge.resp)+sizeof(challenge.rnum),
				dkeysched)) < 0)
			terminate_connection_nomsg(rc); 

		    /*=======================================================
		     * Get the next request
		     *======================================================*/
		    do {
			if((rc = getdata(0,(char *)&req,sizeof(req),
							dkeysched)) <= 0)
				terminate_connection(rc,"Request failed",
							dkeysched); 
		    } while (ntohl(req.component) == CTIMER &&
					ntohl(req.cmd) == KEEPALIVE);
		    order_bytes_in_req(&req);

		    /*=======================================================
		     * Make sure it is for authentication
		     *======================================================*/
		    if (req.component != CLOGIN || req.cmd != AUTH) {
			syslog(LOG_ERR,"Error: Expected response to "
				"challenge not received.  Remote host %s.",
				inet_ntoa(addr.sin_addr));
			terminate_connection(-1,"Request failed",dkeysched);
		    }

		    if (req.msglen == 0) {
			syslog(LOG_ALERT,"Security Alert: Empty "
				"response to strong user authentication.  "
				"Remote host %s",
				inet_ntoa(addr.sin_addr));
			terminate_connection(-1,DENIED,dkeysched);
		    }

		    if (req.msglen < 0 || req.msglen > 9) {
			syslog(LOG_ALERT,"Security Alert: Number of "
				"characters in the response to strong user "
				"authentication is less than 0 or greater "
				"than 8.  Remote host %s",
				inet_ntoa(addr.sin_addr));
			terminate_connection(-1,DENIED,dkeysched);
		    }

		    /*=======================================================
		     * Get response to the challenge
		     *======================================================*/
		    if((rc = getdata(0,userresp,req.msglen,dkeysched)) <= 0)
			terminate_connection(rc,"Request failed",dkeysched); 

		    if (snkverify(challenge.rnum, userresp, &gwuser) < 0) {
			syslog(LOG_ALERT,"Security Alert: Authentication "
				"failure for user %s.  Remost host %s",
				username, inet_ntoa(addr.sin_addr));
			terminate_connection(-1,DENIED,dkeysched);
		    }

		    /*=======================================================
		     * Make sure user has the appropriate permission to login
		     *======================================================*/
		    if (!(gwuser.tnaccess & ADM_TN)) {
			char tmp[BUFSIZ];
			sprintf(tmp,"User %s does not have ADM_TN permission "
				"to access padmin", username);
			syslog(LOG_NOTICE,tmp);
			terminate_connection(-1,tmp,dkeysched);
		    }

		    /*=======================================================
		     * Record user and group name for later use
		     *======================================================*/
		    memcpy(admin.user, gwuser.username, UNMSIZ);
		    memcpy(admin.group, gwuser.groupname, GRPSIZ);

		    /*=======================================================
		     * See if this is to be treated as a root user
		     * If so, set admin.group to all zeros to indicate
		     *======================================================*/
		    for (i=0; i < nrootusers; i++) {
			if (!strcmp(admin.user,rup[i].username) &&
				(rup[i].addr == 0 ||
				 ((addr.sin_addr.s_addr & rup[i].mask) ==
			    		(rup[i].addr & rup[i].mask)))) {
			    memset(admin.group,0,GRPSIZ);
			    break;
			}
		    }

		    /*=======================================================
		     * Send a response to tell client he passed challenge
		     *======================================================*/
		    response.resp.rc = OK;
		    response.resp.msglen = 0;
		    order_bytes_in_resp(&(response.resp));
		    if (putdata(0, (char *)&response.resp, sizeof(SCRESP),
							dkeysched) < 0)
			terminate_connection_nomsg(-1);

		    /*======================================================
		     * setup for encryption
		     *=====================================================*/
		    des_set_key((des_cblock *)md5key(gwuser.userkey),
					ckeysched);

		    break;
	
/*		case SDI:	*/	/* Security Dynamics Support    */
/*
		    return(tnsecdyn(gwuser.,buf));
		    break;  
*/
	      
		case NONE:    
		    syslog(LOG_ERR,"Error: no security token specified for "
			"user = '%s', group = '%s'",
			gwuser.username,gwuser.groupname);
		    terminate_connection(-1,DENIED,dkeysched);
		    break;

		default:
		    terminate_connection_nomsg(0);
		    break;

	    }    

	    /*==============================================================
	     * just in case gwuser returned failure earlier, we terminate
	     * the connection right here
	     *=============================================================*/
	    if (gwuserrc)
		terminate_connection(-1,DENIED,dkeysched);
	}
	else {
	    int shmid;
	    void *shmptr;
	    des_cblock random;
	    /*==============================================================
	     * special case local connection, we may need to read in some
	     * garbage bytes just to get rid of them
	     * After that, empty out username so there's no confusion
	     *=============================================================*/
	    if (req.msglen > 0)
		getdata(0, username, req.msglen, dkeysched);
	    strcpy(username,"");

	    /*==============================================================
	     * Set up a random key to be used for encrypting the session
	     * and copy it into shared memory for the local client to
	     * pick up
	     *=============================================================*/
	    pid = getpid();
	    if ((shmid = shmget(pid,sizeof(des_key_schedule),
				IPC_CREAT | IPC_EXCL | 0600)) < 0)
		terminate_connection(-1,"Cannot get shared memroy",dkeysched);

	    if ((shmptr = shmat(shmid,(void *)0,0)) == (void *)-1)
		terminate_connection(-1,"Cannot attach shared memory",
				dkeysched);

	    des_random_key(random);
	    des_set_key(&random,ckeysched);
	    memcpy((char *)shmptr, (char *)ckeysched, sizeof(des_key_schedule));

	    shmdt(shmptr);

	    /*=======================================================
	     * Record user and group name for later use
	     *======================================================*/
	    strcpy(username, "root");
	    memcpy(admin.user, username, UNMSIZ);
	    memset(admin.group, 0, GRPSIZ);

	    /*==============================================================
	     * Send an OK response to client
	     * The response.passwd field contains the pid so the client
	     * knows which shared memory to attach to
	     *=============================================================*/
	    response.resp.rc = OK;
	    sprintf(response.passwd, "%d", pid);
	    response.resp.msglen = strlen(response.passwd) + 1;
	    order_bytes_in_resp(&(response.resp));
	    if (putdata(0,(char *)&response,
			sizeof(SCRESP)+strlen(response.passwd)+1,
							dkeysched) < 0)
		terminate_connection_nomsg(-1);
	}

	/*=================================================================
	 * Log the connection
	 *================================================================*/
	if (strlen(admin.group) == 0)
	    syslog(LOG_NOTICE,"padmin connection with root privilege "
			"permitted for client %s, "
			"username %s",inet_ntoa(addr.sin_addr),username);
	else
	    syslog(LOG_NOTICE,"padmin connection permitted for client %s, "
			"username %s",inet_ntoa(addr.sin_addr),username);

	/*=================================================================
	 * Client has passed authentication
	 *================================================================*/
	while (1) {
	    if ((rc = getreq(0, (char *)&req)) < 0)
		terminate_connection(-1,"Server cannot read request",ckeysched);
	    if (rc == 0) {
		shutdown(0,2);
		return(CLOSED);
	    }
	    switch (req.cmd) {
		case GETCONF:
		    if (strlen(admin.group) != 0) {
			sprintf(error_resp.msg,"Must be root to run this "
					"function (currently logged in as %s)",
					admin.user);
			syslog(LOG_NOTICE,error_resp.msg);
			rc = NONFATAL;
			break;
		    }
		    rc = confile_to_client(0,req.component,filelog,username);
		    break;
		case PUTCONF:
		    if (strlen(admin.group) != 0) {
			sprintf(error_resp.msg,"Must be root to run this "
					"function (currently logged in as %s)",
					admin.user);
			syslog(LOG_NOTICE,error_resp.msg);
			rc = NONFATAL;
			break;
		    }
		    rc = confile_from_client(0,req.component,filelog);
		    break;
		case CALLBACK:
		    rc = call_client(0,req.msglen);
if (rc > 0)
fprintf(stderr,"fd conflict\n");
		    break;
		case PROXY:
		    rc = proxy(0,req.msglen);
		    break;
		case TAILLOG:
		    rc = taillog(0,req.component,req.msglen);
		    break;
		case GETGWUSER:
		    rc = getgwrecord(0,req.component,req.msglen);
		    break;
		case ADDGWUSER:
		    rc = put_new_gwuser(0,req.component,req.msglen);
		    break;
		case MODGWUSER:
		    rc = put_exist_gwuser(0,req.component,req.msglen);
		    break;
		case GETDEFGWUSER:
		    rc = getgwdefault(0);
		    break;
		case DELGWUSER:
		    rc = delgwrecord(0,req.component,req.msglen);
		    break;
		case BACKUP:
		case RESTORE:
		    if (strlen(admin.group) != 0) {
			sprintf(error_resp.msg,"Must be root to run this "
					"function (currently logged in as %s)",
					admin.user);
			syslog(LOG_NOTICE,error_resp.msg);
			rc = NONFATAL;
			break;
		    }
		    if (req.cmd == BACKUP)
			rc = backup(0,req.msglen);
		    else
			rc = restore(0,req.msglen);
		    break;
		case STARTCOMP:
		case STOPCOMP:
		case REFCOMP:
		    if (strlen(admin.group) != 0) {
			sprintf(error_resp.msg,"Must be root to run this "
					"function (currently logged in as %s)",
					admin.user);
			syslog(LOG_NOTICE,error_resp.msg);
			rc = NONFATAL;
			break;
		    }
		    rc = ssref(0,req.component,req.cmd);
		    break;
		case GETLIST:
		    if (strlen(admin.group) != 0 && req.component != CALL) {
			sprintf(error_resp.msg,"Must be root to run this "
					"function (currently logged in as %s)",
					admin.user);
			syslog(LOG_NOTICE,error_resp.msg);
			rc = NONFATAL;
			break;
		    }
		    rc = getlist(0,req.component);
		    break;
		case GETIFLIST:
		    if (strlen(admin.group) != 0 && req.component != CALL) {
			sprintf(error_resp.msg,"Must be root to run this "
					"function (currently logged in as %s)",
					admin.user);
			syslog(LOG_NOTICE,error_resp.msg);
			rc = NONFATAL;
			break;
		    }
		    rc = getiflist(0,req.component);
		    break;
		case KEEPALIVE:
		    /*=======================================================
		     * If cmd = KEEPALIVE, component = CTIMER and msglen = 0
		     * all we need do is loop back and listen for another
		     * request.  This will extend the connection for another
		     * timeout.  If cmd = KEEPALIVE but the oter fields are
		     * not as expected, issue an error message
		     *======================================================*/
		    if (req.component != CTIMER || req.msglen != 0) {
			rc = NONFATAL;
			sprintf(error_resp.msg,"Invalid KEEPALIVE request");
		    }
		    else
			rc = 0;
		    break;
		default:
		    rc = NONFATAL;
		    sprintf(error_resp.msg,"Invalid request command = %x",
				req.cmd);
	    }

	    if (rc == CLOSED) {
		shutdown(0,2);
		return(CLOSED);
	    }
	    if (rc == NONFATAL) {
		nonfatal_error(error_resp.msg,ckeysched);
		continue;
	    }
	    if (rc == FAILED)
		terminate_connection(-1, error_resp.msg,ckeysched);
	}

    }
}
