/*============================================================================
 * filename:		padmin/pafiles.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1998-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1998-2000.  
 * All rights reserved.   
 *
 * Date written:	June	  24, 1998
 * Last modified:	February   8, 2000 - added LPL
 *
 *===========================================================================*/
#include <stdio.h>
#include <errno.h>
#include <syslog.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include "des.h"
#include "cs.h"

int getdata(int fd, char *buf, int len, des_key_schedule ckeysched);
int getsomedata(int fd, char *buf, int len, des_key_schedule ckeysched);
int putdata(int fd, char *buf, int len, des_key_schedule ckeysched);

#ifndef linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif
extern RESPPKT error_resp;
extern des_key_schedule ckeysched;
extern des_key_schedule dkeysched;

/*===========================================================================
 * Returns the filename of the config file for the component passed as an
 * argument
 * Returns -1 if component is not on our list
 *==========================================================================*/
int trexfile(int component, char *path) {
    switch (component) {
	case CAPACHE:
	    strcpy(path,"/etc/apache/httpd.conf");
	    return(0);
	case CFTPROXY:
	    strcpy(path,FW_CFG_DIR "/ftproxy.conf");
	    return(0);
	case CFWPULSE:
	    strcpy(path,FW_CFG_DIR "/fwpulse.conf");
	    return(0);
	case CAPROXY:
	    strcpy(path,FW_CFG_DIR "/aproxy.conf");
	    return(0);
	case CGWUSER:
	    strcpy(path,FW_CFG_DIR "/gwuser.conf");
	    return(0);
	case CHTTPD:
	    strcpy(path,"/etc/httpd.conf");
	    return(0);
	case CINETD:
	    strcpy(path,"/etc/inetd.conf");
	    return(0);
	case CPADMIN:
	    strcpy(path,FW_CFG_DIR "/padmin.conf");
	    return(0);
	case CFWMON:
	    strcpy(path,FW_CFG_DIR "/fwmon.conf");
	    return(0);
	case CRAPROXY:
	    strcpy(path,FW_CFG_DIR "/raproxy.conf");
	    return(0);
	case CRPCPROXY:
	    strcpy(path,FW_CFG_DIR "/rpcproxy.conf");
	    return(0);
	case CSMALIASES:
	    strcpy(path,FW_CFG_DIR "/aliases");
	    return(0);
	case CSMWRAP:
	    strcpy(path,FW_CFG_DIR "/smwrap.conf");
	    return(0);
	case CTNPROXY:
	    strcpy(path,FW_CFG_DIR "/tnproxy.conf");
	    return(0);
	case CSECNETS:
	    strcpy(path,FW_CFG_DIR "/securenets");
	    return(0);
	case CSECPORTS:
	    strcpy(path,FW_CFG_DIR "/secureports");
	    return(0);
	case CSOCKD:
	    strcpy(path,"/etc/sockd.conf");
	    return(0);
	case CSYSLOG:
	    strcpy(path,"/etc/syslog.conf");
	    return(0);
	case CWEBGATE:
	    strcpy(path,FW_CFG_DIR "/webgate.conf");
	    return(0);
	case CTREXLIST:
	    strcpy(path,FW_CFG_DIR "/trex.bulist");
	    return(0);
	case CAPACHELIST:
	    strcpy(path,FW_CFG_DIR "/apache.bulist");
	    return(0);
/*
	case CPADMINLIST:
	    strcpy(path,FW_CFG_DIR "/padmin.bulist");
	    return(0);
	case CSOCKSLIST:
	    strcpy(path,FW_CFG_DIR "/socks.bulist");
	    return(0);
*/
	case CSYSTEMLIST:
	    strcpy(path,FW_CFG_DIR "/system.bulist");
	    return(0);
	case CCUSTOMLIST:
	    strcpy(path,FW_CFG_DIR "/custom.bulist");
	    return(0);
	default:
	    return(-1);
    }
}

/*======================================================================
 * Send the content of the config file for the passed component to the
 * client.
 * input:
 *	ofd	file descriptor for socket to client
 *	component the component for which the config file is for
 *	filelog	the pointer returned from mmap() for /etc/firewall/padmin.binlog
 *	username	the username currently logged in
 * output:
 *	returns 0 if OK
 *	returns -1 if failed - connection should be closed
 *	returns -2 if connection is closed during transfer
 *	returns -3 if non fatal error - loop back for new request
 *	error_resp.msg will contain a message if rc is negative
 *=====================================================================*/
int confile_to_client(int ofd, int component, char *filelog, char *username) {
    char fn[BUFSIZ];
    char buf[BUFSIZ];
    char *logp;
    int ifd;
    int rc;
    int total;
    int nread;
    struct stat statbuf;
    SCRESP resp;

    if (trexfile(component, fn) < 0) {
	strcpy(error_resp.msg,"Requested component does not have a config "
			"file that is managed by padmin");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    /*======================================================================
     * See if this file has been 'checked out'
     *=====================================================================*/
    logp = filelog + component * UNMSIZ;
    if (strlen(logp) > 0 && strcmp(username,logp) != 0) {
	/*==================================================================
	 * File has been checked out, send a prompt string to the client
	 *=================================================================*/
	CSREQ req;
	resp.rc = CONTINUE;
	sprintf(buf,"User %s may be editing config file %s also.", logp,fn);
	resp.msglen = strlen(buf) + 1;

	/*==================================================================
	 * Send a CONTINUE response with the prompt string to the client
	 *=================================================================*/
	order_bytes_in_resp(&resp);
	if ((rc = putdata(ofd, (char *)&resp, sizeof(resp), ckeysched)) < 0) {
	    sprintf(error_resp.msg,"putdata() failed.  %s",
	    sys_errlist[errno]);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    return(rc);
	}
	if ((rc = putdata(ofd, buf, ntohl(resp.msglen), ckeysched)) < 0) {
	    sprintf(error_resp.msg,"putdata() failed.  %s",
	    sys_errlist[errno]);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    return(rc);
	}

	/*==================================================================
	 * Look for the next request and skip over NO-OPs
	 *=================================================================*/
	do {
	    if ((rc = getdata(ofd, (char *)&req, sizeof(req), ckeysched)) < 0) {
		sprintf(error_resp.msg,"getdata() failed.  %s",
		sys_errlist[errno]);
		syslog(LOG_ERR,"Error: %s",error_resp.msg);
		return(rc);
	    }
	    if (rc == 0)
		return(CLOSED);
        } while (ntohl(req.cmd) == KEEPALIVE && ntohl(req.component) == CTIMER);
	order_bytes_in_req(&req);

        if (req.cmd != GETCONF || req.component != component)
	    return(0);	/* Not a failure	*/
    }

    /*======================================================================
     * Open the local file to be downloaded
     *=====================================================================*/
    if ((ifd = open(fn, O_RDONLY, 0)) < 0) {
	sprintf(error_resp.msg,"Cannot open %s.  %s",fn,sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    if (fstat(ifd, &statbuf) < 0) {
	sprintf(error_resp.msg,"stat() on file %s failed.  %s",
		fn,sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    resp.rc = OK;
    resp.msglen = statbuf.st_size;
    order_bytes_in_resp(&resp);

    if ((rc = putdata(ofd, (char *)&resp, sizeof(resp), ckeysched)) < 0) {
	sprintf(error_resp.msg,"putdata() failed.  %s",
			sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(rc);
    }

    for (total=0; total < statbuf.st_size; total += nread) {
	if ((nread = read(ifd, buf, sizeof(buf))) < 0) {
	    sprintf(error_resp.msg,"Read error reading %s.  %s",
			fn, sys_errlist[errno]);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    return(NONFATAL);
	}

	if (nread == 0) {
	    sprintf(error_resp.msg,"File %s size conflict - stat() "
			"returns %d bytes but file contains %d bytes",
			fn, statbuf.st_size, total);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    return(NONFATAL);
	}

	if ((rc = putdata(ofd, buf, nread, ckeysched)) < nread) {
	    sprintf(error_resp.msg,"Error writing to client.  %s",
			sys_errlist[errno]);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    return(-1);
	}
    }
    close(ifd);

    /*======================================================================
     * Write the username in the log file to indicate that this config file
     * has been checked out by this user.
     *=====================================================================*/
    strcpy(logp,username);

    return(0);
}

/*======================================================================
 * Get the content of the conf file for the passed component from the
 * client.
 * input:
 *	ifd	file descriptor for socket to client
 *	component the component for which the config file is for
 *	filelog	the pointer returned from mmap() for /etc/firewall/padmin.binlog
 * output:
 *	returns 0 if OK
 *	returns -1 if failed - connection should be closed
 *	returns -2 if connection is closed during transfer
 *	returns -3 if non fatal error - loop back for new request
 *	error_resp.msg will contain a message if rc is negative
 *=====================================================================*/
int confile_from_client(int ifd, int component, char *filelog) {
    char fn[BUFSIZ];
    char newfn[BUFSIZ];
    char buf[BUFSIZ];
    int ofd;
    int rc;
    int total;
    int nread;
    struct stat statbuf;
    CSREQ req;
    SCRESP resp;
    mode_t mode;

    if (trexfile(component, fn) < 0) {
	strcpy(error_resp.msg,"Requested component does not have a config "
			"file that is managed by padmin");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    strcpy(newfn,fn);
    strcat(newfn,".new");

    /*======================================================================
     * Get the current config file's mode and use it to create the new file
     * If no current file exists, then use 0640 for everyone except sockd
     * which uses 0644
     *=====================================================================*/
    if (stat(fn, &statbuf) < 0) {
	if (errno != ENOENT) {
	    sprintf(error_resp.msg,"stat() on file %s failed.  %s",
		fn,sys_errlist[errno]);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    return(NONFATAL);
	}
	else {
	    if (component == CSOCKD)
		mode = 0644;
	    else
		mode = 0640;
	}
    }
    else
	mode = statbuf.st_mode;

    /*======================================================================
     * Open the local file to be written
     *=====================================================================*/
    if ((ofd = open(newfn, O_WRONLY|O_CREAT, mode)) < 0) {
	sprintf(error_resp.msg,"Cannot open %s.  %s",fn,sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    resp.rc = OK;
    resp.msglen = 0;
    order_bytes_in_resp(&resp);

    if ((rc = putdata(ifd, (char *)&resp, sizeof(resp), ckeysched)) < 0) {
	sprintf(error_resp.msg,"putdata() failed.  %s",
			sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close(ofd);
	return(rc);
    }

    /*======================================================================
     * Get the second putconf request from the client (this time with msglen
     * containing the length of the file
     *=====================================================================*/
    if ((rc = getdata(ifd, (char *)&req, sizeof(req), ckeysched)) < 0) {
	sprintf(error_resp.msg,"getdata() for second putconf request "
			"failed.  %s",sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close(ofd);
	return(rc);
    }
    order_bytes_in_req(&req);

    if (req.cmd != PUTCONF || req.component != component) {
	strcpy(error_resp.msg,"Protocol error, expected second PUTCONF "
			"request not found");
	close(ofd);
	return(rc);
    }

    if (req.msglen <= 0) {
	strcpy(error_resp.msg,"Upload file length in second putconf request "
			"cannot be zero or negative");
	close(ofd);
	return(-1);
    }

    for (total=0; total < req.msglen; total += nread) {
	int len;
	char *p;

	if ((nread = getsomedata(ifd, buf, sizeof(buf), ckeysched)) < 0) {
	    sprintf(error_resp.msg,"Read error reading %s.  %s",
			fn, sys_errlist[errno]);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    close(ofd);
	    return(NONFATAL);
	}

	if (nread == 0) {
	    sprintf(error_resp.msg,"Remote host disconnected after uploading "
			"%d of %d bytes intended for file %s",
			total, req.msglen, fn);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    close(ofd);
	    return(-1);
	}

	/*===============================================================
	 * If we find a byte of zero, we assume everything that follows
	 * is a filler and can be ignored.
	 *==============================================================*/
	if (p = memchr(buf, 0, nread))
	    len = p - buf;
	else
	    len = nread;

	if ((rc = write(ofd, buf, len)) < len) {
	    sprintf(error_resp.msg,"Error writing to %s.  %s",
			newfn, sys_errlist[errno]);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    close(ofd);
	    return(-1);
	}
    }
    close(ofd);

    if (rename(newfn,fn) < 0) {
	sprintf(error_resp.msg,"Could not rename %s to %s",newfn,fn);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    /*=======================================================================
     * Notify the client that upload is OK
     *======================================================================*/
    resp.rc = OK;
    resp.msglen = 0;
    order_bytes_in_resp(&resp);

    if ((rc = putdata(ifd, (char *)&resp, sizeof(resp), ckeysched)) < 0) {
	sprintf(error_resp.msg,"putdata() failed.  %s",
			sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(rc);
    }

    /* Remove any indicator that this file is checked out */
    memset(filelog + component * UNMSIZ, 0, UNMSIZ);

    return(0);
}
