/*============================================================================
 * filename:		padmin/pagwuser.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1998-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1998-2000.  
 * All rights reserved.   
 *
 * Date written:	June	  24, 1998
 * Last modified:	February   8, 2000 - added LPL
 *
 *===========================================================================*/
#include <stdio.h>
#include <errno.h>
#include <syslog.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#ifndef linux
#include <ndbm.h>
#else
#include <db1/ndbm.h>
#endif
#include <sys/stat.h>
#include <netinet/in.h>
#include "des.h"
#include "cs.h"
#include "gwuser.h"

int getdata(int fd, char *buf, int len, des_key_schedule ckeysched);
int getsomedata(int fd, char *buf, int len, des_key_schedule ckeysched);
int putdata(int fd, char *buf, int len, des_key_schedule ckeysched);
int put_new_gwgroup(int ofd, struct GWUSER *user);

#ifndef linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif
extern struct GWADMIN admin;
extern struct GWCONF conf;
extern RESPPKT error_resp;
extern des_key_schedule ckeysched;
extern des_key_schedule dkeysched;
extern DBM *db;
extern struct  GWINFO info;

/*====================================================================
 * Gets the contents of a gwuser or gwgroup and send it to the client.
 * Since the client always expects a GWUSER structure, in cases where
 * a gwgroup record is requested, this function translates the GWGROUP
 * structure to a GWUSER structure before transmitting.
 * input:
 *	ofd	file descriptor for socket to client
 *	component the component is GWUSER or GWGROUP
 *	msglen	is the length of the user/group name that follows in the
 *		request
 * output:
 *	returns 0 if OK
 *	returns -1 if failed - connection should be closed
 *	returns -2 if connection is closed during transfer
 *	returns -3 if non fatal error - loop back for new request
 *	error_resp.msg will contain a message if rc is negative
 *=====================================================================*/
int getgwrecord(int ofd, int component, int msglen) {
    char username[UNMSIZ];
    int rc;
    struct GWGROUP gwg;
    char *gwbuf;
    int gwlen;
    struct {
	SCRESP resp;
	struct GWUSER gwu;
    } pkt;
    datum key, data;
    GWUSER_DECRYPT decrypted_data;

    /*======================================================================
     * Read in the rest of the request to get the user/group name
     * Assume the terminating null is in the data stream
     *=====================================================================*/
    if (msglen <= 0) {
	strcpy(error_resp.msg,"Get gwuser request contains empty name");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    /*=======================================================================
     * Make sure the name fits into our buffer
     *======================================================================*/
    if (msglen > sizeof(username)) {
	strcpy(error_resp.msg,"User/Group name too long");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    if ((rc = getdata(ofd,username,msglen,ckeysched)) < 0) {
	strcpy(error_resp.msg,"padmin getgwuser cannot read user/group name");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(rc);
    }
    if (rc == 0)
	return(CLOSED);

    /*=======================================================================
     * Determine whether to get a gwuser or a gwgroup record
     *======================================================================*/
    if (component == CGWUSER) {
	gwbuf = (char *)&pkt.gwu;
	gwlen = sizeof(pkt.gwu);
    }
    else if (component == CGWGROUP) {
	gwbuf = (char *)&gwg;
	gwlen = sizeof(gwg);

	/*===================================================================
	 * Make sure user has rights to access requested group
	 *==================================================================*/
	if (strlen(admin.group) > 0 && strcmp(admin.group,username)) {
	    sprintf(error_resp.msg,"Permission denied for user %s to access "
			"group %s",admin.user, username);
	    syslog(LOG_NOTICE,error_resp.msg);
	    return(NONFATAL);
	}
    }
    else {
	strcpy(error_resp.msg,"GETGWUSER request must target component "
		"GWUSER or GWGROUP");
	syslog(LOG_ERR,"Error: %s", error_resp.msg);
	return(NONFATAL);
    }

    /*======================================================================
     * Open NDBM database.
     *    The function open_db() locks the entire data base.
     *    The function close_db() unlocks the data base.
     *======================================================================*/
    if ((rc = open_db(EXIST|NO_LOCK)) < 0) {
	strcpy(error_resp.msg,"Open of gwuser database failed");
	syslog(LOG_ERR,"Error: %s", error_resp.msg);
	return(NONFATAL);
    }

    /* Retrieve user record */
    key.dptr = username;
    key.dsize = strlen(username);
    data = dbm_fetch(db,key);

    /* Check that record is actually for a user */
    if (data.dptr == NULL) {
	if (component == CGWUSER)
	    sprintf(error_resp.msg,"User '%s' not found.",username);
	else
	    sprintf(error_resp.msg,"Group '%s' not found.",username);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close_db();
	return(NONFATAL);
    }
    else if (*(char *)data.dptr != USER && component == CGWUSER) {
	sprintf(error_resp.msg,"'%s' is not a user record.",username);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close_db();
	return(NONFATAL);
    }
    else if (*(char *)data.dptr != GROUP && component == CGWGROUP) {
	sprintf(error_resp.msg,"'%s' is not a group record.",username);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    decrypt_record(data.dptr,decrypted_data,data.dsize);
    close_db();
    memcpy(gwbuf,decrypted_data,gwlen);

    if (component == CGWGROUP) {
	memset((char *)&pkt.gwu, 0, sizeof(pkt.gwu));
	strcpy(pkt.gwu.username, gwg.groupname);
	strcpy(pkt.gwu.realname, gwg.realname);
	pkt.gwu.tokentype = gwg.tokentype;
	pkt.gwu.tnaccess = gwg.tnaccess;
	pkt.gwu.ftpaccess = gwg.ftpaccess;
	memcpy(pkt.gwu.accesstimes,gwg.accesstimes,sizeof(pkt.gwu.accesstimes));
    }
    else {
	/*===============================================================
	 * Make sure user has rights to access this user
	 *==============================================================*/
	if (strlen(admin.group) > 0 && strcmp(admin.group,pkt.gwu.groupname)) {
	    sprintf(error_resp.msg, "Permission denied for user %s to access "
			"user %s",admin.user,pkt.gwu.username);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    return(NONFATAL);
	}
    }

    pkt.resp.rc = OK;
    pkt.resp.msglen = sizeof(pkt.gwu);
    order_bytes_in_resp(&(pkt.resp));

    if ((rc = putdata(ofd, (char *)&pkt, sizeof(pkt), ckeysched)) < 0) {
	sprintf(error_resp.msg,"putdata() failed.  %s",
			sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(rc);
    }

    return(0);
}

/*====================================================================
 * Gets the defaults from the gwuser.conf file and returns the info
 * in a GWUSER structure
 * input:
 *	ofd	file descriptor for socket to client
 * output:
 *	returns 0 if OK
 *	returns -1 if failed - connection should be closed
 *	returns -2 if connection is closed during transfer
 *	returns -3 if non fatal error - loop back for new request
 *	error_resp.msg will contain a message if rc is negative
 *=====================================================================*/
int getgwdefault(int ofd) {
    int rc;
    struct {
	SCRESP resp;
	struct GWUSER gwu;
    } pkt;

    memset((char *)&pkt.gwu, 0, sizeof(pkt.gwu));

    if (gwuserconf() < 0) {
	sprintf(error_resp.msg, "Encountered problem accessing gwuser.conf");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    /*=======================================================================
     * Copy default values into a gwuser structure
     * gwuserconf() was called in main (padminsvr.c)
     *======================================================================*/
    strcpy(pkt.gwu.groupname, conf.groupname);
    pkt.gwu.tokentype = conf.tokentype;
    pkt.gwu.tnaccess = conf.tnaccess;
    pkt.gwu.ftpaccess = conf.ftpaccess;
    memcpy(pkt.gwu.accesstimes, conf.accesstimes, sizeof(pkt.gwu.accesstimes));

    /*=======================================================================
     * Send response to client
     *======================================================================*/
    pkt.resp.rc = OK;
    pkt.resp.msglen = sizeof(pkt.gwu);
    order_bytes_in_resp(&(pkt.resp));

    if ((rc = putdata(ofd, (char *)&pkt, sizeof(pkt), ckeysched)) < 0) {
	sprintf(error_resp.msg,"putdata() failed.  %s", sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(rc);
    }
    return(OK);
}

/*====================================================================
 * Adds a new gwuser record from info received in a ADDGWUSER request.
 * If the request is to add a new gwgroup, put_new_gwgroup is called.
 * input:
 *	ofd	file descriptor for socket to client
 *	component the component is GWUSER or GWGROUP
 *	msglen	is the length of the user/group structure that follows in the
 *		request
 * output:
 *	returns 0 if OK
 *	returns -1 if failed - connection should be closed
 *	returns -2 if connection is closed during transfer
 *	returns -3 if non fatal error - loop back for new request
 *	error_resp.msg will contain a message if rc is negative
 *=====================================================================*/
int put_new_gwuser(int ofd, int component, int msglen) {
    int rc;
    datum userkey, userdata;
    datum groupkey, groupdata;
    GWUSER_ENCRYPT encrypted_data;
    struct GWUSER user;
    char buf[BUFSIZ];
    char *cp;
    RESPPKT pkt;

    /*======================================================================
     * Read in the rest of the request to get the user/group info
     * Assume the terminating null is in the data stream
     *=====================================================================*/
    if (msglen <= 0) {
	strcpy(error_resp.msg,"Add gwuser/gwgroup request contains no info");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    /*=======================================================================
     * Make sure the gwuser info matches our expected length
     *======================================================================*/
    if (msglen != sizeof(user)) {
	sprintf(error_resp.msg,"Info following an Add gwuser/gwgroup request "
		"does not match the expected length of %d",sizeof(user));
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    if ((rc = getdata(ofd,(char *)&user,msglen,ckeysched)) < 0) {
	strcpy(error_resp.msg,"padmin cannot read user/group info");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(rc);
    }
    if (rc == 0)
	return(CLOSED);

    /*=======================================================================
     * Make sure user/group name is non_blank
     *======================================================================*/
    if ((cp = strtok(user.username," \t")) == NULL) {
	strcpy(error_resp.msg,"User/Group name must be nonblank");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    /*=======================================================================
     * If GWGROUP, call its processor instead
     *======================================================================*/
    if (component == CGWGROUP)
	return(put_new_gwgroup(ofd, &user));

    /*=======================================================================
     * Make sure groupname for the new user is non_blank
     *======================================================================*/
    if ((cp = strtok(user.groupname," \t")) == NULL) {
	strcpy(error_resp.msg,"Group name must be nonblank");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    /*=======================================================================
     * Check admin group priviledge
     *======================================================================*/
    if (strlen(admin.group) > 0 && strcmp(admin.group,user.groupname)) {
	sprintf(error_resp.msg,"User %s cannot add user to group %s",
		admin.user,user.groupname);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    /*=======================================================================
     * Open NDM database
     *======================================================================*/
    if ((rc = open_db(EXIST)) < 0) {
	strcpy(error_resp.msg,"Open of gwuser database failed");
	syslog(LOG_ERR,"Error: %s", error_resp.msg);
	return(NONFATAL);
    }

    /*=======================================================================
     * Check that user record doesn't exist
     *======================================================================*/
    userkey.dptr = user.username;
    userkey.dsize = strlen(user.username);
    userdata = dbm_fetch(db,userkey);
    if (userdata.dptr != NULL) {
	if (*(char *)userdata.dptr == USER)
	    sprintf(error_resp.msg,"User %s already exists, cannot add.",
			userkey.dptr);
	else if (*(char *)userdata.dptr == GROUP)
	    sprintf(error_resp.msg,"Group %s already exists, cannot add.",
			userkey.dptr);
	else
	    sprintf(error_resp.msg,"%s already exists, cannot add.",
			userkey.dptr);
	syslog(LOG_ERR,"Error: %s", error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    /*=======================================================================
     * Check that user's group exists
     *======================================================================*/
    groupkey.dptr = user.groupname;
    groupkey.dsize = strlen(user.groupname);
    groupdata = dbm_fetch(db,groupkey);
    if (groupdata.dptr == NULL) {
	sprintf(error_resp.msg,"Group %s does not exist, user %s not added.",
			groupkey.dptr,userkey.dptr);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close_db();
	return(NONFATAL);
    }
    else if (*(char *)groupdata.dptr != GROUP) {
	sprintf(error_resp.msg,"%s is not a group, user %s not added.",
			groupkey.dptr,userkey.dptr);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    /*=======================================================================
     * If password changed, process it
     *======================================================================*/
    if (strlen(user.password[0]) > 0) {
	char tmppw[MAXPWDS][PWDSIZ];
	memset(tmppw, 0, sizeof(tmppw));
	/*===================================================================
	 * Check new password to make sure it meets restrictions in gwuser.conf
	 *==================================================================*/
	if (check_password(user.password[0], tmppw, &conf,
		error_resp.msg) < 0) {
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close_db();
	return(NONFATAL);
	}
    }

    /*=======================================================================
     * Make sure start and stop dates are consistent
     *======================================================================*/
    if (user.stopdate > 0 && user.startdate > user.stopdate) {
	strcpy(error_resp.msg,"Start date cannot be later than stop date.");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    /*=======================================================================
     * Complete the user record
     *======================================================================*/
    user.type = USER;
    user.passworddate = time((time_t *)NULL);
    user.recorddate = user.passworddate;

    /*=======================================================================
     * Store the record
     *======================================================================*/
    encrypt_record((char *)&user,encrypted_data,sizeof(user));
    userdata.dptr = encrypted_data;
    userdata.dsize = encrypted_sizeof(user);
    if (dbm_store(db,userkey,userdata,DBM_INSERT)) {
	sprintf(error_resp.msg,"dbm_store failed for %s.",userkey.dptr);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    /*=======================================================================
     * Update summary information record
     *======================================================================*/
    info_db(GWINFO_MODIFY,NUMUSERS,1,0);

    close_db();

    /*=======================================================================
     * Send an OK response to the client
     *======================================================================*/
    pkt.hdr.rc = OK;
    sprintf(pkt.msg,"User %s added.",userkey.dptr);
    pkt.hdr.msglen = strlen(pkt.msg) + 1;
    order_bytes_in_resp(&(pkt.hdr));
    if ((rc = putdata(ofd, (char *)&pkt, sizeof(SCRESP) + ntohl(pkt.hdr.msglen),
			ckeysched)) < 0) {
	syslog(LOG_ERR,"Error: putdata() failed");
	return(rc);
    }

    return(OK);
}

/*====================================================================
 * Adds a new gwgroup record from info received in a ADDGWUSER request.
 * that has a component of CGWGROUP
 * Since the client always sends a GWUSER structure, in cases where
 * a gwgroup record is to be added, this function translates the GWUSER
 * structure to a GWGROUP structure before processing.
 * input:
 *	ofd	file descriptor for socket to client
 *	user	pointer to the GWUSER structure that was just received
 * output:
 *	returns 0 if OK
 *	returns -1 if failed - connection should be closed
 *	returns -2 if connection is closed during transfer
 *	returns -3 if non fatal error - loop back for new request
 *	error_resp.msg will contain a message if rc is negative
 *=====================================================================*/
int put_new_gwgroup(int ofd, struct GWUSER *user) {
    int rc;
    struct GWGROUP group;
    datum key, data;
    GWGROUP_ENCRYPT encrypted_data;
    RESPPKT pkt;

    memset(&group, 0, sizeof(group));
    /*=======================================================================
     * Translate the received GWUSER structure to a GWGROUP structure
     *======================================================================*/
    group.type = GROUP;
    strcpy(group.groupname, user->username);
    strcpy(group.realname, user->realname);
    group.tokentype = user->tokentype;
    group.tnaccess = user->tnaccess;
    group.ftpaccess = user->ftpaccess;
    memcpy(group.accesstimes, user->accesstimes, sizeof(group.accesstimes));
    group.recorddate = time((time_t *)NULL);

    /*=======================================================================
     * Check for root user
     *======================================================================*/
    if (strlen(admin.group) != 0) {
	sprintf(error_resp.msg, "Must be root to add group (admin = %s).",
		admin.user);
	syslog(LOG_NOTICE, error_resp.msg);
	return(NONFATAL);
    }

    /*=======================================================================
     * Open NDM database
     *======================================================================*/
    if ((rc = open_db(EXIST|CREATE)) < 0) {
	strcpy(error_resp.msg,"Open of gwuser database failed");
	syslog(LOG_ERR,"Error: %s", error_resp.msg);
	return(NONFATAL);
    }

    /*=======================================================================
     * Check that group record doesn't exist
     *======================================================================*/
    key.dptr = group.groupname;
    key.dsize = strlen(group.groupname);
    data = dbm_fetch(db,key);
    if (data.dptr != NULL) {
	if (*(char *)data.dptr == USER)
	    sprintf(error_resp.msg,"User %s already exists, cannot add.",
			key.dptr);
	else if (*(char *)data.dptr == GROUP)
	    sprintf(error_resp.msg,"Group %s already exists, cannot add.",
			key.dptr);
	else
	    sprintf(error_resp.msg,"%s already exists, cannot add.",
			key.dptr);
	syslog(LOG_ERR,"Error: %s", error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    /*=======================================================================
     * Store new group record
     *======================================================================*/
    encrypt_record((char *)&group,encrypted_data,sizeof(group));
    data.dptr = encrypted_data;
    data.dsize = encrypted_sizeof(group);
    if (dbm_store(db,key,data,DBM_INSERT)) {
	sprintf(error_resp.msg, "dbm_store failed for %s", key.dptr);
	syslog(LOG_ERR,"Error: %s", error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    /*=======================================================================
     * Update summary information record
     *======================================================================*/
    info_db(GWINFO_MODIFY,NUMGROUPS,0,1);

    close_db();

    /*=======================================================================
     * Send an OK response to the client
     *======================================================================*/
    pkt.hdr.rc = OK;
    sprintf(pkt.msg,"Group %s added.",key.dptr);
    pkt.hdr.msglen = strlen(pkt.msg) + 1;
    order_bytes_in_resp(&(pkt.hdr));
    if ((rc = putdata(ofd, (char *)&pkt, sizeof(SCRESP) + ntohl(pkt.hdr.msglen),
			ckeysched)) < 0) {
	syslog(LOG_ERR,"Error: putdata() failed");
	return(rc);
    }

    return(OK);
}

/*====================================================================
 * Modify an existing gwuser record from info received in a MODGWUSER request.
 * If the request is to modify an existing gwgroup, put_exist_group is called.
 * input:
 *	ofd	file descriptor for socket to client
 *	component the component is GWUSER or GWGROUP
 *	msglen	is the length of the user/group structure that follows in the
 *		request
 * output:
 *	returns 0 if OK
 *	returns -1 if failed - connection should be closed
 *	returns -2 if connection is closed during transfer
 *	returns -3 if non fatal error - loop back for new request
 *	error_resp.msg will contain a message if rc is negative
 *=====================================================================*/
int put_exist_gwuser(int ofd, int component, int msglen) {
    int rc;
    datum userkey, userdata;
    datum groupkey, groupdata;
    GWUSER_ENCRYPT encrypted_data;
    GWUSER_DECRYPT decrypted_data;
    struct GWUSER moduser;	/* Original record	*/
    struct GWUSER user;		/* New record		*/
    char buf[BUFSIZ];
    char *cp;
    RESPPKT pkt;

    if (component != CGWUSER && component != CGWGROUP) {
	strcpy(error_resp.msg,"Component not GWUSER or GWGROUP for "
		"Add/Mod User");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    /*======================================================================
     * Read in the rest of the request to get the user/group info
     * Assume the terminating null is in the data stream
     *=====================================================================*/
    if (msglen <= 0) {
	strcpy(error_resp.msg,"Mod gwuser/gwgroup request contains no info");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    /*=======================================================================
     * Make sure the gwuser info matches our expected length
     *======================================================================*/
    if (msglen != sizeof(user)) {
	sprintf(error_resp.msg,"Info following a Mod gwuser/gwgroup request "
		"does not match the expected length of %d",sizeof(user));
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    if ((rc = getdata(ofd,(char *)&user,msglen,ckeysched)) < 0) {
	strcpy(error_resp.msg,"padmin cannot read user/group info");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(rc);
    }
    if (rc == 0)
	return(CLOSED);

    /*=======================================================================
     * Make sure user/group name is non_blank
     *======================================================================*/
    if ((cp = strtok(user.username," \t")) == NULL) {
	strcpy(error_resp.msg,"User/Group name must be nonblank");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    /*=======================================================================
     * If GWGROUP, call its processor instead
     *======================================================================*/
    if (component == CGWGROUP)
	return(put_exist_gwgroup(ofd, &user));

    /*=======================================================================
     * Open NDM database
     *======================================================================*/
    if ((rc = open_db(EXIST)) < 0) {
	strcpy(error_resp.msg,"Open of gwuser database failed");
	syslog(LOG_ERR,"Error: %s", error_resp.msg);
	return(NONFATAL);
    }

    /*=======================================================================
     * Check that user record exists
     *======================================================================*/
    userkey.dptr = user.username;
    userkey.dsize = strlen(user.username);
    userdata = dbm_fetch(db,userkey);
    if (userdata.dptr == NULL) {
	sprintf(error_resp.msg,"User %s does not exist.",userkey.dptr);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    /*=======================================================================
     * Check that record is actually for a user
     *======================================================================*/
    if (*(char *)userdata.dptr == GROUP) {
	sprintf(error_resp.msg,"%s is a group, cannot modify.", userkey.dptr);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close_db();
	return(NONFATAL);
    }
    else if (*(char *)userdata.dptr != USER) {
	sprintf(error_resp.msg,"%s is not a user, cannot modify.",
			userkey.dptr);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    /*=======================================================================
     * Check admin group priviledge
     *======================================================================*/
    decrypt_record(userdata.dptr,decrypted_data,userdata.dsize);
    memcpy(&moduser,decrypted_data,sizeof(moduser));
    if (strlen(admin.group) > 0) {
	if (strcmp(admin.group,moduser.groupname)) {
	    sprintf(error_resp.msg,"Admin %s can modify users in group %s "
			"only, user %s not modified.",
			admin.user,admin.group,moduser.username);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    close_db();
	    return(NONFATAL);
	}
	if (strcmp(user.groupname,moduser.groupname)) {
	    sprintf(error_resp.msg,"Must be root user to change group "
			"(admin = %s).", admin.user);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    close_db();
	    return(NONFATAL);
	}
    }

    /*=======================================================================
     * Group must exist
     *======================================================================*/
    groupkey.dptr = user.groupname;
    groupkey.dsize = strlen(user.groupname);
    groupdata = dbm_fetch(db,groupkey);
    if (groupdata.dptr == NULL) {
	sprintf(error_resp.msg,"Group %s does not exist, user %s not modified.",
		groupkey.dptr,userkey.dptr);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close_db();
	return(NONFATAL);
    }
    else if (*(char *)groupdata.dptr != GROUP) {
	sprintf(error_resp.msg,"%s is not a group, user %s not modified.",
		groupkey.dptr,userkey.dptr);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    /*=======================================================================
     * If password changed, process it
     *======================================================================*/
    if (strlen(user.password[0]) > 0) {
	int i;

	/*===================================================================
	 * Check new password to make sure it meets restrictions in gwuser.conf
	 *==================================================================*/
	if (check_password(user.password[0], moduser.password, &conf,
		error_resp.msg) < 0) {
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close_db();
	return(NONFATAL);
	}

	/*===================================================================
	 * Save old passwords.  Copy passwords from old moduser struct to
	 * new user struct
	 *==================================================================*/
	for (i = MAXPWDS-1; i > 0; i--)
	    strncpy(user.password[i],moduser.password[i-1],PWDSIZ);

	/*===================================================================
	 * Update passowrd mod time
	 *==================================================================*/
	user.passworddate = time((time_t *)NULL);
    }
    else
	/*===================================================================
	 * Password not changed, just copy all old and existing passwords
	 *==================================================================*/
	memcpy(user.password,moduser.password,sizeof(user.password));

    /*=======================================================================
     * Make sure start and stop dates are consistent
     *======================================================================*/
    if (user.stopdate > 0 && user.startdate > user.stopdate) {
	strcpy(error_resp.msg,"Start date cannot be later than stop date.");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    /*=======================================================================
     * Update user record
     *======================================================================*/
    user.type = moduser.type;
    user.no_login_attempts = 0;
    user.recorddate = time((time_t *)NULL);
    encrypt_record((char *)&user,encrypted_data,sizeof(user));
    userdata.dptr = encrypted_data;
    userdata.dsize = encrypted_sizeof(user);
    if (dbm_store(db,userkey,userdata,DBM_REPLACE)) {
	sprintf(error_resp.msg,"dbm_store failed for %s.",userkey.dptr);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    /*=======================================================================
     * Update summary information record
     *======================================================================*/
    info_db(GWINFO_MODIFY,MODIFYDATE,0,0);

    close_db();

    /*=======================================================================
     * Send an OK response to the client
     *======================================================================*/
    pkt.hdr.rc = OK;
    sprintf(pkt.msg,"User %s modified.",userkey.dptr);
    pkt.hdr.msglen = strlen(pkt.msg) + 1;
    order_bytes_in_resp(&(pkt.hdr));
    if ((rc = putdata(ofd, (char *)&pkt, sizeof(SCRESP) + ntohl(pkt.hdr.msglen),
			ckeysched)) < 0) {
	syslog(LOG_ERR,"Error: putdata() failed");
	close_db();
	return(rc);
    }

    return(OK);
}

/*====================================================================
 * Modifies an existing gwgroup record from info received in a MODGWUSER
 * request that has a component of CGWGROUP
 * Since the client always sends a GWUSER structure, this function translates
 * the GWUSER structure to a GWGROUP structure before processing.
 * input:
 *	ofd	file descriptor for socket to client
 *	user	pointer to the GWUSER structure that was just received
 * output:
 *	returns 0 if OK
 *	returns -1 if failed - connection should be closed
 *	returns -2 if connection is closed during transfer
 *	returns -3 if non fatal error - loop back for new request
 *	error_resp.msg will contain a message if rc is negative
 *=====================================================================*/
int put_exist_gwgroup(int ofd, struct GWUSER *user) {
    int rc;
    struct GWGROUP group;
    datum key, data;
    GWGROUP_ENCRYPT encrypted_data;
    RESPPKT pkt;

    memset(&group, 0, sizeof(group));
    /*=======================================================================
     * Translate the received GWUSER structure to a GWGROUP structure
     *======================================================================*/
    group.type = GROUP;
    strcpy(group.groupname, user->username);
    strcpy(group.realname, user->realname);
    group.tokentype = user->tokentype;
    group.tnaccess = user->tnaccess;
    group.ftpaccess = user->ftpaccess;
    memcpy(group.accesstimes, user->accesstimes, sizeof(group.accesstimes));
    group.recorddate = time((time_t *)NULL);

    /*=======================================================================
     * Check for root user
     *======================================================================*/
    if (strlen(admin.group) != 0) {
	sprintf(error_resp.msg, "Must be root to modify group (admin = %s).",
		admin.user);
	syslog(LOG_NOTICE, error_resp.msg);
	return(NONFATAL);
    }

    /*=======================================================================
     * Open NDM database
     *======================================================================*/
    if ((rc = open_db(EXIST)) < 0) {
	strcpy(error_resp.msg,"Open of gwuser database failed");
	syslog(LOG_ERR,"Error: %s", error_resp.msg);
	return(NONFATAL);
    }

    /*=======================================================================
     * Check that group record doesn't exist
     *======================================================================*/
    key.dptr = group.groupname;
    key.dsize = strlen(group.groupname);
    data = dbm_fetch(db,key);
    if (data.dptr == NULL) {
	sprintf(error_resp.msg,"Group %s doesn't exist.", key.dptr);
	syslog(LOG_ERR,"Error: %s", error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    /*=======================================================================
     * Check that record is actually for a group
     *======================================================================*/
    if (*(char *)data.dptr == USER) {
	sprintf(error_resp.msg,"%s is a user, cannot modify as a group.",
			key.dptr);
	syslog(LOG_ERR,"Error: %s", error_resp.msg);
	close_db();
	return(NONFATAL);
    }
    else if (*(char *)data.dptr != GROUP) {
	sprintf(error_resp.msg,"%s is not a group, cannot modify.", key.dptr);
	syslog(LOG_ERR,"Error: %s", error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    /*=======================================================================
     * Store new group record
     *======================================================================*/
    encrypt_record((char *)&group,encrypted_data,sizeof(group));
    data.dptr = encrypted_data;
    data.dsize = encrypted_sizeof(group);
    if (dbm_store(db,key,data,DBM_REPLACE)) {
	sprintf(error_resp.msg, "dbm_store failed for %s", key.dptr);
	syslog(LOG_ERR,"Error: %s", error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    /*=======================================================================
     * Update summary information record
     *======================================================================*/
    info_db(GWINFO_MODIFY,MODIFYDATE,0,0);

    close_db();

    /*=======================================================================
     * Send an OK response to the client
     *======================================================================*/
    pkt.hdr.rc = OK;
    sprintf(pkt.msg,"Group %s modified.",key.dptr);
    pkt.hdr.msglen = strlen(pkt.msg) + 1;
    order_bytes_in_resp(&(pkt.hdr));
    if ((rc = putdata(ofd, (char *)&pkt, sizeof(SCRESP) + ntohl(pkt.hdr.msglen),
			ckeysched)) < 0) {
	syslog(LOG_ERR,"Error: putdata() failed");
	return(rc);
    }

    return(OK);
}

/*====================================================================
 * Deletes an existing gwuser or gwgroup record
 * input:
 *	ofd	file descriptor for socket to client
 *	component the component is GWUSER or GWGROUP
 *	msglen	is the length of the user/group name that follows in the
 *		request
 * output:
 *	returns 0 if OK
 *	returns -1 if failed - connection should be closed
 *	returns -2 if connection is closed during transfer
 *	returns -3 if non fatal error - loop back for new request
 *	error_resp.msg will contain a message if rc is negative
 *=====================================================================*/
int delgwrecord(int ofd, int component, int msglen) {
    char username[UNMSIZ];
    int rc;
    datum key, data;
    GWUSER_DECRYPT decrypted_data;
    struct GWUSER user;
    struct GWGROUP group;
    char * gwbuf;
    int gwlen;
    RESPPKT pkt;

    if (component != CGWUSER && component != CGWGROUP) {
	strcpy(error_resp.msg,"DELGWUSER request must target GWUSER or "
		"GWGROUP");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    if (component == CGWUSER) {
	gwbuf = (char *)&user;
	gwlen = sizeof(user);
    }
    else {
	gwbuf = (char *)&group;
	gwlen = sizeof(group);
    }

    /*======================================================================
     * Read in the rest of the request to get the user/group name
     * Assume the terminating null is in the data stream
     *=====================================================================*/
    if (msglen <= 0) {
	strcpy(error_resp.msg,"Delete gwuser request contains empty name");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    /*=======================================================================
     * Make sure the name fits into our buffer
     *======================================================================*/
    if (msglen > sizeof(username)) {
	strcpy(error_resp.msg,"User/Group name too long");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    if ((rc = getdata(ofd,username,msglen,ckeysched)) < 0) {
	strcpy(error_resp.msg,"padmin delgwuser cannot read user/group name");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(rc);
    }
    if (rc == 0)
	return(CLOSED);

    /*======================================================================
     * Open NDBM database.
     *    The function open_db() locks the entire data base.
     *    The function close_db() unlocks the data base.
     *======================================================================*/
    if ((rc = open_db(EXIST)) < 0) {
	strcpy(error_resp.msg,"Open of gwuser database failed");
	syslog(LOG_ERR,"Error: %s", error_resp.msg);
	return(NONFATAL);
    }

    /*=======================================================================
     * Check that record exists
     *======================================================================*/
    key.dptr = username;
    key.dsize = strlen(username);
    data = dbm_fetch(db,key);
    if (data.dptr == NULL) {
	if (component == CGWUSER)
	    sprintf(error_resp.msg,"User %s doesn't exist.",key.dptr);
	else
	    sprintf(error_resp.msg,"Group %s doesn't exist.",key.dptr);
	syslog(LOG_NOTICE,error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    decrypt_record(data.dptr,decrypted_data,data.dsize);
    memcpy(gwbuf,decrypted_data,gwlen);

    /*=======================================================================
     * Check that record is actually for a user if component GWUSER and
     * for a group if component GWGROUP
     *======================================================================*/
    if (component == CGWUSER && *(char *)data.dptr != USER) {
	sprintf(error_resp.msg,"%s is not a user, cannot delete.",key.dptr);
	syslog(LOG_NOTICE,error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    if (component == CGWGROUP && *(char *)data.dptr != GROUP) {
	sprintf(error_resp.msg,"%s is not a group, cannot delete.",key.dptr);
	syslog(LOG_NOTICE,error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    /*=======================================================================
     * If GWGROUP, check that no users exist with this group
     *======================================================================*/
    if (component == CGWGROUP) {
	for (key = dbm_firstkey(db); key.dptr != NULL; key = dbm_nextkey(db)) {
	    data = dbm_fetch(db,key);
	    if (data.dptr == NULL) {
		sprintf(error_resp.msg,"dbm_fetch failed for %s.",key.dptr);
		syslog(LOG_ERR,"Error: %s",error_resp.msg);
		close_db();
		return(NONFATAL);
	    }
	    if (*(char *)data.dptr == USER) {
		decrypt_record(data.dptr,decrypted_data,data.dsize);
		memcpy(&user,decrypted_data,sizeof(user));
		if (!strcmp(user.groupname,group.groupname)) {
		    sprintf(error_resp.msg,"User(s) exist with this group, "
			"cannot delete.");
		    syslog(LOG_NOTICE,error_resp.msg);
		    close_db();
		    return(NONFATAL);
		}
	    }
	}
	/*===================================================================
	 * Reset pointers so the correct group record is deleted
	 *==================================================================*/
	key.dptr = group.groupname;
	key.dsize = strlen(group.groupname);
    }

    /*=======================================================================
     * Make sure user has right to delete user/group
     * To delete a group, user must be root
     * To delete a user, user must be in that group
     *======================================================================*/
    if (component == CGWGROUP && strlen(admin.group) > 0) {
	sprintf(error_resp.msg,"Must be root user to delete a group");
	syslog(LOG_NOTICE,error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    if (component == CGWUSER && strlen(admin.group) > 0 &&
		strcmp(admin.group,user.groupname)) {
	sprintf(error_resp.msg,"Admin %s can delete users from group %s only, "
		"user %s not deleted.",admin.user,admin.group,user.username);
	syslog(LOG_NOTICE,error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    /*=======================================================================
     * Delete record
     *======================================================================*/
    if (dbm_delete(db,key)) {
	sprintf(error_resp.msg,"dbm_delete failed for %s.",key.dptr);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	close_db();
	return(NONFATAL);
    }

    /*=======================================================================
     * Update summary information record
     *======================================================================*/
    if (component == CGWUSER)
	info_db(GWINFO_MODIFY,NUMUSERS,-1,0);
    else
	info_db(GWINFO_MODIFY,NUMGROUPS,0,-1);

    /*=======================================================================
     * Send an OK response to the client
     *======================================================================*/
    pkt.hdr.rc = OK;
    if (component == CGWUSER)
	sprintf(pkt.msg,"User %s deleted.",key.dptr);
    else
	sprintf(pkt.msg,"Group %s deleted.",key.dptr);
    pkt.hdr.msglen = strlen(pkt.msg) + 1;
    order_bytes_in_resp(&(pkt.hdr));
    if ((rc = putdata(ofd, (char *)&pkt, sizeof(SCRESP) + ntohl(pkt.hdr.msglen),
			ckeysched)) < 0) {
	syslog(LOG_ERR,"Error: putdata() failed");
	return(rc);
    }

    close_db();
    return(OK);
}
