/*============================================================================
 * filename:		padmin/proxy.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1998-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1998-2000.  
 * All rights reserved.   
 *
 * Date written:	June	  24, 1998
 * Last modified:	February   8, 2000 - added LPL
 *
 *===========================================================================*/
#include <stdio.h>
#include <errno.h>
#include <syslog.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#if !defined(__hppa) && !defined(__Lynx__)
#include <sys/select.h>
#endif
#include "firewall.h"
#include "gwuser.h"
#include "des.h"
#include "cs.h"
#include "enc.h"

#define IS_NUMERIC_IP(a)        (strspn(a,"0123456789.") == strlen(a))

#ifndef linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif

extern RESPPKT error_resp;
extern des_key_schedule ckeysched;
extern des_key_schedule dkeysched;
extern PACONF *padmconf;
extern  struct sockaddr_in addr;                /* Remote host's address */
extern struct GWADMIN admin;

int getdata(int fd, char *buf, int len, des_key_schedule keysched);
int putdata(int fd, char *buf, int len, des_key_schedule keysched);

int proxycallback(int *clientfd, int *serverfd) {
    int rc;
    pid_t pid;
    int len;
    int newfd;
    int one = 1;
    short clientport;
    char urg[4];
    char buf[BUFSIZ];
    struct sockaddr_in myaddr;
    struct sockaddr_in serveraddr;
    struct sockaddr_in clientaddr;
    CSREQ req;
    struct {
	CSREQ	req;
	char	buf[BUFSIZ];
    } pkt;

    /*======================================================================
     * Read the next few bytes as a client to server request
     *=====================================================================*/
#ifdef __sun
    if ((rc = enc_read(*clientfd,(char *)&req,sizeof(CSREQ),dkeysched)) < 0) {
#else
    if ((rc = getdata(*clientfd, (char *)&req, sizeof(CSREQ), dkeysched)) < 0) {
#endif
	syslog(LOG_ERR,"Error: Cannot get proxy request from client.  %s",
		sys_errlist[errno]);
	return(NONFATAL);
    }

    if (rc == 0) {
	syslog(LOG_ERR,"Error: Connection timed out");
	return(NONFATAL);
    }
    order_bytes_in_req(&req);

    if (req.cmd != CALLBACK || req.component != CLOGIN) {
	syslog(LOG_ERR,"Error: Did not receive the expected CALLBACK request ");
	return(NONFATAL);
    }

    /*======================================================================
     * Read in the request to get the server's address
     * Assume the terminating null is in the data stream
     *=====================================================================*/
    if (req.msglen <= 0) {
	strcpy(error_resp.msg,"Proxy request contains empty ip address");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    /*=======================================================================
     * Make sure the port number fits into our buffer
     *======================================================================*/
    if (req.msglen > sizeof(buf)) {
	strcpy(error_resp.msg,"Remote host name too long");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    if (getdata(*clientfd,buf,req.msglen,dkeysched) < 0) {
	strcpy(error_resp.msg,"padmin proxy cannot read client port number");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }
/* decrypt ??? */

    clientport = atoi(buf);

    if (clientport == 0) {
	strcpy(error_resp.msg,"Invalid client callback port number in proxy "
			"request");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    /*======================================================================
     * Get the address of the server
     *=====================================================================*/
    len = sizeof(serveraddr);
    if (getpeername(*serverfd,(struct sockaddr *)&serveraddr,
						(socklen_t *)&len) < 0) {
	syslog(LOG_ERR,"Error: Cannot get server's address.  %s",
		sys_errlist[errno]);
	return(NONFATAL);
    }

    /*=======================================================================
     * Create the socket and bind a port
     *======================================================================*/
    if ((newfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	sprintf(error_resp.msg,"Error creating socket.  %s",sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }
    if (setsockopt(newfd,SOL_SOCKET,SO_REUSEADDR,
				(char *)&one,sizeof(int)) < 0)
	syslog(LOG_ERR,"Error: setsockopt failed.  %s",sys_errlist[errno]);

/*
    if (setsockopt(newfd,SOL_SOCKET,SO_OOBINLINE,(char *)&one,sizeof(int)) < 0)
	syslog(LOG_ERR,"Error: setsockopt failed %s\n",sys_errlist[errno]);
*/

    memset(&myaddr, 0, sizeof(myaddr));
    myaddr.sin_family = AF_INET;
    myaddr.sin_port = INADDR_ANY;
    myaddr.sin_addr.s_addr = INADDR_ANY;
    if (bind(newfd,(struct sockaddr *)&myaddr,sizeof(serveraddr)) < 0) {
	sprintf(error_resp.msg,"Bind error.  %s",sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    /*======================================================================
     * Listen for the callback
     *=====================================================================*/
    if (listen(newfd,1) < 0) {
	syslog(LOG_ERR,"listen() failed.  %s",sys_errlist[errno]);
	return(NONFATAL);
    }

    /*======================================================================
     * Get the bound port
     *=====================================================================*/
    len = sizeof(myaddr);
    if (getsockname(newfd, (struct sockaddr *)&myaddr, (socklen_t *)&len) < 0) {
	syslog(LOG_ERR,"Error: Cannot get the bound port number");
	return(NONFATAL);
    }

    /*======================================================================
     * Send a requset to the server to callback so another socket can be opened
     * between the server and the client
     *=====================================================================*/
    sprintf(pkt.buf,"%d",myaddr.sin_port);
    pkt.req.cmd = htonl(CALLBACK);
    pkt.req.component = htonl(CLOGIN);
    pkt.req.msglen = htonl(strlen(pkt.buf) + 1);

/* encrypt ??? */
    /*======================================================================
     * Send the request
     *=====================================================================*/
    memset(&urg,0,sizeof(urg));
    if ((rc = send(*serverfd, urg, 1, MSG_OOB)) < 0) {
	syslog(LOG_ERR,"Error: Cannont send OOB data.  %s",sys_errlist[errno]);
	return(NONFATAL);
    }

    if ((rc = putdata(*serverfd, (char *)&pkt, sizeof(CSREQ) + pkt.req.msglen,
						dkeysched)) < 0) {
	syslog(LOG_ERR,"Error: Cannot send a CALLBACK request.  %s",
		sys_errlist[errno]);
	return(NONFATAL);
    }

    /*======================================================================
     * Hold here for the server's connect back to our port
     *=====================================================================*/
    if ((rc = accept(newfd,0,0)) < 0) {
	syslog(LOG_ERR,"Error: accept() failed.  %s",sys_errlist[errno]);
	return(NONFATAL);
    }

    /*======================================================================
     * Close the file descriptor that is no longer needed and then save
     * the file descriptor returned from accept()
     * Don't close the original fd (unhand(handle)->fd) here since the
     * caller is still using it
     *=====================================================================*/
    close(newfd);
    newfd = rc;

    /*======================================================================
     * fork a child to handle further activities
     *=====================================================================*/
    if ((pid = fork()) < 0) {
	strcpy(error_resp.msg,"Cannot fork child for proxy");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(NONFATAL);
    }

    if (pid != 0) {
	close(newfd);
	return(0);
    }

    close(*serverfd);
    *serverfd = newfd;

    if ((rc = callback(*clientfd, clientport)) < 0)
	exit(rc);

    close(*clientfd);
    *clientfd = rc;

    return(0);
}

#define terminate_and_return()	{error_resp.hdr.rc = FAILED; \
				error_resp.hdr.msglen =  \
					strlen(error_resp.msg) + 1; \
				order_bytes_in_resp(&(error_resp.hdr)); \
				return(NONFATAL); }

/*===========================================================================
 * This function implements the proxy for T.Rex remote admin
 * It spawns a child to handle the proxy and the parent returns to the
 * main loop
 *
 * If successful, a connection is established on the server side.
 * If unsuccessful, this function returns -1 if component is not on our list
 *==========================================================================*/
int proxy(int clientfd, int msglen) {
    int serverfd;
    int maxfd;
    short serverport;
    int one=1;
    int len;
    int serverbytes=0;
    int clientbytes=0;
    char buf[BUFSIZ];
    char strcip[16];
    char strsip[16];
    char ip[BUFSIZ];
    char *tmp;
    pid_t pid;
    struct sockaddr_in clientaddr;
    struct sockaddr_in serveraddr;
    SCRESP resp;

    /*=======================================================================
     * Make sure the hostname/ip fits into our buffer
     *======================================================================*/
    if (msglen <= 0) {
	strcpy(error_resp.msg,"Proxy request contains empty ip address");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	terminate_and_return();
    }

    if (msglen > sizeof(buf)) {
	strcpy(error_resp.msg,"Remote host name too long");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	terminate_and_return();
    }

    /*======================================================================
     * Read in the request to get the server's address
     *=====================================================================*/
    if ((len = getdata(clientfd, buf, msglen, ckeysched)) < 0) {
	strcpy(error_resp.msg,"padmin proxy cannot read server ip address");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	terminate_and_return();
    }

    if (len == 0) {
	syslog(LOG_NOTICE,"Connection timed out.  Remote host %s",
			inet_ntoa(addr.sin_addr));
	return(CLOSED);
    }

    strcpy(ip,buf);
    tmp = strtok(ip,":");
    tmp = strtok(NULL,":");
    serverport = tmp ? atoi(tmp) : DEFAULT_PADMIN_PORT;

    /*=======================================================================
     * Create the socket and bind a port
     *======================================================================*/
    if ((serverfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	sprintf(error_resp.msg,"Error creating socket.  %s",sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	terminate_and_return();
    }
    if (setsockopt(serverfd,SOL_SOCKET,SO_REUSEADDR,
				(char *)&one,sizeof(int)) < 0)
	syslog(LOG_ERR,"Error: setsockopt error.  %s",sys_errlist[errno]);

    memset(&serveraddr, 0, sizeof(serveraddr));
    serveraddr.sin_family = AF_INET;
    serveraddr.sin_port = INADDR_ANY;
    serveraddr.sin_addr.s_addr = INADDR_ANY;
    if (bind(serverfd,(struct sockaddr *)&serveraddr,sizeof(serveraddr)) < 0) {
	sprintf(error_resp.msg,"Bind error.  %s",sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	shutdown(serverfd,2);
	terminate_and_return();
    }

    /*======================================================================
     * If remote host is specified as a name, look it up
     *=====================================================================*/
    if (!IS_NUMERIC_IP(ip)) {
	struct hostent *Hostent;
	if ((Hostent = gethostbyname(buf)) == NULL) {
	    char tmp[BUFSIZ];
	    sprintf(error_resp.msg,"Host lookup error.  %s",sys_errlist[errno]);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    shutdown(serverfd,2);
	    terminate_and_return();
	}
	memcpy((char *)&(serveraddr.sin_addr.s_addr), Hostent->h_addr,
                                Hostent->h_length);
    }
    else
	serveraddr.sin_addr.s_addr = inet_addr(buf);

    if (serveraddr.sin_addr.s_addr == INADDR_NONE ||
		serveraddr.sin_addr.s_addr == INADDR_ANY) {
	strcpy(error_resp.msg,"Invalid host IP address");
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	shutdown(serverfd,2);
	terminate_and_return();
    }

    /*=======================================================================
     * Make sure a proxy request does not connect to 127.0.0.1 which can
     * be potentially used by an admin to obtain rootuser access 
     *======================================================================*/
    if (serveraddr.sin_addr.s_addr == htonl(0x7f000001)) {
	syslog(LOG_ALERT,"Security Alert: User %s attempted to proxy to "
			"the loopback address 127.0.0.1",admin.user);
	strcpy(error_resp.msg,"Proxy request denied");
	shutdown(serverfd,2);
	terminate_and_return();
    }

    /*=======================================================================
     * Connect to the server
     *======================================================================*/
    serveraddr.sin_port = htons(serverport);

    if (connect(serverfd, (struct sockaddr *)&serveraddr,
				sizeof(serveraddr)) < 0) {
	sprintf(error_resp.msg,"Cannot connect to server.  %s",
			sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	shutdown(serverfd,2);
	terminate_and_return();
    }

    /*=======================================================================
     * Get the response from the server
     *======================================================================*/
    if (getdata(serverfd, (char *)&resp, sizeof(SCRESP), dkeysched) <= 0) {
	sprintf(error_resp.msg,"Cannot get response from server %s.  %s",
		ip, sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	shutdown(serverfd,2);
	terminate_and_return();
    }
    resp.rc = ntohl(resp.rc);
    resp.msglen = ntohl(resp.msglen);

    if (resp.rc != OK) {
	sprintf(error_resp.msg,"Server %s rejected proxy connection request",
		ip);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	shutdown(serverfd,2);
	terminate_and_return();
    }

    /*=======================================================================
     * Get the client's address so we can log it
     *======================================================================*/
    len = sizeof(clientaddr);
    memset(&clientaddr,0,sizeof(clientaddr));
    if (getpeername(clientfd,(struct sockaddr *)&clientaddr,
					(socklen_t *)&len) < 0) {
	syslog(LOG_ERR,"Error: getpeername error.  %s",sys_errlist[errno]);
    }

    strcpy(strcip,inet_ntoa(clientaddr.sin_addr));
    strcpy(strsip,inet_ntoa(serveraddr.sin_addr));
    syslog(LOG_NOTICE,"padmin proxy connected source %s port %d to "
		"destination %s (%s) port %d",strcip,clientaddr.sin_port,
		strsip,ip,ntohs(serveraddr.sin_port));

    /*=======================================================================
     * Send a response to the client that everything is OK
     *======================================================================*/
    resp.rc = OK;
    resp.msglen = 0;
    order_bytes_in_resp(&resp);
    if (putdata(clientfd, (char *)&resp, sizeof(SCRESP), ckeysched) < 0) {
	syslog(LOG_ERR,"Error: Failed to send response to client");
	shutdown(serverfd,2);
	terminate_and_return();
    }

    /*=======================================================================
     * Loop to implement the proxy
     *======================================================================*/
    maxfd = serverfd > clientfd ? serverfd : clientfd;
    while (1) {
	int rc;
	int nread;
	char urg[4];
	struct timeval timeout;
	fd_set rdy;
	fd_set xcpt;

	FD_ZERO(&rdy);
	FD_ZERO(&xcpt);
	FD_SET(serverfd,&rdy);
	FD_SET(clientfd,&rdy);
	FD_SET(clientfd,&xcpt);

	timeout.tv_sec = padmconf->timeout;
	timeout.tv_usec = 0;

#ifdef __hppa
	if((rc = select(maxfd+1,(int *)&rdy,0,(int *)&xcpt,&timeout)) < 0) {
#else
	if((rc = select(maxfd+1,&rdy,(fd_set *)0,(fd_set *)&xcpt,
							&timeout)) < 0) {
#endif
	    syslog(LOG_ERR,"Error: select failed. %s.", sys_errlist[errno]);
	    break;
	}

	if (rc == 0) {
	    syslog(LOG_NOTICE,"Connection timed out.");
	    break;
	}

	/*==================================================================
	 * Callback request from client
	 *=================================================================*/
	if (FD_ISSET(clientfd,&xcpt)) {
	    if ((rc = recv(clientfd, urg, 1, MSG_OOB)) < 0) {
		syslog(LOG_ERR,"Error: Error receiving OOB byte from client.  "
		"%s", sys_errlist[errno]);
		exit(-1);
	    }

	    /*==============================================================
	     * Process the following bytes as a callback request
	     *=============================================================*/
	    proxycallback(&clientfd,&serverfd);
	    continue;
	}

	/*==================================================================
	 * Input from client
	 *=================================================================*/
	if (FD_ISSET(clientfd,&rdy)) {
	    if ((nread = read(clientfd, buf, sizeof(buf))) < 0) {
		syslog(LOG_ERR,"Error: Read error from client %s  %s",
			strcip,sys_errlist[errno]);
		break;
	    }

	    if (nread == 0)
		break;

	    if (write(serverfd, buf, nread) < 0) {
		syslog(LOG_ERR,"Error: Write error from server %s  %s",
			strsip,sys_errlist[errno]);
		break;
	    }
	    clientbytes += nread;
	}

	/*==================================================================
	 * Input from server
	 *=================================================================*/
	if (FD_ISSET(serverfd,&rdy)) {
	    if ((nread = read(serverfd, buf, sizeof(buf))) < 0) {
		syslog(LOG_ERR,"Error: Read error from server %s  %s",
			strsip,sys_errlist[errno]);
		break;
	    }

	    if (nread == 0)
		break;

	    if (write(clientfd, buf, nread) < 0) {
		syslog(LOG_ERR,"Error: Write error from client %s  %s",
			strcip,sys_errlist[errno]);
		break;
	    }
	    serverbytes += nread;
	}
    }

    syslog(LOG_NOTICE,"padmin proxy exited source %s port %d (input bytes = "
		"%d) to destination %s (%s) port %d (input bytes = %d)",
		strcip,clientaddr.sin_port,clientbytes,
		strsip,ip,serveraddr.sin_port,serverbytes);
    exit(0);
}
