/*============================================================================
 * filename:		padmin/runcmd.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1999-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1999-2000.  
 * All rights reserved.   
 *
 * Date written:	June	  24, 1998
 * Last modified:	February   8, 2000 - added LPL
 *
 *===========================================================================*/
#include <stdio.h>
#include <errno.h>
#include <syslog.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#if !defined(__hppa) && !defined(__Lynx__)
#include <sys/select.h>
#endif
#ifdef _AIX41
#include <procinfo.h>
#else
#ifdef __sun
#define         _KMEMUSER 1
#include <sys/proc.h>
#include <kvm.h>
#include <fcntl.h>
#else
#ifdef __hppa
#include <sys/pstat.h>
#else
#ifdef __Lynx__
#include <info.h>
#include <proc.h>
#else
#ifdef linux
#include <dirent.h>
#endif
#endif
#endif
#endif
#endif

#include "des.h"
#include "cs.h"
#include "gwuser.h"

int getdata(int fd, char *buf, int len, des_key_schedule ckeysched);
int getsomedata(int fd, char *buf, int len, des_key_schedule ckeysched);
int putdata(int fd, char *buf, int len, des_key_schedule ckeysched);

extern struct GWADMIN admin;

#ifndef linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif

extern RESPPKT error_resp;
extern des_key_schedule ckeysched;
extern des_key_schedule dkeysched;

/*===========================================================================
 * Returns the command for the component passed as an argument
 * Returns -1 if component is not on our list
 *==========================================================================*/
int trexcmd(int component, char *path) {
    switch (component) {
	case CAPACHE:
	    strcpy(path,"/usr/local/etc/httpd");
	    return(0);
	case CAPROXY:
	    strcpy(path,"/usr/local/etc/aproxy");
	    return(0);
	case CFWPULSE:
	    strcpy(path,"/usr/local/etc/fwpulse");
	    return(0);
	case CFWPULSED:
	    strcpy(path,"/usr/local/etc/fwpulsed");
	    return(0);
	case CHTTPD:
	    strcpy(path,"/usr/local/etc/httpd_3.0A");
	    return(0);
	case CINETD:
	    strcpy(path,"/usr/sbin/inetd");
	    return(0);
	case CPADMIN:
	    strcpy(path,"/usr/local/etc/padminsvr");
	    return(0);
	case CFWMON:
	    strcpy(path,"/usr/local/etc/fwmon");
	    return(0);
	case CSOCKD:
	    strcpy(path,"/usr/local/etc/sockd &");
	    return(0);
	case CSMWRAPD:
	    strcpy(path,"/usr/local/etc/smwrapd");
	    return(0);
	case CSMALIASES:
	    strcpy(path,"/usr/local/etc/adam -b");
	    return(0);
	case CSYSLOG:
	    strcpy(path,"/usr/sbin/syslogd");
	    return(0);
	case CRPCPROXY:
	    strcpy(path,"/usr/local/etc/rpcproxy");
	    return(0);
	case CWEBGATE:
	    strcpy(path,"/usr/local/etc/webgate");
	    return(0);
	default:
	    return(-1);
    }
}

/*=======================================================================
 * Input:
 *	input	input buffer containing the argument
 *	buf	buffer of BUFSIZ to hold the arguments
 *	incl1	returned ptr to incl1 string
 *	incl2	returned ptr to incl2 string
 *	excl	returned ptr to excl string
 * Returns	0	if OK
 *		NONFATAL	if failed
 *======================================================================*/
int inclexcl(char *input, char *strbuf,
			char **incl1, char **incl2, char **excl) {
    char * cp;

    *incl1 = NULL;
    *incl2 = NULL;
    *excl = NULL;
    memcpy(strbuf, input, BUFSIZ);
    if ((cp = strtok(strbuf," ")) == NULL)
	return(0);

    /*=======================================================================
     * 1st include
     *======================================================================*/
    if (strcmp(cp,"include") == 0) {
	if ((cp = strtok(NULL,"\"")) == NULL) {
	    strcpy(error_resp.msg,"Empty include search argument");
	    return(NONFATAL);
	}
	*incl1 = cp;
	if ((cp = strtok(NULL," ")) == NULL)
	    return(0);

	/*===============================================================
	 * 2nd include
	 *==============================================================*/
	if (strcmp(cp,"include") == 0) {
	    if ((cp = strtok(NULL,"\"")) == NULL) {
		strcpy(error_resp.msg,"Empty include search argument");
		return(NONFATAL);
	    }
	    *incl2 = cp;
	    if ((cp = strtok(NULL," ")) == NULL)
		return(0);
	}
    }

    /*===============================================================
     * exclude
     *==============================================================*/
    if (strcmp(cp,"exclude") == 0) {
	if ((cp = strtok(NULL,"\"")) == NULL) {
	    strcpy(error_resp.msg,"Empty exclude search argument");
	    return(NONFATAL);
	}
	*excl = cp;
    }
}

/*=======================================================================
 * Given tokens in a '|' separated string in inclstr, determines if the
 * string line contains one of the tokens
 * Returns	0	if line doesn't contain any of the tokens
 *		1	if line contains one of the tokens
 *======================================================================*/
int incl(char *line, char *inclstr) {
    char *cp;
    char buf[BUFSIZ];

    strcpy(buf, inclstr);	/* Make a copy so we don't mess up the orig */
    for (cp = strtok(buf,"|"); cp != NULL; cp = strtok(NULL,"|")) {
	if (strstr(line,cp))
	    return(1);
    }
    return(0);
}

/*======================================================================
 * Run a requested command and send the output to the client
 * input:
 *	ofd	file descriptor for socket to client
 *	component the component to run
 *	msglen	the length of the argument string to be received
 * output:
 *	returns 0 if OK
 *	returns -1 if failed - connection should be closed
 *	returns -2 if connection is closed during transfer
 *	returns -3 if non fatal error - loop back for new request
 *	error_resp.msg will contain a message if rc is negative
 *=====================================================================*/
int taillog(int ofd, int component, int msglen) {
    int ifd;
    int rc;
    int pipefds[2];
    int maxfd;
    int cmdlen;
    int pause = 0;
    fd_set rdy;
    fd_set wdy;
    pid_t pid;
    char *incl1 = NULL;
    char *incl2 = NULL;
    char *excl = NULL;
    char buf[BUFSIZ];
    char strbuf[BUFSIZ];
    char fn[BUFSIZ];
    char *syslogfile;
    SCRESP resp;
    int leftover = 0;

    memset(fn,0,sizeof(fn));
    /*======================================================================
     * Set up the command buffer
     *=====================================================================*/
    switch (component) {
	case CSYSLOG:
	    /*===============================================================
	     * arg can contain include/exclude strings
	     *==============================================================*/
	    #ifdef _AIX
		syslogfile = "/var/adm/syslog";
		strcpy(buf,"/bin/tail ");
	    #else
		# ifdef __sun
		    syslogfile = "/var/adm/messages";
		    strcpy(buf,"/usr/xpg4/bin/tail ");
		# else
		    #  ifdef __hppa
			syslogfile = "/var/adm/syslog/syslog.log";
			strcpy(buf,"/bin/tail ");
		    #  else
			#  ifdef __Lynx__
			    syslogfile = "/etc/log/messages";
			    strcpy(buf,"/bin/tail ");
		 	# else
			    # ifdef linux
				syslogfile = "/var/log/messages";
				strcpy(buf,"/usr/bin/tail ");
			    # endif
			#  endif
		    #  endif
		# endif
	    #endif
	    strcpy(fn,syslogfile);
	    strcat(fn," ");
	    break;
	case CGWUSER:
	    /*===============================================================
	     * arg can contain \* or name or -a
	     *==============================================================*/
	    strcpy(buf,"/usr/local/etc/gwuser -l ");
	    break;
	case CGWGROUP:
	    /*===============================================================
	     * arg can contain \* or name or -a
	     *==============================================================*/
	    strcpy(buf,"/usr/local/etc/gwgroup -l ");
	    break;
	case CAPACHE:
	case CAPACHEERR: {
	    /*===============================================================
	     * arg can (potentially, but not yet) contain include/exclude
	     * strings
	     *==============================================================*/
	    FILE *httpfile;
	    char linebuf[BUFSIZ];
	    char *search;
	    if (component == CAPACHE)
		search = "TransferLog";
	    else
		search = "ErrorLog";
	    if (!(httpfile = fopen("/etc/apache/httpd.conf","r"))) {
		sprintf(error_resp.msg,"Open of /etc/apache/httpd.conf "
					"failed.  %s", sys_errlist[errno]);
		syslog(LOG_ERR,"Error: %s",error_resp.msg);
		if ((rc = flushreq(ofd, msglen, ckeysched)) < 0)
		    return(rc);
		return(NONFATAL);
	    }
	    #ifdef __sun
		strcpy(buf,"/usr/xpg4/bin/tail ");
	    #else
		#ifdef linux
		    strcpy(buf,"/usr/bin/tail ");
		#else
		    strcpy(buf,"/bin/tail ");
		#endif
	    #endif
	    while( fgets(linebuf, BUFSIZ, httpfile) != NULL) {
		char *cp;
		if ((cp = strtok(linebuf," \t\n")) == NULL)
			continue;
		if (strcasecmp(cp,search) != 0)
			continue;
		if ((cp = strtok(NULL," \t\n")) == NULL) {
		    sprintf(error_resp.msg,"Encountered error on %s line "
				"in  /etc/apache/httpd.conf",search);
		    syslog(LOG_ERR,"Error: %s",error_resp.msg);
		    if ((rc = flushreq(ofd, msglen, ckeysched)) < 0)
			return(rc);
		    return(NONFATAL);
		}
		strcpy(fn,cp);
		strcat(fn," ");
		break;
	    }
	    if (feof(httpfile)) {
		sprintf(error_resp.msg,"No %s line found in "
				"/etc/apache/httpd.conf",search);
		syslog(LOG_ERR,"Error: %s",error_resp.msg);
		if ((rc = flushreq(ofd, msglen, ckeysched)) < 0)
		    return(rc);
		return(NONFATAL);
	    }
	    break;
	}
	case CPS:
	    /*===============================================================
	     * arg is empty
	     *==============================================================*/
#if defined(__Lynx__)
	    strcpy(buf,"/bin/ps ax ");
#else
#if defined(linux)
	    strcpy(buf,"/bin/ps lx ");
#else
	    strcpy(buf,"/bin/ps -ef ");
#endif
#endif
	    break;
	case CVMSTAT:
	    /*===============================================================
	     * arg can contain a number for the interval in seconds
	     *==============================================================*/
	    #ifdef linux
		strcpy(buf,"/usr/bin/vmstat ");
	    #else
		strcpy(buf,"/bin/vmstat ");
	    #endif
	    break;
	case CTNKEY:
	    /*===============================================================
	     * arg is a user name or name pattern
	     *==============================================================*/
	    strcpy(buf,"/usr/local/etc/tnkey `/usr/local/etc/gwuser -l \"");
	    break;
	default:
	    sprintf(error_resp.msg,"Component %d not supported\n",component);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    if ((rc = flushreq(ofd, msglen, ckeysched)) < 0)
		return(rc);
	    return(NONFATAL);
    }

    /*======================================================================
     * Process argument string if present
     *=====================================================================*/
    cmdlen = strlen(buf);
    if (msglen > 0) {

	/*==================================================================
	 * Make sure arg string fits in buf
	 *=================================================================*/
	if (msglen > sizeof(buf) - strlen(fn) - cmdlen) {
	    sprintf(error_resp.msg,"Argument string longer than %d",
						sizeof(buf) - cmdlen);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    if ((rc = flushreq(ofd, msglen, ckeysched)) < 0)
		return(rc);
	    return(NONFATAL);
	}

	/*==================================================================
	 * Get the argument string
	 *=================================================================*/
	if ((rc = getdata(ofd, &buf[cmdlen], msglen, ckeysched)) < 0) {
	    sprintf(error_resp.msg,"getdata() failed.  %s", sys_errlist[errno]);
	    syslog(LOG_ERR,"Error: %s",error_resp.msg);
	    return(-1);
	}
    }

    /*=======================================================================
     * Assume for these components, the arg is in the form
     * -60 -f include "kw" ...
     * I.e. the -number and -f always exist.
     * -60 -f include "kw" ...
     *    ^
     *    |
     * strchr() points here
     * then strtok() delivers the "-f"
     * then the subsequent strtok(NULL) in inclexcl() delivers "include if
     *	present
     *======================================================================*/
    if (component == CSYSLOG || component == CAPACHE
		|| component == CAPACHEERR || component == CHTTPD){
	char *cp;
#ifdef linux
	/* Make -60 into -n 60 */
	cp = &buf[cmdlen];
	if (*cp == '-' && isdigit(*(cp+1))) {
	    memmove(cp+2, cp+1, msglen-1);
	    *(cp + 1) = 'n';
	}
#endif
	cp = strchr(&buf[cmdlen],' ');	/* Skip over -number	*/
	cp = strtok(cp," ");	/* Skip over -f	*/
	if (inclexcl(cp + strlen(cp) + 1, strbuf, &incl1, &incl2, &excl) < 0)
	    return(NONFATAL);
	strcat(buf," ");
    }

    /*=======================================================================
     * Now merge in the filename we want to tail
     * Since the arg string from the client may be in the forms:
     * "" or
     * "-60 " or
     * "-60 include "Error""
     * and fn is always "filename "
     * If fn is present, we need to insert it between the - flags and the
     * first '|'.
     *======================================================================*/
    if (strlen(fn) > 0) {
	char *cp;
	if ((cp = strchr(buf,'|')) == NULL) {
	    strcat(buf,fn);
	}
	else {
	    /* Insert the filename in the middle just before '|' */
	    memmove(cp+strlen(fn), cp, strlen(cp));
	    memcpy(cp, fn, strlen(fn));
	}
    }

    /*=======================================================================
     * Make sure non-root users don't get to run tnkey
     *======================================================================*/
    if (component == CTNKEY) {
	if (strlen(admin.group) > 0) {
	    sprintf(error_resp.msg,"Must be root user to run tnkey, currently "
			"logged in as  %s", admin.user);
	    syslog(LOG_NOTICE,error_resp.msg);
	    return(NONFATAL);
	}
	strcat(buf, "\"`");
    }

    /*=======================================================================
     * Special processing to make sure non-root users don't see users outside
     * their group
     *======================================================================*/
    if (component == CGWUSER && strlen(admin.group) > 0) {
	strcat(buf," -g ");
	strcat(buf,admin.group);
    }

    /*=======================================================================
     * Special processing to make sure non-root users don't see any gwgroup
     * info
     *======================================================================*/
    if (component == CGWGROUP && strlen(admin.group) > 0) {
	sprintf(error_resp.msg,"Must be rootuser to list gwgroup, currently "
		"logged in as %s",admin.user);
	syslog(LOG_NOTICE,error_resp.msg);
	return(NONFATAL);
    }

    /*======================================================================
     * Send an OK response
     *=====================================================================*/
    resp.rc = OK;
    resp.msglen = 0;
    order_bytes_in_resp(&resp);

    if ((rc = putdata(ofd, (char *)&resp, sizeof(SCRESP), ckeysched)) < 0) {
	sprintf(error_resp.msg,"putdata() failed.  %s",
			sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(-1);
    }

    /*======================================================================
     * Create a pipe so we can receive output from the command
     *=====================================================================*/
    if (pipe(pipefds) < 0) {
	sprintf(error_resp.msg,"Cannot create pipe.  %s", sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(-1);
    }

    /*======================================================================
     * Fork a child to run the command
     *=====================================================================*/
    if ((pid = fork()) < 0) {
	sprintf(error_resp.msg,"Cannot create pipe.  %s", sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(-1);
    }

    /*======================================================================
     * The child sets up the pipe and then runs the command
     *=====================================================================*/
    if (pid == 0) {
	/*==================================================================
	 * Set up a process group for the child so it and its children can
	 * be terminated as a group when needed
	 *=================================================================*/
#ifdef __hppa
	setpgid(0, 0);
#else
	setpgrp();
#endif
	/*==================================================================
	 * Make the write side of the pipe the child's stdout.
	 * Also make stderr the same as stdout.
	 *=================================================================*/
	close(0);
	dup(pipefds[0]);
	close(1);
	dup(pipefds[1]);
	close(2);
	dup(pipefds[1]);
	close(pipefds[1]);
	close(pipefds[0]);
	system(buf);

	exit(0);
    }

    /*======================================================================
     * The parent listens on the pipe and sends output to the client
     *=====================================================================*/
    ifd = pipefds[0];
    close(pipefds[1]);
    maxfd = ifd > ofd ? ifd : ofd;
    while (1) {
	int rc;
	int nread;
	CSREQ req;
	struct {
	    SCRESP resp;
	    char buf[BUFSIZ];
	} pkt;

	FD_ZERO(&rdy);
	FD_ZERO(&wdy);
	if (pause) {
	    FD_CLR(ifd,&rdy);
	    FD_SET(ofd,&rdy);
	}
	else {
	    FD_SET(ifd,&rdy);
	    FD_SET(ofd,&rdy);
	}

#ifdef __hppa
	if((rc = select(maxfd + 1,(int *)&rdy,0,0,NULL)) < 0) {
#else
	if((rc = select(maxfd + 1,&rdy,(fd_set *)0,(fd_set *)0,NULL)) < 0) {
#endif
	    syslog(LOG_ERR,"Error: select error.  %s",sys_errlist[errno]);
	    close(ifd);
	    killpg(pid,SIGTERM);
	    exit(-1);
	}

	if (rc == 0) {
	    close(ifd);
	    killpg(pid,SIGTERM);
	    exit(0);
	}

	/*==============================================================
	 * Process pause, resume, stop or EOF from client
	 *=============================================================*/
	if (FD_ISSET(ofd,&rdy)) {
	    if ((rc = enc_read(ofd,(char *)&req,sizeof(req),ckeysched)) < 0) {
		syslog(LOG_ERR,"Error: Error reading pause/stop request");
		close(ifd);
		killpg(pid,SIGTERM);
		exit(-1);
	    }

	    if (rc == 0) {
		close(ifd);
		killpg(pid,SIGTERM);
		exit(0);
	    }
	    order_bytes_in_req(&req);

	    /*==========================================================
	     * Only process if component matches the current one
	     *=========================================================*/
	    if (req.component == component) {
		if (req.cmd == CANCEL) {
		    close(ifd);
		    killpg(pid,SIGTERM);
		    resp.rc = CLOSED;
		    resp.msglen = 0;
		    order_bytes_in_resp(&resp);
		    if ((rc = putdata(ofd,(char *)&resp,sizeof(resp),
							ckeysched)) < 0) {
			sprintf(error_resp.msg,"putdata() failed.  %s",
				sys_errlist[errno]);
			syslog(LOG_ERR,"Error: %s",error_resp.msg);
			return(-1);
		    }
		    return(OK);
		}
		if (req.cmd == PAUSE)
		    pause = 1;
		if (req.cmd == RESUME)
		    pause = 0;
	    }
	}

	if (pause)
	    continue;

	/*==============================================================
	 * Read in the data from tail
	 *=============================================================*/
	if (FD_ISSET(ifd,&rdy)) {
	    if ((nread = read(ifd, &pkt.buf[leftover],
				sizeof(pkt.buf)-leftover)) < 0) {
		syslog(LOG_ERR,"Error: read error from %s.  %s",
				syslogfile,sys_errlist[errno]);
		close(ifd);
		killpg(pid,SIGTERM);
		exit(-1);
	    }
	    if (nread == 0) {
		pause = 1;	/* Don't select on ifd anymore */
		continue;
	    }

	    /*==========================================================
	     * Adjust the read data length if we had leftover
	     * Make sure there is a null terminator at the end
	     *=========================================================*/
	    nread += leftover;
	    pkt.buf[nread] = '\0';
	    leftover = 0;
	    
	    /*==========================================================
	     * Make sure socket is writable before proceeding
	     *=========================================================*/
	    FD_CLR(ifd,&rdy);
	    FD_SET(ofd,&rdy);
	    FD_SET(ofd,&wdy);
#ifdef __hppa
	    if((rc = select(maxfd + 1,(int *)&rdy,(int *)&wdy,0,NULL)) < 0) {
#else
	    if((rc = select(maxfd + 1,&rdy,&wdy,(fd_set *)0,NULL)) < 0) {
#endif
		syslog(LOG_ERR,"Error: select error.  %s",sys_errlist[errno]);
		close(ifd);
		killpg(pid,SIGTERM);
		exit(-1);
	    }

	    /*======================================================
	     * Process pause, resume, stop or EOF from client
	     *=====================================================*/
	    if (FD_ISSET(ofd,&rdy)) {
		if ((rc = enc_read(ofd,(char *)&req,sizeof(req),
							ckeysched)) < 0) {
		    syslog(LOG_ERR,"Error: Error reading pause/stop "
					"request");
		    close(ifd);
		    killpg(pid,SIGTERM);
		    exit(-1);
		}

		if (rc == 0) {
		    close(ifd);
		    killpg(pid,SIGTERM);
		    return(OK);
		}
		order_bytes_in_req(&req);

		/*==================================================
		 * Only process if component matches the current one
		 *=================================================*/
		if (req.component == component) {
		    if (req.cmd == CANCEL) {
			close(ifd);
			killpg(pid,SIGTERM);
			resp.rc = CLOSED;
			resp.msglen = 0;
			order_bytes_in_resp(&resp);
			if ((rc = putdata(ofd,(char *)&resp,sizeof(resp),
							ckeysched)) < 0) {
			    sprintf(error_resp.msg,"putdata() failed.  %s",
					sys_errlist[errno]);
			    syslog(LOG_ERR,"Error: %s",error_resp.msg);
			    return(-1);
			}
			return(OK);
		    }
		    if (req.cmd == PAUSE) {
			pause = 1;
			continue;
		    }
		    if (req.cmd == RESUME)
			pause = 0;
		}
	    }

	    if (FD_ISSET(ofd,&wdy)) {
		/*===========================================================
		 * Loop to break the read buffer into lines (separated by
		 * newline)
		 *==========================================================*/
		pkt.resp.rc = htonl(OK);
		while (nread > 0) {
		    int wlen;
		    char *ncp;

		    /*=======================================================
		     * If no newline character, this is a partial line, set
		     * leftover and break out to read the next buffer
		     *======================================================*/
		    if (!strchr(pkt.buf,'\n')) {
			leftover = nread;
			nread = 0;
			break;
		    }

		    ncp = strtok(pkt.buf,"\n");
		    wlen = strlen(pkt.buf) + 1;
		    ncp = pkt.buf + wlen;
		    pkt.resp.msglen = htonl(wlen);
		    if ((incl1 == NULL || incl(pkt.buf,incl1)) &&
			(incl2 == NULL || incl(pkt.buf,incl2)) &&
			(excl == NULL || !incl(pkt.buf,excl))) {
			if ((rc = enc_write(ofd, (char *)&pkt,
					sizeof(SCRESP) + wlen,
					ckeysched)) < 0) {
			    syslog(LOG_ERR,"Error: Error writing to socket.  "
				"%s", sys_errlist[errno]);
			    close(ifd);
			    killpg(pid,SIGTERM);
			    exit(-1);
			}
		    }
		    nread -= wlen;
		    if (nread > 0)
			memmove(pkt.buf,ncp,nread+1);
		}
	    }
	}
    }
}

/*=======================================================================
 * Given a component, this function finds the pathname to the pid file for
 * the component.
 * Returns	1	if component has no pid file
 *		0	if component has pid file
 *		-1	if component is not on the list to be managed
 *		-2	if unsuccessful in opening/reading httpd.conf
 *	error_resp.msg will contain a message if rc is negative
 *======================================================================*/
int pidfile(int component, char *path) {
    FILE *cf;
    char *cp;
    char linebuf[BUFSIZ];
    char filename[BUFSIZ];

    switch (component) {
	case CAPACHE:
	case CHTTPD:
	    if (component == CAPACHE) {
		strcpy(linebuf, "/etc/apache/httpd.conf");
		strcpy(filename, "httpd.pid");
		strcpy(path, "/etc/apache/");
	    }
	    else {
		strcpy(linebuf, "/etc/httpd.conf");
		strcpy(filename, "httpd-pid");
		strcpy(path, "/etc/");
	    }

	    memset(filename, 0, sizeof(filename));

	    /*===============================================================
	     * Get the pathname to the httpd pid file.
	     *==============================================================*/
	    cf =  fopen(linebuf,"r");
	    if (cf == NULL) {
		sprintf(error_resp.msg, "Could not open %s.  %s",
			linebuf, sys_errlist[errno]);
		return(-2);
	    }
	    while( fgets(linebuf, BUFSIZ, cf) != NULL) {
		if ((cp = strtok(linebuf," \t\n")) == NULL)
		    continue;

		if (!strcasecmp(cp,"ServerRoot")) {
		    strcpy(path,strtok(NULL," \t\n"));
		}
		if (!strcasecmp(cp,"PidFile")) {
		    strcpy(filename,strtok(NULL," \t\n"));
		}
	    }
	    fclose(cf);

	    if (strlen(filename) == 0) {
		sprintf(error_resp.msg, "Could not find pid file name in %s",
			linebuf);
		return(-2);
	    }

	    if (path[strlen(path)-1] != '/')
		strcat(path,"/");

	    /*===============================================================
	     * If filename is absolute, overwrite the ServerRoot path
	     *==============================================================*/
	    if (*filename == '/')
		strcpy(path, filename);
	    else
		strcat(path, filename);

	    return(0);
	case CAPROXY:
	    strcpy(path,FW_CFG_DIR "/aproxy.pid");
	    return(0);
	case CPADMIN:
	    strcpy(path,PIDFILE);
	    return(0);
	case CFWMON:
	    strcpy(path,FWMONPIDS);
	    return(0);
	case CSOCKD:
	    strcpy(path,"/etc/sockd.pid");
	    return(0);
	case CSYSLOG:
#if defined(linux)
	    strcpy(path,"/var/run/syslogd.pid");
#else
	    strcpy(path,"/etc/syslog.pid");
#endif
	    return(0);
	case CINETD:
#if defined(linux)
	    /* Only linux has inetd.pid.  Other platforms will return file
		not found						*/
	    strcpy(path,"/var/run/inetd.pid");
	    return(0);
#endif
	case CRPCPROXY:
	case CFWPULSE:
	case CSMWRAPD:
	    strcpy(path,"");
	    return(1);
	case CWEBGATE:
	    strcpy(path,FW_CFG_DIR "/webgate.pid");
	    return(0);
	default:
	    strcpy(path,"");
	    strcpy(error_resp.msg,"Component is not managed by the padmin "
			"start/stop function");
	    return(-1);
    }
}

/*=======================================================================
 * Look for processes with procname in the process table and returns 
 * all the matching process ids in char format.  The ids are separated by
 * spaces.
 * input:
 *	procname	char string containing the process name to look for
 *	procnums	char buffer of size BUFSIZ to hold the pids
 * output:
 *	returns	0	if successful
 *	returns	-1	if failed
 *======================================================================*/
int findproc(char *procname, char *procnums) {
    int rc;
    pid_t pid;
#ifdef _AIX
    struct procsinfo	monitor;
    #define PIDNO	monitor.pi_pid
#else
#ifdef __sun
    kvm_t		*kd;
    struct proc		*monitor;
    #define PIDNO	monitor->p_pidp->pid_id
#else
#ifdef __hppa
    struct pst_status	monitor;
    #define PIDNO	monitor.pst_pid
#else
#ifdef __Lynx__
    int			kmem;
    int			nproc;
    long		proctab;
    struct pentry	monitor;
    #define PIDNO	monitor.pid
#else
#ifdef linux
    DIR			*dp;
    struct dirent	*dirp;
    char		buf[BUFSIZ];
    #define PIDNO       dirp->d_name
#endif
#endif
#endif
#endif
#endif

    memset(procnums,0,BUFSIZ);
    /*=============================================================
     * Loop thru all the process table entries
     *============================================================*/
    pid = 0;
#ifdef _AIX
    while ((rc = getprocs(&monitor,sizeof(monitor),NULL,0,&pid,1)) == 1) {
	if (strcmp(procname,monitor.pi_comm) == 0) {
#else
#ifdef __sun
    kd = kvm_open(NULL,NULL,NULL,O_RDONLY,"padmin:");
    if (kd == NULL) {
	sprintf(error_resp.msg, "kvm_open returned NULL.  %s.",
			sys_errlist[errno]);
	syslog(LOG_ERR, "Error: %s", error_resp.msg);
	return(-1);
    }
    while ((monitor = kvm_nextproc(kd)) != NULL) {
	int error;
	char *cmd, *cmd1;
	char **arg;
	struct user *user;

	cmd = NULL;
	cmd1 = NULL;
	/*===================================================================
	 * kvm_getu() returns a NULL for a zombie.
	 * Just to make sure we don't core dump here
	 *==================================================================*/
	if ((user = kvm_getu(kd, monitor)) == NULL)
	    continue;
	
	/*===================================================================
	 * Get the command from the user area
	 *==================================================================*/
	if ((error = kvm_getcmd(kd, monitor, user, &arg, NULL)) == -1) {
	    if (user->u_comm[0] != '\0')
		cmd = user->u_comm;
	}
	else
	    cmd = arg[0];

	if (cmd == NULL) {
	    if (error != -1)
		free(arg);
	    continue;
	}
	/*===============================================================
	 * Get rid of path info
	 *==============================================================*/
	if ((cmd1 = strrchr(cmd, '/')) != NULL)
	    cmd1++;
	else
	    cmd1 = cmd;

	if (strcmp(procname, cmd1) == 0) {
#else
#ifdef __hppa
    while ((rc = pstat_getproc(&monitor,sizeof(monitor),1,pid++)) == 1) {
	if (strcmp(procname,monitor.pst_ucomm) == 0) {
#else
#ifdef __Lynx__
    if ((kmem = open("/dev/mem", O_RDONLY)) < 0) {
	sprintf(error_resp.msg, "Could not open /dev/mem.  %s",
			sys_errlist[errno]);
	syslog(LOG_ERR, "Error: %s", error_resp.msg);
	return(-1);
    }
    proctab = info(I_PROCTAB);
    nproc = info(I_NPROCTAB);
    /* Start with process #2, i.e. skip nullpr and init */
    for (pid = 2; pid < nproc; pid++) {
	lseek(kmem, proctab + pid * sizeof(monitor), 0);
	if (read(kmem, &monitor, sizeof(monitor)) != sizeof(monitor)) {
	    sprintf(error_resp.msg, "Failed reading process info for pid %d.  "
			"%s", pid, sys_errlist[errno]);
	    syslog(LOG_ERR, "Error: %s", error_resp.msg);
	    return(-1);
	}
	if (monitor.pstate != PRFREE && strcmp(procname,monitor.pname) == 0) {
#else
#ifdef linux
    dp = opendir("/proc");
    if (dp == NULL) {
	sprintf(error_resp.msg, "Could not open /proc directory.  %s",
			sys_errlist[errno]);
	syslog(LOG_ERR, "Error: %s", error_resp.msg);
	return(-1);
    }
    while ((dirp = readdir(dp)) != NULL) {
	int ffd;
	char *cp;
	/*===============================================================
	 * Build a pseudo filename of the form /proc/<pidno>/stat and
	 * try to open the file
	 * If we encounter any error in the open or the read, do not treat
	 * them as errors.  I assume this is correct way to handle them
	 * since /proc is a pseudo filesystem anyway and some of these
	 * pseudo files may be "weird" in the eyes of open and read.
	 *==============================================================*/
	strcpy(buf, "/proc/");
	strcat(buf, PIDNO);
	strcat(buf, "/stat");
	if ((ffd = open(buf, 0)) < 0)
	    continue;
	if (read(ffd, buf, sizeof(buf)) <= 0)
	    continue;
	close(ffd);
	cp = strtok(buf, " \t()");
	if (cp == NULL)
	    continue;
	/*===============================================================
	 * Get 2nd field which has proc name
	 *==============================================================*/
	if ((cp = strtok(NULL, " \t()")) == NULL)
	    continue;
	if (strcmp(procname, cp) == 0) {
#endif
#endif
#endif
#endif
#endif
	    char tmp[16];
#ifdef linux
	    /*===========================================================
	     * Shouldn't happen but juts in case . . .  
	     * If the pid number is too long, just silently ignore it.
	     *==========================================================*/
	    if (strlen(PIDNO) > sizeof(tmp) - 1)
		continue;
	    strcpy(tmp, PIDNO);
#else
	    sprintf(tmp," %d",PIDNO);
#endif

	    /*===========================================================
	     * Make sure the string fits in the buffer.
	     * BUFSIZ - strlen(tmp) - 1 - 16
	     * strlen is the length of the latest pid
	     * 1 for the separating blank
	     * 16 for the string " >/dev/null 2>&1" to be added to the
	     * buffer later.
	     *==========================================================*/
	    if (strlen(procnums) < BUFSIZ - strlen(tmp) - 1 - 16)
		strcat(procnums,tmp);
	    else {
		sprintf(error_resp.msg,"Too many processes named %s to be "
			"handled by padmin",procname);
		syslog(LOG_NOTICE,error_resp.msg);
		return(-1);
	    }
	}
#ifdef __sun
	if (error != -1)
	    free(arg);
    }
    kvm_close(kd);
#else
#ifdef linux
    }
    closedir(dp);
#else
    }
#endif
#endif
    if (strlen(procnums) == 0) {
	sprintf(error_resp.msg,"Cannot find any process named %s",procname);
	return(-1);
    }
    return(0);

}

/*======================================================================
 * Start/Stop/Refresh a component
 * input:
 *	ofd	file descriptor for socket to client
 *	component the component to start/stop/refresh
 *	cmd	indicate whether to start, stop or refresh
 * output:
 *	returns 0 if OK
 *	returns -1 if failed - connection should be closed
 *	returns -2 if connection is closed during transfer
 *	returns -3 if non fatal error - loop back for new request
 *	error_resp.msg will contain a message if rc is negative
 *=====================================================================*/
int ssref(int ofd, int component, int cmd) {
    int rc;
    char cmdstr[BUFSIZ];
    char buf[BUFSIZ];
    char name[16];

#define RESPOK(a)	{RESPPKT resp; \
			strcpy(resp.msg,a); \
			resp.hdr.msglen = strlen(resp.msg) + 1; \
			resp.hdr.rc = OK; \
			order_bytes_in_resp(&(resp.hdr)); \
			if ((rc = putdata(ofd,(char *)&resp, \
				sizeof(SCRESP)+ntohl(resp.hdr.msglen), \
					ckeysched)) <0) { \
			    syslog(LOG_ERR,"Error: putdata() failed"); \
			    return(FAILED); \
			} \
			return(0); \
			}

    if (trexcmd(component,cmdstr) < 0) {
	strcpy(error_resp.msg,"Component is not managed by the padmin "
		"start/stop function");
	return(NONFATAL);
    }

    /*=======================================================================
     * Get the name of the component
     *======================================================================*/
    switch (component) {
	case CAPACHE:
	    strcpy(name,"apache");
	    break;
	case CAPROXY:
	    strcpy(name,"aproxy");
	    break;
	case CFWPULSE:
	    strcpy(name,"fwpulse");
	    break;
	case CFWPULSED:
	    strcpy(name,"fwpulsed");
	    break;
	case CHTTPD:
	    strcpy(name,"cern httpd");
	    break;
	case CINETD:
	    strcpy(name,"inetd");
	    break;
	case CPADMIN:
	    strcpy(name,"padmin");
	    break;
	case CFWMON:
	    strcpy(name,"fwmon");
	    break;
	case CRPCPROXY:
	    strcpy(name,"rpcproxy");
	    break;
	case CSOCKD:
	    strcpy(name,"sockd");
	    break;
	case CSMWRAPD:
	    strcpy(name,"smwrapd");
	    break;
	case CSMALIASES:
	    strcpy(name,"smwrap aliases");
	    break;
	case CSYSLOG:
	    strcpy(name,"syslogd");
	    break;
	case CWEBGATE:
	    strcpy(name,"webgate");
	    break;
    }

    switch (cmd) {
	int tmpfd;
	case STARTCOMP:
	    /*===============================================================
	     * The switch here processes the special cases where the normal
	     * pathname alone does not handle
	     *==============================================================*/
	    switch (component) {
		case CINETD:
		    /*=======================================================
		     * Currently not allowing starting inetd with padmin
		    #ifdef _AIX
			strcpy(cmdstr,"startsrc -s inetd");
		    #endif
		    #ifdef __sun
			strcat(cmdstr," -s");
		    #endif
		    break;
		     *======================================================*/
		case CPADMIN:
		case CSMALIASES:
		    strcpy(error_resp.msg,"This component cannot be started "
				"with padmin");
		    return(NONFATAL);
		case CSYSLOG:
		    #ifdef _AIX
			strcpy(cmdstr,"startsrc -s syslogd");
		    #endif
		    #if defined(__sun) || defined(linux)
			if (trexcmd(component, cmdstr) != 0)
			    return(NONFATAL);
		    #endif
		    break;
		case CFWPULSED:
		   /*=======================================================
		    * Kludge to get around problem with fwpulsed using a
		    * getpeername() on fd 0 to determine if it was started by
		    * inetd. We temporarily closes 0 and then start it up
		    * later.
		    *======================================================*/
		   tmpfd = dup(0);
		   close(0);
		   break;
	    }
	    strcat(cmdstr," >/dev/null 2>&1");
	    rc = system(cmdstr);
	    if (component == CFWPULSED) {
		close(0);
		dup(tmpfd);
	    }
	    if (rc != 0) {
		sprintf(error_resp.msg,"%s returned %d",cmdstr,rc);
		return(NONFATAL);
	    }
	    sprintf(buf,"%s started",name);
	    RESPOK(buf);
	case STOPCOMP:
	    switch (component) {
		case CAPACHE:
		case CAPROXY:
		case CHTTPD:
		case CFWMON:
		case CSOCKD:
		case CSYSLOG:
		case CWEBGATE:
		    if (pidfile(component,buf) != 0)
			return(NONFATAL);
		    sprintf(cmdstr,"kill `cat %s`",buf);
		    break;
		case CINETD:
		case CPADMIN:
		case CSMALIASES:
		    strcpy(error_resp.msg,"This component cannot be stopped "
				"with padmin");
		    return(NONFATAL);
		case CFWPULSE:
		case CFWPULSED:
		case CRPCPROXY:
		case CSMWRAPD:
		    if (findproc(name,buf) < 0)
			return(NONFATAL);
		    sprintf(cmdstr,"kill %s",buf);
		    break;
	    }
	    strcat(cmdstr," >/dev/null 2>&1");
	    if ((rc = system(cmdstr)) != 0) {
		sprintf(error_resp.msg,"Stopping \"%s\" processes returned rc "
			"of %d",name,rc);
		return(NONFATAL);
	    }
	    sprintf(buf,"%s stopped",name);
	    RESPOK(buf);
	case REFCOMP:
	    switch (component) {
		case CAPACHE:
		case CSOCKD:
		    if (pidfile(component,buf) != 0)
			return(NONFATAL);
		    sprintf(cmdstr,"kill -HUP `cat %s`",buf);
		    break;
		case CHTTPD:
		    strcat(cmdstr," -restart");
		    break;
		case CINETD:
		    #ifdef _AIX
			strcpy(cmdstr,"refresh -s inetd");
		    #endif
		    #if defined(__sun) || defined(__Lynx__)
			if(findproc("inetd", buf) != 0)
			    return(NONFATAL);
			sprintf(cmdstr,"kill -HUP %s", buf);
		    #endif
		    #ifdef linux
			if (pidfile(component,buf) != 0)
			    return(NONFATAL);
			sprintf(cmdstr,"kill -HUP `cat %s`",buf);
		    #endif
		    #ifdef __hppa
			strcat(cmdstr," -c");
		    #endif
		    break;
		case CPADMIN:
		    if (pidfile(component,buf) != 0)
			return(NONFATAL);
		    sprintf(cmdstr,"kill -USR1 `cat %s`",buf);
		    break;
		case CFWMON:
		    break;
		case CSYSLOG:
		    #ifdef _AIX
			strcpy(cmdstr,"refresh -s syslogd");
		    #endif
		    #if defined(__sun) || defined (__hppa) || defined (linux)
			if (pidfile(component,buf) != 0)
			    return(NONFATAL);
			sprintf(cmdstr,"kill -HUP `cat %s`",buf);
		    #endif
		    break;
		case CAPROXY:
		case CWEBGATE:
		    if (pidfile(component, buf) != 0)
			return(NONFATAL);
		    sprintf(cmdstr,"kill `cat %s`",buf);
		    strcat(cmdstr," >/dev/null 2>&1");
		    if ((rc = system(cmdstr)) != 0) {
			sprintf(error_resp.msg,"Stopping %s processes "
				"returned rc of %d",name,rc);
			return(NONFATAL);
		    }
		    trexcmd(component,cmdstr);
		    sleep(1);
		    break;
		case CFWPULSE:
		case CFWPULSED:
		case CRPCPROXY:
		case CSMWRAPD:
		    if (findproc(name,buf) < 0)
			return(NONFATAL);
		    sprintf(cmdstr,"kill %s",buf);
		    strcat(cmdstr," >/dev/null 2>&1");
		    if ((rc = system(cmdstr)) != 0) {
			sprintf(error_resp.msg,"Stopping \"%s\" processes "
				"returned rc of %d",name,rc);
			return(NONFATAL);
		    }
		    trexcmd(component,cmdstr);
		    break;
	    }
	    strcat(cmdstr," >/dev/null 2>&1");
	    if (component == CFWPULSED) {
		tmpfd = dup(0);
		close(0);
	    }
	    rc = system(cmdstr);
	    if (component == CFWPULSED) {
		close(0);
		dup(tmpfd);
	    }
	    if (rc != 0) {
		sprintf(error_resp.msg,"%s returned %d",cmdstr,rc);
		return(NONFATAL);
	    }
	    sprintf(buf,"%s refreshed",name);
	    RESPOK(buf);
	default:
	    strcpy(error_resp.msg,"Request is not Start, Stop or Refresh");
	    return(NONFATAL);
    }
}

/*======================================================================
 * Get the directory listing for LISTDIR and send the output
 * to the client
 * input:
 *	ofd	file descriptor for socket to client
 *	component the component to get dir list for
 * output:
 *	returns 0 if OK
 *	returns -1 if failed - connection should be closed
 *	returns -2 if connection is closed during transfer
 *	returns -3 if non fatal error - loop back for new request
 *	error_resp.msg will contain a message if rc is negative
 *=====================================================================*/
#define MAXDIRENT	100	/* The client code (webblock.java) should
				   have the same max			*/
int getlist(int ofd, int component) {
    int n = 0;
    int rc;
    char *dirname;
    DIR *dir;
    struct dirent *dp;
    RESPPKT resp;
    char buf[sizeof(resp.msg)];

    if (component != CALL && component != CAPACHE) {
	strcpy(error_resp.msg,"GETLIST request must be for APACHE or ALL");
	return(NONFATAL);
    }

    if (component == CALL)
	dirname = PROGDIR;
    else
	dirname = LISTDIR;

    memset(resp.msg,0,sizeof(buf));
    memset(buf,0,sizeof(buf));
    /*=======================================================================
     * Open the directory
     *======================================================================*/
    if ((dir = opendir(dirname)) == NULL) {
	sprintf(error_resp.msg,"Cannot open directory %s",dirname);
	return(NONFATAL);
    }

    /*=======================================================================
     * Get the directory entries one at a time
     *======================================================================*/
    for (dp = readdir(dir); dp != NULL; dp = readdir(dir)) {
	if (n >= MAXDIRENT ||
#if defined(__sun) || defined(linux)
		(strlen(buf)+strlen(dp->d_name)+1) > sizeof(buf)) {
#else
		(strlen(buf) + dp->d_namlen + 1) > sizeof(buf)) {
#endif
	    closedir(dir);
	    sprintf(error_resp.msg,"%s directory listing too long",dirname);
	    return(NONFATAL);
	}
	if (strcmp(dp->d_name,".") == 0 || strcmp(dp->d_name,"..") == 0)
	    continue;
	if (n > 0)
	    strcat(buf," ");
	strcat(buf,dp->d_name);
	n++;
    }
    closedir(dir);

#define append_component(a)	sprintf(comp, "%d", a); \
				if (strlen(resp.msg) + strlen(comp) + 1 > \
						sizeof(resp.msg)) { \
				    closedir(dir); \
				    sprintf(error_resp.msg,"%s directory " \
						"listing too long",dirname); \
				    return(NONFATAL); \
				} \
				if (strlen(resp.msg) == 0) \
				    strcpy(resp.msg, comp); \
				else { \
				    strcat(resp.msg, " "); \
				    strcat(resp.msg, comp); \
				}
					
    /*=======================================================================
     * If component == CALL, translate the directory listing to component
     * numbers before responding.
     *======================================================================*/
    if (component == CALL) {
	char *cp;
	char comp[5];
	for (cp = strtok(buf, " "); cp; cp = strtok(NULL, " ")) {
	    if (!strcmp(cp, "httpd")) {
		append_component(CAPACHE);
		continue;
	    }
	    if (!strcmp(cp, "ftproxy")) {
		append_component(CFTPROXY);
		continue;
	    }
	    if (!strcmp(cp, "fwpulse")) {
		append_component(CFWPULSE);
		continue;
	    }
	    if (!strcmp(cp, "fwpulsed")) {
		append_component(CFWPULSED);
		continue;
	    }
	    if (!strcmp(cp, "aproxy")) {
		append_component(CAPROXY);
		continue;
	    }
	    if (!strcmp(cp, "gwuser")) {
		append_component(CGWUSER);
		continue;
	    }
	    if (!strcmp(cp, "httpd_3.0A")) {
		append_component(CHTTPD);
		continue;
	    }
	    if (!strcmp(cp, "fwmon")) {
		append_component(CFWMON);
		continue;
	    }
	    if (!strcmp(cp, "raproxy")) {
		append_component(CRAPROXY);
		continue;
	    }
	    if (!strcmp(cp, "rpcproxy")) {
		append_component(CRPCPROXY);
		continue;
	    }
	    if (!strcmp(cp, "smwrap")) {
		append_component(CSMWRAP);
		continue;
	    }
	    if (!strcmp(cp, "sockd")) {
		append_component(CSOCKD);
		continue;
	    }
	    if (!strcmp(cp, "tnproxy")) {
		append_component(CTNPROXY);
		continue;
	    }
	    if (!strcmp(cp, "webgate")) {
		append_component(CWEBGATE);
		continue;
	    }
	}
    }
    else
	strcpy(resp.msg, buf);

    /*=======================================================================
     * Send response to the client
     *======================================================================*/
    resp.hdr.rc = OK;
    resp.hdr.msglen = strlen(resp.msg) + 1;
    order_bytes_in_resp(&(resp.hdr));

    if ((rc = putdata(ofd, (char *)&resp,
				sizeof(SCRESP) + ntohl(resp.hdr.msglen),
				ckeysched)) < 0) {
	sprintf(error_resp.msg,"putdata() failed.  %s",
			sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(-1);
    }

    return(OK);
}

/*======================================================================
 * Get the list of all the firewall local IP addresses to the client
 *
 * input:
 *	ofd	file descriptor for socket to client
 *	component the component to get dir list for
 * output:
 *	returns 0 if OK
 *	returns -1 if failed - connection should be closed
 *	returns -2 if connection is closed during transfer
 *	returns -3 if non fatal error - loop back for new request
 *	error_resp.msg will contain a message if rc is negative
 *=====================================================================*/
int getiflist(int ofd, int component) {
    int i;
    int n = 0;
    static struct in_addr *iplist = NULL;
    int rc;
    RESPPKT resp;
    
    /*==================================================================
     * Free any IP address lists from previous calls
     *=================================================================*/
    if (iplist != NULL) {
	free(iplist);
	iplist = NULL;
    }

    if (component != CAPACHE) {
	strcpy(error_resp.msg,"GETIFLIST request must be for APACHE");
	return(NONFATAL);
    }

    /*=======================================================================
     * Get the IP list
     *======================================================================*/
    if (getlocal((in_addr_t **)&iplist, NULL, &n) != 0) {
	sprintf(error_resp.msg,"Getting local IP address list failed");
	return(NONFATAL);
    }
    if (n > MAXDIRENT) {
	sprintf(error_resp.msg,"Too many local IP addresses, %d is the limit.",
			MAXDIRENT);
	return(NONFATAL);
    }
    for (i = 0; i < n; i++) {
	if (i == 0)
	    strcpy(resp.msg, inet_ntoa(iplist[i]));
	else {
	    strcat(resp.msg, " ");
	    strcat(resp.msg, inet_ntoa(iplist[i]));
	}
    }

    /*=======================================================================
     * Send response to the client
     *======================================================================*/
    resp.hdr.rc = OK;
    resp.hdr.msglen = strlen(resp.msg) + 1;
    order_bytes_in_resp(&(resp.hdr));

    if ((rc = putdata(ofd, (char *)&resp,
				sizeof(SCRESP) + ntohl(resp.hdr.msglen),
				ckeysched)) < 0) {
	sprintf(error_resp.msg,"putdata() failed.  %s",
			sys_errlist[errno]);
	syslog(LOG_ERR,"Error: %s",error_resp.msg);
	return(-1);
    }

    return(OK);
}
