/*============================================================================
 * filename:		plog/plog.c					
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *										
 * written by:		Tommy Chan
 *
 * (C) COPYRIGHT R. James Livermore 1998-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1998-2000.
 * All rights reserved. 
 *           
 *		
 * Date written:	September 15, 1998   
 * Last modified:	February   8, 2000 - added LPL
 *  
 * function:
 *============================================================================*/
static	char	ident[] = "@(#) plog, V 1.0, 00/01/21"
			  " (C) Freemont Avenue Software, Inc."
			  " 1998-2000";
#include <syslog.h>
#include <netdb.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/sem.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <limits.h>
#include "firewall.h"

#define IS_NUMERIC_IP(a)        (strspn(a,"0123456789.") == strlen(a))
#define IS_NUMERIC(a)        (strspn(a,"0123456789") == strlen(a))
#define DEFAULT_PORT 741
#define MAX_NAME_LENGTH PATH_MAX+1
#define PIDFILE FW_CFG_DIR "/plog.pid"

#ifndef linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif

struct RMLOG {
    char infile[MAX_NAME_LENGTH];	/* input file to be logged	*/
    struct in_addr ip;			/* ip address of log server	*/
    unsigned short port;		/* port number for log server	*/
    long size;				/* cache buffer size in bytes	*/
    long interval;			/* max seconds between sends	*/

    char *bp;				/* pointer to cache buffer	*/
    long bytes_filled;			/* total bytes from infile	*/
    int fd;				/* fd for the socket to server	*/
    int hist_fd;			/* fd for the history file	*/
    off_t hist_offset;			/* offset to the matching hist
					   entry			*/
};
typedef struct RMLOG RMLOG;
static RMLOG rmlog;			/* static so int handler can use*/
static int semid = -1;

void alrm(int sig);
int update_hist(char *fn, int increment);
#define get_hist_count(a)	update_hist(a, 0)
/*===========================================================================
 *
 *==========================================================================*/
void sendbuf() 
{
    int nwrite;
    int nmore;
    if (rmlog.bytes_filled == 0) 
    {
	alarm(rmlog.interval);
	signal(SIGALRM, alrm);
	return;
    }
    nmore = rmlog.bytes_filled;
    while (nmore > 0) 
    {
	if ((nwrite = write(rmlog.fd, rmlog.bp, nmore)) < 0) {
	    fprintf(stderr, "Error: socket write failed.  %s\n",
			sys_errlist[errno]);
	    syslog(LOG_ERR, "Error: socket write failed.  %s",
			sys_errlist[errno]);
	    exit(1);
	}
	nmore = nmore - nwrite;
	if (nmore > 0)
	    syslog(LOG_INFO, "Partial write completed for %d bytes from %d "
			"byte buffer", rmlog.bytes_filled - nmore,
			rmlog.bytes_filled);
    }
    
    /*=======================================================================
     * Update the count in the history file
     *======================================================================*/
    update_hist(NULL, rmlog.bytes_filled);

    rmlog.bytes_filled = 0;
    alarm(rmlog.interval);
    signal(SIGALRM, alrm);
    return;
}
/*===========================================================================
 *
 *===========================================================================*/
void term(int sig) 
{
    killpg(getpid(), SIGTERM);
    exit(0);
}

void nop(int sig) {
    return;
}

void alrm(int sig) {
    alarm(rmlog.interval);
    signal(SIGALRM, alrm);
    return;
}

/*============================================================================
 * Lock, read and update the history file
 * If an entry cannot be found (that corrensponds with info in the rmlog
 * structure), a new entry will be created with increment as count.
 * If increment is nonzero, it will be added to the count in the history entry
 * and written back to the history file.
 * If increment is 0, the count in the history entry is read and returned.
 *
 * fn can be NULL when updating count since the file has already been opened.
 *
 * Returns: count in history entry + increment
 *	    -1 if failed
 *
 * Format of the history file
 * optional '[' to indicate the beginning of an entry
 * log_file_name server_ip_addr count
 * log_file_name is the full pathname of the file to be logged.  This is used
 *			together with server_ip_addr as the search key to the
 *			history entries.
 * server_ip_addr is the ip address of the remote log receiver in dot decimal
 *			representation
 * count is the number of bytes from the log file that has been written to the
 *			receiver.  It must be a 10-digit number, with leading
 *			zeros if necessary.  This is so that an update of the
 *			count will not require re-writing the entire file.
 * mandatory ']' to indicate the end of an entry
 *===========================================================================*/
int update_hist(char *fn, int increment) 
{
#define LOCK_HIST()	if (semop(semid,&op_lock[0],2) < 0) { \
				fprintf(stderr, "Error: semop lock error.  " \
					"%s\n", sys_errlist[errno]); \
				syslog(LOG_ERR, "Error: semop lock error.  " \
					"%s", sys_errlist[errno]); \
				return(-1); \
    }
#define UNLOCK_HIST()	if (semop(semid,&op_unlock[0],1) < 0) { \
				fprintf(stderr, "Error: semop unlock error.  " \
					"%s\n", sys_errlist[errno]); \
				syslog(LOG_ERR, "Error: semop unlock error.  " \
					"%s", sys_errlist[errno]); \
			}
    int file_offset = 0;
    int en = 0;
    int buflen = 0;
    int found = 0;
    char buf[BUFSIZ];
    char *cp;
    char *nextcp;
    char *logfn;
    char *server;
    char *count;
    struct stat statbuf;
    static int cnt;			/* Current count in history entry */
    register off_t offset;
    register int whence;
    char tokensep[] = "[ \t\n]";
    static struct sembuf op_lock[2] =
    {
	0,0,0,				/* Wait for sem#0 to become 0 */
	0,1,SEM_UNDO			/* then increment sem#0 by 1  */
    };
    static struct sembuf op_unlock[1] =
    {
	0,-1,(IPC_NOWAIT|SEM_UNDO)	/* decrement sem#0 by 1 (sets to 0) */
    };
    
    /*=======================================================================
     * If we have already been here before, then we already have the file
     * offset and don't need to re-read the file.
     *======================================================================*/
    if (rmlog.hist_offset < 0) {

	/*===================================================================
	 * Open the history file
	 *==================================================================*/
	if ((rmlog.hist_fd = open(fn, O_RDWR|O_CREAT, S_IRUSR|S_IWUSR|S_IRGRP))
									< 0) {
	    fprintf(stderr, "Error: open failed for history file \"%s\".  %s\n",
			fn, sys_errlist[errno]);
	    syslog(LOG_ERR, "Error: open failed for history file \"%s\".  %s",
			fn, sys_errlist[errno]);
	    return(-1);
	}
	
	/*===================================================================
	 * Serialize this access code
	 *==================================================================*/
	LOCK_HIST();
	
	/*===================================================================
	 * Read a buffer from the file.
	 *==================================================================*/
	if ((buflen = read(rmlog.hist_fd, buf, BUFSIZ-1)) < 0) {
	    UNLOCK_HIST();
	    fprintf(stderr, "Error: read failed from history file \"%s\".  "
			"%s\n", fn, sys_errlist[errno]);
	    syslog(LOG_ERR, "Error: read failed from history file \"%s\".  %s",
			fn, sys_errlist[errno]);
	    return(-1);
	}

	/*===================================================================
	 * Search thru the file if it is not a brand new file
	 *==================================================================*/
	if (buflen != 0) {

	    /*===============================================================
	     * Make sure the buffer has a terminator.
	     *==============================================================*/
	    buf[buflen] = '\0';

	    /*===============================================================
	     * Loop to read in one entry at a time
	     *==============================================================*/
	    for (cp = buf; buflen > 0; cp = nextcp) {

		en++;
		/*===========================================================
		 * See if there is at least one ']' in the buffer indicating
		 * there is at least one more complete entry.
		 *==========================================================*/
		if ((nextcp = strchr(cp, ']')) == NULL) {
		    int len = 0;
		    /*=======================================================
		     * This last entry is not complete, read in more
		     *======================================================*/
		    if (cp < buf + buflen - 1) {
			len = strlen(cp);
			memmove(buf, cp, len);
		    }
		    /*=======================================================
		     * Read in another buffer
		     *======================================================*/
		    if ((buflen = read(rmlog.hist_fd, &buf[len], BUFSIZ-1-len))
								< 0) {
			UNLOCK_HIST();
			fprintf(stderr, "Error: read failed from history file "
				"\"%s\".  %s\n", fn, sys_errlist[errno]);
			syslog(LOG_ERR, "Error: read failed from history file "
				"\"%s\".  %s", fn, sys_errlist[errno]);
			return(-1);
		    }
		    /*=======================================================
		     * We're done in this loop if we hit EOF
		     *======================================================*/
		    if (buflen == 0)
			break;

		    buflen += len;
		    buf[buflen] = '\0';
		    cp = buf;
		    /*=======================================================
		     * If still no ']' in buffer and there is a non-separator in
		     * the buffer, there must be something wrong with the file
		     *======================================================*/
		    if ((nextcp = strchr(buf, ']')) == NULL &&
			strspn(buf, tokensep) != strlen(buf)) {
			UNLOCK_HIST();
			fprintf(stderr, "Error: Possible corrupted history "
				"file at entry %d\n", en);
			syslog(LOG_ERR, "Error: Possible corrupted history "
				"file at entry %d", en);
			return(-1);
		    }
		}
		nextcp++;
		/*===========================================================
		 * Get the fields in the current entries
		 *==========================================================*/
		if ((logfn = strtok(cp, tokensep)) == NULL ||
			(server = strtok(NULL, tokensep)) == NULL ||
			(count = strtok(NULL, tokensep)) == NULL) {
		    UNLOCK_HIST();
		    fprintf(stderr, "Error: Possible corrupted history file "
				"at entry %d\n", en);
		    syslog(LOG_ERR, "Error: Possible corrupted history file "
				"at entry %d", en);
		    return(-1);
		}
		/*===========================================================
		 * See if this is the entry we are looking for
		 *==========================================================*/
		if (strcmp(logfn, rmlog.infile) == 0 &&
		    strcmp(server, inet_ntoa(rmlog.ip)) == 0) {
		    found = 1;
		    break;
		}

		/*===========================================================
		 * Update the file offset so it reflects the number of bytes
		 * from the beginning of the history file to the beginning of
		 * this entry.
		 *==========================================================*/
		file_offset += nextcp - cp;
	    }

	    /*===============================================================
	     * file_offset now points to the beginning of the entry, adjust it
	     * so that it points to the beginning of the count field
	     *==============================================================*/
	    file_offset += count - cp;	/* cp still points to beginning of
						this entry */
	}

	/*===================================================================
	 * Position the file pointer to the end of the file if a matching entry
	 * is not found; otherwise, position file pointer to the beginning of
	 * the count field.
	 *==================================================================*/
	if (found) {
	    cnt = atoi(count);
	    offset = file_offset;
	    whence = SEEK_SET;
	}
	else {
	    cnt = 0;
	    offset = 0;
	    whence = SEEK_END;
	}

	if ((rmlog.hist_offset = lseek(rmlog.hist_fd, offset, whence)) < 0) {
	    rmlog.hist_offset = 0;
	    UNLOCK_HIST();
	    fprintf(stderr, "Error: lseek of history file failed.  %s\n",
				sys_errlist[errno]);
	    syslog(LOG_ERR, "Error: lseek of history file failed.  %s",
				sys_errlist[errno]);
	    return(-1);
	}

	/*===================================================================
	 * If the fife size is smaller than the count in our history file, then
	 * assume the count is old and not valid anymore.
	 *==================================================================*/
	if (stat(rmlog.infile, &statbuf) < 0) {
	    UNLOCK_HIST();
	    fprintf(stderr, "Error: stat failed for file \"%s\".  %s\n",
				rmlog.infile, sys_errlist[errno]);
	    syslog(LOG_ERR, "Error: stat failed for file \"%s\".  %s",
				rmlog.infile, sys_errlist[errno]);
	    return(-1);
	}
	if (statbuf.st_size < cnt) {
	    syslog(LOG_INFO, "Info: File \"%s\" has size less than count kept "
				"in history file, assume history file is "
				"out of date",
				rmlog.infile, sys_errlist[errno]);
	    cnt = 0;
	}
    }
    else {
	/*===================================================================
	 * We have already found the offset from a previous call, so we just
	 * seek over there to prepare for the update
	 *==================================================================*/
	found = 1;	/* make it look like we just found the entry	*/
	LOCK_HIST();
	if (lseek(rmlog.hist_fd, rmlog.hist_offset, SEEK_SET) < 0) {
	    UNLOCK_HIST();
	    fprintf(stderr, "Error: lseek of history file failed.  %s\n",
				sys_errlist[errno]);
	    syslog(LOG_ERR, "Error: lseek of history file failed.  %s",
				sys_errlist[errno]);
	    return(-1);
	}
    }


    cnt += increment;
    /*=======================================================================
     * Update the history file
     *======================================================================*/
    if (found)
	sprintf(buf, "%010d", cnt);
    else {
	char *ip = inet_ntoa(rmlog.ip);
	sprintf(buf, "[%s %s %010d]", rmlog.infile, ip, cnt);
	if (rmlog.hist_offset < 0)
	    rmlog.hist_offset = 0;
	rmlog.hist_offset += 1 + strlen(rmlog.infile) + 1 + strlen(ip) + 1;
    }

    if (write(rmlog.hist_fd, buf, strlen(buf)) != strlen(buf)) {
	UNLOCK_HIST();
	fprintf(stderr, "Error: Update of history entry failed.  %s\n",
				sys_errlist[errno]);
	syslog(LOG_ERR, "Error: Update of history entry failed.  %s",
				sys_errlist[errno]);
	return(-1);
    }

    UNLOCK_HIST();
    return(cnt);
}

/*=======================================================================
 * Read the config file and spawn a child process for each "remotelog"
 * Child processes all return 0 to the caller
 * Parent process returns 1
 *======================================================================*/
int readconf(char *conffn) {
    FILE *fp;
    char line[BUFSIZ];
    char hn[BUFSIZ];
    int linen = 0;
    int port;
    pid_t pid;
    struct servent *sp;

    if (!(fp = fopen(conffn, "r"))) {
	fprintf(stderr, "Error: Could not open config file \"%s\"\n", conffn);
	syslog(LOG_ERR, "Error: Could not open config file \"%s\"", conffn);
	exit(1);
    }

    /*=======================================================================
     * Read config file statements
     *======================================================================*/
    while (fgets(line, BUFSIZ, fp) != NULL) {
	char *token;
	linen++;

	/*===================================================================
	 * Skip over comment and null lines
	 *==================================================================*/
	if ((token = strtok(line, "\t,:;= \n")) == NULL || *token == '#')
	    continue;

	/*===================================================================
	 * Process "remotelog" statement
	 *==================================================================*/
	if (strcasecmp(token, "remotelog") == 0) {
	    /*===============================================================
	     * Set defaults
	     *==============================================================*/
	    memset((void *)&rmlog, 0, sizeof(rmlog));
	    port = DEFAULT_PORT;
	    rmlog.size = 16384;
	    rmlog.interval = 10;
	    rmlog.hist_offset = -1;

	    /*===============================================================
	     * input file name
	     *==============================================================*/
	    if ((token = strtok(NULL, "\t,:;= \n")) == NULL) {
		fprintf(stderr, "Error: No remotelog filename specfiied on "
				"line %d of %s\n", linen, conffn);
		syslog(LOG_ERR, "Error: No remotelog filename specfiied on "
				"line %d of %s", linen, conffn);
		exit(1);
	    }
	    strcpy(rmlog.infile, token);

	    /*===============================================================
	     * server name or address
	     *==============================================================*/
	    if ((token = strtok(NULL, "\t,:;= \n")) == NULL) {
		fprintf(stderr, "Error: No server name/address specfiied on "
				"line %d of %s\n", linen, conffn);
		syslog(LOG_ERR, "Error: No server name/address specfiied on "
				"line %d of %s", linen, conffn);
		exit(1);
	    }
	    strcpy(hn, token);

	    /*===============================================================
	     * server port number
	     *==============================================================*/
	    if ((token = strtok(NULL, "\t,:;= \n")) == NULL) {
		fprintf(stderr, "Error: No server port specfiied on "
				"line %d of %s\n", linen, conffn);
		syslog(LOG_ERR, "Error: No server port specfiied on "
				"line %d of %s", linen, conffn);
		exit(1);
	    }

	    /*===============================================================
	     * Lookup service name for port number if necessary
	     *==============================================================*/
	    if (IS_NUMERIC(token)) {
		port = atoi(token);
		if (port <= 0) {
		    fprintf(stderr, "Error: Invalid port number \"%s\" "
				"specified on line %d of %s\n",
				token, linen, conffn);
		    syslog(LOG_ERR, "Error: Invalid port number \"%s\" "
				"specified on line %d of %s",
				token, linen, conffn);
		    exit(1);
		}
	    }
	    else {
		if((sp = getservbyname(token,"tcp")) != (struct servent *) 0)
		    port = ntohs(sp->s_port);
		else {
		    fprintf(stderr, "Error: Unknown service \"%s\" on line "
				"%d of %s\n", token, linen, conffn);
		    syslog(LOG_ERR, "Error: Unknown service \"%s\" on line "
				"%d of %s", token, linen, conffn);
		    exit(1);
		}
	    }

	    /*===============================================================
	     * Optional buffer size
	     *==============================================================*/
	    if ((token = strtok(NULL, "\t,:;= \n")) != NULL) {
		rmlog.size = atoi(token);
		if (rmlog.size == 0) {
		    fprintf(stderr, "Error: Invalid buffer size \"token\" on "
				"line %d of %s\n", token, linen, conffn);
		    syslog(LOG_ERR, "Error: Invalid buffer size \"token\" on "
				"line %d of %s", token, linen, conffn);
		    exit(1);
		}

		/*===========================================================
		 * Optional time interval to force a send
		 *==========================================================*/
		if ((token = strtok(NULL, "\t,:;= \n")) != NULL) {
		    rmlog.interval = atoi(token);
		    if (rmlog.interval == 0) {
			fprintf(stderr, "Error: Invalid buffer size \"token\" "
				"on line %d of %s\n", token, linen, conffn);
			syslog(LOG_ERR, "Error: Invalid buffer size \"token\" "
				"on line %d of %s", token, linen, conffn);
			exit(1);
		    }
		}
	    }

	    /*===============================================================
	     * Fork a child to handle this "remotelog"
	     *==============================================================*/
	    if ((pid = fork()) < 0) {
		fprintf(stderr, "Error: Could not fork a child process.  %s\n",
			sys_errlist[errno]);
		syslog(LOG_ERR, "Error: Could not fork a child process.  %s",
			sys_errlist[errno]);
		exit(1);
	    }

	    /*===============================================================
	     * Parents continues to read the next config file line
	     * while child goes on to connect to the server
	     *==============================================================*/
	    if (pid != 0)
		continue;

	    /*===============================================================
	     * Connect to the server
	     *==============================================================*/
	    if ((rmlog.fd = connectserver(hn, port, 0, NULL)) < 0) {
		fprintf(stderr, "Error: Could not connect to server \"%s\" at "
				"port %d\n", hn, port);
		syslog(LOG_ERR, "Error: Could not connect to server \"%s\" at "
				"port %d", hn, port);
		exit(1);
	    }

	    /*===============================================================
	     * Log the connection in syslog
	     *==============================================================*/
	    syslog(LOG_NOTICE, "remotelog file %s to server %s at port %d",
			rmlog.infile, hn, port);

	    /*===============================================================
	     * Child process returns 0
	     *==============================================================*/
	    return(0);
	}
	else {
	    fprintf(stderr, "Error: Invalid statement on line %d of %s\n",
			linen, conffn);
	    syslog(LOG_ERR, "Error: Invalid statement on line %d of %s",
			linen, conffn);
	    exit(1);
	}
    }
    /*=======================================================================
     * Parent process returns 1
     *======================================================================*/
    return(1);
}
/*============================================================================
 *
 *===========================================================================*/
main(int argc, char *argv[]) 
{
    int i;
    int bytes;
    int size = 16384;
    pid_t pid;
    char conffn[BUFSIZ];
    char portname[BUFSIZ];
    struct sockaddr_in addr;
    unsigned long len;
    char histfn[BUFSIZ] = FW_CFG_DIR "/plog.hist";

    openlog("plog", LOG_PID|LOG_NDELAY, LOG_DAEMON);

    strcpy(conffn, FW_CFG_DIR "/plog.conf");
    signal(SIGCHLD, SIG_IGN);

    for (i = 1; i < argc; i++) {
	if (strcmp(argv[i], "-f") == 0) {
	    if (i == argc) {
		fprintf(stderr, "Error: -f flag not followed by a file name\n");
		syslog(LOG_ERR, "Error: -f flag not followed by a file name");
		exit(1);
	    }
	    if (strlen(argv[i+1]) > BUFSIZ) {
		fprintf(stderr, "Error: File name following -f is too long\n");
		syslog(LOG_ERR, "Error: File name following -f is too long");
		exit(1);
	    }
	    strcpy(conffn, argv[++i]);
	    continue;
	}
	if (strcmp(argv[i], "-h") == 0) {
	    if (i == argc) {
		fprintf(stderr, "Error: -h flag not followed by a file name\n");
		syslog(LOG_ERR, "Error: -h flag not followed by a file name");
		exit(1);
	    }
	    if (strlen(argv[i+1]) > BUFSIZ) {
		fprintf(stderr, "Error: File name following -h is too long\n");
		syslog(LOG_ERR, "Error: File name following -h is too long");
		exit(1);
	    }
	    strcpy(histfn, argv[++i]);
	    continue;
	}
    }

    /*===================================================================
     * Init the lock
     *==================================================================*/
    if ((semid=semget(IPC_PRIVATE,1,S_IRUSR|S_IWUSR|IPC_CREAT|IPC_EXCL))
									< 0) {
	fprintf(stderr, "Error: semget failed.  %s\n", sys_errlist[errno]);
	syslog(LOG_ERR, "Error: semget failed.  %s", sys_errlist[errno]);
	return(-1);
    }

    /*===================================================================
     * Make a daemon and write a pid file
     *==================================================================*/
    if ((pid = fork()) < 0) {
	fprintf(stderr, "Error: cannot fork.  %s\n", sys_errlist[errno]);
	syslog(LOG_ERR, "Error: cannot fork.  %s", sys_errlist[errno]);
	return(-1);
    }
    if (pid != 0)
	exit(0);

    writepid(PIDFILE);
#ifdef __hppa
    setpgid(0, 0);
#else
    setpgrp();
#endif
    signal(SIGTERM, term);

    /*=======================================================================
     * Read the config file
     * Parent process pauses forever
     *======================================================================*/
    if (readconf(conffn)) {
	pause();
	exit(0);
    }

    syslog(LOG_NOTICE, "plog Version 1.0 started.");

    /*=======================================================================
     * malloc the buffer
     *======================================================================*/
    if ((rmlog.bp = malloc(rmlog.size)) == NULL) {
	fprintf(stderr, "Error: Could not allocate the requested buffer\n");
	syslog(LOG_ERR, "Error: Could not allocate the requested buffer");
	exit(1);
    }

    /*=======================================================================
     * Open the input file
     *======================================================================*/
    close(0);
    if (open(rmlog.infile, 0) != 0) {
	fprintf(stderr, "Error: open failed for input file \"%s\"\n",
			rmlog.infile);
	syslog(LOG_ERR, "Error: open failed for input file \"%s\"",
			rmlog.infile);
	exit(1);
    }

    /*=======================================================================
     * Get the receiver's ip address and put it in rmlog to be used for
     * searching the history file
     *======================================================================*/
    len = sizeof(addr);
    if (getsockname(rmlog.fd, (struct sockaddr *)&addr, (socklen_t *)(&len)) < 0) {
	fprintf(stderr, "Error: getsockname failed.  %s\n", sys_errlist[errno]);
	syslog(LOG_ERR, "Error: getsockname failed.  %s", sys_errlist[errno]);
	exit(1);
    }
    bcopy(&addr.sin_addr, &rmlog.ip, sizeof(rmlog.ip));

    /*===============================================================
     * Get the last count from the history file and position the file
     * pointer at the spot.
     *==============================================================*/
    if ((bytes = get_hist_count(histfn)) > 0) {
	if (lseek(0, bytes, SEEK_SET) < 0) {
	    fprintf(stderr, "Error: seek error in \"%s\".  %s\n",
				sys_errlist[errno]);
	    syslog(LOG_ERR, "Error: seek error in \"%s\".  %s",
				sys_errlist[errno]);
	    exit(1);
	}
    }
    else if (bytes < 0)
	exit(1);

    /*=======================================================================
     * Catch SIGPIPE so we can inform when the receiver closed the connection
     * on the other side.
     * Note that this process won't know the other side has closed until it
     * tries to write to the closed socket.
     * This is what will happen when the receiver closes the connection:
     * 1. this child process will continue to block in the read() call until
     *    new input comes in or timer alarm went off
     * 2. if new input has arrived, sendbuf() will be called
     * 3. the write() call in sendfbuf() now generates a SIGPIPE because of
     *    the closed socket
     * 4. nop() is invoked and returns without doing anything
     * 5. code in sendbuf() detects the -1 return from write() and issues
     *    the error messages and then exits.
     *
     * Without providing a SIGPIPE handler, the process will just exit with
     * no messages.
     *======================================================================*/
    signal(SIGPIPE, nop);

    /*=======================================================================
     * Infinite loop
     *======================================================================*/
    signal(SIGALRM, alrm);
    alarm(rmlog.interval);
    while ((bytes = read(0, rmlog.bp + rmlog.bytes_filled,
				rmlog.size - rmlog.bytes_filled)) != -1 ||
			errno == EINTR) {
	if (bytes == 0 || errno == EINTR) {
	    sendbuf();
	    if (bytes == 0)
		sleep(1);
	    continue;
	}
	rmlog.bytes_filled += bytes;
	if (rmlog.bytes_filled == rmlog.size)
	    sendbuf();
    }

    exit(0);
}
