/*============================================================================
 * filename:		raproxy/raprules.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  All rights
 * reserved.
 *		
 * Date written:	April     11, 1995
 * Last modified:	February   8, 2000 - added LPL
 *	                         
 * function:                       
 *	The rarules program reads the permit and deny rules in the Real Audio
 *	proxy configuration file. 
 *
 * inputs:
 *	prt to RAPROXY record: 
 *	raproxy configuration file  
 *
 *      file format:           
 * 
 *      Manditory Commands are ignored
 *
 *		groupid  = name
 *		timeout = nnn 		- number of seconds 
 * 		ftproxpath = pathname
 *		userid  = name   
 *
 *	Process permit and deny commands: 
 *	
 * 	permit from src_addr to dest_addr  
 *	deny   from src_addr to dest_addr 
 *
 * outputs:	 
 * 		return  1 if Denied Access 
 *              return  0 if Permitted 
 *		return -1 if error
 * 
 *
 *============================================================================*/
#include	<sys/types.h>
#include	<stdio.h>
#include	<syslog.h>
#include	<string.h>
#include	<sys/time.h> 
#include	<pwd.h>
#include	<grp.h> 
#include	<netdb.h>               /* defines servent structure      */
#include	<netinet/in.h>
#include 	"firewall.h"  
#include	"rap.h"  
char 	raproxcfgfile[] = RAPROXCFGFILE ;
int rulematch(char *cp, struct RAPROXY *proxyptr, int line);
/*=============================================================================
 * function: raprules    Real Audio Proxy rules
 *===========================================================================*/
int raprules(struct RAPROXY *proxptr)
{    
   char		record[MAXRECLN];
   FILE 	*cfgfp; 
   char 	*cp; 
   char		*cp2;
   char		*token1;
   char		*token2; 
   char		tokensep[]   = "\t,:;= \n";
   int		hostfound    = 0;
   int		line	     = 0;  
   int		match        = 0;
   struct servent	*sp; 		/* ptr to service entry 	*/ 
   struct passwd	*pw; 
   struct group		*gp; 
   /* open raproxy conf file */
   if (!(cfgfp = fopen(raproxcfgfile, "r")))
   {
      syslog(LOG_ERR,"Error: open failed for %s: %m",raproxcfgfile);
      return(-1);    			/* open failed signal error */
   } 
   cp = (char *) &record; 

   /*===================================================== 
    * Get permit or deny rules from raproxy config file. 
    *====================================================*/      
   while((  cp = fgets(record, MAXRECLN, cfgfp)) != NULL)
   { 
      line++;   
      if( cp == NULL) break;
      /* skip over comments and null lines */
      if( (*cp == '#') || (*cp == '\n') ) 
	 goto getnextline; 
      if( (token1 = strtok(record,tokensep)) == NULL) 
	 goto getnextline; 
      /*==================================================================
       * if permit rule then match source host name then source port
       *==================================================================*/
      if(strcmp(token1,"permit") == 0)
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL) /* get 'from'  */
         { 
            syslog(LOG_ERR,"Error: permit is missing parameters on line %d of %s.",
 				line,raproxcfgfile); 
	    fclose(cfgfp);
	    return(-1);
         }
         match = rulematch(token2, proxptr,line);   
         if( match == 1 )  
         {                 
            fclose(cfgfp);
            return(0);     	/* access permitted	*/
         }
         if( match == 0) 
	    goto getnextline;   
         fclose(cfgfp);
         return(-1);         
      } 
      /*===================================================================
       * if deny rule match source name and source port
       *==================================================================*/ 
      if(strcmp(token1,"deny") == 0)
      {
         if( (token2 = strtok(NULL,tokensep)) == NULL) /* get source host  */
         { 
            syslog(LOG_ERR,"Error: deny is missing parameters on line %d of %s.",
 				line,raproxcfgfile); 
	    fclose(cfgfp);
	    return(-1);
         }
         match = rulematch(cp+7, proxptr, line); 
         if( match == 0 ) 
	    goto getnextline;
         if( match == 1 )
	 { 
            fclose(cfgfp);
            return(1); 
         } 
         fclose(cfgfp);
         return(-1);    
      } 
      getnextline:;
   }
   fclose(cfgfp); 
    syslog(LOG_ERR,"Permission denied to connect player from %s to %s.",
 			proxptr->srcIP, proxptr->destIP); 
   return(1);   /* deny accees - no match found */


}  
/*=========================================================================  
 * rulematch: 
 *    
 * inputs:  	pointer to second byte past permit command
 *		pointer to TNPROXY structure	
 * 
 * output: -1 on error
 *	    0 if rule not matched
 *          1 if rule applies 
 *========================================================================*/
int rulematch(char *cp, struct RAPROXY *proxptr, int line)
{   
   int		usermatch;
   int		groupmatch;
   int		srcmatch; 
   int		destmatch;
   char		*token2;  	/* from keyword			*/
   char		*token3;        /* source IP address  		*/
   char		*token4;        /* from keyword			*/
   char		*token5;        /* source IP address		*/
   char		*token6; 
   char		*token7;
   char		tokensep[]   = "\t,:;= \n"; 
   /* initialze variables	*/
   usermatch  = 0;
   groupmatch = 0;
   srcmatch   = 0; 
   destmatch  = 0;     
   token2 = cp;
   if( strcmp(token2,"from") ) 
   {
      syslog(LOG_ERR,"Error: missing 'from' keyword on line %d "
                     "of %s. found '%s'", line,raproxcfgfile,token2); 
      return(-1);
   }
   if( (token3 = strtok(NULL,tokensep)) == NULL)
   {
      syslog(LOG_ERR,"Error: missing source IP address on line %d of %s.",
 				line,raproxcfgfile); 
      return(-1);
   }
   if( strcmp(token3,"*") == 0 )
      srcmatch++;
   else   
   {  
      /*================================================================ 
       * If source address ends with wildcard then shorten string
       * and do a compare for length of short string, else compare
       * strings as is. 
       *===============================================================*/
      if( token3[strlen(token3) -1] == '*') 
      {
         token3[strlen(token3)-1] = '\0'; 
         if( (strncmp(token3,proxptr->srcIP,strlen(token3)) != 0) ) 
	 return(0);  /* rule not matched	*/   
      }
      else
      {
         if( (strcmp(token3,proxptr->srcIP) != 0) ) 
            return(0);  /* rule not matched	*/  
      }
         srcmatch++;
   }
   if( (token4 = strtok(NULL,tokensep)) == NULL)
   {
      syslog(LOG_ERR,"Error: missing 'to' parameters on line %d of %s.",
 				line,raproxcfgfile); 
      return(-1);
   }
   if( (token4 = strtok(NULL,tokensep)) == NULL)
   {
      syslog(LOG_ERR,"Error: missing destination IP parameters on line %d of %s.",
 				line,raproxcfgfile); 
      return(-1);
   }
   if( strcmp(token4,"*") == 0 )
      destmatch++; 
   else   
   {  
      /*================================================================ 
       * If destination address ends with wildcard then shorten string
       * and do a compare for length of short string, else compare
       * strings as is. 
       *===============================================================*/
      if( token4[strlen(token4) -1] == '*')
      {
         token4[strlen(token4)-1] = '\0'; 
         if( (strncmp(token4,proxptr->destIP,strlen(token4)) != 0) ) 
	    return(0);  /* rule not matched	*/
      }
      else
      {
         if( (strcmp(token4,proxptr->destIP) != 0) ) 
	    return(0);  /* rule not matched	*/  
      }               
      destmatch++;
   }
   
   if( srcmatch && destmatch ) 
      return(1);	/* match found	*/     
   
   return(0);
}
