/*============================================================================
 * filename:		rpcproxy/rpcproxy.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1996-2000.  
 * All rights reserved.   
 *
 *
 * Date written:	November   1, 1996
 * Last modified:	February   8, 2000
 *			
 *
 * function:		rpc proxy server	
 *			udp proxy server 
 *			NFS support
 *
 *
 * Copyright 1993 Tom Fitzgerald <fitz@wang.com>; permission to use, modify
 * and distribute this source is granted provided only that this notice is
 * retained intact.
 *
 * Parts of this program were drawn from tcp-wrapper by Wietse Venema.
 *
 *
 *===========================================================================*/

/*=========================================================================
 * Terminology:
 * A fanout is a structure controlling access to a single unencapsulated
 *	local port.  All "permit from" lines which share access to a localport
 *	or a range of localports (on either side) share a fanout structure.
 *	On receiving a packet, the fanout decides which permit line or instance
 *	is to get the packet.
 *	Fanout is created from config file and has generics (also created from
 *	config file) and instances (created on the fly) hanging from it, for
 *	all config file directives and connections through this port or range
 *	of ports.  Fanout owns the socket for the port.
 * A dynamic fanout is a fanout which is created as a clone of a statically
 *	built fanout (from config file).  A dynamic fanout always owns a
 *	specific localport while it can be cloned from a regular fanout which
 *	lists a range of ports.  A dynamic fanout also has flags set which
 *	make it subject to timeouts.
 * An encaps defines a single local port mentioned in "encapsulate" lines from
 *	the config file.  Has generics hanging from it, one for each
 *	"encapsulate" directive, specifying the wildcarded name or addr that
 *	can use the encapsulator.  Also has instances, connections set up
 *	through the specified localport.  Created only from config file.
 * A plex (multiplexor) is either an encaps or a fanout; it controls a single
 *	local port of either kind or a range of local ports.
 * A generic is a single remote host/port specification.  The spec can be
 *	wildcarded ("*.dom.ain" or address/mask), but even if it's not, it is
 *	still treated as a generic until a packet arrives, to handle multi-
 *	homed hosts.  A generic hanging from an encaps has no partner, since
 *	the partner is determined by the incoming packet.  A generic hanging
 *	from a fanout has a partner which is a template of the far side of
 *	the connection for each instance created from the generic.
 * A dynamic generic is a generic which is subject to timeouts.
 * An instance is a single exactly matched remote address/port, local port
 *	set.  It receives packets and is matched with a partner instance which
 *	is where the packets are sent out to.  This is the one assoc type
 *	which is always created on the fly.
 * An assoc (association) is any one of the above; half of a conversation.
 *===========================================================================*/
static  char	ident[] = "@(#) rpcproxy, V 1.0, 00/01/21"
 			  " (C) Freemont Avenue Software, Inc."
			  " 1996-2000";

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <syslog.h>
#include <errno.h>
#include <poll.h>
#include <string.h>
#include <rpc/rpc.h>
#include <rpc/pmap_prot.h>
#ifdef _AIX
#  include <sys/utsname.h>
#endif
#if !defined(_AIX) || defined(_AIX41)
#include <rpc/rpc_msg.h>
#endif
#ifdef __sun
#include <rpc/clnt_soc.h>
#include <sys/filio.h>
#endif
#ifdef _AIX
#include <sys/ioctl.h>
#endif
#include <pwd.h>
#include <grp.h>
#ifdef linux
#include <asm/ioctls.h>
#include <rpc/clnt.h>
#endif
#include "firewall.h"
#include "rpcproxy.h"

#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE  1
#endif

#ifndef INADDR_NONE
#define INADDR_NONE ((long) 0xffffffff)
#endif

#ifndef linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif

/* extern char *malloc (); */

/* #define LogInstances	1 */
#define	MAXBUFLEN	65536
#define InitPktBufLen	MAXBUFLEN

#define giderror(x)  { \
			struct group *gp; \
			int en=errno; \
			if ((gp = getgrgid(x)) == NULL) \
			    syslog(LOG_ERR,"Error: Could not get group id " \
				"or set effective group id to %d, %m", x, \
				sys_errlist[en]); \
			else \
			    syslog(LOG_ERR,"Error: Could not set effective " \
				"group id to %s, %m",gp->gr_name, \
				sys_errlist[en]); \
		     }

#define uiderror(x)  { \
			struct passwd *pw; \
			int en=errno; \
			if ((pw = getpwuid(x)) == NULL) \
			    syslog(LOG_ERR,"Error: Could not get user id " \
				"or set effective user id to %d, %m", x, \
				sys_errlist[en]); \
			else \
			    syslog(LOG_ERR,"Error: Could not set effective " \
				"user id to %s, %s",pw->pw_name, \
				sys_errlist[en]); \
		     }
/*===========================================================================
 * Return TRUE if both addresses point to the same address/port pair
 *==========================================================================*/
#define sameaddr(a1,a2)	((a1)->sin_addr.s_addr == (a2)->sin_addr.s_addr \
			  && (a1)->sin_port == (a2)->sin_port)

typedef struct assoc {
	struct assoc	*next;		/* Global chain */
	struct assoc	*partner;
	struct assoc	*generics;	/* Used fanout or encaps */
	struct assoc	*instances;	/* Used if fanout or encaps */
	struct assoc	*nextinst;	/* Used if generic or instance */
	struct assoc	*plex;		/* Fanout/encaps for this localport */
	SockIn		sin;		/* Far end of connection */
	struct in_addr	addrmask;
	struct in_addr	localip;	/* 0 if "any" and not yet open */
	u_short		localport;	/* 0 if "any" and not yet open */
	u_short		highport;	/* 0 means no range */
	int		socket;		/* -1 if not open; copy if instance */
	u_long		nbytes;		/* #bytes received with this instance */
	char		fanout;		/* 1 if this just runs a localport */
	char		inst;		/* 1 if this is an instance */
	char		receiver;	/* 1 means this assoc receives pkts */
	char		priv;		/* 1 means this assoc needs privport */
	char		dynamic;	/* 1 means this assoc is dynamic */
	char		tcp;		/* 1 means this assoc is tcp */
	char		tftp;		/* 1 means this assoc is tftp */
	char		encaps;		/* 1 if pkts are encapsulated */
	char		transient;	/* 1 if session will time out */
	char		ownport;	/* 0 if port is copy of plex's */
	time_t		faccess;	/* time of first access */
	time_t		lastaccess;	/* timestamp */
} Assoc;

static char		*pktbuf = NULL;
static int		pktbuflen = InitPktBufLen;	/* Packet buffer size */

static Assoc		*assoclist = NULL;	/* Global chain of assocs */
static Assoc		*lastassoc = NULL;	/* Last assoc in list */

static int		nfilemax;
static struct pollfd	*fdlist;
static Assoc		**fdlistassocs;
static int		fdlistlen;
static char		rpcconffile[PATH_MAX]=FW_CFG_DIR "/rpcproxy.conf";

static	RPCCONF		rpcconf;
static	PMAPREQ		*pmapreq;
static	char		fixed_ports_done=FALSE;
static	int		one=1;

#ifdef  Debug
#define	DBPRINTF(n,X)	if (rpcconf.traceflag >= n) printf X
#else
#define	DBPRINTF
#endif

#ifdef Debug
/*============================================================================
 * Debug function to dump an assoc
 *==========================================================================*/
void dumpassoc(Assoc *assoc, char *msg)
{
   unsigned short port;

   if (msg)
	DBPRINTF(4, ("%s",msg));
   DBPRINTF(4, ("ASSOC 0x%08x\n", assoc));
   DBPRINTF(4, (" next = 0x%08x	partner = 0x%08x	generics = 0x%08x\n",
		assoc->next, assoc->partner, assoc->generics));
   DBPRINTF(4, (" instances = 0x%08x	nextinst = 0x%08x	"
		"plex = 0x%08x\n",
		assoc->instances, assoc->nextinst, assoc->plex));
   DBPRINTF(4, (" sin_addr = %s	sin_port = %d\n",
		inet_ntoa(assoc->sin.sin_addr), ntohs(assoc->sin.sin_port)));
   DBPRINTF(4, (" addrmask = %s\n",
		inet_ntoa(assoc->addrmask)));
   DBPRINTF(4, (" local_ip = %s	local_port = %d	highport = %d\n",
		inet_ntoa(assoc->localip), ntohs(assoc->localport),
		ntohs(assoc->highport)));
   DBPRINTF(4, (" socket = %d	fanout = %d	receiver = %d\n",
		assoc->socket, assoc->fanout, assoc->receiver));
   DBPRINTF(4, (" priv = %d	dynamic = %d	tcp = %d\n",
		assoc->priv, assoc->dynamic, assoc->tcp));
   DBPRINTF(4, (" encaps = %d	transient = %d	ownport = %d\n",
		assoc->encaps, assoc->transient, assoc->ownport));
   DBPRINTF(4, (" lastaccess = 0x%08x\n", assoc->lastaccess));
   return;
}

void dumpallassoc()
{
   Assoc *next;

   for (next=assoclist; next; next=next->next)
	dumpassoc(next,0);
   DBPRINTF(4, ("------------------\n"));
   return;
}
#endif

/*============================================================================
 * Re-allocate packet buffer for bigger packets (use smallest power of 2
 * that's at least as large as len).  If len == 0, allocate initial buffer.
 * Return TRUE if ok, FALSE if failed.
 *==========================================================================*/
int makepktbuf (int len)
{
   char		*newbuf;

   if (len > 0)
   {
      while (pktbuflen < len)
         pktbuflen <<= 1;
      syslog (LOG_WARNING, "Boosting buf size to %d\n", pktbuflen);
   }
   if ((newbuf = malloc (pktbuflen)) == NULL)
   {
      syslog (LOG_ERR, "Error: malloc (%d) failed, out of memory", len);
      return (FALSE);
   }
   if (pktbuf != NULL)
   {
      free (pktbuf);
   }
   pktbuf = newbuf;
   return (TRUE);
}


/*============================================================================
 * Allocate an assoc
 *===========================================================================*/
Assoc *getassoc()
{
   Assoc	*assoc;

   assoc = (Assoc *) malloc (sizeof (Assoc));
   if (assoc == NULL)
      return (NULL);
   bzero ((char *) assoc, sizeof (Assoc));
   assoc->socket	 = -1;
   assoc->sin.sin_family = AF_INET;

   /*=========================================================================
    * When reading in the config file, we want to thread new assoc to the end
    * so that the permit rules are placed in the same order as they are coded.
    * After config file done, we want to thread new assoc at the top so that
    * a matching instance is found before a matching generic.
    * assoclist == NULL && lastassoc == NULL is an indication that we just
    * started the program and haven't read a permit yet.
    * At the end of processing the config file, lastassoc is set to NULL so
    * that assoclist != NULL && lastassoc == NULL means we want to thread at
    * the top.
    *========================================================================*/
   if (assoclist == NULL && lastassoc == NULL)
   {
	assoc->next	= NULL;
	assoclist	= assoc;
	lastassoc	= assoc;
   }
   else if (lastassoc)
   {
	assoc->next	= NULL;
	lastassoc->next	= assoc;	/* Thread it at end of list */
	lastassoc	= assoc;
   }
   else
   {
	assoc->next	= assoclist;	/* Thread it at top of list */
	assoclist 	= assoc;
   }
   return (assoc);
}


/*============================================================================
 * Delete an instance assoc, removing it from the global chain and from the
 * list of instances hanging off the same plex.
 *==========================================================================*/
void deleteassoc (assoc)
   Assoc	*assoc;
{
   Assoc	**prev;

   /*================================================================
    * Remove from global chain
    *===============================================================*/
   for (prev = &assoclist; *prev != assoc; prev = &(*prev)->next)
		;
   *prev = assoc->next;
   /*===============================================================
    * Remove from chain of instances hanging off the same multiplex
    *==============================================================*/
   if (assoc->plex != NULL)
   {
      for (prev = &assoc->plex->instances; (*prev && *prev != assoc);
           prev = &(*prev)->nextinst)
			;
      if (*prev)
      {
         *prev = assoc->nextinst;
         DBPRINTF(2, ("Removed assoc 0x%08x from instances list\n",assoc));
      }
      else
      /*===============================================================
       * Since it was not on the instances list, it may be on the
       * generics list.
       * Remove from chain of generics hanging off the same multiplex
       *==============================================================*/
      {
         for (prev = &assoc->plex->generics; (*prev && *prev != assoc);
           prev = &(*prev)->nextinst)
			;
         if (*prev)
         {
            *prev = assoc->nextinst;
            DBPRINTF(2, ("Removed assoc 0x%08x from generics list\n",assoc));
         }
      }
   }

   free (assoc);
}


/*===================================================================
 * Return the fanout associated with a particular local port
 * Find an assoc
 * 1. which is a fanout
 * 2. whose localip matches the localip
 * 3. in the same protocol			AND 4 or 5
 *
 * 4. whose localport = port or whose localport is within range
 * 5. the fanout assoc and the search arguments both use a port range
 * Note that if the localip argument is INADDR_ANY, it should not automatically
 * be a match because the fanout may be just for a specific local ip and
 * not for all.
 *===================================================================*/
Assoc *findfanout (struct in_addr localip, u_short port, u_short highport,
				int tcp)
{
   Assoc	*assoc;

   for (assoc = assoclist; assoc != NULL; assoc = assoc->next)
      if (assoc->fanout	&&
		assoc->localip.s_addr == localip.s_addr &&
		assoc->tcp == tcp)
      {
         if ((highport == 0 &&
		((assoc->localport == port && assoc->highport == 0)	||
			(assoc->highport &&
			 ntohs(port) >= ntohs(assoc->localport) &&
			 ntohs(port) <= ntohs(assoc->highport))))	||
	     (highport != 0 && assoc->highport != 0))
            break;
      }
   return (assoc);
}


/*=============================================================================
 * makefdlist()
 * Rebuild the socket list from a (potentially new) set of assocs.
 * Sets fdlist, fdlistassocs and fdlistlen
 *===========================================================================*/
void makefdlist ()
{
   int		j;
   Assoc	*assoc;

   j = 0;
   for (assoc = assoclist; assoc != NULL; assoc = assoc->next)
   {
      /*===================================================
       * !!! This test picks up fanouts with no receiving
       * instances or generics; should filter these out.
       *==================================================*/
      if (assoc->ownport && assoc->socket >= 0
            && (assoc->fanout || assoc->encaps || assoc->receiver))
      {
         fdlist [j].fd = assoc->socket;
	 fdlist [j].events = POLLIN;
	 fdlist [j].revents = 0;
	 fdlistassocs [j] = assoc;
	 DBPRINTF (2, ("polling socket %d, assoc 0x%x\n",
			  assoc->socket, assoc));
	 ++j;
      }
   }
   if (j == 0)
   {
      syslog (LOG_ERR, "Error: No sockets to listen on, exiting.");
      exit (1);
   }
   fdlistlen = j;
}

struct preopened {
   struct in_addr	localip;
   u_short		port;
   int			socket;
   int			tcp;
   struct preopened	*next;
};

typedef struct preopened preopened;
static preopened *preo_list=NULL;

/*============================================================================
 * Given a socket number, this function looks thru the preo_list and see if
 * the socket number is recorded in any entry.  If it is, it will open a
 * socket and bind it to the port recorded in the corresponding entry.
 *==========================================================================*/
void close_and_preopen(int sock)
{
   preopened *preo=NULL;
   SockIn address;

   close(sock);

   for (preo=preo_list; preo; preo = preo->next)
   {
	if (sock == preo->socket)
	    break;
   }
   if (!preo)
	return;

   /*========================================================================
    * Open a socket
    *=======================================================================*/
   if (preo->tcp)
   {
	preo->socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (preo->socket >= 0 &&
		setsockopt(preo->socket,SOL_SOCKET,SO_REUSEADDR,(char *)&one,
							sizeof(int)) < 0)
	    syslog(LOG_ERR, "Error: setsockopt failed to make socket %d "
				"reuseable.  %m",preo->socket);
   }
   else
	preo->socket = socket(AF_INET, SOCK_DGRAM,  IPPROTO_UDP);
   if (preo->socket < 0)
   {
	syslog (LOG_ERR, "Error: Failed to preopen %s socket (port %d). %m.",
		preo->tcp?"TCP":"UDP",ntohs(preo->port));
	return;
   }
   /*========================================================================
    * Prebind to the port
    *=======================================================================*/
   memset(&address,0,sizeof(address));
   address.sin_family = AF_INET;
   address.sin_addr = preo->localip;
   address.sin_port = preo->port;
   if (bind (preo->socket, (struct sockaddr *)&address, sizeof (address)) < 0)
   {
	syslog(LOG_ERR,"Error: Failed to re-prebind %s socket %d to port %d. "
		"%m.", preo->tcp?"TCP":"UDP",preo->socket,ntohs(preo->port));
	close(preo->socket);
	preo->socket = -1;
	return;
   }

   DBPRINTF(2,("      Successfully re-preopened/prebound %s socket %d as IP %s "
			"port %d.\n",
			preo->tcp?"TCP":"UDP",
			preo->socket,inet_ntoa(preo->localip),preo->port));
   return;
}

/*============================================================================
 * Handle pre-opening of ports that are permitted for RPC connections.
 * This essentially reserves these ports for use by the permitted RPC
 * connections only.
 * The strategy is to preopen and then prebin dthe port in question.  When
 * a request comes in later and rpcproxy really needs to open the port,
 * the preopened socket will then be closed, making the port available for
 * opensocket().  Later, when the connection timeout, the socket opened by
 * opensocket() will then be closed and another preopen will be done on
 * the port to reserve it for later use.
 * Note the *preo->tcp == tcp test should work because they are either
 * specifically set to TRUE or left as zero, or 'computed' by a comparison.
 * I.e. there should not be any non-canonical TRUE values (non-zero other
 * than 1).
 *==========================================================================*/
void preopen(struct in_addr localip, u_short port, int tcp)
{
   preopened **preo,*new;
   SockIn address;

   for (preo=&preo_list; *preo; preo=&(*preo)->next)
   {
	/*===================================================================
	 * If these all match, it means we already opened this port
	 *==================================================================*/
	if ((*preo)->localip.s_addr == localip.s_addr &&
		(*preo)->port == port &&
		(*preo)->tcp == tcp)
	{
	    DBPRINTF(2,("      Already preopened %s IP %s port %d.  Must be "
			"duplicate.\n",
			tcp?"TCP":"UDP",inet_ntoa(localip),port));
	    return;
	}
   }
   if ((new = malloc(sizeof(preopened))) == NULL)
   {
	syslog(LOG_ERR,"Error: malloc for a preopen list failed.");
	exit(1);
   }
   /*========================================================================
    * Attach the new structure at the end of the list
    *=======================================================================*/
   if (preo_list)
	*preo = new;
   else
	preo_list = new;

   new->localip = localip;
   new->port = port;
   new->tcp = tcp;
   new->next = NULL;
   /*========================================================================
    * Preopen a socket
    *=======================================================================*/
   if (tcp)
   {
	new->socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (new->socket >= 0 &&
		setsockopt(new->socket,SOL_SOCKET,SO_REUSEADDR,(char *)&one,
							sizeof(int)) < 0)
	    syslog(LOG_ERR, "Error: setsockopt failed to make socket %d "
				"reuseable.  %m",new->socket);
   }
   else
	new->socket = socket(AF_INET, SOCK_DGRAM,  IPPROTO_UDP);
   if (new->socket < 0)
   {
	syslog (LOG_ERR, "Error: Failed to preopen %s socket (port %d). %m.",
		tcp?"TCP":"UDP",ntohs(port));
	exit(1);
   }
   /*========================================================================
    * Prebind to the port
    *=======================================================================*/
   memset(&address,0,sizeof(address));
   address.sin_family = AF_INET;
   address.sin_addr = localip;
   address.sin_port = port;
   if (bind (new->socket, (struct sockaddr *)&address, sizeof (address)) < 0)
   {
	syslog(LOG_ERR,"Error: Failed to prebind %s socket %d to port %d. %m.",
		tcp?"TCP":"UDP",new->socket,ntohs(port));
	close(new->socket);
	exit(1);
   }

   DBPRINTF(2,("      Successfully preopened/prebound %s socket %d as IP %s "
			"port %d.\n",
			tcp?"TCP":"UDP",new->socket,inet_ntoa(localip),port));
   return;
}

/*============================================================================
 * Loop thru all the assoc's and for each fanout with the portmapper port
 * and which has an opened socket, call the server's (ip in sin.sin_addr)
 * portmapper to get a list of all the RPC ports that are active.
 *==========================================================================*/
void get_active_rpc_ports()
{
   Assoc *cur111;	/* points to a fanout that has port 111 */
   Assoc *generic;
   struct sockaddr_in server;
   struct timeval to;

   to.tv_sec = 20;
   to.tv_usec = 0;

   memset(&server,0,sizeof(server));
   server.sin_family = AF_INET;
   server.sin_port = htons(PMAPPORT);

   for (cur111=assoclist; cur111; cur111=cur111->next)
   {
	int sock;
	char *dummy;
	char tmp[16];
	struct pmaplist *curport,*portlist;
	CLIENT *client;
	enum clnt_stat clnt_stat;

	/*===================================================================
	 * If not portmap port or if socket not opened yet, don't have to do
	 * anything
	 *==================================================================*/
	if (cur111->localport != PMAPPORT ||
		cur111->socket < 0)
	    continue;

	/*===================================================================
	 * Loop thru the generics list
	 *==================================================================*/
	for (generic = cur111->generics; generic; generic = generic->nextinst)
	{
	    sock = RPC_ANYSOCK;

	    server.sin_addr = generic->partner->sin.sin_addr;

	    strcpy(tmp,inet_ntoa(server.sin_addr));
	    DBPRINTF (4, ("Found a portmapper fanout.  Remote server %s, "
			"Local IP %s\n", tmp, inet_ntoa(cur111->localip)));

	    /*===============================================================
	     * Ask RPC to create the client structure
	     *==============================================================*/
	    if ((client = clnttcp_create(&server,PMAPPROG,PMAPVERS,
					&sock,BUFSIZ,BUFSIZ)) == NULL &&
		(client = clntudp_create(&server, PMAPPROG,PMAPVERS,to,&sock))
				== NULL)
	    {
		syslog(LOG_ERR,"Error: Failed to create portmap client for "
			"server %s.\n", inet_ntoa(server.sin_addr));
		exit(1);
	    }

	    /*===============================================================
	     * Call the server's portmapper
	     *==============================================================*/
	    portlist = NULL;
	    clnt_stat = clnt_call(client,PMAPPROC_DUMP,(xdrproc_t)xdr_void,
			(char *)&dummy,(xdrproc_t)xdr_pmaplist,(char *)&portlist,
					to);
	    if (clnt_stat != RPC_SUCCESS)
	    {
		syslog(LOG_ERR,"Error: Call to portmapper on %s failed.",
			inet_ntoa(server.sin_addr));
		/* Don't exit here - otherwise, if one of the servers is down,
			we will not be able to successfully start rpcproxy */
		return;
	    }

	    clnt_destroy(client);
	    close(sock);

	    /*===============================================================
	     * For each port in the returned list, check to see if there is
	     * some permit rule that specified it.
	     *==============================================================*/
	    for (curport=portlist; curport; curport = curport->pml_next)
	    {
		Assoc *temp;

		DBPRINTF(4, ("   Server %s returned %s RPC port %d.\n",
			inet_ntoa(server.sin_addr),
			curport->pml_map.pm_prot == IPPROTO_TCP?"TCP":"UDP",
			curport->pml_map.pm_port));
		if (curport->pml_map.pm_port != htons(PMAPPORT) &&
		    (temp = findfanout(cur111->localip,curport->pml_map.pm_port,
				0,(curport->pml_map.pm_prot == IPPROTO_TCP))) &&
				temp->socket < 0)
		{
		    preopen(cur111->localip,curport->pml_map.pm_port,
				(curport->pml_map.pm_prot == IPPROTO_TCP));
		}
	    }
	}
   }
   return;
}


/*============================================================================
 * Locate any instances that have not been used in the last idletime seconds
 * and delete them and their partners.  Returns number of milliseconds to
 * next timeout, or -1 if no timeouts pending.
 *===========================================================================*/
int checktimeouts ()
{
   Assoc	*assoc;
   Assoc	*next;
   Assoc	*partner;
   time_t	curtime;
   time_t	nexttime;
   char		changed=0;

   curtime = time (NULL);
   nexttime = -1;
   for (assoc = assoclist; assoc != NULL; assoc = next)
   {
	next = assoc->next;
	if (!assoc->transient)
	    continue;		/* Don't timeout this one */
	if (assoc->lastaccess + rpcconf.idletime <= curtime)
	{
	    partner = assoc->partner;
	    DBPRINTF (1, ("Timed out assoc 0x%x with socket %d\n",
			  assoc, assoc->socket));
	    if (partner != NULL)
	    {
		DBPRINTF (1, ("...and partner 0x%x with socket %d\n",
			  partner, partner->socket));
		if (partner->ownport && partner->socket >= 0)
		{
		    DBPRINTF (1, ("closing socket %d\n", partner->socket));
		    close_and_preopen (partner->socket);
		    changed = 1;
		}

		/*============================================================
		 * Log statistics before we get rid of instances
		 * If the faccess member is non-zero, the instance is the LHS.
		 *===========================================================*/
		if (assoc->inst)
		{
		    Assoc	*lhs;
		    char	th1[16],th2[16];
		    time_t	endtime;

		    if (assoc->faccess)
			lhs = assoc;
		    else
			lhs = partner;
		    strcpy(th1,inet_ntoa(lhs->sin.sin_addr));
		    strcpy(th2,inet_ntoa(lhs->partner->sin.sin_addr));
		    endtime = time(NULL);
		    syslog(LOG_NOTICE,"exit %s src_ip = %s, src_port = %d, "
			"local_ip = %s, local_port = %d, local_port2 = %d, "
			"dest_ip = %s, dest_port = %d bytes_from_src = %d "
			"bytes_to_src = %d duration = %d sec",
			lhs->tcp?"TCP":"UDP",
			th1, ntohs(lhs->sin.sin_port),
			inet_ntoa(lhs->localip), ntohs(lhs->localport),
			ntohs(lhs->partner->localport),
			th2, ntohs(lhs->partner->sin.sin_port),
			lhs->nbytes, lhs->partner->nbytes,
			endtime - lhs->faccess);
		}

		deleteassoc (partner);
		/*============================================================
		 * Set "next" again since it may be pointing to "partner"
		 * which was just deleted.
		 *===========================================================*/
		next = assoc->next;
	    }
	    if (assoc->ownport && assoc->socket > 0)
	    {
		DBPRINTF(1, ("closing socket %d\n", assoc->socket));
		close_and_preopen(assoc->socket);
		changed = 1;
	    }
	    deleteassoc (assoc);
	}
	else if (nexttime == -1 ||
		assoc->lastaccess + rpcconf.idletime < nexttime)
		nexttime = assoc->lastaccess + rpcconf.idletime;
   }
   if (changed)
	makefdlist ();
   if (nexttime != -1)
	nexttime = (nexttime - curtime) * 1000;
   return (nexttime);
}


/*============================================================================
 * Open socket - return TRUE for success, FALSE for failure
 *===========================================================================*/
int opensocket (Assoc *assoc)
{
   int	sock;
   int	len;
   int  rc=0;
   int	maxb;
   int	lmaxb;
   SockIn	myaddr;
   preopened *preo=NULL;

   DBPRINTF (1, ("opensocket assoc=0x%x, %s localport=%d\n", assoc,
	  assoc->tcp?"TCP":"UDP",htons (assoc->localport)));
   /*=========================================================================
    * If we need to get a priviledged local port, we will switch back to root
    * first.  The seteuid needs to be here because Solaris insists that the
    * socket be owned by root to bind to a priv port.  AIX only needs to be
    * root at the bind for priv ports.
    *========================================================================*/
   if (((assoc->localport == INADDR_ANY && assoc->priv) ||
        (assoc->localport != INADDR_ANY && ntohs(assoc->localport) <= 1023)
       ) &&
#ifdef __hppa
       geteuid() != 0 && setresuid(-1,0,-1) < 0)
#else
       geteuid() != 0 && seteuid(0) < 0)
#endif

      uiderror(0);

   /*=========================================================================
    * Check to see if the port is a preopened port.  Close the port if it
    * was preopened.
    *========================================================================*/
   for (preo=preo_list; preo; preo = preo->next)
   {
      if (assoc->localip.s_addr == preo->localip.s_addr &&
		assoc->localport == preo->port &&
		assoc->tcp == preo->tcp)
         break;
   }
   if (preo)
   {
      if (preo->socket < 0)
      {
         syslog(LOG_ERR,"Error: An expected preopened socket number does not "
			"exist for %s IP = %s, port = %d.\n",
			preo->tcp?"TCP":"UDP",
			inet_ntoa(preo->localip),preo->port);
      }
      else
      {
         DBPRINTF(4,("Closing socket %d to make prebound port %d available.\n",
				preo->socket,preo->port));
         close(preo->socket);
         preo->socket = -1;
      }
   }

   if (assoc->tcp)
   {
      sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
      /*===================================================================
       * Set socket option to make the local address reuseable
       *==================================================================*/
      if (sock >= 0 &&
	  setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,(char *)&one,sizeof(int)) < 0)
	   syslog(LOG_ERR, "Error: setsockopt failed to make socket %d "
				"reuseable.  %m",sock);
   }
   else
      sock = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP);

   if (sock < 0)
   {
      syslog (LOG_ERR, "Error: Failed to open %s socket (port %d). %m.",
		assoc->tcp?"TCP":"UDP",assoc->localport);
      return (FALSE);			/* Probably out of sockets */
   }

   /*=========================================================================
    * If the send buffer size is less than our buffer size, make it equal
    * to our buffer size.
    * Note that we continue on even if we encounter a problem with getsockopt
    * or setsockopt.  Hopefully the current system buffer size is OK.
    *========================================================================*/
   lmaxb = sizeof(maxb);
   if (getsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char *)&maxb,
					(socklen_t *)&lmaxb) < 0)
      syslog (LOG_ERR, "Error: getsockopt for socket %d failed. %m", sock);

   if (maxb < MAXBUFLEN)
   {
      maxb = MAXBUFLEN;
      lmaxb = sizeof(maxb);
      if (setsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char *)&maxb, lmaxb) < 0)
         syslog (LOG_ERR, "Error: setsockopt for socket %d failed. %m", sock);
   }

   /*=========================================================================
    * If the recv buffer size is less than our buffer size, make it equal
    * to our buffer size.
    * Note that we continue on even if we encounter a problem with getsockopt
    * or setsockopt.  Hopefully the current system buffer size is OK.
    *========================================================================*/
   lmaxb = sizeof(maxb);
   if (getsockopt(sock, SOL_SOCKET, SO_RCVBUF, (char *)&maxb,
					(socklen_t *)&lmaxb) < 0)
      syslog (LOG_ERR, "Error: getsockopt for socket %d failed. %m", sock);

   if (maxb < MAXBUFLEN)
   {
      maxb = MAXBUFLEN;
      lmaxb = sizeof(maxb);
      if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF, (char *)&maxb, lmaxb) < 0)
         syslog (LOG_ERR, "Error: setsockopt for socket %d failed. %m", sock);
   }

   if (!assoc->tcp && ioctl(sock, FIONBIO, &one) < 0)
      syslog (LOG_ERR, "Error: ioctl FIONBIO failed for UDP socket %d. %m",
			sock);
   
   bzero (&myaddr, sizeof (myaddr));
   myaddr.sin_family = AF_INET;
   myaddr.sin_port = assoc->localport;
   myaddr.sin_addr = assoc->localip;

   /*=========================================================================
    * Bind the socket to the localport.
    *========================================================================*/
   if (assoc->localport == INADDR_ANY && assoc->priv)
   {
      myaddr.sin_port = htons(1023);
      while ((rc = bind(sock,(struct sockaddr *)&myaddr,sizeof(myaddr))) < 0 &&
             (errno == EADDRNOTAVAIL || errno == EADDRINUSE) &&
             ntohs(myaddr.sin_port) > 512)
         myaddr.sin_port = htons(ntohs(myaddr.sin_port) - 1);
   }
   else
         rc = bind (sock, (struct sockaddr *)&myaddr, sizeof (myaddr));

   /*=========================================================================
    * If the port was preopened, record the socket number in the preo_list.
    *========================================================================*/
   if (preo)
      preo->socket = sock;

   /*=========================================================================
    * Switch back to regular user if necessary
    *========================================================================*/
#ifdef __hppa
   if (rpcconf.uid != geteuid() && setresuid(-1,rpcconf.uid,-1) < 0)
#else
   if (rpcconf.uid != geteuid() && seteuid(rpcconf.uid) < 0)
#endif
      uiderror(rpcconf.uid);

   if (rc < 0)
   {
      syslog (LOG_ERR, "Error: bind failed for %s localip=%s, localport=%d, %m",
		assoc->tcp?"TCP":"UDP",
		inet_ntoa(myaddr.sin_addr),ntohs(myaddr.sin_port));
      close_and_preopen(sock);
      return (FALSE);
   }

   /*=========================================================================
    * If TCP, Make the connection now
    * If assoc is a fanout or if its partner's socket is not established,
    * we do a listen
    * otherwise, we do a connect
    *========================================================================*/
   if (assoc->tcp)
   {
      if (assoc->fanout || assoc->partner->socket < 0)
      {
         if (listen(sock, rpcconf.qlen) < 0)
         {
            syslog (LOG_ERR, "Error: listen failed for TCP port %d, %m",
				myaddr.sin_port);
            close_and_preopen(sock);
            return(FALSE);
         }
         DBPRINTF (2, ("listening to socket %d for %s 0x%x from %s "
			"port %d\n",sock,assoc->fanout ? "fanout":"assoc",
			assoc,
			inet_ntoa(myaddr.sin_addr),myaddr.sin_port));
      }
      else
      {
         if (connect(sock, (struct sockaddr *)&assoc->sin,
			sizeof(assoc->sin)) < 0)
         {
            syslog(LOG_ERR, "Error: connect failed for TCP remote_ip = %s "
				"port %d, %m, errno = %d",
				inet_ntoa(assoc->sin.sin_addr),
				assoc->sin.sin_port,errno);
            close_and_preopen(sock);
            return(FALSE);
         }
#ifndef __sun
         if (ioctl(sock, FIONBIO, &one) < 0)
            syslog (LOG_ERR, "Error: ioctl FIONBIO failed for TCP socket %d.  "
			"%m", sock);
#endif
         DBPRINTF (2, ("connected socket %d to TCP %s port %d\n",sock,
			inet_ntoa(assoc->sin.sin_addr),
			assoc->sin.sin_port));
      }
   }

   /*=========================================================================
    * We may be handling the RHS of a connection here.  On the RHS, the
    * localport member may be INADDR_ANY.  We want to have this assoc reflect
    * the bound port in the localport member so we do a getsockname here
    * after the bind and then set the localport value in the assoc.
    *========================================================================*/
   if (assoc->localport == INADDR_ANY)
   {
      len = sizeof (myaddr);
      if (getsockname (sock, (struct sockaddr *)&myaddr,
					(socklen_t *)&len) < 0)
      {
         syslog (LOG_ERR, "Error: getsockname for RHS remote name failed, %m");
         close_and_preopen(sock);
	 return (FALSE);
      }
      assoc->localport = myaddr.sin_port;
      DBPRINTF (1, ("opensocket: localport bound to port %d\n",
		  ntohs (assoc->localport)));
   }
   DBPRINTF (1, ("opened %s socket %d for assoc 0x%x\n",
			assoc->tcp?"TCP":"UDP", sock, assoc));
   assoc->socket = sock;
   assoc->ownport = TRUE;

   return (TRUE);
}
/*============================================================================
 *
 *===========================================================================*/
char *skipwhite (char *cp)
{
   while (*cp == ' ' || *cp == '\t')
      ++cp;
   return (cp);
}
/*============================================================================
 *
 *==========================================================================*/
char *findeos (cp)
	char		*cp;
{
	while (*cp != ' ' && *cp != '\t' && *cp != '\0')
		++cp;
	return (cp);
}
/*===========================================================================
 *
 *===========================================================================*/
void parsehostname (char **lscanp, Assoc *assoc, int lno)
{
   char	*lscan;
   char	*end;
   struct hostent	*host;

   lscan = *lscanp;
   DBPRINTF (2, ("parsehostname ('%s', 0x%x, %d)\n", lscan, assoc, lno));
   end = findeos (lscan);
   if (*end)
      *end++ = '\0';
   assoc->sin.sin_addr.s_addr = inet_addr (lscan);
   if (assoc->sin.sin_addr.s_addr == INADDR_NONE)
   {
      assoc->sin.sin_addr.s_addr = INADDR_ANY;
      if (strcmp(lscan,"*") != 0)
      {
         syslog(LOG_ERR, "Error: '%s' is specified in the IP field on line %d "
		"of %s, it must be a valid IP address.",lscan,lno,rpcconffile);
         exit(1);
      }
   }
   /*
    * Got an IP address, check for mask
    */
   lscan = skipwhite (end);
   end = findeos (lscan);
   if (end - lscan != 4 || strncmp (lscan, "mask", 4) != 0)
   {
      assoc->addrmask.s_addr = INADDR_NONE;
      DBPRINTF (2, ("parsehostname got addr=%s, no mask\n",
		  inet_ntoa (assoc->sin.sin_addr)));
      *lscanp = lscan;
      return;				/* No mask */
   }
   lscan = skipwhite (end);		/* Get mask value */
   end   = findeos (lscan);
   if (*end)
      *end++ = '\0';
   assoc->addrmask.s_addr = inet_addr (lscan);
   DBPRINTF (2, ("parsehostname got addr=%s, mask=%s\n",
	  inet_ntoa (assoc->sin.sin_addr), lscan));
   *lscanp = end;
   return;
}


/*============================================================================
 * Return a UDP portnumber pointed at by *lscanp, or INADDR_ANY if it
 * matches wildstr.  Adjusts lscanp to point after end of scanned string.
 *==========================================================================*/
u_short getportnum (char **lscanp, char *wildstr, int lno, int tcp)
{
   struct servent	*servent;
   char	*lscan;
   char	*end;
/*
   lscan = skipwhite (*lscanp);
   end   = findeos (lscan);
   *end++ = '\0';
*/
   lscan = strtok(*lscanp," \t-");
   DBPRINTF (2, ("getportnum ('%s', '%s', %d)\n", lscan, wildstr, lno));
   *lscanp = lscan + strlen(lscan) + 1;
   if (wildstr != NULL && strcmp (lscan, wildstr) == 0)
      return (INADDR_ANY);
   else if ((servent = getservbyname (lscan, tcp ? "tcp":"udp")) != NULL)
         return (servent->s_port);
   	else if (!isdigit (*lscan))
        {
	   syslog (LOG_ERR, "Error: Bad %s port number %s on line %d of %s",
			tcp ? "TCP":"UDP", lscan, lno, rpcconffile);
	   exit (1);
	}
   return (htons (atoi (lscan)));
}
/*===========================================================================
 * Scan a single hostname, remote-port, local-port triplet.  Return pointer
 * to end of string.
 * flag == 0 means parsing the srcip/srcport/localip/localport combo
 * flag == 1 means parsing the destip/destport/localport2 combo
 *==========================================================================*/
char *parsehost (char *lscan, Assoc *assoc, int lno, int flag)
{
   char *end;

   DBPRINTF (2, ("parsehost ('%s', 0x%x, %d)\n", lscan, assoc, lno));
   parsehostname (&lscan, assoc, lno);
   assoc->sin.sin_port = getportnum (&lscan, "*", lno, assoc->tcp);
   lscan = skipwhite(lscan);
   end   = findeos(lscan);
   *end++ = '\0';
   if (flag == 0 && strcmp(lscan,"to") != 0)
   {
	syslog(LOG_ERR, "Error: Missing 'to' keyword on line %d of %s",
			lno, rpcconffile);
	exit(1);
   }
   if (flag == 0 || strcmp(lscan,"using") == 0)
   {
	lscan = skipwhite(end);
	end   = findeos (lscan);
	*end++ = '\0';
	assoc->localip.s_addr = inet_addr (lscan);
	if (assoc->localip.s_addr == INADDR_NONE)
	{
	    syslog(LOG_ERR,"Error: Invalid%s local IP address '%s' specified "
			"on line %d of %s",
			flag?" \"using\"":"",lscan,lno,rpcconffile);
	    exit(1);
	}
	lscan = skipwhite (end);
   }
   assoc->localport = getportnum (&lscan, "*", lno, assoc->tcp);

   /*=========================================================================
    * See if port range is specified
    *========================================================================*/
   lscan = skipwhite(lscan);
   if (flag == 0 && (*lscan == '-' || isdigit(*lscan)))
   {
	if (!assoc->dynamic)
	{
	    syslog(LOG_ERR,"Error: Either 'tcprpc' or 'udprpc' is required "
			"to use a port range on line %d of %s.",
			lno, rpcconffile);
	    exit(1);
	}
	if (*lscan == '-')
	    lscan = skipwhite(lscan+1);
	assoc->highport = getportnum (&lscan, "*",lno, assoc->tcp);
	/*====================================================================
	 * Make sure the port range is valid
	 *===================================================================*/
	if (ntohs(assoc->highport) < ntohs(assoc->localport))
	{
	    syslog(LOG_ERR,"Error: The hiport value (%d) is less than the "
			"localport value (%d) on line %d of %s",
			ntohs(assoc->highport),ntohs(assoc->localport),
			lno,rpcconffile);
	    exit(1);
	}
   }
   DBPRINTF (2, (
	  "parsehost: host=%s, remport=%d, localport=%d, highport=%d\n",
	  inet_ntoa (assoc->sin.sin_addr),
	  ntohs (assoc->sin.sin_port), ntohs (assoc->localport),
	  ntohs (assoc->highport)));
   return skipwhite (lscan);
}
/*============================================================================
 * Parse a config file "permit" command line.  lscan points after the "permit".
 * Creates a generic assoc and partner, hooked into the appropriate fanouts.
 * Exits on error.
 *===========================================================================*/
Assoc *parsepermit (char *lscan, int lno)
{
   Assoc	*assoc;
   Assoc	*partner;
   Assoc	*fanout;
   Assoc	**scan;
   char		oneway=FALSE;
   char		priv=FALSE;
   char		*end;

   lscan = skipwhite (lscan);
   if ((assoc = getassoc ()) == NULL || (partner = getassoc ()) == NULL)
   {
      syslog (LOG_ERR, "Error: Out of memory, exiting");
      exit (1);
   }

   /*=========================================================================
    * Check for tcp permits
    *========================================================================*/
   end = findeos(lscan);
   *end++ = '\0';
   if (strcmp(lscan,"tcprpc") == 0)
   {
      fixed_ports_done = TRUE;
      assoc->tcp = TRUE;
      assoc->dynamic = TRUE;
      partner->tcp = TRUE;
      partner->dynamic = TRUE;
      lscan = skipwhite(end);
      end = findeos(lscan);
      *end++ = '\0';
   }
   else if (strcmp(lscan,"udprpc") == 0)
   {
      fixed_ports_done = TRUE;
      assoc->dynamic = TRUE;
      partner->dynamic = TRUE;
      lscan = skipwhite(end);
      end = findeos(lscan);
      *end++ = '\0';
   }
   else if (strcmp(lscan,"tftp") == 0)
   {
      if (fixed_ports_done)
      {
         syslog(LOG_ERR,"Error: \"permit tftp from\" on line %d of %s must be "
			"positioned before all \"permit tcprpc/udprpc\" lines",
			lno,rpcconffile);
         exit(1);
      }
      assoc->tftp = TRUE;
      partner->tftp = TRUE;
      lscan = skipwhite(end);
      end = findeos(lscan);
      *end++ = '\0';
   }
   else if (strcmp(lscan,"tcp") == 0)
   {
      if (fixed_ports_done)
      {
         syslog(LOG_ERR,"Error: \"permit tcp from\" on line %d of %s must be "
			"positioned before all \"permit tcprpc/udprpc\" lines",
			lno,rpcconffile);
         exit(1);
      }
      assoc->tcp = TRUE;
      partner->tcp = TRUE;
      lscan = skipwhite(end);
      end = findeos(lscan);
      *end++ = '\0';
   }
   else if (strcmp(lscan,"udp") == 0)
   {
      if (fixed_ports_done)
      {
         syslog(LOG_ERR,"Error: \"permit udp from\" on line %d of %s must be "
			"positioned before all \"permit tcprpc/udprpc\" lines",
			lno,rpcconffile);
         exit(1);
      }
      lscan = skipwhite(end);
      end = findeos(lscan);
      *end++ = '\0';
   }
   else if (fixed_ports_done)
   {
      syslog(LOG_ERR,"Error: \"permit from\" on line %d of %s must be "
			"positioned before all \"permit tcprpc/udprpc\" lines",
			lno,rpcconffile);
      exit(1);
   }

   if (strcmp(lscan,"from") != 0)
   {
      syslog(LOG_ERR,"Error: Missing 'from' keyword on line %d of %s",lno,
		rpcconffile);
      exit(1);
   }
   lscan = skipwhite(end);

   lscan = parsehost (lscan, assoc, lno, 0);
   if (assoc->localport == INADDR_ANY)
   {
      syslog (LOG_ERR, "Error: Missing localport on line %d of %s", lno,
			rpcconffile);
      exit (1);
   }

   end = findeos(lscan);
   *end++ = '\0';

   /*=========================================================================
    * Create a new fanout if one doesn't already exist
    *========================================================================*/
   if ((fanout = findfanout (assoc->localip,assoc->localport,
				assoc->highport,assoc->tcp))
                   == NULL)
   {
      DBPRINTF (2, ("Making new fanout for port %d\n",
     		  ntohs (assoc->localport)));
      if ((fanout = getassoc ()) == NULL)
      {
         syslog (LOG_ERR, "Error: Out of memory, exiting");
	 exit (1);
      }
      fanout->fanout = TRUE;
      fanout->tcp = assoc->tcp;
      fanout->localip = assoc->localip;
      fanout->localport = assoc->localport;
      fanout->highport = assoc->highport;
      fanout->dynamic = assoc->dynamic;
      fanout->tftp = assoc->tftp;
      if (!fanout->dynamic)
         if (!opensocket (fanout))
            exit (1);
   }
   else
   {
      if (assoc->highport && fanout->highport)
      {
         if (assoc->localport < fanout->localport)
            fanout->localport = assoc->localport;
         if (assoc->highport > fanout->highport)
            fanout->highport = assoc->highport;
      }
   }
   /*========================================================================
    * Append to end of generics list, so list gets searched in the
    * order in which things were specified in the config file
    *======================================================================*/
   for (scan = &fanout->generics; *scan; scan = &(*scan)->nextinst)
		;
   *scan = assoc;
   assoc->plex = fanout;
   assoc->receiver = TRUE;

   if (strcmp(lscan,"redirect") != 0)
   {
      syslog(LOG_ERR,"Error: Missing 'redirect' keyword on %d of %s",lno,
		rpcconffile);
   }
   lscan = skipwhite(end);
   /*=======================================================================
    * Same thing again, for the partner.  (Even though the partner can't
    * be wildcarded, we delay resolving the IP address until instance
    * creation time, so that if the partner's IP address is changed,
    * rpcproxy catches the change without needing to be restarted.)
    *=====================================================================*/
   assoc->partner = partner;
   lscan = parsehost (lscan, partner, lno, 1);
   partner->partner = assoc;

   if (partner->sin.sin_addr.s_addr == INADDR_ANY)
   {
	syslog(LOG_ERR, "Error: Invalid redirect host IP on line %d of %s.", 
			lno, rpcconffile);
	exit(1);
   }
   if (partner->sin.sin_port == INADDR_ANY &&
	assoc->highport == 0)
   {
	partner->sin.sin_port = assoc->localport;
   }
   if (partner->addrmask.s_addr != INADDR_ANY
	  && partner->addrmask.s_addr != INADDR_NONE)
   {
      syslog (LOG_ERR, "Error: Can't use mask on redirect host, line %d of %s.",
			 lno,rpcconffile);
      exit (1);
   }

   /*=======================================================================
    * Preopen the port if localport2 is specified
    *======================================================================*/
   if (partner->localport != INADDR_ANY)
	preopen(partner->localip,partner->localport,partner->tcp);

   lscan = skipwhite (lscan);
   while (*lscan)
   {
      end = findeos (lscan);
      if (*end != '\0')
         *end++ = '\0';
      else if (strcmp (lscan, "priv") == 0)
         priv = TRUE;
      else
      {
         syslog (LOG_ERR, "Error: Invalid parameter(s) at end of line %d of "
			"%s.", lno,rpcconffile);
	 exit (1);
      }
      lscan  = end;
   }

   /*========================================================================
    * Disallow inbound tftp permits
    *=====================================================================*/
   if (assoc->tftp && assoc->sin.sin_addr.s_addr == INADDR_ANY)
   {
      syslog(LOG_ERR, "Error: Line %d of %s - permit tftp from a wildcarded "
			"source IP address (*) is not supported.",
			lno,rpcconffile);
      exit(1);
   }
   if (assoc->tftp && !ipsecure(inet_ntoa(assoc->sin.sin_addr)))
   {
      syslog(LOG_ERR, "Error: Line %d of %s - permit tftp from an unprotected "
			"source IP address is not supported.",
			lno,rpcconffile);
      exit(1);
   }

   /*========================================================================
    * !!! If LHS is wildcarded and localport2 is specified, then what?
    * can only allow a single session at a time to use the relay (how
    * do we notify others that session is in use?  ICMP?), or can force
    * oneway
    *=====================================================================*/
/*
   if (partner->localport != INADDR_ANY)
   {
      if ((fanout = findfanout (partner->localip,partner->localport,FALSE))
                         == NULL)
      {
         DBPRINTF (2, ("Making new fanout for port %d\n",
			  ntohs (partner->localport)));
	 if((fanout = getassoc ()) == NULL)
         {
	    syslog (LOG_ERR, "Error: Out of memory, exiting");
	    exit (1);
	 }
         fanout->fanout = TRUE;
	 fanout->localport = partner->localport;
         if (!fanout->dynamic)
	    if (!opensocket (fanout))
	       exit (1);
      }
*/
      /*======================================================================
       * Point plex back to the fanout so instances of this generic
       * get the right plex pointer; however, we leave it off the
       * list of generics for the plex so that it doesn't get found
       * for uninstantiated connections.  (Sessions can be initiated
       * from right-to-left).
       *===================================================================*/
/*
      partner->plex = fanout;
   }
*/
   if (!oneway)
      partner->receiver = TRUE;
   partner->priv = priv;

}
/*===========================================================================
 *
 *==========================================================================*/
void parseencaps (char *lscan, int lno)
{
   Assoc	*assoc;
   Assoc	*encaps;

   lscan = skipwhite (lscan);
   if ((assoc = getassoc ()) == NULL)
   {
      syslog (LOG_ERR, "Error: Out of memory, exiting");
      exit (1);
   }
   parsehostname (&lscan, assoc, lno);
   assoc->localport = getportnum (&lscan, NULL, lno, 0);
   if (assoc->localport == INADDR_ANY)
   {
      syslog (LOG_ERR, "Error: Missing localport on line %d of %s",
			lno,rpcconffile);
      exit (1);
   }
   for (encaps = assoclist; encaps != NULL; encaps = encaps->next)
   {
      if (encaps->encaps && encaps->localport == assoc->localport)
         break;
   }
   if (encaps == NULL)
   {
      DBPRINTF (2, ("Making new encaps for port %d\n",
   		  ntohs (assoc->localport)));
      if ((encaps = getassoc ()) == NULL)
      {
         syslog (LOG_ERR, "Error: Out of memory, exiting");
	 exit (1);
      }
      encaps->encaps = TRUE;
      encaps->localport = assoc->localport;
      if (!opensocket (encaps))
         exit (1);
   }
   assoc->nextinst = encaps->generics;
   encaps->generics = assoc;
   assoc->plex = encaps;
   lscan = skipwhite (lscan);
   if (*lscan)
   {
      syslog (LOG_ERR, "Error: Too much stuff on line %d", lno);
      exit (1);
   }
   assoc->receiver = TRUE;
}
/*============================================================================
 * Read configuration file and build initial associations
 *==========================================================================*/
void readconfig ()
{
   FILE	*configfile;
   char	line [LINE_MAX];
   char	*lscan;
   char	*end;
   int	lno;

   configfile = fopen (rpcconffile, "r");
   if (configfile == NULL)
   {
      syslog (LOG_ERR, "Error: opening config file %s, %m", rpcconffile);
      exit (1);
   }

   rpcconf.pmapport = htons(PMAPPORT);

   lno = 0;
   while (fgets (line, sizeof (line), configfile) != NULL)
   {
      ++lno;
      if (strlen (line) >= sizeof (line) - 1)
      {
         syslog (LOG_ERR, "Error: Config file %s line %d too long",
			rpcconffile, lno);
	 exit (1);
      }
      lscan = strchr (line, '#');
      if (lscan != NULL)
         *lscan = '\0';
      else
         line [strlen (line) - 1] = '\0';	/* Nuke \n */
      lscan = skipwhite (line);
      if (*lscan == '\0')		/* Skip blank lines */
         continue;
      end = findeos (lscan);
      if (*end != '\0')
         *end++ = '\0';
      if (strcmp (lscan, "permit") == 0)
         parsepermit (end, lno);
/* We don't support encaps
      else if (strcmp (lscan, "encapsulate") == 0)
         parseencaps (end, lno);
*/
      else if (strcmp (lscan, "timeout") == 0)
      {
         lscan = strtok(end, " :=");
         if (lscan == NULL)
         {
            syslog(LOG_ERR, "Error: Missing value following timeout keyword in "
				"line %d of %s.", lno, rpcconffile);
            exit(1);
         }
         rpcconf.idletime = atoi(lscan);
      }
      else if (strcmp (lscan, "qlen") == 0)
      {
         lscan = strtok(end, " :=");
         if (lscan == NULL)
         {
            syslog(LOG_ERR, "Error: Missing value following qlen keyword in "
				"line %d of %s.", lno, rpcconffile);
            exit(1);
         }
         rpcconf.qlen = atoi(lscan);
      }
      else if (strcmp (lscan, "pmapcmd") == 0)
      {
         lscan = strtok(end, " :=");
         if (lscan == NULL)
         {
            syslog(LOG_ERR, "Error: Missing value following pmapcmd keyword in "
				"line %d of %s.", lno, rpcconffile);
            exit(1);
         }
         if (strcmp(lscan,"nofilter") == 0)
            rpcconf.filter = FALSE;
         else
         {
            syslog(LOG_ERR, "Error: Value (%s) following pmapcmd keyword is "
				"not valid.", lscan);
            exit(1);
         }
      }
      else if (strcmp (lscan, "groupid") == 0)
      {
         struct group *grp;

         lscan = strtok(end, " :=");
         if (lscan == NULL)
         {
            syslog(LOG_ERR, "Error: Missing value following groupid keyword in "
				"line %d of %s.", lno, rpcconffile);
            exit(1);
         }
         if ((grp = getgrnam(lscan)) == NULL)
         {
            syslog(LOG_ERR, "Error: Line %d of %s - Could not get group id for "
			"%s, %m.",lno,rpcconffile,lscan);
            exit(1);
         }
         rpcconf.gid = grp->gr_gid;

         /*==================================================================
          * Lose root group now
          *=================================================================*/
         if (rpcconf.gid != getegid() && setgid (rpcconf.gid) < 0)
         {
            giderror(rpcconf.gid);
            exit(1);
         }
      }
      else if (strcmp (lscan, "userid") == 0)
      {
         struct passwd *pw;
         lscan = strtok(end, " :=");
         if (lscan == NULL)
         {
            syslog(LOG_ERR, "Error: Missing value following userid keyword in "
				"line %d of %s.", lno, rpcconffile);
            exit(1);
         }
         if ((pw = getpwnam(lscan)) == NULL)
         {
            syslog(LOG_ERR, "Error: Line %d of %s - Could not get user id for "
			"%s, %m.",lno,rpcconffile,lscan);
            exit(1);
         }
         rpcconf.uid = pw->pw_uid;
      }
#ifdef Debug
      else if (strcmp (lscan, "trace") == 0)
      {
         lscan = strtok(end, " :=");
         if (lscan == NULL)
         {
            syslog(LOG_ERR, "Error: Missing value following trace keyword in "
				"line %d of %s.", lno, rpcconffile);
            exit(1);
         }
         rpcconf.traceflag = atoi(lscan);
      }
#endif
      else
      {
         syslog (LOG_ERR, "Error: Bad config file command on line %d of %s.",
			lno, rpcconffile);
	 exit (1);
      }
   }
   fclose (configfile);

   /*==========================================================================
    * Reset lastassoc so that new assoc will be threaded at the top
    *========================================================================*/
   lastassoc = NULL;

   get_active_rpc_ports();

dumpallassoc();
}
/*=============================================================================
 * Return TRUE if the specified sockaddr matches the wildcarded assoc; must
 * match both the address/mask in the assoc, the possibly wildcarded
 * hostname, and the port number.
 *===========================================================================*/
int validate (Assoc *assoc, SockIn *remote, SockIn *local)
{
   struct hostent	*hp;
   char		namebuf [512];
   char		th1[16];
   char		*cp;
   int		mismatch, i;
   u_int	nblen;

   strcpy(th1, inet_ntoa(local->sin_addr));
   DBPRINTF (2, ("validate (assoc=0x%x, remote=%s/port=%d) local=%s/port=%d\n",
          assoc, inet_ntoa (remote->sin_addr), ntohs (remote->sin_port),
          th1, ntohs (local->sin_port)));
   /*==========================================================================
    * Check IP address and port number
    *========================================================================*/
   if ((assoc->sin.sin_addr.s_addr != INADDR_ANY) &&
	(remote->sin_addr.s_addr & assoc->addrmask.s_addr)
	  != (assoc->sin.sin_addr.s_addr & assoc->addrmask.s_addr) )
   {
      DBPRINTF (2, ("Failed mask test, remote=%s, mask=0x%08x\n",
		  inet_ntoa (assoc->sin.sin_addr),
		  ntohl (assoc->addrmask.s_addr)));
      return FALSE;
   }

   if (assoc->sin.sin_port != INADDR_ANY &&
       assoc->sin.sin_port != remote->sin_port)
   {
      DBPRINTF (2, ("Mismatched ports, assoc.port=%d, have %d\n",
		  ntohs (assoc->sin.sin_port), ntohs (remote->sin_port)));
      return FALSE;
   }

   if (assoc->localip.s_addr != INADDR_ANY &&
       local->sin_addr.s_addr != INADDR_ANY &&
       assoc->localip.s_addr != local->sin_addr.s_addr)
   {
      DBPRINTF (2, ("Mismatched local ip address, assoc.localip=%s, "
		"local.sin_addr=%s\n",inet_ntoa(assoc->localip),th1));
      return FALSE;
   }

   if (assoc->localport != local->sin_port &&
       (assoc->highport == INADDR_ANY ||
        ntohs(assoc->localport) > ntohs(local->sin_port) ||
        ntohs(assoc->highport) < ntohs(local->sin_port)))
   {
      DBPRINTF (2, ("Mismatched local port, assoc->localport=%d, "
		"local->sin_port=%d\n",ntohs(assoc->localport),
		ntohs(local->sin_port)));
      return FALSE;
   }

   return TRUE;
}
/*===========================================================================
 *
 * inputs:
 * 	Assoc	- Fanout/encaps assoc
 *
 *=========================================================================*/
Assoc *makeinstance (Assoc *assoc, SockIn *fromaddr)
{
   Assoc	*instance;
   Assoc	*partner;

   DBPRINTF (1, ("makeinstance (0x%x) plex=0x%x\n", assoc, assoc->plex));
   if ((instance = getassoc ()) == NULL)
   {
      syslog (LOG_ERR, "Error: Out of memory");
      return (NULL);
   }
   if ((partner = getassoc ()) == NULL)
   {
      deleteassoc (instance);
      syslog (LOG_ERR, "Error: Out of memory");
      return (NULL);
   }
   DBPRINTF (1, ("creating instance 0x%x, partner 0x%x\n",
	  instance, partner));
   instance->partner   = partner;
   instance->sin       = *fromaddr;
   instance->plex      = assoc->plex;
   if (!assoc->tcp)
      instance->socket    = assoc->plex->socket;	/* clone socket if udp*/
   instance->priv      = assoc->priv;
   instance->dynamic   = assoc->dynamic;
   instance->inst      = TRUE;
   instance->receiver  = TRUE;
   instance->transient = TRUE;
   instance->tftp      = assoc->tftp;
   partner->partner    = instance;
   partner->inst       = TRUE;
   partner->tftp       = assoc->tftp;
   instance->faccess   = time(NULL);
   /*
    * Thread onto list of instances for this fanout/encaps
    */
   instance->nextinst  = assoc->plex->instances;
   assoc->plex->instances = instance;
   DBPRINTF (1, ("makeinstance: done\n"));
   return (instance);
}
/*============================================================================
 * Send a packet out to the outgoing half of an assoc pair.  pkt points to
 * the packet, but *must* have space available before it, in its buffer,
 * to hold an encapsulating wrapper if necessary.
 *==========================================================================*/
void sendtopartner(Assoc *assoc, char *pkt, int pktlen, SockIn *fromaddr,
                   u_short rcvport)
{
   char		justopened=0;

   /*
    * If socket isn't open yet, open it
    */
   if (assoc->socket < 0)
   {
      if (opensocket (assoc))
      {
         if (assoc->receiver)
	    makefdlist ();	/* Rebuild socket list */
      }
      else
      {
         syslog (LOG_WARNING, "Packet dropped because of opensocket problems");
	 return;
      }
      justopened = 1;
   }

   if (pktlen)
   {
      if (assoc->tcp)
      {
         int len;

         while ((len = write(assoc->socket, pkt, pktlen)) != pktlen)
         {
            if (len < 0 && (errno != EAGAIN && errno != EWOULDBLOCK))
            {
               syslog (LOG_ERR, "Error: write failed, %m, errno = %d.  "
			"dest_ip = %s, dest_port = %d, local_port = %d.",
			errno,
			inet_ntoa(assoc->sin.sin_addr),
			ntohs(assoc->sin.sin_port),
			ntohs(assoc->localport));
               return;
            }
            else 
            {
               if (len < 0)
                  len = 0;
               DBPRINTF(2,("Repeated attempt to write %d of %d bytes\n",
			pktlen-len,pktlen));
               pktlen = pktlen - len;
            }
          }
      }
      else
      {
         if (sendto (assoc->socket, pkt, pktlen, 0,
             (struct sockaddr *)&assoc->sin, sizeof (SockIn)) != pktlen)
         {
            syslog (LOG_ERR, "Error: sendto failed, %m.  dest_ip = %s, "
			"dest_port "
			"= %d, local_port=%d.",inet_ntoa(assoc->sin.sin_addr),
			ntohs(assoc->sin.sin_port),ntohs(assoc->localport));
            return;
         }
      }
      DBPRINTF (2, (
	  "packet sent to assoc=0x%x socket=%d addr=%s %s port=%d localport="
          "%d\n",
	  assoc, assoc->socket, inet_ntoa (assoc->sin.sin_addr),
          assoc->tcp?"TCP":"UDP",
	  ntohs (assoc->sin.sin_port), ntohs (assoc->localport)));
   }

   /*=======================================================================
    * Special handling for tftp permits
    * If the socket was just opened, assoc is on the LHS of the connection
    * and we just sent an initial request to the redirect host's tftp port.
    * tftp will reply to my local port but from an ephemeral port and not
    * from the wellknown tftp port.  So we close the socket to the server's
    * tftp port and then bind my local port to listen to (unfortunately) all
    * ports from the server.
    * We first do a getsockname so we can get the local ip address that was
    * used to send the request - just to narrow the hole a little bit.
    *======================================================================*/
   if (assoc->tftp && justopened)
   {
      SockIn		myaddr;
      int		len;

      len = sizeof(myaddr);
      if (getsockname(assoc->socket, (struct sockaddr *)&myaddr,
						(socklen_t *)&len) < 0)
      {
         syslog (LOG_ERR, "Error: getsockname failed for tftp reply, %m");
         close_and_preopen(assoc->socket);
         assoc->socket = -1;
         return;
      }

      close_and_preopen(assoc->socket);

      assoc->socket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
      if (assoc->socket < 0)
      {
         syslog(LOG_ERR, "Error: Failed to open UDP socket to receive tftp "
			"replies.  %m.");
         return;
      }
      if (bind(assoc->socket,(struct sockaddr *)&myaddr, sizeof(myaddr)) < 0)
      {
         syslog(LOG_ERR, "Error: Failed to re-bind localport %d for tftp "
			"replies.  %m",ntohs(myaddr.sin_port));
         close_and_preopen(assoc->socket);
         assoc->socket = -1;
         return;
      }
      makefdlist();
   }
}

/*==========================================================================
 * Handles requests to a portmapper
 * If packet is a request to the portmapper to get the port number for an
 * application, this function saves the request id, the from ip and the
 * protocol (tcp or udp).
 * Note: This function assumes that the last 4 (4-byte) words in the packet
 * are the arguments to the portmapper for PMAPPROC_GETPORT.  It is a kludge
 * but should work since all OS's send the same length packet for getport.
 * I just couldn't find a correct mapping for the request packet.
 *==========================================================================*/
void do_pmap_request(SockIn *remote, struct rpc_msg *msg, int len)
{
   PMAPREQ	*newpmapreq;
   struct pmap	*pm;

   /*=======================================================================
    * If this is not a CALL (request), or
    * If program requested is not the portmapper, or
    * If procedure requested is not PMAPPROC_GETPORT
    * then there is nothing to be done
    *=======================================================================*/
   if (msg->rm_direction != htonl(CALL) ||
       msg->ru.RM_cmb.cb_prog != htonl(PMAPPROG) ||
       msg->ru.RM_cmb.cb_proc != htonl(PMAPPROC_GETPORT))
      return;

   if ((newpmapreq = malloc(sizeof(PMAPREQ))) == 0)
   {
	syslog(LOG_ERR,"Error: Out of memory for pmapreq");
	exit(1);
   }

   pm = (struct pmap *)((char *)msg + len - 4*sizeof(unsigned long));
   /*=======================================================================
    * Save the info from this request
    *=======================================================================*/
   newpmapreq->xid = msg->rm_xid;
   newpmapreq->clnt_ip = remote->sin_addr;
   newpmapreq->tcp = pm->pm_prot==htonl(IPPROTO_TCP);

   DBPRINTF(2, ("saved portmap request: xid=0x%x, client_ip=%s, "
		"protocol=%s\n",
		ntohl(newpmapreq->xid),inet_ntoa(newpmapreq->clnt_ip),
		newpmapreq->tcp?"TCP":"UDP"));

   /*=======================================================================
    * Add new request to top of the list
    *=======================================================================*/
   newpmapreq->next = pmapreq;
   pmapreq = newpmapreq;

   return;
}

/*==========================================================================
 * make a fanout for a dynamic RPC port
 * Returns FALSE if access denied, TRUE otherwise
 *=========================================================================*/
int makedynamic(Assoc *rassoc, SockIn *remote, PMAPREQ *pmr, unsigned long port)
{
   char		th1[16];
   Assoc	*ofanout,*fanout,*generic,*assoc,*partner;

   /*========================================================================
    * We use rassoc->localip for findfanout here - rassoc points to the LHS
    * of the permit for portmapper port.  Since we don't have the real RPC
    * request yet, we will have to search base on the portmapper request.
    * Note that a response with port=0 from the portmapper means the appl
    * is not registered.  The problem here is that there will most likely
    * not be a permit for port 0 and we do want the info to get back to the
    * client, so we always pass port=0 back to the requester.  Given that
    * the requester is permitted to the portmapper port already, this should
    * not be a big exposure.
    *======================================================================*/
   if ((fanout = findfanout (rassoc->localip,port,0,pmr->tcp)) == NULL)
   {
	if (port != 0)
	{
	    strcpy(th1,inet_ntoa(rassoc->sin.sin_addr));
	    syslog(LOG_ALERT,"Security Alert: PMAP_GETPORT reply not "
		"forwarded, src host %s is not permitted to "
		"access local IP %s at %s port %d",th1,
		inet_ntoa(rassoc->localip),
		pmr->tcp?"TCP":"UDP",ntohs(port));
	    return(FALSE);
	}
	else
	    return(TRUE);
   }

   if ((assoc = getassoc ()) == NULL || (partner = getassoc ()) == NULL)
   {
	syslog (LOG_ERR, "Error: Out of memory, exiting");
	exit (1);
   }

   /*========================================================================
    * We have found a matching fanout.  If a socket has not been opened,
    * create a new fanout (threaded on top of assoclist) and open a socket
    * for the new fanout.  Mark it transient so it expires.
    *======================================================================*/
   if (fanout->socket < 0)
   {
	ofanout = fanout;
	DBPRINTF (2, ("Making new fanout for dynamic %s port %d\n",
			pmr->tcp?"TCP":"UDP",ntohs (port)));
	if ((fanout = getassoc ()) == NULL)
	{
	    syslog (LOG_ERR, "Error: Out of memory, exiting");
	    exit (1);
	}

	/*=====================================================================
	 * Clone some of the info from the original fanout
	 *====================================================================*/
	fanout->localip.s_addr    = ofanout->localip.s_addr;
	fanout->generics   = ofanout->generics;

	fanout->fanout     = TRUE;
	fanout->localport  = port;
	fanout->tcp        = pmr->tcp;
	if (!opensocket (fanout))
	    exit (1);
	fanout->dynamic    = TRUE;
	fanout->transient  = TRUE;
	fanout->lastaccess = time(NULL);
   }


   /*========================================================================
    * Find the generic that will permit this access
    *=======================================================================*/
   for (generic = fanout->generics; generic; generic = generic->nextinst)
	if (generic->tcp == pmr->tcp &&
	    (generic->sin.sin_addr.s_addr == INADDR_ANY ||
	     (generic->sin.sin_addr.s_addr & generic->addrmask.s_addr) ==
			(pmr->clnt_ip.s_addr & generic->addrmask.s_addr)) &&
	    (generic->localip.s_addr == rassoc->localip.s_addr ||
	     generic->localip.s_addr == INADDR_ANY) &&
	    (generic->localport == port ||
	     (generic->highport &&
	      ntohs(generic->highport) >= ntohs(port) &&
	      ntohs(generic->localport) <= ntohs(port))))
	    break;

   if (generic == NULL)
   {
	strcpy(th1,inet_ntoa(rassoc->sin.sin_addr));
	syslog(LOG_ALERT,"Security Alert: PMAP_GETPORT reply not forwarded, "
		"src host %s is not permitted to access local IP "
		"%s at %s port %d",th1, inet_ntoa(rassoc->localip),
		pmr->tcp?"TCP":"UDP",ntohs(port));
	return(FALSE);
   }	

   /*========================================================================
    * thread dynamic generics at the top
    *=======================================================================*/
   assoc->nextinst = fanout->generics;
   fanout->generics = assoc;

   assoc->partner = partner;
   assoc->plex = fanout;
   assoc->sin.sin_addr.s_addr = rassoc->sin.sin_addr.s_addr;
   assoc->sin.sin_port = INADDR_ANY;
   assoc->addrmask.s_addr = INADDR_NONE;
   assoc->localip.s_addr = rassoc->localip.s_addr;
   assoc->localport = port;
   assoc->receiver = TRUE;
   assoc->dynamic = TRUE;
   assoc->transient = TRUE;
   assoc->lastaccess = time(NULL);

   DBPRINTF (2, ("Made dynamic assoc 0x%x for local %s port %d\n",assoc,
			pmr->tcp?"TCP":"UDP",ntohs(assoc->localport)));

   partner->partner = assoc;
   partner->sin.sin_addr.s_addr = rassoc->partner->sin.sin_addr.s_addr;
   partner->sin.sin_port = port;
   partner->addrmask.s_addr = INADDR_NONE;
   partner->localip = generic->partner->localip;
   partner->localport = generic->partner->localport;
   partner->receiver = TRUE;
   partner->priv = generic->partner->priv;
   partner->dynamic = TRUE;
   partner->transient = TRUE;
   partner->lastaccess = assoc->lastaccess;

   DBPRINTF (2, ("Made dynamic partner 0x%x for remote %s port %d\n",partner,
			pmr->tcp?"TCP":"UDP",ntohs(partner->sin.sin_port)));

   return(TRUE);
}

/*==========================================================================
 * Handles replies from a portmapper
 * Note: This function assumes that the last 4-byte word in the packet
 * is the return value from the portmapper for PMAPPROC_GETPORT.  It is a kludge
 * but should work since all OS's reply with the same length packet for getport.
 * I just couldn't find a correct mapping for the reply packet.
 * Returns FALSE if access is denied, TRUE otherwise
 *=========================================================================*/
int do_pmap_reply(Assoc *assoc, SockIn *remote, struct rpc_msg *msg, int len)
{
   PMAPREQ	*pmr;
   PMAPREQ	*prev=NULL;
   Assoc	*partner,*instance;
   unsigned short	port;
   int		rc=TRUE;

   /*=======================================================================
    * Do nothing if this is not an accepted and successful reply
    * We want to return TRUE since we want the portmap response to reach
    * the requesting host
    * Check packet length?
    *======================================================================*/
   if (msg->rm_direction != htonl(REPLY) ||
       msg->ru.RM_rmb.rp_stat != htonl(MSG_ACCEPTED) )
       /* msg->ru.RM_rmb.ru.RP_ar.ar_stat != htonl(SUCCESS)) */
	return(TRUE);

   /*=======================================================================
    * If the poll list is full, we can't do anything with opening another
    * port
    *======================================================================*/
   if (fdlistlen >= nfilemax)
   {
	syslog(LOG_ERR,"Error: Already have MaxSockets=%d files opened, "
			"cannot handle any new sockets.",nfilemax);
	return(TRUE);
   }

   /*==================================================================
    * Get the returned port number
    *=================================================================*/
   memcpy(&port,(char *)msg + len - sizeof(unsigned short),
			sizeof(unsigned short));

   /*=======================================================================
    * Match up the reply with a request we saved earlier
    *======================================================================*/
   for (pmr=pmapreq; pmr; prev=pmr,pmr=pmr->next)
   {
	if (pmr->xid != msg->rm_xid ||
	    pmr->clnt_ip.s_addr != assoc->partner->sin.sin_addr.s_addr)
	    continue;

	DBPRINTF(2, ("portmap reply port=%d to remote IP %s, xid=0x%x, "
			"protocol=%s\n",ntohs(port),inet_ntoa(pmr->clnt_ip),
			ntohl(msg->rm_xid),pmr->tcp?"TCP":"UDP"));

	if (port)
	    rc = makedynamic(assoc->partner,remote,pmr,port);

	/*==================================================================
	 * Remove the request from our saved list
	 *=================================================================*/
	if (prev == NULL)
	    pmapreq = pmr->next;
	else
	    prev->next = pmr->next;
	free(pmr);

	/*==================================================================
	 * Add instance to the fdlist to poll
	 *=================================================================*/
	if (port)
	    makefdlist();

	return(rc);
   }

   /*=======================================================================
    * Couldn't find a match, it must be a reply to a portmapper request
    * other than a PMAPPROC_GETPORT
    *======================================================================*/
   return(TRUE);
}

/*==========================================================================
 * Provides special handling for packets coming from or to the portmap port.
 * First it filters out questionable requests/replies to or from a portmapper.
 * Returns FALSE if access is denied, TRUE otherwise
 *==========================================================================*/
int dynamic_port(Assoc *assoc, SockIn *remote, struct rpc_msg *msg, int len)
{
   int		rc=TRUE;

   /*=======================================================================
    * If the packet is not from or to a portmapper port, nothing to do here.
    *=======================================================================*/
   if (assoc->localport != rpcconf.pmapport &&
		remote->sin_port != rpcconf.pmapport)
	return(TRUE);

   /*=======================================================================
    * Filter out questionable requests and replies.
    *=======================================================================*/
   if (rpcconf.filter && msg->ru.RM_cmb.cb_prog == htonl(PMAPPROG) &&
		(msg->ru.RM_cmb.cb_proc == htonl(PMAPPROC_SET) ||
		 msg->ru.RM_cmb.cb_proc == htonl(PMAPPROC_UNSET) ||
		 msg->ru.RM_cmb.cb_proc == htonl(PMAPPROC_CALLIT)))
   {
	char	*procset="PMAPPROC_SET";
	char	*procunset="PMAPPROC_UNSET";
	char	*proccallit="PMAPPROC_CALLIT";
	char	*procname;
	char	th1[16];
	if (msg->ru.RM_cmb.cb_proc == htonl(PMAPPROC_SET))
	    procname = procset;
	else if (msg->ru.RM_cmb.cb_proc == htonl(PMAPPROC_UNSET))
	    procname = procunset;
	else if (msg->ru.RM_cmb.cb_proc == htonl(PMAPPROC_CALLIT))
	    procname = proccallit;
	strcpy(th1,inet_ntoa(remote->sin_addr));
	syslog(LOG_ALERT, "Security Alert: src_ip = %s src_port = %d attempts "
			"to send unallowed packet to local_ip = %s "
			"local_port = %d for the portmapper procedure %s",
			th1,ntohs(remote->sin_port),
			inet_ntoa(assoc->localip),ntohs(assoc->localport),
			procname);
	return(FALSE);
   }

   /*=======================================================================
    * If local port is the PMAP port, assume this is a request to the
    * portmapper
    *=======================================================================*/
   if (assoc->localport == rpcconf.pmapport)
      do_pmap_request(remote, msg, len);
   else
      /*====================================================================
       * Foreign port is the PMAP port, assume this is a reply to an
       * earlier portmapper request to PMAPPROC_GETPORT
       *====================================================================*/
      rc = do_pmap_reply(assoc,remote, msg, len);

   return(rc);
}

/*==========================================================================
 *
 *
 *==========================================================================*/
void fwdpacketfrom (Assoc *assoc)
{
   Assoc	*instance;
   Assoc	*partner;
   Assoc	*fano;
   Assoc	*generic=NULL;
   int		maxpktlen;
   int		pktlen,mpktlen;
   int		newsock;
   int		fromaddrlen;
   SockIn	fromaddr;
   char		*pkt,*mpkt;
   int		len;
   SockIn	myaddr;

   DBPRINTF (2, ("Reading from assoc 0x%x, %s socket %d\n", assoc,
	  assoc->tcp?"TCP":"UDP",assoc->socket));
   fromaddrlen = sizeof (fromaddr);
   /*=======================================================================
    * Get packet in buffer, with room before it to encapsulate it if
    * necessary
    *=======================================================================*/
   pkt = pktbuf;
   maxpktlen = pktbuflen;

   if (assoc->tcp)
   {
      int	on=1;

      newsock = assoc->socket;
      /*====================================================================
       * If assoc is a fanout, we have not accepted this connection yet
       *===================================================================*/
      if (assoc->fanout)
      {
         /*=================================================================
          * Accept the new TCP connection
          *================================================================*/
         int off=0;
         len = sizeof(fromaddr);
         if ((newsock = accept(assoc->socket,(struct sockaddr *)&fromaddr,
				(addrlen_t *)&len)) < 0)
         {
            syslog (LOG_ERR, "Error: accept failed, %m");
            return;
         }
#ifndef __sun
         if (ioctl(newsock, FIONBIO, &on) < 0)
            syslog (LOG_ERR, "Error: ioctl FIONBIO failed for TCP socket %d.  ",
			newsock);
#endif
         DBPRINTF(2, ("accepted TCP socket %d\n",newsock));
         pktlen = 0;
      }
      else
      {

         /*====================================================================
          * Get peer's address
          *===================================================================*/
         len = sizeof(fromaddr);
         if (getpeername(assoc->socket, (struct sockaddr *)&fromaddr,
							(socklen_t *)&len))
            syslog(LOG_ERR,"Error: failed in getpeername.  %m");

         /*====================================================================
          * Read the tcp data
          *===================================================================*/
         if ((pktlen = read(assoc->socket, pkt, maxpktlen)) < 0)
         {
            if (errno == EWOULDBLOCK || errno == EAGAIN)
            {
               syslog(LOG_NOTICE,"read() returned EWOULDBLOCK or EAGAIN");
               DBPRINTF(2,("read() returned EWOULDBLOCK or EAGAIN"));
               return;
            }
            syslog (LOG_ERR, "Error: read failed socket %d src_ip = %s"
			"src_port = %d, pktlen=%d, %m",
			assoc->socket,
			inet_ntoa(fromaddr.sin_addr),ntohs(fromaddr.sin_port),
			pktlen);
            close_and_preopen(newsock);
            return;
         }

         if (pktlen == 0 && !assoc->fanout)
         {
            DBPRINTF(2, ("Zero length packet read for TCP assoc 0x%8x.  "
			"Closing socket %d\n",assoc,assoc->socket));
            close_and_preopen(assoc->socket);
            if (assoc->partner->socket != -1)
            {
               DBPRINTF(2, ("Closing socket %d\n",assoc->partner->socket));
               close_and_preopen(assoc->partner->socket);
            }
            if (assoc->partner)
               deleteassoc(assoc->partner);
            deleteassoc(assoc);
            makefdlist();
            return;
         }
      }
   }
   else
   {
      if ((pktlen = recvfrom (assoc->socket, pkt, maxpktlen, 0,
	    (struct sockaddr *)&fromaddr, (addrlen_t *)&fromaddrlen)) <= 0)
      {
         syslog (LOG_ERR, "Error: recvfrom failed on socket %d src_ip = %s "
			"src_port = %d pktlen=%d, %m",
			assoc->socket,
			inet_ntoa(fromaddr.sin_addr),ntohs(fromaddr.sin_port),
			pktlen);
         return;
      }
   }
   DBPRINTF (2, ("pkt rcvd from addr=%s, %s port=%d, pktlen=%d\n",
	  inet_ntoa (fromaddr.sin_addr), assoc->tcp?"TCP":"UDP",
          ntohs (fromaddr.sin_port), pktlen));
   if (pktlen > maxpktlen)
   {
      syslog (LOG_WARNING, "Dropped pkt (pktlen=%d, maxpktlen=%d)",
      		  pktlen, maxpktlen);
      (void) makepktbuf (pktlen);	/* Boost pkt buffer size */
      return;				/* Drop this packet */
   }

   /*=======================================================================
    * Do special processing if packet is to/from portmapper.
    * If access was denied, return right away, don't even send the reply.
    * If packet is tcp, there are 4 extra bytes in the packet before the
    * portmapper info.  We just have to skip those bytes.
    *=======================================================================*/
   if (assoc->tcp)
   {
      mpkt = pkt + 4;
      mpktlen = pktlen - 4;
   }
   else
   {
      mpkt = pkt;
      mpktlen = pktlen;
   }
   if (!dynamic_port(assoc,&fromaddr,(struct rpc_msg *)mpkt,mpktlen))
      return;

   if (assoc->fanout)
   {		/* Wildcarded source? */
      char	th1[16];

      len = sizeof(myaddr);
      if (getsockname(assoc->socket, (struct sockaddr *)&myaddr,
						(socklen_t *)&len) < 0)
      {
         syslog(LOG_ERR,"Error: getsockname error: %m");
         return;
      }

      strcpy(th1,inet_ntoa(fromaddr.sin_addr));
      /*====================================================================
       * Find only non fanouts that match
       *====================================================================*/
      for (generic = assoc->generics; generic != NULL;
	     		  generic = generic->nextinst)
      {
         if (validate (generic, &fromaddr, &myaddr) && !generic->fanout)
             break;
      }
      if (generic == NULL)
      {
          char th1[16];
          strcpy(th1,inet_ntoa(myaddr.sin_addr));
          syslog(LOG_ALERT,
	           "Security Alert: %s pkt from unapproved source src_host %s "
				"src_port %d to local_IP %s local_port %d",
				  assoc->tcp?"TCP":"UDP",
				  inet_ntoa (fromaddr.sin_addr),
				  ntohs(fromaddr.sin_port),
            			  th1, ntohs (assoc->localport));
          return;
       }
       if (!generic->dynamic && !generic->receiver)
       {
          syslog (LOG_ALERT,
		"Security Alert: Pkt rcvd at wrong end of 1-way link: %s/%d",
				  inet_ntoa (fromaddr.sin_addr),
				  ntohs (assoc->localport));
          return;
       }

      for (instance = assoc->instances; instance != NULL;
		  instance = instance->nextinst)
      {
         if (instance->tcp == generic->tcp &&
		instance->sin.sin_addr.s_addr == fromaddr.sin_addr.s_addr &&
		instance->sin.sin_port == fromaddr.sin_port &&
		(instance->localip.s_addr == myaddr.sin_addr.s_addr ||
		 myaddr.sin_addr.s_addr == INADDR_ANY) &&
		(instance->localport == myaddr.sin_port ||
		 instance->localport == INADDR_ANY ||
		 myaddr.sin_port == INADDR_ANY))
         {
            DBPRINTF(4, ("found existing instance 0x%08x\n",instance));
	    break;
         }
      }
      if (instance == NULL)
      {

	 instance = makeinstance (generic, &fromaddr);
	 if (instance == NULL)
	    return;
	 instance->localip.s_addr   = myaddr.sin_addr.s_addr;
	 instance->localport = myaddr.sin_port;
	 partner            = instance->partner;
	 partner->sin       = generic->partner->sin;
	 partner->localip   = generic->partner->localip;
	 partner->localport = generic->partner->localport;
	 partner->receiver  = generic->partner->receiver;
	 partner->priv      = generic->partner->priv;
	 partner->dynamic   = generic->partner->dynamic;
         if (assoc->tcp)
         {
            instance->socket   = newsock;
            instance->ownport  = TRUE;
            instance->receiver = TRUE;
            instance->tcp      = TRUE;
            partner->tcp       = TRUE;
            makefdlist();
         }

         /*==================================================================
          * if partner's remote port was wildcarded, fix it up
          *=================================================================*/
         if (partner->sin.sin_port == INADDR_ANY)
            partner->sin.sin_port = myaddr.sin_port;

	 if (partner->sin.sin_addr.s_addr == INADDR_ANY)
         {
            {
	       DBPRINTF (2, ("Send generic\n"));
               syslog (LOG_WARNING, "Assoc 0x%x - sendto address cannot "
			"be generic", partner);
					deleteassoc (instance);
					deleteassoc (partner);
	       return;
	    }
         }
#if LogInstances
         syslog (LOG_INFO, "creating new instance from %s/%d",
			  inet_ntoa (fromaddr.sin_addr),
			  ntohs (fromaddr.sin_port));
	 syslog (LOG_INFO, "... to %s/%d (localport %d)",
			  inet_ntoa (partner->sin.sin_addr),
			  ntohs (partner->sin.sin_port),
			  ntohs (partner->localport));
#endif
	 if (generic->partner->plex != NULL)
         {
	    /*
	     * Port is slaved to wildcard - thread on list
	     * of ports under the fanout.
	     */
	    partner->plex            = generic->partner->plex;
	    partner->nextinst        = partner->plex->instances;
	    partner->plex->instances = partner;
            partner->socket          = partner->plex->socket;
         }

         /*==================================================================
          * Open the partner socket if tcp.  telnet, e.g. will hang until
          * a key is hit on the client side if this open is not done.
          *=================================================================*/
         if (assoc->tcp && partner->socket == -1)
         {
            if (opensocket (partner))
            {
               if (partner->receiver)
	          makefdlist ();	/* Rebuild socket list */
            }
            else
            {
               syslog (LOG_ERR, "Error: Failed opening socket to %s port %d",
			inet_ntoa(partner->sin.sin_addr),
			ntohs(partner->sin.sin_port));
	       return;
            }
         }

         /*==================================================================
          * Log this new connection
          *=================================================================*/
         {
	    char	th1[16],th2[16];

	    strcpy(th1,inet_ntoa(instance->sin.sin_addr));
	    strcpy(th2,inet_ntoa(partner->sin.sin_addr));
	    syslog(LOG_NOTICE,"connect %s src_ip = %s src_port = %d "
		"local_ip = %s local_port = %d local_port2 = %d "
		"dest_ip = %s dest_port = %d",
		instance->tcp?"TCP":"UDP",
		th1, ntohs(instance->sin.sin_port),
		inet_ntoa(instance->localip), ntohs(instance->localport),
		ntohs(partner->localport),
		th2, ntohs(partner->sin.sin_port));
         }
      }
#ifdef Debug
      else
      {
         DBPRINTF (2, ("mapped to instance 0x%x, %s/%d\n",
	               instance, inet_ntoa (instance->sin.sin_addr),
			  ntohs (instance->sin.sin_port)));
      }
#endif
      assoc             = instance;
   }
   /*========================================================================
    * Not a fanout, must be a socket on the right
    * if this assoc is for tftp and sin.sin_port is the LHS localport, we will
    * have to allow all remote ports thru.  We will then change sin.sin_port
    * so that it has the new remote port number.
    *=======================================================================*/
   else
   {
      if (assoc->tftp && assoc->sin.sin_port == assoc->partner->localport)
         assoc->sin.sin_port = fromaddr.sin_port;
      else if (fromaddr.sin_addr.s_addr != assoc->sin.sin_addr.s_addr ||
		 fromaddr.sin_port != assoc->sin.sin_port)
      {
         char th1[16];
         strcpy(th1,inet_ntoa(fromaddr.sin_addr));
         syslog(LOG_ALERT,"Security Alert: Attempt to access localport %d "
		"from IP %s/Port %d.  Expected IP %s/Port %d",
		assoc->localport,th1,ntohs(fromaddr.sin_port),
		inet_ntoa(assoc->sin.sin_addr),ntohs(assoc->sin.sin_port));
         return;
      }
   }

   assoc->nbytes             += pktlen;
   assoc->lastaccess          = time (NULL);
   assoc->partner->lastaccess = assoc->lastaccess;
   if (assoc->plex && assoc->plex->transient)
	assoc->plex->lastaccess = assoc->lastaccess;
   if (generic && assoc->transient)
   {
	generic->lastaccess          = assoc->lastaccess;
	generic->partner->lastaccess = assoc->partner->lastaccess;
   }
   sendtopartner (assoc->partner, pkt, pktlen, &fromaddr,
                   assoc->localport);
}
/*=============================================================================
 *   main()
 *
 *===========================================================================*/
int main (int argc, char *argv[])
{
   extern char 	*optarg;
   int		pollval;
   int		i;
   int		opt;
   int		timeout;
   pid_t	pid;
#ifdef __sun
   int		fdzero;
#endif

   openlog ("rpcproxy", LOG_PID, LOG_DAEMON);

#ifdef _AIX
   {
	struct utsname utsname;

	/*====================================================================
	 * If AIX 3.2, issue message and exit, it's not supported.
	 *===================================================================*/
	if (uname(&utsname) < 0)
	{
	    syslog(LOG_ERR,"Error: Uname call failed, =%m.\n");
	    exit(1);
	}
	if (strcmp(utsname.version,"3") == 0)
	{
	    syslog(LOG_ERR,"Error: rpcproxy is not supported on AIX 3.2");
	    exit(1);
	}
   }
#endif

   memset(&rpcconf,0,sizeof(rpcconf));
#ifdef Debug
   rpcconf.traceflag = Debug;
#endif
   rpcconf.idletime = 1800;	/* Default to 30 minutes timeout */
   rpcconf.qlen = 100;		/* Default listen queue length */
   rpcconf.filter = TRUE;	/* Always filter unless specifically disabled */

   while ((opt = getopt (argc, argv, "d:f:")) != -1)
   {
      switch (opt)
      {
#ifdef Debug
         case 'd':
	 	rpcconf.traceflag    = atoi(optarg);
		rpcconf.idletime = 20;		/* Short timeouts */
		setbuf(stdout, NULL);
		break;
#endif
	 case 'f':
		strncpy(rpcconffile,optarg,sizeof(rpcconffile));
		break;
	 default:
#ifdef Debug
	 	printf ("usage: rpcproxy [ -d ] [ -f configfile ]\n");
	 	syslog (LOG_ERR,"usage: rpcproxy [ -d ] [ -f configfile ]\n");
#else
		printf ("usage: rpcproxy [ -f configfile ]\n");
		printf (LOG_ERR,"usage: rpcproxy [ -f configfile ]\n");
#endif
		exit (1);
      }
   }

   syslog(LOG_NOTICE,"rpcproxy started.");

   /*======================================================================
    * Use the allowable limit for number of opened files
    *====================================================================*/
#ifndef _AIX
   {
	struct rlimit fl;
	int rc;
	rc = getrlimit(RLIMIT_NOFILE,&fl);
	if (rc)
	{
	    syslog(LOG_ERR,"Error: Cannot do getrlimit.");
	    exit(1);
	}
	fl.rlim_cur = fl.rlim_max;
	rc = setrlimit(RLIMIT_NOFILE,&fl);
	if (rc)
	{
	    syslog(LOG_ERR,"Error: Cannot do setrlimit.");
	    exit(1);
	}
	nfilemax = fl.rlim_cur;
   }
#else
	nfilemax = OPEN_MAX;
#endif

   fdlist = malloc(sizeof(struct pollfd) * nfilemax);
   bzero(fdlist,sizeof(struct pollfd) * nfilemax);
   fdlistassocs = malloc(sizeof(caddr_t) * nfilemax);
   bzero(fdlistassocs,sizeof(caddr_t) * nfilemax);

   readconfig ();

   /*========================================================================
    * Make this a daemon
    * The parent exits leaving the child to run forever
    *=======================================================================*/
   if (rpcconf.traceflag == 0)
   {
      if ((pid = fork()) < 0)
      {
         syslog(LOG_ERR, "Error: Could not fork");
         printf("rpcproxy: Could not fork");
         exit(1);
      }

      if (pid != 0)
         exit(0);
   }

   /*========================================================================
    * Ignore SIGCHLD so we don't get zombies
    *=======================================================================*/
   if (signal(SIGCHLD, SIG_IGN) == SIG_ERR)
   {
      syslog(LOG_ERR, "Error: signal failed to ignore SIGCHLD.  %m");
      exit(1);
   }

   /*===================================================================
    * Create initial packet buffer; die if no memory available
    *==================================================================*/
   if (!makepktbuf (0))
      exit (1);
   makefdlist ();		/* Build list of sockets to poll */
   while (TRUE)
   {
      if (rpcconf.idletime)
         timeout = checktimeouts();
      else
         timeout = -1;
      DBPRINTF (2, ("polling %d sockets, timeout %d\n", fdlistlen,
		  timeout));
      pollval = poll(fdlist, (u_long) fdlistlen, timeout);
      DBPRINTF (2, ("returned from poll\n"));
      if (pollval < 0)
      {
         syslog (LOG_ERR, "Error: poll failed, pollval=%d, %m",
	 pollval);
	 exit (1);
      }
      if (pollval == 0)
        continue;	/* Timeout only */
      for (i = 0; i < fdlistlen; ++i)
      {
         if ((fdlist [i].revents & POLLIN) != 0)
            fwdpacketfrom (fdlistassocs [i]);
         /*================================================================
          * Report any failures and set the lastaccess in the failed assoc
          * to zero so it will expire right away the next time checktimeouts()
          * is called.
          *===============================================================*/
         else if ((fdlist [i].revents & ~POLLIN) != 0)
         {
	    syslog (LOG_ERR, "Error: Error on polled socket %d, %m, "
				"revent=0x%x",
				fdlistassocs[i]->socket,
				fdlist [i].revents);
            fdlistassocs[i]->lastaccess = 0;
            dumpassoc(fdlistassocs[i],"Bad Assoc?	");
	 }
      }
   }
   /*NOTREACHED*/
}

