/*============================================================================
 * filename:	smwrap/smwrap.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=============================================================================
 *
 * written by:		R. James Livermore
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  
 * All rights reserved.  
 *
 * Date written:	December  16, 1994
 * Last modified:	February   8, 2000 - LPL
 *			February  29, 2000 - added call to externalrelay
 *
 * function:		Secure Mail Wrapper	
 *       
 * 	The secure mail wrapper (smwrap) provides a SMTP service which runs in
 *  	an unpriviledged state in a chrooted directory. 
 *      
 *	Smwrap creates a temporay file in the spool directory with a name
 *	smwrapXXXXXX where the six X's represent a random character string
 *	created to guarentee uniqueness. When the file is complete it is closed
 *	and renamed with a leading capital 'S' to signal smwrapd the file is
 *	ready for delivery.  If smwrap discovers any hacker signatures it will
 *	move the file to the sequestered directory. If the transmissions is
 *	broken for any reason the temporary file will removed (unlinked).     
 *              
 *
 *	Check for Hacker attacks:
 *		Bogus Helo command
 *		Helo buffer overflow
 *		SMTP commands that exceed 512 characters	
 *		Blocked sender|receiver
 *		Unauthorized internal user
 *		Maximum total Header size exceeded
 *		Excessive number of recipients
 *		Use of VRFY & EXPN commands
 *  
 * Note: The command replies conform to RFC 822.
 *                                               
 * To run in DEBUG mode execute as follows:   
 *
 *	./smwrap smwrap -D nnnn   then telnet to port nnn
 *============================================================================*/
   static  char ident[] = "@(#) smwrap, V 1.0.1, 00/02/29"
                          " (C) Freemont Avenue Software, Inc."
                          " 1995-2000";
#include	<sys/types.h>
#include	<sys/file.h>
#include	<sys/stat.h>
#include	<stdio.h> 
#include	<string.h>		
#include	<time.h> 
#include	<sys/times.h>		/* cpu timings		*/
#include	<ctype.h>
#include	<syslog.h>
#include	<netdb.h>
#include	<sys/socket.h>
#include	<sys/signal.h>
#include	<netinet/in.h> 
#include	<unistd.h>  		
#include	<stdlib.h>		  
#include	"firewall.h"
#include	"smwrap.h"   


/*===========================================================================
 * External variables
 *==========================================================================*/
extern	int	errno;
extern	char	*strtok();  			/* ptr to                   */ 
extern  mailblk *mailgate;		/* ptr to top of mail block list    */	
extern	char	*optarg;  

struct	towhom
{
   char		*who; 				/* ptr to RCPT		   */ 
   char         *aka;				/* ptr to RCPT aliases     */
   char		*forward;			/* ptr to forwarding addr  */
   int		authcode;			/* mailauth code	   */
   struct	towhom	*nxt;
};

static	struct	towhom	*recip   = (struct towhom *) 0;
static	char		*ruser   = (char *) 0;
static	char		PeerHost[MAXHOSTLN+1];
static	char		PeerIP[MAXADDRLN];
static	char	*tempfile  = (char *) 0; /* ptr to name of temp file	    */ 
static	int	curbytes   = 0;		 /* Current message byte count	    */
static	int	hdrbytecnt = 0;		 /* current header bye count	    */
static	int	currecip   = 0;
static	int	timeout    = 0;
static	int	livetim    = 0;
/*===========================================================================
 * function: burring(int timeout)
 * 
 * The first time it is called livetim should be greater than zero. It 
 * schedules a SIGALRM signal 'timeout' seconds in the future, and sets
 * livetim = 0.
 * 
 *===========================================================================*/
void burring(int sig)
{ 
   if(livetim == 0) 
   {
      syslog(LOG_NOTICE,"timeout after %d seconds - exiting.",timeout);
      if(tempfile != (char *)0)
         unlink(tempfile);
      exit(1);
   }
   livetim = 0;
   alarm(timeout);
   return;				/* added for C++ */
}
/*============================================================================
 *  main routine
 *                    av
 *===========================================================================*/
main(int ac, char* argv[])
{  
   char		actualname[128];	/* user's actual name           */  
   char		aka[MAXADDRLN+2];	/* mail aliases name    	*/    
   int		authcode	= 0;	/* mailauth authorization code  */
   char		buf[BUFSIZ]; 		/* input buffer			*/
   int		cmdlen;			/* SMTP command length		*/
   char		obuf[BUFSIZ + 100];     /* output buffer		*/
   static  long	clktck 	  	= 0;    /* clock ticks/sec		*/   
   char		*cmdp;			/* command pointer	   	*/
   static  float cpu; 
   char		*cptr;                  /* character ptr		*/
   char		*cptr2;  
   char		*cptr3;
   char		*ctimeptr;		/* point at current time	*/
   int		debugflg    = 0;  
   unsigned short debugport = 9000; 	/* default debug port		*/ 
   char		*domptr;		/* domain name pointer		*/ 

   char		forward[MAXADDRLN];
   char		extRcpt[MAXADDRLN+2];	/* external Rcpt: e-mail name 	*/
   char		extFrom[MAXADDRLN+2];	/* external From: e-mail name 	*/
   char		fromname[MAXADDRLN+2];	/* orig from: name		*/
   int		i; 
   int		IPsecure;               /* set = 1 for secured host 	*/ 
   char		novaname[MAXADDRLN+2];	/* new name			*/
   long		now    		= 0; 
   char		lochostname[MAXHOSTLN+1];/* local hostname		*/   
   char		locdomain[MAXHOSTLN+1]; /* local domain name		*/
   FILE		*vout = (FILE *)0;
   int		gotdata   	= 0; 

   char		newnamebuf[128];	/* new random filename      	*/
   int		numVRFY       	= 1;  
   int		peerdropped	= 0;	/* peer dropped connection  	*/
   char		*RCPTp;			/* RCPT pointer			*/
   int		reclen;			/* record length		*/
   int		relayflag	= 0;	/* external relay flag		*/	
   int		sequesterflg 	= 0; 
   char		*token1;
   int		trecno    	= 0;		/* number of trace records  */   
   int		on;                             /* setsockopt option value  */  
   int		rcode	  	= 0;		/* return code		    */ 
   char		rnamebuf[64];			/* random filename buffer   */
   char		*topptr;			/* top domain ptr	    */

   struct hostent	*hp;
   struct SMWRAP 	smwrap;
   struct SMWRAP 	*smwpptr;
   struct trace  	*tracebufadr;  
   long			tempport; 

   struct tracehdr 	traceh;  
   

   static time_t	starttime;
   static time_t	offtime; 
   static  struct tms	tmsstart;
   static  struct tms	tmsend;   
   fromname[0]   = '\0';
   fromname[sizeof(fromname)-1] = '\0';		/* terminate string	*/
   extFrom[sizeof(extFrom)-1]   = '\0';	
   smwpptr       = & smwrap; 
   rnamebuf[0]   = '\0';
   rnamebuf[sizeof(rnamebuf)-1]  = '\0'; 
   /* Initialize wall clock and cpu clock timings */
   if( (clktck = sysconf(_SC_CLK_TCK)) < 0)  /* Initialize clock ticks/sec */ 
      clktck = 100;		 
   time(&starttime); 
   /*=========================================================================
    * Open the system log and get intial cpu time. If CPU timer isn't working
    * set flag. Note: Solaris 2.6 CPU timer will overflow if system has not 
    * been re-booted for 248.55 days.  Yes this does happen  
    *========================================================================*/
   openlog("smwrap",LOG_PID|LOG_NDELAY,LOG_DAEMON);     
   if( times(&tmsstart) < 0 )
   {
      syslog(LOG_NOTICE,"Error: CPU timer returned negative number to times()"
			"function. CPU timer may have overflowed.");
       clktck = 0;	/* set to zero to stop cpu timing.	*/    
   } 

#define DEBUG
#ifdef DEBUG
   /*===================================================================
    *	If Debug is specified this it will bind the port and listen  
    *	directly (circumventing inetd) so you can run the program under
    *	a debugger
    *=================================================================*/
   if( ac > 2)
   {
      if( strcmp(argv[2],"-D") == 0)
      {  
	 /* check for debug port number  */
         if( ac > 3 )
         {
	    if( (strlen(argv[3]) < 6)  && isalldigits(argv[3]) )
            {
               if( (tempport = atoi(argv[3])) > 64000)
               {
                  printf("\ntnproxy: portnumber > 64000. Please try again.\n");
	          return(-1);
               }
               debugport = tempport;
            }
            else
            {
               printf("\ntnproxy: invalid portnumber (%s), must be a positive "
		      "integer less than 64000.\n          Please try again.\n",
		      argv[3]);
               return(-1); 
            }   
         }
         printf("\n smwrap: Debug mode specified. Use port %d.\n",
                 debugport);
         debugbind(debugport);
	 debugflg = 1; 
      }
      else       
      {
         usage();
         return(-1);
      }	
   } 
#endif      
 

   /*==========================================
    * test peer name 
    *=========================================*/
   if(tstpname(0,PeerHost,PeerIP,sizeof(PeerIP))) 
   {
      syslog(LOG_ERR,"Error: cannot get remote host");
      exit(1);
   } 
   else 
   {
      syslog(LOG_NOTICE,"connect host = %s (%s).",PeerHost,PeerIP);
   }
   /*======================================================================
    * If the client's host is secure IPsecure = 1. If it is not secure then
    * IPsecure = 0.  If IPsecure = -1 then the securenets file is invalid
    * mail server should be stopped. 
    *=====================================================================*/
   if( (IPsecure = ipsecure(PeerIP) ) < 0)
   {
      exit(1);
   }   
   /*=========================================================================
    * Read config file to obtain:
    * userid, groupid, maxbytes, maxreceipts, timout and spool directory  
    *======================================================================*/    
   if( getsmwp(smwpptr) < 0) 
   {
      exit(1);		/* exit if unable to read smwrapper config file */  
   }   
   timeout = smwpptr->timeout; 
   if(smwpptr->timeout > 0)
   { 
      signal(SIGALRM,burring);
      alarm(timeout);
   }
   /*=========================================================================
    * Change the root directory to directory specified in smwrap config file.
    * Must be root user to do this.
    *=======================================================================*/
   if( smwpptr->spoolpath[0] != '\0') 
   {  
      chdir("/");   		/* change current directory  */
      if( ( chdir(smwpptr->spoolpath) || chroot(smwpptr->spoolpath) ) ) 
      {
         syslog(LOG_ERR,"Error: cannot chroot to %s: %m",smwpptr->spoolpath);
         exit(1);
      }
      chdir("/");
   } 
   else
   {  
      syslog(LOG_ERR,"Error: NULL spool directory returned"); 
      exit(1);
   }
   /*==================================================================
    * Set group id.
    *==================================================================*/   
   if( setgid(smwpptr->groupid)) 
   {
      syslog(LOG_ERR,"Error: cannot set gid to %d: %m",smwpptr->groupid);
      exit(1);
   }
   /*================================================================
    * Set user id.
    *================================================================*/
   if( setuid(smwpptr->userid) ) 
   {
      syslog(LOG_ERR,"Error: cannot set uid to %d: %m",smwpptr->userid);
      exit(1);
   }
   /*================================================
    * get name of the local host 
    *===============================================*/  

   if(gethostname(lochostname,sizeof(lochostname)))
   {  
      syslog(LOG_ERR,"Error: gethostname failed %m"); 
      exit(1); 
   } 
 
   /*=======================================================================
    * If address is of the form hostname.domain.toplevel (two dots) then
    * extract domain name from hostname                              
    *======================================================================*/    
   if( (domptr = strchr(lochostname,'.')) != NULL)
   {                                                      
      domptr++;
      if( (topptr = strchr(domptr,'.')) != NULL)
      {
         strcpy(locdomain,domptr);
         strcpy(lochostname,locdomain); 
      }
   } 
   /*======================================================================
    * if trace facility is specified initialize trace array  
    *======================================================================*/
   if( smwpptr->traceflg ) 
   {
      if( (tracebufadr = trace_init(&traceh)) == NULL) 
      {  
         syslog(LOG_ERR,"Error: trace_init returned NULL. strerror(errno)");
         smwpptr->traceflg = 0; 
      }/* else syslog(LOG_NOTICE,"trace on");*/
   }
   time(&now);  
   ctimeptr = ctime(&now);
   cptr = ctimeptr + strlen(ctimeptr) - 1;
   *cptr = '\0';
   sprintf(obuf,"220 %s Secure Mail Wrapper ready at %s %d\r\n",
		lochostname, ctimeptr, -(timezone-daylight*3600)/36); 
   printf("%s",obuf);
   fflush(stdout); 
   if( smwpptr->traceflg)
            trace_it(&traceh, tracebufadr, 'W', obuf);
         

   on = 1;                 	/* C++ requires (char *) &x */
   if( setsockopt(fileno(stdin),SOL_SOCKET,SO_KEEPALIVE,(char *)&on,sizeof(on)) < 0)
   {
      syslog(LOG_ERR,"Error: setsockopt failed for SO_KEEPALIVE. %s",
		strerror(errno)); 
   }   
   i = 0;
   /*======================================================================
    * Read data from remote host until:
    * 	QUIT command or Max Recipients or EOF or peer drops connection,  
    *======================================================================*/ 
   while(1) 
   {
      if((reclen = getswocrlf(buf,sizeof(buf),stdin)) < 0) 
      {  

	 if(feof(stdin))
         {  
            syslog(LOG_NOTICE,"Error: peer dropped connection while reading"\
			" SMTP commands: unexpected EOF from %s.",
			PeerHost);
	 }
         else if( ferror(stdin))
         {
            syslog(LOG_NOTICE,"Error: peer dropped connection: %s.",
				strerror(errno));
         }
         else
         {
            syslog(LOG_NOTICE,"Error: (3) peer dropped connection: %m."); 
	 }
         rcode = 1;
	 break;
      }
      if( smwpptr->traceflg)
      {      
        trace_it(&traceh, tracebufadr,'R', buf);
      }
      /*=====================================================================
       * Check for excessive SMTP Command length            
       *===================================================================*/
      if( reclen  > MAXSMTPLEN )
      {
         sprintf(obuf,"500 Invalid SMTP command - excessive length. \r\n");
	 printf("%s",obuf);
	 syslog(LOG_ALERT,"Securiy Alert: Invalid SMTP command. length %d "
	 		  "exceeds %d. %.100s...'",
			  reclen,MAXSMTPLEN,buf);
	 fflush(stdout); 
	 if( smwpptr->traceflg)
            trace_it(&traceh, tracebufadr, 'W', obuf);
	 continue;
      }
      if((cmdp = strtok(buf," \r\t\n")) == (char *)0) 
      {
         sprintf(obuf,"500 The command is not known. \r\n");
	 printf("%s",obuf);
	 fflush(stdout); 
	 if( smwpptr->traceflg)
            trace_it(&traceh, tracebufadr, 'W', obuf);
	 continue;
      } 
      if( strcasecmp(cmdp,"VRFY") && strcasecmp(cmdp,"EXPN")) 
         numVRFY = 0;
      /*=================================================
       * process the MAIL command
       *================================================*/
      if(!strcasecmp(cmdp,"MAIL")) 
      {
         char	*q;           
         char	*q2;

         if((q = strtok((char *)0,"\r\n")) == (char *)0) 
         {
            sprintf(obuf,"501 There is a syntax error. 'From:' should "
	    		 "follow 'MAIL'.\r\n");
	    printf("%s",obuf);
	    fflush(stdout); 
	    if( smwpptr->traceflg)
               trace_it(&traceh, tracebufadr, 'W', obuf);
	    continue;
	 }

         if(strncasecmp(q,"From:",5)) 
         {
            sprintf(obuf,"501 There is a syntax error. 'From:' should "
	    		 "follow 'MAIL'.\r\n");
	    printf("%s",obuf);
	    fflush(stdout); 
	    if( smwpptr->traceflg)
               trace_it(&traceh, tracebufadr, 'W', obuf);
	    continue;
	 } 
	 /* skip over "From:" and any white spaces */
	 q += 5;
         while(isspace(*q))
         {
	    q++;
         } 
         /*============================================================
          * If there is an actual name in front of <name@domain> then
          * preserve it.
          *===========================================================*/
         actualname[0] = '\0';
         if(( cptr2 = strchr(q,'<')) != NULL)
         {
            strncpy(actualname,q,sizeof(actualname)-1); /* save actual name     */
            q = cptr2;			/* point to e-mail addr */
	    cptr2 = strchr(actualname,'<');
            *cptr2 = '\0'; 
	 } 
         strncpy(fromname,q,sizeof(fromname)-1);/* save orig from name for bldbody*/

         /*===============================================================
          * If the peer host is a secured system (IPsecure == 1) 
          * then translate sender name to external name.
	  * If transend is unable to translate internal name@host then use
	  * name@domain.    
	  *==============================================================*/
	 if( IPsecure == SECURED )
         {                                     

            /*============================================================= 
             * Translate the senders name. 
             * If an aliases is required and there is not an entry in the
             * DB then issue an Alert and Sequester the file.
             *===========================================================*/
            if( transend(q,extFrom,lochostname,smwpptr->translate) != 0)
            {  
               if( smwpptr->aliasflag & ALIASREQ  )
               {
                  syslog(LOG_ERR,"Security Alert: unauthorized user %s"
                                 " trying to send outbound mail",
                                 q ); 
                  sequesterflg = 1;
               }
               /* let mail go but hide internal host name */
	       extFrom[0] = '\0';
               strcat(extFrom,q);  
               q2 = (char *) &extFrom;
               while( *q2 != '@') 
               {
                  if( *q2 == '\0')
                  {
                     sprintf(obuf,"501 Syntax error: no '@' found in From "
				  "address.\r\n");
	             printf("%s",obuf);
	             fflush(stdout); 
	             if( smwpptr->traceflg)
                        trace_it(&traceh, tracebufadr, 'W', obuf);  
		     rcode = 1; 
                    /* close(stdin); */
	             goto theend;
                  } 
                  q2++;
               } 
               q2++;
               *q2 = '\0';
               strcat(q2,lochostname); 
               
               if( extFrom[0] == '<')
                  strcat(q2,">");
               syslog(LOG_NOTICE,"unable to translate %s to external name. "
				 "used %s  %s(%s)",
		      		 q, extFrom, PeerHost,PeerIP ); 
               q = (char *) &extFrom;    
	    }
            else
            {
            /*   syslog(LOG_NOTICE," translated %s to %s", q, extFrom); */
               q = (char *) &extFrom; 
            }
         }        
	 /*===============================================================
	  * Free storage used to hold name&address of previous sender and 
	  * acquire storage to hold new sender's name&address         
	  *==============================================================*/ 
         if(ruser != (char *)0)
	    free(ruser);
	 ruser = (char *) malloc(strlen(q) + strlen(actualname) + 1);
	 if(ruser == (char *)0) 
         {  /* 421 ? or 410?? */
	    sprintf(obuf,"421 out of memory. Shutting down.\r\n");
	    printf("%s",obuf);
	    fflush(stdout); 
	    if( smwpptr->traceflg)
               trace_it(&traceh, tracebufadr, 'W', obuf);
	    syslog(LOG_CRIT,"Error: out of memory: %m.");
	    unlink(tempfile); 
            rcode = 1;
	    break; 
	 } 
         if( IPsecure == SECURED)
	 {
            strcpy(ruser,actualname);	/* preserve "actual name" */
         }
         else
         {
            *ruser = '\0';
         }     
         strcat(ruser,q);
	 sprintf(obuf,"250 %s... Sender Ok\r\n",ruser);
         printf("%s",obuf);
	 fflush(stdout); 
	 if( smwpptr->traceflg)
            trace_it(&traceh, tracebufadr, 'W', obuf);
	 continue;
      }
      /*======================================================
       * process RCPT command
       *======================================================*/
      if(!strcasecmp(cmdp,"RCPT")) 
      {
         struct	towhom	*nr;
         if(ruser == (char *)0) 
         {
            sprintf(obuf,"503 need MAIL From: first.\r\n");
	    printf("%s",obuf);
	    fflush(stdout); 
	    if( smwpptr->traceflg)
               trace_it(&traceh, tracebufadr, 'W', obuf);
	    continue;
	 }

         if((RCPTp = strtok((char *)0,"\r\n")) == (char *)0) 
         {
	    sprintf(obuf,"501 There is a syntax error. 'To:' should "
	    		 "follow RCPT.\r\n");
            printf("%s",obuf);
	    fflush(stdout); 
	    if( smwpptr->traceflg)
               trace_it(&traceh, tracebufadr, 'W', obuf);
	    continue;
	 }

	 if(strncasecmp(RCPTp,"To:",3)) 
         {
	    sprintf(obuf,"501 There is a syntax error. 'To:' should "
	    		 "follow RCPT.\r\n");
	    printf("%s",obuf);
	    fflush(stdout); 
	    if( smwpptr->traceflg)
               trace_it(&traceh, tracebufadr, 'W', obuf);
	    continue;
	 }
         /* skip over "To:" and any white spaces */
	 RCPTp += 3;
	 while(isspace(*RCPTp))
	    RCPTp++; 
         *aka = '\0';
	 strncpy(extRcpt,RCPTp,MAXADDRLN-5);	/* save for mailauth	*/
         extRcpt[MAXADDRLN-5] = '\0';          
	 /*===============================================================    
          * If the peer host is a non-secured system (IPsecure != 1)          
          * First check for external mail relay and reject mail if found.
	  * second check to see if the recipient name is valid.  	 
          *==============================================================*/
	 if( IPsecure == NOTSECURED )          
	 { 
	    relayflag = externalrelay(RCPTp);
	    switch(relayflag) 
            {	 
               case 0:
   		   	break;      
	     
	       case 1:
	       		syslog(LOG_ALERT,"Security Alert: External Relay "
			                 " attempted %s from %s (%s)",
					 RCPTp, PeerHost, PeerIP);
	       		sprintf(obuf,"551 We do not do relay.\n");
	                printf("%s",obuf);
	                fflush(stdout);
			continue;
			break; 
               default:
	       		syslog(LOG_ERR,"Error: invalid recipient %s "
				       " from %s (%s)",
				       RCPTp, PeerHost, PeerIP);
	       		sprintf(obuf,"551 Invalid recipient.\n");
	                printf("%s",obuf);
	                fflush(stdout);
			continue;	       	    
            }
            /*============================================================ 
             * If noalias is specified send mail to specified postoffice.
	     *===========================================================*/
	    if( smwpptr->aliasflag & NOALIAS )
            {
               strncpy(aka,RCPTp,250); 
	       token1 = strchr(aka,'@'); 
               if( token1 == NULL)
               { 
                  syslog(LOG_NOTICE,"rejected invalid recip %s from %s (%s),"
				    " missing '@'.",
		                     RCPTp, PeerHost,PeerIP );  
                  sprintf(obuf,"550 Invalid recipient %s.\r\n",RCPTp);
	          printf("%s",obuf);
	          fflush(stdout);
                  if( smwpptr->traceflg)
                     trace_it(&traceh, tracebufadr, 'W', obuf); 
                  continue;  
               } 
	       token1++;
               strcpy(token1,smwpptr->postoffice);
               strcat(token1,">");   
            }
            else if( validrcpt(RCPTp,aka) != 0)
            {
               syslog(LOG_NOTICE,"rejected invalid recip %s from %s (%s).",
		      RCPTp, PeerHost,PeerIP );  
               sprintf(obuf,"550 Invalid recipient %s.\r\n",RCPTp);
	       printf("%s",obuf);
	       fflush(stdout);
               if( smwpptr->traceflg)
                  trace_it(&traceh, tracebufadr, 'W', obuf); 
               continue;
	    }
         }     
         else  
         /*=================================================================== 
          * Else process mail comming from a Secured System.  
          * Remove explicit source routing and %Hack routing if found. 
          *==================================================================*/
	 {   

            /*=================================================================     
             * If host is secured and explicit source routing is used then  
             * remove any source route information.  
             * <@one,@two,@three:joe@four.com>
             *================================================================*/ 
            if( *(RCPTp+1) == '@' )
            {
               strcpy(aka,"<");		/* address starts with '<'	*/ 
               /* find first ':'	*/
               if( ( cptr = strchr(RCPTp,':')) == (char *) 0)
	       {
                  syslog(LOG_ERR,"Error: invalid explicit source route %s.",
				 RCPTp);  
                  sprintf(obuf,"550 Invalid explicit source route %s.\r\n",
		  		RCPTp);
	          printf("%s",obuf);
	          fflush(stdout); 
                  if( smwpptr->traceflg)
                     trace_it(&traceh, tracebufadr, 'W', obuf);
                  continue;
               }  
               cptr++;
               strcat(aka,cptr); 
               if (*(aka+strlen(aka)-1) != '>')
	          strcat(aka,">");
	       if(debugflg)  
                  syslog(LOG_NOTICE," source route %s changed to %s.",
		  	            RCPTp,aka);
               RCPTp = (char *) &novaname;  
               strcpy(RCPTp,aka);		/* save name minus source route info */
               
            }       
            /*================================================================
             * Check for the "%" Hack.  
             * If the address has the form:
             *     username%domain%relay@gateway 
             * replace it with
	     *     username%domain@relay
	     * If the hostname of the firewall does not match gateway then
             * send the entire address through unchanged. 
             *===============================================================*/
            if( ( cptr = strrchr(RCPTp,'%')) == NULL) /* find last % */
               goto notHack;              
	    if( ( cptr2 = strchr(cptr,'@')) == NULL)
               goto notHack;         
            cptr2++;  
            if( (cptr3 = strchr(cptr2,'>')) != NULL)
            {
               if( strncmp(cptr2,lochostname,cptr3-cptr2) != 0)
                  goto notHack;
            }
            if( strncmp(cptr2,lochostname,strlen(lochostname)) != 0)
               goto notHack;
	    strcpy(aka,"<");  
            *cptr = '\0';
            strcat(aka,RCPTp+1);
            strcat(aka,"@");
            cptr2 = strtok(cptr+1,"@");
            strcat(aka,cptr2);
            strcat(aka,">");
	    RCPTp = (char *) &novaname;
	    strcpy(RCPTp,aka);     
notHack:;
            /*=================================================================
             * Change to allow an internal user to send mail to another
             * internal user (different domain.name) using the smwrap.
             *================================================================*/
            if( internalrcpt(RCPTp,aka) == 0)
            {
               RCPTp = (char *) &aka;
            } 
         } /*=============================================================
            * end of Internal Mail handler 
            *============================================================*/ 
         if(smwpptr->maxreceipts > 0 && currecip++ > smwpptr->maxreceipts) 
         {
            sprintf(obuf,"552 There are too many recipients.\r\n");
	    printf("%s",obuf);
	    fflush(stdout); 
	    if( smwpptr->traceflg)
               trace_it(&traceh, tracebufadr, 'W', obuf);
	    syslog(LOG_NOTICE,"exiting - number of recipients exceeds max (%d).",
                             smwpptr->maxreceipts);
	    unlink(tempfile); 
            rcode = 1;
	    break; 
	 }
	 /*===================================================================
          * Check if mail for sender-receiver pair should be blocked.
          *==================================================================*/
         authcode = authmail(fromname, extRcpt, mailgate,forward); 
         switch(authcode) 
         {	 
            case 0:
   		break;
            case 1:   
	    	syslog(LOG_ALERT,"Security Alert: mail from %s to %s has been"
		   	     	 " blocked by rule.",
			     	 fromname,RCPTp);   
                strcpy(obuf,"550 Mail rejected.\r\n");
	        printf("%s",obuf);
	        fflush(stdout);
                if( smwpptr->traceflg)
                   trace_it(&traceh, tracebufadr, 'W', obuf); 
		goto end_of_RCPT;
            	break;                         
 	    case 2:
	    	syslog(LOG_ALERT,"Security Alert: mail from %s to %s has been"
		   	     	 " blocked by rule. Mail forwarded to %s",
			     	 fromname,RCPTp, forward); 
                strcpy(aka,forward);  
                /*RCPTp = forward;*/                
                break;
	    case 4: 
	    	syslog(LOG_ALERT,"Security Alert: mail from %s to %s has been"
		   	     	 " blocked by rule. Mail sequestered.",
			     	 fromname,RCPTp);
             	sequesterflg = 1;
		break;
	    default:
		syslog(LOG_ERR,"Error: invalid return code from authmail.");
		return(-1);     

         }
         /*================================================== 
          * Allocate storage for towhom structure.	
	  *=================================================*/
	 nr = (struct towhom *)malloc(sizeof(struct towhom));
	 if(nr == (struct towhom *)0) 
         {
            sprintf(obuf,"410 out of memory.\r\n");
	    printf("%s",obuf);
	    fflush(stdout); 
	    if( smwpptr->traceflg)
               trace_it(&traceh, tracebufadr, 'W', obuf);
	    syslog(LOG_CRIT,"Error: out of memory: %m.");
	    unlink(tempfile); 
            rcode = 1;
	    break; 
	 } 
	 /*==============================================================
	  * Acquire storage to hold RCPT address, Alias Address and 
          * forwarding address.	
	  *===============================================================*/
         nr->who = (char *) malloc( strlen(RCPTp)   + 1 + strlen(aka) + 1 
				    + strlen(forward) + 1);
	 if(nr->who == (char *)0) 
         {
            sprintf(obuf,"410 out of memory.\r\n");
	    printf("%s",obuf);
	    fflush(stdout); 
	    if( smwpptr->traceflg)
               trace_it(&traceh, tracebufadr, 'W', obuf);
	    syslog(LOG_CRIT,"Error: out of memory: %m.");
	    unlink(tempfile);
            rcode = 1;
	    break;
	 }  
         nr->aka = nr->who + strlen(RCPTp) + 1;
         nr->forward = nr->aka + strlen(aka) + 1;
         /*==========================================
	  * copy recipient address to nr->who and
          * copy alaises   address to nr->aka
          *=========================================*/ 
	 strcpy(nr->who,RCPTp); 
         strcpy(nr->aka,aka);  
         strcpy(nr->forward,forward);
         nr->authcode = authcode;
	 nr->nxt      = recip;  	/* chain to previous structure	*/
	 recip        = nr;             /* update ptr			*/
	 sprintf(obuf,"250 %s OK\r\n",RCPTp);
         printf("%s",obuf);
	 fflush(stdout); 
	 if( smwpptr->traceflg)
            trace_it(&traceh, tracebufadr, 'W', obuf);
end_of_RCPT:
	 continue; 
      }
      /*======================================================
       * process DATA command
       *=====================================================*/
      if(!strcasecmp(cmdp,"DATA")) 
      {
         struct	towhom	*tp;
         int		msgsize;  

	 curbytes   = 0;	/* initialze message byte count	*/
	 hdrbytecnt = 0;	/* initialize header byte count */
	 currecip   = 0;
	 if(recip == (struct towhom *)0) 
         {
            sprintf(obuf,"503 RCPT must preceed DATA command.\r\n");
	    printf("%s",obuf);
	    fflush(stdout); 
	    if( smwpptr->traceflg)
               trace_it(&traceh, tracebufadr, 'W', obuf);
	    continue;
	 }      
         /*=====================================
          * open temporay file to store e-mail    
          *====================================*/
         if((vout = smwrapopen(rnamebuf)) == (FILE *)0) 
         {
            syslog(LOG_ERR,"Error: cannot open temporary file: %m.");
	    exit(1);
	 }

	 sprintf(obuf,"354 Enter mail, end with \".\" on a line by itself."
	 	      "\r\n");
         printf("%s",obuf);
	 fflush(stdout); 
	 if( smwpptr->traceflg)
            trace_it(&traceh, tracebufadr, 'W', obuf);         

	 if( fprintf(vout,"FROM %s\n",ruser) < 0)
         {
            syslog(LOG_ERR,
	           "Error: cannot write \"FROM %s\" to temporary file: %m.");
	    sprintf(obuf,"555 Unable to receive data.  Bye.\r\n");
            printf("%s",obuf); 
            unlink(tempfile); 
	    exit(1);
         }  
	 smwpptr->linecnt = 1; 		/* initialize line cnt for message */
         /*=================================================================
          * write each RCPT to the temporary file 
	  * If sender is unsecured translate RCPT using aliases DB  
          *================================================================*/
	 for(tp = recip; tp != (struct towhom *)0; tp = tp->nxt) 
         {   
	    if( IPsecure != 1)
            {
               fprintf(vout,"RCPT %s\n",tp->aka); 
            }
            else
            {
	       fprintf(vout,"RCPT %s\n",tp->who); 
            } 
	    smwpptr->linecnt++;
         }
	 fprintf(vout,"BODY\n");
	 smwpptr->linecnt++;
	 time(&now); 
         if(IPsecure == 1)
         {   
            fprintf(vout,"Received: by %s via smwrap %s\n\t id %s; %s",
                 lochostname,"Version 1.0",tempfile,ctime(&now));
         }
         else
         {
	    fprintf(vout,"Received: from %s(%s) by %s via smwrap %s\n\t id %s; %s",
                 PeerHost,PeerIP,lochostname,"Version 1.0",tempfile,ctime(&now));
	 }  
	 smwpptr->linecnt++;
         /*================================================================
          * If mail is blocked insert Error message into body. 
          *===============================================================*/
	 for(tp = recip; tp != (struct towhom *)0; tp = tp->nxt) 
         {   
            switch(tp->authcode)
            {
       	       case 2:
            	   fprintf(vout,"Alert: mail from %s to %s has been blocked by "
			     "rule.\n"
		   	     "\tMail forwarded to %s.\n",
			     fromname,tp->who, tp->forward);                 
            	   break;
               case 4:  	          
            	   fprintf(vout,"Alert: mail from %s to %s has been blocked by "
		   	     "rule.\n"
			     "\tMail sequestered.\n",
			     	 fromname,tp->who);                       
		   break;
               default:
		   break;
            }
         }
	 /*================================================================
          * process message body and examine return code
          *===============================================================*/ 
         switch( bldbody(stdin,vout,smwpptr,IPsecure,extFrom,lochostname,
		         fromname, &traceh, tracebufadr))  
         {
	    case MSGOK:
	            sprintf(obuf,"250 Mail accepted.\r\n");
                    printf("%s",obuf);
		    fflush(stdout);
		    if( smwpptr->traceflg)
            		trace_it(&traceh, tracebufadr, 'W', obuf); 
		    gotdata++;
	 	    break;
            case UNEXPEXIT:  
		    if(feof(stdin))
         	    {  
            	       syslog(LOG_NOTICE,"Error: peer dropped connection while "
				"reading body: unexpected EOF from %s.",
				PeerHost);
	 	    }
         	    else if( ferror(stdin))
         	    {
            		syslog(LOG_NOTICE,"Error: peer dropped connection "
				"while reading body: %s.",
				strerror(errno));
         	    }
                    peerdropped++;
	            sprintf(obuf,"550 you could say goodbye\r\n"); 
                    printf("%s",obuf);
		    fflush(stdout);
		    if( smwpptr->traceflg)
            		trace_it(&traceh, tracebufadr, 'W', obuf);        
		    break;
	    case MSGTOOBIG:
	            sprintf(obuf,"550 Maximum message size exceeded."
		    		 "  Bye.\r\n");
		    printf("%s",obuf);
		    fflush(stdout);
		    syslog(LOG_WARNING,"Warning: maximum message size %d"
		    		       " exceeded.  Message deleted.",
			    smwpptr->maxbytes);
		    unlink(tempfile); 
                    if( smwpptr->traceflg)
            		trace_it(&traceh, tracebufadr, 'W', obuf);
		    rcode = 1;
		    goto theend;
	    case HDRTOOBIG:
	            sprintf(obuf,"550 Header exceeds maximum size.  Bye.\r\n");
		    printf("%s",obuf);
		    fflush(stdout);
		    syslog(LOG_ALERT,"Security Alert: maximum header size %d"
		    		     "exceeded. File %s sequestered.",
			    smwpptr->maxheadersz,rnamebuf);
		    sequesterflg = 1; 
                    if( smwpptr->traceflg)
            		trace_it(&traceh, tracebufadr, 'W', obuf);
		    rcode = 1;
		    goto theend;
	    case HDRADRBAD:
	    	    sprintf(obuf,"550 Invalid Header address.  Bye.\r\n");
		    printf("%s",obuf);
		    fflush(stdout);
		    syslog(LOG_ALERT,"Security Alert: Dangerous header size %d"
		    		     "exceeded. File %s sequestered.",
			    smwpptr->maxheadersz,rnamebuf);
		    sequesterflg = 1; 
                    if( smwpptr->traceflg)
            		trace_it(&traceh, tracebufadr, 'W', obuf);
		    rcode = 1;
		    goto theend;
		    	    
            case FPRTERR:
	            sprintf(obuf,"451 System problem.\r\n"); 
                    printf("%s",obuf);
		    fflush(stdout); 
                    unlink(tempfile); 
                    if( smwpptr->traceflg)
            		trace_it(&traceh, tracebufadr, 'W', obuf);
		    continue;
            default:
	            sprintf(obuf,"250 Mail accepted.\r\n");
                    printf("%s",obuf);
		    fflush(stdout);
		    if( smwpptr->traceflg)
            		trace_it(&traceh, tracebufadr, 'W', obuf); 
		    gotdata++;
					/* .... fall through... */
         }

	 fflush(vout);
	 msgsize = filesize(fileno(vout));
         /*============================================================= 
	  * log mail for each recipient
	  *=============================================================*/
	 for(tp = recip; tp != (struct towhom *)0; tp = tp->nxt)
         {     
            if( IPsecure != 1)
            {
               syslog(LOG_NOTICE,"host = %s (%s) bytes = %d from = %s to = %s"
				 " aka %s",
		                  PeerHost,PeerIP,msgsize,ruser,tp->who,tp->aka);  
            } 
            else
            {
	       syslog(LOG_NOTICE,"host = %s (%s) bytes = %d from = %s to = %s",
		       PeerHost,PeerIP,msgsize,ruser,tp->who);  
            }
         }
	 if(recip != (struct towhom *)0) 
         {
            currecip = 0;
            /* should really free them */
	    recip = (struct towhom *) 0;
	 }
         /*=================================================================
	  * If peer host issues QUIT command without a message body delete
	  * file and issue message. 
          *================================================================*/ 
         if(!peerdropped)
         {
	    if(!gotdata) 
            {
               unlink(tempfile);
	       syslog(LOG_NOTICE,"SMTP QUIT with no message %s (%s)",
	       PeerHost,PeerIP);
	    } 
         }
         /*=================================================================
          * Close the file then rename it using a leading 'Smwrap' to
          * indicate completion.             
          * If the sequesterflag is on move the file into the sequestered 
          * directory. 
          *===============================================================*/
         if( fclose(vout) < 0)
         {
            syslog(LOG_ERR,"Error: fclose(%s) failed,  %s.",
	   	   	   tempfile, strerror(errno));    
         }
         vout = (FILE *)0;  
         if(!peerdropped)
         {   
            if( sequesterflg )
            {          
               strcpy(newnamebuf,"sequestered/");          
               strcat(newnamebuf,rnamebuf);
               newnamebuf[12] = 'S';   
               sequesterflg = 0;
            }
            else
            {
               strcpy(newnamebuf,rnamebuf);
               newnamebuf[0] = 'S';   
            }
            if( rename(rnamebuf,newnamebuf) < 0)
            {
	       syslog(LOG_ERR,"Error: rename from %s to %s failed,  %s.",
	      	      rnamebuf, newnamebuf,  strerror(errno));
               exit(1);              
            }
         }                                                   
         /*=====================================  
          * gather and log statistics  
          *====================================*/
         time(&offtime); 
         if( times(&tmsend) < 0)
         {
            syslog(LOG_NOTICE,"Error: times(0 error. cpu timing stopped."); 
            clktck = 0;
         }
         if( clktck > 1)
         {
            cpu = ( tmsend.tms_utime   + tmsend.tms_stime 
	          -tmsstart.tms_utime - tmsstart.tms_stime)/(1.0*clktck);
         } 
         if(peerdropped)
         {
            unlink(tempfile); 
	    syslog(LOG_NOTICE,"exiting host = %s (%s) bytes = %d, cpu = %.2f,"
			      " file = %s deleted.",
			      PeerHost, PeerIP, msgsize, cpu, tempfile); 
            exit(1);
         }  
	 syslog(LOG_NOTICE,"exiting host = %s (%s) bytes = %d, cpu = %.2f, "
	 		   "file = %s.",
			   PeerHost, PeerIP, msgsize, cpu, newnamebuf); 

         continue;
      }
      /*==========================================================
       * Process VRFY and EXPN commands.
       *==========================================================*/
      if(!strcasecmp(cmdp,"VRFY") || !strcasecmp(cmdp,"EXPN")) 
      {
         char	*q;

         if((q = strtok((char *)0," \t\r\n")) == (char *)0)
               sprintf(obuf,"550 User Unknown.\r\n");
         else
            sprintf(obuf,"250 <%s>\r\n",q); 
         
	 syslog(LOG_ALERT,"Security Alert: '%s %s' request from %s (%s).",
                          cmdp,q,PeerHost,PeerIP);  
         sleep(numVRFY);
         printf("%s",obuf);
         fflush(stdout);  
         numVRFY++;
	 if( smwpptr->traceflg)
            trace_it(&traceh, tracebufadr, 'W', obuf);
	 continue;
      }
      /*============================================================
       * process HELP command
       *===========================================================*/
      if(!strcasecmp(cmdp,"HELP")) 
      {   
         char	*q;  
         if((q = strtok((char *)0,"\r\n")) == (char *)0)
         {  
            printf("214-Commands:\r\n");
            printf("214-   HELO    MAIL    RCPT    DATA    RSET\r\n");
	    printf("214-   NOOP    QUIT    HELP    VRFY    EXPN\r\n");
            sprintf(obuf,"214 End of HELP information.\r\n"); 
	    printf("%s",obuf);
	    fflush(stdout); 
	    if( smwpptr->traceflg)
               trace_it(&traceh, tracebufadr, 'W', obuf);
            continue; 
         }
         if(strncasecmp(q,"HELO",4) == 0) 
         {
            printf("214-Usage: HELO hostname\r\n"); 
            printf("214-            Identifies the sending host.\r\n"); 
	 }
         else 
         if(strncasecmp(q,"MAIL",4) == 0) 
         {
            printf("214-Usage: MAIL From: sender\r\n"); 
            printf("214-            Specifies the sender.\r\n"); 
	 } 
         else 
         if(strncasecmp(q,"RCPT",4) == 0) 
         {
            printf("214-Usage: RCPT To: recipient\r\n"); 
            printf("214-            Specifies the recipient. Can be used "
	    	   "multiple times.\r\n"); 
	 } 
         else 
         if(strncasecmp(q,"DATA",4) == 0) 
         {
            printf("214-Usage: DATA\r\n"); 
            printf("214-       The following text is collected as the "
	    	   "message.\r\n"); 
	    printf("214-       End with a line containing only the '.' "
	    	   "character.\r\n");
	 }
	 else
         if(strncasecmp(q,"RSET",4) == 0) 
         {
            printf("214-Usage: RSET\r\n"); 
            printf("214-       Resets the protocol state.\r\n");
         }
	 else
         if(strncasecmp(q,"NOOP",4) == 0) 
         {
            printf("214-Usage: NOOP\r\n"); 
            printf("214-       Does nothing.\r\n");
         } 
         else
         if(strncasecmp(q,"QUIT",4) == 0) 
         {
            printf("214-Usage: QUIT\r\n"); 
            printf("214-       Exits the sendmail wrapper.\r\n");
         } 
         else
         if(strncasecmp(q,"HELP",4) == 0) 
         {
            printf("214-Usage: HELP [topic]\r\n"); 
            printf("214-       The HELP command provides help information.\r\n");
            printf("214-       The topic is optional.\r\n");
         } 
         else
         if(strncasecmp(q,"VRFY",4) == 0) 
         {
            printf("214-Usage: VRFY Recipient\r\n"); 
            printf("214-       Allows verification of recipient.  "
	    	   "Not part of protocol.\r\n");
         } 
         else
         if(strncasecmp(q,"EXPN",4) == 0) 
         {
            printf("214-Usage: EXPN Recipient\r\n"); 
            printf("214-       Same as VRFY. Not part of protocol.\r\n");
         }  
	 sprintf(obuf,"214 End of help information.\r\n"); 
	 printf("%s",obuf);
	 fflush(stdout); 
	 if( smwpptr->traceflg)
            trace_it(&traceh, tracebufadr, 'W', obuf);
	 continue;
      }  
      /*============================================================
       * process NOOP command
       *===========================================================*/     
      if( !strcasecmp(cmdp,"NOOP") ) 
      {
         sprintf(obuf,"220 Ok.\r\n");
         printf("%s",obuf);
	 fflush(stdout);
	 if( smwpptr->traceflg)
            trace_it(&traceh, tracebufadr, 'W', obuf);
	 continue;
      }
      /*============================================================
       * process QUIT command
       *============================================================*/
      if( !strcasecmp(cmdp,"QUIT") ) 
      {
         sprintf(obuf,"221 %s is closing the connection.\r\n",lochostname); 
         printf("%s",obuf);
	 fflush(stdout);
	 if( smwpptr->traceflg)
            trace_it(&traceh, tracebufadr, 'W', obuf);
	 break;
      }
      /*=============================================================
       * process RSET command
       *============================================================*/
      if(!strcasecmp(cmdp,"RSET")) 
      {
         recip = (struct towhom *)0;
         ruser = (char *)0;
	 if(vout != (FILE *)0) 
         {
            unlink(tempfile);
	    fclose(vout);
	    if((vout = smwrapopen(rnamebuf)) == (FILE *)0) 
            {
	       syslog(LOG_ERR,"Error: cannot open temporary file %m.");
	       exit(1);
	    }
         }
	 sprintf(obuf,"250 Reset state.\r\n"); 
         printf("%s",obuf);
	 fflush(stdout); 
         if( smwpptr->traceflg)
               trace_it(&traceh, tracebufadr, 'W', obuf);
	 continue;
      }                      
      /*===================================================================
       * process HELO command 
       * If hostname is missing and bogus helo is allowed then echo real 
       * peer host name else send error message.  If hostname is provided
       * check for HELO Buffer overflow and return error message
       * if found.
       *==================================================================*/
      if(!strcasecmp(cmdp,"HELO")) 
      {
         char	*hnp;			/* hostname ptr       		*/ 
	 int	hnln; 			/* hostname length 	        */
         if((hnp = strtok((char *)0,"\r\n")) == (char *)0) 
         { 
	    if(smwpptr->bogushelo)
	       sprintf(obuf,"250 Hello (%s) charmed, Im sure.\r\n",PeerHost); 
	    else
	    {  
	       syslog(LOG_NOTICE,"Bogus HELO - no hostname after HELO." 
	                        " src host %s (%s).",
	       			PeerHost,PeerIP); 
	       sprintf(obuf,"501 Bogus Helo - no hostname after HELO.\r\n");  
	    }  
         } 
         else 
         {
	    if((hnln = strlen(hnp)) > MAXADDRLN )
	    {
	       syslog(LOG_ALERT,"Security Alert: Host name length '%d' after"
	                        " HELO exceeds maximum value %d. '%.120s...'",
	       			hnln,MAXADDRLN,hnp); 
	       sprintf(obuf,"501 Invalid hostname following HELO\r\n"); 			
	    } else
	    {
               sprintf(obuf,"250 %s Hello %s (%s)\r\n",
			lochostname,hnp,PeerHost);
	    } 		 
         }
	 printf("%s",obuf);
         fflush(stdout);  
 	 if( smwpptr->traceflg)
            trace_it(&traceh, tracebufadr, 'W', obuf);
         continue;
      } 
     

      /*===============================================================
       * process VERB command
       *==============================================================*/
      if(!strcasecmp(cmdp,"VERB")) 
      {
         sprintf(obuf,"200 Verbose request acknowledged. Still using "
	 	      "concise mode.\r\n");
         printf("%s",obuf);
	 fflush(stdout); 
         if( smwpptr->traceflg)
            trace_it(&traceh, tracebufadr, 'W', obuf);
         continue;
      }
      /*================================================================
       * process ONEX command
       *==============================================================*/  
      if(!strcasecmp(cmdp,"ONEX")) 
      {
         sprintf(obuf,"200 Only one transaction is allowed.\r\n");
	 printf("%s",obuf);
         fflush(stdout); 
	 if( smwpptr->traceflg)
            trace_it(&traceh, tracebufadr, 'W', obuf);     
         continue;
      }
      /*================================================================
       * process DEBUG command
       *===============================================================*/
      if(!strcasecmp(cmdp,"DEBUG")) 
      {
         sprintf(obuf,"200 Debug is NOT supported! \r\n"); 
	 printf("%s",obuf);
         fflush(stdout);
	 if( smwpptr->traceflg)
            trace_it(&traceh, tracebufadr, 'W', obuf);
         continue;
      }

      sprintf(obuf,"500 The command is not known. \r\n"); 
      printf("%s",obuf);
      fflush(stdout);
      if( smwpptr->traceflg)
            trace_it(&traceh, tracebufadr, 'W', obuf);
   }                            
theend:
   if( (smwpptr->traceflg == 2) || ( smwpptr->traceflg == 1 && rcode != 0 ) )
   {
      trace_dump(&traceh, tracebufadr, smwpptr);
   }
   closelog();
   exit(rcode);
}

/*============================================================================
 *  
 * function:	create temporaray file with random name to hold mail. 
 *		open file and return FILE descriptor. 
 *
 * output:	FILE descriptor
 *		NULL makstemp or open failed
 *===========================================================================*/
static FILE *smwrapopen(char rnamebuf[])
{
   FILE		*ret;
   int		fd;
   /*======================================================================== 
    * Create template for mkstemp routine then create file with random name. 
    *=======================================================================*/
   strcpy(rnamebuf,"smwrapXXXXXX");
   if((fd = mkstemp(rnamebuf)) < 0) 
   {
      return((FILE *)0);            /* file creation failed  */
   }                                                          
   chmod(rnamebuf,0600);            /* set intial filemode:   -rw------- */
   if((ret = fdopen(fd,"w")) == (FILE *)0)
   {
      return((FILE *)0);
   } 
   if((tempfile = (char *) malloc(strlen(rnamebuf) + 1)) == (char *)0) 
   {
      syslog(LOG_CRIT,"Error: out of memory: %m"); 
      unlink(rnamebuf);
      exit(1);
   }
   strcpy(tempfile,rnamebuf);
   return(ret);
} 

/*============================================================================
 *
 *
 *==========================================================================*/
int trace_dump( struct tracehdr *th, struct trace  *tracebufadr, struct SMWRAP *smwpptr )
{  
   int  j;
   int	k;                                        
   char traceout[120];
   char tracefile[256];
   int		fh;			/* file handle 		*/  
   FILE		*fd;                  	/* tracefile fd         */
   char		rnamebuf[512];
   struct trace *trptr;  
#ifdef DONTDOIT   
   /* create template for mkstemp routine */
   strcpy(tracefile,"traceXXXXXX");
  if((fh = mkstemp(tracefile)) < 0) 
   {  
      syslog(LOG_ERR,"Error: create failed for tracefile %s. strerror(%s)",
                         tracefile, strerror(errno));
      return(-1);  
   }
   
   if((fd = fdopen(fh,"w")) == (FILE *)0)
   {
      syslog(LOG_ERR,"Error: open failed for tracefile %s. strerror(%s)",
                         tracefile, strerror(errno));
      return(-1);    
   }
#endif
   if((fd = fopen("trace.smwrap","a")) == (FILE *)0)
   {
      syslog(LOG_ERR,"Error: open failed for tracefile %s. strerror(%s)",
                         tracefile, strerror(errno));
      return(-1);    
   }  
   /* print trace header      */
   k = fprintf(fd,"%s\n\n",th->title);
   /* print every trace entry */
   trptr = tracebufadr;
   for( j = 0; j < MAXTRACE; j++)
   {   
      if( trptr->sr == '-' ) break; 
      k = fprintf(fd,"%d %c:%.94s\n",trptr->entryno, trptr->sr, trptr->entry);
      trptr++;
   }
   fprintf(fd,"\n");  
   return(0);  
}

/*============================================================================
 * function: trace_init 
 *
 * inputs:
 *
 * outputs:  
 *	returns address of trace buffer if ok  
 *		0 on error  
 *===========================================================================*/
struct trace *trace_init(struct tracehdr *th)
{
   int	i; 
   struct trace *tracebufadr; 
   struct trace *trptr;
   tracebufadr  = (struct trace *) malloc( sizeof(struct trace)*(MAXTRACE + 1) );
   trptr = tracebufadr;
   th->recno = 0;
   th->tindx = 0;
   strncpy(th->title,"smwrap trace output",25);
   for( i =0; i< MAXTRACE;i++)
   {
      trptr->entryno  = 0;
      trptr->sr       = '-';	    
      trptr->entry[0] = '\0'; 
      trptr++;  
   }
   return(tracebufadr);
} 
/*============================================================================
 * function: trace_it 
 *
 * inputs:
 *
 * outputs:  
 *	returns 0	OK 
 *		1       error 
 *===========================================================================*/
int trace_it(struct tracehdr *th, struct trace tracea[],char sr, char *obuf)
{
   tracea[th->tindx].entryno  = th->recno++;
   tracea[th->tindx].sr       = sr;	    
   strncpy(tracea[th->tindx].entry,obuf,94);   
   th->tindx++;
   if(th->tindx > (MAXTRACE -1) )  th->tindx  = 0;
   return(0);
}    
/*===========================================================================
 * 		get string without CRLF 
 * 
 * inputs: 	pointer to buffer
 *		size of buffer 
 *		file ptr
 * function:
 * 		read the input until EOF or buffer is filled.   
 *
 * outputs:
 *		updated buffer
 *		If EOF return -1 else return buffer string length.
 *===========================================================================*/
extern int getswocrlf(char* buf, int siz, FILE *fp)
{
   int	i;
   char	*cp = buf;

   if(ferror(fp) || feof(fp))
      return(-1); 
   /*===================================================
    * loop until buffer is full or CRLF is encountered	
    *==================================================*/
   while(--siz) 
   {
      if((i= fgetc(fp)) == EOF) 
      {
         if( cp == buf)
         {
	    return(-1);
         }
         break;
      }
      if(i == '\n') 
      {
         /* back up over carriage returns */
         if( cp > buf && *(cp - 1) == '\r')
	    cp--;
         break;
      }
      *cp++ = i;
   }
   *cp = '\0';        	/* terminate string */
   livetim = 1;
   i =  ((int)cp  - (int) (&buf[0]));
   return(i);
}
/*============================================================================
 *
 *===========================================================================*/
int filesize(int fd)
{
   struct	stat	sbuf;

   if(fstat(fd,&sbuf))
      return(-1);
   return(sbuf.st_size);
}
/*============================================================================
 * print usage message
 *===========================================================================*/
int usage(void)
{
   printf("\nusage: smwrap smwrap -D nnnn\n\n");
   printf("\n          where:\n");
   printf("                    nnnn is the debugging port.\n");
   return(1);
}     
#ifdef DEBUG
/*======================================================================
 * debugbind
 * input: 	debug port number
 *======================================================================*/
void debugbind(unsigned short debugport)
{
   struct	sockaddr_in	mya;
   int	x;
   int	nread;

   if((x = socket(AF_INET,SOCK_STREAM,0)) < 0) 
   {
      perror("socket");
      exit(1);
   }
   mya.sin_family = AF_INET;
   bzero(&mya.sin_addr,sizeof(mya.sin_addr));
   mya.sin_port = htons(debugport);

   if(bind(x,(struct sockaddr *)&mya,sizeof(mya))) 
   {
      perror("bind");
      exit(1);
   }
   if(listen(x,1) < 0) 
   {
      perror("listen");
      exit(1);
   }
   if((nread = accept(x,0,0)) < 0) 
   {
      perror("accept");
      exit(1);
   }
   close(0);
   dup(nread);
   close(1);
   dup(nread);
}
#endif

