/*============================================================================
 * filename:		smwrap/transend.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=============================================================================
 *
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.
 * All rights reserved.   
 *
 *
 *
 * Date written:	July       5, 1995 
 * Last modified:	February   8, 2000 - added LPL
 * 
 * functions:   This file contains the following functions:
 *
 *  transend()	Translates internal sender name and address to an external
 *		e-mail address using the reverse aliases DB
 *
 * transcc()	Translates internal e-mail address found in the Cc: list to 
 *		external names and leaves external e-mail anmes alone.
 *

 *============================================================================*/
#include	<sys/types.h> 
#include	<stdio.h>
#include	<string.h>
#include	<syslog.h>
#include	<ctype.h> 
#include	"firewall.h"
#include	"smwrap.h"
/*=========================================================
 * function prototypes
 *========================================================*/    
int getaliases(char *keyptr, char *dataptr, char *dbname);
/*============================================================================
 * transend()	Translate sender e-mail address to an external name.         
 *
 * inputs:	character pointer to sender name and address 
 *		character pointer to array to hold external name
 *		character pointer to local host name 
 *		translate on|off switch   
 *
 * function:	Translate internal sender name and address to an external
 *              e-mail address using the reverse aliases DB. 
 *
 *		If sender name is found in the reverse aliases table then
 *			substitute external name. If the external name is
 *			not fully qualified (name@host.domain) then 
 *			concatenate gateway host name to username.
 *			Return to smwrap.
 *		Repeat the preceeding process process without the domain name.
 *		Repeat the preceeding process without the hostname. This allows
 *		internal users to send mail from any interanl host and always
 *		get the same external e-mail address. 
 *		  
 *		Check to see if the internal client has used an external
 *		e-mail address by searching the external alaises DB.   		
 * 
 *
 *	All names are reduced to lower case.
 *
 * outputs: 
 *
 *   If translation worked:		        
 *	returns	 0  &
 *      external e-mail name in area pointed to by aka. 
 *   else	
 *============================================================================*/
int transend(char *name, char *aka, char *lochostname, char translate)  
{ 
#define KEYSIZE 200
   static	char	cprght[] = 
  	  	"(C) COPYRIGHT Freemont Avenue Software, Inc. 1995-2000 "
  	  	"transend, V 1.0, 00/01/21";
   int  akalen;
   int	rc; 
   char *cp1;
   char	*cp2;
   char	*cp3;
   char *cp4; 
   char *token1; 
   char *token2;
   char *token3;
   char *akap1;
   char key[KEYSIZE+1]; 
   char key2[KEYSIZE+1];  
   char DomainName[MAXHOSTLN+1];		/* added 8/01/96        */
   char username[MAXHOSTLN+1];    		/* added 8/01/96	*/

   key[sizeof(key)-1] = '\0';
   /*========================================================================
    * First check for Null address used for notification messages, ala
    * RFC 821 section 3.6. 
    *========================================================================*/
   if( strcmp(name,"<>") == 0)
   {
      strcpy(aka,"<>");
      return(0);
   }
   strncpy(key,name,KEYSIZE);
   if( key[0] == '<' ) key[0] = ' '; 
   token1 = strtok(key," <>");
   akap1 = aka + 1;
   /**aka = '<';*/
   strcpy(aka,"<");   			   
   /*=========================================================================
    * Check for DEC special formats  
    * <"gw::intermediatehost::name"@host.domain>
    *========================================================================*/
   if( *token1 == '\"')
   {  
      /* find last occurence of ':' */
      if( (cp3 = strrchr(token1,':')) == NULL) 
         goto notDEC;
      if( (cp4 = strchr(cp3+1,'\"')) == NULL)
         goto notDEC;  
      if( strncmp(cp4,"\"@",2) != 0)
         goto notDEC;
      /* looks like a DEC format	*/
      token1 = strtok(cp3,":\""); 	/* get name preceeding @ */
      strcpy(key2,token1);
      strcat(key2,cp4+1);
      strcpy(key,key2); 
      token1 = key;
      goto GetAliases;
   }  
notDEC: 
   /*====================================================
    * Check for the "%" Hack
    * take jim%domain1@gateway and convert to jim@domain
    *====================================================*/  
   if( (cp3 = strchr(token1,'%')) == NULL)
      goto notHack; 
   if( (cp4 = strchr(cp3,'@')) == NULL)
      goto notHack; 
   token1 = strtok(key," %");
   strcpy(key2,token1);
   strcat(key2,"@");
   token1 = strtok(NULL," %@");
   strcat(key2,token1);
   strcpy(key,key2);
   token1 = key; 
notHack: 
GetAliases:  
   /*======================================================
    * first look for domain transfer record  
    *   example:    @lsli.com    @lsli2
    *=====================================================*/  
   strcpy(username,token1); 
   token2 = strchr(username,'@');
   if( token2 == NULL)
   {
      syslog(LOG_ERR,"Error: invalid e-mail address '%s'",
	 	 name); 
      goto fqname;
   }
   *token2 = '\0';
   token2 = strchr(token1,'@'); 
   if( ( rc = getaliases(token2, DomainName, "raliasesDB")) == 0)
   {              
      if( (akalen = strlen(DomainName)) > MAXNAME-1)
      {  
	 syslog(LOG_ERR,"Error: aliases name length %d is > %d for %s.",
	 	        akalen,MAXNAME,token1);
         goto badrtn;	 
      }
      strcat(aka,username);              
      strcat(aka,DomainName);
      strcat(aka,">");	
      return(0);
   } 
   /*========================================================================
    * If sender does not come from local domains and translate is off
    * then do not translate the senders e-mail address as it could be 
    * extrenal mail thatn is being forwarded to an outside host.
    *=======================================================================*/
   if( translate == TRANSLATEOFF)
   {
      strcpy(aka,name);  
      return(0);
   }  
fqname:
   /*=========================================================================
    * second try fully qualified name.
    *========================================================================*/
   if( ( rc = getaliases(token1, akap1, "raliasesDB")) == 0)
   {              
      if( (akalen = strlen(akap1)) > KEYSIZE)
      {  
	 syslog(LOG_ERR,"Error: reverse aliases name length %d is > %d '%s'.",
	 	        akalen,KEYSIZE,token1);
         goto badrtn;	 
      } 
      /*======================================================================
       * If external name already has the form "name@host.domain" then use it.
       * That is it has a '@' one or two periods.  Else concatenate the 
       * gateway name to the user name.   
       *=====================================================================*/
      if( ( cp1 = strchr(akap1,'@') ) != NULL)
      {
         if( (cp2 = strchr(cp1,'.') ) == NULL)
         {
            syslog(LOG_ERR,"Error: invalid entry in reverse aliases DB. %s",
				akap1);
             strcat(aka,"@");
             strcat(aka,lochostname);
             strcat(aka,">");	
             return(0);
         }
         strcat(aka,">");   
         return(0);
      }   
      strcat(aka,"@");    
      strcat(aka,lochostname);
      strcat(aka,">");	
      return(0);
   } 
   /*===================================================
    * third try name without domain name 
    *==================================================*/ 
   token1 = strtok(key," <>");
   if((token2 = strchr(key,'@')) == NULL)
   {
      syslog(LOG_ERR,"Error: e-mail address missing '@', %s",token1);
      goto badrtn; /* ??? */
   }	
   if( (token3 = strchr(token2,'.')) != NULL)
      *token3 = '\0'; 
   if( ( rc = getaliases(token1, akap1, "raliasesDB")) == 0)
   { 
      if( (akalen = strlen(akap1)) > KEYSIZE)
      {  
	 syslog(LOG_ERR,"Error: aliases name length %d is > %d for %s.",
	 	 	akalen,KEYSIZE,token1);
         goto badrtn;	 
      }
      /*======================================================================
       * If external name already has the form "name@host.domain" then use it.
       * That is it has a '@' one or two periods.  Else concatenate the 
       * gateway name to the user name.   
       *=====================================================================*/
      if( ( cp1 = strchr(akap1,'@') ) != NULL)
      {
         if( (cp2 = strchr(cp1,'.') ) == NULL)
         {
            syslog(LOG_ERR,"Error: invalid entry in reverse aliases DB. %s",
				akap1);
             strcat(aka,"@");
             strcat(aka,lochostname);
             strcat(aka,">");	
             return(0);
         } 
         strcat(aka,">");   
         return(0);
      }               
      strcat(aka,"@");
      strcat(aka,lochostname);
      strcat(aka,">");	
      return(0);
   } 
   token1 = strtok(key,"@ <>");
   /*========================================================================
    * fourth try name without @host.domain name 
    *=======================================================================*/
   if( ( rc = getaliases(token1, akap1, "raliasesDB")) == 0)
   { 
      if( (akalen = strlen(akap1)) > KEYSIZE)
      {  
	 syslog(LOG_ERR,"Error: aliases name length %d is > %d for %s.",
	 	 	akalen,KEYSIZE,token1);
         goto badrtn;	 
      }
      /*======================================================================
       * If external name already has the form "name@host.domain" then use it.
       * That is it has a '@' one or two periods.  Else concatenate the 
       * gateway name to the user name.   
       *=====================================================================*/
      if( ( cp1 = strchr(akap1,'@') ) != NULL)
      {
         if( (cp2 = strchr(cp1,'.') ) == NULL)
         {
            syslog(LOG_ERR,"Error: invalid entry in reverse aliases DB. %s",
				akap1);
             strcat(aka,"@");
             strcat(aka,lochostname);
             strcat(aka,">");	
             return(0);
         } 
         strcat(aka,">");   
         return(0);
      }          
      strcat(aka,"@");
      strcat(aka,lochostname);
      strcat(aka,">");	
      return(0);
   }  
   /*=======================================================================
    * Check to see if the protected client has sent us an external e-mail
    * address that is to be translated by this firewall. 
    *======================================================================*/ 
   strncpy(key,name,KEYSIZE);
   if( key[0] == '<' ) key[0] = ' '; 
   token1 = strtok(key," <>");
   if( ( rc = getaliases(token1, akap1, "aliasesDB")) == 0)
   {  
      strcpy(aka,"<");            
      strcat(aka,token1);
      /*======================================================================
       * If external name already has the form "name@host.domain" then use it.
       * That is it has a '@' one or two periods.  Else concatenate the 
       * gateway name to the user name.   
       *=====================================================================*/
      if( ( cp1 = strchr(aka,'@') ) != NULL)
      {
         if( (cp2 = strchr(cp1,'.') ) == NULL)
         {
            syslog(LOG_ERR,"partially qualified external name '%s' appended "
			   "with @%s.",
				aka,lochostname);
             strcat(aka,"@");
             strcat(aka,lochostname);
             strcat(aka,">");	
             return(0);
         }
         strcat(aka,">");   
         return(0);
      }  
      /* Need to concatenate firewall hostname. */ 
      strcat(aka,"@");    
      strcat(aka,lochostname);
      strcat(aka,">");	
      return(0);
   } 

badrtn:
   *aka = '\0';		/* lookup failed so set aka to NULL string */ 
   return(-1);
} 
/*============================================================================
 * function:	transcc()
 *		Translate internal e-mail names found in the Cc: list to
 *		external names using the reverse aliases DB, while leaving 
 *		external e-mail address unchanged.     
 *                        
 *  example:
 *
 *  jim         	-> jim@lsli.com
 *  jim@lsli2   	-> jim@lsli.com  
 *  jim@lsli2.lsli.com  -> jim@lsli.com
 *  jim@else.where.com 	-> jim@else.where.com
 *
 *
 * inputs:	character pointer to input buffer
 *		character pointer to local host name of the firewall. 
 *
 * outputs:	Updated buffer with translated and untranslated e-mail
 *		addresses.    
 *
 *===========================================================================*/
int transcc(char *buf, char *lochostname)
{  
   char actualname[MAXNAME];   
   char	*cptr;            	/* ptr to current buf position */
   char *cptr1;
   char *cptr2; 
   char *cptr3;
   char *cptr4;                                    
   char *cptr5;
   char DomainName[MAXNAME];		/* added 8/13/96        */ 
   char emaddr[MAXNAME];    		/* e-mail address	*/
   int	endingcomma	= 0;
   int	i; 
   char newline[MAXLINELN+100];	/* used to build a new line	*/
   char	newname[MAXNAME]; 
   int	numadr	= 0;		/* number of addresses found	*/
   int	rc;          
   char tempname[MAXNAME];
   char	username[MAXNAME];  	/* user name extracted from emaddr     */
   actualname[MAXNAME-1] = '\0';
   emaddr[MAXNAME-1]     = '\0';/* e-mail address of form jim@lsli.com */
   memset(&newline[0],0,10);
   newline[sizeof(newline)-1] = '\0';

   if( buf[strlen(buf)-1] == ',')
   { 
      buf[strlen(buf)-1] = '\0';
      endingcomma = 1; 
   }
   if( (strncasecmp(buf,"Cc:",3) == 0) || (strncasecmp(buf,"To:",3) == 0) ) 
   {
      strncpy(newline,buf,3);
      strcat(newline," "); 
      cptr = &buf[3];
   }
   else  
   {
      strcpy(newline,"    ");   	/* Indent line to match Cc: */ 
      cptr = &buf[0];
   } 
   
   /*=============================================================
    * examine every entry in list 
    *============================================================*/
   while( (cptr != NULL) &&  ( cptr < &buf[MAXLINELN-1]  ) && (*cptr != '\0') )
   {
      actualname[0] = '\0';  
      username[0]   = '\0';
      emaddr[0]     = '\0';
      /*====================================================================
       * If there is an actual name in front of the email address then
       * preserve it:    "Jim Livermore" <jim@lsli.com> 
       * Also check for: "Jim Livermore" jim@lsli.com
       *===================================================================*/
      if(( cptr2 = strchr(cptr,'<')) != NULL)
      { 
         /* skip over spaces */  
         cptr3 = cptr;    
         while( isspace(*cptr3) )
            cptr3++;  
         /*====================================================================
          * MS OutLook inserts real e-mail addresses between double quotes
          * so we must skip over quoted data to prevent translation.  ie.
          * don't translate "<jim@lsli.com>" and translate <jim@lsli2.lsli.com>.
          *===================================================================*/        
         if( *cptr3 == '\"' )
         {
            if( (cptr4 = strchr(cptr3+1,'\"')) == NULL)
            {
               syslog(LOG_ERR,"Error: unbalanced '\"'in e-mail address, "
	       		      "'%.80s'",
	       		      buf);
               goto failed;
            }     
            if( (cptr2 = strchr(cptr4,'<')) == NULL)
               cptr2 = cptr4 + 1;   
         }

         strncpy(actualname,cptr3,MAXNAME-1);     /* save actual name     */ 
         strncpy(emaddr,cptr2+1,MAXNAME-1);     /* save email addr without ','*/
         if(( cptr = strchr(cptr2,'>')) == NULL)
         {
            syslog(LOG_ERR,"Error: missing '>' in e-mail address, '%.80s'",
	    		   buf);
            goto failed;
         }    
         cptr++;  
         if( *cptr == ',')
            cptr++;
         if( actualname[0] == '\"' )
         {
            cptr2 = strchr(&actualname[1],'\"') + 1;     
         }  
         else
         {
	    cptr2  = strchr(actualname,'<');	/* truncate actual name	     */   
         }
         if(cptr2)
            *cptr2 = '\0';
         else
            actualname[0] = '\0';
         strcat(actualname," ");                    
         cptr2  = strchr(emaddr,'>');		/* strip '>' from email addr */ 
         *cptr2 = '\0'; 

      } 
      /*=====================================================================
       * process  "actual name" name@domain      - without <> characters   
       *====================================================================*/
      else
      {
         if( (cptr3 = strchr(cptr,'"')) !=NULL)
         {
            if((cptr4 = strchr(cptr3+1,'"')) != NULL)           
            {  
               /*=========================================================== 
		* If There is no '@' then treat rest of line as a comment.
        	*==========================================================*/
               if(( cptr5 = strchr(cptr,'@')) == NULL) 
               {
                  strncpy(actualname,cptr3,MAXNAME-1); /* save actual name    */
                  emaddr[0] = '\0';
                  if( newline[strlen(newline)-1] == '>')
                     strcat(newline,", ");  
                  strcat(newline,actualname); 
                  strcpy(buf,newline);
                  return(0); 
               }  
               /*==================================================== 
		* If e-mail address follows actual name then save 
		* actual name and concatenate with a blank.
		*===================================================*/ 
               if( cptr5 < cptr4 )
               {  
                  if( (cptr4 - cptr3) < MAXNAME-1 )
                  {
                     strncpy(actualname,cptr3,cptr4-cptr3+1);
                     actualname[cptr4-cptr3+1] = '\0'; 
                  }
                  else  
                     strncpy(actualname,cptr3,MAXNAME-1); /* save actual name*/
                  emaddr[0] = '\0';
                  if( newline[strlen(newline)-1] == '>')
                     strcat(newline,", ");  
                  strcat(newline,actualname); 
                  cptr = cptr4+1; 
                  goto getnextaddress;
               }
               else
               {
                  strncpy(actualname,cptr3,MAXNAME-1); /* save actual name    */ 
                  cptr2 = &actualname[1];
                  cptr3 = strchr(cptr2,'"');
                  strcpy(cptr3+1," ");		
                  cptr = cptr4 + 1;  
 
                  while( isspace(*cptr) )
                     cptr++;  
                  strncpy(emaddr,cptr,MAXNAME-1);     
                                                
                  cptr = cptr5 + 1;		/* skip over '@' 	*/  
                  while( isspace(*cptr) )	/* skip over blanks     */ 
                     cptr++;	             	
                  while( !isspace(*cptr) )    /* skip over host.domain name */ 
                  {
                     if( *cptr == '\0' )
                        goto EndOfLine1;
                     cptr++; 
                  }                                                 
                  while( isspace(*cptr) )     /* skip blanks following host */
                     cptr++; 
                  if( *cptr == ',')
                     cptr++;
EndOfLine1:                                                 
                  cptr1 =  strchr(emaddr,' ');
                  cptr2 = strchr(emaddr,'@');                 
                  if( cptr1 != NULL && cptr2 != NULL && (cptr1 < cptr2 ) )
                  {
                     *cptr1 = '@';
                     cptr1++;           
                     cptr2++;
                     while( isspace(*cptr2) )
                        cptr2++;
	             strcpy(cptr1,cptr2);   
                  } 
                  strtok(emaddr," ,");                   

               }
 
            } 
            else
            {
               syslog(LOG_ERR,"Error: invalid address '%.80s'",buf);
               goto failed;      
            }
         }         
         /*===========================================================
          * There is no actual name within '"'s and there are no '<'
          * or '>' characters. 
          * Check for trailing actual name between '(' and ')'. 	
          *==========================================================*/
         else		 
         {
            while( isspace(*cptr) )
               cptr++;  
            if( *cptr == '(' )
            {   
               strcat(newline," ");
               strcat(newline,cptr);
               strcpy(buf,newline); 
               if( endingcomma)
                  strcat(buf,",");
               return(0);
            }
            strncpy(emaddr,cptr,MAXNAME-1);  
            cptr = strchr(cptr,'@');	/* point at first @ */
            if( cptr == NULL)
            {
               syslog(LOG_NOTICE,"invalid e-mail address '%s' in Cc: line.",
                                 emaddr); 

               goto getnextaddress;
            }
            cptr++;
            while( isspace(*cptr) )     /* skip over any blanks after @ */ 
               cptr++;	             	
            while( !isspace(*cptr) )    /* skip over host.domain name */ 
            {
               if( *cptr == '\0' )
                  goto EndOfLine2;
               if( *cptr == ',' )
               {
                  cptr++;
                  goto EndOfLine2; 
               } 
               cptr++; 
            }                                                 
            while( isspace(*cptr) )     /* skip over blanks following host */
               cptr++; 
            if( *cptr == ',')
               cptr++;
EndOfLine2:
            cptr1 =  strchr(emaddr,' ');
            cptr2 = strchr(emaddr,'@');                 
            if( cptr1 != NULL && cptr2 != NULL && (cptr1 < cptr2 ) )
            {
               *cptr1 = '@';
               cptr1++;           
               cptr2++;
               while( isspace(*cptr2) )
                  cptr2++;
	       strcpy(cptr1,cptr2);   
            } 
            strtok(emaddr," ,"); 
 
         }
      }
      if( (numadr > 0) && (newline[strlen(newline)-1] != '\"') )
         strcat(newline,", "); 
      else
         strcat(newline," "); 
      /*=============================================================
       * First look for domain transfer record    
       *  example:	@lsli.com	@mailserver.lsli.com
       *============================================================*/
      strcpy(username,emaddr);
      cptr1 = strchr(username,'@');
      if( cptr1 == NULL)
      {
         syslog(LOG_NOTICE,"invalid e-mail address '%s' in Cc: line.",
                        username); 
         strcat(newline,actualname);
         strcat(newline,"<");
         strcat(newline,username);
         strcat(newline,">");
         goto getnextaddress;
      }
      *cptr1 = '\0';		/* terminate username	*/
      cptr2 = strchr(emaddr,'@');
      if( ( rc = getaliases(cptr2, DomainName, "raliasesDB")) == 0)
      { 
         strcat(newline,actualname); 
         strcat(newline,"<"); 
         strcat(newline,username); 
         strcat(newline,DomainName);  
         strcat(newline,">"); 
         goto getnextaddress;
      }
      newname[0] = '\0';	 	 

      /*======================================================================
       * Second, since there was no domain entry try e-mail name as is.
       * example:  jim@mailserver.lsli.com
       *=====================================================================*/
      if( ( rc = getaliases(emaddr, newname, "raliasesDB")) == 0)
      {                          
         if( strlen(newline) + strlen(newname) + strlen(actualname) < MAXLINELN)
         {   
             
            strcat(newline,actualname);  
            strcat(newline,"<");   
            strcat(newline,newname); 
            if( ( cptr1 = strchr(newname,'@') ) == NULL)  
            {   
               strcat(newline,"@");
               strcat(newline,lochostname);  
            } 
            strcat(newline,">");                   
         }
         else
            goto failed;     
      }  
      /*=======================================================================
       * Third, if e-mail address is of the form  name@host
       * ( ie no domain is included) then try name without hostname. 
       *======================================================================*/   
      else if( (( cptr2 = strchr(emaddr,'@')) != NULL))
      {        
         if(( cptr3 = strchr(emaddr,'.')) == NULL)  
         {  
            strcpy(tempname,emaddr);
            cptr4 = strchr(tempname,'@'); 
            *cptr4 = '\0';                             
            if( ( rc = getaliases(tempname, newname, "raliasesDB")) == 0)
            {                          
               if( strlen(newline) + strlen(newname) < MAXLINELN)
               {  
                  strcat(newline,actualname);
                  strcat(newline,"<");  
                  strcat(newline,newname); 
                  if( ( cptr1 = strchr(newname,'@') ) == NULL)  
                  {   
                     strcat(newline,"@");
                     strcat(newline,lochostname);  
                  } 
                  strcat(newline,">");      
               }
               else
	          goto failed;
            }
            else
            {                       
               /* If name not found use name@domain.name  */   
               strcat(newline,actualname);
               strcat(newline,"<"); 
               strcat(newline,tempname);
               strcat(newline,"@");
               strcat(newline,lochostname);  
               strcat(newline,">");
            }
         }
         /*==============================================================
          * Else if not found and e-mail address is of the form 
          * name@host.domain.toplevel  or
          * name@host.domain
          *=============================================================*/  
         else
         {   
            /*==================================================
             * Only one '.' was found  so use as is. 
             *=================================================*/ 
            if(( cptr4 = strchr(cptr3+1,'.')) == NULL)
            {           
               strcat(newline,actualname); 
               strcat(newline,"<"); 
               strcat(newline,emaddr); 
               strcat(newline,">"); 
            } 
            /*================================================================
             * Two '.'s were found so we must check to see if domain.toplevel
             * matchs firewall's domain name. If it does then translate name.
             *===============================================================*/         
            else
            {
               if( strcmp(cptr3+1,lochostname) == 0)
               {
                  *cptr2 = '\0'; 
                  strcat(newline,actualname); 
                  strcat(newline,"<");  
                  if( getaliases(username,tempname,"raliasesDB") == 0)
                     strcpy(username,tempname);   
                  strcat(newline, username);
                  if( strchr(username,'@') == NULL )
                  {
                     strcat(newline,"@");
                     strcat(newline,lochostname);  
                  }
                  strcat(newline,">");
               }  
               else
               {  
                  strcat(newline,actualname);
                  strcat(newline,"<"); 
                  strcat(newline,emaddr);
                  strcat(newline,">");
              }
            }
         }       
      }
      else
      /*=============================================================  
      * Use the name as is, since it must be an external name.  
       *============================================================*/
      {
         strcat(newline,cptr); 
         /*strcat(newline,", ");*/ 
      }                    
getnextaddress:                   
      numadr++;
      /*cptr = strtok(NULL," ,\r\t\n"); */   /* get next token */
   }     
   if( endingcomma )
      strcat(newline,",");                            
   strcpy(buf,newline);

   return(0);            
failed:      
   syslog(LOG_ERR,"Error: Unable to translate %.3s names, mail sent as is.",
          buf);
   return(0);
}
