/*============================================================================
 * filename:		smwrap/tstpname.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=============================================================================
 *
 * written by:		R. James Livermore
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.
 * All rights reserved. 
 *
 * Date written:	March      2, 1995
 * Last modified:	February   8, 2000 - added LPL
 * 
 * inputs:	file descriptor for the socket 
 *		pointer to long hostname of remote host  
 *		pointer to area for remote host dot-decimal Internet Address 
 *              size of internet address area
 *
 * function:	test peer host 
 *       
 *      Get the IP address of the peer host. Do a DNS lookup by address to get
 *	the host name. Then do a lookup by name to see if the IP address
 *	matches.  If there is a miss match substitute "unknown" for the 
 *	host name, and issue a warning message in syslog.    
 *
 *	All names are reduced to lower case.  Output strings are truncated to
 *	maximum length passed as the third parameter.
 *
 * outputs: 
 *		Hostname of remote system         
 *		IP Address of remote peer in ASCII dot-decimal format	
 *
 *============================================================================*/

#include	<sys/types.h>
#include	<sys/socket.h>
#include	<netinet/in.h>
#include	<netdb.h>
#include	<syslog.h>
#include	<ctype.h>
#include	"firewall.h"
extern int h_errno;
static	char	unknown[] = "unknown"; 
/*=================================================================
 * function prototypes
 *==================================================================*/
struct hostent *Pgethostbyaddr(const char *addr, int len, int type); 
struct hostent *Pgethostbyname(const char *name); 
/*========================================================
 *
 *========================================================*/
int tstpname(int fd, char *rhostname, char *rhostIP, int z)  
{ 
   static	char	cprght[] = 
   "(C) COPYRIGHT Freemont Avenue Software, Inc. 1995-2000 "
   "tstpname, V 1.0 00/01/21";
   struct sockaddr_in	a;       	/* peer socket 	*/
   struct in_addr	*p_addr,*addr1;
   struct hostent	*p;
   int		y;
   int		eq = 0; 
   int	  errno1; 
   

   y = sizeof(a);
   /*===============================================
    * get the peer socket for the remote host
    *===============================================*/
   if(getpeername(fd,(struct sockaddr *)&a,(socklen_t *)&y) < 0)
      	return(1);
   /* convert Internet address to ASCII string in dot-decimal format */
   strncpy(rhostIP,inet_ntoa(a.sin_addr),z);
   rhostIP[z - 1] = '\0';  		/* guarentee end of string */
   /*==================================================================
    * Get remote host info using DNS pointed to by /etc/resolv.conf.
    * If gethostbyaddr fails try the internal DNS pointed to by
    * /etc/firewall/resolv.inside.conf. 
    *=================================================================*/
    p = gethostbyaddr((const char *)&a.sin_addr,sizeof(a.sin_addr),AF_INET); 
   errno1 = h_errno;    
   if( p == (struct hostent *) 0)
   { 
      p = Pgethostbyaddr((const char *)&a.sin_addr,sizeof(a.sin_addr),AF_INET);
   }
   if(p != (struct hostent *)0) 
   {
      char	*x;
      /* copy remote host logical name */
      strncpy(rhostname,p->h_name,z);
      rhostname[z - 1] = '\0';

      if(( p = gethostbyname(rhostname)) == (struct hostent *) 0)
      {
         p = Pgethostbyname(rhostname); 
      }
      if(p == (struct hostent *)0) 
      {
         syslog(LOG_NOTICE,"Host name lookup failed for %s (%s).",
                           rhostname,rhostIP);
	 goto badguy;
      }
      /*======================================================================
       * Check list of addresses from DNS if a match is not found then this
       * is a possible spoof. 	
       *=====================================================================*/
      addr1 = (struct in_addr *)*p->h_addr_list;
      while((p_addr = (struct in_addr *)*p->h_addr_list++) != (struct in_addr *)0) 
      {
         if(bcmp(p_addr,(char *)&a.sin_addr,p->h_length) == 0) 
          {
	    eq = 1;                	/* match found  */
	    break;
	 }
      }
      if(!eq) 
      {
	 syslog(LOG_NOTICE,"Security Alert: possible spoof %s (%s) != %s name"
			   "lookup mismatch", 	
                           rhostname,rhostIP,inet_ntoa(*addr1));
 	 goto badguy;
      }

#ifdef IP_OPTIONS
      {
         unsigned char	optbuf[1024 / 3];
	 unsigned char	*cp;
	 char		lbuf[1024];
	 char		*lp;
	 int		optsize;
	 int		ipproto;
	 struct protoent	*ip;

	 if((ip = getprotobyname("ip")) != (struct protoent *)0)
	    ipproto = ip->p_proto;
	 else
	    ipproto = IPPROTO_IP;
	 optsize = sizeof(optbuf);
	 if(getsockopt(0,ipproto,IP_OPTIONS,(char *)optbuf,
			(socklen_t *)&optsize) == 0 && optsize != 0) 
         {
	    lp = lbuf;
	    for(cp = optbuf; optsize > 0; cp++, optsize--, lp += 3)
	       sprintf(lp," %2.2x",*cp);
	    syslog(LOG_NOTICE,"Security Alert: Connection received using IP"
			      "	options (ignored):%s",lbuf);
 	    if(setsockopt(0,ipproto,IP_OPTIONS,(char *)0,optsize) != 0)              {
	       syslog(LOG_NOTICE,"setsockopt IP_OPTIONS NULL: %m");
	       goto badguy;
	    }
         }
      }
#endif
      /* convert to lower case */
      for(x = rhostname; *x != '\0'; x++)
         if(isupper(*x))
   	    *x = tolower(*x);
	 return(0);
   }
   /*================================================
    * cross examination failed. Set name to "unknown". 
    *===============================================*/
badguy: 
   syslog(LOG_NOTICE,"gethostbyaddr errno = %d, Pgethostbyaddr errno = %d",
			errno1, h_errno);
   strcpy(rhostname,unknown);
   return(0);
}
