/*============================================================================
 * filename:		smwrap/validrcpt.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=============================================================================
 *
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.
 * All rights reserved.  
 *
 * Date written:	June      26, 1995
 * Last modified:	February   8, 2000 - added LPL
 *			February  29, 2000 - added externalrelay function to
 *					     check for mail relay when using
 *					     multiple "@domain" aliases 
 *	
 *============================================================================*/

#include	<sys/types.h>
#include	<sys/file.h>
#include	<stdio.h>
#include	<string.h>
#include	<syslog.h>
#include	<ctype.h>
#include	<firewall.h>
#include	<smwrap.h>
/*=========================================================
 * function prototypes
 *========================================================*/    
int getaliases(char *keyptr, char *dataptr, char *dbname);
/*============================================================================ 
 * validrcpt
 * inputs:	character pointer to recipient name and address 
 *		character pointer to array to hold aliases name    
 *
 * function:	Determine if recipient is valid. Check the aliases Data 
 *              Base for valid name. Check for name with and without domain
 *		name. This routine is called when sender is on an external host.
 *
 *	All names are reduced to lower case.
 *
 * outputs: 
 *		        
 *	returns	 0 if ok
 *		-1 on error 	
 *      updated aliases name if ok else area is a null string
 *===========================================================================*/
int validrcpt(char *name, char *aka)  
{ 
   static	char	cprght[] = 
  	"(C) COPYRIGHT Freemont Avenue Software, Inc. 1995-2000 validrcpt,"
 	" V 1.0 00/01/21";
   int  akalen;
   char *akap1;
   int	rc;  
   char *token1;  
   char *token2;
   
   char key[MAXADDRLN];   
   char emailsvr[MAXADDRLN];
   char username[MAXADDRLN];
   strncpy(key,name,MAXADDRLN-1);
   key[MAXADDRLN-1] = '\0';
   if( key[0] == '<' ) key[0] = ' '; 
   token1 = strtok(key," <>");
   akap1 = aka + 1;
   *aka = '<';                                      
   /*=========================================================================
    * first look for domain record transfer record  
    *   example:    @lsli.com    @lsli2
    *========================================================================*/  
   strncpy(username,token1,MAXADDRLN-3); 
   username[MAXADDRLN-3] = '\0';
   token2 = strchr(username,'@'); 
   if( token2 == NULL)
   {
      return(-1);
   }
   *token2 = '\0';   				/* terminate username	*/
   token2 = strchr(token1,'@'); 
   if( ( rc = getaliases(token2, emailsvr, "aliasesDB")) == 0)
   {              
      if( (akalen = strlen(emailsvr) + strlen(username)) > MAXADDRLN-3)
      {  
	 syslog(LOG_ERR,"Error: aliases name for %s%s is longer than %d: %d.",
	 	 username,token1,MAXADDRLN-1,akalen);
         goto badrtn;	 
      }  
      strncpy(akap1,username,MAXADDRLN-2);
      strcat(aka,emailsvr);
      strcat(aka,">");	
      return(0);
   } 
   /*======================================================
    * next try fully qualified name.
    *=====================================================*/
   if( ( rc = getaliases(token1, akap1, "aliasesDB")) == 0)
   {              
      if( (akalen = strlen(akap1)) > MAXADDRLN-2)
      {  
	 syslog(LOG_ERR,"Error: aliases name for %s is longer than %d: %d.",
	 	 token1,MAXADDRLN-2,akalen);
         goto badrtn;	 
      }
      strcat(aka,">");	
      return(0);
   } 
   token1 = strtok(key,"@ <>");
   /*===================================================
    * now try name without @host.domain 
    *==================================================*/
   if( ( rc = getaliases(token1, akap1, "aliasesDB")) == 0)
   { 
      if( (akalen = strlen(akap1)) > MAXADDRLN-2)
      {  
	 syslog(LOG_ERR,"Error: aliases name for %s is longer than %d: %d.",
	 	 token1,MAXADDRLN-2,akalen);
         goto badrtn;	 
      }
      strcat(aka,">");	
      return(0);
   } 
badrtn:
   *aka = '\0';		/* lookup failed so set aka to NULL string */ 
   return(-1);
}  /* end of validrcpt */            

/*=============================================================================
 * internalrcpt
 * inputs:	character pointer to recipient name and address 
 *		character pointer to array to hold aliases name    
 *
 * function:	This function will translate a fully qualified e-mail 
 *		addresss if a match is found in the aliases data base. 
 *              This routine is called when sender is on a protected host.
 *		This function was designed to support multiple domain names
 *		for mail. Thus allowing one internal user to send mail to 
 *		another internal user using their external mail address. 
 *		
 *
 *	All names are reduced to lower case.
 *
 * outputs: 
 *		        
 *	returns	 0 if ok
 *		-1 on error 	
 *      updated aliases name if ok else area is a null string
 *
 *===========================================================================*/
int internalrcpt(char *name, char *aka)  
{ 
   int  akalen;
   char *akap1;
   int	rc;  
   char *token1;
   char *token2;
   char emailsvr[MAXADDRLN];
   char key[MAXADDRLN];     
   char username[MAXADDRLN];
   strncpy(key,name,MAXADDRLN-1);
   if( key[0] == '<' ) key[0] = ' '; 
   token1 = strtok(key," <>");
   akap1 = aka + 1;
   *aka = '<'; 
   /*======================================================
    * first look for domain record transfer record  
    *   example:    @lsli.com    @lsli2
    *=====================================================*/  
   strncpy(username,token1,MAXADDRLN-3); 
   username[MAXADDRLN-3] = '\0';
   token2 = strchr(username,'@'); 
   if( token2 == NULL)
   {
      return(-1);
   }
   *token2 = '\0';   				/* terminate username	*/
   token2 = strchr(token1,'@'); 
   if( ( rc = getaliases(token2, emailsvr, "aliasesDB")) == 0)
   {              
      if( (akalen = strlen(emailsvr) + strlen(username)) > MAXADDRLN-2)
      {  
	 syslog(LOG_ERR,"Error: aliases name for %s is longer than %d: %d.",
 	 	        token1,MAXADDRLN-2, akalen);
         goto badrtn;	 
      }  
      strcpy(akap1,username);
      strcat(aka,emailsvr);
      strcat(aka,">");	
      return(0);
   } 
   /*===================================================
    * Try fully qualified name.
    *==================================================*/
   if( ( rc = getaliases(token1, akap1, "aliasesDB")) == 0)
   {              
      if( (akalen = strlen(akap1)) > MAXADDRLN-2)
      {  
	 syslog(LOG_ERR,"Error: aliases name for %s is longer than %d: %d.",
	 	 token1, MAXADDRLN, akalen);
         *aka = '\0';
         return(-1);	 
      }
      strcat(aka,">");	
      return(0);
   } 
badrtn:   
   *aka = '\0';		/* lookup failed so set aka to NULL string */ 
   return(-1);
}
/*=============================================================================
 * external relay
 * inputs:	character pointer to recipient name and address     
 *
 * function:	This function checks for external third party mail relay.
 *		If the email has the form "user%domain@localdomain" then
 *		reject it.   
 *
 * outputs: 
 *		        
 *	returns	 0 if ok
 *		 1 if external relay found
 *		-1 on error 	
 *      
 *
 *===========================================================================*/
int externalrelay(char *name)  
{ 
   int	rc;  
   char *token1;
   char *token2;  
   if( (token1 = strchr(name,'%')) == NULL)
      return(0);
   if( (token2 = strchr(name,'@')) == NULL)
      return(-1);
   if( token1 < token2 )
      return(1);
   else
      return(0);
}
