/*============================================================================
 * filename:	smwrapd/smwrapd.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=============================================================================
 *
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.
 * All rights reserved.   
 *   
 *
 * Date written:	March      8, 1995
 * Last modified:	February   8, 2000
 *
 * function:		secure mail wrapper daemon	
 *       
 *      The smwrapd program sits in a loop scanning the queue directory defined
 *	in the /etc/firewall/smwrap.conf file.  When smwrapd finds a file in
 *	the queue directory it locks the file, forks a copy of itself with the
 *	name of the file and clears the lock on the file.  
 *
 *	The child process is invoked with the filename, which allows it to 
 *	process the file through the sendmail program for delivery to the 
 *	receipent host. When the delivery process exits, it deletes the 
 *	work file from the queue directory.     
 *	
 *      The process is designed to minimize the time spent waiting for 
 *	processes to complete and to eliminate situations where a system crash
 *	might cause a message NOT to be delivered.
 *	
 *
 *
 *============================================================================*/
#include	<stdio.h>
#include	<syslog.h>
#include	<sysexits.h>
#include	<ctype.h>
#include	<sys/param.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<sys/file.h>
#include	<fcntl.h>
#ifdef _AIX
#include	<sys/dir.h> 
#endif
#include	<sys/signal.h> 
#include	<sys/wait.h>
#include	<dirent.h>
#include	<firewall.h>
#include	"smwrap.h"  
    
/*===========================================================================
 * function prototypes
 *===========================================================================*/
int allblanks(char *s);
int chkaddr(char *s); 
int deliver(char *path); 
int getsmwp(struct SMWRAP *smwpptr); 
/*===========================================================================
 * External variables
 *==========================================================================*/
extern	char	*index();
extern  int	errno;

struct	towhom	
{
   char		*who;
   struct	towhom	*nxt;
};
     
int	currPIDs 	= 0;
int	debugflag  	= 0; 
struct 	SMWRAP 	*smwpptr;     	/* pointer to smwrap config record */
/*=============================================================================
 * reap child
 *============================================================================*/
static void reapchild(int sig)
{
   int	i;
   int	j;

   while((i = wait(&j)) != -1) 
   {
      currPIDs--;
      if(debugflag)
         fprintf(stderr,"reap %d exit = %d current pids = %d max pids = %d.\n",
			i, j, currPIDs, smwpptr->maxpids);
   }
   if(currPIDs < 0)
      currPIDs = 0;
   signal(SIGCHLD,reapchild);
}
/*=============================================================================
 * input arguments:
 *			normally none.
 *			-d [level]     debug flag
 *				levels 1 to 3 are supported wtih 1 = default
 *============================================================================*/
main(int argc, char *argv[]) 
{ 
   static  char ident[] = "@(#) smwrapd, V 1.0, 00/01/21"
                          " (C) Freemont Avenue Software, Inc."
                          " 1995-2000";

   DIR	*dp;      			/* directory pointer	*/
   struct 	dirent	*fn;    	/* directory structure 	*/ 
   struct 	SMWRAP 	smwrap;
   int		delvpid;		/* deliver process pid	*/ 

   smwpptr = & smwrap;
   openlog("smwrapd",LOG_PID|LOG_NDELAY,LOG_DAEMON); 
   /*================================
    * check for debug flags
    *================================*/
   if(argc > 1 )
   {
      if( !strcmp(argv[1],"-d")) 
      {
         debugflag = 1; 
         if( argc > 2)
         {
            if(!isalldigits(argv[2]))
            {
               fprintf(stderr,"\nError: invalid parameter %s. Must be an "
                              "integer > 1 and < 4.\n", argv[2]);
	       return(-1);
            } 
            if( ((debugflag = atoi(argv[2])) > 3) || 
		((debugflag = atoi(argv[2])) < 1)  )
            {
               fprintf(stderr,"\nError: invalid input parameter '%s'.\n",
				argv[2]);
	       return(-1);
            }
         }
      }
      else
      {
         fprintf(stderr,"\nError: invalid input parameter '%s'.\n",
				argv[1]);
         return(-1);
      }
   } 
   if( debugflag )
      fprintf(stderr,"\nsmwrapd starting: debug level = %d.\n\n",debugflag); 
   /*===============================================================
    * Read config file to obtain:
    * userid, groupid, maxbytes, maxreceipts, timout, spool directory,
    * Then set grouipd, userid. 
    *===============================================================*/    
   if( getsmwp(smwpptr) < 0) 
      return(-1);   


   if( setgid(smwpptr->groupid) ) 
   {
      syslog(LOG_CRIT,"Error: cannot setgid %d: %m", smwpptr->groupid);
      exit(1);
   }  
   if(setuid(smwpptr->userid)) 
   {
      syslog(LOG_CRIT,"Error: cannot setuid %d: %m",smwpptr->userid);
      exit(1);
   }
   syslog(LOG_NOTICE,"smwrapd Version 1.0 started. wakeup = %d sec." 
		     " userid = %d ",
		smwpptr->wakeup, smwpptr->userid);   
   if( chdir(smwpptr->spoolpath) ) 
   {
      syslog(LOG_CRIT,"Error: cannot chdir to spool directory %s: %m: "
                      " groupid(%d) userid(%d)",
		smwpptr->spoolpath, smwpptr->groupid,smwpptr->userid);
      exit(1);
   }

   if(!debugflag)
      daemonize();
   signal(SIGCHLD,reapchild);

   if((dp = opendir(".")) == (DIR *)0) 
   {
      syslog(LOG_CRIT,"Error: cannot scan spool directory: %m");
      exit(1);
   }
   /*======================================================================
    * Continuous loop:  Look for files to process. If any are found deliver
    * them. When files are delivered sleep for specified duration. 
    *=====================================================================*/
   while(1) 
   {      
      if(debugflag)
         fprintf(stderr,"re-scanning directory\n"); 
      /*================================================================
       * Scan the directory for files with names that fit the pattern
       * "Smwrapxxxxxx" and deliver them.  
       *===============================================================*/
      while((fn = ( struct dirent *) readdir(dp)) != (struct dirent *)0) 
      {
         

	 while(currPIDs >= smwpptr->maxpids) 
         {
	    if(debugflag)
	       fprintf(stderr,"Current pids = %d, max pids = %d. Will sleep 2 "
                              "seconds before scheduling more work.\n",
  			      currPIDs,smwpptr->maxpids);
	    sleep(2);
         }

	 if(fn->d_name[0] == '.')
	    continue;

	 if( strncmp(fn->d_name,"Smwrap",6) )
	    continue;

	 currPIDs++;
         /*============================================================
          * create child process. The child process executes deliver
	  * and exits.
          *============================================================*/
	 if((delvpid = fork()) == 0) 
         {  
	    exit(deliver(fn->d_name));
         }
         if(debugflag)
	    fprintf(stderr,"Child %d scheduled to deliver %s. Current pids = "
			   "%d, max pids = %d.\n", 
			   delvpid,fn->d_name,currPIDs,smwpptr->maxpids);
	 if( delvpid < 0) 
         {
	    syslog(LOG_ERR,"Error: cannot fork. Current # pids = %d. "
			   "Unable to deliver %s: %m. Will retry.", 
			   currPIDs, fn->d_name);
	    currPIDs--;
	    goto takenap;
	 }
      }
      rewinddir(dp);


      if(debugflag > 2)
      {
         int	highfd;
	 if((highfd = open("/dev/null",O_RDONLY)) > -1) 
         {
	    fprintf(stderr,"\nhigh fd = %d\n", highfd);
	    close(highfd);
	 }
      }
  
      /*====================================================
       * take a nap for wakeup seconds
       *====================================================*/
takenap:
      if(debugflag)
         fprintf(stderr,"Parent napping for %d seconds.\n",smwpptr->wakeup);
      sleep(smwpptr->wakeup);
   }
}
/*===========================================================================
 * deliver
 *
 * input:	pointer to pathname of a complete file "Smwrap..."
 *
 * outputs:
 *	return codes	0 mail delivered ok
 *			1 mail not delivered 
 *	
 *==========================================================================*/
int deliver(char *path)   
{
   char		*addrptr;           	/* mail address ptr		*/
   int		bad = 0; 
   off_t	bodyoffset;  		/* offset of body of the letter */
   char		buf[BUFSIZ];  
   struct	towhom	*recips = (struct towhom *) 0;
   struct	towhom	*nrec;
   int		fileMT  = 0;		/* File MT flag			*/
   FILE		*fp;
 
   char		*fromaddr;
   char		**h;
   char		*hdrptr;
   int		wstatus;                /* status returned by wait()	*/
   struct	stat	sbuf;   	/* file stat buffer  		*/
   
   
   int		pid;

   /* stuff used for parsing addresses.	*/
   enum	{ H_NONE, H_ANY, H_ADDR, H_BODY } state;
   
   static	char	*addr_hdrs[] = 
   {
	"bcc:",
	"cc:",
	"errors-to:",
	"from:",
	"reply-to:",
	"return-path:",
	"return-receipt-to:",
	"to:",
	0,
   }; 

   if(debugflag)
   {
      fprintf(stderr,"deliver function entered. file = '%s'.\n",path);  
   }
   signal(SIGCHLD,SIG_IGN); 
   /*================================================
    * open file for update (reading and writing) 
    *===============================================*/
   if((fp = fopen(path,"r+")) == (FILE *)0) 
   {
      syslog(LOG_ERR,"Error: cannot open '%s': %m",path);
      return(1);
   }
   if(debugflag)
         fprintf(stderr,"deliver: file '%s' fileno(%d) opened.\n",
			path, fileno(fp));
   if(locktest_fd(fileno(fp))) 
   {  
      if(debugflag)
         fprintf(stderr,"locktest failed for file %s. error: %s\n",
			path, strerror(errno));
      fclose(fp);
      return(1);
   }
   /*=====================================================================
    * Get the file stat information and test for a regular file. 
    *====================================================================*/
   if(fstat(fileno(fp),&sbuf)) 
   {
      syslog(LOG_ERR,"Error: fstat failed for '%s': %m",path);
      goto leave_gracefully;
   } 
   if((sbuf.st_mode & S_IFMT) != S_IFREG) 
   {
      syslog(LOG_ALERT,"Security Alert: '%s' is not a regular file.",path);
      goto sequester;
   }
   
   /*======================================================================
    * The file will be sequestered if it is empty or if does not start with
    * the keyword FROM or the FROM address is larger than the input buffer
    * (4KB).  
    *=====================================================================*/
   if(fgets(buf,sizeof(buf),fp) == (char *)0) 
   {  
      if( feof(fp) )
      {
         syslog(LOG_ERR,"Error: file '%s' is empty and was deleted.",
			path);       
      } else if( ferror(fp) )
      {
         syslog(LOG_ERR,"Error: file '%s' is empty and was deleted. %s",
			path, strerror(errno));
      }
      fileMT = 1;
      goto sequester;
   }
   if(strncmp(buf,"FROM ",5)) 
   {
      syslog(LOG_ERR,"Error: file '%s' does not begin with 'FROM'. "
		     "Status sequestered.",
 			path);
      goto sequester;
   }
   if( (addrptr = index(buf,'\n')) == (char *) 0) 
   {
      syslog(LOG_ALERT,"Security Alert: file %s's FROM line too long"
		       " (> %d).  Status sequestered.  %.40s",
 			path, sizeof(buf),buf);
      goto sequester;
   } 
   
   *addrptr = '\0';		/* terminate the 'FROM' address string 	*/
   addrptr = &buf[5];         	/* start with first char after FROM 	*/
   				/* validate FROM address		*/
   if(chkaddr(addrptr) < 0) 
   {
      syslog(LOG_ALERT,"Security Alert: file %s contains a bad FROM address."
			" Status sequestered.	%.50s.",
 			path,addrptr);
      goto sequester;
   }  

   if( (fromaddr = (char *) malloc(strlen(addrptr) + 1)) == (char *) 0) 
   {
      syslog(LOG_ERR,"Error: malloc for fromaddr failed: %m. Will retry later.");
      goto leave_gracefully;
   }
   strcpy(fromaddr,addrptr);

   if(debugflag)
      fprintf(stderr,"FROM %s\n",fromaddr);
                                            
   /*================================================================
    * FROM is followed by RCPT and a list of one or more receipients  
    *===============================================================*/
   while(1) 
   {
      if(fgets(buf,sizeof(buf),fp) == (char *)0) 
      {
         syslog(LOG_ERR,"Error: file %s contains no receipients or data!",
			path);
	 goto sequester;
      }

      if(!strncmp(buf,"RCPT ",5)) 
      {
         if( (addrptr = index(buf,'\n')) == (char *)0 ) 
         {
	    syslog(LOG_ALERT,"Security Alert: file %s RCPT line too "
			     " long.  Status sequestered. %.100s...",
			     path,addrptr);
	     	    goto sequester;
	 }
	 *addrptr = '\0';
	 addrptr = &buf[5];
	 while(isspace(*addrptr))
	    addrptr++;

	 /*===================================================================
          * Look for hackers sending parameters to sendmail.           
	  *==================================================================*/
	 if(*addrptr == '-')           
	 {                                      
            syslog(LOG_ALERT,"Security Alert: possible attempt to send "
			     "parameter to sendmail. Address starts with '-':"
			     " %.120s",
  	 	     	     addrptr);
 	    continue;
 	 }
	 if(chkaddr(addrptr) < 0) 
         {
	    syslog(LOG_ALERT,"Security Alert: sequestered file %s contains a "
			     "bad RCPT.  %.120s",
	 			path,addrptr);
	    goto sequester;
	 }
         /*==================================================================
	  * Acquire space for towhom struct to link next recipient.
          * Acquire space to hold next recipient address and link.  
          *==================================================================*/
	 nrec = (struct towhom *) malloc(sizeof(struct towhom));
	 if(nrec == (struct towhom *)0) 
         {
	    syslog(LOG_ERR,"Error: malloc failed for recipient pointer: %m."
                           " Delivery rescheduled.");
	    goto leave_gracefully;
	 }
	 nrec->who = (char *) malloc(strlen(addrptr) + 1);
	 if(nrec->who == (char *)0) 
         {
	    syslog(LOG_ERR,"Error: malloc failed for recipient: %m. "
                           "Delivery rescheduled.");
	    goto leave_gracefully;
	 }

	 strcpy(nrec->who,addrptr);
	 nrec->nxt = recips;
	 recips = nrec;
	 continue;
      }
      /*============================================================
       * BODY must follow the RCPT line
       *============================================================*/
      if(!strncmp(buf,"BODY",4))
         break;
      syslog(LOG_ERR,"Error: file %s has no BODY, unexpected "
                        "envelope.  Status sequestered.  %.32s",
			path, buf);
      goto sequester;
   }

   /* save the body offset for later use */
   bodyoffset = ftell(fp);

   /*==========================================================
    * look for dangerous mail header stuff 
    *==========================================================*/
   state = H_NONE;
   while(state != H_BODY && fgets(buf,sizeof(buf),fp) != (char *)0) 
   {
      if ((hdrptr = index(buf,'\n')) == 0) 
      {
         syslog(LOG_ALERT,"Security Alert: file %s's header is too"
			  "long. Status sequesterd. %.64s...",
  			  path, buf);
 	 goto sequester;
      }
      *hdrptr = 0;
      if(isspace(buf[0])) 
      {
         if(state == H_NONE || allblanks(buf))
	    state = H_BODY;
	 else
	    hdrptr = buf;
      } 
      else 
      {
         state = H_BODY;
	 for(hdrptr = buf; *hdrptr && isgraph(*hdrptr); hdrptr++) 
         {
	    if(*hdrptr == ':') 
            {
	       state = H_ANY;
	       for(h = addr_hdrs; *h; h++) 
               {
	          if(!strncasecmp(buf,*h,hdrptr-buf)) 
                  {
		     state = H_ADDR;
		     break;
		  }
	       }
	       break;
	    }
	 }
      }
      if(state == H_ADDR && chkaddr(hdrptr + 1) < 0) 
      {
         syslog(LOG_ALERT,"Security Alert: file %s has a bad address"
			  " in header. Status sequestered. %.40s",
 			   path, buf);
	 goto sequester;
      }
   }

   nrec = recips;
   /*======================================================= 
    * process all the recipient records on the list	
    *=======================================================*/
   while(nrec != (struct towhom *)0) 
   { 
      int	i;
      pid_t	cpid;			/* child pid returned by wait	*/
      char	*smarg[MAXSMARG];     	/* pointers to sendmail args 	*/ 

      i = 0;
      smarg[i++] = smwpptr->smtppath; 
      smarg[i++] = "-f";
      smarg[i++] = fromaddr;
      /*================================================================
       * Process all the records up to the maximum allowed for sendmail 
       *===============================================================*/
      while( nrec != (struct towhom *) 0 ) 
      {
         if(i > MAXSMARG - 2)
	    break;
	 smarg[i++] = nrec->who;
	 nrec = nrec->nxt;
      }
      smarg[i] = (char *) 0;      	/* terminate sendmail argument list*/
      /*===========================================
       * create child process to exec sendmail.
       *==========================================*/
      if((pid = fork()) == 0) 
      {
         int	fd;

         /*========================================
	  * opened twice!!!
          * fix to get around mach 2.0 error
	  *========================================*/
         if((fd = open(path,O_RDONLY|O_RDWR)) < 0) 
         {
	    syslog(LOG_ERR,"Error: child cannot open %s: %m",path);
	    exit(1);
	 } 
	 /* seek to the beginning of the body of the letter	*/ 
	 if(lseek(fd,bodyoffset,0) != bodyoffset) 
         {
	    syslog(LOG_ERR,"Error: seek to offset %d of %s failed: %m",
				bodyoffset, path);
	    exit(1);
	 } 
	 /* close stdin and dup file fd to stdin	*/
	 close(0);  			
	 if(dup(fd) < 0) 
         {
	    syslog(LOG_NOTICE,"Error: cannot dup %s: %m", path);
	    exit(1);
	 } 
         /*==============================================
          * exec sendmail 
          *==============================================*/
	 execv(smarg[0],smarg);
	 lockun_fd(fileno(fp));
	 syslog(LOG_ERR,"Error: cannot exec %s: %m",smarg[0]);
	 exit(1);
      }
      /*========================================================
       * Parent waits for children ( just like in real life).
       * POSIX.1 macros are used to examine wait status.
       * For normal termination get the exit status.
       * For abnormal terminaton get the exit status code.
       * Sendmail uses exit codes beginning with EX__BASE (64)   
       *
       *=======================================================*/
      while((cpid = wait((int *) &wstatus)) != pid && cpid != -1) ; 
      if( WIFEXITED(wstatus) ) 
      {
         wstatus = WEXITSTATUS(wstatus);   
      }
      else
      {
         if( WIFSIGNALED(wstatus) ) 
         {  
            /*=======================================================
	     * AIX 4.1.3 sendmail returns signal 20 on exit.  
             * AIX 3.2.5 sendmail sometimes returns 56 or 44 on exit. 
	     * Solaris 2.5 sendmail returns 2,4,8,16,32, & 64
 	     *======================================================*/ 
/*
            if( WTERMSIG(wstatus) == 20  || WTERMSIG(wstatus) == 56 || 
                WTERMSIG(wstatus) == 44    ) 
*/
            if( WTERMSIG(wstatus) < EX__BASE ) 
            {
               lockun_fd(fileno(fp));
               close(fileno(fp));
               unlink(path);		/* remove the file	*/
               syslog(LOG_NOTICE,"file %s was delivered.",path); 
               return(0);
            }
	    syslog(LOG_ERR,"Error: child process terminated with signal %d.",
			WTERMSIG(wstatus) ); 
         }
         else
         if( WIFSTOPPED(wstatus) )
         {  
            lockun_fd(fileno(fp));
            close(fileno(fp));
            unlink(path); 		/* remove the file 	*/
            syslog(LOG_NOTICE,"file %s was delivered.",path);
            return(0);
            
         }          
                             
         /*===================================================================
          * If the termination status is >= EX__BASE then it came from sendmail. 
          *==================================================================*/
         if( wstatus >= EX__BASE )
         {
            switch(wstatus) 
            {              
	    case EX_OK:
                 goto leave_gracefully; 
#ifndef __hppa
            case EX_CONFIG: 
                 syslog(LOG_NOTICE,"Error: 'Configuration Error', wait status "
                                   "= %d. Will retry.",
					wstatus);
		 goto leave_gracefully; 
#endif
	    case EX_IOERR: 
                 syslog(LOG_NOTICE,"Error: I/O Error, wait status = %d. Will "
                                   "retry.",
					wstatus);
		 goto leave_gracefully; 

            case EX_NOHOST:
	         syslog(LOG_NOTICE,"Error: Host Unknown: wait status = %d. "
                                   " File sequestered.", wstatus);
		 goto sequester;

	    case EX_NOPERM:         
	         syslog(LOG_NOTICE,"Error: Permission Denied: wait status = %d."
                                   " Will retry.",
					wstatus);
		 goto leave_gracefully;

	    case EX_NOUSER:
	         syslog(LOG_NOTICE,"Error: User Unknown: wait status = %d. "
                                   "File sequestered.", wstatus);
		 goto sequester;  
            
            case EX_OSERR:
                 syslog(LOG_NOTICE,"Error: 'OSERR', wait status = %d. Will "
                                   "retry.",
		 			wstatus);
		 goto leave_gracefully; 
          	
	    case EX_OSFILE: 
                 syslog(LOG_NOTICE,"Error: 'OSFILE', wait status = %d. Will "
                                   "retry.",
		 			wstatus);
		 goto leave_gracefully; 

	    case EX_TEMPFAIL: 
                 syslog(LOG_NOTICE,"Error: 'Temporary Error', wait status = %d. Will retry.",
		 			wstatus);
		 goto leave_gracefully;  

             case EX_UNAVAILABLE: 
                 syslog(LOG_NOTICE,"Service Unavailable: wait status = %d. "
                                   "Will retry. ",	wstatus);
		 goto leave_gracefully; 


	    default:
	         syslog(LOG_NOTICE,"Error: other wait status error = %d. "
                                   " %s. File sequestered.", 
                                   WTERMSIG(wstatus), strerror(errno));
	         goto sequester;
            } /* end of switch */
         } 
      }
   }
   lockun_fd(fileno(fp));
   close(fileno(fp));
   unlink(path);
   syslog(LOG_NOTICE,"file %s was delivered.",path);
   return(0);
   /*=======================================================================
    * Unlock the offending file, close it, place it in the sequestered file
    * directory and unlink it.  If the file is MT simply remove it.
    *======================================================================*/ 
sequester:  
   lockun_fd(fileno(fp));
   if(fp != (FILE *)0)
      fclose(fp);
   if(smwpptr->undelivpath && !fileMT) 
   {
      sprintf(buf,"%s/%s",smwpptr->undelivpath,path);
      (void)rename(path,buf);
   }
   (void)unlink(path);
   return(1);
/*============================================================================
 * unlock the file, close, and return;
 *===========================================================================*/
leave_gracefully:
   lockun_fd(fileno(fp));
   fclose(fp);
   return(1);
} /* end of deliver function     */
/*==========================================================================
 * input: a pointer to a character string
 * output: returns 1 if string is all blanks
 *	   returns 0 is string is not all blanks 
 *=========================================================================*/
int allblanks(char *s)  
{
   int	ch;
   /* loop through string */
   while((ch = *s++) != 0)
   {
      if(!isascii(ch) || !isspace(ch))
         return(0);
   }
   return(1);
}
/*=======================================================================
 * function: chkaddr
 *         
 * inputs: pointer to e-mail address
 *
 * functions:
 * 
 * For each address in this list, find the user part that would be
 * seen on the final destination host. The code below understands
 * user@host, @host:user, host!user, host::user, host:user, user%host
 * and combinations thereof (for instance, @host:user%host). After
 * stripping quotes from the user part, require that it does not
 * start with |. If the user part begins with /, require that the /
 * is followed by attribute=value.
 *                           
 * outputs:  
 *	returns  1 if recipient address are ok
 *		-1 if any recipient address is questionable
 *
 *========================================================================*/
int chkaddr(char *s)
{
   char		*open_parens = "(<";
   char		*close_parens = ")>";
   char		paren_stack[10];
   char		*sp = paren_stack;
   unsigned	ch;
   char		*userpart;
   char		*start;
   char		*cp;

#define separator(c) (isspace(c) || (c) == ',')

   while((ch = *s) != 0) 
   {  
      /* skip over separators	*/
      if(separator(ch)) 
      {
         ++s;
	 continue;
      }
      start = userpart = s;

      while((ch = *s) != 0 && (sp > paren_stack || !separator(ch))) 
      {
         if(ch == '\\') 
         {
	    if(*++s)
	       	++s;
	 } 
         else if((cp = index(open_parens, ch)) != 0) 
         {
	    if(sp >= paren_stack + sizeof(paren_stack))
	       return(-1);
            *sp++ = close_parens[cp - open_parens];
	    ++s;
         } 
         else if(sp > paren_stack && ch == sp[-1]) 
         {
	    sp--;
	    ++s;
	 } 
         else if(ch == '"') 
         {
	    while((ch = *++s) != 0) 
            {
	       if(ch == '\\' && s[1]) 
               {
	           ++s;
	       } 
               else if(ch == '"') 
               {
	          ++s;
		  break;
	       }
	    }
	 } 
         else 
         {
	    ++s;
	 }
	 if(index("!:<",ch) != 0)
	    userpart = s;
      }
      while(*userpart == '"')
         userpart++;
      if(*userpart == '|')
         return(-1);
      if(*userpart == '/') 
      {
         cp = userpart + 1;
	 while((ch = *cp++) != '=') 
         {
	    if(ch == 0 || index("/%@>",ch) != 0)
	       return(-1);
	 }
      }
   }
   return(1);
}

