/*============================================================================
 * filename:		tnproxy/authenticate.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*========================================================================== 
 *
 * written by:		R. James Livermore
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  All rights
 * reserved.  
 * 
 *
 * Date written:	April     23, 1995
 * Last modified:	February   8, 2000
 *	
 * functions:
 *		Prompt user for name and read gwuser record. 
 *		If user not found ask for password and reject so that they
 *		can Not tell which was wrong.  Allow MAXTRIES.  If user
 *		record is found then determine type of authentication.  
 * 
 * outputs: 
 *	return  0 if ok 
 *	return -1 error or not found.
 *             -2 error unable to write name prompt
 *             -3 error while reading user name 
 *	       -4 error writing "longID" message  
 *             -5 error writing password prompt
 *             -6 error reading password 
 *	       -7 error writing "invalid name or password" message
 *	       -8 error user locked or access outside allowed times/dates.
 * 
 *=========================================================================*/  
static	char	cprght[] = 
                "(C) COPYRIGHT Freemont Avenue Software, Inc. 1995-2000 "
                "tnproxy V 1.0 00/01/21";
#include	<stdio.h>  
#include	<string.h>
#include	<ctype.h>
#include	<syslog.h>
#include	<fcntl.h>
#include	<sys/ioctl.h>
#include	<sys/errno.h> 
#include	<sys/signal.h>
#include	<arpa/telnet.h>
#include	<sys/time.h>
#include	<sys/types.h>
#ifndef __hppa
#include 	<sys/select.h>		/* to support fd_set 		  */
#endif
#include	<sys/socket.h>
#include	<netinet/in.h>
#include	<netdb.h> 
#include	"firewall.h" 
#include	"tnproxy.h" 
#include	"gwuser.h"

/*===========================================================================
 * function prototypes
 *===========================================================================*/
int authuser(struct GWUSER *gwup, char *buf, struct TNPROXY *proxyptr);  
int dssapi(struct GWUSER *gwup);
int getline(unsigned char *buf, int siz, char *pro, int plen, int echo);  
int tnsecdyn(char *username, char *buffer);
static char	nprompt[] = "Username: ";
static char	pprompt[] = "Password: "; 
extern struct GWCONF conf;
/*===========================================================================
 *
 *=========================================================================*/
int authenticate(struct GWUSER *gwuptr, char *sourceip, char *failedusers,
                 struct TNPROXY *proxyptr)
{ 
   static char  invalnpwd[] = "You entered an invalid login name or password.\r\n";
   static char	longID[]   = "User-ID is too long";
   char		buf[CBUFSIZ];
   int		i;
   int		j;
   int		namelen; 
   int		ntries   = MAXTRIES;
   int		numalpha = 0;
   int		rc;
   char		username[UNMSIZ];
   failedusers[0] = '\0';	/* initialize end of string	*/  
   /*============================================================
    * Check for use of DSS.  
    *===========================================================*/          
   if( proxyptr->srcIPsecure != SECURED )
   {
      if( (rc = dssapi(gwuptr)) == 0)  
      {  
         syslog(LOG_NOTICE,"user '%s' authenticated by DSS",gwuptr->username); 
         /* look for gwuser record in case user is a T.Rex administrator */
         strncpy(username,gwuptr->username,UNMSIZ);
         get_gwuser(username, gwuptr);
         return(0); 
      }
      switch (rc)
      {  
         case 1002:     	/* DSS not used	*/
            break;                
         case 5:
            syslog(LOG_NOTICE,"unknown user '%s' denied by DSS, rc = %d",
	           gwuptr->username,rc);
            return(-1); 
            break;  
         case 8:
            syslog(LOG_NOTICE,"user '%s' denied by DSS, rc = %d",
	           gwuptr->username,rc);
            return(-1); 
            break; 
         case 9:
            syslog(LOG_NOTICE,"user '%s' denied, DSS server busy, rc = %d",
	           gwuptr->username,rc);
            return(-1); 
            break;  
         default:
            syslog(LOG_NOTICE,"user '%s' denied, DSS rc = %d",
	           gwuptr->username,rc);
            return(-1); 
            break;  
      }
   }
   /*=========================================================== 
    * Give the user ntries to get it right	
    *===========================================================*/
   while(ntries--) 
   {
      int gwuserrc;
      /* prompt for user ID */
      if(write(0,nprompt,sizeof(nprompt)) != sizeof(nprompt)) 
      {  
         syslog(LOG_ERR,"Error: while writing '%s'.",nprompt);
         return(-2);
      }
      /* get Username from user */
      if(getline((unsigned char *)buf,sizeof(buf) - 1,nprompt,sizeof(nprompt),1)
          < 0 )
      {  
         syslog(LOG_ERR,"Error: while reading user name.");
         return(-3);    
      }
      /*====================================================================
       * Validate Username: length > 0 and < 17, at least one alpha character 
       *===================================================================*/
      if((namelen = strlen(buf)) == 0)
         continue;
      if( namelen > (UNMSIZ-1)) 
      {  
	 strncat(failedusers,buf,UNMSIZ);
         strcat(failedusers," ");
         if(sayn(0,longID,sizeof(longID)))
         {  
            syslog(LOG_ERR,"Error: while writing '%s'.",longID);
	    return(-4);
         }
	 continue;
      }
      strcat(failedusers,buf);
      strcat(failedusers," ");
      numalpha = 0; 
      for (i=0; i < namelen;i++)
      {
         j = buf[i];
         if( j < 128 && j > 0)
	    if (isalpha(j) !=0) numalpha++; 
      }
      if( numalpha  == 0 )  continue; 
      /*====================================================================
       * Get user record.  If not found (negative return code) ask for
       * password and send message indicating invalid name or password. Let
       * them try again.  If record found (non-negative return code) then
       * authenticate the user and if the user authenticated, we then handle
       * the positive retrn codes from authuser() by telling the user what
       * was wrong.  If the user did not authenticate, we won't tell them
       * and will prompt them for username/password again.
       *===================================================================*/
      if( (gwuserrc = get_gwuser(buf, gwuptr)) >= 0)
      {   
         int authrc;
         if( (authrc = authuser(gwuptr,buf,proxyptr)) == 0) 
         {
            switch (gwuserrc) 
            {
               case 0: return(0);       
               case 1: say(0,"User record locked.");
                       return(-8);
               case 2: say(0,"Current date outside of effective start and "
				"stop dates for this user.");
                       return(-8);
               case 3: say(0,"Current time is not within the allowed access "
				"times.");
                       return(-8);
            }
         }
         else if (authrc == -1)
            continue;	/* try again	*/ 	
         else	/* authuser returned -2 - don't give any other tries */
            return(-8);
      }
      else	/* user or group record cannot be read or found */
      {  
         if(write(0,pprompt,sizeof(pprompt)) != sizeof(pprompt))
         {  
            syslog(LOG_ERR,"Error: while writing '%s'.",pprompt);
	    return(-5); 
         }
	 if(getline((unsigned char *)buf,sizeof(buf) - 1,pprompt,sizeof(pprompt),0) < 0) 
         {  
            syslog(LOG_ERR,"Error: while reading password.");
	    return(-6);  
         }
         if(write(0,invalnpwd,sizeof(invalnpwd)) != sizeof(invalnpwd)) 
         {
            syslog(LOG_ERR,"Error: while writing '%s'.",invalnpwd);
	    return(-7); 
         }
         continue;	/*  try again */
      } 
   } 
   return(MAXTRIES);  /* user not found	*/

} 
/*===========================================================================
 * authenticate user using method specified in gwuser record
 * inputs:
 *		ptr to GWUSER record 
 *		ptr to buffer
 *		ptr to source IP address
 * outputs:
 *		 0 if OK
 *		-1 if failed	
 *		-2 if failed and the user should not be given another chance
 *==========================================================================*/ 
int authuser(struct GWUSER *gwup, char *buf, struct TNPROXY *proxyptr)
{
#define APE	1
#define NOAPE	0
   struct GWCONF *gwconf = &conf;
   char         challenge[9];
   int		msglen;  
   int    	ichall;   
   int		rtncode;
   time_t	epoch; 
   memset(challenge,0,sizeof(challenge));  
   if (gwuserconf()) 
   {
      say(0,"Cannot access gwuser config info, please contact "
                        "administrator.");
      return(-1);
   }
   /*=================================================================== 
    * If user is on a protected host use password for authentication
    * unless Strong User authentication is required for protected users.	
    *==================================================================*/ 
   if( (proxyptr->srcIPsecure == SECURED) && !(gwup->tnaccess & PRO_TNSA) )
   {
      if(write(0,pprompt,sizeof(pprompt)) != sizeof(pprompt))
         return(-1); 
      if( getline((unsigned char *)buf,CBUFSIZ - 1,
          pprompt,sizeof(pprompt),NOAPE) < 0 )
         return(-1);
      if( strlen(gwup->password[0]) == 0)
      {  
         syslog(LOG_NOTICE,"Access denied to %s. NULL password.",
			   gwup->username);
         return(-1);
      }                      
      /*=======================================================
       * Compare password from user with password in record. 	
       *======================================================*/ 
      if( strncmp(buf,gwup->password[0],UNMSIZ-1) != 0)
      {  
         set_nologins(gwup->username,1,gwconf->no_login_attempts);
         return(-1);
      }               
      else 
      {
         int rc;
         time_t cur = time((time_t *)NULL);
         
         /*=======================================================
          * Only prompt for password change if it has expired AND
          * the user has not been locked.
          *======================================================*/ 
         if (gwconf->password_max_age != 0 &&
		cur - gwup->passworddate > gwconf->password_max_age*24*60*60 &&
		gwup->status != LOCKED)
         {
            say(0,"Password has expired.  Change password now or this gwuser "
		"will be disabled from further access.");
            if ((rc = cmd_passwd(0, NULL, NULL)))
            {
               if (rc < 0) 
               {
                  say(0,"Password change cancelled.  This gwuser id is now "
			"disabled for further access.");
                  lockuser(gwup->username);
                  return(-2);
               }
               else
                  return(-1);
            }
            else
                say(0,"Password changed.");
            /*===============================================================
             * So the user just changed his expired password, if he doesn't
             * have PRO_TN access, we assume he is an http user and only
             * uses tnproxy to change his expired password so we just exit
             * quietly rather than putting up the no PRO_TN message and
             * prompt for username again.
             *==============================================================*/
            if( !(gwup->tnaccess & PRO_TN) )
               return(-2);
         } 
         set_nologins(gwup->username,0,gwconf->no_login_attempts);
         if( !(gwup->tnaccess & PRO_TN) )
         {  
            say(0, "User has no permission to telnet from protected network.");
            syslog(LOG_NOTICE,"Access denied to %s - no PRO_TN permission.",
				gwup->username);
            return(-1);
         }  
         return(0);
      }
   } 
   /*========================================================================
    * If you arrive here then you will be challenged. First
    * select token type from users record
    *========================================================================*/
   switch (gwup->tokentype)  
   {
      case CRYPTO:              /* CRYPTOCard support	  */
      case SNK:           	/* SecureNet Key  Support */   
      
         epoch = time(&epoch); 
         srand48((int) epoch);		/* seed the random # generator */
         ichall = (lrand48() & 0x03FFFFFF) + 32800000; 
         sprintf(challenge,"%d",ichall);
         sprintf(buf,"Challenge: %s \r\nEnter Response: ",challenge);
	 msglen = strlen(buf); 
         if(write(0,buf,msglen) != msglen)
	    return(-1); 
         /* get response and echo challenge */ 
         if( getline((unsigned char *)buf,CBUFSIZ - 1,
		     buf,sizeof(buf),APE) < 0 ) 
         {
	    syslog(LOG_ERR,"Error: unable to read response to challenge %m.");
            return(-1); 
         } 
         rtncode = snkverify(challenge, buf, gwup);
         if(rtncode)
            set_nologins(gwup->username,1,gwconf->no_login_attempts);  
         else {
            /* added UNPRO_TN check,  rjl  7/10/87 */
            if( !(gwup->tnaccess & UNPRO_TN) )
            {  
               say(0, "User has no permission to telnet from unprotected "
			"network.");
               syslog(LOG_NOTICE,"Access denied to %s.  UNPRO_TN not allowed "
				"for user.",gwup->username);
               return(-1);
            }  
            set_nologins(gwup->username,0,gwconf->no_login_attempts);   
         }
         return(rtncode);   
         break;
   
      case SDI:			/* Security Dynamics Support    */
         return(tnsecdyn(gwup->username,buf));
         break;  
        
      case NONE:    
         syslog(LOG_ERR,"Error: no security token specified for user = '%s',"
			" group = '%s'",
			gwup->username,gwup->groupname);
         sprintf(buf,"No security token specified.  Call your systems "
                     "administrator.\r\n"); 
         msglen = strlen(buf); 
         if(write(0,buf,msglen) != msglen)
	    return(-1); 
         return(-1);
         break;

      default:
         return(-1);
         break;

   }   
}            

