/*============================================================================
 * file:	util/adam/adam.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 * function:	Adam builds the aliases and reverse aliases data bases
 *		from an aliases file.   
 *		Named after adam who named all the animals.  
 *
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  All rights
 * reserved.  
 *                                                             
 * Date written:	June 	  26, 1995
 * Last modified:	February   8, 2000
 *
 *============================================================================*/
#include	<sys/types.h>  
#include	<stdio.h>
#include	<ctype.h>
#include	<syslog.h> 
#include	<string.h>
#include	<fcntl.h>
#include	<sys/ioctl.h>
#include	<sys/errno.h> 
#include	<sys/signal.h>
#include	<sys/stat.h>
#include	<arpa/inet.h>
#include	<sys/time.h>
#include	<time.h> 
#include	<sys/times.h>
#include	<unistd.h> 
#include	<stdlib.h>
#include	"adam.h"
#include	"firewall.h"
#include	"smwrap.h"
extern int h_errno;   


/*========================================================================
 * function prototypes
 *=======================================================================*/
int addaliases(char * newaliasesfile, char *aliasesDB, int maxusers); 
int addraliases(char * newaliasesfile, char *aliasesDB, int maxusers);    
int bldaliases(char * aliasesfile, char *aliasesdb, int maxusers);  
int bldraliases(char * aliasesfile, char *raliasesDB, int maxusers);
int delaliases(char * delaliases, char *aliasesDB, int maxusers); 
#ifdef _AIX
extern int strcasecmp( char *, char *);
#endif 
/*
extern "C" int getaliases(char *keyptr, char *dataptr, char *aliasesDB);  
*/
int getaliases(char *keyptr, char *dataptr, char *aliasesDB);  
int prtaliasesDB(char *aliasesDB, int maxaddrs, int maxusers);  
void usage(void);
/*============================================================================
 * main function  
 * inputs: optional input arguments
 * 	   -h		for help
 *	   -p [all,aliases,rev] print aliases, revers alaiases or both		
 *
 * outputs:
 *    returns 
 *
 *
 *=========================================================================*/ 
main(int argc, char *argv[])
{ 
   static char ident[]
          = "@(#) adam, V 1.0, 00/01/21" 
            " (C) COPYRIGHT Freemont Avenue Software, Inc. 1995-2000";

   static char aliasesfile[256] = ALIASESFILE;  
   static char newaliasesfile[256] = NEWALIASESFILE; 
   char aliasesDB[256]; 
   char raliasesDB[256]; 
   int	addDB		= 0;
   int	bldDB		= 0; 
   int	daliases	= 0;
   int	drev		= 0; 
   int	i;  
   int	maxusers	= INT_MAX;
   int	prtaliases	= 0;
   int	prtrev		= 0; 
   int	radd1		= 0;
   int	radd2		= 0;
   int	rc1		= 0;  
   int	rc2		= 0;  
   int	rc3		= 0;
   char data[200];
   char dalname[200];
   char	drevname[200];
   struct tms	tmsstart, tmsend; 
   static long	clktck;
   static	clock_t		starttime;
   static	clock_t		endtime; 
   long		duration;

   struct SMWRAP smwrap;
   struct SMWRAP *smwpptr;
   smwpptr = & smwrap;   
     
   dalname[0] 	= '\0';
   drevname[0]	= '\0'; 
   if( ( starttime = times(&tmsstart)) == -1)
   {
      printf("start times error.\n");   
   } 
   /*=========================================================================
    * Open the system log
    *========================================================================*/
   openlog("adam",LOG_PID|LOG_NDELAY,LOG_NOTICE);
   /*===================================================
    * check for input arguments
    *==================================================*/
   if( argc > 1)
   {
      if( argv[1][0] != '-')   
      {
         printf("\n Error:  invalid leading character '%c' \n",argv[1][0]);
         usage();
         return(-1);        
      }
      
      for(i = 1; i < argc ; i++)
      {
         switch( argv[i][1])
         {                  
                        /*=============================================
			 * add new aliases records
			 *============================================*/  
            case 'a':
	       addDB = 1;	                     

               break;
                        /*=============================================
			 * build aliases or reverse aliases data bases
			 *============================================*/ 
            case 'b':  
	       bldDB = 1;
               break; 
                        /*=============================================
			 * delete aliases or reverse aliases record(s)
			 *============================================*/ 
            case 'd': 
               if( argc < i+2)
               {
                  printf("\nError: no parameters following '-%s'.\n",
				argv[i]);
                  usage();
                  return(-1);
               }
               if( argv[i][2] == 'a' )
               {
                  daliases = 1; 

                  if( argv[i+1][0] != '-') 
                  {
		     if( strlen(argv[i+1]) > 200 )
                     {
                        printf("\n Error: parameter following '-da' is longer"
			       " than 200 characters.\n"); 
		        usage();
			return(-1);
                     }
                     strcpy(dalname,argv[i+1]);  
                  }
                  else
                  {
                     printf("\nError: aliases name or file name must follow "
			    " '-da'.\n");   
		     usage();
		     return(-1);
                  }          
               }   
	       else
	       {
                  if( argv[i][2] == 'r' )
                  {
                     drev = 1;
                     if( argv[i+1][0] != '-')  
                     {
		        if( strlen(argv[i+1]) > 200 )
                        {
                           printf("\n parameter following '-dr' is longer than"
				  " 200 characters.\n"); 
		           usage();
			   return(-1);
                        }
                        strcpy(drevname,argv[i+1]); 
                     }   
                  }   
                  else
                  {
                     printf("\nError:Invalid parameter '%s'.\n",argv[i]);
                     usage();
	             return(-1);
		  }
               }
	       i++;
               break;


            case 'h':  
            case '?':
               usage();   
	       return(0);                       
                        /*=============================================
			 * print aliases or reverse aliases data bases
			 *============================================*/
	    case 'p': 
               if( argc <= i+1 )
               {
                  printf("\n Error: no parameter after -p.\n");
                  usage();
                  return(-1);
               }
               if( strcasecmp(argv[i+1],"all") == 0)
               {
		  prtaliases = 1;                     
                  prtrev     = 1;	
               } 
	       if( strcasecmp(argv[i+1],"aliases") == 0)
               {
		  prtaliases = 1;  
               } 
               if( strcasecmp(argv[i+1],"rev") == 0)
               {
		  prtrev     = 1;  
               }
	       i++;	
               break;
            
            default:
               printf("\n adam: invalid input parameter %s \n",argv[i]);   
               break; 
         }
      }
   } 
   else
   {
      printf("\nError: Missing input parameters.\n");
      usage();
      return(-1);
   }  


   printf("\n Aliases Database Administrator (adam) V 1.0\n"); 
   /*=========================================================================
    * Read sendmail wrapper configuration file to obtain the name of the
    * spool directory.  This is the directory the alaises DB must reside.  
    *========================================================================*/    
   if( getsmwp(smwpptr) < 0) 
   {  
      printf("\n Error returned from getsmwp. Check the error message in"
             " syslog.\n");
      return(-1);	/* exit if unable to read smwrapper config file */  
   } 
   strcpy(aliasesDB,smwpptr->spoolpath);  
   strcat(aliasesDB,"/aliasesDB"); 
   strcpy(raliasesDB,smwpptr->spoolpath);     
   strcat(raliasesDB,"/raliasesDB");      


   if( addDB )
   {  
      /*====================================================================== 
       * add new records to aliases data base
       *=====================================================================*/
      printf("\nUpdating Aliases Data Base.\n\n");  
      if( (radd1 = addaliases(newaliasesfile, aliasesDB, maxusers)) < 0)
      {        
         printf("\nError: addaliases rc = %d\n",radd1);
         return(radd1);
      }
      /*====================================================================== 
       * add new records to reverse aliases data base
       *=====================================================================*/
      printf("\nUpdating Reverse Aliases Data Base.\n\n");                         

      if( (radd2 = addraliases(newaliasesfile, raliasesDB, maxusers)) < 0)
      {  
         printf("\nError: addraliases rc = %d\n",radd2);
         return(1);
      }
   }
 
   if( bldDB )
   {  
      /*====================================================================== 
       * build aliases data base
       *=====================================================================*/

      if( (rc1 = bldaliases(aliasesfile, aliasesDB, maxusers)) < 0)
      {        
         printf("\nError: bldaliases rc = %d\n",rc1);
         return(1);
      }
      /*====================================================================== 
       * build reverse aliases data base
       *=====================================================================*/
      printf("\nStarting to build Reverse Aliases Data Base.\n\n");                         

      if( (rc2 = bldraliases(aliasesfile, raliasesDB, maxusers)) < 0)
      {  
         printf("\nError: bldraliases rc = %d\n",rc2);
         return(1);
      }
   }  
   if( daliases ) 
   {
      if( (rc3 = delaliases(dalname, aliasesDB, maxusers) ) < 0)
      {
         printf("\nError: delete aliases function failed.");
         return(rc3);
      }
   }    

   if( drev ) 
   {  
      if( (rc3 = delaliases(drevname, raliasesDB, maxusers) ) < 0)
      {
         printf("\nError: delete reverse aliases function failed.");
         return(rc3);
      }
     
   }

   /*=========================================================================
    * print aliases data base and reverse aliases data base
    *=======================================================================*/ 
   if( prtaliases)   
   {  
      if( rc1 == 0)
         rc1 = 10000;
      
      printf("\n\n Aliases Data Base\n");
      prtaliasesDB(aliasesDB, rc1, maxusers);  
   }
   if( prtrev)
   {  
      if( rc2 == 0)
         rc2 = 10000;
      printf("\n\n Reverse Aliases Data Base\n");
      prtaliasesDB(raliasesDB, rc2, maxusers);
   }                                  
   if( (endtime = times(&tmsend)) == -1) 
   {
      printf("end times error.\n");
   }
   if( (clktck = sysconf(_SC_CLK_TCK)) < 0)
   {
      printf("sysconf error getting clock ticks.\n");
      return(1);
   }
   duration = endtime - starttime;
   printf("\n\nElapsed time       = %8.2f seconds.\n",(duration/(double) clktck));
   printf("User CPU time      = %8.2f seconds.\n",
	 (tmsend.tms_utime - tmsstart.tms_utime)/((double) clktck));
   printf("System CPU time    = %8.2f seconds.\n",
	 (tmsend.tms_stime - tmsstart.tms_stime)/((double) clktck));
   return(0);
}  



/*============================================================================ 
 * print usage message
 *===========================================================================*/
void usage(void)
{
   
   printf("\n usage: adam [-a] [-b] [-da name|filename] [-dr name|filename]\n\
             [-p [all|aliases|rev] ]");
   printf("\n             [-h]");
   printf("\n        where:\n");
   printf("        -a    add new records using newaliases file as input.\n");
   printf("        -b    build aliases and reverse aliases data bases.\n");
   printf("        -da   delete aliases. Use name or fully qualified filename\n");
   printf("        -dr   delete revers aliases. Use name or fully qualified filename.\n");   
   printf("        -p    print aliases DB | reverse Aliases DB or both. \n");
   printf("        -h    print usage information.\n");
   return;
}   
