/*============================================================================
 * file:	util/adam/bldaliases.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*===========================================================================
 * function:	The bldaliases and bldraliases functions build the aliases
 *		and reverse aliases Data Bases using the aliases file. 
 *
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.
 * All rights reserved. 
 * 
 *                                                                      
 * Author:	Jim Livermore                                                          
 * Date written:	June 	  26, 1995
 * Last modified:	February   8, 2000
 *
 *=========================================================================*/
#include	<sys/types.h>  
#include	<stdio.h>
#include	<ctype.h>
#include	<syslog.h> 
#include	<string.h>
#include	<fcntl.h>
#include	<sys/ioctl.h>
#include	<sys/errno.h> 
#include	<sys/signal.h>
#include	<sys/stat.h>
#include	<arpa/inet.h>
#include	<sys/time.h>
#include	<time.h> 
#include	<sys/times.h>
#include	<unistd.h> 
#include	<stdlib.h> 
#include	"firewall.h"
#include	"adam.h"
#include	"smwrap.h"

extern int h_errno;  


/*================================================================
 * function prototypes
 *===============================================================*/
int updatealiasescnt(DBM *db, int count);
 
/*===========================================================================
 * function: build the aliases data base used by sendmail wrapper 
 * inputs:
 *
 * outputs:
 *
 * returns:  	 0 if ok
 * 		-1 on error 
 *==========================================================================*/
int bldaliases(char * aliasesfile, char *aliasesDB, int maxusers)
{  
#define MAXLEN 1024 
   DBM	*db;
   datum key; 
   
   datum data;  
   FILE	*alfp;                            
   char *cp;
   char	*cp2;              
   float avgkeysize	= 0;
   float avgrecsize     = 0;
   int	commentlines	= 0;
   int	i;
   int	line		= 0;
   int	largestkey	= 0;		/* largest key found    */ 
   int	largestrec  	= 0;		/* largest record found */
   int	numduplicates 	= 0; 
   int	numrecs 	= 0; 
   int  numrejects      = 0;
   int	rc		= 0;
   char	record[MAXLEN]; 
   static char	tokensep[]   = "\t,:;= \n";
   char	*token1; 
   char	*token2;  
   int totalkeysize	= 0;
   int totalrecsize	= 0;
   /*==============================
    * open aliases file 
    *==============================*/
   if (!(alfp = fopen(aliasesfile, "r")))
   {           
      printf("\nError: open failed for %s.",aliasesfile);
      syslog(LOG_ERR,"Error: open failed for %s: %m",aliasesfile);
      return(-1);    			/* open failed signal error */
   } 
   /*========================================================================
    * open aliases Data Base
    *======================================================================*/
   if( (db = dbm_open(aliasesDB,O_RDWR | O_CREAT | O_TRUNC,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) == NULL)
   {
      printf("\nError: unable to open %s.\n",aliasesDB); 
      fclose(alfp);
      return(-1);
   }

   cp = (char *) &record;         
   /*===================================================== 
    * Read each line of the aliases file. 
    *====================================================*/        
   while( fgets(record, MAXLEN, alfp) != NULL)  
   {
      line++;   
      /* skip over comments and null lines */
      if( (*cp == '#') || (*cp == '\n') )
      {
	 commentlines++; 
         goto getnextline1; 
      }
      if( (token1 = strtok(record,tokensep)) == NULL) goto getnextline1; 
      /*================================================================
       * check for valid name@host.domain
       *================================================================*/
      if( ( key.dsize = strlen(token1)) > 200 )
      {
         printf("\nError: invalid key name found on line %d of %s.\n \
      Length is > 200 characters.\n       '%.44s...'",line,aliasesfile,token1); 
         goto breakout;
      } 
      if( key.dsize > largestkey) largestkey = key.dsize;  
      totalkeysize = totalkeysize + key.dsize;
      /*================================================================
       * get second address and check for valid name@host.domain
       *================================================================*/
      if( (token2 = strtok(NULL,tokensep)) == NULL)
      {
         printf("\nError: NULL secondary name@host.domain found on line %d "
		"of %s.",
                line,aliasesfile); 
      }
      if( ( data.dsize = strlen(token2)) > 200 )
      {
         printf("\nError: invalid aliases name found on line %d of %s.\n "
                "Length is > 200 characters.\n      "
                " '%.44s...'",line,aliasesfile,token2);
         goto breakout; 
      } 
      if( data.dsize > largestrec) largestrec = data.dsize;
      totalrecsize = totalrecsize + data.dsize; 
      /*==================================================
       * Convert key and data to lower case
       *=================================================*/
      cp2 = token1;
      for( i = 0; i < strlen(token1); i++)
      {
         *cp2 =  tolower((int)*cp2);
         cp2++;      
      }
      cp2 = token2;
      for( i = 0; i < strlen(token2); i++)
      {
         *cp2 =  tolower((int)*cp2);
         cp2++;      
      }     
      key.dptr  = token1;
      data.dptr = token2;   
      /*=======================================================================
       * If the number of records is less than the maximum then store the 
       * record in Data Base, else stop. 
       *=====================================================================*/ 
      if( numrecs >= maxusers)
      {
         printf("\nError: The record '%s %s' \n"
                "       on line %d of %s was not added as the maximum \n"
                "       number of allowed users (%d) has been reached.",
                 key.dptr, data.dptr, line , aliasesfile, maxusers);   
         printf("\n       None of the remaining entries were processed.\n");  
         break;
      }     
      if( (rc = dbm_store(db, key, data, DBM_INSERT)) != 0)
      {      
         if(rc == 1)
         {        
	    printf("\nWarning: duplicate key, record not stored, input line ="
                   " %d of file %s.\n",
	           line, aliasesfile);  
            numduplicates++;
         }
         else
         {
	    printf("\nError: dbm_store failed for record number %d, input line ="
                   " %d of file %s. rc = %d\n",
	           numrecs, line, aliasesfile, rc);  
            numrejects++;
         }
         printf("      key = %s, data = %s",key.dptr,data.dptr);
      } 
      else
         numrecs++;
      getnextline1:; 
   } 
   /*=======================================================================
    * Update the alias count in the data base.
    *=====================================================================*/ 
   if( updatealiasescnt(db, numrecs) < 0)
   {
      printf("\n Error: update of total record count failed.\n");       
      return(-1);
   }  
   /*=======================================================================
    * print  statistics 
    *=======================================================================*/
   printf("\n\nSummary of Build for aliasesDB\n\n"); 
   printf("input file      = %s\n",aliasesfile);              
   printf("data base name  = %s\n", aliasesDB);
   printf("number of lines read from input file = %7d\n", line);     
   printf("number of records stored in DB       = %7d\n", numrecs);  
   printf("number of comment lines              = %7d\n", commentlines);
   printf("number of duplicate records          = %7d\n", numduplicates);
   printf("number of rejected records           = %7d\n", numrejects);
   printf("maximum number authorized            = %7d\n", maxusers);
   printf("number of unused records available   = %7d\n", maxusers-numrecs);
   printf("largest key size                     = %7d\n", largestkey); 
   if( numrecs > 0 )
      printf("average key size                     = %9.1f\n",
                     totalkeysize/(1.0*numrecs));
   printf("total key size                       = %7d\n",totalkeysize);
   printf("largest record size                  = %7d\n",largestrec);
   if( numrecs > 0 )
      printf("average record size                  = %9.1f\n",
                     totalrecsize/(1.0*numrecs));
   printf("total record size                    = %7d\n",totalrecsize); 
   fclose(alfp);
   dbm_close(db);
   return(numrecs);

updatefailed:
   /*=======================================================================
    * print  partial statistics 
    *=======================================================================*/
   printf("\n\nSummary of Build for aliasesDB prior to Error message\n\n"); 
   printf("input file      = %s\n",aliasesfile);              
   printf("data base name  = %s\n", aliasesDB);
   printf("number of lines read from input file = %7d\n", line);     
   printf("number of records stored in DB       = %7d\n", numrecs);
   printf("maximum number authorized            = %7d\n",maxusers);
   printf("number of unused records available   = %7d\n",maxusers-numrecs);
   printf("largest key size                     = %7d\n",largestkey); 
   if( numrecs > 0 )
      printf("average key size                     = %9.1f\n",
                       totalkeysize/(1.0*numrecs));
   printf("total key size                       = %7d\n",totalkeysize);
   printf("largest record size                  = %7d\n",largestrec);
   if( numrecs > 0 )
      printf("average record size                  = %9.1f\n",
                       totalrecsize/(1.0*numrecs));
   printf("total record size                    = %7d\n",totalrecsize);   

breakout:
   fclose(alfp);
   dbm_close(db);
   return(-1);
}    
/*===========================================================================
 * function: build the Reverse aliases data base used by sendmail wrapper 
 * inputs:
 *
 * outputs:
 *
 * returns:  	 record count if ok
 * 		-1 on error 
 *==========================================================================*/
int bldraliases(char * aliasesfile, char *raliasesDB, int maxusers)
{  
#define MAXLEN 1024 
   DBM	*db;
   datum key; 
   
   datum data;  
   FILE	*alfp;                            
   char	workarea[256];
   char	*cp;              
   char	*cp2;
   float avgkeysize	= 0;
   float avgrecsize     = 0;
   int	count		= 0;   
   int	i;
   int	line		= 0;
   int	largestkey	= 0;		/* largest key found    */ 
   int	largestrec  	= 0;		/* largest record found */
   int	numduplicates 	= 0;
   int  numrecs         = 0; 
   int	rc		= 0;
   char	record[MAXLEN]; 
   static char	tokensep[]   = "\t,:;= \n";
   char	*token1; 
   char	*token2;
   char	*token3;
   int totalkeysize	= 0;
   int totalrecsize	= 0;
   /*==============================
    * open aliases file 
    *==============================*/
   if (!(alfp = fopen(aliasesfile, "r")))
   {           
      printf("\nError: open failed for %s.",aliasesfile);
      syslog(LOG_ERR,"Error: open failed for %s: %m",aliasesfile);
      return(-1);    			/* open failed signal error */
   } 
   /*========================================================================
    * open reverse aliases Data Base
    *======================================================================*/
   if( (db = dbm_open(raliasesDB,O_RDWR | O_CREAT | O_TRUNC,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) == NULL)
   {
      printf("\nError: unable to open %s.\n",raliasesDB); 
      fclose(alfp);
      return(-1);
   }

   cp = (char *) &record;         
   /*===================================================== 
    * Read each line of the aliases file. 
    *====================================================*/ 
   while( fgets(record, MAXLEN, alfp) != NULL)  
   {
      line++;       
   
      /* skip over comments and null lines */
      if( (*cp == '#') || (*cp == '\n') ) goto getnextline1; 
      if( (token1 = strtok(record,tokensep)) == NULL) goto getnextline1; 
      /*================================================================
       * check for valid name@host.domain
       *================================================================*/
      if( ( data.dsize = strlen(token1)) > 200 )
      {
         printf("\nError: invalid key name found on line %d of %s.\n \
      Length is > 200 characters.\n       '%.44s...'",line,aliasesfile,token1); 
         goto breakout;
      } 

      /*================================================================
       * get second address and check for valid name@host.domain
       *================================================================*/
      if( (token2 = strtok(NULL,tokensep)) == NULL)
      {
         printf("\nError: NULL secondary name@host.domain found on line %d of %s."
                 ,line,aliasesfile); 
      }
      if( ( key.dsize = strlen(token2)) > 200 )
      {
         printf("\nError: invalid aliases name found on line %d of %s.\n \
      Length is > 200 characters.\n       '%.44s...'",line,aliasesfile,token2);
         goto breakout; 
      } 
      /*==================================================
       * Convert key and data to lower case
       *=================================================*/
      cp2 = token1;
      for( i = 0; i < strlen(token1); i++)
      {
         *cp2 = tolower((int)*cp2);
         cp2++;      
      }
      cp2 = token2;
      for( i = 0; i < strlen(token2); i++)
      {
         *cp2 = tolower((int)*cp2);
         cp2++;      
      }  
      key.dptr = token2;
      data.dptr = token1;
      /*====================================================
       * store record using internal name      
       *====================================================*/
      if( (rc = dbm_store(db, key, data, DBM_INSERT)) != 0)
      {  
         if( rc == 1)
         {
            printf("Duplicate record ignored, key  = %s,\n\
                          data = %s,\n\
                          record number = %d, input line = %d.\n",
	    	    key.dptr, data.dptr, numrecs, line);
            numduplicates++;
            goto getnextline1;
         }
         printf("\nError: dbm_store failed for record %d, input line = %d. rc = %d\n",
	        numrecs,line,rc);  
         goto getnextline1;
      }    
      else
      {   
         if( data.dsize > largestrec) largestrec = data.dsize;  
         totalrecsize = totalrecsize + data.dsize;
         if( key.dsize > largestkey) largestkey = key.dsize;
         totalkeysize = totalkeysize + key.dsize;
         numrecs++;   
         count++;
      }             
      /*====================================================
       * store record using internal name minus @host.domain      
       *====================================================*/
      strcpy(workarea,token2);   
      token3 = strtok(workarea,"@"); 
      key.dptr = token3; 
      key.dsize = strlen(token3);
      if( strcmp(token3,token2) >= 0)
         goto getnextline1; 
      if( (rc = dbm_store(db, key, data, DBM_INSERT)) != 0)
      { 
         if( rc == 1)
         {
            printf("Duplicate record ignored, key  = %s,\n\
                          data = %s,\n\
                          record number = %d, input line = %d.\n",
	    	    key.dptr, data.dptr, numrecs, line);
            numduplicates++;
            goto getnextline1;
         }
         printf("\nError: dbm_store failed for record %d, input line = %d. rc = %d\n",
	        numrecs,line,rc);  
         goto getnextline1;     
         
      }    
      else
      {   
         if( data.dsize > largestrec) largestrec = data.dsize;  
         totalrecsize = totalrecsize + data.dsize;
         if( key.dsize > largestkey) largestkey = key.dsize;
         totalkeysize = totalkeysize + key.dsize;
        
         numrecs++;  
      } 


      getnextline1:; 
   }                                                              
   /*=======================================================================
    *
    *=====================================================================*/ 
   if( updatealiasescnt(db, count) < 0)
   {
      printf("\n Error: update of total record count failed fro Reverse Aliases DB.\n");       
      return(-1);
   }  
   /*=======================================================================
    * print  statistics 
    *=======================================================================*/
   printf("\n\nBuild Summary for Reverse Aliases Data Base\n\n"); 
   printf("input file      = %s\n",aliasesfile);              
   printf("data base name  = %s\n", raliasesDB);
   printf("number of lines read from input file  = %7d\n", line);     
   printf("number of records stored in Rev DB    = %7d\n", numrecs);    
   printf("number of unique records stored in DB = %7d\n",count);
   printf("maximum number authorized             = %7d\n",maxusers);
   printf("number of unused records available    = %7d\n",maxusers-count);
   printf("number of duplicates rejected         = %7d\n", numduplicates);
   printf("largest key size                      = %7d\n",largestkey); 
   if( numrecs > 0 )
      printf("average key size                      = %9.1f\n",totalkeysize/(1.0*numrecs));
   printf("total key size                        = %7d\n",totalkeysize);
   printf("largest record size                   = %7d\n",largestrec);
   if( numrecs > 0 )
      printf("average record size                   = %9.1f\n",totalrecsize/(1.0*numrecs));
   printf("total record size                     = %7d\n",totalrecsize); 

   dbm_close(db);
   return(numrecs);
breakout:
   fclose(alfp);
   dbm_close(db);
   return(-1);
} 

