/*============================================================================
 * file:	util/adam/delaliases.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 * function:	Deletes aliases records from aliases & reverse aliases
 *		Data Bases.  This routine is called by adam.
 * 
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  All rights
 * reserved. 
 *                                                        
 * Date written:	July 	  24, 1995
 * Last modified:	February   8, 2000 - added LPL
 *		
 *
 *===========================================================================*/
#include	<sys/types.h>  
#include	<stdio.h>
#include	<ctype.h>
#include	<syslog.h> 
#include	<string.h>
#include	<fcntl.h>
#include	<sys/ioctl.h>
#include	<sys/errno.h> 
#include	<sys/signal.h>
#include	<sys/stat.h>
#include	<arpa/inet.h>
#include	<sys/time.h>
#include	<time.h> 
#include	<sys/times.h>
#include	<unistd.h> 
#include	<stdlib.h>
#include	"adam.h"
#include	"firewall.h"
#include	"smwrap.h"

extern int h_errno;  


/*================================================================
 * function prototypes
 *===============================================================*/   
int cntaliasesDB(DBM *db);  
int updatealiasescnt(DBM *db, int count);
/*============================================================================
 * function: delete user(s) from the aliases data base 
 * inputs:
 *		Name of the user to be deleted or the file containing multiple
 *		names to be deleted. 
 *		Name of aliases Data Baase 
 *
 * outputs:
 *
 * returns:  	 0 if ok
 * 		-1 on error 
 *==========================================================================*/
int delaliases(char * delaliases, char *aliasesDB, int maxusers)
{  
#define MAXLEN 1024 
   DBM		*db;  
   datum	data;
   datum 	key;  
   FILE		*alfp; 
   int		count	  = 0;
   int 		origcount = 0;  
   char		*cp;
   char         *cp2; 
   int		i;
   char		Key[32];             
   int		line	= 0;
   int		numrecs	= 0; 
   int		rc	= 0;
   char	record[MAXLEN]; 
   static char	tokensep[]   = "\t,:;= \n";
   char	*token1; 
   /*===========================================================
    * If name starts with '/' then open the delete aliases file 
    *==========================================================*/
   if( *delaliases == '/')
   {
      if (!(alfp = fopen(delaliases, "r")))
      {           
         printf("\nError: open failed for %s.", delaliases);
         syslog(LOG_ERR,"Error: open failed for %s: %m", delaliases);
         return(-1);    			/* open failed signal error */
      }  
   }
   /*========================================================================
    * open aliases Data Base
    *======================================================================*/
   if( (db = dbm_open(aliasesDB,O_RDWR,0) ) == NULL)
   {
      printf("\nError: unable to open %s.\n",aliasesDB); 
      fclose(alfp);
      return(-1);
   }
   /*========================================================================
    * Fetch the aliases record count. 
    *======================================================================*/ 
   strcpy(Key,"ALIASES_Count");  
   key.dsize = strlen(Key); 
   key.dptr  = &Key[0];
   data = dbm_fetch(db,key); 
   if( data.dptr == NULL)
   {   
      printf("\n Error: aliases record count is missing.\n");
      origcount = cntaliasesDB(db);
      count = origcount; 
      if( updatealiasescnt(db, count) < 0)
      {
         printf("\n Error: update of record count failed.\n");       
         return(-1);
      }  
      else
      {
         printf("\n Aliases count automatically updated. Error condition \
automatically fixed.\n");
      }
   }
   else
   {
      origcount = atoi(data.dptr); 
      count = origcount; 
   }
   cp = (char *) &record; 
   if( *delaliases == '/')
   {        
    /*===================================================== 
     * Read each line of the delete aliases file. 
     *====================================================*/        
     while( fgets(record, MAXLEN, alfp) != NULL)  
     {
        line++;   
        /* skip over comments and null lines */
        if( (*cp == '#') || (*cp == '\n') ) goto getnextline1; 
        if( (token1 = strtok(record,tokensep)) == NULL) goto getnextline1; 
        /*================================================================
         * check for valid name@host.domain
         *================================================================*/
         if( ( key.dsize = strlen(token1)) > 200 )
         {
            printf("\nError: invalid key name found on line %d of %s.\n \
      Length is > 200 characters.\n       '%.44s...'",line,delaliases,token1); 
            goto breakout;
         } 
         /*==================================================
          * Convert key to lower case
          *=================================================*/
         cp2 = token1;
         for( i = 0; i < strlen(token1); i++)
         {
            *cp2 = tolower((int)*cp2);
            cp2++;      
         }
 
         key.dptr  = token1; 
         /*==================================================
          * Delete record from Data Base
          *=================================================*/      
         if( (rc = dbm_delete(db, key)) != 0)
         {  
            printf("\nError: dbm_delete failed for %s. rc = %d", key.dptr, rc);    
 
         }  
         else
         {
            count--; 
            numrecs++; 
         }
         getnextline1:; 
      } 
      /*=======================================================================
       * update count record
       *======================================================================*/
      if( updatealiasescnt(db, count) < 0)
      {
         printf("\n Error: update of record count failed.\n");       
         return(-1);
      }  
      /*=======================================================================
       * print  statistics 
       *======================================================================*/
      printf("\n\nSummary of Aliases DB deletes\n\n"); 
      printf("input file      = %s\n",delaliases);              
      printf("data base name  = %s\n", aliasesDB);
      printf("number of lines read from input file = %d\n", line);
      printf("beginning number of records in DB    = %d\n",origcount);      
      printf("number of records deleted from DB    = %d\n", numrecs);  
      printf("number of records remaining in DB    = %d\n", count);  
      printf("number of unused records available   = %d\n",maxusers-count);
      fclose(alfp);
      dbm_close(db);
      return(numrecs);

#ifdef never
updatefailed:
      /*=======================================================================
       * print  partial statistics 
       *=======================================================================*/
      printf("\n\nSummary of Aliases DB deletes prior to Error message\n\n"); 
      printf("input file      = %s\n",delaliases);              
      printf("data base name  = %s\n", aliasesDB);
      printf("number of lines read from input file = %d\n", line);     
      printf("number of records deleted from DB    = %d\n", numrecs);
#endif

breakout:
      fclose(alfp);
      dbm_close(db);
      return(-1);  
   } /* end of code to process delaliases file	*/ 
   key.dptr  = delaliases;
   key.dsize = strlen(delaliases); 
   /*==================================================
    * Delete record from Data Base
    *=================================================*/      
   if( (rc = dbm_delete(db, key)) != 0)
   {      
      printf("\nError: dbm_delete failed for %s. rc = %d", key.dptr, rc);
      return(rc);
   }
   printf("\nRecord deleted. key = %s.",key.dptr);   
   /*=======================================================================
    * update count record
    *======================================================================*/ 
   count--;
   if( updatealiasescnt(db, count) < 0)
   {
      printf("\n Error: update of record count failed.\n");       
      return(-1);
   }  
   dbm_close(db);
   return(0);
}    

