/*============================================================================
 * file:	util/adam/prtaliasesDB.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 * function:	Print the aliases Data Base sorted by key.
 *
 * written by:		R. James Livermore
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  All rights
 * reserved.   	
 *                                                          
 * Date written:	July      17, 1995
 * Last modified:	February   8, 2000
 *              		
 *
 *===========================================================================*/
#include	<sys/types.h> 
#include	<search.h> 
#include	<stdio.h>
#include	<ctype.h>
#include	<syslog.h> 
#include	<string.h>
#include	<fcntl.h>
#include	<sys/ioctl.h>
#include	<sys/errno.h> 
#include	<sys/signal.h>
#include	<sys/stat.h>
#include	<arpa/inet.h>
#include	<sys/time.h> 
/*#include	<unistd.h>*/
#include	<stdlib.h>
#include	"adam.h"
#include	"firewall.h"
#include	"smwrap.h"

extern int h_errno;  
/*==========================================
 * structure used to sort e-mail addresses
 *=========================================*/
struct EMAILADDRS
{
   char from[200]; 
   char to[200];
};

/*================================================================
 * function prototypes
 *===============================================================*/   

int compraddrs( char *s1, char *s2);
/*extern "C" int getsmwp(struct SMWRAP *smwpptr); */
/*int prtaliasesDB(char *aliasesDB, int maxaddrs);*/ 

/*===========================================================================
 * function: print aliasesDB used by sendmail wrapper 
 * inputs:
 *
 * outputs: 
 *		name of aliases Data Base
 *		maximum number of records for sort
 *
 * returns:  	 0 if ok
 * 		-1 on error 
 *==========================================================================*/
int prtaliasesDB(char * aliasesDB, int maxrecords, int maxusers)
{ 
   DBM	*db;
   datum key; 
   datum data; 
   char	Key[200];
   char Data[200]; 
   int	i;  
   int	numrecs	  	= 0;
   int	totalrecs 	= 0; 
   int	totalkeysize 	= 0;
   int	totaldatasize	= 0;
   struct	EMAILADDRS	*p; 	/* pointer to data area used by sort*/
   struct	EMAILADDRS	*p2; 
   if( maxrecords < 1)
   {      
      printf("\n Invalid maxrecords passed to printaliasesDB.\n");
      return(-1); 
   }             
    p = (struct EMAILADDRS *) malloc(sizeof(struct EMAILADDRS)*(maxrecords+1));
   if( p == NULL )
   {
      printf("\nError: malloc failed for %d bytes.\n",
		sizeof(struct EMAILADDRS)*(maxrecords+1));
      return(-1);
   }             
   p2 = p;
#ifndef linux
   if( (db = dbm_open(aliasesDB,O_RDONLY,NULL)) == NULL)
#else
   if( (db = dbm_open(aliasesDB,O_RDONLY,0600)) == NULL)
#endif
   {
      printf("\ndbm_open Error for %s.\n",aliasesDB); 
      return(-1);
   } 
   printf("\n  Data Base Name: %s.\n",aliasesDB);
   printf("\n                    record_key  data\n"); 
   /*==================================================================
    * read the entire data base
    *==================================================================*/
   for( key = dbm_firstkey(db); key.dptr != NULL; key = dbm_nextkey(db))
   {  
      if( key.dptr == NULL)
      {
         printf("\nError: null key pointer returned.\n");  
         dbm_close(db);   
         free(p);
         return(-1); 
      }
      data = dbm_fetch(db,key);
      if( data.dptr == NULL)
      {
         printf("\nError: null data pointer returned in prtaliasesDB. key = %s\n",key.dptr);
         dbm_close(db); 
         free(p);
         return(-1);
      } 
      if( strncmp(key.dptr,"ALIASES_Count",key.dsize) == 0 )
      {       
         strncpy(Data,data.dptr,data.dsize); 
         Data[data.dsize] = '\0';   		/* truncate record */
         totalrecs = atoi(Data);
         continue;
      }
      numrecs++;                     
      strncpy(Key,key.dptr,key.dsize);
      Key[key.dsize] = '\0';   
      strncpy(Data,data.dptr,data.dsize); 
      Data[data.dsize] = '\0';   
      totalkeysize = totalkeysize + key.dsize;
      totaldatasize = totaldatasize + data.dsize;
      /*============================================= 
       * copy key and data into array for sorting
       *===========================================*/
      strcpy(p2->from,Key);
      strcpy(p2->to,Data);
      p2++;   
   } 
   dbm_close(db);
   qsort(p,numrecs, sizeof(struct EMAILADDRS),
	 	(int(*) (const void *, const void *))compraddrs); 
   p2 = p;
   for( i = 0; i < numrecs; i++)
   {
      printf("\n%30s %s",p2->from, p2->to); 
      p2++;
   }
 
   printf("\n\nPrint Aliases DB Summary Report\n\n");
   printf("Number of records read              = %d\n",numrecs);
   printf("Internal Record count               = %d\n",totalrecs); 
   printf("Maximum number authorized           = %d\n", maxusers);
   printf("Number of unused records available  = %d\n",maxusers-totalrecs); 
   printf("Total key size                      = %d\n",totalkeysize);    
   printf("Total data size                     = %d\n",totaldatasize); 
   free(p);
   return(0);
}
/*===========================================================================
 * compare E-mail addresses
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int compraddrs(char *s1, char *s2)
{
   struct EMAILADDRS *p1;
   struct EMAILADDRS *p2;
   p1 = (struct EMAILADDRS *) s1;
   p2 = (struct EMAILADDRS *) s2;
   return( strcmp(p1->from,p2->from) );
} 
