/*============================================================================
 * filename:		util/gwuser/convert_db.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * (C) COPYRIGHT R. James Livermore 1997-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1997-2000.  All rights
 *	reserved.    
 *										
 * written by:		Jim Livermore
 *
 * Date written:	January   28, 1997
 * Last modified:	February   8, 2000 - added LPL
 *	                         
 * function:	Convert the /etc/firewall/gwuser database to NDBM format 
 *              (.dir, .pag) if the latter files don't exist
 *
 *============================================================================*/
#include 	<fcntl.h> 
#include 	<errno.h> 

#include 	<stdio.h>
#include 	<stdlib.h>
#include 	<strings.h>
#include 	<syslog.h>
#include 	<sys/file.h>
#include 	<sys/stat.h> 
#include	<time.h>
#include 	<unistd.h>
#include 	"firewall.h"
#include 	"gwuser.h"

#ifdef linux
extern const char *const sys_errlist[];
#else
extern char *sys_errlist[];
#endif

extern DBM *db;
extern struct GWCONF conf;
extern struct GWADMIN admin;
extern int maxusers;

#ifndef linux
#define DBMDIRF db->dbm_dirf
#else
#define DBMDIRF dbm_dirfno(db)
#endif
/*============================================================================
 * Convert DB
 *===========================================================================*/
int convert_db()
{
   static char dirfile[] = GWUSERFILE".dir";
   static char pagfile[] = GWUSERFILE".pag";
   struct stat sbuf;
   datum key;
   datum data;
   FILE *fp;
   struct GWUSER_OLD oldrec;
   struct GWUSER newrec;
   struct GWGROUP group;
   GWUSER_ENCRYPT encrypted_data;
   int statdirfile;
   int statpagfile;
   int ret = 0;
   int nrec = 0;
   int nusers = 0;
   int ngroups = 0;
   int i;
   static char lastgroup[GRPSIZ];
   char string1[14];
   char string2[14];

   /* Check for authorized admin */
   if (strcmp(admin.user,"root")) 
   {
      fprintf(stderr,"Must be root to convert database (admin = %s).\n",
		     admin.user);
      syslog(LOG_NOTICE,"Must be root to convert database (admin = %s).\n",
		     admin.user);
      return(-1);
   }

   /*=======================================================
    * Check that NDBM database files don't already exist.
    *=======================================================*/
   statdirfile = stat(dirfile,&sbuf);
   statpagfile = stat(pagfile,&sbuf);
   if (statdirfile == 0 && statpagfile == 0) 
   {
      printf("NDBM database already exists, conversion not needed.\n");
      return(1);
   }
   else if (statdirfile == 0) 
   {
      fprintf(stderr,"Error: NDBM database partially exists, %s.dir is "
		     "missing.\n",GWUSERFILE);
      syslog(LOG_ERR,"NDBM database partially exists, %s.dir is "
		     "missing.\n",GWUSERFILE);
      return(-1);
   }
   else if (statpagfile == 0) 
   {
      fprintf(stderr,"Error: NDBM database partially exists, %s.pag is "
		     "missing.\n",GWUSERFILE);
      syslog(LOG_ERR,"Error: NDBM database partially exists, %s.pag is "
		     "missing.\n",GWUSERFILE);
      return(-1);
   }
  /*========================================================================
   * If non-NDBM database exists and is non-zero size then convert the 
   * old file to NDBM format, else create a new NDBM file using the default
   * user group. 
   *=======================================================================*/
/*
   if (stat(GWUSERFILE,&sbuf) != 0) 
   {
      fprintf(stderr,"Database %s does not exist.\n",GWUSERFILE);
      return(-1);
   }
   else if (sbuf.st_size == 0) 
   {
      fprintf(stderr,"Database %s is zero size.\n",GWUSERFILE);
      return(-1);
   }
*/
   if( stat(GWUSERFILE,&sbuf) != 0 || sbuf.st_size == 0 ) 
   {
      printf("Creating new gwuser database.\n");
      syslog(LOG_NOTICE,"Creating new gwuser database.");      
      return(create_db());
   }


   printf("Converting gwuser database to NDBM format...\n");
   syslog(LOG_NOTICE,"Converting gwuser database to NDBM format..."); 

   /* Open old database for reading */
   if ((fp = fopen(GWUSERFILE,"r")) == NULL) 
   {
      fprintf(stderr,"Error: fopen failed for %s: %s.\n",
		     GWUSERFILE,sys_errlist[errno]);
      syslog(LOG_ERR,"Error: fopen failed for %s: %m.",GWUSERFILE);
      return(-1);
   }

   /* Open NDBM files for writing */
#ifndef linux
   if ((db = dbm_open(GWUSERFILE,O_WRONLY|O_CREAT,0600)) == NULL) 
#else
   if ((db = dbm_open(GWUSERFILE,O_RDWR|O_CREAT,0600)) == NULL)
#endif
   {
      fprintf(stderr,"Error: dbm_open failed for %s.\n",GWUSERFILE);
      syslog(LOG_ERR,"Error: dbm_open failed for %s.",GWUSERFILE);
      fclose(fp);
      return(-1);
   }

   /* Lock NDBM database (.dir file) */
/*   if (lockf(db->dbm_dirf,F_LOCK,0) < 0) */
   if (lockf(DBMDIRF,F_LOCK,0) < 0)
   {
      fprintf(stderr,"Error: lockf lock failed (%s).\n",sys_errlist[errno]);
      syslog(LOG_ERR,"Error: lockf lock failed (%s).\n",sys_errlist[errno]);
      dbm_close(db);
      return(-1);
   }

   /* Read old database and create new records */
   while (fread(&oldrec,sizeof(oldrec),1,fp)) 
   {
      nrec++;

      /* Check that max user count not reached */
      if (nusers >= maxusers) 
      {
         fprintf(stderr,"Error: Maximum licensed user count of %d reached.\n",
			maxusers);
         syslog(LOG_ERR,"Error: Maximum licensed user count of %d reached.\n",
			maxusers);
         ret = -1;
         break;
      }
 
      /*=====================================
       * Write group record, if needed 
       *====================================*/
      if (strcmp(oldrec.groupname,lastgroup)) 
      {
         strncpy(lastgroup,oldrec.groupname,GRPSIZ);
         key.dptr = oldrec.groupname;
         key.dsize = strlen(oldrec.groupname);
         group.type = GROUP;
         strncpy(group.groupname,oldrec.groupname,GRPSIZ);
         group.realname[0] = '\0';
         group.tokentype = conf.tokentype;
         group.tnaccess = conf.tnaccess;
         group.ftpaccess = conf.ftpaccess;
         memcpy(group.accesstimes,conf.accesstimes,sizeof(group.accesstimes));
         group.recorddate = time((time_t *)NULL);
/*
         data.dptr = (char *) &group;
         data.dsize = sizeof(group);
*/
         encrypt_record((char *)&group,encrypted_data,sizeof(group));
         data.dptr = encrypted_data;
         data.dsize = encrypted_sizeof(group);
         if (ret = dbm_store(db,key,data,DBM_REPLACE)) 
         {
            fprintf(stderr,"Error: dbm_store failed for record number %d: "
			   "ret = %d.", nrec,ret);
            fprintf(stderr,"\n       group = %s, data size = %d.\n",
			   key.dptr,data.dsize);
            syslog(LOG_ERR,"Error: dbm_store failed for record number %d: "
			   "ret = %d.",nrec,ret);
            syslog(LOG_ERR,"group = %s, data size = %d.",key.dptr,data.dsize);
            ret = -1;
            break;
         }
         ngroups++;

         /*===================================================================
          * Write summary information record - update this record for every 
          * group record so will be up-to-date in case of an error termination 
          *==================================================================*/
         if (info_db(GWINFO_CREATE,0,nusers,ngroups)) 
         {
            ret = -1;
            break;
         }
      }

      /*==============================================
       * Write user record 
       *=============================================*/
      newrec.type = USER;
      strncpy(newrec.username,oldrec.username,UNMSIZ);
      newrec.realname[0] = '\0';
      newrec.phonenum[0] = '\0';
      strncpy(newrec.groupname,oldrec.groupname,GRPSIZ);
      memcpy(newrec.password[0],oldrec.password,PWDSIZ);
      for (i = 1; i < MAXPWDS; i++)
         newrec.password[i][0] = '\0';
      newrec.passworddate = time((time_t *)NULL);
      newrec.tokentype = oldrec.tokentype;
      if (strlen(oldrec.userkey))
         strncpy(newrec.userkey,oldrec.userkey,sizeof(newrec.userkey));
      else
         randomkey(newrec.userkey);
      newrec.ftpaccess = oldrec.ftpaccess;
      newrec.tnaccess = oldrec.tnaccess;
      memset(newrec.accesstimes,0,sizeof(newrec.accesstimes));
      newrec.startdate = 0;
      newrec.stopdate = 0;
      newrec.recorddate = time((time_t *)NULL);
      newrec.no_login_attempts = 0;
      newrec.status = UNLOCKED;  
      
      key.dptr = newrec.username;
      key.dsize = strlen(newrec.username);
/*
      data.dptr = (char *) &newrec;
      data.dsize = sizeof(newrec);
*/
      encrypt_record((char *)&newrec,encrypted_data,sizeof(newrec));
      data.dptr = encrypted_data;
      data.dsize = encrypted_sizeof(newrec);
      if (ret = dbm_store(db,key,data,DBM_INSERT)) 
      {
         fprintf(stderr,"Error: dbm_store failed for record number %d: ret = %d.", nrec,ret);
         fprintf(stderr,"\n       user = %s, data size = %d.\n",key.dptr,data.dsize);
         syslog(LOG_ERR,"Error: dbm_store failed for record number %d: ret = %d.",nrec,ret);
         syslog(LOG_ERR,"user = %s, data size = %d.",key.dptr,data.dsize);
         ret = -1;
         break;
      }
      nusers++;

    /*====================================================================== 
     * Write summary information record - update this record for every
     * user record so will be up-to-date in case of an error termination 
     *=====================================================================*/
      if (info_db(GWINFO_CREATE,0,nusers,ngroups)) 
      {
         ret = -1;
         break;
      }
   }

   fclose(fp);

   /*========================================
    * Unlock NDBM database (.dir file) 
    *=======================================*/
/*   if (lockf(db->dbm_dirf,F_ULOCK,0) < 0) */
   if (lockf(DBMDIRF,F_ULOCK,0) < 0)
   {
      fprintf(stderr,"Error: lockf unlock failed (%s).\n",sys_errlist[errno]);
      syslog(LOG_ERR,"Error: lockf unlock failed (%s).\n",sys_errlist[errno]);
   }

   dbm_close(db);

   /* Print summary */
   if (ngroups == 1)
      strcpy(string1,"group record");
   else
     strcpy(string1,"group records");
   if (nusers == 1)
     strcpy(string2,"user record");
   else
     strcpy(string2,"user records");
   if (ret >= 0) 
   {
      printf("Database conversion complete: %d %s, %d %s written.\n",
             ngroups,string1,nusers,string2);
      syslog(LOG_NOTICE,"Database conversion complete: %d %s, %d %s written.",
                        ngroups,string1,nusers,string2);
   }
   else 
   {
      fprintf(stderr,"Database conversion incomplete: %d %s, %d %s written.\n",
                     ngroups,string1,nusers,string2);
      syslog(LOG_ERR,"Database conversion incomplete: %d %s, %d %s written.",
                     ngroups,string1,nusers,string2);
   }

   return(ret);
} 
/*============================================================================  
 * Create a new gwuser DB when old file does not exist. 
 *===========================================================================*/
int create_db()
{
   static char dirfile[] = GWUSERFILE".dir";
   static char pagfile[] = GWUSERFILE".pag";
   struct stat sbuf;
   datum key;
   datum data;


   struct GWGROUP group;
   GWUSER_ENCRYPT encrypted_data;
   int statdirfile;
   int statpagfile;
   int ret = 0;
   int nrec = 0;
   int nusers = 0;
   int ngroups = 1;
   int i;
   static char lastgroup[GRPSIZ];

          
   /* Open NDBM files for writing */
#ifndef linux
   if ((db = dbm_open(GWUSERFILE,O_WRONLY|O_CREAT,0600)) == NULL) 
#else
   if ((db = dbm_open(GWUSERFILE,O_RDWR|O_CREAT,0600)) == NULL)
#endif
   {
      fprintf(stderr,"Error: dbm_open failed for %s.\n",GWUSERFILE);
      syslog(LOG_ERR,"Error: dbm_open failed for %s.",GWUSERFILE);
      return(-1);
   }

   /*==================================
    * Lock NDBM database (.dir file) 
    *==================================*/
/*   if (lockf(db->dbm_dirf,F_LOCK,0) < 0) */
   if (lockf(DBMDIRF,F_LOCK,0) < 0)
   {
      fprintf(stderr,"Error: lockf lock failed (%s).\n",sys_errlist[errno]);
      syslog(LOG_ERR,"Error: lockf lock failed (%s).\n",sys_errlist[errno]);
      dbm_close(db);
      return(-1);
   }
   /*============================================
    * Format Default group record. 
    *============================================*/
   key.dptr          = conf.groupname;
   key.dsize         = strlen(conf.groupname);
   group.type	     = GROUP;
   strncpy(group.groupname,conf.groupname,GRPSIZ);
   group.realname[0] = '\0';
   group.tokentype   = conf.tokentype;
   group.tnaccess    = conf.tnaccess;
   group.ftpaccess   = conf.ftpaccess;
   memcpy(group.accesstimes,conf.accesstimes,sizeof(group.accesstimes));
   group.recorddate  = time((time_t *)NULL);  
   /*============================================
    * Encrypt the group record. 
    *============================================*/
   encrypt_record((char *)&group,encrypted_data,sizeof(group));
   data.dptr = encrypted_data;
   data.dsize = encrypted_sizeof(group);
   if (ret = dbm_store(db,key,data,DBM_REPLACE)) 
   {
      fprintf(stderr,"Error: dbm_store failed for record number %d: "
			   "ret = %d.", nrec,ret);
      fprintf(stderr,"\n       group = %s, data size = %d.\n",
			   key.dptr,data.dsize);
      syslog(LOG_ERR,"Error: dbm_store failed for record number %d: "
			   "ret = %d.",nrec,ret);
      syslog(LOG_ERR,"group = %s, data size = %d.",key.dptr,data.dsize);
      ret = -1;
            
   }
   /*===================================================================
    * Write summary information record - update this record for every 
    * group record so will be up-to-date in case of an error termination 
    *==================================================================*/
   if (info_db(GWINFO_CREATE,0,nusers,ngroups)) 
   {
      ret = -1;
     
   }   
   /*===================================================================
    * Unlock NDBM database (.dir file) 
    *==================================================================*/
/*   if (lockf(db->dbm_dirf,F_ULOCK,0) < 0) */
   if (lockf(DBMDIRF,F_ULOCK,0) < 0)
   {
      fprintf(stderr,"Error: lockf unlock failed (%s).\n",sys_errlist[errno]);
      syslog(LOG_ERR,"Error: lockf unlock failed (%s).\n",sys_errlist[errno]);
   }

   dbm_close(db);  
   if (ret >= 0) 
   {
      printf("Database creation complete: %d group record, %d user record "
	     "written.\n",
             ngroups,nusers);
      syslog(LOG_NOTICE,"Database creation complete: %d group record, %d user "
		        "record written.",
                        ngroups,nusers);
   }
   else 
   {
      fprintf(stderr,"Database creation incomplete: %d group record, %d user"
		     "records %s written.\n",
                     ngroups,nusers);
      syslog(LOG_ERR,"Database creation incomplete: %d group record, %d "
		     "user record written.",
                     ngroups,nusers);
   }

   return(ret); 
}
