/*============================================================================
 * filename:		util/gwuser/get_times.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * (C) COPYRIGHT R. James Livermore 1997-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1997-2000.
 * All rights reserved.    
 *										
 * written by:		Jim Livermore
 *
 * Date written:	April     23, 1997
 * Last modified:	February   8, 2000 - added LPL
 *	                         
 * function:	Get gwuser/gwgroup access times argument   
 *
 * inputs:	argument count
 *		ptr to arg ptrs
 *		ptr to access times array    
 *
 * outputs:     returns
 *			n number of arguments following "-times"
 *			-1 on error
 *	              
 *
 *============================================================================*/

#include 	<fcntl.h>
#include 	<stdio.h>
#include 	<stdlib.h>
#include 	<strings.h>
#include 	"firewall.h"
#include 	"gwuser.h"

#if defined(_AIX) && ! defined(_AIX41)
   extern "C" int strcasecmp( char *, char *);  
   extern "C" int tolower(int);
#endif 

extern struct GWOPTIONS options;

int get_times(int argc, char **argv, char *access)
{

   typedef	struct
   {
      char	*name;
     unsigned char hrs[3];
   } DAYS;

   static	DAYS	days[] = 
   {  
      "Sunday",		'\0','\0','\0',    
      "Monday",		'\0','\0','\0',  
      "Tuesday",	'\0','\0','\0', 
      "Wednesday",	'\0','\0','\0',   
      "Thursday",	'\0','\0','\0',  
      "Friday",		'\0','\0','\0', 
      "Saturday",	'\0','\0','\0', 

   };
   int	argcount  =  0;        
   char	*cp;
   int	firstday  = -1;
   int	firsthour = -1; 
   unsigned char hours[3];
   int  i;
   int	index1;
   int  index2; 
   int  j;
   int	lastday	  = -1;
   int  lasthour  = -1;  
   char temp[128];
   memset(hours,'\0',3); 
   /*========================================================================= 
    *Initialize days[i].hrs to zeros, since function is called multiple times.
    *=======================================================================*/     
   for( i = 0; i < 7; i++)
   {   
      memset(days[i].hrs,0x00,3);
   } 
   if (argc == 0 || **argv == '-') 
   {
      fprintf(stderr,"Error: access times must follow -times.\n");
      return(-1);
   }

   if (!strcasecmp(*argv,"ALL"))
   {
      memset(access,0xff,21);
      return(1);
   }
   else if (!strcasecmp(*argv,"NONE"))
   {
      memset(access,0,21);
      return(1);
   }             
   /*=========================================================
    * If not ALL|NONE then must have two arguments: day time	
    *========================================================*/
   if( argc < 2 ) 
   { 
      fprintf(stderr,"Error: access times must have at least two arguments"
		     " following -times, unless using ALL|NONE");
      return(-1); 
   }         
   /*===============================================
    * Check which days are allowed 
    *==============================================*/
   while( (argcount < argc ) && (**argv != '-')  )
   {
      argcount++; 
      switch( **argv )
      {
      		case 'm': 
		case 'M':
			firstday = 1;  
			lastday  = 1;
                        break;
		case 't':
		case 'T':
                        if( *(*(argv)+1) == 'u')
                        {
			   firstday = 2; 
			   lastday  = 2; 
			}
                        else if( *(*(argv)+1) == 'h')
			{ 	               
                           firstday = 4;  
			   lastday  = 4;
			}
			else
                        {
                           fprintf(stderr,"Error: '%s' is invalid, "
					  "must use Tu or Th.\n",
					  *argv);
                           return(-1);                           
                        }
			break;
          	case 'w':
		case 'W':
			   firstday = 3;  
			   lastday  = 3;
                        break;      
         	case 'f':
		case 'F':
			firstday = 5;  
			lastday  = 5;
                        break;       
         	case 's':  
		case 'S':
                        if( *(*(argv)+1) == 'a') 
			{
			   firstday = 6;
			   lastday  = 6;
			}
                        else if( *(*(argv)+1) == 'u') 
			{ 	               
                           firstday = 0;
                           lastday  = 0;   
			}
                        else
                        {
                           fprintf(stderr,"Error: '%s' is invalid, "
					  "must use Sa or Su.\n",*argv);
                           return(-1);
                        }
			break;
		default:
			return(-1);  
         }
      
      /*=================================================
       * If there is a dash then a last day is specified 
       *=================================================*/                               
      if( (cp = strchr(*argv,'-')) != NULL)
      {  
         cp++;
         switch( *cp )
         {
      		case 'm':
		case 'M':
			lastday  = 1;
                        break;
		case 't':
		case 'T':
                        if( *(cp+1) == 'u')
			   lastday  = 2;
                        else if( *(cp+1) == 'h')	               
			   lastday  = 4;   
                        else 
                        {
                           fprintf(stderr,"Error: '%s' is invalid, "
					  "must use Tu or Th.\n",
					  *argv); 
                           return(-1);
                        }  
			break;
          	case 'w':
		case 'W':
			   lastday  = 3;
                        break;      
         	case 'f':
		case 'F':
			lastday  = 5;
                        break;       
         	case 's':
		case 'S':   
                        if( *(cp+1) == 'a')
			   lastday  = 6;
                        else if( *(cp+1) == 'u')	               
	                   lastday  = 0;         
                        else
                        {
                           fprintf(stderr,"Error: '%s' is invalid, "
					  "must use Sa or Su.\n",
					  *argv); 
                           return(-1);
                        }  
			break;
		default:        
                        fprintf(stderr,"Error: '%s' contains an invalid, "
					  "day.\n",
					  *argv); 
			return(-1);
         }
      }           
      /*==========================
       * Get times 
       *=========================*/	
      argcount++;
      argv++;     
      strncpy(temp,*argv,127);
      if( (cp = strchr(temp,'-')) == NULL )
      {	
         return(-1);
      }
      *cp = '\0';  
      if( temp[strlen(temp)-1] == 'p' ||  temp[strlen(temp)-1] == 'P' )
      {
         temp[strlen(temp)-1] = '\0'; 
         firsthour = atoi(temp);
         if( firsthour < 1 || firsthour > 12 )
         {
            fprintf(stderr,"Error: the starting time '%s' is invalid, "
			   "must be > 0 and < 13.\n", *argv); 
            return(-1);           
         }  
         /*  Make exception for 12p which is noon.  */     
         if( firsthour < 12)
            firsthour = firsthour + 12;  
      } 
      else if( temp[strlen(temp)-1] == 'a' ||  temp[strlen(temp)-1] == 'A' )
      { 
         temp[strlen(temp)-1] = '\0';    
         firsthour = atoi(temp); 
         if( firsthour < 1 || firsthour > 12 )
         { 
            fprintf(stderr,"Error: the starting time '%s' is invalid, "
			   "must be > 0 and < 13.\n",*argv); 
            return(-1);    
         }  
         /* Make exception for 12a which is midnight  */
         if( firsthour == 12 )
            firsthour = 0;      
      }
      else 
      /*====================================================================
       * If neither am or pm is specified then assume use of 24 hour clock. 
       *===================================================================*/
      {
         firsthour = atoi(temp);
      }                 
      /*==========================
       * get lasthour 
       *=========================*/
      cp = strchr(*argv,'-');
      cp++;
      strncpy(temp,cp,127);
      if( temp[strlen(temp)-1] == 'p' ||  temp[strlen(temp)-1] == 'P' )
      {
         temp[strlen(temp)-1] = '\0';   
	 lasthour = atoi(temp);  
         if( lasthour < 1 || lasthour > 12 )
         {
            fprintf(stderr,"Error: the ending time '%s' is invalid, "
			   "must be > 0 and < 13.\n", *argv); 
            return(-1);           
         } 
         /*  Make exception for 12p which is noon.  */     
         if( lasthour < 12)
            lasthour = lasthour + 12;  
      } 
      else if( temp[strlen(temp)-1] == 'a' ||  temp[strlen(temp)-1] == 'A' )
      { 
         temp[strlen(temp)-1] = '\0';    
         lasthour = atoi(temp); 
         if( lasthour < 1 || lasthour > 12 )
         { 
            fprintf(stderr,"Error: the ending time '%s' is invalid, "
			   "must be > 0 and < 13.\n",*argv);  
            return(-1);   
         }  
         /* Make exception for 12a which is midnight  */
         if( lasthour == 12 )
            lasthour = 24;    
      }
      else
      {
         lasthour = atoi(temp);
      }              
      if( lasthour <= firsthour )
      {
         return(-1);
      }
      /* set times for each day */     
      for( i = firstday; i <= lastday; i++)
      {
         for(j = firsthour; j < lasthour; j++)
         {
             /* accesstimes is 21 X 8 hourly bitmap */
            index1 = j / 8;
            index2 = 1 << (j - index1 * 8);
            hours [index1] = hours[index1] | index2;
           
         }  
         days[i].hrs[0] |= hours[0];
         days[i].hrs[1] |= hours[1];
         days[i].hrs[2] |= hours[2];
         memset(hours,0x00,3);
      } 
      argv++;

   }
   cp = access;
   for(i = 0; i < 8; i++)
   {
      memcpy(cp,days[i].hrs,3);
      cp = cp +3; 
   }
   return(argcount);
}
