/*============================================================================
 * filename:		util/gwuser/gwuserconf.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=============================================================================	
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  
 * All rights reserved.    
 *										 *										
 * written by:		Jim Livermore
 *
 * Date written:	January   13, 1995
 * Last modified:	February   8, 2000 - added LPL
 *	                         
 * function:	Read the /etc/firewall/gwuser.conf configuration file.
 *
 *============================================================================*/
#include	<ctype.h>
#include	<errno.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<strings.h>
#include	<syslog.h>
#include	"firewall.h"
#include	"gwuser.h"

#if defined(_AIX) && ! defined(_AIX41)
   extern "C" int strcasecmp( char *, char *);
#endif 

#ifdef linux
extern const char *const sys_errlist[];
#else
extern char *sys_errlist[];
#endif

struct GWCONF conf;
extern struct GWOPTIONS options;

/* Maximum record length */
#define MAXLEN 512

/* Maximum unrecognized keyword error messages */
#define MAXUNRECOGNIZED 10

/* get_token macro */
#define get_token \
if ((token = strtok(NULL,tokensep)) == NULL) \
{ \
   if (print) \
   { \
      if (ret < 0) fprintf(stderr,"\n"); \
      fprintf(stderr,"Error: no value specified for %s on line %d of %s\n.", \
              keyword,line,GWUSERCFGFILE); \
   } \
   syslog(LOG_ERR,"Error: no value specified for %s on line %d of %s.", \
           keyword,line,GWUSERCFGFILE); \
   ret = -1; \
   continue; \
}
/*============================================================================  
 *
 *============================================================================*/
int gwuserconf()
{
   struct 
   {
     bool group;
     bool token;
     bool telnet;
     bool ftp;
     bool times;
     bool len;
     bool alpha;
     bool other;
     bool age;
     bool unique;
     bool attempts;
   } params;
   FILE *fp;
   static char gwusercfgfile[] = GWUSERCFGFILE;
   static char tokensep[] = "\t,:;= \n";
   char record[MAXLEN];
   char *keyword, *token;
   int print = 0;
   int line = 0;
   int nunrecognized = 0;
   int ret = 0;

   /*=========================================
    * Initialize mandatory parameter flags. 
    *=========================================*/

   params.group    = 0;
   params.token    = 0;
   params.telnet   = 0;
   params.ftp      = 0;
   params.times    = 0;
   params.len      = 0;
   params.alpha    = 0;
   params.other    = 0;
   params.age      = 0;
   params.unique   = 0;
   params.attempts = 0;

   /*===================================================
    * Print to stdout, stderr only for interactive use. 
    *==================================================*/

   if(!strcmp(options.program,"gwuser") ||
      !strcmp(options.program,"gwgroup"))
      print = 1;

   /*============================================
    * Read configuration file 
    *===========================================*/

   if ((fp = fopen(GWUSERCFGFILE,"r")) == NULL) 
   {
      if (print)
         fprintf(stderr,"Error: fopen failed for %s: %s.\n",GWUSERCFGFILE,
                 sys_errlist[errno]);
      syslog(LOG_ERR,"Error: fopen failed for %s: %m.",GWUSERCFGFILE);
      return(-1);
   }

   while (fgets(record,MAXLEN,fp))
   {
      line++;

      /* Skip over comments and null lines */
      if (record[0] == '#' || record[0] == '\n')
         continue;

      /* Get keyword */
      if ((keyword = strtok(record,tokensep)) == NULL)
         continue;
      /*======================================
       * Default group 
       *======================================*/  
      if (!strcasecmp(keyword,"default_group"))
      {
         params.group = 1;
         get_token;
         if (strlen(token) > (GRPSIZ - 1))
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"Error: Default group \"%s\" exceeds maximum "
                              "length\nof %d characters on line %d of %s.\n",
			      token,GRPSIZ-1,line,GWUSERCFGFILE);
            }
            syslog(LOG_ERR,"Error: Default group \"%s\" exceeds maximum "
			   "length of %d characters on line %d of %s.",
			   token,GRPSIZ-1,line,GWUSERCFGFILE);
            ret = -1;
         }
         else
            strcpy(conf.groupname,token);
      }  
      /*============================================
       * Default token type 
       *==========================================*/
      else if (!strcasecmp(keyword,"default_token"))
      {
         params.token = 1;
         get_token;
         if (!strcasecmp(token,"CRYPTO"))
            conf.tokentype = CRYPTO;
         else if (!strcasecmp(token,"SNK"))
            conf.tokentype = SNK;
#ifdef _AIX
         else if (!strcasecmp(token,"SDI"))
            conf.tokentype = SDI;
#endif
         else if (!strcasecmp(token,"NONE"))
            conf.tokentype = NONE;
         else
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"Error: Invalid default token type \"%s\"\n"
                              "on line %d of %s\n.",token,line,GWUSERCFGFILE);
            }
            syslog(LOG_ERR,"Error: Invalid default token type \"%s\" on "
                           "line %d of %s.",token,line,GWUSERCFGFILE);
            ret = -1;
         }
      }
      /*============================================
       * Default telnet access 
       *===========================================*/
      else if (!strcasecmp(keyword,"default_telnet"))
      {
         params.telnet = 1;
         get_token;
         conf.tnaccess = 0;
         do {
            if (!strcasecmp(token,"NO_TN"))
               conf.tnaccess |= NO_TN;
            else if (!strcasecmp(token,"UNPRO_TN"))
               conf.tnaccess |= UNPRO_TN;
            else if (!strcasecmp(token,"PRO_TNSA"))
               conf.tnaccess |= PRO_TNSA;
            else if (!strcasecmp(token,"PRO_TN"))
               conf.tnaccess |= PRO_TN;
            else if (!strcasecmp(token,"ADM_TN"))
               conf.tnaccess |= ADM_TN;
            else
            {
               if (print)
               {
                  if (ret < 0) fprintf(stderr,"\n");
                  fprintf(stderr,"Error: Invalid default telnet access "
				 "option \"%s\"\non line %d of %s.\n",
				 token,line,GWUSERCFGFILE);
               }
               syslog(LOG_ERR,"Error: Invalid default telnet access "
			      "option \"%s\" on line %d of %s.",
			      token,line,GWUSERCFGFILE);
               ret = -1;
            }
         } while ((token = strtok(NULL,tokensep)) != NULL);
         if (conf.tnaccess & PRO_TN && conf.tnaccess & PRO_TNSA)
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"Error: Invalid default telnet access option "
			      "on line %d of\n%s: PRO_TN and PRO_TNSA are "
			      "mutually exclusive.\n",line,GWUSERCFGFILE);
            }
            syslog(LOG_ERR,"Error: Invalid default telnet access option on "
			   "line %d of %s: PRO_TN and PRO_TNSA are mutually "
			   "exclusive.",line,GWUSERCFGFILE);
            ret = -1;
         }
         if (conf.tnaccess & ADM_TN && 
           !(conf.tnaccess & PRO_TNSA || conf.tnaccess & UNPRO_TN))
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"Error: Invalid default telnet access option on line %d of\n%s: ADM_TN also requires the specification of PRO_TNSA, UNPRO_TN, or both.\n",line,GWUSERCFGFILE);
            }
            syslog(LOG_ERR,"Error: Invalid default telnet access option on line %d of %s: ADM_TN also requires the specification of PRO_TNSA, UNPRO_TN, or both.",line,GWUSERCFGFILE);
            ret = -1;
         }
      }
      /*==========================================
       * Default FTP access 
       *=========================================*/
      else if (!strcasecmp(keyword,"default_ftp"))
      {
         params.ftp = 1;
         get_token;
         conf.ftpaccess = 0;
         do {
            if (!strcasecmp(token,"NO_FTP"))
               conf.ftpaccess |= NO_FTP;
            else if (!strcasecmp(token,"UNPRO_FTP"))
               conf.ftpaccess |= UNPRO_FTP;
            else if (!strcasecmp(token,"PRO_FTP"))
               conf.ftpaccess |= PRO_FTP;
            else if (!strcasecmp(token,"PRO_FTPSA"))
               conf.ftpaccess |= PRO_FTPSA;
            else if (!strcasecmp(token,"GET_FTP"))
               conf.ftpaccess |= GET_FTP;
            else if (!strcasecmp(token,"PUT_FTP"))
               conf.ftpaccess |= PUT_FTP;
            else if (!strcasecmp(token,"DEL_FTP"))
               conf.ftpaccess |= DEL_FTP;
            else
            {
               if (print)
               {
                  if (ret < 0) fprintf(stderr,"\n");
                  fprintf(stderr,"Error: Invalid default FTP access "
				 "option \"%s\"\non line %d of %s.\n",
				 token,line,GWUSERCFGFILE);
               }
               syslog(LOG_ERR,"Error: Invalid default FTP access "
			      "option \"%s\" on line %d of %s.",
			      token,line,GWUSERCFGFILE);
               ret = -1;
            }
         } while ((token = strtok(NULL,tokensep)) != NULL);
         if ((conf.ftpaccess & GET_FTP || conf.ftpaccess & PUT_FTP || 
              conf.ftpaccess & DEL_FTP) && !(conf.ftpaccess & PRO_FTP || 
              conf.ftpaccess & PRO_FTPSA || conf.ftpaccess & UNPRO_FTP))
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"GET_FTP, PUT_FTP, and DEL_FTP also require the specification of\nPRO_FTP, PRO_FTPSA, or UNPRO_FTP.\n");
            }
            syslog(LOG_ERR,"GET_FTP, PUT_FTP, and DEL_FTP also require the specification of PRO_FTP, PRO_FTPSA, or UNPRO_FTP.");
            ret = -1;
         }
      }
      /*=============================================
       * Default time access 
       *=============================================*/
      else if (!strcasecmp(keyword,"default_times"))
      {
         char **av;
         int ac=0;

         params.times = 1;
         memset(conf.accesstimes,0,sizeof(conf.accesstimes));

         av = (char **)malloc(sizeof(char *));
         if (!av)
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"Error: malloc failed for accesstimes, line %d "
			"of %s.\n",line,GWUSERCFGFILE);
            }
            syslog(LOG_ERR,"Error: malloc failed for accesstimes, line %d "
			"of %s.\n",line,GWUSERCFGFILE);
            ret = -1;
            continue;
         }
         get_token;
         *av = token;
         ac++;
         if (ret != -1)
         {
            while ((token = strtok(NULL,tokensep)))
            {
                av = (char **)realloc(av, (ac+1) * sizeof(char *));
                if (!av)
                {
                   if (print)
                   {
                      if (ret < 0) fprintf(stderr,"\n");
                      fprintf(stderr,"Error: realloc failed for accesstimes, "
				"line %d of %s.\n",line,GWUSERCFGFILE);
                   }
                   syslog(LOG_ERR,"Error: realloc failed for accesstimes, "
				"line %d of %s.\n",line,GWUSERCFGFILE);
                   ret = -1;
                   continue;
                }
		*(av+ac++) = token;
            }
            ret = get_times(ac, av, conf.accesstimes);
            if (ret > 0)
                ret = 0;
         }
      }
      /*==================================================
       * Password min length 
       *=================================================*/
      else if (!strcasecmp(keyword,"password_min_length"))
      {
         params.len = 1;
         get_token;
         if (!isalldigits(token))
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"Error: Non-numeric mininum password "
			      "length \"%s\"\non line %d of %s.\n",
			      token,line,GWUSERCFGFILE);
            }
            syslog(LOG_ERR,"Error: Non-numeric mininum password "
			   "length \"%s\" on line %d of %s.",
			   token,line,GWUSERCFGFILE);
            ret = -1;
            continue;
         }
         conf.password_min_len = atoi(token);
         if (conf.password_min_len < 4 || conf.password_min_len > 16)
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"Error: Invalid minimum password length of %d on line %d\nof %s: must be between 4 and 16.\n",conf.password_min_len,line,GWUSERCFGFILE);
            }
            syslog(LOG_ERR,"Error: Invalid minimum password length of %d on line %d of %s: must be between 4 and 16.",conf.password_min_len,line,GWUSERCFGFILE);
            conf.password_min_len = 0;
            ret = -1;
         }
      }
      /*==================================================
       * Password min alpha characters 
       *==================================================*/
      else if (!strcasecmp(keyword,"password_min_alpha"))
      {
         params.alpha = 1;
         get_token;
         if (!isalldigits(token))
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"Error: Non-numeric mininum number of password "
			      "alpha characters \"%s\"\non line %d of %s.\n",
			      token,line,GWUSERCFGFILE);
            }
            syslog(LOG_ERR,"Error: Non-numeric mininum number of password "
			   "alpha characters \"%s\" on line %d of %s.",
			   token,line,GWUSERCFGFILE);
            ret = -1;
            continue;
         }
         conf.password_min_alpha = atoi(token);
         if (conf.password_min_alpha < 0 || conf.password_min_alpha > 16)
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"Error: Invalid minimum number of password alpha characters value of %d\non line %d of %s: must be between 0 and 16.\n",conf.password_min_alpha,line,GWUSERCFGFILE);
            }
            syslog(LOG_ERR,"Error: Invalid minimum number of password alpha characters value of %d on line %d of %s: must be between 0 and 16.",conf.password_min_alpha,line,GWUSERCFGFILE);
            ret = -1;
         }
      }
      /*=================================================
       * Password min non-alpha characters 
       *================================================*/
      else if (!strcasecmp(keyword,"password_min_other"))
      {
         params.other = 1;
         get_token;
         if (!isalldigits(token))
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"Error: Non-numeric mininum number of password "
			     "non-alpha characters \"%s\"\non line %d of %s.\n",
			      token,line,GWUSERCFGFILE);
            }
            syslog(LOG_ERR,"Error: Non-numeric mininum number of password "
			   "non-alpha characters \"%s\" on line %d of %s.",
			   token,line,GWUSERCFGFILE);
            ret = -1;
            continue;
         }
         conf.password_min_other = atoi(token);
         if (conf.password_min_other < 0 || conf.password_min_other > 16)
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"Error: Invalid minimum number of password "
			      "non-alpha characters value of %d\non line %d "
			      "of %s: must be between 0 and 16.\n",
			      conf.password_min_other,line,GWUSERCFGFILE);
            }
            syslog(LOG_ERR,"Error: Invalid minimum number of password non-alpha"
			   " characters value of %d on line %d of %s: must be"
			   " between 0 and 16.",
			   conf.password_min_other,line,GWUSERCFGFILE);
            ret = -1;
            continue;
         }
         if (conf.password_min_len && params.alpha && params.other && 
            (conf.password_min_alpha + conf.password_min_other) > 
             conf.password_min_len)
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"Error: Minimum number of password alpha plus "
			      "non-alpha characters exceeds\nminimum password"
			      " length of %d in %s.\n",
			      conf.password_min_len,GWUSERCFGFILE);
            }
            syslog(LOG_ERR,"Error: Minimum number of password alpha plus "
			   "non-alpha characters exceeds minimum password "
			   "length of %d in %s.",
			   conf.password_min_len,GWUSERCFGFILE);
            ret = -1;
         }
      }
      /*==================================================
       * Password max change interval 
       *=================================================*/
      else if (!strcasecmp(keyword,"password_max_age"))
      {
         params.age = 1;
         get_token;
         if (!strcasecmp(token,"UNLIMITED"))
         {
            conf.password_max_age = 0;
            continue;
         }
         else if (!isalldigits(token))
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"Error: Unrecognized maximum password change "
			      "interval \"%s\"\non line %d of %s.\n",
			      token,line,GWUSERCFGFILE);
            }
            syslog(LOG_ERR,"Error: Unrecognized maximum password change "
			   "interval \"%s\" on line %d of %s.",
			   token,line,GWUSERCFGFILE);
            ret = -1;
            continue;
         }
         conf.password_max_age = atoi(token);
         if (conf.password_max_age < 1 || conf.password_max_age > 365)
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"Error: Invalid maximum password change "
			      "interval of %d on line %d\nof %s: must be "
			      "between 1 and 356.\n",
			      conf.password_max_age,line,GWUSERCFGFILE);
            }
            syslog(LOG_ERR,"Error: Invalid maximum password change interval "
			   "of %d on line %d of %s: must be between 1 and 356.",
			   conf.password_max_age,line,GWUSERCFGFILE);
            ret = -1;
         }
      }
      /*==================================================
       * Number of unique passwords before reuse 
       *==================================================*/
      else if (!strcasecmp(keyword,"no_unique_passwords"))
      {
         params.unique = 1;
         get_token;
         if (!strcasecmp(token,"NONE"))
         {
            conf.no_unique_passwords = 0;
            continue;
         }
         else if (!isalldigits(token))
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"Error: Unrecognized number of unique passwords before re-use \"%s\"\non line %d of %s.\n",token,line,GWUSERCFGFILE);
            }
            syslog(LOG_ERR,"Error: Unrecognized number of unique passwords before re-use \"%s\" on line %d of %s.",token,line,GWUSERCFGFILE);
            ret = -1;
            continue;
         }
         conf.no_unique_passwords = atoi(token);
         if (conf.no_unique_passwords < 1 || conf.no_unique_passwords > 12)
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"Error: Invalid number of unique passwords before re-use value of %d\non line %d of %s: must be between 1 and 12.\n",conf.no_unique_passwords,line,GWUSERCFGFILE);
            }
            syslog(LOG_ERR,"Error: Invalid number of unique passwords before re-use value of %d on line %d of %s: must be between 1 and 12.",conf.no_unique_passwords,line,GWUSERCFGFILE);
            ret = -1;
         }
      }
      /*================================================
       * Number of failed logins before locked 
       *===============================================*/
      else if (!strcasecmp(keyword,"no_login_attempts"))
      {
         params.attempts = 1;
         get_token;
         if (!strcasecmp(token,"UNLIMITED"))
         {
            conf.no_login_attempts = 0;
            continue;
         }
         else if (!isalldigits(token))
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"Error: Unrecognized number of login attempts "
			      "before locked \"%s\"\non line %d of %s.\n",
			      token,line,GWUSERCFGFILE);
            }
            syslog(LOG_ERR,"Error: Unrecognized number of login attempts "
			   "before locked \"%s\" on line %d of %s.",
			   token,line,GWUSERCFGFILE);
            ret = -1;
            continue;
         }
         conf.no_login_attempts = atoi(token);
         if (conf.no_login_attempts < 1 || conf.no_login_attempts > 12)
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"Error: Invalid number of login attempts before locked value of %d\non line %d of %s: must be between 1 and 12.\n",conf.no_login_attempts,line,GWUSERCFGFILE);
            }
            syslog(LOG_ERR,"Error: Invalid number of login attempts before locked value of %d on line %d of %s: must be between 1 and 12.",conf.no_login_attempts,line,GWUSERCFGFILE);
            ret = -1;
         }
      }
      /*=======================================
       * Unrecognized keyword 
       *=======================================*/
      else
      {
         if (nunrecognized < MAXUNRECOGNIZED)
         {
            if (print)
            {
               if (ret < 0) fprintf(stderr,"\n");
               fprintf(stderr,"Error: Unrecognized keyword \"%s\" on line %d\n"
                              "of %s.\n",keyword,line,GWUSERCFGFILE);
            }
            syslog(LOG_ERR,"Error: Unrecognized keyword \"%s\" on line %d "
                           "of %s.",keyword,line,GWUSERCFGFILE);
         }
         nunrecognized++;
         ret = -1;
      }
   }
   /*=====================================
    * Parsing complete 
    *====================================*/
   if (nunrecognized > MAXUNRECOGNIZED)
   {
      if (print)
      {
         if (ret < 0) fprintf(stderr,"\n");
         fprintf(stderr,"Error: %d additional unrecognized keywords found.\n",
                 nunrecognized-MAXUNRECOGNIZED);
      }
      syslog(LOG_ERR,"Error: %d additional unrecognized keywords found.",
             nunrecognized-MAXUNRECOGNIZED);
   }
   /*=====================================================
    * Check that mandatory parameters were specified 
    *====================================================*/
   if (!params.group)
   {
      if (print)
      {
         if (ret < 0) fprintf(stderr,"\n");
         fprintf(stderr,"Error: Default group \"default_group = <name>\"\n"
                        "must be specified in %s.\n",GWUSERCFGFILE);
      }
      syslog(LOG_ERR,"Error: Default group \"default_group = <name>\" "
                    "must be specified in %s.",GWUSERCFGFILE);
      ret = -1;
   }
   if (!params.token)
   {
      if (print)
      {
         if (ret < 0) fprintf(stderr,"\n");
         fprintf(stderr,"Error: Default token type \"default_token = <type>\"\n"
                        "must be specified in %s.\n",GWUSERCFGFILE);
      }
      syslog(LOG_ERR,"Error: Default token \"default_token = <type>\" "
                     "must be specified in %s.",GWUSERCFGFILE);
      ret = -1;
   }
   if (!params.telnet)
   {
      if (print)
      {
         if (ret < 0) fprintf(stderr,"\n");
         fprintf(stderr,"Error: Default telnet access \"default_telnet = "
		        "<options>\"\nmust be specified in %s.\n",
			GWUSERCFGFILE);
      }
      syslog(LOG_ERR,"Error: Default telnet access \"default_telnet = "
		     "<options>\" must be specified in %s.",GWUSERCFGFILE);
      ret = -1;
   }
   if (!params.ftp)
   {
      if (print)
      {
         if (ret < 0) fprintf(stderr,"\n");
         fprintf(stderr,"Error: Default FTP access \"default_ftp = "
			"<options>\"\n"
                        "must be specified in %s.\n",GWUSERCFGFILE);
      }
      syslog(LOG_ERR,"Error: Default FTP access \"default_ftp = <options>\" "
                     "must be specified in %s.",GWUSERCFGFILE);
      ret = -1;
   }
   if (!params.times)
   {
      if (print)
      {
         if (ret < 0) fprintf(stderr,"\n");
         fprintf(stderr,"Error: Default time access \"default_times = <option>\"\n"
                        "must be specified in %s.\n",GWUSERCFGFILE);
      }
      syslog(LOG_ERR,"Error: Default time access \"default_times = <option>\" "
                     "must be specified in %s.",GWUSERCFGFILE);
      ret = -1;
   }
   if (!params.len)
   {
      if (print)
      {
         if (ret < 0) fprintf(stderr,"\n");
         fprintf(stderr,"Error: Minimum password length \"password_min_length"
		        " = <value>\" must be\nspecified in %s.\n",
			GWUSERCFGFILE);
      }
      syslog(LOG_ERR,"Error: Minimum password length \"password_min_length = "
		     "<value>\" must be specified in %s.",GWUSERCFGFILE);
      ret = -1;
   }
   if (!params.alpha)
   {
      if (print)
      {
         if (ret < 0) fprintf(stderr,"\n");
         fprintf(stderr,"Error: Minimum number of password alpha characters"
			" \"password_min_alpha =\n<value>\" must be specified"
			" in %s.\n",GWUSERCFGFILE);
      }
      syslog(LOG_ERR,"Error: Minimum number of password alpha characters"
		     " \"password_min_alpha = <value>\" must be specified"
		     " in %s.",GWUSERCFGFILE);
      ret = -1;
   }
   if (!params.other)
   {
      if (print)
      {
         if (ret < 0) fprintf(stderr,"\n");
         fprintf(stderr,"Error: Minimum number of password non-alpha "
		        "characters \"password_min_other =\n<value>\" must "
			"be specified in %s.\n",GWUSERCFGFILE);
      }
      syslog(LOG_ERR,"Error: Minimum number of password non-alpha characters "
		    "\"password_min_other = <value>\" must be specified in %s.",
		     GWUSERCFGFILE);
      ret = -1;
   }
   if (!params.age)
   {
      if (print)
      {
         if (ret < 0) fprintf(stderr,"\n");
         fprintf(stderr,"Error: Maximum password change interval "
			"\"password_max_age = <value>\"\nmust be specified "
			"in %s.\n",GWUSERCFGFILE);
      }
      syslog(LOG_ERR,"Error: Maximum password change interval "
		     "\"password_max_age = <value>\" must be specified in %s.",
		     GWUSERCFGFILE);
      ret = -1;
   }
   if (!params.unique)
   {
      if (print)
      {
         if (ret < 0) fprintf(stderr,"\n");
         fprintf(stderr,"Error: Number of unique passwords before re-use "
		        "\"no_unique_passwords = <value>\"\nmust be specified "
			"in %s.\n",GWUSERCFGFILE);
      }
      syslog(LOG_ERR,"Error: Number of unique passwords before re-use "
		     "\"no_unique_passwords = <value>\" must be specified in"
		     " %s.",GWUSERCFGFILE);
      ret = -1;
   }
   if (!params.attempts)
   {
      if (print)
      {
         if (ret < 0) fprintf(stderr,"\n");
         fprintf(stderr,"Error: Number of login attempts before locked "
		        "\"no_login_attempts = <value>\"\nmust be specified "
			"in %s.\n",GWUSERCFGFILE);
      }
      syslog(LOG_ERR,"Error: Number of login attempts before locked "
		     "\"no_login_attempts = <value>\" must be specified in %s.",
		     GWUSERCFGFILE);
      ret = -1;
   }

   fclose(fp);
   return(ret);
}  
