/*============================================================================
 * filename:		util/gwuser/info_db.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * (C) COPYRIGHT R. James Livermore 1997-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1997-2000.  
 * All rights reserved.    
 * 
 * written by:  	Walt Croom
 *
 * Date written:	January   28, 1997
 * Last modified:	October    8, 1999 - added Linux support
 *			February   8, 2000 - added LPL
 *	                         
 * function:	Create/update gwuser database summary information record
 *
 *   mode:
 *     GWINFO_CREATE   Create record
 *     GWINFO_MODIFY   Modify record
 *     GWINFO_VERIFY   Verify/update record
 *
 *   option (not used for create)
 *     NUMUSERS    Update total number of user records
 *     NUMGROUPS   Update total number of group records
 *     MODIFYDATE  Update modify date for summary information record
 *
 *   nusers:
 *     Total (create & verify) or +/- (modify) number of user records
 *
 *   ngroups:
 *     Total (create & verify) or +/- (modify) number of group records
 *
 *   Note that this function neither opens nor closes the NDBM database.
 *                
 *============================================================================*/

#include 	<fcntl.h> 
#include 	<stdio.h>
#include 	<stdlib.h>
#include 	<strings.h>
#include 	<syslog.h> 
#include	<time.h>
#include 	"firewall.h"
#include 	"gwuser.h"

extern DBM *db;
extern struct GWOPTIONS options;

int info_db(char mode, char option, long nusers, long ngroups)
{
   datum key, data;
   struct GWINFO info;
   GWINFO_ENCRYPT encrypted_data;
   GWINFO_DECRYPT decrypted_data;
   int print = 0;
   int ret = 0;

   /* Print to stdout, stderr only for interactive use */
   if(!strcmp(options.program,"gwuser") ||
      !strcmp(options.program,"gwgroup"))
      print = 1;

   key.dptr = GWINFO_KEY;
   key.dsize = strlen(GWINFO_KEY);

   /* Create summary information record */
   if (mode == GWINFO_CREATE) 
   {
      info.version = GWVERSION;
      info.type = INFO;
      info.total_users = nusers;
      info.total_groups = ngroups;
      info.createdate = time((time_t *)NULL);
      info.modifydate = time((time_t *)NULL);
/*
      data.dptr = (char *) &info;
      data.dsize = sizeof(info);
*/
      encrypt_record((char *)&info,encrypted_data,sizeof(info));
      data.dptr = encrypted_data;
      data.dsize = encrypted_sizeof(info);
      if (ret = dbm_store(db,key,data,DBM_REPLACE)) 
      {
         if (print)
            fprintf(stderr,"Error: dbm_store failed for summary information "
                           "record.\n");
         syslog(LOG_ERR,"Error: dbm_store failed for summary information "
			"record.\n");
         ret = -1;
      }
   }

   /* Modify summary information record */
   else if (mode == GWINFO_MODIFY) 
   {
      data = dbm_fetch(db,key);
      if (data.dptr == NULL) 
      {
         if (print)
            fprintf(stderr,"Error: dbm_fetch failed for summary information "
                           "record.\n");
         syslog(LOG_ERR,"Error: dbm_fetch failed for summary information "
			"record.");
         ret = -1;
      }
      else 
      {
/*         bcopy(data.dptr,(char *)&info,data.dsize); */
         decrypt_record(data.dptr,decrypted_data,data.dsize);
         memcpy(&info,decrypted_data,sizeof(info));
         info.modifydate = time((time_t *)NULL);
         if (option & NUMUSERS)
           info.total_users += nusers;
         if (option & NUMGROUPS)
           info.total_groups += ngroups;
/*         data.dptr = (char *) &info; */
         encrypt_record((char *)&info,encrypted_data,sizeof(info));
         data.dptr = encrypted_data;
         if (ret = dbm_store(db,key,data,DBM_REPLACE)) 
         {
            if (print)
               fprintf(stderr,"Error: dbm_store failed for summary information "
                              "record.\n");
            syslog(LOG_ERR,"Error: dbm_store failed for summary information "
			   "record.");
            ret = -1;
         }
      }
   }

   /* Verify summary information record */
   else if (mode == GWINFO_VERIFY) 
   {
      data = dbm_fetch(db,key);
      if (data.dptr == NULL) 
      {
         if (print)
            fprintf(stderr,"Error: dbm_fetch failed for summary information "
                           "record.\n");
         syslog(LOG_ERR,"Error: dbm_fetch failed for summary information "
			"record.");
         ret = -1;
      }
      else 
      {
/*         bcopy(data.dptr,(char *)&info,data.dsize); */
         decrypt_record(data.dptr,decrypted_data,data.dsize);
         memcpy(&info,decrypted_data,sizeof(info));
         info.modifydate = time((time_t *)NULL);
         if (option & NUMUSERS) 
         {
            if (info.total_users != nusers) 
            {
               if (print)
                  fprintf(stderr,"\nWarning: Database summary information "
                                 "inconsistency for number of users.\n");
               syslog(LOG_WARNING,"Warning: Database summary information "
			      "inconsistency for number of users.");
               if (print)
                  fprintf(stderr,"Changing database summary information number "
                                 "of users from %d to %d.\n",info.total_users,
                                 nusers);
               syslog(LOG_WARNING,"Changing database summary information "
				  "number of users from %d to %d.",
			       info.total_users,nusers);
               info.total_users = nusers;
/*               data.dptr = (char *) &info; */
               encrypt_record((char *)&info,encrypted_data,sizeof(info));
               data.dptr = encrypted_data;
               if (dbm_store(db,key,data,DBM_REPLACE)) 
               {
                  if (print)
                     fprintf(stderr,"Error: dbm_store failed for summary "
                                    "information record.\n");
                  syslog(LOG_ERR,"Error: dbm_store failed for summary "
				 "information record.");
                  ret = -1;
               }
            }
         }
         if (option & NUMGROUPS) 
         {
            if (info.total_groups != ngroups) 
            {
               if (print)
                  fprintf(stderr,"\nWarning: Database summary information "
                                 "inconsistency for number of groups.\n");
               syslog(LOG_WARNING,"Warning: Database summary information inconsistency for number of groups.");
               if (print)
                  fprintf(stderr,"Changing database summary information number of groups from %d to %d.\n",info.total_groups,ngroups);
               syslog(LOG_WARNING,"Changing database summary information number of groups from %d to %d.",info.total_groups,ngroups);
               info.total_groups = ngroups;
/*               data.dptr = (char *) &info; */
               encrypt_record((char *)&info,encrypted_data,sizeof(info));
               data.dptr = encrypted_data;
               if (dbm_store(db,key,data,DBM_REPLACE)) 
               {
                  if (print)
                     fprintf(stderr,"Error: dbm_store failed for summary information record.\n");
                  syslog(LOG_ERR,"Error: dbm_store failed for summary information record.");
                  ret = -1;
               }
            }
         }
      }
   }

   return(ret);
}
