/*============================================================================
 * filename:		util/gwuser/listgroup.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================	
 *
 * (C) COPYRIGHT R. James Livermore 1997-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1997-2000. 
 * All rights reserved.    
 *
 * written by:  	Walt Croom
 * Date written:	January   28, 1997
 * Last modified:	October    8, 1999 - added Linux support
 *			February   8, 2000 - added LPL
 *	                         
 * function:	List gwuser group records
 *
 *============================================================================*/

#include 	<errno.h>
#include 	<fcntl.h>   
#include 	<stdio.h>
#include 	<stdlib.h>
#include 	<strings.h>
#include 	<syslog.h>   
#include	<time.h>
#include 	"firewall.h"
#include 	"gwuser.h"

#ifdef linux
extern const char *const sys_errlist[];
#else
extern  char    *sys_errlist[];
#endif
                                  
extern char *hourstring(char *hours);

extern struct GWOPTIONS options;
extern struct GWADMIN admin;
extern DBM *db;
extern int maxusers;

/* Max number of dbm_fetch errors to explicity report */
#define MAXERRORS 10

/* Key structure to sort */
struct KEY 
{
   char name[GRPSIZ];
   int len;
};

int compare_key(const void *arg1, const void *arg2)
{
   struct KEY *p1, *p2;

   p1 = (struct KEY *) arg1;
   p2 = (struct KEY *) arg2;
 
   return(strcmp((const char *)p1->name,(const char *)p2->name));
}

void display_group(struct GWGROUP group)
{
   struct table 
   {
      unsigned char value;
      char *name;
   } *ptable;

   static struct table token_table[] = 
   {
      {NONE,    "None"  },
      {CRYPTO,  "CRYPTO"},
      {SNK,     "SNK"   },
      {SDI,     "SDI"   },
      {0xff,    "Null"  }
   };

   static struct table tnaccess_table[] = 
   {
      {NO_TN,     "NO_TN"   },
      {PRO_TN,    "PRO_TN"  },
      {PRO_TNSA,  "PRO_TNSA"},
      {UNPRO_TN,  "UNPRO_TN"},
      {ADM_TN,    "ADM_TN"  },
      {0xff,      "Null"    }
   };

   static struct table ftpaccess_table[] =
   {    
      {NO_FTP,     "NO_FTP"   },
      {PRO_FTP,    "PRO_FTP"  },
      {UNPRO_FTP,  "UNPRO_FTP"},
      {PRO_FTPSA,  "PRO_FTPSA"},
      {GET_FTP,    "GET_FTP"  },
      {PUT_FTP,    "PUT_FTP"  },
      {DEL_FTP,    "DEL_FTP"  },
      {0xff,       "Null"     }
   };
   static struct table Days[] = 
   { 
      {0,      	"Sunday   "},
      {1,	"Monday   "},
      {2,    	"Tuesday  "},
      {3,  	"Wednesday"},
      {4,  	"Thursday "},
      {5,    	"Friday   "},  
      {6,    	"Saturday "}

   };
   int i;
   int defined;

   printf("Group ID:          %s\n",group.groupname);

   if (strlen(group.realname) > 0)
      printf("Real name:         %s\n",group.realname);
   else
      printf("Real name:         Null\n");

   for (ptable = token_table; ptable->value != (unsigned char) 0xff; ptable++) 
   {
      if (ptable->value == group.tokentype) 
      {
         printf("Token type:        %s\n",ptable->name);
         break;
      }
   }

   printf("Telnet access:     ");
   for (ptable = tnaccess_table; ptable->value != (unsigned char) 0xff; ptable++) 
   {
      if (ptable->value & group.tnaccess || ptable->value == group.tnaccess)
          printf("%s ",ptable->name);
   }
   printf("\n");

   printf("FTP access:        ");
   for (ptable = ftpaccess_table; ptable->value != (unsigned char) 0xff; ptable++) 
   {
      if (ptable->value & group.ftpaccess || ptable->value == group.ftpaccess)
         printf("%s ",ptable->name);
   }
   printf("\n");

   defined = 0;
   for (i = 0; i < sizeof(group.accesstimes); i++) 
   {
      if (group.accesstimes[i]) 
      {
         defined = 1;
         break;
      }
   }
   if(defined) 
   {
      printf("Access times:  Day/Hours\n");        
      printf("               000000000011111111112222\n"); 
      printf("     Day       012345678901234567890123\n");       
      for(i = 0; i < 7; i++)
      {
         printf("    %s  %s\n",
		Days[i].name, hourstring(&group.accesstimes[i*3]));   
      } 
      printf("\n");
   }
   else
      printf("Access times:      None\n");

   printf("Record date:       %s",ctime(&group.recorddate));
}   

/*============================================================================
 *
 *============================================================================*/
char *hourstring(char *hours)
{
static  char hrstring[25];
   int	i;
   int index1;
   int index2;

   memset(hrstring,'.',24);
   hrstring[24] = '\0';
   /* accesstimes is 21 X 8 hourly bitmap */
   for(i = 0; i < 24; i++)
   {
      index1 = i / 8;
      index2 = 1 << (i - index1 * 8);
      if((hours[index1] & index2))
      {
         hrstring[i] = 'Y';
      }
   }      
   return(&hrstring[0]);
}
/*============================================================================
 *
 *============================================================================*/
int listgroup()
{
   struct KEY *pkey, *pmalloc;
   int sizekey = sizeof(struct KEY);
   datum key, data;
   struct GWGROUP group;
   GWGROUP_DECRYPT decrypted_data;
   char reg[128];
   unsigned int nkeys = 0;
   int i;
   int j; 
   int match;
   int allrecords = 0;
   int allgroups  = 0;
   int wildgroup  = 0;
   int ngroups    = 0;
   int nerrors    = 0;
   int nlisted    = 0;
   int ret        = 0;

   /* Check whether any groups specified */
   if (options.ngroups == 0)
      allgroups = 1;

   else 
   {

      /* Check admin group priviledge */
      if (strcmp(admin.user,"root")) 
      {
         if (options.ngroups > 1 || strcmp(admin.group,options.groups[0])) 
         {
            fprintf(stderr,"Admin %s can list group %s only.\n",
                           admin.user,admin.group);
            syslog(LOG_NOTICE,"Admin %s can list group %s only.\n",
			   admin.user,admin.group);
            return(-1);
         }
      }

      /* Check for group wildcards */
      for (i = 0; i < options.ngroups; i++) 
      {
         if (!strcmp(options.groups[i],"*")) 
         {
            allgroups = 1;
            break;
         }
         else if (glob2reg(options.groups[i],reg,sizeof(reg)) > 0) 
         {
            wildgroup = 1;
            break;
         }
      }
   }

   /* Open NDBM database */
   if (open_db(EXIST) < 0)
      return(-1);

   /* Check whether to retrieve all keys */
   if(allgroups == 1 || wildgroup == 1)
      allrecords = 1;

   /* Retrieve all keys */
   if(allrecords == 1) 
   {
      if ((pmalloc = (struct KEY *) malloc(maxusers*sizekey)) == NULL) 
      {
         fprintf(stderr,"Error: malloc failed for %d bytes (%s).\n",
              		maxusers*sizekey,sys_errlist[errno]);
         syslog(LOG_ERR,"Error: malloc failed for %d bytes (%s).",
         		maxusers*sizekey,sys_errlist[errno]);
         close_db();
         return(-1);
      }
      pkey = pmalloc;
      for (key = dbm_firstkey(db); key.dptr != NULL; key = dbm_nextkey(db)) 
      {
         strncpy(pkey->name,key.dptr,key.dsize);
         pkey->len = key.dsize;
         pkey++;
         nkeys++;
      }
   }

   /* Use specified keys */
   else 
   {
      nkeys = options.ngroups;
      if ((pmalloc = (struct KEY *) malloc(nkeys*sizekey)) == NULL) 
      {
         fprintf(stderr,"Error: malloc failed for %d bytes (%s).\n",
              		nkeys*sizekey,sys_errlist[errno]);
         syslog(LOG_ERR,"Error: malloc failed for %d bytes (%s).",
             		nkeys*sizekey,sys_errlist[errno]);
         close_db();
         return(-1);
      }
      pkey = pmalloc;
      for (i = 0; i < nkeys; i++) 
      {
         strcpy(pkey->name,options.groups[i]);
         pkey->len = strlen(options.groups[i]);
         pkey++;
      }
   }

   /* Sort key list */
   pkey = pmalloc;
   if (nkeys > 1)
      qsort(pkey,nkeys,sizekey,compare_key);

   /* Retrieve records for sorted key list */
   for (i = 0; i < nkeys; i++, pkey++) 
   {
      key.dptr = pkey->name;
      key.dsize = pkey->len;
      data = dbm_fetch(db,key);

      /* Looping over all records */
      if (allrecords == 1) 
      {
        /*=====================================================================
         * Record should always exist since keys generated by 
         * dbm_first/nextkey 
         *==================================================================*/
         if (data.dptr == NULL) 
         {
            if (nerrors < MAXERRORS) 
            {
               fprintf(stderr,"Error: dbm_fetch failed for %s.\n",key.dptr);
               syslog(LOG_ERR,"Error: dbm_fetch failed for %s.\n",key.dptr);
            }
            nerrors++;
         }

         /* Check whether group record */
         else if (*(char *)data.dptr == GROUP) 
         {
            ngroups++;

            /* Match group pattern */
            if (!allgroups) 
            {
               match = 0;
               for (j = 0; j < options.ngroups; j++) 
               {
                  if (pattern_match(options.groups[j],key.dptr) == 0) 
                  {
                     match = 1;
                     break;
                  }
               }

               /* Go on to next record if no match */
               if (!match)
                  continue;
            }

            /* Check admin group priviledge */
            if( !strcmp(admin.user,"root") || 
                !strcmp(admin.group,key.dptr)) 
            {

               /* Display entire record */
               if (options.allinfo == 1) 
               {
                  /* bcopy(data.dptr,(char *)&group,data.dsize); */
                  decrypt_record(data.dptr,decrypted_data,data.dsize);
                  memcpy(&group,decrypted_data,sizeof(group));
                  if (nlisted > 0) printf("\n");
                  display_group(group);
               }

/* Display group ID */
          else {
            ((char *)key.dptr)[key.dsize] = '\0';
            printf("%s\n",key.dptr);
          }
          nlisted++;
        }
      }
    }

/* Looping over specified group ID's */
    else {

/* Check that record exists */
      if (data.dptr == NULL) {
        ((char *)key.dptr)[key.dsize] = '\0';
        fprintf(stderr,"%s not found.\n",key.dptr);
      }

/* Check whether group record */
      else if (*(char *)data.dptr == GROUP) {

/* Display entire record */
        if (options.allinfo == 1) {
/*          bcopy(data.dptr,(char *)&group,data.dsize); */
          decrypt_record(data.dptr,decrypted_data,data.dsize);
          memcpy(&group,decrypted_data,sizeof(group));
          if (nlisted > 0) printf("\n");
          display_group(group);
        }

        /* Display group ID */
        else 
        {
           ((char *)key.dptr)[key.dsize] = '\0';
           printf("%s\n",key.dptr);
        }
        nlisted++;
      }

/* Specified group ID is actually a user */
      else if (*(char *)data.dptr == USER) {
        ((char *)key.dptr)[key.dsize] = '\0';
        fprintf(stderr,"%s is a user.\n",key.dptr);
      }
    }
  }
  free(pmalloc);

/* Print group record summary for long listing */
  if (options.allinfo)
    printf("\nNumber of group records listed = %d.\n",nlisted);

/* Verify summary record number of groups, even though should never need */
/* to update */
  if (allgroups) {
    if(info_db(GWINFO_VERIFY,NUMGROUPS,0,ngroups))
      ret = -1;
  }

/* Print error summary (this is bad!) */
  if (nerrors > MAXERRORS) {
    fprintf(stderr,"Error: dbm_fetch failed for %d additional keys.\n",nerrors-MAXERRORS);
    syslog(LOG_ERR,"Error: dbm_fetch failed for %d additional keys.\n",nerrors-MAXERRORS);
  }

  close_db();
  return(ret);
}
