/*============================================================================
 * filename:		util/gwuser/listuser.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *	
 * (C) COPYRIGHT R. James Livermore 1997-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1997-2000.  All rights
 *	reserved.
 *
 * written by:  	Walt Croom
 * Date written:	January   28, 1997
 * Last modified:	February   8, 2000
 *	                         
 * function:	List gwuser user records
 *
 *============================================================================*/

#include 	<errno.h>
#include 	<fcntl.h>  
#include 	<stdio.h>
#include 	<stdlib.h>
#include 	<strings.h>
#include 	<syslog.h>  
#include	<time.h>
#include 	"firewall.h"
#include 	"gwuser.h"
#ifdef linux
extern const char *const sys_errlist[];
#else
extern char *sys_errlist[];
#endif

extern char *uhourstring(char *hours);

extern struct GWOPTIONS options;
extern struct GWADMIN admin;
extern struct GWCONF conf;
extern DBM *db;
extern int maxusers;

/* Max number of dbm_fetch errors to explicity report */
#define MAXERRORS 10

/* Key structure to sort */
struct KEY 
{
  char name[UNMSIZ];
  int len;
};

int compare_key(const void *arg1, const void *arg2)
{
  struct KEY *p1, *p2;

  p1 = (struct KEY *) arg1;
  p2 = (struct KEY *) arg2;
 
  return(strcmp((const char *)p1->name,(const char *)p2->name));
}
/*============================================================================
 *
 *===========================================================================*/
void display_user(struct GWUSER user)
{
   struct table 
   {
      unsigned char value;
      char *name;
   } *ptable;

   static struct table token_table[] = 
   {
      {NONE,    "Default"},
      {CRYPTO,  "CRYPTO" },
      {SNK,     "SNK"    },
      {SDI,     "SDI"    },
      {0xff,    "Null"   }
   };

   static struct table tnaccess_table[] = 
   {
      {NO_TN,     "Default" },
      {PRO_TN,    "PRO_TN"  },
      {PRO_TNSA,  "PRO_TNSA"},
      {UNPRO_TN,  "UNPRO_TN"},
      {ADM_TN,    "ADM_TN"  },
      {0xff,      "Null"    }
   };

   static struct table ftpaccess_table[] = 
   {
      {NO_FTP,     "Default"  },
      {PRO_FTP,    "PRO_FTP"  },
      {UNPRO_FTP,  "UNPRO_FTP"},
      {PRO_FTPSA,  "PRO_FTPSA"},
      {GET_FTP,    "GET_FTP"  },
      {PUT_FTP,    "PUT_FTP"  },
      {DEL_FTP,    "DEL_FTP"  },
      {0xff,       "Null"     }
   };
   static struct table Days[] = 
   {    
      {0,      	"Sunday   "},
      {1,	"Monday   "},
      {2,    	"Tuesday  "},
      {3,  	"Wednesday"},
      {4,  	"Thursday "},
      {5,    	"Friday   "},  
      {6,    	"Saturday "}

   };


   static struct table status_table[] = 
   {
      {LOCKED,    "Locked"},
      {UNLOCKED,  "Unlocked"},
   };

   time_t expdate;
   int password_length;
   int i;
   int defined;

   printf("User ID:             %s\n",user.username);

   if (strlen(user.realname) > 0)
      printf("Real name:           %s\n",user.realname);
   else
      printf("Real name:           Null\n");

   if (strlen(user.phonenum) > 0)
      printf("Phone number:        %s\n",user.phonenum);
   else
      printf("Phone number:        Null\n");

   if (strlen(user.groupname) > 0)
      printf("Group:               %s\n",user.groupname);
   else
      printf("Group:               Null\n");

   printf("Password:            ");
   if ((password_length = strlen(user.password[0])) > 0) 
   {
      for (i = 0; i < password_length; i++)
         printf("#");
      printf("\n");
      if (conf.password_max_age > 0) 
      {
         expdate = user.passworddate + conf.password_max_age * 86400;
         printf("Password expires:    %s",ctime(&expdate));
      }
      else
         printf("Password expires:    Never\n");
   }
   else
      printf("Null\n");

   for (ptable = token_table; ptable->value != (unsigned char) 0xff; ptable++) 
   {
      if (ptable->value == user.tokentype) 
      {
         printf("Token type:          %s\n",ptable->name);
         break;
      }
   }

   if (strlen(user.userkey))
    printf("Token key:           %3.3o %3.3o %3.3o %3.3o %3.3o %3.3o %3.3o %3.3o\n",
      user.userkey[0],user.userkey[1],user.userkey[2],user.userkey[3],
      user.userkey[4],user.userkey[5],user.userkey[6],user.userkey[7]);
   else
   /*======================================================================== 
    * Should never see this, since random key is generated when not specified 
    *=======================================================================*/
   printf("Token key:           Null\n");

   printf("Telnet access:       ");
   for (ptable = tnaccess_table; ptable->value != (unsigned char) 0xff; ptable++) 
   {
      if (ptable->value & user.tnaccess || ptable->value == user.tnaccess)
         printf("%s ",ptable->name);
   }
   printf("\n");

   printf("FTP access:          ");
   for (ptable = ftpaccess_table; ptable->value != (unsigned char) 0xff; ptable++) 
   {
      if (ptable->value & user.ftpaccess || ptable->value == user.ftpaccess)
         printf("%s ",ptable->name);
   }
   printf("\n");

   if (user.startdate > 0)
      printf("Start date:          %s",ctime(&user.startdate));
   else
      printf("Start date:          Now\n");

   if (user.stopdate > 0)
      printf("Stop date:           %s",ctime(&user.stopdate));
   else
      printf("Stop date:           Never\n");

   defined = 0;
   for (i = 0; i < sizeof(user.accesstimes); i++) 
   {
      if (user.accesstimes[i]) 
      {
         defined = 1;
         break;
      }
   }
   if (defined)
   {
   /*====================================================================== 
    * 
    *======================================================================*/ 
      printf("Access times:    Day/Hours\n");        
      printf("                 000000000011111111112222\n"); 
      printf("     Day         012345678901234567890123\n");       
      for(i = 0; i < 7; i++)
      {
         printf("      %s  %.24s\n",
		Days[i].name, uhourstring(&user.accesstimes[i*3]));   
      } 
      printf("\n");
   } 
   else
      printf("Access times:        Default\n");

   printf("Record date:         %s",ctime(&user.recorddate));
   printf("Unsuccessful logins: %d\n",
           user.no_login_attempts);
   for (ptable = status_table; ptable->value != (unsigned char) 0xff; ptable++) 
   {
      if (ptable->value == user.status)
      {
         printf("User status:         %s\n",ptable->name);
         break;
      }
   }  

}
/*============================================================================
 *
 *============================================================================*/

char *uhourstring(char *hours)
{
static  char hrstring[25];
   int	i;
   int index1;
   int index2;

   memset(hrstring,'.',24);
   hrstring[24] = '\0';
   /* accesstimes is 21 X 8 hourly bitmap */
   for(i = 0; i < 24; i++)
   {
      index1 = i / 8;
      index2 = 1 << (i - index1 * 8);
      if((hours[index1] & index2))
      {
         hrstring[i] = 'Y';
      }
   }      
   return(&hrstring[0]);
}

/*============================================================================
 *
 *============================================================================*/
int listuser()
{
   struct KEY *pkey, *pmalloc;
   int sizekey = sizeof(struct KEY);
   datum key, data;
   struct GWUSER user;
   GWUSER_DECRYPT decrypted_data;
   char reg[128];
   unsigned int nkeys = 0;
   int i;
   int j;
   int match;
   int allrecords = 0;
   int allgroups = 0;
   int wildgroup = 0;
   int allusers = 0;
   int wilduser = 0;
   int nerrors = 0;
   int nlisted = 0;
   int nusers = 0;
   int ret = 0;

/* Check whether any groups specified */
  if (options.ngroups == 0)
    allgroups = 1;

  else {

/* Check admin group priviledge */
    if (strcmp(admin.user,"root")) {
      if (options.ngroups > 1 || strcmp(admin.group,options.groups[0])) {
        fprintf(stderr,"Admin %s can list users in group %s only.\n",admin.user,admin.group);
        syslog(LOG_NOTICE,"Admin %s can list users in group %s only.\n",admin.user,admin.group);
        return(-1);
      }
    }

/* Check for group wildcards */
    for (i = 0; i < options.ngroups; i++) {
      if (!strcmp(options.groups[i],"*")) {
        allgroups = 1;
        break;
      }
      else if (glob2reg(options.groups[i],reg,sizeof(reg)) > 0) {
        wildgroup = 1;
        break;
      }
    }
  }

/* Check whether any users specified */
  if (options.nusers == 0)
    allusers = 1;

/* Check for user wildcards */
  else {
    for (i = 0; i < options.nusers; i++) {
      if (!strcmp(options.users[i],"*")) {
        allusers = 1;
        break;
      }
      else if (glob2reg(options.users[i],reg,sizeof(reg)) > 0) {
        wilduser = 1;
        break;
      }
    }
  }

/* Open NDBM database */
  if (open_db(EXIST) < 0)
    return(-1);

/* Check whether to retrieve all keys */
  if(allusers == 1 || wilduser == 1 || options.ngroups > 0)
    allrecords = 1;

/* Retrieve all keys */
  if(allrecords == 1) {
    if ((pmalloc = (struct KEY *) malloc(maxusers*sizekey)) == NULL) {
      fprintf(stderr,"Error: malloc failed for %d bytes (%s).\n",
              maxusers*sizekey,sys_errlist[errno]);
      syslog(LOG_ERR,"Error: malloc failed for %d bytes (%s).",
             maxusers*sizekey,sys_errlist[errno]);
      close_db();
      return(-1);
    }
    pkey = pmalloc;
    for (key = dbm_firstkey(db); key.dptr != NULL; key = dbm_nextkey(db)) {
      strncpy(pkey->name,key.dptr,key.dsize);
      pkey->len = key.dsize;
      pkey++;
      nkeys++;
    }
  }

/* Use specified keys */
  else {
    nkeys = options.nusers;
    if ((pmalloc = (struct KEY *) malloc(nkeys*sizekey)) == NULL) {
      fprintf(stderr,"Error: malloc failed for %d bytes (%s).\n",
              nkeys*sizekey,sys_errlist[errno]);
      syslog(LOG_ERR,"Error: malloc failed for %d bytes (%s).",
             nkeys*sizekey,sys_errlist[errno]);
      close_db();
      return(-1);
    }
    pkey = pmalloc;
    for (i = 0; i < nkeys; i++) {
      strcpy(pkey->name,options.users[i]);
      pkey->len = strlen(options.users[i]);
      pkey++;
    }
  }

/* Sort key list */
  pkey = pmalloc;
  if (nkeys > 1)
    qsort(pkey,nkeys,sizekey,compare_key);

/* Retrieve records for sorted key list */
  for (i = 0; i < nkeys; i++, pkey++) {
    key.dptr = pkey->name;
    key.dsize = pkey->len;
    data = dbm_fetch(db,key);

/* Looping over all records */
    if (allrecords == 1) {

/* Record should always exist since keys generated by dbm_first/nextkey */
      if (data.dptr == NULL) {
        if (nerrors < MAXERRORS) {
          fprintf(stderr,"Error: dbm_fetch failed for %s.\n",key.dptr);
          syslog(LOG_ERR,"Error: dbm_fetch failed for %s.\n",key.dptr);
        }
        nerrors++;
        continue;
      }

/* Check whether user record */
      if (*(char *)data.dptr == USER) {
        nusers++;

/* Match user name or pattern */
        if (!allusers) {
          match = 0;
          for (j = 0; j < options.nusers; j++) {
            if (pattern_match(options.users[j],key.dptr) == 0) {
              match = 1;
              break;
            }
          }

/* Go on to next record if no match */
          if (!match)
            continue;
        }

/* Copy record if needed */
        if (options.allinfo || !allgroups || strcmp(admin.user,"root"))
/*          bcopy(data.dptr,(char *)&user,data.dsize); */
          decrypt_record(data.dptr,decrypted_data,data.dsize);
          memcpy(&user,decrypted_data,sizeof(user));

/* Match group name or pattern */
        if (!allgroups) {
          match = 0;
          for (j = 0; j < options.ngroups; j++) {
            if (pattern_match(options.groups[j],user.groupname) == 0) {
              match = 1;
              break;
            }
          }

/* Go on to next record if no match */
          if (!match)
            continue;
        }

/* Check admin group priviledge */
        if (!strcmp(admin.user,"root") || 
            !strcmp(admin.group,user.groupname)) {

/* Display entire record */
          if (options.allinfo == 1) {
            if (nlisted > 0) printf("\n");
            display_user(user);
          }

/* Display user ID */
          else {
            ((char *)(key.dptr))[key.dsize] = '\0';
            printf("%s\n",key.dptr);
          }
          nlisted++;
        }
      }
    }

/* Looping over specified user ID's */
    else {

/* Check that record exists */
      if (data.dptr == NULL) {
        ((char *)key.dptr)[key.dsize] = '\0';
        fprintf(stderr,"%s not found.\n",key.dptr);
        continue;
      }

/* Check whether user record */
      if (*(char *)data.dptr == USER) {

/* Copy record if needed */
        if (options.allinfo || strcmp(admin.user,"root"))
/*          bcopy(data.dptr,(char *)&user,data.dsize); */
          decrypt_record(data.dptr,decrypted_data,data.dsize);
          memcpy(&user,decrypted_data,sizeof(user));

/* Check admin group priviledge */
        if (!strcmp(admin.user,"root") || 
            !strcmp(admin.group,user.groupname)) {

/* Display entire record */
          if (options.allinfo == 1) {
            if (nlisted > 0) printf("\n");
            display_user(user);
          }

/* Display user ID */
          else {
            ((char *)key.dptr)[key.dsize] = '\0';
            printf("%s\n",key.dptr);
          }
          nlisted++;
        }
      }

/* Specified user ID is actually a group */
      else if (*(char *)data.dptr == GROUP) {
        if (!strcmp(admin.user,"root") || !strcmp(admin.group,key.dptr)) {
          ((char *)key.dptr)[key.dsize] = '\0';
          fprintf(stderr,"%s is a group.\n",key.dptr);
        }
      }

    }
  }
  free(pmalloc);

/* Print user record summary for long listing */
  if (options.allinfo)
    printf("\nNumber of user records listed = %d.\n",nlisted);

/* Verify summary record number of users, even though should never need */
/* to update */
  if (allrecords) {
    if(info_db(GWINFO_VERIFY,NUMUSERS,nusers,0))
      ret = -1;
  }

/* Print error summary (this is bad!) */
  if (nerrors > MAXERRORS) {
    fprintf(stderr,"Error: dbm_fetch failed for %d additional keys.\n",nerrors-MAXERRORS);
    syslog(LOG_ERR,"Error: dbm_fetch failed for %d additional keys.\n",nerrors-MAXERRORS);
  }

  close_db();
  return(ret);
}
