/*============================================================================
 * filename:		util/gwuser/moduser.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================	
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  All rights
 *	reserved.
 *
 * writen bY:		Marut Dalahl
 * Date written:	January   26, 1995
 * Last modified:	January   28, 1997 - major update by Walt Croom
 *			October    8, 1999 - added Linux support
 *			February   8, 2000 - added LPL
 *	                         
 * function:	Modify gwuser user record
 *
 *============================================================================*/

#include 	<fcntl.h> 
#
#include 	<stdio.h>
#include 	<stdlib.h>
#include 	<strings.h>
#include 	<syslog.h>
#include	<time.h>                     
#include 	"firewall.h"
#include 	"gwuser.h"

extern struct GWOPTIONS options;
extern struct GWADMIN admin;
extern struct GWUSER user;
extern struct GWCONF conf;
extern DBM *db;
/*===========================================================================
 *
 *===========================================================================*/
int moduser()
{
   datum userkey, userdata;
   datum groupkey, groupdata;
   struct GWUSER moduser;
   GWUSER_ENCRYPT encrypted_data;
   GWUSER_DECRYPT decrypted_data;
   int 	i;
   int	ret;

   /*=====================================
    * Open NDBM database and fetch record 
    *====================================*/
   if (open_db(EXIST) < 0)
      return(-1);

   /* Check that user record exists */
   userkey.dptr = user.username;
   userkey.dsize = strlen(user.username);
   userdata = dbm_fetch(db,userkey);
   if (userdata.dptr == NULL) 
   {
      fprintf(stderr,"Error: user %s does not exist.\n",userkey.dptr);
      close_db();
      return(-1);
   }

   /* Check that record is actually for a user */
   if (*(char *)userdata.dptr == GROUP) 
   {
      fprintf(stderr,"%s is a group, cannot modify with gwuser.\n",
                     userkey.dptr);
      close_db();
      return(-1);
   }
   else if (*(char *)userdata.dptr != USER) 
   {
      fprintf(stderr,"%s is not a user, cannot modify.\n",userkey.dptr);
      close_db();
      return(-1);
   }

   /*===================================================================== 
    * Check admin group priviledge 
    *    bcopy(userdata.dptr,(char *)&moduser,userdata.dsize); 
    *========================================================================*/
   decrypt_record(userdata.dptr,decrypted_data,userdata.dsize);
   memcpy(&moduser,decrypted_data,sizeof(moduser));
   if (strcmp(admin.user,"root") && strcmp(admin.group,moduser.groupname)) 
   {
      fprintf(stderr,"Admin %s can modify users in group %s only,\n"
                     "user %s not modified.\n",
		     admin.user,admin.group,moduser.username);
      syslog(LOG_NOTICE,"Admin %s can modify users in group %s only,"
		        " user %s not modified.",
			admin.user,admin.group,moduser.username);
      close_db();
      return(-1);
   }

   /* Update user record */
   if (options.real)
      strcpy(moduser.realname,user.realname);
   if (options.phone)
      strcpy(moduser.phonenum,user.phonenum);
   if (options.group) 
   {
      /* Check for authorized admin */
      if (strcmp(admin.user,"root")) 
      {
         fprintf(stderr,"Must be root to change group (admin = %s).\n",
                        admin.user);
         syslog(LOG_NOTICE,"Must be root to change group (admin = %s).\n",
			   admin.user);  
         close_db();
         return(-1);
      }
      /* Group must exist */
      groupkey.dptr = user.groupname;
      groupkey.dsize = strlen(user.groupname);
      groupdata = dbm_fetch(db,groupkey);
      if (groupdata.dptr == NULL) 
      {
         fprintf(stderr,"Group %s does not exist, user %s not modified.\n",
              		groupkey.dptr,userkey.dptr);
         close_db();
         return(-1);
      }
      else if (*(char *)groupdata.dptr != GROUP) 
      {
         fprintf(stderr,"%s is not a group, user %s not modified.\n",
         		groupkey.dptr,userkey.dptr);
         close_db();
         return(-1);
      }
      strcpy(moduser.groupname,user.groupname);
   }
   if (options.pass) 
   {
      /* Save old passwords */
      for (i = MAXPWDS-1; i > 0; i--)
         strncpy(moduser.password[i],moduser.password[i-1],PWDSIZ);
      strcpy(moduser.password[0],user.password[0]);
      moduser.passworddate = time((time_t *)NULL);
   }
   if (options.token)
      moduser.tokentype = user.tokentype;
   if (options.key)
      strcpy(moduser.userkey,user.userkey);
   if (options.telnet)
      moduser.tnaccess = user.tnaccess;
   if (options.ftp)
      moduser.ftpaccess = user.ftpaccess;
   if (options.start)
      moduser.startdate = user.startdate;
   if (options.stop) 
   {
       moduser.stopdate = user.stopdate;
   }
   if (options.start) 
   {
      if (moduser.startdate > moduser.stopdate && moduser.stopdate > 0) 
      {
         fprintf(stderr,"Start date cannot be later than stop date.\n");
         close_db();
         return(-1);
      }
   }
   if (options.stop) 
   {
      if (moduser.stopdate < moduser.startdate && moduser.stopdate > 0) 
      {
         fprintf(stderr,"Stop date cannot be earlier than start date.\n");
         close_db();
         return(-1);
      }
   }
   if (options.times)
     memcpy(moduser.accesstimes,user.accesstimes,sizeof(moduser.accesstimes));
   if (options.stat) {
     moduser.status = user.status;
     if (user.status == UNLOCKED)
       moduser.no_login_attempts = 0;
   }
   moduser.recorddate = time((time_t *)NULL);

   /* Update user record */
/*
   userdata.dptr = (char *) &moduser;
   userdata.dsize = sizeof(moduser);
*/
   encrypt_record((char *)&moduser,encrypted_data,sizeof(moduser));
   userdata.dptr = encrypted_data;
   userdata.dsize = encrypted_sizeof(moduser);
   if (dbm_store(db,userkey,userdata,DBM_REPLACE)) 
   {
      fprintf(stderr,"Error: dbm_store failed for %s.\n",userkey.dptr);
      syslog(LOG_ERR,"Error: dbm_store failed for %s.",userkey.dptr);
      ret = -1;
   }

   /* Update summary information record */
   else 
   {
      printf("User %s modified.\n",userkey.dptr);
      info_db(GWINFO_MODIFY,MODIFYDATE,0,0);
      ret = 0;
   }

   close_db();
   return(ret);
}
