/*============================================================================
 * filename:		util/gwuser/set_password.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================	
 *
 * (C) COPYRIGHT R. James Livermore 1997-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1997-2000.  
 * All rights reserved.    
 *
 *
 * written by:		R. James Livermore 
 *
 *
 * Date written:	July       9, 1997
 * Last modified:	February   8, 2000
 *	                         
 * function:	Set the number of unsuccessful login attempts.
 *
 * inputs:
 *	ptr to user name
 *	integer	if 0 reset no_login_attempts to zero,else increment by value
 *                 passed.
 *                         
 * outputs:
 *	return  0 if ok
 *	       -1 on error   
 *	updated user record
 *============================================================================*/

#include	<fcntl.h>
#include 	<stdio.h>
#include 	<stdlib.h>
#include 	<strings.h>
#include 	<syslog.h>
#include 	<time.h>
#include 	"firewall.h"
#include 	"gwuser.h"

extern DBM *db;

int set_nologins(char *username, int value, int maxattempts)
{
   datum key, data;
   struct GWUSER user;
#ifndef linux
   static char gwuserfiles[] = GWUSERFILE".<dir,pag>";
#else
   static char gwuserfiles[] = GWUSERFILE".db";
#endif
   GWUSER_ENCRYPT encrypted_data;
   GWUSER_DECRYPT decrypted_data;
   int status;
   int i;
   int ret;

   /*======================================================================
    * Open NDBM database & fetch user record. 
    * The function open_db() locks the entire data base.
    * The function close_db() unlocks the data base.
    *====================================================================*/
   if ((status = open_db(EXIST)) < 0)
      return(-1);

   /* Retrieve user record */
   key.dptr = username;
   key.dsize = strlen(username);
   data = dbm_fetch(db,key);

   /* Check that user exists (but shouldn't be necessary) */
#ifdef linux
   if (*data.dptr == 0)
#else
   if (*(char *)data.dptr == NULL)
#endif
   {
      syslog(LOG_ERR,"Error: user '%s' not found in %s, cannot set "
		     "number of login attempts.",
             	     username,gwuserfiles);
      close_db();
      return(-1);
   }
   else if (*(char *)data.dptr != USER)
   {
      syslog(LOG_ERR,"Error: '%s' is not a user record in %s, "
                     "cannot set number of login attempts.",
                     username,gwuserfiles);
      close_db();
      return(-1);
   }

   /*==========================================================
    * Decrypt user record
    *==========================================================*/
   decrypt_record(data.dptr,decrypted_data,data.dsize);
   memcpy(&user,decrypted_data,sizeof(user));

   if( value == 0  && user.status != LOCKED)
   {
      user.no_login_attempts = 0;   
   }
   else
   {
      user.no_login_attempts = user.no_login_attempts + value; 
   }  
   /*==========================================
    * If maxattempts exceeded then lock record.
    *=========================================*/
   if( user.no_login_attempts > maxattempts )
      user.status = LOCKED;
   /* Set record modified date */
   user.recorddate = time((time_t *)NULL);

   /*========================================================
    * Encrypt the user record then store it in DB.
    *=======================================================*/
   encrypt_record((char *)&user,encrypted_data,sizeof(user));
   data.dptr = encrypted_data;
   data.dsize = encrypted_sizeof(user);
   if (dbm_store(db,key,data,DBM_REPLACE))
   {
      syslog(LOG_ERR,"Error: dbm_store failed for %s.",key.dptr);
      ret = -1;
   }

   /* Update summary information record */
   else
   {
      info_db(GWINFO_MODIFY,MODIFYDATE,0,0);
      ret = 0;
   }

   close_db();
   return(ret);
}
