/*============================================================================
 * filename:		util/gwuser/userargs.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================	
 *
 * (C) COPYRIGHT R. James Livermore 1997-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1997-2000.  All rights
 *	reserved.
 *
 * written by:		Walt Croom 
 * Date written:	January   28, 1997
 * Last modified:	February   8, 2000
 *	                         
 * function:	Get/check gwuser arguments
 *
 *============================================================================*/
#include 	<fcntl.h>  
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include "firewall.h"
#include "gwuser.h"

struct GWOPTIONS options;
struct GWUSER user;
struct GWGROUP group;

int userargs(int argc, char **argv)
{
   int first = 1;
   int groups_specified = 0;
   int ntokens;  
   int numargs;

   argc--; 
   argv++;

   /* Help summary */
   if (argc == 0 || strcmp(*argv,"?") == 0 || strcmp(*argv,"-h") == 0) 
   {
      userhelp(GWHELP,stdout);
      return(1);
   }

   /*========================================================================
    * Add user 
    *=======================================================================*/
   else if (strcmp(*argv,"-a") == 0) 
   {
      argc--; argv++;
      options.action = GWADD;
      options.user   = 0;
      options.real   = 0;
      options.phone  = 0;
      options.group  = 0;
      options.pass   = 0;
      options.token  = 0;
      options.key    = 0;
      options.telnet = 0;
      options.ftp    = 0;
      options.start  = 0;
      options.stop   = 0;
      options.times  = 0;
      while (argc) 
      {
         if (strcmp(*argv,"?") == 0 || strcmp(*argv,"-h") == 0) 
         {
            userhelp(GWADD_HELP,stdout);
            return(1);
         }
         else if (**argv == '*') 
         {
            fprintf(stderr,"Invalid argument %s.\n\n",*argv);
            userhelp(GWADD_HELP,stdout);
            return(-1);
         }
         else if (first == 1) 
         {
            first = 0;
            if (**argv == '-') 
            {
               fprintf(stderr,"User name must be first argument.\n");
               return(-1);
            }
            else if (get_string(argc,argv,"User name",user.username,UNMSIZ))
               return(-1);
            options.user = 1;
            argc--; 
            argv++;
         }
         else if (strcmp(*argv,"-g") == 0) 
         {
            argc--; 
            argv++;
            if (get_string(argc,argv,"Group name",user.groupname,GRPSIZ))
               return(-1);
            options.group = 1;
            argc--; 
            argv++;
         }
         else if (strcmp(*argv,"-n") == 0) 
         {
            argc--; 
            argv++;
            if (get_string(argc,argv,"Real name",user.realname,REALSIZ))
               return(-1);
            options.real = 1;
            argc--; 
            argv++;
         }
         else if (strcmp(*argv,"-ph") == 0) 
         {
            argc--; 
            argv++;
            if (get_string(argc,argv,"Phone number",user.phonenum,PHONESIZ))
               return(-1);
            options.phone = 1;
            argc--; 
            argv++;
         }
         else if (strcmp(*argv,"-p") == 0) 
         {
            argc--; 
            argv++;
            if (get_password(argc,argv,user.password[0]))
               return(-1);
            options.pass = 1;
            argc--; 
            argv++;
         }
         else if (strcmp(*argv,"-s") == 0) 
         {
            argc--; 
            argv++;
            if (get_token(argc,argv,&user.tokentype))
               return(-1);
            options.token = 1;
            argc--; 
            argv++;
         }
         else if (strcmp(*argv,"-t") == 0) 
         {
            argc--; 
            argv++;
            if ((ntokens = get_telnet(argc,argv,&user.tnaccess)) <= 0)
               return(-1);
            options.telnet = 1;
            argv += ntokens;
            argc -= ntokens;
         }
         else if (strcmp(*argv,"-f") == 0) 
         {
            argc--; 
            argv++;
            if ((ntokens = get_ftp(argc,argv,&user.ftpaccess)) <= 0)
               return(-1);
            options.ftp = 1;
            argv += ntokens;
            argc -= ntokens;
         }
         else if (strcmp(*argv,"-k") == 0) 
         {
            argc--; 
            argv++;
            if ((ntokens = get_key(argc,argv,user.userkey)) <= 0)
               return(-1);
            options.key = 1;
            argv += ntokens;
            argc -= ntokens;
         }
         else if (strcmp(*argv,"-start") == 0) 
         {
            argc--; 
            argv++;
            options.start = 1;
            if (get_date(argc,argv,"Start date",&user.startdate))
               return(-1);
            argc--; 
            argv++;
         }
         else if (strcmp(*argv,"-stop") == 0) 
         {
            argc--; 
            argv++;
            options.stop = 1;
            if (get_date(argc,argv,"Stop date",&user.stopdate))
               return(-1);
            argc--; 
            argv++;
         }
         else if (strcmp(*argv,"-times") == 0) 
         {
            argc--; 
            argv++;
            options.times = 1;
            if(( numargs = get_times(argc,argv,(char *)&user.accesstimes)) < 0)
               return(-1); 
            argc = argc - numargs; 
            argv = argv + numargs;
         }
         /* Add additional argument parsing here ... */

         else 
         {
            fprintf(stderr,"Invalid argument %s.\n\n",*argv);
            userhelp(GWADD_HELP,stderr);
            return(-1);
         }
      }
      if (!options.user) 
      {
         fprintf(stderr,"A user name must be specified.\n\n");
         userhelp(GWADD_HELP,stderr);
         return(-1);
      }
      return(0);
   }

   /*=========================================================================
    * Convert database 
    *========================================================================*/
   else if (strcmp(*argv,"-c") == 0) 
   {
      argc--; 
      argv++;
      options.action = GWCONVERT;    
      if (argc) 
      {
         if (strcmp(*argv,"?") == 0 || strcmp(*argv,"-h") == 0) 
         {
            userhelp(GWCONVERT_HELP,stdout);
            return(1);
         }
         else 
         {
            fprintf(stderr,"Invalid argument %s.\n\n",*argv);
            userhelp(GWCONVERT_HELP,stderr);
            return(-1);
         }
      }
      return(0);
   }

   /*===============================
    * Delete users 
    *==============================*/
   else if (strcmp(*argv,"-d") == 0) 
   {
      argc--; 
      argv++;
      options.action = GWDEL;    
      options.user = 0;
      while (argc) 
      {
         if (strcmp(*argv,"?") == 0 || strcmp(*argv,"-h") == 0) 
         {
            userhelp(GWDEL_HELP,stdout);
            return(1);
         }
         else if (**argv == '*' || **argv == '-') 
         {
            fprintf(stderr,"Invalid argument %s.\n\n",*argv);
            userhelp(GWDEL_HELP,stderr);
            return(-1);
         }
         else if (options.user) 
         {
            fprintf(stderr,"Only one user name can be specified.\n");
            return(-1);
         }
         else if (get_string(argc,argv,"User name",user.username,UNMSIZ))
            return(-1);
         options.user = 1;
         argc--; 
         argv++;
      }
      if (!options.user) 
      {
         fprintf(stderr,"A user name must be specified.\n\n");
         userhelp(GWDEL_HELP,stderr);
         return(-1);
      }
      return(0);
   }

   /*========================================================================
    * List users 
    *======================================================================*/
   else if (strcmp(*argv,"-l") == 0) 
   {
    argc--; argv++;
    options.action   = GWLIST;
    options.nusers   = 0;
    options.ngroups  = 0;
    options.allinfo  = 0;
    groups_specified = 0;
    while (argc) {
      if (strcmp(*argv,"?") == 0 || strcmp(*argv,"-h") == 0) {
        userhelp(GWLIST_HELP,stdout);
        return(1);
      }
      else if (strcmp(*argv,"-a") == 0) {
        options.allinfo = 1;
      }
      else if (strcmp(*argv,"-g") == 0) {
        groups_specified = 1;
      }
      else if (**argv == '-') {
        fprintf(stderr,"Invalid argument %s.\n\n",*argv);
        userhelp(GWLIST_HELP,stderr);
        return(-1);
      }
      else if (groups_specified) {
        if (options.ngroups > (MAX_GROUPS - 1)) {
          fprintf(stderr,"Too many groups specified, maximum is %d.\n",MAX_GROUPS);
          return(-1);
        }
        if (get_string(argc,argv,"Group name",options.groups[options.ngroups],GRPSIZ))
          return(-1);
        options.ngroups++;
      }
      else {
        if (options.nusers > (MAX_USERS - 1)) {
          fprintf(stderr,"Too many users specified, maximum is %d.\n",MAX_USERS);
          return(-1);
        }
        if (get_string(argc,argv,"User name",options.users[options.nusers],UNMSIZ))
          return(-1);
        options.nusers++;
      }
      argc--; argv++;
    }
    return(0);
  }

   /*=========================================================================
    * Modify user 
    *=======================================================================*/
   else if (strcmp(*argv,"-m") == 0) 
   {
      argc--; 
      argv++;
      options.action = GWMOD;
      options.user   = 0;
      options.real   = 0;
      options.phone  = 0;
      options.group  = 0;
      options.pass   = 0;
      options.token  = 0;
      options.key    = 0;
      options.telnet = 0;
      options.ftp    = 0;
      options.start  = 0;
      options.stop   = 0;
      options.times  = 0;
      options.stat   = 0;
      while (argc) 
      {
         if (strcmp(*argv,"?") == 0 || strcmp(*argv,"-h") == 0) 
         {
            userhelp(GWMOD_HELP,stdout);
            return(1);
         }
         else if (**argv == '*') 
         {
            fprintf(stderr,"Invalid argument %s.\n\n",*argv);
            userhelp(GWMOD_HELP,stdout);
            return(-1);
         }
         else if (first == 1) 
         {
            first = 0;
            if (**argv == '-') 
            {
               fprintf(stderr,"User name must be first argument.\n");
               return(-1);
            }
            else if (get_string(argc,argv,"User name",user.username,UNMSIZ))
               return(-1);
            options.user = 1;
            argc--; 
            argv++;
            if(!argc) 
            {
               fprintf(stderr,"A parameter to modify must be specified.\n\n");
               userhelp(GWMOD_HELP,stdout);
               return(-1);
            }
         }
         else if (strcmp(*argv,"-g") == 0) 
         {
            argc--; 
            argv++;
            if (get_string(argc,argv,"Group name",user.groupname,GRPSIZ))
               return(-1);
            options.group = 1;
            argc--; 
            argv++;
         }
         else if (strcmp(*argv,"-n") == 0) 
         {
            argc--; 
            argv++;
            if (get_string(argc,argv,"Real name",user.realname,REALSIZ))
               return(-1);
            options.real = 1;
            argc--; 
            argv++;
         }
         else if (strcmp(*argv,"-ph") == 0) 
         {
            argc--; 
            argv++;
            if (get_string(argc,argv,"Phone number",user.phonenum,REALSIZ))
               return(-1);
            options.phone = 1;
            argc--; 
            argv++;
         }
         else if (strcmp(*argv,"-p") == 0) 
         {
            argc--; 
            argv++;
            if (get_password(argc,argv,user.password[0]))
               return(-1);
            options.pass = 1;
            argc--; 
            argv++;
         }
         else if (strcmp(*argv,"-s") == 0) 
         {
            argc--; 
            argv++;
            if (get_token(argc,argv,&user.tokentype))
               return(-1);
            options.token = 1;
            argc--; 
            argv++;
         }
         else if (strcmp(*argv,"-t") == 0) 
         {
            argc--; 
            argv++;
            if ((ntokens = get_telnet(argc,argv,&user.tnaccess)) <= 0)
               return(-1);
            options.telnet = 1;
            argv += ntokens;
            argc -= ntokens;
         }
         else if (strcmp(*argv,"-f") == 0) 
         {
            argc--; 
            argv++;
            if ((ntokens = get_ftp(argc,argv,&user.ftpaccess)) <= 0)
               return(-1);
            options.ftp = 1;
            argv += ntokens;
            argc -= ntokens;
         }
         else if (strcmp(*argv,"-k") == 0) 
         {
            argc--; 
            argv++;
            if ((ntokens = get_key(argc,argv,user.userkey)) <= 0)
               return(-1);
            options.key = 1;
            argv += ntokens;
            argc -= ntokens;
         }
         else if (strcmp(*argv,"-start") == 0) 
         {
            argc--; 
            argv++;
            options.start = 1;
            if (get_date(argc,argv,"Start date",&user.startdate))
               return(-1);
            argc--; 
            argv++;
         }
         else if (strcmp(*argv,"-stop") == 0) 
         {
            argc--; 
            argv++;
            options.stop = 1;
            if (get_date(argc,argv,"Stop date",&user.stopdate))
               return(-1);
            argc--; 
            argv++;
         }
         else if (strcmp(*argv,"-times") == 0) 
         {
            argc--; 
            argv++;
            options.times = 1;
            if(( numargs = get_times(argc,argv,(char *)&user.accesstimes)) < 0)
               return(-1); 
            argc = argc - numargs; 
            argv = argv + numargs;
         }


/* Add additional argument parsing here ... */

         else if (strcmp(*argv,"-lock") == 0) 
         {
            user.status = LOCKED;
            options.stat = 1;
            argc--; 
            argv++;
         }
         else if (strcmp(*argv,"-unlock") == 0) 
         {
            user.status = UNLOCKED;
            options.stat = 1;
            argc--; 
            argv++;
         }
         else 
         {
            fprintf(stderr,"Invalid argument %s.\n\n",*argv);
            userhelp(GWMOD_HELP,stderr);
            return(-1);
         }
      }
      if (!options.user) 
      {
         fprintf(stderr,"A user name must be specified.\n\n");
         userhelp(GWMOD_HELP,stderr);
         return(-1);
      }
      return(0);
   }

   /*=========================================================================
    * Query database summary information 
    *=======================================================================*/
   else if (strcmp(*argv,"-q") == 0) 
   {
      argc--; 
      argv++;
      options.action = GWQUERY;
      if (argc) 
      {
         if (strcmp(*argv,"?") == 0 || strcmp(*argv,"-h") == 0) 
         {
            userhelp(GWQUERY_HELP,stdout);
            return(1);
         }
         else 
         {
            fprintf(stderr,"Invalid argument %s.\n\n",*argv);
            userhelp(GWQUERY_HELP,stderr);
            return(1);
         }
      }
      return(0);
   }

   else 
   {
      fprintf(stderr,"Invalid argument %s.\n\n",*argv);
      userhelp(GWHELP,stderr);
      return(-1);
   }
}
