/*============================================================================
 * filename:		util/gwuser/userhelp.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================	
 *
 * (C) COPYRIGHT R. James Livermore 1997-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1997-2000.  All rights
 *	reserved.
 *
 * written by:		Walt Croom
 * Date written:	January   28, 1997
 * Last modified:	February   8, 2000 - added LPL
 *	                         
 * function:	Display gwuser usage
 *
 *============================================================================*/
#include 	<fcntl.h> 

#include 	<stdio.h>
#include 	<stdlib.h>
#include 	"firewall.h"
#include 	"gwuser.h"

extern struct GWUSER user;

void userhelp(int option, FILE *stream)
{
  switch(option)
  {
    case GWADD_HELP:
      fprintf(stream,"gwuser add user usage:\n\n");
      fprintf(stream,"gwuser -a user [ -g group ] [ -n \"name\" ] [ -ph \"phone\" ]\n");
#ifdef _AIX
      fprintf(stream,"       [ -p password ] [ -s CRYPTO | SNK | SDI ]\n");
#else
      fprintf(stream,"       [ -p password ] [ -s CRYPTO | SNK ]\n");
#endif
      fprintf(stream,"       [ -k nnn nnn nnn nnn nnn nnn nnn nnn ]\n");
      fprintf(stream,"       [ -t [ UNPRO_TN ] [ PRO_TN | PRO_TNSA ] [ ADM_TN ] ]\n");
      fprintf(stream,"       [ -f [ UNPRO_FTP ] [ PRO_FTP | PRO_FTPSA ]\n");
      fprintf(stream,"            [ GET_FTP ] [ PUT_FTP ] [ DEL_FTP ] ]\n");
      fprintf(stream,"       [ -start mmdd[yy] ] [ -stop mmdd[yy] ]\n");
      fprintf(stream,"       [ -times ALL | \"times\" ]\n");
      fprintf(stream,"\n");
      fprintf(stream,"   -a     Add gwuser database user record.\n");
      fprintf(stream,"   user   User name (max %d chars).\n",UNMSIZ-1);
      fprintf(stream,"   -g     Group name (max %d chars).\n",GRPSIZ-1);
      fprintf(stream,"   -n     Real name (max %d chars).\n",REALSIZ-1);
      fprintf(stream,"   -ph    Phone number (max %d chars).\n",PHONESIZ-1);
      fprintf(stream,"   -p     Password (max %d chars).\n",PWDSIZ-1);
      fprintf(stream,"   -s     Security token (group default).\n");
      fprintf(stream,"   -k     Security token key.\n");
      fprintf(stream,"   -t     Telnet access (group default).\n");
      fprintf(stream,"   -f     FTP access (group default).\n");
      fprintf(stream,"  -start  Activation date (default is now).\n");
      fprintf(stream,"  -stop   Termination date (default is never).\n");
      fprintf(stream,"  -times  days hours ... [days hours].\n");
      fprintf(stream,"          example m-f 8a-5p sa 9a-4p\n"); 
      fprintf(stream,"          If not specified group defaults are used.\n");               
      break;

    case GWDEL_HELP:
      fprintf(stream,"gwuser delete usage:\n\n");
      fprintf(stream,"gwuser -d user\n");
      fprintf(stream,"   -d    Delete gwuser user database record.\n");
      fprintf(stream,"  user   User name to delete.\n");
      break;

    case GWLIST_HELP:
      fprintf(stream,"gwuser list user usage:\n\n");
      fprintf(stream,"gwuser -l [ -a ] [ users ] [ -g groups ]\n");
      fprintf(stream,"   -l    List gwuser user database records.\n");
      fprintf(stream,"   -a    List all information.\n");
      fprintf(stream,"  users  User name(s); if not specified, list all users.\n");
      fprintf(stream,"         If wild card specified, enclose name in quotes.\n");
      fprintf(stream,"   -g    List users in following groups; if not specified, list\n");
      fprintf(stream,"         users in all groups.\n");
      fprintf(stream," groups  Group name(s). If wild card specified, enclose name in quotes.\n");
      break;

    case GWMOD_HELP:
      fprintf(stream,"gwuser modify user usage:\n\n");
      fprintf(stream,"gwuser -m user [ -g group ] [ -n \"name\" ] [ -ph \"phone\" ]\n");
#ifdef _AIX
      fprintf(stream,"       [ -p password ] [ -s CRYPTO | SNK | SDI | NONE ]\n");
#else
      fprintf(stream,"       [ -p password ] [ -s CRYPTO | SNK | NONE ]\n");
#endif
      fprintf(stream,"       [ -k nnn nnn nnn nnn nnn nnn nnn nnn ]\n");
      fprintf(stream,"       [ -t [ NO_TN ] | [ [ UNPRO_TN ] [ PRO_TN | PRO_TNSA ] [ ADM_TN ] ]\n");
      fprintf(stream,"       [ -f [ NO_FTP ] | [ [ UNPRO_FTP ] [ PRO_FTP | PRO_FTPSA ]\n");
      fprintf(stream,"            [ GET_FTP ] [ PUT_FTP ] [ DEL_FTP ] ]\n");
      fprintf(stream,"       [ -start NOW | mmdd[yy] ] [ -stop NEVER | mmdd[yy] ]\n");
      fprintf(stream,"       [ -times ALL | NONE | \"times\" ]\n");
      fprintf(stream,"       [ -lock | -unlock ]\n");
      fprintf(stream,"\n");
      fprintf(stream,"   -m      Modify gwuser database user record.\n");
      fprintf(stream,"   user    User name (max %d chars).\n",UNMSIZ-1);
      fprintf(stream,"   -g      Group name (max %d chars).\n",GRPSIZ-1);
      fprintf(stream,"   -n      Real name (max %d chars).\n",REALSIZ-1);
      fprintf(stream,"   -ph     Phone number (max %d chars).\n",PHONESIZ-1);
      fprintf(stream,"   -p      Password (max %d chars).\n",PWDSIZ-1);
      fprintf(stream,"   -s      Security token.\n");
      fprintf(stream,"   -k      Security token key.\n");
      fprintf(stream,"   -t      Telnet access.\n");
      fprintf(stream,"   -f      FTP access.\n");
      fprintf(stream,"  -start   Activation date.\n");
      fprintf(stream,"  -stop    Termination date.\n");
      fprintf(stream,"  -times   days hours ...[days hours].\n");
      fprintf(stream,"           example m-f 8a-5p sa 9a-4p\n"); 
      fprintf(stream,"           If not specified group defaults are used.\n");  
      fprintf(stream,"  -lock    Lock user access.\n");
      fprintf(stream,"  -unlock  Unlock user access.\n");
      break;

    case GWQUERY_HELP:
      fprintf(stream,"gwuser query usage:\n\n");
      fprintf(stream,"gwuser -q  List gwuser database summary information.\n");
      break;

    case GWCONVERT_HELP:
      fprintf(stream,"gwuser convert usage:\n\n");
      fprintf(stream,"gwuser -c  Convert gwuser database to NDBM format.\n");
      break;

    case GWHELP:
    default:
      fprintf(stream,"gwuser usage:\n\n");
      fprintf(stream,"help:     gwuser [ -a | -m | -d | -l | -q | -c ] [ ? | -h ]\n");
      fprintf(stream,"add:      gwuser -a user [ options ]\n");
      fprintf(stream,"modify:   gwuser -m user options\n");
      fprintf(stream,"delete:   gwuser -d user\n");
      fprintf(stream,"list:     gwuser -l [ -a ] [ users ] [ -g groups ]\n");
      fprintf(stream,"query:    gwuser -q\n");
      fprintf(stream,"convert:  gwuser -c\n");
  }
}
