/*============================================================================
 * file:	util/portscan/portscan.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=======================================================================
 * function:	Scan ports on the firewall 
 * 
 * (C) COPYRIGHT Freemont Avenue Software, Inc. 1995-2000.
 *                                                                      
 * Author:	Jim Livermore
 *
 * Date written:	January   14, 1995
 * Last modified:	February   8, 2000 - added LPL
 *                                                                      
 *=========================================================================*/
#include	<ctype.h>  
#include	<fcntl.h>
#include	<stdio.h>
#include 	<sys/stat.h>
#include	<sys/types.h>
#include	<sys/socket.h>
#include	<netinet/in.h>
#include	<netdb.h>
#include	"servlist.h"
extern	char		*optarg;
extern	int		optind;
/*==========================================================================
 * function prototypes
 *=========================================================================*/
extern	struct servent	*getservbyport();
struct servlist	*Getservbyport(int portnum, char *protocol, char *buffptr, int numrec);
int safeport(int portnum, char * protocol);
void usage();
/*==========================================================================
 *
 *=========================================================================*/
main(int ac,char **av) 
{  
   static  char	ident[] = "@(#) portscan, V 1.0, 00/01/21"
 			  " (C) Freemont Avenue Software, Inc."
			  " 1995-2000";
   char		*bufptr;			/* pointer to service buffer*/
   int		bufsize;			/* buffer size              */
   int		loport  = 0;			/* default low port number  */
   int		hiport  = 32000;		/* default high port number */	                                             	
   int		numscan	= 0; 			/* number of ports scanned  */
   int		numact	= 0;			/* number active	    */
   int		numrec  = 0;			/* number of service records*/
   char		*s;
   struct	sockaddr_in	addr;           /* for server               */
   struct       sockaddr_in	cli_addr; 	/* for client               */
   struct	hostent		*hp;
   int		fd;  
   int		fdu;
   int		warn 	= 0;   			/* number of warnings	   */
   int		x;
   int		xit;
   char		*p;
    
   char 	servfn[256] = "/etc/services";
   /*========================================================================
    * Get the input arguments where l, h, f & v are recognised as flags.   
    * The extern optind indexes the next argument and optarg points to string
    * following the flag letter.
    *=======================================================================*/
   while((x = getopt(ac,av,"l:h:f:")) != -1) 
   {
      switch(x) 
      {
         case 'h':
		hiport = atoi(optarg);
		break;
	 case 'l':
		loport = atoi(optarg);
		break;
         case 'f':
                strcpy(servfn,optarg); 
		break;
      }
   }  
   if(optind == ac) 
   {
      usage();
      return(-1);
   }
   /* optind should be index of hostname or IP address  */
   for(xit = optind; xit < ac; xit++) 
   {
      s = av[xit]; 
      p = s; 
      /* check for IP address  */
      while(*p != '\0' && (*p == '.' || isdigit(*p)))
			p++;  
      /* If not all digits or dots then lookup host name */
      if(*p != '\0') 
      {
         if((hp = gethostbyname(s)) == (struct hostent *)0) 
         {
	    fprintf(stderr,"%s: address unknown\n",s);
	    return(-1); 
         }  
         (void)bcopy(hp->h_addr,(char *)&addr.sin_addr,hp->h_length);
      } 
      else 
      {
         unsigned long	f;
         /* convert Internet dot address to long int */
         if((f = inet_addr(s)) == -1L) 
         {
            fprintf(stderr,"%s: address unparsable\n",s);
	    return(-1);
         }
         /* copy inet address in f to addr.sin_addr */
         (void)bcopy((char *)&f,(char *)&addr.sin_addr,sizeof(f));
	 /* added by rjl on 5/30/95 */
         addr.sin_family = AF_INET; 
         /* get host by address for use later */  
	 if((hp = gethostbyaddr((const char *)&addr.sin_addr,sizeof(addr.sin_addr),AF_INET)) 
		   == (struct hostent *)0) 
         {
	    printf("portscan: %s: host address unknown\n",s);
	    return(-1);
         } 
      }
      if( loport > hiport)
      {
         printf("portscan: low port (%d) > high port (%d). Please try again.\n",loport,hiport); 
         usage();
         return(-1);
      }
      /*==============================================================
       * load services file into memory to speed up scan 
       *==============================================================*/
      if((numrec = loadserv(servfn, &bufptr, &bufsize)) < 0) return(-1);
      printf("\nportscan: trying stream ports between %d and %d on %s\n\n",
		loport,hiport,s);
      printf("service  name  port/tcp  message \n");
      /*===========================================
       * loop through the ports    
       *==========================================*/
      for(x = loport; x < hiport + 1; x++) 
      {   
         addr.sin_port = htons(x);   		/* set port number  */
         addr.sin_family = AF_INET;
         /*=========================================== 
	  * create tcp socket and try to connect
          *===========================================*/
         fd = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
   	
         if(fd < 0) 
         {
            printf("portscan: socket rc = %d, port number = %d/tcp.\n",
		    fd, x);
            return (-1); 
         }

         if(connect(fd,(struct sockaddr *)&addr,sizeof(addr)) == 0) 
         {
            struct servlist	*sp;
            numact++; 
	    if(safeport(x, "tcp")) warn++;    
            /*==============================================================
	     * Use our version of getservbyport that scans memory resident
             * services file.
             *=============================================================*/ 
            if((sp = Getservbyport(x,"tcp",bufptr,numrec)) != (struct servlist *)0) 
            {
   	       printf("%13s %5d/tcp  is alive on %s (%s) \n",
			 sp->service,x,hp->h_name,inet_ntoa(addr.sin_addr));
	    } 
            else 
            {
	       printf("              %5d/tcp  is alive on %s (%s) \n",
				x,hp->h_name,inet_ntoa(addr.sin_addr));
	    }
         }
         numscan++;
         close(fd);
         /*==================================================
          * Create UDP socket and try to bind 
          *==================================================*/  
         /*
         cli_addr.sin_family	  = AF_INET;
         cli_addr.sin_addr.s_addr = htonl(INADDR_ANY); 
         cli_addr.sin_port	  = htons(0);	
         fdu = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);
   	 
         if(fdu < 0) 
         {
            printf("portscan: socket rc = %d, port number = %d/udp.\n",
		    fdu, x);
            return (-1); 
         }
         if(bind(fdu,(struct sockaddr *)&cli_addr,sizeof(cli_addr)) == 0) 
         {
            struct servent	*sp;
            numact++;
	    if((sp = getservbyport(x,"udp")) != (struct servent *)0) 
            {
   	       if(vflg)
	          fprintf(stderr,"\n%c%s\n",07,sp->s_name);
               else
	          printf("%12s %4d/udp is alive on %s (%s) \n",
			 sp->s_name,x,hp->h_name,inet_ntoa(addr.sin_addr));
	    } 
            else 
            {
	       if(vflg)
	          fprintf(stderr,"\n%c%d\n",07,x);
	       else
	          printf("             %4d/udp is alive on %s (%s) \n",
				x,hp->h_name,inet_ntoa(addr.sin_addr));
	    }
         }
         close(fdu); 
         */

      }
      printf("\n ports scanned = %d, number active = %d\n",numscan,numact);
      if(warn > 0)
         printf(" %d WARNING(S) ISSUED.\n",warn);
   }
   return(0);
} 
/*=============================================================================== 
 * function:	loadserv
 * 
 * The loadserv function reads the services file into memory   
 * 
 * inputs:
 *	A ptr to the file name
 *	A ptr to the buffer area
 *	A ptr to the buffer size   
 * 
 * processing:
 *	 
 * 	The function copies the file into a memory buffer equal to the file size.
 * 	Then it scans the buffer looking for a match for both the internal shell      
 * 	and the external shell. If EITHER is not found in the file pointed to by
 *      fp, the function returns (-1), else it reurns 0. 
 *
 * outputs:
 *	-1 = error
 *	 number of entries      
 *                          
 *============================================================================*/

int loadserv(char *filename, char **bufptr, int *bufsize)
{

   struct servlist *slp; 
   struct servlist *slp2;  
   struct servlist list;
   FILE	*fp;			/* file ptr                             */
   long length;                 /* file size * 2                     	*/
   int 	succeeded;              /* flag for dynamic memory allocation 	*/
   
   char *bptr;                  /* pointer to buffer                    */
   char *shellsp;		/* pointer to 'shells '                 */ 
   char	tokensep[] = "\t\n,:; ";
   int	bytes_rd;

   int numrec;               	/* number of records                    */
   char current;                /* current character to put into buffer */
   struct stat statbuf;		/* buffer for stat record 		*/
   /* Determine the length of the file pointed to by fp 		*/ 
   numrec = 0;
   if( stat(filename, &statbuf) < 0)
   {
      printf("portscan: loadserv: call to stat function failed for file '%s'.\n",
		filename);
      return(-1);
   }             
   length = statbuf.st_size*3 + 1000;
   /*=======================================================================
    * Allocate the memory buffer to contain the contents of the services 
    * file.
    *======================================================================*/
   if ((bptr = (char *)malloc(length)) == NULL)
   {
      printf("\n portscan:loadserv: Memory allocation failure. size = %x \n",
	     length);
      return(-1);
   } 
   /* open the services file */
   if ( (fp = fopen(filename,"r") ) == NULL) 
   {
      printf("portscan: loadserv: Unable to open %s.\n",
	      filename);
      return(-1);
   }
   slp2 = (struct servlist *) bptr;
   /* Read the file into the memory buffer       */ 
   while( ( slp = (struct servlist *)Getservent(fp, &list)) != NULL)
   {             
      if(slp2 + sizeof(struct servlist) > (struct servlist *) (bptr + length))
      {
         printf("portscan: buffer not large enough for file.\n");
         return(-1); 
      }       
      memcpy(slp2,&list,sizeof(struct servlist)); 
      slp2++;  
      numrec++;
   }
   
   *bufptr = bptr;    
   *bufsize = length;
   return(numrec);
}
/*=============================================================================== 
 * function:	Getservbyport
 * 
 * The Getservbyport retrieves an entry from the service file that has been 
 * loaded into a memory buffer.    
 * 
 * inputs:
 *	The port number.
 *	A ptr to a string containing protocol (tcp | udp). 
 *	A ptr to the buffer.
 *      Buffer size.   
 * 
 * processing:
 *	 
 * 	The function copies the file into a memory buffer equal to the file size.
 * 	Then it scans the buffer looking for a match for both the internal shell      
 * 	and the external shell. If EITHER is not found in the file pointed to by
 *      fp, the function returns (-1), else it reurns 0. 
 *
 * outputs:
 *	0 = not found
 *	pointer to the service entry       
 *                          
 *============================================================================*/

struct servlist	*Getservbyport(int portnum, char *protocol, char *buffptr, int numrec)
{

  
  
   struct servlist *slp;
   char *cptr1;                 /* pointer during memory scanning       */ 
   char *cptr2;   
   char *cptr3;
   char *shellsp;		/* pointer to 'shells '                 */ 
   char	tokensep[] = "\t\n,/ ";
   int	bytes_rd;
   int  first; 
                                                              
   int 	i;		 	/* counter 				*/
   char current;                /* current character to put into buffer */
   struct stat statbuf;		/* buffer for stat record 		*/
   struct servent entry;
   
   /*=============================================
    * look for port number ( second token in the line )
    *============================================*/ 
   slp = (struct servlist *) buffptr;      
   first = 1;                     
   for(i=0; i < numrec; i++)
   {  
      if(( slp->portnum == portnum) && (strncmp(protocol,slp->protocol,
                                        sizeof(protocol)-1) == 0))
      return(slp);
      slp++;  
   }  
   return(0);
}
/*===========================================================================
 * function: safeport
 * returns:	0 if port is safe
 *		1 if port not safe
 * also issues a WARNING message
 *
 *==========================================================================*/
int safeport(int portnum, char * protocol)
{
   struct unsafe
   {
      int 	port;
      char	proto[5];
      char	name[12];
   };
   struct unsafe table[] =
   {   
	{    69, "udp", "tftp"},
        {    79, "tcp", "finger"},
        {    80, "tcp", "http"},
	{    87, "tcp", "link"},			/* tty link     */
	{   111, "tcp", "sunrpc"},
        {   111, "udp", "sunrpc"},
        {   117, "tcp", "uucp-path"}, 
        {   512, "tcp", "rexecd"},			/* rexecd	*/
        {   513, "tcp", "rlogin"},			/* rlogin 	*/ 
        {   514, "tcp", "rshd"},			/* rshd 	*/ 
	{   515, "tcp", "printer"},			/* lpr		*/  
        {   540, "tcp", "uucp"},
	{   560, "udp", "rmonitor"}, 
	{  2000, "tcp", "openwindows"},
        { 66000, "end", "end"}
   };  
   struct unsafe *p;
   p = (struct unsafe *) & table;
   for( ; p->port != 66000; p++)
   {
      if( portnum == p->port && (strcmp(protocol, p->proto) == 0) )
      {   
          printf("\nWARNING! Use of %s on port %d is considered UNSAFE!\n",
			p->name,p->port); 
          return(1);          
      }      
   } 
   /* test for special case of DNS */
   if( (portnum == 53) &&
       ((strcmp(protocol, "tcp") == 0) || (strcmp(protocol,"udp") == 0) ) )
      {   
          printf("\nWARNING! DNS (53/tcp) is running. Make certain DNS is running as a \
caching-only\n\
         server on the firewall, and can only resolve external host names.\n");
      } 
   return(0);
}  
/*============================================================================
 *  print program usage instructions
 *===========================================================================*/
void usage(void)
{
   printf("usage: portscan [-l low port] [-h high port] [-f services_file_name]\
 host\n");  
   return; 
}
