/*============================================================================
 * filename:		util/reports/aproxsum/aproxsum.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=============================================================================
 *										
 * written by:		Jim Livermore						
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  All rights 
 * reserved.
 *
 * Date written:	April 	  14, 1995
 * Last modified:	February   8, 2000 - added LPL
 *
 * function:		aproxy daemon summary report program 	
 *       
 * 	The aproxsum program reads the system log and produces a report
 *	on aproxy usage.
 *
 *  	1. read the file until eof
 *	2. capture all aproxy messages. 
 *
 *============================================================================*/

static  char	ident[] = "@(#) aproxsum, V 1.0, 00/01/21"
 			  " (C) Freemont Avenue Software, Inc."
			  " 1995-2000";
#include	<stdio.h> 
#include	<search.h>		/* for qsort	*/
#include	<sys/types.h>
#include	<sys/file.h>
#include	<sys/stat.h>
#include	<string.h>		/* added for C++ */
#include	<time.h>
#include	<ctype.h>
#include	<syslog.h>
#include	<netdb.h>
 
#include	<unistd.h>  		/* added for C++ */
#include	<stdlib.h>		/* added for C++ */  
   
struct APROXHOST 
{
   int	connects;
   int	receives;
   int	bytess;
   int  bytesr;
   int	duration;
   int  port; 
   char ipaddr[16];
   char	host[84];  
}; 
/*==========================
 * external variables
 *==========================*/  
extern   int	errno;
#define		MAXHOSTS 	8000  
   int	maxhosts;
/*======================================================================
 * function prototypes
 *======================================================================*/ 
int comprhosts(char **s1, char **s2);  /* compare host names		*/ 
int comprbr(char **s1, char **s2);   /* compare bytes received		*/
int comprbs(char **s1, char **s2);   /* compare bytes sent		*/
int comprcon(char **s1, char **s2);  /* compare connects 		*/
int comprrecv(char **s1, char **s2);  /* compare receives 		*/
int isalldigits(char *s); 
int prthosts(struct APROXHOST *p, int top); 
int procaprox(char *from, char *sipaddr, char *sport, char *to, char *tport,
		int fbytes, int tbytes, int duration, struct APROXHOST *p, int cntflg);
void usage(void);  
 
/*============================================================================                        
 * main
 * inputs: optional input arguments
 *         -h 		for help
 *	   -l logfile	specify logfile name		
 *         -t nnn	specify printing of top nnn hosts only
 * output:
 *	return  0 if ok
 *	return -1 on error
 *===========================================================================*/
int main(int argc, char *argv[])
{   
   /* log file statistics		*/ 
   int		trunc_recs	  = 0; /* eof adds 1 so must sub 1 !!*/
   int		trunc_recnum	  = 0;
   int		total_recs	  = 0;
   int		total_connects 	  = 0;
   int		total_disconnects = 0;
   float	total_bytess 	  = 0; 
   float	total_bytesr      = 0;
   int		total_aproxrec  = 0; 
   int		total_alerts	  = 0;
   int		total_logbytes	  = 0;
   int		total_apbytes	  = 0; 
   char		firstdate[17];
   char		lastdate[17]; 
   /* 					*/ 
   struct	APROXHOST	*p;     /* ptr to data area		*/
   struct	APROXHOST	*p2; 
   int		duration;		/* duration in seconds		*/
   int		total_duration  = 0;    /* duration of all aproxies   */
   int		i;
   int		fbytes; 
   int		tbytes; 
   int		numhosts	= 0;
   int		reclen;
   int		top		= 0;
   #define	MAXLEN 1024
   static char	alertfile[128] = "/tmp/aproxy.alerts";
#ifdef _AIX
   static char 	logfile[256] = "/var/adm/syslog";
#elif __hppa
   static char 	logfile[256] = "/var/adm/syslog/syslog.log";
#elif linux
   static char 	logfile[256] = "/var/log/messages";
#else   
   static char 	logfile[256] = "/var/adm/messages";
#endif   
   char		buffer[MAXLEN];
   char		record[MAXLEN];
   FILE 	*lfp;			/* log file ptr		*/ 
   FILE 	*alfp;			/* alert file ptr	*/
   char 	*cp;   
   char		*localIP;
   char		*token1;
   char		*token2;
   char		*token3;
   char		*token4;
   char		*token5;
   char		*from; 
   char		*tipaddr;
   char		*sipaddr;		/* source IP address 	*/
   char 	*sport;			/* source port		*/ 
   char		*tport;			/* destination port	*/
   static char	tokensep[] 	= "\t,:;=[] \n";
   maxhosts = MAXHOSTS;
   /*===================================================
    * check for input arguments
    *==================================================*/
   if( argc > 1)
   {
      if( argv[1][0] != '-')   
      {
         printf("\n aproxsum:  invalid leading character '%c' \n",argv[1][0]);
         usage();
         return(-1);        
      }
      /* check to see if user is asking for help */
      if ((strcmp(argv[1], "-h")    == 0) ||
          (strcmp(argv[1], "-help") == 0) ||
          (strcmp(argv[1], "help")  == 0) ||
          (strcmp(argv[1], "?")     == 0)
         ) 
      {
         usage();
         return(0);
      } 
      for(i = 1; i < argc ; i++)
      {
         switch( argv[i][1])
         { 
            case 'l': 
               if( strlen(argv[i+1]) > 255)
               {
                  printf("\n aproxsum: file name too long. length = %d\n",
			strlen(argv[i+1])); 
                  usage();
                  return(-1);
               }
	       strcpy(logfile,argv[i+1]);
               i++;
               break;
            case 't':
               if( !isalldigits(argv[i+1]) )
               {
                  printf("\n proxsum: the paramter following -t is not all digits.\n"); 
                  usage();
		  return(-1);
               }
               top = atoi(argv[i+1]);
               i++;
               break;
            default:
               printf("\n aproxsum: invalid input parameter %s \n",argv[i][1]);   
               break; 
         }
      }
   }
   /* open log file		*/  
   if (!(lfp = fopen(logfile, "r")))
   {  
      
      fprintf(stderr,"aproxsum: open failed for logfile  %s.\n       %s. \n",
			logfile, strerror(errno));
      return(-1);    			
   } 
   /* open security alert file	*/
   if (!(alfp = fopen(alertfile, "w")))
   {  
      
      fprintf(stderr,"aproxsum: open failed for alert file  %s.\n       %s. \n",
			alertfile, strerror(errno));
      return(-1);    			
   }
   cp = (char *) &record;
   for( i = 0; i <17; i++)
   {
      firstdate[i] = '\0';
      lastdate[i]  = '\0';             
   } 
   /*=======================================================================
    * acquire storage for maximum number of users and intialize every record
    *======================================================================*/
   p = (struct APROXHOST *) malloc(sizeof(struct APROXHOST)*(maxhosts+1)); 
   p->connects 	= 0;
   p->receives	= 0;
   p->bytess	= 0;
   p->bytesr 	= 0; 
   p->duration  = 0;
   p->port      = 0;
   p->ipaddr[0] = '\0';
   p->host[0]   = '\0';
   p2 = p;
   for ( i=0; i < maxhosts+1; i++)
   {
      memcpy(p2,p,sizeof(struct APROXHOST));
      p2++; 
   }  
   /*======================================================== 
    * Read the log file and look for aproxy record 
    *=======================================================*/        
   while(!feof(lfp))
   {   
      cp = fgets(buffer, MAXLEN, lfp);
      if( cp == NULL)
      {
         if(feof(lfp)) break; 
         if(ferror(lfp))
	 {
            fprintf(stderr,"aproxsum: error reading record %d in file  %s. %s\n       %s. \n",
			total_recs, logfile, strerror(errno));
            return(-1);   
         }
      } 
      total_recs++; 
      reclen = strlen(buffer);  
      total_logbytes = total_logbytes + reclen;   
      strcpy(record,buffer);   
      /* skip over comments, null lines and short records */
      if( (*cp == '#') || (*cp == '\n') || (reclen < 23) ) 
      { 
         trunc_recs++;
         goto getnextline; 
      } 
      if(total_recs == 1)
      {
         memcpy(firstdate,record,16);
      }
      memcpy(lastdate,record,16);
      /* skip over date & time */ 
      if( (token1 = strtok(record+16,"\t \n")) == NULL) 
         goto getnextline; 
      /* get name of program that wrote the message	*/
      if( (token1 = strtok(NULL,tokensep)) == NULL)
         goto getnextline;  
      token3 = "\n";
      token4 = "\n";
      /*========================================================
       * select aproxy records
       *========================================================*/
      if( strcmp(token1,"aproxy") == 0)
      {
         total_aproxrec++;
         total_apbytes = total_apbytes + reclen;   
	 token2 = strtok(NULL,":");	/* skip over [xxxxx]:	*/ 
         token2 = strtok(NULL," "); 
	 /*=====================================================
	  * process connect record 	
	  *=====================================================*/
         if( strcmp(token2,"connect") == 0)  
         {   
            if( reclen < 107 )
            {
               trunc_recs++;
               goto getnextline;
            }  
            token2  = strtok(NULL," ");	/* skip over "src_host"	*/
            token2  = strtok(NULL," ");  /* skip over "=" 	*/
            from    = strtok(NULL," ");  /* get from host	*/ 
            sipaddr = strtok(NULL,"() ");/* get src IP addr  	*/ 
            token2  = strtok(NULL," ");	/* get next token   	*/  
            if( strcmp(token2,"localIP") == 0)
            {
                localIP = strtok(NULL," ="); /* get local IP     */ 
                token2  = strtok(NULL," ");  /* get 'local_port  */	
            }
            if( strcmp(token2,"src_port") != 0 && strcmp(token2,"local_port") != 0  )
            {
               trunc_recs++;
               goto getnextline;
            }  
            token2  = strtok(NULL," ");  /* skip over "=" 	*/
            sport   = strtok(NULL," ");  /* get source port      */ 
            if ( !isalldigits(sport) ) 
            {
               trunc_recs++;
               goto getnextline;
            }
            token2  = strtok(NULL," ");  /* skip over "dest_host"*/
            token2  = strtok(NULL," ");  /* skip over "="        */
            tipaddr = strtok(NULL," \n");  /* get to host	 */  
            token2  = strtok(NULL," ");  /* skip over "dest_port"*/
            token2  = strtok(NULL," ");  /* skip over "="        */ 
            tport   = strtok(NULL,"\n ");
            if ( !isalldigits(tport) ) 
            {
               trunc_recs++;
               goto getnextline;
            }
            fbytes  = 0; 
            tbytes  = 0;
            numhosts = numhosts + 
	    		procaprox(from, sipaddr, sport, tipaddr, tport, 
					 fbytes, tbytes, 0, p, 1);
	    total_connects++; 
            goto getnextline;
         }
	 /*=================================================================
	  * process disconnect record 
          * This code will handle both old and new record formats
          * The new format contains localIP = x.x.x.x	
	  *=================================================================*/
         if( strcmp(token2,"disconnect") == 0)  
         {  
            if( reclen < 152 )
            {
               trunc_recs++;
               goto getnextline;
            }
            token2  = strtok(NULL," ");	/* skip over "src_host"	*/
            token2  = strtok(NULL," ");  /* skip over "=" 	*/
            from    = strtok(NULL," ");  /* get from host	*/ 
            sipaddr = strtok(NULL,"() ");/* skip over "("	*/
            token2  = strtok(NULL," ");	/* skip over "src_port"	*/  
            /* Skip over "localIP = x.x.x.x" if present	*/ 
            if( strcmp(token2,"localIP") == 0) 
            {
               token2 = strtok(NULL," ="); 
               token2 = strtok(NULL," ");
            } 
            if( strcmp(token2,"src_port") != 0 && strcmp(token2,"local_port") != 0 )
            {
               trunc_recs++;
               goto getnextline;
            }  
            token2  = strtok(NULL," ");  /* skip over "=" 		*/
            sport   = strtok(NULL,", "); /* get source port      	*/
 
            token2  = strtok(NULL," ");  /* skip over "dest_host"	*/
            token2  = strtok(NULL," ");  /* skip over "="        	*/
            tipaddr = strtok(NULL," \n");  /* get to host	 	*/  
            token2  = strtok(NULL," ");  /* skip over "dest_port"	*/
            token2  = strtok(NULL," ");  /* skip over "="        	*/ 
            tport   = strtok(NULL,"\n "); 
            token2  = strtok(NULL," ");  /* skip over "input"	 	*/
            token2  = strtok(NULL," ");  /* skip over "="        	*/  
            token2  = strtok(NULL," ");  /* get input bytes 	 	*/
            if ( !isalldigits(token2) )
            {
               trunc_recs++;
               goto getnextline;
            }
	    fbytes  = atoi(token2);  
	    total_bytesr = total_bytesr + fbytes;
            token2  = strtok(NULL," ");  /* skip over "bytes"	 	*/
	    token2  = strtok(NULL," ");  /* skip over "output"	 	*/
            token2  = strtok(NULL," ");  /* skip over "="        	*/ 
	    token2  = strtok(NULL," ");  /* get output bytes 	 	*/
	    if ( !isalldigits(token2) ) 
            {
               trunc_recs++;
               goto getnextline;
            }
            tbytes  = atoi(token2);    
	    total_bytess = total_bytess + tbytes; 
            token2  = strtok(NULL,", "); /* skip over "bytes,"   	*/ 
            token2  = strtok(NULL," ");  /* skip over "Blocks"   	*/    
            token2  = strtok(NULL," ");  /* skip over "="	 	*/ 
            token2  = strtok(NULL,", "); /* skip over "block size,"   	*/ 
            token2  = strtok(NULL," ");  /* skip over "duration"   	*/    
            token2  = strtok(NULL," ");  /* skip over "="	 	*/ 
            token2  = strtok(NULL,": "); /* get hrs		 	*/
            token3  = strtok(NULL,":");  /* get min		 	*/
            token4  = strtok(NULL," \n ");/* get sec		 	*/
            if( !isalldigits(token2) || !isalldigits(token3) || 
		!isalldigits(token4) )
	    {
               trunc_recs++;
               goto getnextline;
	    } 
            duration = atoi(token2)*3600 + atoi(token3)*60 + atoi(token4);
	    total_duration = total_duration + duration; 

	    /*============================================================
             * process 
	     *===========================================================*/
	    procaprox(from, sipaddr, sport, tipaddr, tport, 
					 tbytes, tbytes, duration, p, 0);
            procaprox(" ",  tipaddr, tport, sipaddr, sport, 
					 fbytes, fbytes, 0, p, 0);
	    total_disconnects++; 
            goto getnextline;
         }
         /*============================================================
	  * process security alerts
	  *============================================================*/
	 if( strcmp(token2,"Security") == 0)  
         {  
            if( reclen < 88 )
            {
               trunc_recs++;
               goto getnextline;
            } 
            total_alerts++;
            fputs(buffer, alfp);
            goto getnextline;
         }
      }

getnextline:;   
   } 
   
   /* print summary reports */ 
   printf("\n Application Proxy Summary Report \n");
   printf("\n log file name: %s\n",logfile);
   printf("\n total records in log file      = %d",total_recs);
   printf("\n total bytes in log file        = %d",total_logbytes); 
   printf("\n number of truncated records    = %d",trunc_recs);
   printf("\n date of first record           = %s",firstdate);
   printf("\n date of last record            = %s",lastdate);
   printf("\n number of aproxy records     = %d",total_aproxrec);
   printf("\n total size of aproxy records = %d",total_apbytes); 
   printf("\n number of unique hosts         = %d",numhosts);
   printf("\n number of connect messages     = %d",total_connects); 
   printf("\n number of disconnect messages  = %d",total_disconnects);
   printf("\n number of Security Alerts      = %d",total_alerts); 
   printf("\n total bytes sent               = %.0f",total_bytess);
   printf("\n total bytes received           = %.0f",total_bytesr);
   printf("\n total duration  (sec)          = %d",total_duration);
   printf("\n\n"); 
   /*====================================================================
    * sort buy user E-mail address
    * sort by bytes sent and print 
    * sort by messages sent and print
    * sort by bytes received and print 
    * sort by messages received and print
    *====================================================================*/

   qsort(p, numhosts, sizeof(struct APROXHOST),
		(int(*) (const void *, const void *))comprhosts);
   printf("\n\n Top hosts by bytes sent\n"); 
   qsort(p, numhosts, sizeof(struct APROXHOST), 
		(int(*) (const void *, const void *))comprbs);
   prthosts(p,top);

   printf("\n\n Top hosts by connect requests\n"); 
   qsort(p, numhosts, sizeof(struct APROXHOST), 
		(int(*) (const void *, const void *))comprcon);
   prthosts(p,top);  

   printf("\n\n Top hosts by bytes received\n"); 
   qsort(p, numhosts, sizeof(struct APROXHOST), 
		(int(*) (const void *, const void *))comprbr);
   prthosts(p,top);
  
   printf("\n\n Top hosts by receives\n"); 
   qsort(p, numhosts, sizeof(struct APROXHOST), 
		(int(*) (const void *, const void *))comprrecv);
   prthosts(p,top);

   free((void *) p);	/* free data array	*/
   return(0);
}  
/*========================================================================
 *
 *========================================================================*/
int prthosts(struct APROXHOST *p, int top)
{                            
   struct APROXHOST *p2; 
   int	i; 
   int	j;
   int		numhosts = 0;
   int		bs 	 = 0;
   int		conns	 = 0;
   int		br	 = 0;
   int		recvs	 = 0; 
   int		totalsec = 0;
   j = top;   
   if( (top < 1) || (top > maxhosts) ) j = maxhosts;
   p2 = p;
   printf("\n             initiated  bytes       received  ");
   printf("\nbytes sent   connects   received    connects   PORT   duration     \
IP address   hostname\n"); 
   for (i =0; i < j; i++)
   {
      printf("\n %10d  %9d  %10d  %9d  %5d %9d %16s %s",
         p2->bytess, p2->connects, p2->bytesr, p2->receives, p2->port, 
	 p2->duration, p2->ipaddr, p2->host); 
      bs    = bs    + p2->bytess;
      conns = conns + p2->connects;
      br    = br    + p2->bytesr;
      recvs = recvs + p2->receives; 
      totalsec = totalsec + p2->duration;
      p2++;
      if( strlen(p2->host) == 0) break;
   }
   if(top) i--; 
   if (top != 1)
   {
      printf("\n\ntotals:");
      printf("\n %10d  %9d  %10d  %9d        %9d   %9d\n",
               bs, conns, br, recvs, totalsec, i+1);  
   }
   else printf("\n");
   return(0);
} 
/*===========================================================================
 * compare host names
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int comprhosts(char **s1, char **s2)
{
   struct APROXHOST *p1;
   struct APROXHOST *p2;
   p1 = (struct APROXHOST *) s1;
   p2 = (struct APROXHOST *) s2;
   return( strcmp(p1->ipaddr,p2->ipaddr) );
} 
/*===========================================================================
 * compare bytes sent
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int comprbs(char **s1, char **s2)
{
   struct APROXHOST *p1;
   struct APROXHOST *p2;
   p1 = (struct APROXHOST *) s1;
   p2 = (struct APROXHOST *) s2;
   if(p1->bytess >= p2->bytess) return(-1); 
   return(1);
} 
/*===========================================================================
 * compare bytes received
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int comprbr(char **s1, char **s2)
{
   struct APROXHOST *p1;
   struct APROXHOST *p2;
   p1 = (struct APROXHOST *) s1;
   p2 = (struct APROXHOST *) s2;
   if(p1->bytesr >= p2->bytesr) return(-1); 
   return(1);
} 
/*===========================================================================
 * compare connects sent
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int comprcon(char **s1, char **s2)
{
   struct APROXHOST *p1;
   struct APROXHOST *p2;
   p1 = (struct APROXHOST *) s1;
   p2 = (struct APROXHOST *) s2;
   if(p1->connects >= p2->connects) return(-1); 
   return(1);
} 
/*===========================================================================
 * compare messages received
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int comprrecv(char **s1, char **s2)
{
   struct APROXHOST *p1;
   struct APROXHOST *p2;
   p1 = (struct APROXHOST *) s1;
   p2 = (struct APROXHOST *) s2;
   if(p1->receives >= p2->receives) return(-1); 
   return(1);
} 
/*============================================================================
 * process aproxy records
 * inputs:
 *	ptr to source host name
 *	ptr to source ip addr
 *	ptr to source port #
 *	ptr to dest source ip addr
 *	ptr to dest port #  
 *      number of bytes from host
 *	number of bytes to host
 *	pointer to first record
 *	number of connects
 *
 * outputs:  
 *	returns number of records added
 *
 *===========================================================================*/
int procaprox(char *from, char *sipaddr, char *sport, char *to, char *tport,
                int fbytes, int tbytes, int duration, struct APROXHOST *p, int cntflg)
{  
   struct APROXHOST *p2;
   int	i;
   int	newhost; 
   int  tofound;
   int	fromfound; 
   p2 = p;
   tofound 	= 0;
   fromfound	= 0; 
   newhost 	= 0;
   for( i = 0; i < maxhosts; i++) 
   {  
      /* match source IP address and source port		*/ 
      if( (strcmp(p2->ipaddr,sipaddr) ==0) && (p2->port == atoi(sport)))
      {
         p2->bytess = p2->bytess + fbytes; 
         if(cntflg) p2->connects++; 
         p2->duration = p2->duration +duration;
         fromfound++;
      } 
      /* match destination IP address and destination port	*/
      if( (strcmp(p2->ipaddr,to)  == 0) && ( p2->port == atoi(tport) ) )
            {
         p2->bytesr = p2->bytesr + tbytes; 
         if(cntflg) p2->receives++; 
         p2->duration = p2->duration +duration;
         tofound++;
      }
      if( strlen(p2->host) == 0) break;
      if( fromfound && tofound) return(0);
      p2++;
   }
   if(!fromfound)
   {
      p2->bytess  = fbytes;
      p2->connects = 1;
      strcpy(p2->host,from); 
      strcpy(p2->ipaddr,sipaddr);
      p2->port = atoi(sport);
      newhost++;
   } 
   if(!tofound)
   {  
      if(!fromfound) p2++;
      p2->bytesr = tbytes;
      p2->receives = 1;
      strcpy(p2->host,to); 
      strcpy(p2->ipaddr,to);
      p2->port = atoi(tport); 
      newhost++; 
   } 
   return(newhost);
} 
/*============================================================================ 
 * print usage message
 *===========================================================================*/
void usage(void)
{
   
   printf("\n usage: aproxsum [-l logfile ] [-t nnn ]  \n");
   printf("\n        where:\n");  
   printf("        -l logfile   full path name of the log file. \n");
   printf("        -t nnn       list top nnn users\n");
   
   return;
}  
/*======================================================================
 * function: 	isalldigits
 * input:	a pointer to a character string 
 * output:	returns 1 if all the characters in the string are digits 
 *=====================================================================*/  
int isalldigits(char *s)  
{
   while(isdigit(*s))
      s++;
   return(*s == '\0');  	/* if at end of string return 1	*/
} 
