/*============================================================================
 * filename:		util/reports/smrpt/smwrpt.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=============================================================================
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  
 * All rights reserved.  
 *
 * written by:		R. James Livermore                    
 * Date written:	March     18, 1995
 * Last modified:	February   8, 2000 - addd LPL
 *
 * function:		secure mail wrapper report program 	
 *       
 * 	The smrpt program reads the system log and produces a report
 *	on e-mail usage.
 *
 *  	1. read the file until eof
 *	2. capture all smwrap and smwrapd messages.
 *
 *============================================================================*/

static  char	ident[] = "@(#) smrpt, V 1.0, 00/01/21"
 			  " (C) Freemont Avenue Software, Inc."
			  " 1995-2000";
#include	<search.h>		/* for qsort	*/
#include	<sys/types.h>
#include	<sys/file.h>
#include	<sys/stat.h>
#include	<stdio.h> 
#include	<string.h>		/* added for C++ */
#include	<time.h>
#include	<ctype.h>
#include	<syslog.h>
#include	<netdb.h>
 
#include	<unistd.h>  		/* added for C++ */
#include	<stdlib.h>		/* added for C++ */  
   
struct SMUSER {
   int		msgsent;
   int		msgrec;
   float	bytess;
   float	bytesr;
   char	user[84];  
}; 
/*==========================
 * external variables
 *==========================*/  
extern   int	errno;
#ifdef __hppa
#define		MAXUSERS 	16000
#else
#define		MAXUSERS 	16000
#endif
   int	maxusers;
/*======================================================================
 * function prototypes
 *======================================================================*/  
int checkdomain(char *domain, char *from, char *to);
int compradr(char **s1, char **s2);  /* compare address			*/ 
int comprbr(char **s1, char **s2);   /* compare bytes received		*/
int comprbs(char **s1, char **s2);   /* compare bytes sent		*/
int comprms(char **s1, char **s2);   /* compare messages sent		*/
int comprmr(char **s1, char **s2);   /* compare messages received	*/
int isalldigits(char *s); 
int prtusers(struct SMUSER *p, int top, char *domain); 
int procsmwrap(char *from, char *to, int bytes, struct SMUSER *p);
void usage(void);  
 
/*============================================================================                        
 * main
 * inputs: optional input arguments   
 *	   -d domain    report on domain only   
 *         -h 		for help
 *	   -l logfile	specify logfile name	
 *	   -m nnnnn     maximum number of users to sort	
 *         -t nnn	specify printing of top nnn users only
 * output:
 *	return  0 if ok
 *	return -1 on error
 *===========================================================================*/
int main(int argc, char *argv[])
{   
   /* log file statistics		*/ 
   float	cpu		= 0;
   int		trunc_recs	= 0;   
   int		trunc_recnum	= 0;
   int		total_recs	= 0;
   int		total_msg 	= 0;
   float	total_bytes 	= 0.; 
   float	total_cpu	= 0.;
   int		total_smwraprec = 0;  
   char		firstdate[17];
   char		lastdate[17]; 
   /* 					*/ 
   struct	SMUSER	*p;             /* ptr to data area		*/
   struct	SMUSER	*p2;
   int		i;
   int		ibytes;  
   int		numaddr		= 0;    
   int		tempmax		= 1;
   int		top		= 0;
#define	MAXLEN 1024
#ifdef _AIX
   static char 	logfile[256] = "/var/adm/syslog";
#elif __hppa
   static char 	logfile[256] = "/var/adm/syslog/syslog.log";
#elif linux
   static char  logfile[256] = "/var/log/messages";
#else   
   static char 	logfile[256] = "/var/adm/messages";
#endif  
   char		buffer[MAXLEN];    
   char		domain[256];
   char		record[MAXLEN];
   FILE 	*lfp;			/* log file ptr	*/ 
   char 	*cp;  
   char		*cp2;
   char		*token1;
   char		*from;
   char		*to;
   char		*bytes; 
   static char	tokensep[] 	= "\t,:;=[] \n";
   maxusers = MAXUSERS;   
   memset(domain,0,sizeof(domain));
   /*===================================================
    * check for input arguments
    *==================================================*/
   if( argc > 1)
   {
      if( argv[1][0] != '-')   
      {
         printf("\n smrpt:  invalid leading character '%c' \n",argv[1][0]);
         usage();
         return(-1);        
      }
      /* check to see if user is asking for help */
      if ((strcmp(argv[1], "-h")    == 0) ||
          (strcmp(argv[1], "-help") == 0) ||
          (strcmp(argv[1], "help")  == 0) ||
          (strcmp(argv[1], "?")     == 0)
         ) 
      {
         usage();
         return(0);
      } 
      for(i = 1; i < argc ; i++)
      {
         switch( argv[i][1])
         { 
            case 'd':  
               if( argc < i+2)
               {
                  printf("\n missing domain name after %s\n",argv[i]);  
                  usage();
                  return(-1);
               } 
               if( strlen(argv[i+1]) > 255)
               {
                  printf("\n smrpt: domain name too long. length = %d\n",
			strlen(argv[i+1])); 
                  usage();
                  return(-1);
               }
	       strcpy(domain,argv[i+1]);
               i++;
               break;
            case 'l':  
               if( argc < i+2)
               {
                  printf("\n missing filename after %s\n",argv[i]);  
                  usage();
                  return(-1);
               } 
               if( strlen(argv[i+1]) > 255)
               {
                  printf("\n smrpt: file name too long. length = %d\n",
			strlen(argv[i+1])); 
                  usage();
                  return(-1);
               }
	       strcpy(logfile,argv[i+1]);
               i++;
               break;  
            case 'm':  
               if( argc < i+2)
               {
                  printf("\n missing argument after %s\n",argv[i]);  
                  usage();
                  return(-1);
               } 
               if( !isalldigits(argv[i+1]) )
               {
                  printf("\n smrpt: the paramter following -m is not all digits.\n"); 
                  usage();
		  return(-1);
               }
               maxusers = atoi(argv[i+1]);
               i++;
               break;
            case 't':  
               if( argc < i+2)
               {
                  printf("\n missing argument after %s\n",argv[i]);  
                  usage();
                  return(-1);
               } 
               if( !isalldigits(argv[i+1]) )
               {
                  printf("\n smrpt: the paramter following -t is not all digits.\n"); 
                  usage();
		  return(-1);
               }
               top = atoi(argv[i+1]);
               i++;
               break;
            default:
               printf("\n smrpt: invalid input parameter '%s' \n",argv[i]);
	       usage();
	       return(-1);               
         }
      }
   }
   /* open log file	*/  
   if (!(lfp = fopen(logfile, "r")))
   {  
      fprintf(stderr,"smrpt: open failed for logfile  '%s'.\n       %s. \n",
			logfile, strerror(errno));
      return(-1);    			
   }
   cp = (char *) &buffer;
   for( i = 0; i <17; i++)
   {
      firstdate[i] = '\0';
      lastdate[i]  = '\0';             
   } 
   /*=======================================================================
    * acquire storage for maximum number of users and intialize every record
    *======================================================================*/
   p = (struct SMUSER *) malloc(sizeof(struct SMUSER)*(maxusers+2)); 
   p->msgsent 	= 0;
   p->msgrec	= 0;
   p->bytess	= 0;
   p->bytesr 	= 0;
   p->user[0]   = '\0';
   p2 = p;
   for ( i=0; i < maxusers+2; i++)
   {
      memcpy(p2,p,sizeof(struct SMUSER));
      p2++; 
   }  
   /*======================================================== 
    * Read the log file and look for smwrap or smwrapd record 
    *=======================================================*/        
   while(!feof(lfp))
   {   
      cp = fgets(buffer, MAXLEN, lfp); 
      if( cp == NULL)
      {
         if(feof(lfp)) break; 
         if(ferror(lfp))
	 {
            fprintf(stderr,"smrpt: error reading record %d in file  %s. %s\n       %s. \n",
			total_recs, logfile, strerror(errno));
            return(-1);   
         }
      }
      total_recs++;   
      /*========================================================
       * skip over comments, null lines and short records 
       *=======================================================*/
      if( (*cp == '#') || (*cp == '\n') || (strlen(buffer) < 23) ) 
      { 
         trunc_recs++;
         goto getnextline; 
      } 
      if(total_recs == 1)
      {
         memcpy(firstdate,buffer,16);
      }
      memcpy(lastdate,buffer,16);
      /* skip over date & time */ 
      if( (token1 = strtok(buffer+16,"\t \n")) == NULL) 
         goto getnextline; 
      /* get name of program that wrote the message	*/
      if( (token1 = strtok(NULL,tokensep)) == NULL)
         goto getnextline;
      /*===================================================================
       * Process smwrap records
       *==================================================================*/
      if( strcmp(token1,"smwrap") == 0)
      {
         total_smwraprec++;   
	 if( (token1 = strtok(NULL,":")) == NULL)
         {
	    trunc_recs++;
            goto getnextline;
         } 
         if( (token1 = strtok(NULL," ")) == NULL)
         {
	    trunc_recs++;
            goto getnextline;
         }  
         /*==============================================
          * Process "exiting host" line.
          *=============================================*/
         if( strcmp(token1,"exiting") == 0)  
         {
            if( (token1 = strtok(NULL,")")) == NULL)
            {
	       trunc_recs++;
               goto getnextline;
            } 
            if( (token1 = strtok(NULL,tokensep)) == NULL) /* get "bytes" */ 
            {
	       trunc_recs++;
               goto getnextline;
            } 
            if( strcmp(token1,"bytes") != 0)
            {   
	        trunc_recs++;
                goto getnextline;
            } 
	    token1 = strtok(NULL,tokensep);	/* number of bytes	*/
            if( (token1 = strtok(NULL,tokensep)) == NULL) /* get "cpu"	*/
            {   
	       trunc_recs++;
               goto getnextline;
            } 
            if( strcmp(token1,"cpu") != 0)
            {   
	        trunc_recs++;
                goto getnextline;
            }
            if( (token1 = strtok(NULL,tokensep)) == NULL) /* get cpu sec */
            {   
	        trunc_recs++;
                goto getnextline;
            } 
            sscanf(token1,"%f",&cpu);
            total_cpu = total_cpu + cpu;
            goto getnextline;
         }  /* end of exiting record	*/ 
         /*==============================================
          * Process "host = " line.
          *=============================================*/
         if( strcmp(token1,"host") != 0)  goto getnextline;
         if( (token1 = strtok(NULL,")")) == NULL)
         {   
	    trunc_recs++;
            goto getnextline;
         }  
         if( (token1 = strtok(NULL,tokensep)) == NULL)
         {   
	    trunc_recs++;
            goto getnextline;
         }   
         if( strcmp(token1,"bytes") != 0)
         {
            trunc_recs++;
            goto getnextline;
         } 
         if( (bytes = strtok(NULL,tokensep)) == NULL)
         {   
	    trunc_recs++;
            goto getnextline;
         }   
         if( !isalldigits(bytes) )        goto getnextline;
         ibytes = atoi(bytes);
         total_bytes = total_bytes + ibytes;
         if( (token1 = strtok(NULL,tokensep)) == NULL)   
         {   
	    trunc_recs++;
            goto getnextline;
         }   
         if( strcmp(token1,"from") != 0)  goto getnextline;
         if( (from = strtok(NULL,tokensep)) == NULL)
         {   
	    trunc_recs++;
            goto getnextline;
         }    
         if( (token1 = strtok(NULL,tokensep)) == NULL)  
         {   
	    trunc_recs++;
            goto getnextline;
         }  
         if( strcmp(token1,"to") != 0)    goto getnextline;
         if( (to = strtok(NULL,tokensep)) == NULL)
         {   
	    trunc_recs++;
            goto getnextline;
         }   
         /*====================================================
          * Convert to and from e-mail addresses to lower case
          *===================================================*/
         cp2 = from;
         for( i = 0; i < strlen(from); i++)
         {
            *cp2 = tolower((int)*cp2);
            cp2++;      
         }
         cp2 = to;
         for( i = 0; i < strlen(to); i++)
         {
            *cp2 = tolower((int)*cp2);
            cp2++;      
         }
         if( strlen(domain) > 0 )
            checkdomain(domain,from,to);
         numaddr = numaddr + procsmwrap(from,to,ibytes, p);
	 total_msg++;  
         if( numaddr > maxusers)
         {
            printf("\n Warning:  The number of unique e-mail addresses"
                   " %d, exceeded the maximum value of %d.\n",
		   numaddr, maxusers);                      
            printf("           Processing of additional records halted.\n"); 
            break; 
         }
      } /* end of smwrap record processing	*/

getnextline:;   
   } 
   
   /* print summary reports */ 
   printf("\n Sendmail Wrapper V 1.0 Summary Report \n");
   printf("\n log file name: %s\n",logfile);  
   if(top)
      printf("\n top %d users were requested",top);
   if( strlen(domain) > 0) 
   {
      printf("\n summary for domain             = %s",domain);
   }
   printf("\n date of first record           = %s",firstdate);
   printf("\n date of last record            = %s",lastdate);
   printf("\n total records in log file      =   %10d",total_recs); 
   printf("\n number of truncated records    =   %10d",trunc_recs);
   printf("\n total number of smwrap records =   %10d",total_smwraprec); 
   printf("\n number of unique addresses     =   %10d",numaddr);
   printf("\n number of E-mail messages      =   %10d",total_msg);
   printf("\n total E-mail bytes             = %12.0f",total_bytes);  
   if( total_msg > 0 )
   {
      printf("\n average E-mail message size    =   %12.1f",
             total_bytes/total_msg); 
   } 
   printf("\n total cpu time (sec)           =    %12.2f",total_cpu);
   if( total_msg > 0 )
   {	 
      printf("\n average cpu sec per message    =    %12.2f",
	     total_cpu/total_msg);
   }
   printf("\n\n"); 
   /*====================================================================
    * sort buy user E-mail address
    * sort by bytes sent and print 
    * sort by messages sent and print
    * sort by bytes received and print 
    * sort by messages received and print
    *====================================================================*/

   qsort(p, numaddr, sizeof(struct SMUSER),
		(int(*) (const void *, const void *))compradr);  
   if( top > numaddr)
   {                 
      printf("\n The top number requested %d is > the number of unique "
	      "addresses %d, will use the smaller value.",
	      top, numaddr);
      top = numaddr;   
   }
   if( top ) 
      printf("\n\n Top %d senders by bytes sent\n",top); 
   else
      printf("\n\n Top senders by bytes sent\n"); 
   qsort(p, numaddr, sizeof(struct SMUSER), 
		(int(*) (const void *, const void *))comprbs);
   prtusers(p,top,domain);
   if( top) 
      printf("\n\n Top %d senders by messages sent\n",top); 
   else
      printf("\n\n Top senders by messages sent\n"); 
   qsort(p, numaddr, sizeof(struct SMUSER), 
		(int(*) (const void *, const void *))comprms);
   prtusers(p,top,domain);  
   if( top)
      printf("\n\n Top %d receiver by bytes received\n",top); 
   else
      printf("\n\n Top receiver by bytes received\n"); 
   qsort(p, numaddr, sizeof(struct SMUSER), 
		(int(*) (const void *, const void *))comprbr);
   prtusers(p,top,domain);
   if ( top)
      printf("\n\n Top %d receiver by messages received\n",top); 
   else  
      printf("\n\n Top receiver by messages received\n"); 
   qsort(p, numaddr, sizeof(struct SMUSER), 
		(int(*) (const void *, const void *))comprmr);
   prtusers(p,top,domain);

   free((void *) p);	/* free data array	*/
   return(0);
} 
/*========================================================================
 *
 *========================================================================*/
int prtusers(struct SMUSER *p, int top, char *domain)
{                            
   struct SMUSER *p2; 
   int	i; 
   int	j;
   int		numusers  = 0;
   float	bs	  = 0;        	/* bytes sent		*/  
   float        bsother   = 0;
   int		msgs	  = 0; 		/* messages		*/      
   int          msgsother = 0;
   float	br	  = 0;		/* bytes received	*/  
   float        brother   = 0;  
   int 		msgr	  = 0;  	
   int		msgrother = 0;
   j = top;   
   if( (top < 1) || (top > maxusers) ) j = maxusers;
   p2 = p;
   printf("\n                msg           bytes      msg  ");
   printf("\nbytes sent      sent          received   received  address\n"); 
   for (i =0; i < j; i++)
   {   
      if( strlen(domain) > 0 && strcmp(p2->user,"other") == 0)
      {
         bsother   = bsother + p2->bytess; 
         msgsother = msgsother  + p2->msgsent;
         brother   = brother + p2->bytesr;
         msgrother = msgrother + p2->msgrec; 
         i--;				/* don't count "other" in top n */
      }
      else
      {
         printf("\n %12.0f  %9d  %12.0f  %9d  %s",
               p2->bytess, p2->msgsent, p2->bytesr, p2->msgrec, p2->user); 
         bs   = bs   + p2->bytess;
         msgs = msgs + p2->msgsent;
         br   = br   + p2->bytesr;
         msgr = msgr + p2->msgrec; 
      }
      p2++;  				/* get next record */
      if( strlen(p2->user) == 0) 
      {  
         i++;
         break;
      }
      
   }
/*   if(top) i--;  */
   if (top != 1)
   {
      printf("\n\n\ntotals:"); 
      printf("\n %12.0f  %9d  %12.0f  %9d  %9d\n",
               bs, msgs, br, msgr, i);  
/*                 bs, msgs, br, msgr, i+1);  */
   }
   else printf("\n");     
   if( strlen(domain) > 0)
   {
      printf("\n %12.0f  %9d  %12.0f  %9d  %s\n",
               bsother, msgsother, brother, msgrother, "other");       
   }
   return(0);
} 
/*===========================================================================
 * check domain names
 * inputs: 
 *	domain name ptr
 *      from e-mail address ptr 
 *      to   e-mail address ptr
 * outputs: 
 *	if "from" address is outside of the domain then change to "other"   
 *	if "to"   address is outside of the domain then change to "other"
 *     
 *===========================================================================*/
int checkdomain(char *domain, char *from, char *to)
{
   char *cp;
   char wrk[256];   
   cp = strchr(from,'@'); 
   strcpy(wrk,cp+1);  
   wrk[strlen(wrk)-1] = '\0';
   if( strcmp(wrk,domain) !=0)
      strcpy(from,"other"); 
   cp = strchr(to,'@'); 
   strcpy(wrk,cp+1); 
   wrk[strlen(wrk)-1] = '\0';
   if( strcmp(wrk,domain) !=0)
      strcpy(to,"other");
   return(0);
} 

/*===========================================================================
 * compare address
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int compradr(char **s1, char **s2)
{
   struct SMUSER *p1;
   struct SMUSER *p2;
   p1 = (struct SMUSER *) s1;
   p2 = (struct SMUSER *) s2;
   return( strcmp(p1->user,p2->user) );
} 
/*===========================================================================
 * compare bytes sent
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int comprbs(char **s1, char **s2)
{
   struct SMUSER *p1;
   struct SMUSER *p2;
   p1 = (struct SMUSER *) s1;
   p2 = (struct SMUSER *) s2;
   if(p1->bytess >= p2->bytess) return(-1); 
   return(1);
} 
/*===========================================================================
 * compare bytes received
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int comprbr(char **s1, char **s2)
{
   struct SMUSER *p1;
   struct SMUSER *p2;
   p1 = (struct SMUSER *) s1;
   p2 = (struct SMUSER *) s2;
   if(p1->bytesr >= p2->bytesr) return(-1); 
   return(1);
} 
/*===========================================================================
 * compare messages sent
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int comprms(char **s1, char **s2)
{
   struct SMUSER *p1;
   struct SMUSER *p2;
   p1 = (struct SMUSER *) s1;
   p2 = (struct SMUSER *) s2;
   if(p1->msgsent >= p2->msgsent) return(-1); 
   return(1);
} 
/*===========================================================================
 * compare messages received
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int comprmr(char **s1, char **s2)
{
   struct SMUSER *p1;
   struct SMUSER *p2;
   p1 = (struct SMUSER *) s1;
   p2 = (struct SMUSER *) s2;
   if(p1->msgrec >= p2->msgrec) return(-1); 
   return(1);
} 
/*============================================================================
 * process smwrap records
 * inputs:
 *	from address pointer
 *	to address pointer  
 *      number of bytes in message
 *	pointer to first record
 *
 * outputs:  
 *	returns number of records added
 *
 *===========================================================================*/
int procsmwrap(char *from, char *to, int bytes, struct SMUSER *p)
{  
   struct SMUSER *p2;
   int	i;
   int	newadr; 
   int  tofound;
   int	fromfound; 
   p2 = p;
   tofound 	= 0;
   fromfound	= 0; 
   newadr 	= 0;
   for( i = 0; i < maxusers; i++) 
   {
      if( strcmp(p2->user,from) == 0)
      {
         p2->bytess = p2->bytess + bytes; 
         p2->msgsent++; 
         fromfound++;
      }
      if( strcmp(p2->user,to)  == 0)
            {
         p2->bytesr = p2->bytesr + bytes; 
         p2->msgrec++; 
         tofound++;
      }
      if( strlen(p2->user) == 0) break;  /* end of list */
      if( fromfound && tofound) return(0);
      p2++;
   }
   if(!fromfound)
   {
      p2->bytess  = bytes;
      p2->msgsent = 1;
      strcpy(p2->user,from);     
      if( strcmp(from,"other") )
         newadr++;
   } 
   if(!tofound)
   {  
      if(!fromfound)
      {
         if( strcmp(to,from) != 0)
            p2++;
      }
      p2->bytesr = bytes;
      p2->msgrec = 1;
      strcpy(p2->user,to); 
      if( strcmp(to,"other") )
         newadr++; 
   } 
   return(newadr);
} 
/*============================================================================ 
 * print usage message
 *===========================================================================*/
void usage(void)
{
   
   printf("\n usage: smrpt [-d domain] [-l logfile ] [-m nnnnnn] [-t nnn ]  \n");
   printf("\n        where:\n"); 
   printf("        -d domain    list e-mail addresses for a single domain.\n"); 
   printf("        -l logfile   full path name of the log file. \n"); 
   printf("        -m nnnnnn    maximum number of e-mail addresses to sort.\n");
   printf("        -t nnn       list top nnn users\n");
   
   return;
}  
/*==================================================================
 * function: 	isalldigits
 * input:	a pointer to a character string 
 * output:	returns 1 if all the characters in the string are digits 
 *==================================================================*/  
int isalldigits(char *s)  
{
   while(isdigit(*s))
      s++;
   return(*s == '\0');
} 
