/*============================================================================
 * filename:		util/reports/smrptx/smwrptx.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=============================================================================
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  
 * All rights reserved.                                 
 *
 * written by:		R. James Livermore
 * Date written:	March     22, 1995
 * Last modified:	February   8, 2000 - LPL
 *
 * function:		secure mail wrapper exceptions report program 	
 *       
 * 	The smrpt program reads the system log and produces an exception report
 *	on e-mail usage.
 *
 *  	1. read the file until eof
 *	2. capture all smwrap and smwrapd messages.
 *
 *============================================================================*/

static  char	ident[] = "@(#) smrptx, V 1.0, 00/01/21"
 			  " (C) Freemont Avenue Software, Inc."
			  " 1995-2000";
#include	<search.h>		/* for qsort	*/
#include	<sys/types.h>
#include	<sys/file.h>
#include	<sys/stat.h>
#include	<stdio.h> 
#include	<string.h>		/* added for C++ */
#include	<time.h>
#include	<ctype.h>
#include	<syslog.h>
#include	<netdb.h>
 
#include	<unistd.h>  		/* added for C++ */
#include	<stdlib.h>		/* added for C++ */  
 
 
/*==========================
 * external variables
 *==========================*/  
extern   int	errno;

/*======================================================================
 * function prototypes
 *======================================================================*/ 

int isalldigits(char *s); 

void usage(void);  
/*============================================================================                        
 * main
 * inputs: optional input arguments
 *         -h 		for help
 *	   -l logfile	specify logfile name		
 *         
 * output:
 *	return  0 if ok
 *	return -1 on error
 *===========================================================================*/
int main(int argc, char *argv[])
{   
   /* log file statistics		*/ 
   int		trunc_recs	= 0;   /* eof adds 1 so must sub 1	*/
   int		trunc_recnum	= 0;
   int		total_recs	= 0;
   int		total_delv	= 0;	/* total delivered		*/
   int		total_smwraprec = 0; 
   int		total_smwrapdrec = 0;   
   int		total_vrfy	= 0;
   int		total_errors	= 0;
   int		total_expn	= 0;
   int		total_secalert  = 0;
   int		total_seq	= 0; 
   int		total_warn	= 0;
   char		firstdate[17];
   char		lastdate[17]; 
   /* 					*/ 
   int		detail		= 0;
   int		i;
   int		top		= 0;
   #define	MAXLEN 1024
   /* files		*/
#ifdef _AIX
   static char 	logfile[256]   = "/var/adm/syslog";
#elif __hppa
   static char 	logfile[256]   = "/var/adm/syslog/syslog.log";
#elif linux
   static char  logfile[256] = "/var/log/messages";
#else   
   static char 	logfile[256]   = "/var/adm/messages";
#endif   
   static char	vrfyfile[128]  = "/tmp/smwrap.vrfy";
   static char	expnfile[128]  = "/tmp/smwrap.expn";
   static char	warnfile[128]  = "/tmp/smwrap.warn";
   static char	errfile[128]   = "/tmp/smwrap.error";
   static char	alertfile[128] = "/tmp/smwrap.alerts";
   static char	seqfile[128]   = "/tmp/smwrap.sequester";
   FILE 	*lfp;			/* log file ptr		*/ 
   FILE		*vfp;			/* vrfy file ptr	*/
   FILE		*exfp;		        /* expn file ptr	*/
   FILE		*erfp;     		/* error file ptr	*/
   FILE		*alfp;                  /* alert file ptr	*/
   FILE		*wfp;
   FILE		*sfp;	
   char		record[MAXLEN]; 
   char		orecord[MAXLEN];
   char 	*cp; 
   char		*token1;
   static char	tokensep[] 	= "\t,:;=[] \n";
   /*===================================================
    * check for input arguments
    *==================================================*/
   if( argc > 1)
   {
      if( argv[1][0] != '-')   
      {
         printf("\n smrptx:  invalid leading character '%c' \n",argv[1][0]);
         usage();
         return(-1);        
      }
      /* check to see if user is asking for help */
      if ((strcmp(argv[1], "-h")    == 0) ||
          (strcmp(argv[1], "-help") == 0) ||
          (strcmp(argv[1], "help")  == 0) ||
          (strcmp(argv[1], "?")     == 0)
         ) 
      {
         usage();
         return(0);
      } 
      for(i = 1; i < argc ; i++)
      {
         switch( argv[i][1])
         {  
            case 'd':
               detail++;
	       break;
            case 'l':  
               if( argc < i+2)
               {
                  printf("\n missing filename after %s\n",argv[i]);  
                  usage();
                  return(-1);
               } 
               if( strlen(argv[i+1]) > 255)
               {
                  printf("\n smrptx: file name too long. length = %d\n",
			strlen(argv[i+1])); 
                  usage();
                  return(-1);
               }
	       strcpy(logfile,argv[i+1]);
               i++;
               break;
            case 't':   
               if( argc < i+2)
               {
                  printf("\n missing argument after %s\n",argv[i]);  
                  usage();
                  return(-1);
               } 
               if( !isalldigits(argv[i+1]) )
               {
                  printf("\n smrptx: the paramter following -t is not all digits.\n"); 
                  usage();
		  return(-1);
               }
               top = atoi(argv[i+1]);
               i++;
               break;
            default:
               printf("\n smrptx: invalid input parameter %s \n",argv[i][1]);   
               break; 
         }
      }
   }
   /*======================================================================
    * open the files and check for errors: 
    *	input: - logfile 
    *   ouput: - vrfy, expn, warn, error, sec, seq 
    *=====================================================================*/  
   if (!(lfp = fopen(logfile, "r")))
   {  
      
      fprintf(stderr,"smrptx: open failed for logfile  %s.\n       %s. \n",
			logfile, strerror(errno));
      return(-1);    			
   }
 
   if (!(vfp = fopen(vrfyfile, "w")))
   {  
      
      fprintf(stderr,"smrptx: open failed for vrfy file  %s.\n       %s. \n",
			vrfyfile, strerror(errno));
      return(-1);    			
   } 

    
   if (!(exfp = fopen(expnfile, "w")))
   {  
      
      fprintf(stderr,"smrptx: open failed for expn file  %s.\n       %s. \n",
			expnfile, strerror(errno));
      return(-1);    			
   }

    
   if (!(wfp = fopen(warnfile, "w")))
   {  
      
      fprintf(stderr,"smrptx: open failed for warn file  %s.\n       %s. \n",
			warnfile, strerror(errno));
      return(-1);    			
   }
   
    
   if (!(erfp = fopen(errfile, "w")))
   {  
      
      fprintf(stderr,"smrptx: open failed for error file  %s.\n       %s. \n",
			errfile, strerror(errno));
      return(-1);    			
   }
    
   if (!(alfp = fopen(alertfile, "w")))
   {  
      
      fprintf(stderr,"smrptx: open failed for alert file  %s.\n       %s. \n",
			alertfile, strerror(errno));
      return(-1);    			
   }
   
    
   if (!(sfp = fopen(seqfile, "w")))
   {  
      
      fprintf(stderr,"smrptx: open failed for sequester file  %s.\n       %s. \n",
			seqfile, strerror(errno));
      return(-1);    			
   }
   cp = (char *) &record;
   for( i = 0; i <17; i++)
   {
      firstdate[i] = '\0';
      lastdate[i]  = '\0';             
   } 
   
   /*========================================================== 
    * Read the log file and look for smwrap or smwrapd records 
    *=========================================================*/        
   while( fgets(record, MAXLEN, lfp) !=NULL )
   {   
      total_recs++;   
      /* skip over comments, null lines and short records */
      if( (*cp == '#') || (*cp == '\n') || (strlen(record) < 23) ) 
      { 
         trunc_recs++;
         goto getnextline; 
      } 
      if(total_recs == 1)
      {
         memcpy(firstdate,record,16);
      }
      memcpy(lastdate,record,16);
      strcpy(orecord,record);
      /* skip over date & time */ 
      if( (token1 = strtok(record+16,"\t \n")) == NULL) 
         goto getnextline; 
      /* get name of program that wrote the message	*/
      if( (token1 = strtok(NULL,tokensep)) == NULL)
         goto getnextline;
      /*=============================== 
       * process smwrap record 
       *==============================*/
      if( strcmp(token1,"smwrap") == 0)
      {
         total_smwraprec++;   
	 token1 = strtok(NULL,":"); 
         if( (token1 = strtok(NULL," ")) == NULL)
         {
	    trunc_recs++;
            goto getnextline;
         } 
         if( strcmp(token1,"Error:") == 0)
         {
            total_errors++; 
            fputs(orecord, erfp); 
	 } 
         if( strcmp(token1,"Security") == 0)
         {
            token1 = strtok(NULL," "); 		/* skip over "Alert"	*/ 
            token1 = strtok(NULL,"\' ");    
            total_secalert++; 
            fputs(orecord, alfp); 
         } 
         if( token1 == NULL || *token1 == '\012')
         {
	    trunc_recs++;
            goto getnextline;
         } 
         if( strcasecmp(token1,"VRFY") == 0 ) 
         {
            total_vrfy++; 
            fputs(orecord, vfp); 
	 }
	 else
         if( strcasecmp(token1,"EXPN") == 0 )
         {
            total_expn++;
            fputs(orecord, exfp); 
	 }
         	 else
         if( strcasecmp(token1,"Warning") == 0)
         {
            total_warn++;
            fputs(orecord, wfp); 
	 }
         goto getnextline;
      }
      /*======================================
       * process smwrapd record 
       *=====================================*/
      if( strcmp(token1,"smwrapd") == 0)
      {
         total_smwrapdrec++;   
	 token1 = strtok(NULL,":");    		/* skip  [pid]:		*/
         if( (token1 = strtok(NULL,"\t,[] \n")) == NULL)  
         {
	    trunc_recs++;
            goto getnextline;
         }       
         if( strstr(orecord,"delivered.") != NULL)
         {
            total_delv++;   
            goto getnextline;   
	 }     
         if( strstr(orecord,"sequestered.") != NULL)
         {
            total_seq++; 
	    fputs(orecord, sfp); 
	 }
         if( strcmp(token1,"Error:") == 0)
         {
            total_errors++; 
            fputs(orecord, erfp); 
	 } 
          else
         if( strcmp(token1,"Security") == 0)
         {
            total_secalert++; 
            fputs(orecord, alfp); 
	 }

         else
         if( strcmp(token1,"Warning") == 0)
         {
            total_warn++; 
	    fputs(orecord, wfp); 
	 }
         goto getnextline;   
	
      }

getnextline:;   
   } 
   fclose(lfp);
   fclose(vfp);
   fclose(erfp);
   fclose(exfp);
   fclose(wfp);
   fclose(alfp);
   fclose(sfp);

   /* print summary reports */ 
   printf("\n Sendmail Wrapper Exception Report \n");
   printf("\n log file name: %s\n",logfile);
   printf("\n total records in log file      = %d",total_recs); 
   printf("\n number of truncated records    = %d",trunc_recs);
   printf("\n date of first record           = %s",firstdate);
   printf("\n date of last record            = %s",lastdate);
   printf("\n number of smwrap records       = %d",total_smwraprec); 
   printf("\n number of smwrapd records      = %d",total_smwrapdrec);
   printf("\n total messages delivered       = %d",total_delv);
   printf("\n number of VRFY attempts        = %d",total_vrfy);
   printf("\n number of Error messages       = %d",total_errors);
   printf("\n number of EXPN attempts        = %d",total_expn);
   printf("\n number of Security Alerts      = %d",total_secalert);
   printf("\n number of Sequestered files    = %d",total_seq);
   printf("\n number of Warning messages     = %d",total_warn); 	
   printf("\n\n"); 

   if(!detail) return(0);
   /*=====================================================================
    * print detail reports
    *------------------------------------------------------------------                  
    * open the files and check for errors: 
    *   input: - vrfy, expn, warn, error, sec, seq 
    *=====================================================================*/  

 
   if (!(vfp = fopen(vrfyfile, "r")))
   {  
      
      fprintf(stderr,"smrptx: open failed for vrfy file  %s.\n       %s. \n",
			vrfyfile, strerror(errno));
      return(-1);    			
   }  
   printf("\n\n\n");
   printf("\n Listing of all VRFY attempts to port 25\n\n");
   while( fgets(record, MAXLEN, vfp) != NULL )
   { 
      printf("%s",record);
   } 
   fclose(vfp); 
   record[0] = '\0';
  
   printf("\n\n\n");
   printf("\n Listing of all EXPN attempts to port 25\n\n");
   if (!(exfp = fopen(expnfile, "r")))
   {   
      fprintf(stderr,"smrptx: open failed for expn file  %s.\n       %s. \n",
			expnfile, strerror(errno));
      return(-1);    			
   }   

   while( fgets(record, MAXLEN, exfp) != NULL )
   { 
      printf("%s",record);
   } 
   fclose(exfp);  
   record[0] = '\0';

   if (!(wfp = fopen(warnfile, "r")))
   {  
      
      fprintf(stderr,"smrptx: open failed for warn file  %s.\n       %s. \n",
			warnfile, strerror(errno));
      return(-1);    			
   }
   printf("\n\n\n");
   printf("\n Listing of all Warning messages from smwrap & smwrapd\n\n");
   while( fgets(record, MAXLEN, wfp) != NULL )
   { 
      printf("%s",record);
   } 
   fclose(wfp); 
   record[0] = '\0';

   if (!(erfp = fopen(errfile, "r")))
   {   
      fprintf(stderr,"smrptx: open failed for error file  %s.\n       %s. \n",
			errfile, strerror(errno));
      return(-1);    			
   }
   printf("\n\n\n");
   printf("\n Listing of all Error messages from smwrap & smwrapd\n\n");
   while(  fgets(record, MAXLEN, erfp) != NULL )
   { 
      printf("%s",record);
   } 
   fclose(erfp); 
   record[0] = '\0';
 
   if (!(alfp = fopen(alertfile, "r")))
   {  
      
      fprintf(stderr,"smrptx: open failed for alert file  %s.\n       %s. \n",
			alertfile, strerror(errno));
      return(-1);    			
   }
   printf("\n\n\n");
   printf("\n Listing of all Security Alert messages from smwrap & smwrapd\n\n");
   while( fgets(record, MAXLEN, alfp) != NULL )
   { 
      printf("%s",record);
   } 
   fclose(alfp);   
   record[0] = '\0';
    
   if (!(sfp = fopen(seqfile, "r")))
   {  
      fprintf(stderr,"smrptx: open failed for sequester file  %s.\n       %s. \n",
			seqfile, strerror(errno));
      return(-1);    			
   }
   printf("\n\n\n");
   printf("\n Listing of all Sequestered file messages \n\n");
   while( fgets(record, MAXLEN, sfp) != NULL )
   { 
      printf("%s",record);
   } 
   fclose(sfp); 




   return(0);
} 


/*============================================================================ 
 * print usage message
 *===========================================================================*/
void usage(void)
{
   
   printf("\n usage: smrpt [-d] [-l logfile ] [-t nnn ]  \n");
   printf("\n        where:\n");
   printf("        -d           for a detail report on all exceptions\n");  
   printf("        -l logfile   full path name of the log file. \n");
   printf("        -t nnn       list top nnn users\n");
   
   return;
}  
/*==================================================================
 * function: 	isalldigits
 * input:	a pointer to a character string 
 * output:	returns 1 if all the characters in the string are digits 
 *==================================================================*/  
int isalldigits(char *s)  
{
   while(isdigit(*s))
      s++;
   return(*s == '\0');
} 
