/*============================================================================
 * filename:		util/reports/sockdsum/sockdsum.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=============================================================================
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  All rights
 * reserved.                   
 *
 * written by:		R. James Livermore
 * Date written:	April 	  12, 1995
 * Last modified:	February   8, 2000 - added LPL
 *
 * function:		sock daemon summary report program 	
 *       
 * 	The smrpt program reads the system log and produces a report
 *	on sockd usage.
 *
 *  	1. read the file until eof
 *	2. capture all sockd messages.
 *
 *============================================================================*/

static  char	ident[] = "@(#) sockdsum, V 1.0, 00/02/08"
 			  " (C) Freemont Avenue Software, Inc."
			  " 1995-2000";
#include	<search.h>		/* for qsort	*/
#include	<sys/types.h>
#include	<sys/file.h>
#include	<sys/stat.h>
#include	<stdio.h> 
#include	<string.h>		/* added for C++ */
#include	<time.h>
#include	<ctype.h>
#include	<syslog.h>
#include	<sys/socket.h>
#include	<netinet/in.h>
#include	<netdb.h>
 
#include	<unistd.h>  		/* added for C++ */
#include	<stdlib.h>		/* added for C++ */  
   
struct SOCKDHOST 
{
   float	connects;
   float	receives;
   float	bytess;
   float  	bytesr;
   char		host[84];  
}; 
/*==========================
 * external variables
 *==========================*/  
extern   int	h_errno;
extern   int	errno;
#define		MAXHOSTS 	15000  
int	maxhosts;
/*======================================================================
 * function prototypes
 *======================================================================*/ 
int comprhosts(char **s1, char **s2);  /* compare host names		*/ 
int comprbr(char **s1, char **s2);   /* compare bytes received		*/
int comprbs(char **s1, char **s2);   /* compare bytes sent		*/
int comprcon(char **s1, char **s2);  /* compare connects 		*/
int comprrecv(char **s1, char **s2);  /* compare receives 		*/
int isalldigits(char *s); 
int prthosts(struct SOCKDHOST *p, int top); 
int procsockd(char *from, char *to, int fbytes, int tbytes, struct SOCKDHOST *p,
int cntflg);
void usage(void);  
 
/*============================================================================                        
 * main
 * inputs: optional input arguments
 *         -h 		for help
 *	   -l logfile	specify logfile name		
 *         -t nnn	specify printing of top nnn hosts only
 * output:
 *	return  0 if ok
 *	return -1 on error
 *===========================================================================*/
int main(int argc, char *argv[])
{   
   /* log file statistics		*/ 
   int		trunc_recs	= 0;   
   int		trunc_recnum	= 0;
   int		total_recs	= 0;
   int		total_connects 	= 0;
   float	total_bytess 	= 0; 
   float	total_bytesr    = 0;
   int		total_sockdrec  = 0;  
   char		firstdate[17];
   char		lastdate[17]; 
   /* 					*/ 
   struct	SOCKDHOST	*p;     /* ptr to data area		*/
   struct	SOCKDHOST	*p2;
   int		i;
   int		fbytes; 
   int		tbytes; 
   int		numhosts	= 0;     
   int		numwarnings	= 0;
   int		top		= 0;
#ifdef _AIX
   static char 	logfile[256] = "/var/adm/syslog";
#elif __hppa
   static char  logfile[256] = "/var/adm/syslog/syslog.log";
#elif linux
   static char  logfile[256] = "/var/log/messages";
#else
   static char  logfile[256] = "/var/adm/messages";
#endif   
   char		record[BUFSIZ];
   FILE 	*lfp;			/* log file ptr	*/ 
   char 	*cp;
   char		*cp2;
   char		*token1;
   char		*token2;
   char		*from;
   char		*to;
   char		*fromip;
   char		*toip;
   char		*logtype;
   int		mhchunk;
   int		no_rev_lookup = 0;
   struct hostent *hp;
   struct in_addr addr;
   static char  tokensep[] 	= "\t,:;= \n";
   maxhosts = MAXHOSTS;
   /*===================================================
    * check for input arguments
    *==================================================*/
   if( argc > 1)
   {
      if( argv[1][0] != '-')   
      {
         printf("\n sockdsum:  invalid leading character '%c' \n",argv[1][0]);
         usage();
         return(-1);        
      }
      /* check to see if user is asking for help */
      if ((strcmp(argv[1], "-h")    == 0) ||
          (strcmp(argv[1], "-help") == 0) ||
          (strcmp(argv[1], "help")  == 0) ||
          (strcmp(argv[1], "?")     == 0)
         ) 
      {
         usage();
         return(0);
      } 
      for(i = 1; i < argc ; i++)
      {
         switch( argv[i][1])
         { 
            case 'l':  			/* specify logfile	*/
               if( strlen(argv[i+1]) > 255)
               {
                  printf("\n sockdsum: file name too long. length = %d\n",
			strlen(argv[i+1])); 
                  usage();
                  return(-1);
               }
	       strcpy(logfile,argv[i+1]);
               i++;
               break;
            case 'm':  			/* specify maximum number of hosts*/
               if( !isalldigits(argv[i+1]) )
               {
                  printf("\n sockdsum: the parameter following -m is not all digits.\n");
                  usage();
                  return(-1);
               }
	       maxhosts = atoi(argv[i+1]);
               if(maxhosts < 100)
               {
                  printf("\nsockdsum: maximum number of hosts %d is < 100",
			 maxhosts);
                  return(-1);
               }
               i++;
               break;
            case 't':
               if( !isalldigits(argv[i+1]) )
               {
                  printf("\n sockdsum: the paramter following -t is not all digits.\n"); 
                  usage();
		  return(-1);
               }
               top = atoi(argv[i+1]);
               i++;
               break;
            case 'R':
	       no_rev_lookup = 1;
	       break;
            default:
               printf("\n sockdsum: invalid input parameter %s \n",argv[i][1]);   
               break; 
         }
      }
   }
   mhchunk = maxhosts;
   /* open log file	*/  
   if (!(lfp = fopen(logfile, "r")))
   {  
      
      fprintf(stderr,"sockdsum: open failed for logfile  %s.\n       %s. \n",
			logfile, strerror(errno));
      return(-1);    			
   }
   cp = (char *) &record;
   memset(firstdate, 0, sizeof(firstdate));
   memset(lastdate, 0, sizeof(lastdate));
   /*=======================================================================
    * acquire storage for maximum number of users and intialize every record
    *======================================================================*/
   p = (struct SOCKDHOST *) calloc(maxhosts + 1, sizeof(struct SOCKDHOST));
   if (p == NULL) {
      fprintf(stderr,"sockdsum: unable to malloc.  %s\n", strerror(errno));
      return(-1);
   }

   /*======================================================== 
    * Read the log file and look for sockd record 
    *=======================================================*/        
   while( fgets(record, sizeof(record), lfp) != NULL)
   {     
      total_recs++;   
      /* skip over comments, null lines and short records */
      if( (record[0] == '#') || (record[0] == '\n') || (strlen(record) < 23) ) { 
         trunc_recs++;
         continue;
      } 
      if(total_recs == 1)
         memcpy(firstdate,record,16);

      /*===================================================================
       * skip over date & time and
       * get name of program that wrote the message
       *==================================================================*/
      if( (token1 = strtok(record+16,tokensep)) == NULL)
         continue;
      if (strchr(token1, '[') == NULL) {
         if ((token1 = strtok(NULL,tokensep)) == NULL)
            continue;
      }
      /*===================================================================
       * The next token can be the hostname (if logfile is the syslog)
       * or the name of the program (sockd[pid]).
       *==================================================================*/
      if( strncmp(token1,"sockd[",6) != 0)
         continue;
      total_sockdrec++;   
      token2 = strtok(NULL," :("); 
      if (token2 == NULL)
         continue;
      /* process connected record 	*/
      if(strcmp(token2,"pass") != 0)
         continue;
      token2 = strtok(NULL,")");	/* get "(nn)"	 	*/
      logtype = strtok(NULL," :");	/* get type "tcp:" 	*/
      fromip  = strtok(NULL," ");	/* get from host	*/ 
      token2 = strtok(NULL," ");	/* skip over "->"	*/ 
      toip   = strtok(NULL," \n");	/* get to host	*/
      cp   = strtok(NULL," \n");	/* byte count	*/
      if (logtype == NULL || fromip == NULL || toip == NULL)
         continue;
      if (cp && strchr(cp,'(') != NULL && (token2 = strtok(cp,"() ")) != NULL) {
         fbytes = atoi(token2);
         tbytes = 0; 
      }
      else {
         if (strcmp(logtype, "connect") != 0)
            continue;
         fbytes = 0; 
         tbytes = 0; 
      }
      /*===================================================
       * Conut the number of connects
       *==================================================*/  
      if (strcmp(logtype, "connect") == 0)
         total_connects++;  
      if ((cp = strrchr(fromip, '.')) != NULL)
         *cp = '\0';
      if ((cp = strrchr(toip, '.')) != NULL)
         *cp = '\0';
      /*===================================================
       * Convert from an IP address to host name
       *==================================================*/  
      if (!no_rev_lookup && strspn(fromip, "0123456789.") == strlen(fromip)) {
#ifdef __sun
         if ((addr.s_addr = inet_addr(fromip)) == (in_addr_t)-1) {
#else
         if (inet_aton(fromip, &addr) == 0) {
#endif
            fprintf(stderr, "sockdsum: error converting \"%s\" to an IP "
			    		"address.", toip);
            return(-1);
         }
         hp = gethostbyaddr((const char *)&addr.s_addr,
			      		sizeof(addr.s_addr), AF_INET);
         if (hp == NULL) {
            fprintf(stderr, "sockdsum: error looking up \"%s\".  h_errno = %d, "
			 "errno = %d\n",
			 fromip, h_errno, errno);
            cp = fromip;
         }
         else if (hp->h_name == NULL)
            cp = fromip;
         else
            cp = hp->h_name;
      }
      else
         cp = fromip;
      from = malloc(strlen(cp) + 1);
      if (from == NULL) {
         fprintf(stderr, "sockdsum: unable to malloc.  %s.\n", strerror(errno));
         return(-1);
      }
      strcpy(from, cp);

      if (!no_rev_lookup && strspn(fromip, "0123456789.") == strlen(fromip)) {
#ifdef __sun
         if ((addr.s_addr = inet_addr(toip)) == (in_addr_t)-1) {
#else
         if (inet_aton(toip, &addr) == 0) {
#endif
            fprintf(stderr, "sockdsum: error converting \"%s\" to an IP "
			    		"address.", toip);
            return(-1);
         }
         hp = gethostbyaddr((const char *)&addr.s_addr,
		      			sizeof(addr.s_addr), AF_INET);
         if (hp == NULL) {
            fprintf(stderr, "sockdsum: error looking up \"%s\".  h_errno = %d, "
			 "errno = %d\n",
			 toip, h_errno, errno);
            cp = toip;
         }
         else if (hp->h_name == NULL)
            cp = toip;
         else
            cp = hp->h_name;
      }
      else
         cp = toip;
      to = malloc(strlen(cp) + 1);
      if (to == NULL) {
         fprintf(stderr, "sockdsum: unable to malloc.  %s.\n", strerror(errno));
         return(-1);
      }
      strcpy(to, cp);

      /*===================================================
       * Convert from and to addresses to lower case
       *==================================================*/  
      cp2 = from;
      for( i = 0; i < strlen(from); i++) {
         *cp2 = tolower((int)*cp2);
         cp2++; 
      }
      cp2 = to;
      for( i = 0; i < strlen(to); i++) {
         *cp2 = tolower((int)*cp2);
         cp2++; 
      }
      numhosts = numhosts + procsockd(from, to, fbytes, fbytes , p,
			 strcmp(logtype, "connect") == 0);
      free(from);
      free(to);
      if( numhosts > maxhosts) {
         maxhosts += mhchunk;
         p = (struct SOCKDHOST *)realloc(p, maxhosts*sizeof(struct SOCKDHOST));
         if (p == NULL) {
            fprintf(stderr,"sockdsum: unable to realloc.  %s\n",
			       strerror(errno));
            return(-1);
         }
      }
   } 
   memcpy(lastdate,record,16);
   
   /* print summary reports */ 
   printf("\n SOCKS Daemon Summary Report \n"); 
   printf("\n sockdsum Version 1.0");
   printf("\n log file name: %s\n",logfile); 
   printf("\n date of first record           = %s",firstdate);
   printf("\n date of last record            = %s",lastdate);
   printf("\n total records in log file      =   %10d",total_recs); 
   printf("\n number of truncated records    =   %10d",trunc_recs);
   printf("\n total number of sockd records  =   %10d",total_sockdrec); 
   printf("\n number of unique hosts         =   %10d",numhosts);
   printf("\n number of connect messages     =   %10d",total_connects);
/*
   printf("\n total bytes sent               = %12.0f",total_bytess);
   printf("\n total bytes received           = %12.0f",total_bytesr);
   printf("\n total bytes sent & received    = %12.0f",total_bytess+total_bytesr);		
*/
   printf("\n\n"); 
   /*====================================================================
    * sort buy user E-mail address
    * sort by bytes sent and print 
    * sort by messages sent and print
    * sort by bytes received and print 
    * sort by messages received and print
    *====================================================================*/
   if( numhosts == 0)
   {
      printf("\n sockdsum: No sockd records were found.\n");
      goto freemem;
   }
   qsort(p, numhosts, sizeof(struct SOCKDHOST),
		(int(*) (const void *, const void *))comprhosts);
   printf("\n\n Top hosts by bytes sent\n"); 
   qsort(p, numhosts, sizeof(struct SOCKDHOST), 
		(int(*) (const void *, const void *))comprbs);
   prthosts(p,top);

   printf("\n\n Top hosts by connect requests\n"); 
   qsort(p, numhosts, sizeof(struct SOCKDHOST), 
		(int(*) (const void *, const void *))comprcon);
   prthosts(p,top);  

   printf("\n\n Top hosts by bytes received\n"); 
   qsort(p, numhosts, sizeof(struct SOCKDHOST), 
		(int(*) (const void *, const void *))comprbr);
   prthosts(p,top);
  
   printf("\n\n Top hosts by receives\n"); 
   qsort(p, numhosts, sizeof(struct SOCKDHOST), 
		(int(*) (const void *, const void *))comprrecv);
   prthosts(p,top);

freemem:
   free((void *) p);	/* free data array	*/
   return(0);
} 
/*========================================================================
 *
 *========================================================================*/
int prthosts(struct SOCKDHOST *p, int top)
{                            
   struct SOCKDHOST *p2; 
   int	i; 
   int	j;
   int		numhosts = 0;
   float	bs 	 = 0;
   float	conns	 = 0;
   float	br	 = 0;
   float	recvs	 = 0;
   j = top;   
   if( (top < 1) || (top > maxhosts) ) j = maxhosts;
   p2 = p;
   printf("\n                 initiated     bytes       received  ");
   printf("\nbytes sent       connects      received    connects   host\n"); 
   for (i = 0; i < j; i++)
   {
      printf("\n %12.0f  %10.0f  %12.0f  %10.0f  %s",
               p2->bytess, p2->connects, p2->bytesr, p2->receives, p2->host); 
      bs    = bs    + p2->bytess;
      conns = conns + p2->connects;
      br    = br    + p2->bytesr;
      recvs = recvs + p2->receives;
      p2++;
      if( strlen(p2->host) == 0) break;
   }
   if(top) i--; 
   if (top != 1)
   {
      printf("\n\n\ntotals:");
      printf("\n %12.0f  %10.0f  %12.0f  %10.0f  %10d\n",
               bs, conns, br, recvs, i+1);  
   }
   else printf("\n");
   return(0);
} 
/*===========================================================================
 * compare host names
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int comprhosts(char **s1, char **s2)
{
   struct SOCKDHOST *p1;
   struct SOCKDHOST *p2;
   p1 = (struct SOCKDHOST *) s1;
   p2 = (struct SOCKDHOST *) s2;
   return( strcmp(p1->host,p2->host) );
} 
/*===========================================================================
 * compare bytes sent
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int comprbs(char **s1, char **s2)
{
   struct SOCKDHOST *p1;
   struct SOCKDHOST *p2;
   p1 = (struct SOCKDHOST *) s1;
   p2 = (struct SOCKDHOST *) s2;
   if(p1->bytess >= p2->bytess) return(-1); 
   return(1);
} 
/*===========================================================================
 * compare bytes received
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int comprbr(char **s1, char **s2)
{
   struct SOCKDHOST *p1;
   struct SOCKDHOST *p2;
   p1 = (struct SOCKDHOST *) s1;
   p2 = (struct SOCKDHOST *) s2;
   if(p1->bytesr >= p2->bytesr) return(-1); 
   return(1);
} 
/*===========================================================================
 * compare connects sent
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int comprcon(char **s1, char **s2)
{
   struct SOCKDHOST *p1;
   struct SOCKDHOST *p2;
   p1 = (struct SOCKDHOST *) s1;
   p2 = (struct SOCKDHOST *) s2;
   if(p1->connects >= p2->connects) return(-1); 
   return(1);
} 
/*===========================================================================
 * compare messages received
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int comprrecv(char **s1, char **s2)
{
   struct SOCKDHOST *p1;
   struct SOCKDHOST *p2;
   p1 = (struct SOCKDHOST *) s1;
   p2 = (struct SOCKDHOST *) s2;
   if(p1->receives >= p2->receives) return(-1); 
   return(1);
} 
/*============================================================================
 * process sockd records
 * inputs:
 *	from host pointer
 *	to   host pointer  
 *      number of bytes from host
 *	number of bytes to host
 *	pointer to first record
 *
 * outputs:  
 *	returns number of records added
 *
 *===========================================================================*/
int procsockd(char *from, char *to, int fbytes, int tbytes, struct SOCKDHOST *p,
int cntflg)
{  
   struct SOCKDHOST *p2;
   int	i;
   int	newhost; 
   int  tofound;
   int	fromfound; 
   p2 = p;
   tofound 	= 0;
   fromfound	= 0; 
   newhost 	= 0;
   for( i = 0; i < maxhosts; i++) 
   {
      if( strcmp(p2->host,from) == 0)
      {
         p2->bytess = p2->bytess + fbytes; 
         if(cntflg) p2->connects++; 
         fromfound++;
      }
      if( strcmp(p2->host,to)  == 0)
            {
         p2->bytesr = p2->bytesr + tbytes; 
         if(cntflg) p2->receives++; 
         tofound++;
      }
      if( strlen(p2->host) == 0) break;
      if( fromfound && tofound) return(0);
      p2++;
   }
   if(!fromfound)
   {
      p2->bytess  = fbytes;
      if (cntflg)
         p2->connects = 1;
      else
         p2->connects = 0;
      strcpy(p2->host,from); 
      newhost++;
   } 
   if(!tofound)
   {  
      if(!fromfound) p2++;
      p2->bytesr = tbytes;
      if (cntflg)
         p2->receives = 1;
      else
         p2->receives = 0;
      strcpy(p2->host,to);
      newhost++; 
   } 
   return(newhost);
} 
/*============================================================================ 
 * print usage message
 *===========================================================================*/
void usage(void)
{
   
   printf("\n usage: sockdsum [-l logfile ] [-m maxhosts] [-t nnn ]  [-R]\n");
   printf("\n        where:\n");  
   printf("        -l logfile   full path name of the log file. \n");
   printf("        -m maxhosts  maximum number of hosts in report. \n");
   printf("        -t nnn       list top nnn users\n");
   printf("        -R           Turns off reverse DNS lookup\n");
   
   return;
}  
/*==================================================================
 * function: 	isalldigits
 * input:	a pointer to a character string 
 * output:	returns 1 if all the characters in the string are digits 
 *==================================================================*/  
int isalldigits(char *s)  
{
   while(isdigit(*s))
      s++;
   return(*s == '\0');
} 
