/*============================================================================
 * filename:		util/reports/tnrprpt/tnprpt.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=============================================================================
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  All rights
 * reserved.                                    
 *
 * written by:		R. James Livermore
 * Date written:	March     31, 1996
 * Last modified:	February   8, 2000 - added LPL
 *
 *			  
 *
 * function:		tnproxy report program  	
 *       
 * 	The tnprpt program reads the system log and produces a report
 *	on tnproxy usage.
 *
 *  	1. read the file until eof
 *	2. capture all tnproxy messages.
 *
 *============================================================================*/

static  char	ident[] = "@(#) tnprpt, V 1.0, 00/01/21"
 			  " (C) Freemont Avenue Software, Inc."
			  " 1995-2000";
#include	<search.h>		/* for qsort	*/
#include	<sys/types.h>
#include	<sys/file.h>
#include	<sys/stat.h>
#include	<stdio.h> 
#include	<string.h>		/* added for C++ */
#include	<time.h>
#include	<ctype.h>
#include	<syslog.h>
#include	<netdb.h>
 
#include	<unistd.h>  		/* added for C++ */
#include	<stdlib.h>		/* added for C++ */  
   
struct TNPUSER {
   int  	cmds;
   float	bytesin;
   float	bytesout;
   int		connect; 
   float 	cpu;
   char		user[84];  
}; 
/*==========================
 * external variables
 *==========================*/  
extern   int	errno;
#ifdef __hppa
#define		MAXUSERS 	10000
#else
#define		MAXUSERS 	10000
#endif
   int	maxusers;
/*======================================================================
 * function prototypes
 *======================================================================*/  
void bldtnpuser(struct TNPUSER *rptr);
int compradr(char **s1, char **s2);  /* compare address			*/ 
int comprbr(char **s1, char **s2);   /* compare bytes received		*/
int comprbs(char **s1, char **s2);   /* compare bytes sent		*/
int comprelt(char **s1, char **s2);   /* compare messages sent		*/
int comprcpu(char **s1, char **s2);   /* compare messages received	*/
int isalldigits(char *s); 
int prtusers(struct TNPUSER *p, int top); 
int proctnproxy(struct TNPUSER *rptr, struct TNPUSER *p);
void usage(void);  
 
/*============================================================================                        
 * main
 * inputs: optional input arguments
 *         -h 		for help
 *	   -l logfile	specify logfile name		
 *         -t nnn	specify printing of top nnn users only
 * output:
 *	return  0 if ok
 *	return -1 on error
 *===========================================================================*/
int main(int argc, char *argv[])
{   
   /* log file statistics		*/   
   int		bytesin		= 0;
   int		bytesout	= 0;   
   struct TNPUSER rec;
   struct TNPUSER *rptr; 
   char		srchost[64];
   int		trunc_recs	= 0;   
   int		trunc_recnum	= 0; 
   int		total_exits	= 0;
   int		total_recs	= 0;
   int		total_msg 	= 0;
   float	total_bytes 	= 0.;
   float	total_bytesin	= 0.; 
   float	total_bytesout  = 0.;
   float	total_cpu	= 0.;
   int		total_tnproxyrec = 0;
   float	total_duration	= 0;  
   char		firstdate[17];
   char		lastdate[17]; 
   /* 					*/ 
   struct	TNPUSER	*p;             /* ptr to data area		*/
   struct	TNPUSER	*p2;
   int		i;
   int		numaddr		= 0;
   int		top		= 0;
#define	MAXLEN 1024
#ifdef _AIX
   static char 	logfile[256] = "/var/adm/syslog"; 
#elif __hppa
   static char 	logfile[256] = "/var/adm/syslog/syslog.log";
#elif linux
   static char  logfile[256] = "/var/log/messages";
#else   
   static char 	logfile[256] = "/var/adm/messages";
#endif  
   char		buffer[MAXLEN];
   char		record[MAXLEN];
   FILE 	*lfp;			/* log file ptr	*/ 
   char 	*cp;  
   char		*token1;
   static char	tokensep[] 	= "\t,:;=[] \n";
   rptr = &rec;
   maxusers = MAXUSERS;
   /*===================================================
    * check for input arguments
    *==================================================*/
   if( argc > 1)
   {
      if( argv[1][0] != '-')   
      {
         printf("\n smrpt:  invalid leading character '%c' \n",argv[1][0]);
         usage();
         return(-1);        
      }
      /* check to see if user is asking for help */
      if ((strcmp(argv[1], "-h")    == 0) ||
          (strcmp(argv[1], "-help") == 0) ||
          (strcmp(argv[1], "help")  == 0) ||
          (strcmp(argv[1], "?")     == 0)
         ) 
      {
         usage();
         return(0);
      } 
      for(i = 1; i < argc ; i++)
      {
         switch( argv[i][1])
         { 
            case 'l': 
               if( strlen(argv[i+1]) > 255)
               {
                  printf("\n tnprpt: file name too long. length = %d\n",
			strlen(argv[i+1])); 
                  usage();
                  return(-1);
               }
	       strcpy(logfile,argv[i+1]);
               i++;
               break;
            case 't':
               if( !isalldigits(argv[i+1]) )
               {
                  printf("\n tnprpt: the paramter following -t is not all digits.\n"); 
                  usage();
		  return(-1);
               }
               top = atoi(argv[i+1]);
               i++;
               break;
            default:
               printf("\n tnprpt: invalid input parameter '%s' \n",argv[i]);
	       return(-1);               
         }
      }
   }
   /* open log file	*/  
   if (!(lfp = fopen(logfile, "r")))
   {  
      
      fprintf(stderr,"tnprpt: open failed for logfile  '%s'.\n       %s. \n",
			logfile, strerror(errno));
      return(-1);    			
   }
   cp = (char *) &buffer;
   for( i = 0; i <17; i++)
   {
      firstdate[i] = '\0';
      lastdate[i]  = '\0';             
   } 
   /*=======================================================================
    * acquire storage for maximum number of users and intialize every record
    *======================================================================*/
   p = (struct TNPUSER *) malloc(sizeof(struct TNPUSER)*(maxusers+1)); 
   bldtnpuser(p);

   p2 = p;
   for ( i=0; i < maxusers+1; i++)
   {
      memcpy(p2,p,sizeof(struct TNPUSER));
      p2++; 
   }  
   /*======================================================== 
    * Read the log file and look for tnproxy record 
    *=======================================================*/        
   while(!feof(lfp))
   {   
      cp = fgets(buffer, MAXLEN, lfp); 
      if( cp == NULL)
      {
         if(feof(lfp)) break; 
         if(ferror(lfp))
	 {
            fprintf(stderr,"tnprpt: error reading record %d in file  %s. %s\n       %s. \n",
			total_recs, logfile, strerror(errno));
            return(-1);   
         }
      }
      total_recs++; 
      bldtnpuser(rptr); 
      srchost[0] = '\0'; 
      /*========================================================
       * skip over comments, null lines and short records 
       *=======================================================*/
      if( (*cp == '#') || (*cp == '\n') || (strlen(buffer) < 23) ) 
      { 
         trunc_recs++;
         goto getnextline; 
      } 
      if(total_recs == 1)
      {
         memcpy(firstdate,buffer,16);
      }
      memcpy(lastdate,buffer,16);
      /* skip over date & time */ 
      if( (token1 = strtok(buffer+16,"\t \n")) == NULL) 
         goto getnextline; 
      /* get name of program that wrote the message	*/
      if( (token1 = strtok(NULL,tokensep)) == NULL)
         goto getnextline;
      /*===================================================================
       * Process tnproxy records
       *==================================================================*/
      if( strcmp(token1,"tnproxy") == 0)
      {
         total_tnproxyrec++;   
	 token1 = strtok(NULL,":"); 
         token1 = strtok(NULL," "); 

         /*==============================================
          * Process "exit user" line.
          *=============================================*/
         if( strcmp(token1,"exit") == 0)  
         {   
            token1 = strtok(NULL," "); 
            if( strcmp(token1,"source") != 0)
            {
               trunc_recs++;
               goto getnextline;     
            } 
            /*==================================
             * Get source host name
             *==================================*/
            if((token1 = strtok(NULL," =")) == NULL)
            {
               trunc_recs++;
               goto getnextline;
            }
            strcpy(srchost,token1); 
            /*==================================
             * Skip over dest = host.name
             *==================================*/ 
            if((token1 = strtok(NULL,"=")) == NULL)
            {
               trunc_recs++;
               goto getnextline;
            } 
            if((token1 = strtok(NULL," ")) == NULL)
            {
               trunc_recs++;
               goto getnextline;
            }                          

            /*==================================
             * Get number of input bytes
             *==================================*/
            token1 = strtok(NULL,tokensep); 	/* get bytes  "in"	*/   
            if( strcmp(token1,"bytesin") != 0)
            {   
	        trunc_recs++;
                goto getnextline;
            }
            if((token1 = strtok(NULL,tokensep)) == NULL) /* get bytesin	*/
            {   
	        trunc_recs++;
                goto getnextline;
            }
            sscanf(token1,"%f",&rptr->bytesin);      
            total_bytes = total_bytes + rptr->bytesin; 
            total_bytesin = total_bytesin + rptr->bytesin; 
            /*==================================
             * Get number of output bytes
             *==================================*/
            token1 = strtok(NULL,tokensep); 	/* get bytes "out" 	*/   
            if( strcmp(token1,"bytesout") != 0)
            {   
	        trunc_recs++;
                goto getnextline;
            } 
            if((token1 = strtok(NULL,tokensep)) == NULL) /* get bytesout*/
            {   
	        trunc_recs++;
                goto getnextline;
            }
            sscanf(token1,"%f",&rptr->bytesout);
            total_bytes = total_bytes + rptr->bytesout; 
            total_bytesout = total_bytesout + rptr->bytesout;  
            /*==================================
             * Get user name 
             *==================================*/ 
            if((token1 = strtok(NULL," ")) == NULL) 
            {   
	        trunc_recs++;
                goto getnextline;
            }  
            if( strcmp(token1,"user") != 0)
            {   
	        trunc_recs++;
                goto getnextline;
            } 
            if((token1 = strtok(NULL,"\'= ")) == NULL) 
            {   
	        trunc_recs++;
                goto getnextline;
            }  
            strcpy(rptr->user,token1);
            strcat(rptr->user,"@");   
            strcat(rptr->user,srchost);
            /*==================================
             * Get session duration 
             *==================================*/
            if((token1 = strtok(NULL,tokensep)) == NULL) /* get "duration" */
            {   
	        trunc_recs++;
                goto getnextline;
            }
            if( strcmp(token1,"duration") != 0)
            {   
	        trunc_recs++;
                goto getnextline;
            } 
            if((token1 = strtok(NULL,tokensep)) == NULL) /* get duration*/
            {   
	        trunc_recs++;
                goto getnextline;
            }
            sscanf(token1,"%d",&rptr->connect);
            total_duration = total_duration + rptr->connect;
            token1 = strtok(NULL,tokensep);	/* skip over "sec"	*/ 
            /*==================================
             * Get cpu time 
             *==================================*/
            if((token1 = strtok(NULL,tokensep)) == NULL) /* get "cpu" */
            {   
	        trunc_recs++;
                goto getnextline;
            }
            if( strcmp(token1,"cpu") != 0)
            {   
	        trunc_recs++;
                goto getnextline;
            } 
            if((token1 = strtok(NULL,tokensep)) == NULL) /* get cpu	*/
            {   
	        trunc_recs++;
                goto getnextline;
            }
            sscanf(token1,"%f",&rptr->cpu);
            total_cpu = total_cpu + rptr->cpu;
            total_exits++; 

         
         }  /* end of exiting record	*/ 

         numaddr = numaddr + proctnproxy(rptr, p);
	 total_msg++;
      } /* end of tnproxy record processing	*/

getnextline:;   
   } 
   
   /* print summary reports */ 
   printf("\n tnproxy V 1.0 Summary Report \n");
   printf("\n log file name: %s\n",logfile);
   printf("\n total records in log file       = %d",total_recs); 
   printf("\n number of truncated records     = %d",trunc_recs);
   printf("\n date of first record            = %s",firstdate);
   printf("\n date of last record             = %s",lastdate);
   printf("\n total number of tnproxy records = %12d",total_tnproxyrec); 
   printf("\n number of unique addresses      = %12d",numaddr);
   printf("\n number of telnet sessions       = %12d",total_exits); 
   printf("\n total bytes sent                = %12.0f",total_bytesout);
   printf("\n total bytes received            = %12.0f",total_bytesin);
   printf("\n total bytes sent and received   = %12.0f",total_bytes);
   printf("\n total duration                  = %12.0f sec",total_duration);
   printf("\n total cpu time                  =    %12.2f",total_cpu); 
   printf("\n");
   if( total_cpu > 0)
      printf("\n Mega Bytes per cpu second        =    %12.2f",
               total_bytes/(total_cpu*1024*1024));  

   if( numaddr > 0 )
   {
      printf("\n Average connect time per user    =   %12.1f",
               total_duration/numaddr);	    
   } 
   if( total_exits > 0 )
   {
      printf("\n Average connect time per session =   %12.1f",
               total_duration/total_exits);	    
   } 
   printf("\n\n"); 
   /*====================================================================
    * sort buy user address
    * sort by bytes sent and print 
    * sort by messages sent and print
    * sort by bytes received and print 
    * sort by messages received and print
    *====================================================================*/

   qsort(p, numaddr, sizeof(struct TNPUSER),
		(int(*) (const void *, const void *))compradr);
   printf("\n\n Top users by bytes sent\n"); 
   qsort(p, numaddr, sizeof(struct TNPUSER), 
		(int(*) (const void *, const void *))comprbs);
   prtusers(p,top);

   printf("\n\n Top users by bytes received\n"); 
   qsort(p, numaddr, sizeof(struct TNPUSER), 
		(int(*) (const void *, const void *))comprbr);
   prtusers(p,top);  

   printf("\n\n Top users by connect time\n"); 
   qsort(p, numaddr, sizeof(struct TNPUSER), 
		(int(*) (const void *, const void *))comprelt);
   prtusers(p,top);
  
   printf("\n\n Top users by cpu time\n"); 
   qsort(p, numaddr, sizeof(struct TNPUSER), 
		(int(*) (const void *, const void *))comprcpu);
   prtusers(p,top);

   free((void *) p);	/* free data array	*/
   return(0);
} 
/*========================================================================
 *
 *========================================================================*/
int prtusers(struct TNPUSER *p, int top)
{                            
   struct TNPUSER *p2; 
   float	tconnect = 0.0; 
   float	tcpu	 = 0;
   int		i; 
   int		j;
   int		numusers = 0;
   float	bs	= 0;
   float	br	= 0;
   int		msgr	= 0;
   j = top;   
   if( (top < 1) || (top > maxusers) ) j = maxusers;
   p2 = p;
   printf("\nbytes             bytes         connect      cpu  ");
   printf("\nsent              received      time         time   address\n"); 
   for (i =0; i < j; i++)
   {
      printf("\n %12.0f  %12.0f   %9d %10.2f  %s",
               	p2->bytesout, p2->bytesin,  
		p2->connect, p2->cpu, p2->user); 
      bs   = bs   + p2->bytesout;
      br   = br   + p2->bytesin;
      
      tconnect = tconnect + p2->connect;
      tcpu = tcpu + p2->cpu;
      p2++;
      if( strlen(p2->user) == 0) break;
   }
   if(top) i--; 
   if (top != 1)
   {
      printf("\n\n\ntotals:");
      printf("\n %12.0f  %12.0f   %9.0f %10.2f  %9d\n",
               bs, br, tconnect, tcpu, i+1);  
   }
   else printf("\n");
   return(0);
} 
/*===========================================================================
 * compare address
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int compradr(char **s1, char **s2)
{
   struct TNPUSER *p1;
   struct TNPUSER *p2;
   p1 = (struct TNPUSER *) s1;
   p2 = (struct TNPUSER *) s2;
   return( strcmp(p1->user,p2->user) );
} 
/*===========================================================================
 * compare bytes sent
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int comprbs(char **s1, char **s2)
{
   struct TNPUSER *p1;
   struct TNPUSER *p2;
   p1 = (struct TNPUSER *) s1;
   p2 = (struct TNPUSER *) s2;
   if(p1->bytesout >= p2->bytesout) return(-1); 
   return(1);
} 
/*===========================================================================
 * compare bytes received
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int comprbr(char **s1, char **s2)
{
   struct TNPUSER *p1;
   struct TNPUSER *p2;
   p1 = (struct TNPUSER *) s1;
   p2 = (struct TNPUSER *) s2;
   if(p1->bytesin >= p2->bytesin) return(-1); 
   return(1);
} 
/*===========================================================================
 * compare connect time
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int comprelt(char **s1, char **s2)
{
   struct TNPUSER *p1;
   struct TNPUSER *p2;
   p1 = (struct TNPUSER *) s1;
   p2 = (struct TNPUSER *) s2;
   if(p1->connect >= p2->connect) return(-1); 
   return(1);
} 
/*===========================================================================
 * compare cpu time
 * inputs: pointers to ptrs for two records 
 * outputs: returns relative sizes  
 *===========================================================================*/
int comprcpu(char **s1, char **s2)
{
   struct TNPUSER *p1;
   struct TNPUSER *p2;
   p1 = (struct TNPUSER *) s1;
   p2 = (struct TNPUSER *) s2;
   if(p1->cpu >= p2->cpu) return(-1); 
   return(1);
} 
/*============================================================================
 * process tnproxy records
 * inputs:     
 *      pointer to current record 
 *	pointer to first record in stack
 *
 * outputs:  
 *	returns number of records added
 *
 *===========================================================================*/
int proctnproxy(struct TNPUSER *rptr, struct TNPUSER *p)
{  
   struct TNPUSER *p2;
   int	i;
   int	newadr; 
   int  found;
    
   p2 = p;
   found	= 0;   
   newadr 	= 0;
   for( i = 0; i < maxusers; i++) 
   {
      if( strcmp(p2->user,rptr->user) == 0)
      {  
         p2->bytesout = p2->bytesout + rptr->bytesout; 
         p2->bytesin  = p2->bytesin  + rptr->bytesin; 
         p2->connect  = p2->connect  + rptr->connect;
         p2->cpu      = p2->cpu      + rptr->cpu;
         found++;
      }
      if( strlen(p2->user) == 0) break;
      if( found ) return(0);
      p2++;
   }
   if(!found)
   {     
      p2->bytesout = rptr->bytesout; 
      p2->bytesin  = rptr->bytesin; 
      p2->connect  = rptr->connect;
      p2->cpu      = rptr->cpu;
      
      strcpy(p2->user,rptr->user); 
      newadr++;
   } 

   return(newadr);
}  
/*============================================================================ 
 * TNPUSER constructor
 *===========================================================================*/
void bldtnpuser(struct TNPUSER *rptr)
{

   rptr->bytesin  = 0;
   rptr->bytesout = 0;
   rptr->connect  = 0; 
   rptr->cpu      = 0.0;
   rptr->user[0]  = '\0';  

   return;
}  


/*============================================================================ 
 * print usage message
 *===========================================================================*/
void usage(void)
{
   
   printf("\n usage: tnprpt [-l logfile ] [-t nnn ]  \n");
   printf("\n        where:\n");  
   printf("        -l logfile   full path name of the log file. \n");
   printf("        -t nnn       list top nnn users\n");
   
   return;
}  
/*==================================================================
 * function: 	isalldigits
 * input:	a pointer to a character string 
 * output:	returns 1 if all the characters in the string are digits 
 *==================================================================*/  
int isalldigits(char *s)  
{
   while(isdigit(*s))
      s++;
   return(*s == '\0');
} 
