/*============================================================================
 * filename:		util/reports/webblocksum/webblocksum.c
 *
 * This software is distributed under the terms of the Livermore Public
 * License version 1.0.
 *	
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0
 *
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA
 *
 * Everyone is permitted to copy and distribute this license document.  The
 * intent of this license is to establish freedom to share and change the
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the
 * copyright holder saying that it may be distributed under the terms of the
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others.
 *
 * 6. You may develop application programs, reusable components and other
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *
 *    b. You must explicitly license all recipients of your items to use and
 *       re-distribute original and modified versions of the items in both
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to
 *       re-distribute to anyone they choose.
 *
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for
 * any damages whatsoever, including - but not restricted to - lost revenue or
 * profits or other direct, indirect, special, incidental or consequential
 * damages, even if they have been advised of the possibility of such damages,
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont
 * Avenue Software, Inc.
 *============================================================================*/
/*=============================================================================
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  All rights
 * reserved.
 *
 * written by:		Tommy Chan
 * Date written:	February   8, 2000
 * Last modified:	February  14, 2000 - added LBL
 *			
 *
 * function:		HTTP Web Block summary report
 *
 * 	The webblocksum program reads the http error log and produces a report
 *	on webblocking activities.
 *
 *============================================================================*/

static  char	ident[] = "@(#) webblocksum, V 1.0, 00/02/14"
 				" (C) Freemont Avenue Software, Inc."
				" 2000";
#include	<sys/types.h>
#include	<sys/file.h>
#include	<sys/stat.h>
#include	<errno.h>
#include	<stdio.h>
#include	<string.h>
#include	<limits.h>
#include	<ctype.h>
#include	<time.h>
#include	"firewall.h"

typedef struct COUNTER COUNTER;
struct COUNTER {
   char		*name;
   int		count;
   COUNTER	*prev;
   COUNTER	*next;
   COUNTER	*client_byfile;		/* Only used in byclient list	*/
};

/*==========================
 * external variables
 *==========================*/
#ifndef linux
extern		char		*sys_errlist[];
#endif
#define		BLOCKSIZE 	1000000
/*======================================================================
 * function prototypes
 *======================================================================*/
void usage(void);

static void *pool = NULL;
static int poolsize = 0;

/*============================================================================
 * Initialize local memory pool
 * Returns 0 if successful
 *         -1 if failed
 *===========================================================================*/
int p_init() {
    pool = (void *) malloc(BLOCKSIZE);
    if (pool == NULL) {
	fprintf(stderr, "webblocksum: Could not malloc memory.  %s\n",
	sys_errlist[errno]);
	return(-1);
    }
    poolsize = BLOCKSIZE;
    return(0);
}
/*============================================================================
 * Allocate memory from a local pool
 * Returns ptr to allocated memory if successful
 *         NULL if failed
 *===========================================================================*/
void *p_malloc(int size) {
    void *retaddr;
    if (poolsize < size) {
	if (p_init() < 0)
	    return(NULL);
    }
    retaddr = pool;
    pool = (char *)pool + size;
    poolsize = poolsize - size;
    return(retaddr);
}

/*============================================================================
 * Add an item to the doubly linked list
 * This function sets the COUNTER structure but does not touch the
 * client_byfile member directly (except to initialize it) and it should
 * stay this way.
 * Returns ptr to the matching record if successful
 *         NULL if failed
 *===========================================================================*/
COUNTER *addItem(char *tmpItem, COUNTER **cntp) {
    COUNTER *last;
    COUNTER *ct = NULL;
    
    for (ct = *cntp; ct != NULL; ct = ct->next) {
	last = ct;		/* Save so we can insert at the end	*/
	if (strcmp(tmpItem, ct->name) == 0) {
	    COUNTER *nextup;
	    COUNTER *tmp;
	    /*===============================================================
	     * Found existing record, increment count
	     *==============================================================*/
	    ct->count++;
	    /*===============================================================
	     * Relocate it if necessary so list is always in descending order
	     * in count
	     *==============================================================*/
	    if (ct->prev && ct->prev->count < ct->count) {
		for (nextup = ct->prev; nextup != NULL; nextup = nextup->prev) {
		    if (nextup->count >= ct->count)
			break;
		}
		/*===========================================================
		 * Remove from the current position
		 *==========================================================*/
		ct->prev->next = ct->next;
		if (ct->next)
		    ct->next->prev = ct->prev;
		/*===========================================================
		 * Insert into the new position
		 *==========================================================*/
		if (nextup == NULL) {
		    ct->prev = NULL;
		    ct->next = *cntp;
		    *cntp = ct;
		}
		else {
		    ct->prev = nextup;
		    ct->next = nextup->next;
		    nextup->next = ct;
		}
		ct->next->prev = ct;
	    }
	    return(ct);
	}
    }
    if (*cntp == NULL || ct == NULL) {
	if ((ct = (COUNTER *)p_malloc(sizeof(COUNTER))) == NULL)
	    return(NULL);
	if ((ct->name = (char *)p_malloc(strlen(tmpItem) + 1)) == NULL)
	    return(NULL);
	strcpy(ct->name, tmpItem);
	ct->client_byfile = NULL;
	ct->count = 1;
	ct->next = NULL;
	if (*cntp == NULL) {
	    *cntp = ct;		/* Head of the list		*/
	    ct->prev = NULL;
	}
	else {
	    last->next = ct;
	    ct->prev = last;
	}
    }
    return(ct);
}
/*============================================================================
 * main
 * inputs: optional input arguments
 *         -h 		for help
 *	   -l logfile	specify logfile name		
 *	   -c cfgfile	specify httpd config file name		
 *         -t nnn	specify printing of top nnn users only
 * output:
 *	return  0 if ok
 *	return -1 on error
 *===========================================================================*/
int main(int argc, char *argv[])
{
   /* log file statistics		*/
   int		trunc_recs	= 0;
   int		total_wb_recs	= 0;
   int		total_recs	= 0;
   char		firstdate[27] = "";
   char		lastdate[27] = "";
   COUNTER	*byfile = NULL;
   COUNTER	*bysite = NULL;
   COUNTER	*byclient = NULL;
   COUNTER	*ct;
   char		*tmpsite;
   char		*tmpfile;
   char		*tmpclient;
   int		i;
   int		j;
   int		line		= 0;
   int		logfound	= 0;
   int		verbose		= 0;
   int		top		= 0;
   static char	httpdconf[PATH_MAX] = "/etc/apache/httpd.conf";
#ifdef _AIX
   static char 	proxylogfile[PATH_MAX] = "/var/adm/error_log";
#elif linux
   static char  proxylogfile[PATH_MAX] = "/var/log/error_log";
#else
   static char 	proxylogfile[PATH_MAX] = "/var/adm/error_log";
#endif
   char		buffer[BUFSIZ];
   FILE 	*confp;			/* httpd.conf file ptr	*/
   FILE		*elfp;			/* Error Log file ptr	*/
   char 	*cp;
   char 	*cp1;
   char		*token1;
   char		*idstring = "] IP Address ";
   char		*idstring2 = ") is blocked by file ";
   time_t	t;
   static char	tokensep[] 	= "\t,:;=[] \n\r";
   /*===================================================
    * check for input arguments
    *==================================================*/
   if( argc > 1) {
      if( argv[1][0] != '-') {
         printf("\n webblocksum:  invalid leading character '%c' \n",
 					argv[1][0]);
         usage();
         return(-1);
      }
      /* check to see if user is asking for help */
      if ((strcmp(argv[1], "-h")    == 0) ||
          (strcmp(argv[1], "-help") == 0) ||
          (strcmp(argv[1], "help")  == 0) ||
          (strcmp(argv[1], "?")     == 0)
         ) {
         usage();
         return(0);
      }
      for(i = 1; i < argc ; i++) {
         switch( argv[i][1]) {
            case 'l':
               if( strlen(argv[i+1]) > 255) {
                  printf("\nwebblocksum: error log file name too long.  "
				"length = %d\n", strlen(argv[i+1]));
                  usage();
                  return(-1);
               }
               strcpy(proxylogfile,argv[i+1]);
               logfound = 1;
               i++;
               break;
            case 'c':
               if( strlen(argv[i+1]) > 255) {
                  printf("\nwebblocksum: httpd config file name too long.  "
				"length = %d\n", strlen(argv[i+1]));
                  usage();
                  return(-1);
               }
               strcpy(httpdconf,argv[i+1]);
               i++;
               break;
            case 't':
               if( argv[i+1] == NULL ||
			strspn(argv[i+1], "0123456789") != strlen(argv[i+1])) {
                  printf("\n webblocksum: the parameter following -t is "
				"missing or is not all digits.\n");
                  usage();
                  return(-1);
               }
               top = atoi(argv[i+1]);
               i++;
               break;
            case 'v':
               verbose = 1;
               break;
            default:
               printf("\n webblocksum: invalid input parameter '%s' \n",
       					argv[i]);
       return(-1);
         }
      }
   }
   cp = (char *) &buffer;
   if (!logfound) {
      /*===========================================================
       * open /etc/apache/httpd.conf file
       *==========================================================*/
      if ((confp = fopen(httpdconf, "r")) == NULL) {
         fprintf(stderr,"webblocksum: open failed for '%s'. %s.\n",
				httpdconf, strerror(errno));
         return(-1);    	
      }
      /*============================================================
       * search for ErrorLog name
       *===========================================================*/
      while(fgets(buffer, BUFSIZ, confp) != NULL) {
         line++;
         /* skip over comments and null lines */
         if( (*cp == '#') || (*cp == '\n') )
            continue;
         token1 = strtok(buffer,tokensep);
         /*=========================================================	
          * If ErrorLog record then get name and validate
          *========================================================*/	
         if( strcasecmp(token1,"ErrorLog") == 0 ) {
            if( (token1 = strtok(NULL,tokensep)) == NULL) {
               fprintf(stderr,"webblocksum: NULL entry after 'ErrorLog' "
				"on line %d of'%s'.\n       %s. \n",
				line, httpdconf);
               return(-1);
            }
            strcpy(proxylogfile,token1);
            logfound = 1;
            break;
         }
      }
   }
   if(!logfound)
   {
      fprintf(stderr,"webblocksum: unable to find TransferLog in '%s'.\n",
				httpdconf, strerror(errno));
         return(-1);
   }
   /*=============================================
    * open proxy-log file
    *============================================*/
   if (!(elfp = fopen(proxylogfile, "r"))) {

      fprintf(stderr,"webblocksum: open failed for '%s'.  %s. \n",
				proxylogfile, strerror(errno));
      return(-1);    			
   }
   cp = (char *) &buffer;
   memset(firstdate, 0, sizeof(firstdate));
   memset(lastdate, 0, sizeof(lastdate));
   /*=======================================================================
    * acquire initial storage
    *======================================================================*/
   if (p_init() < 0)
	return(-1);
   /*========================================================
    * Read the log file and process records
    *=======================================================*/
   while( ( cp = fgets(buffer, BUFSIZ, elfp) ) != NULL) {
      COUNTER *ct;
      total_recs++;
      /*========================================================
       * skip over comments, null lines and short records
       *=======================================================*/
      if( (*cp == '#') || (*cp == '\n') || (strlen(buffer) < 26) ) {
         trunc_recs++;
         continue;
      }
      /*============================================
       * Save Date and time
       *===========================================*/
      if(total_wb_recs == 0)
         memcpy(firstdate,&buffer[1],24);

      memcpy(lastdate, &buffer[1], 24);
      lastdate[26] = '\0';

      /*===========================================
       * See if this is an entry we want
       *===========================================*/
      if ((cp = strstr(buffer, idstring)) == NULL)
         continue;
      if ((cp1 = strstr(buffer, idstring2)) == NULL)
         continue;
      /*==================================================
       * Get pieces of interesting information
       *==================================================*/
      token1 = strtok(cp+1," ");	/* skip "IP"		*/
      token1 = strtok(NULL," ");	/* skip "Address"	*/
      tmpsite = strtok(NULL," ");	/* Site	name		*/
      token1 = strtok(NULL," ");	/* skip "and"		*/
      token1 = strtok(NULL," ");	/* skip "path"		*/
      token1 = strtok(cp1+1," ");	/* skip "is"		*/
      token1 = strtok(NULL," ");	/* skip "blocked"	*/
      token1 = strtok(NULL," ");	/* skip "by"		*/
      token1 = strtok(NULL," ");	/* skip "file"		*/
      tmpfile = strtok(NULL," \"");	/* Filename		*/
      token1 = strtok(NULL,". ");	/* skip "Requesting"	*/
      token1 = strtok(NULL," ");	/* skip "Host"		*/
      tmpclient = strtok(NULL, " \n");	/* Client address	*/

      if (tmpsite == NULL || tmpfile == NULL || tmpclient == NULL) {
         fprintf(stderr, "webblocksum: Error parsing record, record "
         "ignored.\n");
         continue;
      }
      
      total_wb_recs++;
      
      /*==================================================
       * Add to the site list
       *==================================================*/
      if (addItem(tmpsite, &bysite) == NULL)
	  return(-1);
      
      /*==================================================
       * Add to the file list
       *==================================================*/
      if (addItem(tmpfile, &byfile) == NULL)
	  return(-1);
      
      /*==================================================
       * Add to the client list
       *==================================================*/
      if ((ct = addItem(tmpclient, &byclient)) == NULL)
	  return(-1);
      
      /*==================================================
       * Add to the client_byfile list
       *==================================================*/
      if (addItem(tmpfile, &ct->client_byfile) == NULL)
	  return(-1);
   }

   /* print summary reports */
   t = time(NULL);
   printf("WEBBLOCK Summary Report " TREX_VERSION "\n\n");
   printf("  proxy-log file name:                 %s\n\n",proxylogfile);
   printf("  config file name:                    %s\n\n",httpdconf);
   printf("  report date:                         = %s\n",ctime(&t));
   printf("  date of first record                 = %s\n",firstdate);
   printf("  date of last record                  = %s\n",lastdate);
   printf("  total records in log file            = %10d\n",total_recs);
   printf("  total number of blocked requests     = %10d\n",total_wb_recs);
   printf("\n\n");
   
   if (top != 0)
      j = top;
   else
      j = total_wb_recs;
   
   printf("  \n\nTop sites by count\n\n");
   printf("    Site Name (truncated to 56 characters)                  "
		      "     Count\n");
   printf("    --------------------------------------------------------"
		      "     ----------\n");
   ct = bysite;
   for (i = 0; i < j; i++) {
      if (ct == NULL)
         break;
      printf("    %56s     %10d\n", ct->name, ct->count);
      ct = ct->next;
   }
   
   printf("  \n\nTop block files by count\n\n");
   printf("    Block File Name (truncated to 56 characters)            "
		      "     Count\n");
   printf("    --------------------------------------------------------"
		      "     ----------\n");
   ct = byfile;
   for (i = 0; i < j; i++) {
      if (ct == NULL)
         break;
      printf("    %56s     %10d\n", ct->name, ct->count);
      ct = ct->next;
   }
   
   if (!verbose) {
      printf("  \n\nTop clients by count\n\n");
      printf("    Client Name (truncated to 56 characters)                "
		      "     Count\n");
      printf("    --------------------------------------------------------"
		      "     ----------\n");
      ct = byclient;
      for (i = 0; i < j; i++) {
         if (ct == NULL)
            break;
         printf("    %56s     %10d\n", ct->name, ct->count);
         ct = ct->next;
      }
   }
   else {
      printf("  \n\nTop clients by count\n\n");
      printf("    Client Name (to 28 characters)   File Name (to 28 chara"
		      "cters)    Count\n");
      printf("    ----------------------------     ----------------------"
		      "-----     ----------\n");
      ct = byclient;
      for (i = 0; i < j; i++) {
	 COUNTER *sct;
         if (ct == NULL)
            break;
         printf("    %28s                                     %10d\n",
			 				ct->name, ct->count);
         for (sct = ct->client_byfile; sct != NULL; sct = sct->next) {
            printf("                                    %28s     %10d\n",
			    sct->name, sct->count);
         }
         ct = ct->next;
      }
   }
}
/*============================================================================
 * print usage message
 *===========================================================================*/
void usage(void) {
   printf("\n usage: webblocksum [-l logfile ] [-c cfgfile ][-t nnn ]\n");
   printf("\n        where:\n");
   printf("        -l logfile   full path name of the log file.\n");
   printf("        -l cfgfile   full path name of the http config file.\n");
   printf("        -t nnn       list top nnn users\n");
   printf("        -v           printf detailed client info\n");
   return;
} 
