/*============================================================================
 * filename:		webgate/httplog.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *										
 * written by:		Jim Livermore
 *			1602 Mossy Stone
 *			Houston, TX 77077
 *
 * (C) COPYRIGHT R. James Livermore 1996-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1996-2000.  All rights
 * reserved.    
 *
 *		
 * Date written:	August     6, 1996
 * Last modified:	February   8, 2000
 *
 * function:                       
 *              
 *
 *
 *
 *============================================================================*/
#include	<fcntl.h>
#include	<stdio.h>
#include	<syslog.h>  
#include	<string.h>
#include	<sys/errno.h> 
#include	<sys/ioctl.h>
#include	<sys/types.h>
#include	<sys/time.h>  
#include	<sys/times.h>           /* cpu times			  */
#ifndef __hppa
#include 	<sys/select.h>		/* to support fd_set 		  */ 
#endif 
#include	<sys/signal.h>
#include	<sys/socket.h> 
#include	<time.h>
#include	<netdb.h>               /* defines servent structure      */
#include	<netinet/in.h> 
#include	"firewall.h"   
#include	"webgate.h"
/*===========================================================================
 * external variables
 *==========================================================================*/
extern long int timezone;
extern int      daylight;
extern char* tzname[];  
/*============================================================================
 * function prototypes
 *===========================================================================*/
int inithttplog(struct HTTPLOG *plog); 
char *log_time(char *GMTdelta); 
void log_wrt(char *srcip, struct HTTPLOG *plog, char * logname,char * gtmdelta,
             struct WGCONF *pwg); 


/*============================================================================
 * function: 	httplog
 * input:	a pointer to the input buffer 
 *		record size
 *		a pointer to the HTTP log record
 *		ptr to WGCONF structure
 *		character: C=>client, S=>Server
 * 
 * output:	returns 0 if function ok
 *		updated HTTP Log record
 *===========================================================================*/  
int httplog(char *buff, int recsize, struct HTTPLOG *plog, 
            struct WGCONF *wproxptr, char cs)  
{        
   char		*cp1;
   char		*cp2;
   int		i; 
   int		len;  
   int		Reflen;
   char		*token1;
   char		*token2;
   char		tokensep[] = "\t,:;= \n\r";  
   char		tmpbuf[4096+2]; 
   /* move data to temporary buffer to enable use of string functions */ 
   tmpbuf[4096+1] = '\0';		/* set EOS */
   if( recsize > 4096 ) recsize = 4096;
   strncpy(tmpbuf,buff,recsize); 
   if( (token1 = strtok(tmpbuf,tokensep)) == NULL)
   {
      return(0);
   } 
   /*===================================================
    * Process Client information 
    *===================================================*/
   if( cs == 'C' )
   {   
      /*=====================================================================
       * Catch state change from Server Last to Client and write httplog
       * record.
       *=====================================================================*/ 
      plog->ServerCnt = 0;
      if( plog->state == ServerLast )   
      {  
         if( wproxptr->DNSflg && (strcmp(wproxptr->ns->srcHost,"unknown")
									!= 0) ) 
            log_wrt(wproxptr->ns->srcHost, plog, wproxptr->logname,
                    wproxptr->GMTdelta, wproxptr);  
         else
            log_wrt(wproxptr->ns->srcIP, plog, wproxptr->logname,
		    wproxptr->GMTdelta, wproxptr);  
         inithttplog(plog);                     
      } 
      /*================================================
       * If not GET or POST or HEAD throw it away.	
       *===============================================*/
      if( strcmp(token1,"GET") == 0)
         wproxptr->ns->totalGETs++;
      else if( strcmp(token1,"POST") == 0)
         wproxptr->ns->totalPOSTs++;
      else if(strcmp(token1,"HEAD") == 0)
            wproxptr->ns->totalHEADs++;
      else
      { 
         plog->Client_Cmd_Length = plog->Client_Cmd_Length + recsize;
         return(0);
      }
      plog->state = ClientLast;  
      strncpy(plog->request,token1,sizeof(plog->request)-1); 
      /*============================== 
       * Get complete URL pathname. 
       *=============================*/  
      if( (token1 = strtok(NULL,"\t,; \n\r")) == NULL)
      {
         inithttplog(plog);
         return(0);
      }     
      strncpy(plog->data,token1,sizeof(plog->data)-1);
      plog->data[sizeof(plog->data)-1] = '\0'; 
      /*==================================
       * Get HTTP version	
       *=================================*/
      if( (token1 = strtok(NULL,tokensep)) == NULL)
      {
         return(0);
      } 
      strncpy(plog->version,token1,sizeof(plog->version)-1); 
      plog->Client_Cmd_Length = plog->Client_Cmd_Length + recsize;
      /*========================================================   
       * Look for Connection:, Refer:, User-Agent:
       *
       *========================================================*/
      while( (token1 = strtok(NULL,tokensep)) != NULL)
      {     
         /* If "Connection" then skip over "Keep-Alive" */
         if( strcmp(token1,"Connection") == 0)
         {
            if( (token1 = strtok(NULL,tokensep)) == NULL) /* skip over */
            {
               inithttplog(plog);
               return(0);
            } 
            continue;
         }           
         /* If "Referer" then get "hhtp://gw/"	*/
         if( strcmp(token1,"Referer") == 0)
         {
            if( (token1 = strtok(NULL,"\t,;= \n\r")) == NULL)
            {
               inithttplog(plog);
               return(0);
            }
            strncpy(plog->Referer,token1,sizeof(plog->Referer)-1);
            Reflen = strlen(plog->Referer);
            if( plog->Referer[Reflen-1] == '/' )
            {
               plog->Referer[Reflen-1] = '\0'; 
            }  
            strncpy(plog->HostRefer,plog->Referer,sizeof(plog->HostRefer)-1);
            if( (cp1 = strchr(plog->HostRefer,':')) != NULL)
            {
               cp2 = cp1 + 3; 
               cp1 = strchr(cp2,'/');
               if( cp1 != NULL) *cp1 = '\0'; 
            }  
            continue;
         }
         /*=============================================================
          * Save User-Agent data and remove unwanted \r & \n characters.
          *===========================================================*/                                         
         if( strcmp(token1,"User-Agent") == 0)
         {  
            token1 = token1 + strlen(token1) + 1; 
            while(isspace(*token1)) 
	       token1++;
            if( (( cp1 = strchr(token1,'(')) != NULL) &&
                (( cp2 = strchr(token1,')')) != NULL)) 
            {
               len = cp2 - token1 + 1; 
               /*=====================================================
	        * If len is too long then shorten and look for other 
                * commands that that should be excluded.
                *====================================================*/             
               if( len >= sizeof(plog->User_Agent))
               {    
                  len =  sizeof(plog->User_Agent) -1;
                  strncpy(plog->User_Agent,token1,len);
	          plog->User_Agent[ len] = '\0';
                  if( (cp1 = strstr(plog->User_Agent,"Connection:")) !=NULL)
                    *cp1 = '\0';
               }
               else
               {
                  strncpy(plog->User_Agent,token1,len);
	          plog->User_Agent[len] = '\0';
               }
               /*=========================================
		*  remove '\r' from User_Agent data 
                *========================================*/
	       cp1 = &plog->User_Agent[0];	
	       for(i = 0; i < len; i++)
               {
                  cp1++;
                  if( *cp1 == '\r' || *cp1 == '\n' )
                     *cp1 = ' ';
               }  	 
            }else /* User-Agent data is not enclosed within () */
            {
               if( (token1 = strtok(NULL,"\t,:  \n\r")) )
               { 
                  if((len = strlen(token1)) >= sizeof(plog->User_Agent))
                     len = sizeof(plog->User_Agent) -1;
                  strncpy(plog->User_Agent,token1,len);
	          plog->User_Agent[len] = '\0';  
               } else
                  strcpy(plog->User_Agent,"NULL");
            }
            continue;
         }  
      }
   } 
   /*==========================================
    * Process Server information
    *=========================================*/               
   else
   {                     
      if( plog->ServerCnt == 0)
      {  
         if( strncmp(token1,"HTTP",4) == 0)
         {         
            if( (token2 = strtok(NULL,tokensep)) != NULL)
            { 
               if( isalldigits(token2))  
                strncpy(plog->Status,token2,sizeof(plog->Status)-1);
            } 
         }  
         plog->ServerCnt++;
      }
      plog->Content_Length = plog->Content_Length + recsize;
      plog->state = ServerLast;  
   }   
   return(0);
}  
/*====================================================================
 * function: 	inithttplog
 * input:	a pointer to HTTP Log record 
 * output:	returns 0
 *		initializes counters to zero and strings to NULL. 
 *====================================================================*/  
int inithttplog(struct HTTPLOG *plog)  
{
   plog->state 		    	= ClientLast;  
   plog->ServerCnt		= 0;
   memset(&plog->srcIP[0],'\0',sizeof(plog->srcIP));
   strcpy(plog->identd," - "); 
   strcpy(plog->user," - "); 
   memset(&plog->CLFtime[0],'\0',sizeof(plog->CLFtime));
   memset(&plog->request[0],'\0',sizeof(plog->request));
   memset(&plog->data[0],'\0',sizeof(plog->data));
   strcpy(plog->Status,"200"); 
   memset(&plog->version[0],'\0',sizeof(plog->version));
   memset(&plog->HostRefer[0],'\0',sizeof(plog->HostRefer));
   memset(&plog->Referer[0],'\0',sizeof(plog->Referer));
   memset(&plog->User_Agent[0],'\0',sizeof(plog->User_Agent)); 
   plog->Client_Cmd_Length  	= 0;
   memset(&plog->Content_Type[0],'\0',sizeof(plog->Content_Type));
   plog->Content_Length	    	= 0;  
   plog->TotalServerBytes 	= 0;
   plog->TotalClientBytes   	= 0;
   return(0);
}  
/*==================================================================
 * function: 	log_time
 * input:	void
 * output:	returns pointer to character string containing log time
 *		in the following format:
 *		day/month/ccxx:hh:mm::ss + xxxx  
 * 		Date is Year 2000 compliant.
 *==================================================================*/  
char *log_time(char *GMTdelta)  
{ 
   static time_t cur_time; 
   static char 	ltime[32];
   static char month_names[12][4] =
   {
      "Jan","Feb","Mar","Apr","May","Jun",
      "Jul","Aug","Sep","Oct","Nov","Dec"
   };   
   int		century;
   int          delta;

   int		tzone;  
   struct tm    *tp;
   struct tm 	*ltp;		/* local time	*/
   struct tm    *gmtp;
   struct tm    GMtime;
   gmtp = &GMtime;              
   time(&cur_time); 
   tp = gmtime(&cur_time);  
   memcpy(gmtp,tp,sizeof(struct tm));
   ltp  = localtime(&cur_time);   
   /* Done for year 2000 compliance    */
   if( ltp->tm_year < 69 )     	
      century = 2000;
   else
      century = 1900; 
   sprintf(ltime,"%02d/%s/%04d:%02d:%02d:%02d %5.4d",
           ltp->tm_mday,
           month_names[ltp->tm_mon],
           century + ltp->tm_year, 
           ltp->tm_hour,
           ltp->tm_min,
	   ltp->tm_sec,
           (ltp->tm_hour - gmtp->tm_hour)*100
          );
   return(ltime);
} 
/*============================================================================
 * function: 	log_wrt
 * input:
 *		ptr to client IP address
 *		ptr to HTTPLOG record
 *		ptr to log name
 *		GMT delta
 * output:	writes http log record in one of the following formats
 *			Common Log Format + Extensions
 *			WebTrends Multi-Homed
 *
 *===========================================================================*/  
void log_wrt(char *srcIP, struct HTTPLOG *plog, char * logname, char* GMTdelta,
             struct WGCONF *pWG)  
{                             
   char		*cp;
   FILE		*logfile; 
   /* If request type is not valid return without writing record */
   if((strcmp(plog->request,"GET")  != 0) && 
      (strcmp(plog->request,"POST") != 0) &&  
      (strcmp(plog->request,"HEAD") != 0) ) 
   {
      syslog(LOG_INFO,"Request = '%s'",plog->request);
      return;
   }
   if ((logfile = fopen(logname, "a"))  == NULL)
   {
      syslog(LOG_ERR,"Error: unable to open httpd log '%s'.",logname);
      return;      
   }  
   /* Check for NULL strings before trying to print	*/
   if( strlen(plog->HostRefer) == 0)
      strcpy(plog->HostRefer," ");
   /*===============================================================
    * Tidy up the URL before print
    *==============================================================*/
   if (strncmp(plog->data, "http://", 7) == 0)
   {
      if ((cp = strchr(&plog->data[7], '/')) == NULL)
         cp = "/";
   }
   else
      cp = plog->data;
   /*===============================================================
    * Write Common Log Format plus Extensions 
    *==============================================================*/
   if(pWG->logformat == CLFExt)
   {
      fprintf(logfile,"%s - - [%s] \"%s http://%s%s %s\" %s %d \"%s\" \"%s\"\n",
                   srcIP, 
                   log_time(GMTdelta),
	           plog->request, pWG->rulep[pWG->ns->rule_index].toname,
	           cp, plog->version,
		   plog->Status, plog->Content_Length,
		   plog->Referer, plog->User_Agent);  
   } 
   /*===============================================================
    * Write WebTrends Multi-Homed Format (violates CLF standards) 
    *==============================================================*/ 
   else if(pWG->logformat == WTMultiHomed)
   {                       
      fprintf(logfile,"%s %s - [%s] \"%s %s %s\" %s %d \"%s\" \"%s\"\n",
                   srcIP,
                   pWG->rulep[pWG->ns->rule_index].toname, 
                   log_time(GMTdelta),
	           plog->request,  plog->data, plog->version,
		   plog->Status, plog->Content_Length,
		   plog->Referer, plog->User_Agent);  
   } 
   fclose(logfile); 
   inithttplog(plog);		/* reset http log	*/
   return;     
}
