/*============================================================================
 * filename:		webgate/main.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *										
 * written by:		Tommy Chan
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.
 * All rights reserved. 
 *           
 *		
 * Date written:	December  20, 1997
 * Last modified:	February   8, 2000
 *	                         
 * function:	Webgate Version 1
 *
 *	Webgate is a WWW gateway for WWW Servers that reside behind the
 *	T.Rex firewall.  It provides access control and extensive logging
 *	of all requests made of the WWW server.  Webgate Vesion 2 is a
 *	stand-alone daemon that preforks processes to produce higher 
 *	performance than using inetd.
 *	  
 *	Webgate reads its configuration file to determine wether or not
 *	the client is permitted to access the server.  If access is NOT 
 *	permitted the attempt is logged. Permissions are granted based on
 *      the source IP address the local IP address and port number. 
 *	Entries in the configuration file also determine the destination host
 *	and port number.                 
 *  
 *	Debug option
 *                     
 * 	Webgate can be run from the command line with the Debug option. 
 *	To use the debug option type in the program name, the port number,
 *      -D and an optional port used for the test. If the optional port is 
 *      not specified the program will use port number 8080. 
 *
 *		webgate 6080 -D
 *      or	webgate 6080 -D 7080 
 *                       
 *	Kill
 *	
 *	To kill all parent and all listening processes that are not 
 *	currently connected issue the folloing command
 *
 *	kill -USR1 `/etc/firewall/webgate.pid`
 *
 *============================================================================*/

   
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <syslog.h>
#include <fcntl.h>
#include <signal.h>
#include <limits.h>
#include <netdb.h>
#if !defined(__hppa) && !defined(linux)
#include <sys/mode.h>
#endif
#include <sys/poll.h> 
#include <sys/socket.h> 
#include <sys/stat.h>		/* chmod defines for Solaris */
#include <sys/types.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <sys/resource.h>
#include <sys/ipc.h>
#include <sys/mman.h>
#include "firewall.h"
#include "webgate.h"
#ifdef _AIX41
#include        <procinfo.h>
#else
#ifdef __sun
#define         _KMEMUSER 1
#include        <sys/proc.h>
#include        <kvm.h>
#include        <fcntl.h>
#else
#ifdef __hppa
#include        <sys/pstat.h>
#endif
#endif
#endif

extern WGCONF *getwgconf();
extern IPSTR *read_secfile(char *fn, int *n);
#ifndef linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif
extern child_main(WGCONF *wgconfp, int slot);
extern char *getshm(int size, char *mfile);

static WGCONF *wgconfp;
static char *cmdname;
#ifdef __sun
static volatile sig_atomic_t received_sigchld = 0;
#else
static sig_atomic_t received_sigchld = 0;
#endif

/*====================================================================
 * signal handler for SIGUSR2 and SIGCHLD
 *===================================================================*/
void upchild(int sig) {
    pid_t pid;
    int i;
    int stat;
    /*===================================================================
     * Call waitpid (repeatedly if necessary) to get rid of any zombies
     *==================================================================*/
    while ((pid = waitpid(-1, &stat, WNOHANG)) > 0) {
	for (i = 0; i < wgconfp->maxprocs; i++) {
	    if (wgconfp->child[i].pid == pid) {
		wgconfp->child[i].pid = 0;
		wgconfp->child[i].nconn = 0;
		wgconfp->child[i].status = CHILD_NONE;
		break;
	    }
	}
    }

    received_sigchld = 1;
    signal(SIGCHLD, upchild);
    signal(SIGUSR2, upchild);
    return;
}

/*====================================================================
 * Do the real work for SIGCHLD
 * This is done outside of a signal handler because we need to call
 * non-ASYNC-SIGNAL-SAFE functions and system calls.
 *===================================================================*/
void do_upchild() {
    pid_t pid;
    int i;
    int count;
    int tcount;
    int ndiff;
    int stats[CHILD_N_STAT];
#define N_IDLE stats[CHILD_IDLE]
#define N_BUSY stats[CHILD_BUSY]
#define N_NONE stats[CHILD_NONE]

    memset(&stats, 0, sizeof(stats));

    /*===================================================================
     * Go thru the child table and collect status statistics
     *==================================================================*/
    for (i = 0; i < wgconfp->maxprocs; i++)
	stats[wgconfp->child[i].status]++;

    DEBUG(1, wgconfp, (DBLOG, "webgate statistics: idle = %d, busy = %d.",
			N_IDLE, N_BUSY));
    DEBUG(2, wgconfp, (DBLOG, "webgate extended statistics: empty slots = %d, "
			"term = %d.", stats[CHILD_NONE], stats[CHILD_TERM]));

    tcount = wgconfp->maxprocs - N_NONE;	/* total n child	*/
    count = tcount;

    /*===================================================================
     * Start some new child processes if necessary
     *==================================================================*/
    if (N_IDLE < wgconfp->minspare) {
	ndiff = wgconfp->minspare - N_IDLE;
	if (ndiff > N_NONE)
	    ndiff = N_NONE;
	if (ndiff == 0)
	    DEBUG(3, wgconfp, (DBLOG, "maxprocs has been reached, empty slot = "
			"%d.  Cannot start new processes.", N_NONE));
	else
	    DEBUG(2, wgconfp, (DBLOG, "To start %d new processes, minspare = "
			"%d, idle = %d, empty = %d.", ndiff, wgconfp->minspare,
			N_IDLE, N_NONE));

	/*===================================================================
	 * Loop thru to find empty slots and start processes in them
	 *==================================================================*/
	i = 0;
	while (ndiff) {
	    if (i >= wgconfp->maxprocs)
		break;

	    if (wgconfp->child[i].status != CHILD_NONE) {
		i++;
		continue;
	    }

	    wgconfp->child[i].nconn = 0;
	    wgconfp->child[i].status = CHILD_IDLE;

	    if ((pid = fork()) <0) {
		syslog(LOG_ERR,"Error: Unable to fork child %d.  %s",i,
			sys_errlist[errno]);
		wgconfp->child[i].status = CHILD_NONE;
		ndiff--;	/* May not want to start all processes	*/
		sleep(wgconfp->checkprocs);	/* Give it a break	*/
		continue;
	    }

	    if (pid != 0)
		wgconfp->child[i].pid = pid;
	    else {
	    /*===============================================================
	     * In the child now
	     *==============================================================*/
		signal(SIGCHLD, SIG_IGN);
		signal(SIGUSR2, SIG_IGN);
		signal(SIGUSR1, SIG_DFL);
		signal(SIGTERM, SIG_DFL);
		child_main(wgconfp, i);		/* never returns	*/
	    }
	    ndiff--;
	    count++;
	    i++;
	}

	if (count > tcount)
	    DEBUG(1, wgconfp, (DBLOG,"Number of child processes increased "
			"from %d to %d.", tcount, count));
	else
	    DEBUG(3, wgconfp, (DBLOG, "Number of child processes did not "
			"increase."));
    }
    /*=======================================================================
     * Kill some child processes if too many idle processes are found
     *======================================================================*/
    else if (N_IDLE > wgconfp->maxspare) {
	ndiff = N_IDLE - wgconfp->maxspare;

	DEBUG(2, wgconfp, (DBLOG, "To terminate %d processes, maxspare = %d, "
			"idle = %d.", ndiff, wgconfp->maxspare, N_IDLE));

	/*===================================================================
	 * Loop thru to find idle slots and kill processes in them
	 *==================================================================*/
	i = 0;
	while (ndiff) {
	    if (i >= wgconfp->maxprocs)
		break;

	    if (wgconfp->child[i].status != CHILD_IDLE) {
		i++;
		continue;
	    }

	    /*===============================================================
	     * Send SIGUSR1 instead of SIGKILL just in case child has
	     * activated since we check.
	     * We send a signal to the child here and let the SIGCHLD handler
	     * handle updating the child slot when it receives the SIGCHLD
	     * from the child.
	     * We use the pid variable to freeze a copy of the pid number from
	     * the child table so we don't have to worry about the signal
	     * handler zeroing it out while we try to kill it.
	     *==============================================================*/
	    pid = wgconfp->child[i].pid;
	    DEBUG(3, wgconfp, (DBLOG, "Terminating child pid %d in slot %d.",
			pid, i));
	    if (pid != 0)
		kill(pid, SIGUSR1);
	    else {
		i++;
		continue;
	    }

	    ndiff--;
	    count--;
	    i++;
	}

	if (count < tcount)
	    DEBUG(1, wgconfp, (DBLOG,"Number of child processes reduced "
			"from %d to %d.", tcount, count));
	else
	    DEBUG(3, wgconfp, (DBLOG, "Number of child processes did not "
			"decrease."));
    }

    return;
}

/*====================================================================
 * signal handler for SIGUSR1
 *===================================================================*/
void refresh(int sig) 
{
    int i;

    killpg(getpid(), SIGUSR1);
    syslog(LOG_NOTICE, "SIGTERM received, webgate exiting.");
    for (i = 0; wgconfp->fdp[i].fd >= 0; i++)
	close(wgconfp->fdp[i].fd);
    unlink(SHRCONF);
    exit(0);
}
/*====================================================================
 * MAIN
 *===================================================================*/
int main (int argc, char *argv[]) 
{ 
   static  char	ident[] = "@(#) webgate, V 1.0, 00/01/21"
 			  " (C) Freemont Avenue Software, Inc."
			  " 1996-2000";
    int i;
    int fd;
    int nfd = 0;
    const int one = 1;
    int nfilemax;
    pid_t pid;
    char *tfn;
    struct sockaddr_in sock;
    FILE *pf;
    struct sigaction act, oact;

    /*=======================================================================
     * Check for -r flag
     *======================================================================*/
    if (argc > 1)
	rflag(argc, argv);	/* Never returns	*/

    openlog("webgate",LOG_PID|LOG_NDELAY,LOG_DAEMON);  

    /*===================================================================
     *  Save command name
     *=================================================================*/
    if ((tfn = strrchr(argv[0], '/')) == NULL)
	tfn = argv[0];
    else
	tfn++;

    if ((cmdname = malloc(strlen(tfn) + 1)) == NULL) {
	syslog (LOG_ERR, "Error: Cannot malloc.  %s", sys_errlist[errno]);
	exit(1);
    }
    strcpy(cmdname, tfn);

    /*===================================================================
     *  Dissociate from the controlling terminal
     *=================================================================*/
    setsid();
    chdir("/");

    /*=======================================================================
     * Make this a daemon
     *======================================================================*/
    if ((pid = fork()) < 0) {
	syslog(LOG_ERR,"Error: Cannot fork.  %s",sys_errlist[errno]);
	exit(1);
    }

    if (pid != 0)
	exit(0);

    /*==================================================================
     * Set up a process group so tha all the child processes can be
     * be signaled as a group when needed
     *=================================================================*/
#ifdef __hppa
    setpgid(0, 0);
#else
    setpgrp();
#endif

    /*=======================================================================
     * Read in the config file
     *======================================================================*/
    if ((wgconfp = getwgconf()) == NULL)
	exit(1);

    /*=======================================================================
     * Read in the securenets file and map it
     *======================================================================*/
    if ((wgconfp->snp = read_secfile(SECURENETFILE, &(wgconfp->nsnp))) == NULL)
	exit(1);

    /*=======================================================================
     * Read in the secureports file and map it
     *======================================================================*/
    if ((wgconfp->spp = read_secfile(SECUREPORTFILE, &(wgconfp->nspp))) == NULL)
	exit(1);

    /*======================================================================
     * Use the allowable limit for number of opened files
     *====================================================================*/
#ifndef _AIX
    {
	struct rlimit fl;
	int rc;
	rc = getrlimit(RLIMIT_NOFILE,&fl);
	if (rc) {
	    syslog(LOG_ERR,"Error: Cannot do getrlimit.");
	    exit(1);
	}
	fl.rlim_cur = fl.rlim_max;
	rc = setrlimit(RLIMIT_NOFILE,&fl);
	if (rc) {
	    syslog(LOG_ERR,"Error: Cannot do setrlimit.");
	    exit(1);
	}
	nfilemax = fl.rlim_cur;
    }
#else
	nfilemax = OPEN_MAX;
#endif

    /*=======================================================================
     * Fill the fd array with -1
     *======================================================================*/
    if ((wgconfp->fdp = (struct pollfd *)malloc(sizeof(struct pollfd)*nfilemax))
					== NULL)
    {
	syslog(LOG_ERR, "Error: malloc failed for fdp array.  %m");
	exit(1);
    }

    memset((char *)(wgconfp->fdp), 0xff, nfilemax*sizeof(struct pollfd));

    /*=======================================================================
     * Get shared memory for child info table
     *======================================================================*/
    if ((wgconfp->child =
		(CHILD *)getshm(wgconfp->maxprocs * sizeof(CHILD), NULL))
		    							== NULL)
	exit(1);

    /*=======================================================================
     * Loop thru all the rules and bind the ports and start listening to each
     *======================================================================*/
    memset(&sock, 0, sizeof(sock));
    sock.sin_family = AF_INET;
    for (i = 0; i < wgconfp->nrules; i++) 
    {
	int tmp;
	int flags;

	/*===================================================================
	 * We could check for and skip deny rules here but if we do that,
	 * we would eliminate a feature here (one that allows us to have
	 * webgate/genproxy activate a port just so we can report on
	 * suspicious scanning of the activated port).
	 * if (op == DENY)
	 *     continue;
	 *==================================================================*/

	if ((tmp = socket(AF_INET, SOCK_STREAM, 0)) < 0) 
        {
	    syslog(LOG_ERR,"Error: socket() failed.  %s",sys_errlist[errno]);
	    exit(1);
	}
	setsockopt(tmp, SOL_SOCKET, SO_REUSEADDR, (const char *)&one,
								sizeof(one));

	/*===================================================================
	 * Set the socket to non_block to avoid the timing problem described
	 * in Stevens UNP Volume 1 page 422-424.
	 *==================================================================*/
	flags = fcntl(tmp, F_GETFL, 0);
	fcntl(tmp, F_SETFL, flags | O_NONBLOCK);

	if (!strcmp(wgconfp->rulep[i].toaddr,"*"))
	    sock.sin_addr.s_addr = INADDR_ANY;
	else
	    sock.sin_addr.s_addr = inet_addr(wgconfp->rulep[i].toaddr);

	sock.sin_port = htons(wgconfp->rulep[i].toport);

	if (bind(tmp, (struct sockaddr *)&sock, sizeof(sock)) != 0) 
        {
	    if (errno == EADDRINUSE) 
            {
		int j;
		int ownport = 0;
		/*===========================================================
		 * It is not an error for EADDRINUSE if we bound the same
		 * port in an earlier round.
		 *==========================================================*/
		for (j = 0; j < i; j++) 
                {
		    if (wgconfp->rulep[j].toport == wgconfp->rulep[i].toport) 
                    {
			ownport = 1;
			break;
		    }
		}
		if (ownport) {
		    close(tmp);
		    continue;
		}
		else {
		    syslog(LOG_ERR,"Error: bind() failed for port %d.  %s",
				wgconfp->rulep[i].toport, sys_errlist[errno]);
		    close(tmp);
		    exit(1);
		}
	    }
	    else if (errno == EADDRNOTAVAIL) {
		syslog(LOG_INFO, "Address \"%s\" not available on local "
				"machine at this time.  permit line not used.",
				wgconfp->rulep[i].toaddr);
		close(tmp);
		continue;
	    }
	    close(tmp);
	    syslog(LOG_ERR,"Error: bind() failed for port %d.  %s",
				ntohs(sock.sin_port), sys_errlist[errno]);
	    exit(1);
	}

	if (listen(tmp, 512) != 0) {
	    close(tmp);
	    syslog(LOG_ERR,"Error: listen() failed.  %s",sys_errlist[errno]);
	    exit(1);
	}
	if (nfd >= nfilemax) {
	    syslog(LOG_ERR,"Error: Too many opened file descriptors");
	    exit(1);
	}
	wgconfp->fdp[nfd].fd = tmp;
	wgconfp->fdp[nfd].events = POLLIN;
	wgconfp->fdp[nfd++].revents = 0;
    }

    wgconfp->nfds = nfd;
    /*=======================================================================
     * Init the accept lock
     *======================================================================*/
    if ((wgconfp->lockfd = my_lock_init(FW_CFG_DIR "/webgXXXXXX")) < 0)
    {
	exit(1);
    }
    /*=======================================================================
     * Init the round robin read/write lock
     *======================================================================*/
    if ((wgconfp->rrfd = my_lock_init(FW_CFG_DIR "/webgXXXXXX")) < 0)
    {
	exit(1);
    }

    /*========================================================================
     * Write a pid file so the daemon can be signalled easily
     *=======================================================================*/
    if (!(pf = fopen(PIDFILE,"w"))) {
	syslog(LOG_ERR,"Error: Failed opening %s for writes.  %s\n",
			PIDFILE,sys_errlist[errno]);
	return(-1);
    }

    if (fprintf(pf,"%d\n",getpid()) < 0) {
	syslog(LOG_ERR,"Error: Failed writing to %s.  %s\n",PIDFILE,
			sys_errlist[errno]);
	return(-1);
    }

    fclose(pf);

    chmod(PIDFILE,S_IRUSR|S_IWUSR|S_IRGRP);

    /*=====================================================================
     * Make sure gproxpath is not NULL.  chdir and chroot will be done in
     * the children.
     *===================================================================*/
    if( wgconfp->gproxpath[0] == '\0') {  
	syslog(LOG_ERR,"Error: NULL gproxpath directory entry returned"); 
	exit(1);
    }

    /*========================================================================
     * Setup the signal handlers
     *=======================================================================*/
    signal(SIGUSR1, refresh);
    signal(SIGTERM, refresh);
    act.sa_handler = refresh;
    sigemptyset(&act.sa_mask);
    sigaddset(&act.sa_mask, SIGUSR2);	/* Disable USR2 in handler	*/
    sigaddset(&act.sa_mask, SIGCHLD);	/* Disable CHLD in handler	*/
    act.sa_flags = 0;
    if (sigaction(SIGUSR1, &act, &oact) < 0) {
	syslog(LOG_ERR,"Error: Unable to set signal handler for USR1.  %s",
			sys_errlist[errno]);
	exit(1);
    }
    if (sigaction(SIGTERM, &act, &oact) < 0) {
	syslog(LOG_ERR,"Error: Unable to set signal handler for TERM.  %s",
			sys_errlist[errno]);
	exit(1);
    }

    /*=======================================================================
     * Spawn the children
     *======================================================================*/
    for (i = 0; i < wgconfp->nprocs; i++) {
	/*===================================================================
	 * Initialize the child slot as much as we can before the fork to
	 * avoid the possibility of the child updating its slot before the
	 * initialization is complete.  The pid will have to be set later but
	 * that's OK since the child is not going to touch the pid member
	 * anyway.
	 *==================================================================*/
	wgconfp->child[i].nconn = 0;
	wgconfp->child[i].status = CHILD_IDLE;

	if ((pid = fork()) <0) {
	    syslog(LOG_ERR,"Error: Unable to fork child %d.  %s",i,
			sys_errlist[errno]);
	    exit(1);
	}

	if (pid != 0)
	    /*===============================================================
	     * The parent initializes the child table slot for this child
	     *==============================================================*/
	    wgconfp->child[i].pid = pid;
	else {
	    /*===============================================================
	     * In the child now
	     *==============================================================*/
	    signal(SIGCHLD, SIG_IGN);
	    signal(SIGUSR2, SIG_IGN);
	    signal(SIGUSR1, SIG_DFL);
	    signal(SIGTERM, SIG_DFL);
	    child_main(wgconfp, i);		/* never returns	*/
	}
    }
    /*========================================================================
     * Log the start of webgate
     *=======================================================================*/
   if(!wgconfp->traceflg)
   {
      syslog(LOG_NOTICE,"webgate version 1.0 started.  Number of processes = "
			"%d, minspare = %d, maxspare = %d, maxprocs = %d, "
			"maxuse = %d, interval = %d",
			wgconfp->nprocs, wgconfp->minspare,
			wgconfp->maxspare, wgconfp->maxprocs,
			wgconfp->maxuse, wgconfp->checkprocs);
   }
   else
   {
      syslog(LOG_NOTICE,"webgate version 1.0 started.  Number of processes = "
			"%d, minspare = %d, maxspare = %d, maxprocs = %d, "
                        "maxuse = %d, interval = %d, trace = yes",
			wgconfp->nprocs, wgconfp->minspare,
			wgconfp->maxspare, wgconfp->maxprocs,
			wgconfp->maxuse, wgconfp->checkprocs);
   }

    signal(SIGUSR2, upchild);
    signal(SIGCHLD, upchild);

   for ( ; ; ) {
	/*===================================================================
	 * See if we need to up the number of processes.
	 * Use while() in case another SIGCHLD is received in do_upchild().
	 * We want to call do_upchild() at least once, though.
	 *==================================================================*/
	do {
	    sleep(wgconfp->checkprocs);
	    received_sigchld = 0;
	    do_upchild();
	} while (received_sigchld);
    }
}
