/*============================================================================
 * filename:		libsrc/cbc_io.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*====================================================================== 
 *
 * (C) COPYRIGHT R. James Livermore 1998-2000.  All rights reserved.
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1998-2000.
 * All rights reserved.
 *
 * Date written:	May        1, 1998
 * Last modified:	February   8, 2000
 *	
 * functions:
 *		cbc_read()
 *		cbc_write()
 * 
 *=====================================================================*/  
static	char	cprght[] = 
  		"(C) COPYRIGHT Freemont Avenue Software, Inc. 1998-2000 cbc_io,"
  		" V 1.0 00/01/21"; 

#ifndef MSDOS
# ifndef __hppa
#  include <sys/select.h>
# else
#  include <netinet/in.h>
# endif
# include <syslog.h>
#else
# include <dos.h>
# include <io.h>
# include <winsock.h>
# include <sys/stat.h>
#endif
#include <errno.h>
#include <limits.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "firewall.h"
#include "tnkey.h"
#include "des.h"
#include "cbc_io.h"
#ifndef MSDOS
# include "global.h"
# include "md5.h"
#endif

#ifndef	linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif

#ifdef TREX_ENC

/*=======================================================================
 * cbc_read()
 *
 * Read cbc encrypted data
 *
 * input:
 *	fd	file descriptor to read from
 *	buf	pointer to data buffer
 *	len	number of bytes to read
 *	keysched des_key_schedule to use
 *	cbcp	pointer to an CBCBLK structure to hold buffers for read/write
 *
 * Note: before calling cbc_read() to read the first block in a file or a
 *	socket, initialize cbcp->ivec to IVEC (defined in cbc_io.h).  After the
 *	first call, ivec is maintained by cbc_read().
 *	Also, initialize cbcp->flag to CBC_BIN for binary files or
 *	to CBC_ASCII for text files.
 *
 * output:
 *	updated structure pointed to by cbcp
 *	returns the number of bytes (clear text) that were put into the
 *		buf buffer.
 *	returns -1 if failed.
 *======================================================================*/
int cbc_read(int fd, unsigned char *buf, int len, des_key_schedule keysched,
		CBCBLK *cbcp)
{
    unsigned char		ebuf[BUFSIZ];
    unsigned char		*tb,tbuf[sizeof(des_cblock)];
    int				n=0,nread=0,n8,readlen,done_read;
    unsigned char		block_len;
#ifdef MSDOS
    struct stat                 stat;
    int isSocket = 1;
    /*=======================================================================
     * stat the fd to find out if it is a socket or a regular file.
     * Only need to do this on Windows since read() on unix works on both
     * files and sockets.
     *======================================================================*/
    if (fstat(fd, &stat) < 0) 
    {
        if (errno != EBADF) 
	{
            printf("Cannot stat the file descriptor.  %s\n", 
	    		sys_errlist[errno]);
            return(-1);
        }
    }
    else if (S_ISREG(stat.st_mode))
        isSocket = 0;
#endif

/*
    {
	char *p = (char *) keysched;
	fprintf(stderr,"cbcread 0x%02x%02x%02x%02x%02x%02x%02x%02x\n",
		*p, *(p+1),*(p+2),*(p+3),*(p+4),*(p+5),*(p+6),*(p+7));
    }
*/

    /*=======================================================================
     * Make sure we can handle the data length.
     *======================================================================*/
    if (len > sizeof(ebuf))
    {
	printf("Buffer limit exceeded trying to read %d bytes.\n",len);
#ifndef MSDOS
	syslog(LOG_ERR,"Error: Buffer limit exceeded trying to read %d bytes.",
			len);
#endif
	return(-1);
    }

    /*=======================================================================
     * If there are left over bytes to hand back to caller, use them first.
     *======================================================================*/
    if (cbcp->left_over > 0)
    {
	if (cbcp->left_over < len)
	{
	    memcpy(buf,cbcp->prev_buf,cbcp->left_over);
	    n = cbcp->left_over;
	    cbcp->left_over = 0;
	}
	else
	{
	    /*===============================================================
	     * We have more left_over bytes than are asked for, just take
	     * out what we need and return immediately.
	     *==============================================================*/
	    memcpy(buf,cbcp->prev_buf,len);
	    memmove(cbcp->prev_buf,cbcp->prev_buf+len,cbcp->left_over-len);
	    cbcp->left_over -= len;
	    return(len);
	}
    }

    /*=======================================================================
     * If the last read contained a a block that's shorter than
     * sizeof(des_cblock) long, use that short block.
     *======================================================================*/
    if (cbcp->nshort > 0)
    {
	memcpy(ebuf,cbcp->shortblk,cbcp->nshort);
    }

    /*=======================================================================
     * Read in enough to satisfy the read request
     *======================================================================*/
    readlen = ( (len - n + sizeof(des_cblock) - 1) / sizeof(des_cblock) )
			* sizeof(des_cblock) - cbcp->nshort;
    done_read = 0;
    do
    {
#ifdef MSDOS
        if (!isSocket)
	    nread=read(fd,&ebuf[cbcp->nshort+done_read],readlen-done_read);
	else
	    nread=recv(fd,&ebuf[cbcp->nshort+done_read],readlen-done_read,0);

   if (nread < 0 && errno != EAGAIN)
#else
	if ((nread=read(fd,&ebuf[cbcp->nshort+done_read],readlen-done_read)) < 0
		&& errno != EAGAIN)
#endif /* MSDOS */
	{
	    printf("Encrypted read from fd %d failed, %s.\n",
			fd,sys_errlist[errno]);
#ifndef MSDOS
	    syslog(LOG_ERR,
		"Error: Encrypted read from fd %d failed, %s.\n",
		fd,sys_errlist[errno]);
#endif
	    return(-1);
	}
	if (nread > 0)
	   done_read += nread;
    }
    while (nread != 0 && cbcp->nshort+done_read < sizeof(des_cblock));

    nread = done_read;

    /*=======================================================================
     * Pretend that the short block from the last round was read in just now.
     *======================================================================*/
    nread = nread + cbcp->nshort;
    cbcp->nshort = 0;

    if (nread == 0)
	return(0);

    /*=======================================================================
     * Decrypt the buffer.
     *======================================================================*/
    n8 = 0;
    while (n8 < nread)
    {

	/*===================================================================
	 * Make sure data is sizeof(des_cblock)-byte multiples.
	 * If it isn't, save this short block and process it next time.
	 *==================================================================*/
	if (nread-n8 < sizeof(des_cblock))
	{
	    cbcp->nshort = nread-n8;
	    memcpy(cbcp->shortblk,&ebuf[n8],cbcp->nshort);
	    break;
	}

	des_cbc_encrypt((des_cblock *)&ebuf[n8],(des_cblock *)&tbuf,
			sizeof(des_cblock),
			keysched,&(cbcp->ivec),DES_DECRYPT);
	n8 += sizeof(des_cblock);

	/*===================================================================
	 * If ascii file, see if the first byte is a length byte
	 *==================================================================*/
	if ((cbcp->flag & CBC_BIN) == 0 &&
		tbuf[0] > LEN_BYTE_MASK && tbuf[0] <= (LEN_BYTE_MASK+0x07))
	{
	    block_len = tbuf[0] & ~LEN_BYTE_MASK; /* Remove len byte ind */
	    tb = &tbuf[1];
	}
	else
	{
	    block_len = sizeof(des_cblock);
	    tb = tbuf;
	}

	/*===================================================================
	 * Make sure we do not overflow cbuf.  If it is already full, start
	 * putting clear text in prev_buf.
	 *==================================================================*/
	if (n+block_len > len)
	{
	    memcpy(cbcp->prev_buf+cbcp->left_over,tb+len-n,block_len-(len-n));
	    cbcp->left_over += block_len-(len-n);

	    if (n < len)
	    {
		memcpy(buf+n,tb,len-n);
		n = len;
	    }
	}
	else
	{
	    memcpy(buf+n,tb,block_len);
	    n += block_len;
	}

    }

    return(n);

}

/*=======================================================================
 * cbc_write()
 *
 * Encrypt and write data
 *
 * input:
 *	fd	file descriptor to write to
 *	buf	pointer to data
 *	len	number of bytes to write
 *	keysched des_key_schedule to use
 *	cbcp	pointer to an CBCBLK structure to hold buffers for read/write
 *
 * Note: before calling cbc_write() to write the first block in a file or a
 *	socket, initialize ivec to IVEC (defined in cbc_io.h).  After the
 *	first call, ivec is maintained by cbc_write().
 *	Also, initialize cbcp->flag to CBC_BIN for binary files or
 *	to CBC_ASCII for text files.
 *
 * output:
 *	returns the number of bytes processed.  So if everything is OK, it
 *		always returns len.
 *	returns -1 if failed.
 *======================================================================*/
int cbc_write(int fd, unsigned char *buf, int len, des_key_schedule keysched,
			CBCBLK *cbcp)
{
    des_cblock		random_bits;
    unsigned char	ebuf[2*BUFSIZ];
    unsigned char	tbuf[sizeof(des_cblock)];
    int			i,nwrite;
    int			bytes_left=0;
    int			written=0;
#ifndef MSDOS
    fd_set		rdy;
    struct timeval	*to=(struct timeval *)0;
#endif
#ifdef MSDOS
    struct stat			stat;
    int isSocket;

    /*=======================================================================
     * stat the fd to find out if it is a socket or a regular file.
     * Only need to do this on Windows since read() on unix works on both
     * files and sockets.
     *======================================================================*/
    if (fstat(fd, &stat) < 0) {
        if (errno != EBADF) {
            printf("Cannot stat the file descriptor.  %s\n", sys_errlist[errno]);
            return(-1);
        }
    }
    else if (S_ISREG(stat.st_mode))
        isSocket = 0;
#endif

/*
    {
	char *p = (char *) keysched;
	fprintf(stderr,"cbcwrite 0x%02x%02x%02x%02x%02x%02x%02x%02x\n",
		*p, *(p+1),*(p+2),*(p+3),*(p+4),*(p+5),*(p+6),*(p+7));
    }
*/

    /*=======================================================================
     * Make sure we can handle the data length.
     *======================================================================*/
    if (len > sizeof(ebuf))
    {
	printf("Buffer limit exceeded trying to write %d bytes.\n",len);
#ifndef MSDOS
	syslog(LOG_ERR,"Error: Buffer limit exceeded trying to write %d bytes.\n",len);
#endif
	return(-1);
    }

    /*=======================================================================
     * Make sure data is sizeof(des_cblock)-byte multiples.
     *======================================================================*/
    for (i=0; i<(len+sizeof(des_cblock)-1)/sizeof(des_cblock); i++)
    {
	/*====================================================================
	 * If ascii file, see if this is a block that is less than
	 * sizeof(des_cblock) bytes long
	 *===================================================================*/
	if ((cbcp->flag & CBC_BIN) == 0 &&
		i*sizeof(des_cblock)+8 > len)
	{
	    unsigned char block_len;

	    /*================================================================
	     * Put some random bits into the buffer
	     *===============================================================*/
	    des_random_key(tbuf);
	    /*================================================================
	     * Put the length byte in the buffer.  Set MSB to 1 to indicate
	     * it is a length byte.
	     *===============================================================*/
	    block_len = len - i*sizeof(des_cblock);
	    tbuf[0] = block_len | LEN_BYTE_MASK;
	    memcpy(&tbuf[1],&buf[i*sizeof(des_cblock)],block_len);
	}
	else
	    memcpy(tbuf,&buf[i*sizeof(des_cblock)],sizeof(des_cblock));

	des_cbc_encrypt((des_cblock *)tbuf,
			(des_cblock *)&ebuf[i*sizeof(des_cblock)],
			sizeof(des_cblock),
			keysched,&(cbcp->ivec),DES_ENCRYPT);
    }

    /*=======================================================================
     * Write the encrypted data.
     * Make sure fd is ready to be written to before writing.
     *======================================================================*/
    errno = 0;
    bytes_left = i*sizeof(des_cblock);
    while (bytes_left > 0)
    {
int rc;
#ifndef MSDOS
	FD_ZERO(&rdy);
	FD_SET(fd,&rdy);
#ifndef __hppa
	if((rc = select(fd + 1,(fd_set *)0,&rdy,(fd_set *)0,to)) < 0)
#else
	if((rc = select(fd + 1,0,(int *)&rdy,0,to)) < 0)
#endif
	{
	    syslog(LOG_ERR,"Error: Select failed for encrypted write, "
			"fd=%d, %s.",
			fd,sys_errlist[errno]);
	    return(-1);
	}

	if ((nwrite=write(fd,&ebuf[written],bytes_left)) != bytes_left
		&& errno != EAGAIN && errno != EPIPE && errno !=0)
#else
        if (!isSocket)
	    nwrite=write(fd,&ebuf[written],bytes_left);
	else
	    nwrite=send(fd,&ebuf[written],bytes_left,0);

	if (nwrite != bytes_left
		&& errno != EAGAIN && errno != EPIPE && errno !=0)
#endif /* MSDOS */
	{
	    printf("Encryptyed write to fd %d failed, nwrite=%d, %s.\n",
			fd,nwrite,sys_errlist[errno]);
#ifndef MSDOS
	    syslog(LOG_ERR,"Error: Encryptyed write to fd %d failed, "
			"nwrite=%d, %s.\n",
			fd,nwrite,sys_errlist[errno]);
#endif
	    return(-1);
	}

	if (nwrite < 0)
	    return(-1);

	if (nwrite > 0)
	{
	    bytes_left -= nwrite;
	    written += nwrite;
	}
    }

    /*=======================================================================
     * If everything is OK, return the length of the clear text, not the
     * length of the cypher text that waas actually written.
     *======================================================================*/
    return (len);
}
#endif  /* TREX_ENC */
