/*============================================================================
 * filename:		padmin/cs.h
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * written by:		R. James Livermore 
 *
 * (C) COPYRIGHT R. James Livermore 1998-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1998-2000.  
 * All rights reserved.   
 *
 * Date written:	June	  24, 1998
 * Last modified:	February   8, 2000 - added LPL
 *
 *===========================================================================*/
#ifdef MSDOS
#include "tn.h"
#undef CALLBACK
#undef FAILED
#endif
#include "firewall.h"

#define multiple8(a)	((a + 7) >> 3) << 3

#define DEFAULT_PADMIN_PORT	741
#define LISTDIR	FW_CFG_DIR "/lists"
#define PROGDIR	"/usr/local/etc"
#define PIDFILE	FW_CFG_DIR "/padmin.pid"

/*======================================================================
 * The default key used to encrypt the data before a use's key is
 * established.  Just some random bits.
 *=====================================================================*/
#define KEY1 0x6a473d21
#define KEY2 0x9e77c56a

/*======================================================================
 * The defines for flag values in gui/data.c.  These are for client
 * side code only, not for the server
 *=====================================================================*/
#define RESTART   0x1	/* True = restart interrupted blocked read/write*/
#define NOTIMEOUT 0x2	/* True = don't timeout a client select call	*/

/*======================================================================
 * The defines for flag values for backup.
 *=====================================================================*/
#define TYPICAL_BU	1
#define CUSTOM_BU	2

/*======================================================================
 * The defines for the gwuser/gwgroup requests
 *=====================================================================*/
#define ADDUSER		0
#define CLONEUSER	1
#define	MODUSER		2
#define ADDGROUP	3
#define CLONEGROUP	4
#define	MODGROUP	5

/*===========================================================================
 * Be sure to keep these define values the same as their counterparts in
 * cglue.java.
 *==========================================================================*/

struct cs_req {
    unsigned long component;		/* Target component for request	*/
	#define CALL		0
	#define CAPACHE		1
	#define CAPACHEERR	2
	#define	CFTPROXY	3
	#define	CFWPULSE	4
	#define	CFWPULSED	5
	#define	CAPROXY		6
	#define	CGWUSER		7
	#define	CGWGROUP	8
	#define	CHTTPD		9
	#define	CINETD		10
	#define CLOGIN		11
	#define	CPADMIN		12
	#define	CFWMON		13
	#define	CPS		14
	#define	CRAPROXY	15
	#define	CRPCPROXY	16
	#define	CSMALIASES	17
	#define	CSMWRAP		18
	#define	CSMWRAPD	19
	#define	CTNKEY		20
	#define	CTNPROXY	21
	#define CSECNETS	22
	#define CSECPORTS	23
	#define	CSOCKD		24
	#define CSYSLOG		25
	#define CTIMER		26
	#define CVMSTAT		27
	#define	CWEBGATE	28
	#define CTREXLIST	29
	#define CAPACHELIST	30
	#define CSYSTEMLIST	31
	#define	CCUSTOMLIST	32
					/* DO NOT define any components	*/
					/* that have values between	*/
					/* 0x81xxxxxx and 0x87xxxxxx	*/
					/* The enc_io.c read and write	*/
					/* functions interprets and 0x81*/
					/* in the 1st of every 8 bytes	*/
					/* as a length byte		*/
	#define NCOMP		CCUSTOMLIST/* Total number of components*/

    unsigned long cmd;			/* The requested action		*/
	#define	GETCONF		1	/* Get the config info		*/
					/* no msg in request		*/
	#define PUTCONF		2	/* Update the config info	*/
					/* no msg in request		*/
	#define STOPCOMP	3	/* Stop the component		*/
					/* no msg in request		*/
	#define STARTCOMP	4	/* Start the component		*/
					/* no msg in request		*/
	#define GETGWUSER	5	/* Get 1 gwuser or gwgroup	*/
					/* msg contains user/group name	*/
	#define ADDGWUSER	6	/* Add 1 gwuser or gwgroup	*/
					/* msg contains struct GWUSER	*/
					/* which has user/group info	*/
	#define MODGWUSER	7	/* Modify 1 gwuser or gwgroup	*/
					/* msg contains struct GWUSER	*/
					/* which has user/group info	*/
	#define RUN		8	/* Display results of a command	*/
					/* msg contains argument to cmd	*/
	#define CHALLENGE	9	/* Generate challenge to usernm	*/
					/* msg is username		*/
	#define AUTH		10	/* Auth response to a challenge	*/
					/* msg is response to challenge	*/
	#define CANCEL		11	/* Cancel last outstanding req	*/
					/* no msg in request		*/
	#define KEEPALIVE	12	/* Extend for another timeout	*/
					/* no msg in request		*/
	#define CALLBACK	13	/* Call client at specified port*/
					/* msg is port			*/
	#define PROXY		14	/* run as proxy to another fw	*/
					/* msg is remote ip, port	*/
	#define TAILLOG		15	/* tail -f a log file		*/
					/* msg is arg string e.g.	*/
					/* "| grep "Alerts"		*/
	#define PAUSE		16	/* pause the current activity	*/
					/* for TAILLOG			*/
					/* no msg in request		*/
	#define RESUME		17	/* resume a paused activity	*/
					/* for TAILLOG			*/
					/* no msg in request		*/
	#define GETDEFGWUSER	18	/* Get default gwuser/gwgroup	*/
					/* parameters from gwuser.conf	*/
					/* no msg in request		*/
	#define DELGWUSER	19	/* Delete gwuser/gwgroup	*/
					/* msg is user/group name	*/
	#define REFCOMP		20	/* Refresh a component		*/
					/* no msg in request		*/
	#define GETLIST		21	/* Get dir list of		*/
					/* /etc/firewall/lists or	*/
					/* /usr/local/etc		*/
					/* no msg in request		*/
	#define BACKUP		22	/* Do tar with selected lists	*/
					/* msg is char rep of bit flag	*/
					/* 0001 = typical backup        */
					/* 0002 = customized backup	*/
					/* 0001 | 0002 = both		*/
	#define RESTORE		23	/* Restore file to server	*/
					/* msg contains selected files	*/
					/* to restore.  If blank,	*/
					/* restore all			*/
	#define GETIFLIST	24	/* Get a list of all the local	*/
    					/* IP addresses			*/
    unsigned long msglen;		/* Length of following struct	*/
					/* including check.  This is	*/
					/* length before encryption	*/
};

typedef struct cs_req CSREQ;

struct sc_resp {
    long rc;				/* Return code			*/
	#define	LOCALFILE	2	/* File is local, not downloaded*/
	#define	CONTINUE	1	/* Continue with next request	*/
					/* At login, CONTINUE means	*/
					/* username accepted, challenge	*/
					/* and response next		*/
					/* At file downloads, CONTINUE	*/
					/* means someone has downloaded */
					/* the same file before, so	*/
					/* client should ask user	*/
	#define	OK		0	/* Accepted/successful		*/
	#define	FAILED		-1	/* fatal error - close connection */
	#define	CLOSED		-2	/* Server closing connection	*/
	#define	NONFATAL	-3	/* non fatal error - try again	*/
    unsigned long msglen;		/* Length of err msg or struct	*/
					/* including check.  This is	*/
					/* length before encryption	*/
};

typedef struct sc_resp SCRESP;

struct resppkt {
    SCRESP hdr;				/* Response header		*/
    char msg[BUFSIZ];
};

typedef struct resppkt RESPPKT;

struct ips {
    unsigned long op;			/* PERMIT or DENY		*/
	#define PERMIT		0
	#define DENY		1
    unsigned long ip;
    unsigned long mask;
};

typedef struct ips IPS;

struct paconf {
    unsigned short port;		/* Port number to use for padmin*/
    unsigned long timeout;		/* Timeout value for padmin	*/
    unsigned long nips;			/* Number of permitted ip's	*/
    IPS hosts[1];			/* Place holder for IPS array	*/
};

typedef struct paconf PACONF;

struct rootuser {
    char username[UNMSIZ];
    unsigned long addr;
    unsigned long mask;
};
typedef struct rootuser rootuser;

#ifndef INADDR_NONE
#define INADDR_NONE		0xffffffff
#endif
