/*============================================================================
 * filename:            hoplite/data.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=============================================================================
 *
 * R. James Livermore grants Freemont Avenue Software, Inc. a nonexclusive
 * license to use, modify, and market this program using their tradename
 * T.Rex.
 *
 *
 * (C) COPYRIGHT R. James Livermore 1997-2000.  All rights reserved.
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1997-1998.
 * All rights reserved.
 *
 * Date written:        September 29, 1997
 * Last modified:	February   8, 2000
 *
 *
 *============================================================================*/ 


#include <stdio.h>
#include <errno.h>
#ifndef MSDOS
#include <sys/time.h>
#ifndef __hppa
#include <sys/select.h>
#endif
#include <syslog.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif
#include "des.h"
#include "cs.h"

#ifndef MSDOS
#ifdef linux
extern const char *const sys_errlist[];
#else
extern  char    *sys_errlist[];
#endif
#endif
int enc_read(int fd, char *buf, int len, des_key_schedule keysched);
int enc_write(int fd, char *buf, int len, des_key_schedule keysched);
extern int cleft_over;
extern int sleft_over;

void order_bytes_in_req(CSREQ *reqp) {
    reqp->component = htonl(reqp->component);
    reqp->cmd = htonl(reqp->cmd);
    reqp->msglen = htonl(reqp->msglen);
}

void order_bytes_in_resp(SCRESP *resp) {
    resp->rc = ntohl(resp->rc);
    resp->msglen = ntohl(resp->msglen);
}
/*=======================================================================
 * This function gets up to len bytes of data from fd.
 *
 * Input:
 *	fd	file descriptor of the input socket
 *	msg	address of the buffer where results are returned
 *	len	number of bytes to get
 *	keysched encryption key to use
 *	flag	RESTART on = read/write syscalls should be restarted if
 *				interrupted
 *		NOTIMEOUT on = Do not timeout this read - for the monitor
 *				function
 * Output:
 *	msg	filled with the expected data if successful
 *	rc 0	EOF
 *	rc n	n bytes read
 *	rc -1	failed
 *======================================================================*/
int getsomedata(int fd, char *buf, int len, des_key_schedule keysched,
			long flag) {
    int rc;
    struct timeval timeout,*tp;
    fd_set rdy;

    FD_ZERO(&rdy);
    FD_SET(fd,&rdy);

    timeout.tv_sec = 60;
    timeout.tv_usec = 0;
    if (flag & NOTIMEOUT)
/*
	tp = (struct timeval *)0;
*/
timeout.tv_sec = 0;
    else
	tp = &timeout;

    /*======================================================================
     * If there are leftover bytes in the enc_io buffer, we don't want to
     * call select.  If we do, we may block unnecessary (perhaps forever)
     * when data is available in the buffer but no more data si coming
     * down the socket.
     *=====================================================================*/
    if ((fd == 0 && cleft_over >= len) || (fd != 0 && sleft_over >= len)) {
#ifdef __hppa
	while ((rc = select(fd+1,(int *)&rdy,0,0,tp)) < 0 && errno == EINTR &&
			(flag & RESTART));
#else
	while ((rc = select(fd+1,&rdy,(fd_set *)0,(fd_set *)0,tp)) < 0 &&
			errno == EINTR && (flag & RESTART));
#endif
	if (rc < 0) {
	    if (errno != EINTR)
#ifndef MSDOS
		syslog(LOG_ERR,"Error: select() failed.  %s",
						sys_errlist[errno]);
#else
		fprintf(stderr,"Error: select() failed.  %s",
						sys_errlist[errno]);
#endif
	    return(-1);
	}

	if (rc == 0) {
if (flag & NOTIMEOUT)
return(CONTINUE);
#ifndef MSDOS
	    syslog(LOG_NOTICE,"Connection terminated.");
#else
	    fprintf(stderr,"Connection terminated.");
#endif
	    return(0);
	}
    }

    while ((rc = enc_read(fd, buf, len, keysched)) < 0 && errno == EINTR &&
				(flag & RESTART));
    if (rc < 0  && errno != EINTR) {
#ifndef MSDOS
	syslog(LOG_ERR,"Error: Read failed.  %s", sys_errlist[errno]);
#else
	fprintf(stderr,"Error: Read failed.  %s", sys_errlist[errno]);
#endif
	return(-1);
    }

    /*===============================================================
     * Set rc for EOF or short read
     *==============================================================*/
    if (rc == 0)
    	return(0);

    /*===============================================================
     * Success!
     *==============================================================*/
    return(rc);
}
/*=======================================================================
 * This function gets exactly len bytes of data from fd.
 *
 * Input:
 *	fd	file descriptor of the input socket
 *	msg	address of the buffer where results are returned
 *	len	number of bytes to get
 *	keysched encryption key to use
 * Output:
 *	msg	filled with the expected data if successful
 *	rc 0	EOF
 *	rc n	n bytes read
 *	rc -1	failed
 *======================================================================*/
int getdata(int fd, char *buf, int len, des_key_schedule keysched,	
				long flag) {
    int rc;
    int n;

    for (n=0; n < len; n += rc) {
	if  ((rc = getsomedata(fd, &buf[n], len-n, keysched,flag)) <= 0)
	    return(rc);
    }
    return(n);
}

/*=======================================================================
 * Note that this function may be called prior to or after authentication
 * so this function does not directly report errors back to the client.
 *
 * Input:
 *	fd	file descriptor of the output socket
 *	msg	address of the buffer where data is to be written
 *	len	number of bytes to put
 *	keysched encryption key to use
 * Output:
 *	rc n	n bytes written
 *	rc -1	failed
 *======================================================================*/
int putdata(int fd, char *buf, int len, des_key_schedule keysched,
					long flag) {
    int rc;

    while ((rc = enc_write(fd, buf, len, keysched)) < 0 && errno == EINTR &&
				flag);
    if (rc <= 0  && errno != EINTR) {
#ifndef MSDOS
	syslog(LOG_ERR,"Error: Write failed.  %s", sys_errlist[errno]);
#else
	fprintf(stderr,"Error: Write failed.  %s", sys_errlist[errno]);
#endif
	return(-1);
    }
    
    if (rc < len) {
#ifndef MSDOS
	syslog(LOG_ERR,"Error: protocol error, attempted to write %d bytes, "
		"%d bytes written");
#else
	fprintf(stderr,"Error: protocol error, attempted to write %d bytes, "
		"%d bytes written");
#endif
	return(-1);
    }

    /*===============================================================
     * Success!
     *==============================================================*/
    return(rc);
}
