/*============================================================================
 * filename:		include/firewall.h
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * written by:		R. James Livermore 
 *
 * R. James Livermore grants Freemont Avenue Software, Inc. a nonexclusive
 * license to use, modify, and market this program using their tradename
 * T.Rex.	
 *
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  All rights
 * reserved.  
 *  
 *		
 * Date written:	March     21, 1995
 * Last modified:	February   8, 2000
 *			March     15, 2000 - added ifndef for socklen_t on
 *					     Solaris 7
 *	                         
 * function:
 *
 *  The firewall header contains function prototypes for T.Rex modules.
 * 
 *
 *===========================================================================*/
#ifndef _FIREWALL_H
#define _FIREWALL_H

#define TREX_VERSION 	"T.Rex Firewall Version 1.0.0"
#define FW_CFG_DIR	"/etc/firewall"
#define ALIASESFILE     FW_CFG_DIR "/aliases" 
#define BLOCKEDSENDERSFILE "blockedsenders"
#define NEWALIASESFILE  FW_CFG_DIR "/newaliases"
#define FTPROXCFGFILE   FW_CFG_DIR "/ftproxy.conf" 
#define GENPROXCFGFILE 	FW_CFG_DIR "/genproxy.conf" 
#define GWUSERFILE	FW_CFG_DIR "/gwuser"
#define GWUSERCFGFILE	FW_CFG_DIR "/gwuser.conf"
#define HTTPPROXCFGFILE FW_CFG_DIR "/http.conf"
#define LOGINCFGFILE    "/etc/security/login.cfg"
#define PRIVATENETFILE  FW_CFG_DIR "/privatenets"
#define PRIVATEPORTFILE FW_CFG_DIR "/privateports" 
#define RAPROXCFGFILE   FW_CFG_DIR "/raproxy.conf" 
#define SECURENETFILE   FW_CFG_DIR "/securenets"  
#define SECUREPORTFILE  FW_CFG_DIR "/secureports" 
#define SMWRAPCFGFILE   FW_CFG_DIR "/smwrap.conf" 
#define TNPROXCFGFILE   FW_CFG_DIR "/tnproxy.conf" 
#define XFORWARDPGM	"/usr/local/etc/xforward" 
#define RESTRICTEDSHELL "/usr/local/etc/restrict.sh" 
#define FWMONPIDS   FW_CFG_DIR "/fwmon.pids" 
#define SYNMONPID	FW_CFG_DIR "/synmon.pid" 
#ifdef MSDOS
#define TNKEYFILE	"trex.key"
#else
#define TNKEYFILE	".trex.key"	/* either in $HOME or /etc        */
#endif
#define PMONCONFFILE	FW_CFG_DIR "/fwmon.conf"
#define SYNMON		"/usr/local/etc/synmon"
#define SYNMONLD	"/usr/local/etc/synmonld"
#define SYNMONKX	"/usr/local/etc/synmonkx"
#define SPOOFMON	"/usr/local/etc/spoofmon"
#define PROCMON		"/usr/local/etc/procmon"
#define WEBGWCFGFILE 	FW_CFG_DIR "/webgate.conf" 
#define DNSOFF		0
#define DNSON		1
#define IPSIZ 		20 
#define UNMSIZ		17             	/* T.Rex user's name size	  */
#define GRPSIZ		17		/* T.Rex user's group name size  */
#define	LOGNMSIZ	128		/* HTTP log name size		  */
#define PWDSIZ		17		/* T.Rex user's password size    */ 
#define AUTHREQ 	1               /* Authorization required	  */
#define NOAUTHREQ   	0               /* No authorzation required       */
#define EXFTP		1		/* Allow External FTP		  */
#define NOEXFTP		0		/* No External FTP allowed	  */
#define MAXHOSTLN 	255		/* Maximum host name length       */
#define MAXPATHLN	255		/* Max Path Length for firewall   */
#define MAXRECLN	1024		/* Max record length (conf files) */
#define MAXTRIES	4		/* maximum login attempts         */
#define SECURED		1		/* host is secured		  */
#define NOTSECURED      0		/* host is not secured	 	  */ 
#define X4WARDYES	1		/* allow xforwarding		  */
#define X4WARDNO	0               /* Do NOT allow xforwarding	  */
/*==========================================================================
 * Function protoypes
 * modules found in libsrc	
 *=========================================================================*/ 
int connectserver(char *serv,int portnum, int priv, char *localIP); 
void daemonize();
int getaliases(char *keyptr, char *dataptr, char *dbname);
int ipsecure(char *IPsecure);		/* returns 1 if secured	*/ 
int isalldigits(char *s);	/* returns 1 if all digits 	*/ 
int peername(int fd, char *lname,char *sname,int z);
int PeerData(int fd, char *lname,char *sname,int z,int DNSflag); 
/*==========================================================================
 * typedefs 
 *=========================================================================*/ 
#ifdef _AIX
#ifndef _AIX43
typedef size_t socklen_t;
#endif
typedef size_t addrlen_t;
#endif
#ifdef __sun
#ifndef _SOCKLEN_T
typedef int socklen_t;
#endif
typedef int addrlen_t;
#endif
#ifdef linux
typedef int addrlen_t;
#endif
#ifdef MSDOS
typedef int socklen_t;
typedef int addrlen_t;
#endif
/*===========================================================================
 * This typedef is needed for hpux and AIX 3.2 but not AIX 4.1
 *==========================================================================*/
#if !defined(_AIX41) && !defined(__sun)
typedef unsigned long   in_addr_t;
#endif
#endif /* _FIREWALL_H */
