/*============================================================================
 * filename:            hoplite/jc.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=============================================================================
 *
 * R. James Livermore grants Freemont Avenue Software, Inc. a nonexclusive
 * license to use, modify, and market this program using their tradename
 * T.Rex.
 *
 *
 * (C) COPYRIGHT R. James Livermore 1997-2000.  All rights reserved.
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1997-2000.
 * All rights reserved.
 *
 * Date written:        September 29, 1997
 * Last modified:	February   8, 2000
 *
 *============================================================================*/

/* #include <StubPreamble.h> */
#ifndef MSDOS
#ifndef __hppa
#include <sys/select.h>
#endif
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/shm.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#else
typedef double __int64;
#include <windef.h>
#include <win32\winnt.h>
#include <tn.h>
#include <io.h>
#endif
#ifndef MSDOS
#include <tar.h>
#else
#define TMAGIC "ustar"
#define TMAGLEN 6
#endif
#include "cglue.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#ifdef linux
#include <sys/param.h>
#define MAXADDRS       35
#define MAXALIASES     35
#endif
#include "pktinfo.h"
#include "firewall.h"
#include "gwuser.h"
#include "des.h"
#include "cs.h"
#include "tnkey.h"
#include "cbc_io.h"

#define CGLUE_INT(id)	(*env)->GetIntField(env, jo, id)
#define DESKEY(ksid)	(struct des_ks_struct *)(CGLUE_INT(ksid))
#define SET_CGLUE_INT(id,value)	(*env)->SetIntField(env, jo, id, value)
#define java_errmsg(a)	{	jstring _jstr; \
				_jstr = (*env)->NewStringUTF(env,a); \
				(*env)->SetObjectField(env,pktrc, \
							pktemsgID,_jstr); \
			}
#define java_syserrmsg(a) {	char _tmp[BUFSIZ]; \
				jstring _jstr; \
				sprintf(_tmp,a "  %s",sys_errlist[errno]); \
				_jstr = (*env)->NewStringUTF(env,_tmp); \
				(*env)->SetObjectField(env,pktrc, \
							pktemsgID,_jstr); \
			  }
#define java_syserrmsg2(a,b) {	char _tmp[BUFSIZ]; \
				jstring _jstr; \
				sprintf(_tmp,a "  %s.  %s", \
					sys_errlist[errno],b); \
				_jstr = (*env)->NewStringUTF(env,_tmp); \
				(*env)->SetObjectField(env,pktrc, \
							pktemsgID,_jstr); \
			  }
#define returnrc(a)	{	(*env)->SetIntField(env,pktrc,pktrcID,a); \
				return(pktrc); \
			}
#define IS_NUMERIC_IP(a)	(strspn(a,"0123456789.") == strlen(a))
#define DOSTMPDIR "\\firewall\\tmp"
#define UNIXTMPDIR "/usr/local/etc/tmp"

#ifndef MSDOS
#ifndef linux
extern char *sys_errlist[];
#endif
#endif
int getsomedata(int fd, char *buf, int len, des_key_schedule keysched);
int getdata(int fd, char *buf, int len, des_key_schedule keysched,long flag);
int putdata(int fd, char *buf, int len, des_key_schedule keysched,long flag);
int savelocal(int fd, int component, char *fn);
int fwext(int component, char *ext);
int gettnkey(char *username, des_cblock *key);
int cbc_read(int fd, unsigned char *buf, int len, des_key_schedule keysched,
                CBCBLK *cbcp);
int cbc_write(int fd, unsigned char *buf, int len, des_key_schedule keysched,
                        CBCBLK *cbcp);
void order_bytes_in_req(CSREQ *req);
void order_bytes_in_resp(SCRESP *resp);

/*=======================================================================
 * These static fields are used to hold the references to the 2 class objects
 * that are referenced in this file.
 *======================================================================*/
static jclass pktinfoClass;
static jclass cglueClass;

/*=======================================================================
 * This holds the method ID for the pktinfo class constructor.  It can be
 * static because it is the same for all instances.
 *======================================================================*/
static jmethodID pktinfoID;

/*=======================================================================
 * These field IDs are used to address member variables in the pktinfo
 * class.  They can be static because they are the same for all instances
 * of the same class.  They are initialized once in openfd (for each
 * instance of fwwin) but the values should be the same each time.
 *======================================================================*/
static jfieldID pktrcID;
static jfieldID pktstrID;
static jfieldID pktemsgID;

/*=======================================================================
 * These field IDs are used to address member variables in the cglue
 * class.  They can be static because they are the same for all instances
 * of the same class.  They are initialized once in openfd (for each
 * instance of fwwin) but the values should be the same each time.
 *======================================================================*/
static jfieldID EintrID;
static jfieldID ErrnoID;
static jfieldID fdID;
static jfieldID cbcblkpID;
static jfieldID ckeyschedpID;
static jfieldID dkeyschedpID;
static jfieldID passed_authID;
static jfieldID localfnID;

/*=======================================================================
 * This WIN32 only variable is used to indicate whether WSAStartup() has
 * been called to initialize the WIN 32 socket system.
 *======================================================================*/
#ifdef MSDOS
static int doneWSAinit = 0;
#endif

/*===========================================================================
 * This function is intended to be called from java to open a socket to the
 * padmin server.
 *
 * Input:
 *	jip	java String containing the remote hostname or IP address
 *
 * Output:
 *	0	successful
 *	-1	failed
 *		If the return value is negative, the errmsg member contains
 *		the error message
 * Note that the return value is a reference to a pktinfo instance.  In order to
 * access the pktinfo class from c, NewObject() will have to be used to obtain
 * a new instance of a class pktinfo object.  pktinfo contains: 
 * int rc	- is the integer return code
 * String str	- java string used by recvpkt to hold the data packet received
 *		  or used by getconf to hold the name of the temp file created
 * String errmsg- java string which hold an error message if rc is negative
 * These members can be accessed using the static field IDs initialized by
 * openfd.
 *
 *==========================================================================*/
JNIEXPORT jobject JNICALL Java_cglue_openfd(JNIEnv *env, jobject jo,
		jstring jip) {

    char ip[BUFSIZ];
    char buf[BUFSIZ];
    int one = 1;
    struct sockaddr_in addr;
    SCRESP resp;
    const char *ipp;	/* Temp variable to hold the ip returned from JNI */
    jboolean isCopy;	/* Needed for GetStringUTFChars but not used here */
    jobject pktrc;

    /*======================================================================
     * WIN32 needs to have the socket system initialized before any socket
     * calls will work
     *=====================================================================*/
#ifdef MSDOS
    if (!doneWSAinit) {
	int rc;
	WSADATA wsadata;
	if ((rc = WSAStartup(MAKEWORD(1,1), &wsadata)) != 0) {
	    java_errmsg("WSAStartup failed");
	    returnrc(-1);
	}
	doneWSAinit = 1;
    }
#endif

    /*======================================================================
     * Initialize the pktinfo class objects that are needed in this file
     * 1. Obtain a jclass reference to the pktinfo class AND make it
     *    a globalref so it doesn't disappear between calls
     * 2. Obtain all the field IDs in pktinfo that we need to reference
     * 3. Obtain the method ID for the pktinfo class constructor
     * Note that we have to exit in some cases instead of returning.  For
     * example, if it failed to obtain a pktinfo class, we can't set a
     * return value so we can't return.
     *=====================================================================*/
    pktinfoClass = (*env)->NewGlobalRef(env, (*env)->FindClass(env, "pktinfo"));
    if (pktinfoClass == NULL) {
	fprintf(stderr,"Could not find class pktinfo");
	exit(-1);
    }

    pktrcID = (*env)->GetFieldID(env, pktinfoClass, "rc", "I");
    pktstrID = (*env)->GetFieldID(env, pktinfoClass, "str",
					"Ljava/lang/String;");
    pktemsgID = (*env)->GetFieldID(env, pktinfoClass, "errmsg",
					"Ljava/lang/String;");

    pktinfoID = (*env)->GetMethodID(env, pktinfoClass, "<init>","()V");
    if (pktinfoID == NULL) {
	fprintf(stderr,"Could not find constructor for pktinfo");
	exit(-1);
    }

    /*======================================================================
     * Establish addressability to the pktrc structure
     *=====================================================================*/
    pktrc = (*env)->NewObject(env, pktinfoClass, pktinfoID);
    if (pktrc == NULL) {
	fprintf(stderr,"pktinfo constructor failed");
	exit(-1);
    }

    /*======================================================================
     * Initialize the cglue class objects that are needed in this file
     * 1. Obtain a jclass reference to the cglue class AND make it
     *    a globalref so it doesn't disappear between calls
     * 2. Obtain all the field IDs in cglue that we need to reference
     *=====================================================================*/
    cglueClass = (*env)->NewGlobalRef(env, (*env)->GetObjectClass(env,jo));
    if (cglueClass == NULL) {
	java_errmsg("Could not find class cglue");
	returnrc(-1);
    }

    fdID = (*env)->GetFieldID(env, cglueClass, "fd", "I");
    cbcblkpID = (*env)->GetFieldID(env, cglueClass, "cbcblkp", "I");
    ckeyschedpID = (*env)->GetFieldID(env, cglueClass, "ckeyschedp", "I");
    dkeyschedpID = (*env)->GetFieldID(env, cglueClass, "dkeyschedp", "I");
    EintrID = (*env)->GetFieldID(env, cglueClass, "Eintr", "I");
    ErrnoID = (*env)->GetFieldID(env, cglueClass, "Errno", "I");
    passed_authID = (*env)->GetFieldID(env, cglueClass, "passed_auth", "I");
    localfnID = (*env)->GetFieldID(env, cglueClass, "localfn",
						"Ljava/lang/String;");

    /*=======================================================================
     * Set up a constant so it can be referenced in java
     *======================================================================*/
    SET_CGLUE_INT(EintrID, EINTR);

    /*=======================================================================
     * Allocate the key schedules and initialize the default one
     *======================================================================*/
    if (CGLUE_INT(ckeyschedpID) == 0) {
	int tmp;
	if ((tmp = (int) malloc(sizeof(des_key_schedule))) == 0) {
	    java_syserrmsg("Cannot malloc.");
	    returnrc(-1);
	}
	SET_CGLUE_INT(ckeyschedpID, tmp);
    }
    if (CGLUE_INT(dkeyschedpID) == 0) {
	int tmp;
	unsigned long tkey[2];

	if ((tmp = (int) malloc(sizeof(des_key_schedule))) == 0) {
	    java_syserrmsg("Cannot malloc.");
	    returnrc(-1);
	}
	SET_CGLUE_INT(dkeyschedpID, tmp);
	tkey[0] = htonl(KEY1);
	tkey[1] = htonl(KEY2);
	des_set_key((des_cblock *)tkey, DESKEY(dkeyschedpID));
    }

    /*=======================================================================
     * Make sure the hostname/ip fits into our buffer
     *======================================================================*/
    if ((*env)->GetStringUTFLength(env, jip) > sizeof(ip)) {
	java_errmsg("Remote host name too long");
	returnrc(-1);
    }
    /*=======================================================================
     * Convert to c string
     * The temp ipp is used to hold the char pointer so that it can be
     * released.
     *======================================================================*/
    ipp = (*env)->GetStringUTFChars(env, jip, &isCopy);
    strcpy(ip, ipp);
    (*env)->ReleaseStringUTFChars(env, jip, ipp);

    if (CGLUE_INT(fdID) < 0) {
	int tmpfd;
	char *portnum;
	/*===================================================================
	 * Create a socket and connect to the server if fd is -1
	 * i.e. we are not already talking to a server and so don't have to
	 * worry about requesting proxy mode.
	 *==================================================================*/
	if ((tmpfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	    java_syserrmsg("Error creating socket");
	    returnrc(-1);
	}
	SET_CGLUE_INT(fdID,tmpfd);

	if (setsockopt(tmpfd,SOL_SOCKET,SO_REUSEADDR,
			(char *)&one,sizeof(int)) < 0)
	    java_syserrmsg("setsockopt error");
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = INADDR_ANY;
	addr.sin_addr.s_addr = INADDR_ANY;
	if (bind(tmpfd, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
	    java_syserrmsg("Bind error");
	    returnrc(-1);
	}

	/*===================================================================
	 * Tokenize ip and an optional port
	 *==================================================================*/
	portnum = strtok(ip,":");
	portnum = strtok(NULL,": ");

	/*===================================================================
	 * If not all numeric, then do a host lookup
	 *==================================================================*/
	
	if (!IS_NUMERIC_IP(ip)) {
#ifndef MSDOS
	    int h_err;
	    struct hostent Hostent;
#ifdef __sun
	    char HostentData[MAXHOSTNAMELEN*MAXALIASES +
			MAXHOSTNAMELEN*MAXADDRS + MAXHOSTNAMELEN];
#else
	    
#ifdef linux
	    char HostentData[MAXHOSTNAMELEN*MAXALIASES +
			MAXHOSTNAMELEN*MAXADDRS + MAXHOSTNAMELEN];
	    struct hostent * HostentResult; 

#else	    
	    struct hostent_data HostentData;
#endif
#endif

	    memset((char *)&Hostent,0,sizeof(Hostent));
	    memset((char *)&HostentData,0,sizeof(HostentData));
#endif

#ifdef __sun
	    if (gethostbyname_r(ip,&Hostent,HostentData,sizeof(HostentData),
			&h_err) == NULL) {
#else
		    
#ifdef linux
	    if (gethostbyname_r(ip,&Hostent,HostentData,sizeof(HostentData),
			&HostentResult, &h_err) < 0) {
#else
		    
#ifdef MSDOS
            HOSTENT *Hostentp;
            if ((Hostentp = gethostbyname(ip)) == NULL) {
#else
	    if (gethostbyname_r(ip,&Hostent,&HostentData) < 0) {
#endif
#endif
#endif
		java_syserrmsg("Host lookup error");
		returnrc(-1);
	    }
#ifndef MSDOS
	    memcpy((char *)&(addr.sin_addr.s_addr), Hostent.h_addr,
                                sizeof(addr.sin_addr.s_addr));
#else
            memcpy((char *)&(addr.sin_addr.s_addr), Hostentp->h_addr,
                                sizeof(addr.sin_addr.s_addr));
#endif            
	}
	else
	    addr.sin_addr.s_addr = inet_addr(ip);

	if (addr.sin_addr.s_addr == INADDR_NONE ||
		addr.sin_addr.s_addr == INADDR_ANY) {
	    java_errmsg("Invalid host IP address");
	    returnrc(-1);
	}

	/*===================================================================
	 * Connect to the server
	 *==================================================================*/
	addr.sin_port = htons(portnum ? atoi(portnum) : DEFAULT_PADMIN_PORT);

	if (connect(CGLUE_INT(fdID), (struct sockaddr *)&addr,
			sizeof(addr)) < 0) {
	    java_syserrmsg("Cannot connect");
	    returnrc(-1);
	}

	/*===================================================================
	 * Get the response packet from the server
	 *==================================================================*/
	if (getdata(CGLUE_INT(fdID),(char *)&resp,sizeof(SCRESP),
			DESKEY(dkeyschedpID),1) <= 0) {
	    java_errmsg("Cannot get response from server");
	    returnrc(-1);
	}
	order_bytes_in_resp(&resp);

	if (resp.rc != OK) {
	    if (resp.msglen > 0 && getdata(CGLUE_INT(fdID), buf,
			resp.msglen,DESKEY(dkeyschedpID),1) > 0) {
		java_errmsg(buf);
	    }
	    else
		java_errmsg("Connection rejected by server");

	    returnrc(-1);
	}

	returnrc(0);
    }
    else {
	/*==================================================================
	 * Since we are already talking with somebody, we need to send a
	 * PROXY request to the server we are talking to and have it
	 * establish a connection to the new server for us.
	 *=================================================================*/
	int rc;
	struct {
	    CSREQ req;
	    char buf[BUFSIZ];
	} pkt;
	SCRESP resp;

	pkt.req.cmd = PROXY;
	pkt.req.component = CLOGIN;
	strcpy(pkt.buf, ip);
	pkt.req.msglen = strlen(pkt.buf) + 1;
	order_bytes_in_req(&(pkt.req));

	/*==================================================================
	 * Send a PROXY request to the server to ask it to connect to the
	 * new server specified in ip
	 *=================================================================*/
	if ((rc = putdata(CGLUE_INT(fdID), (char *)&pkt,
		sizeof(CSREQ) + ntohl(pkt.req.msglen),
				DESKEY(ckeyschedpID),1)) < 0) {
	    java_syserrmsg("Failed sending PROXY request to server");
	    returnrc(rc);
	}

	/*==================================================================
	 * Get the response
	 *=================================================================*/
	if ((rc=getdata(CGLUE_INT(fdID),(char *)&resp,sizeof(SCRESP),
					DESKEY(ckeyschedpID),1)) <= 0) {
	    java_errmsg("Cannot get response from server for PROXY request");
	    returnrc(-1);
	}
	order_bytes_in_resp(&resp);
	if (resp.rc != OK) {
	    char buf[BUFSIZ];
	    if (resp.msglen > 0 &&
			getdata(CGLUE_INT(fdID),buf,resp.msglen,
					DESKEY(ckeyschedpID),1) > 0) {
		java_errmsg(buf);
	    }
	    else
		java_errmsg("Remote server rejected PROXY request");
	    returnrc(resp.rc);
	}

	/*==================================================================
	 * Reset the encryption so it goes back to use the default key
	 * to send login requests
	 *=================================================================*/
	SET_CGLUE_INT(passed_authID, 0);

	returnrc(0);
    }
}


/*===========================================================================
 * This function is intended to be called from java to open a callback socket
 * to the current padmin server.
 *
 * Input:
 *	fd	The caller's current fd.  This is used to send the initial
 *		callback request to the server.  Once the server opened up
 *		the callback socket, this new window frame no longer needs
 *		this fd (It is not closed because the caller still needs it).
 *
 * Output:
 *	0	successful
 *	-1	failed
 *		If the return value is negative, the errmsg member contains
 *		the error message
 * Note that the return value is a reference to a pktinfo instance.  In order to
 * access the pktinfo class from c, NewObject() will have to be used to obtain
 * a new instance of a class pktinfo object.  pktinfo contains: 
 * int rc	- is the integer return code
 * String str	- java string used by recvpkt to hold the data packet received
 *		  or used by getconf to hold the name of the temp file created
 * String errmsg- java string which hold an error message if rc is negative
 * These members can be accessed using the static field IDs initialized by
 * openfd.
 *
 *==========================================================================*/
JNIEXPORT jobject JNICALL Java_cglue_opencb(JNIEnv *env, jobject jo,
			jint fd, jint auth, jint ckeyschedp, jint dkeyschedp) {
    int rc;
    int len;
    int one = 1;
    int newfd;
    int tmpcsched;
    int tmpdsched;
    char urg[4];
    struct sockaddr_in newaddr;
    struct {
	CSREQ req;
	char buf[BUFSIZ];
    } pkt;
    SCRESP resp;
    jobject pktrc;

    /*======================================================================
     * Establish addressability to the pktrc structure
     * Have to exit on error because we don't have the return structure to
     * return with.
     *=====================================================================*/
    pktrc = (*env)->NewObject(env, pktinfoClass, pktinfoID);
    if (pktrc == NULL) {
	fprintf(stderr,"pktinfo constructor failed");
	exit(-1);
    }
    /*=======================================================================
     * Set up a constant so it can be referenced in java
     *======================================================================*/
    SET_CGLUE_INT(EintrID, EINTR);

    /*======================================================================
     * Create a socket
     *=====================================================================*/
    if ((newfd = socket(AF_INET,SOCK_STREAM,0)) < 0) {
	java_syserrmsg("Error creating socket for callback");
	returnrc(-1);
    }


    if (setsockopt(newfd,SOL_SOCKET,SO_REUSEADDR,
			(char *)&one,sizeof(int)) < 0)
	java_syserrmsg("setsockopt error");

    /*======================================================================
     * Bind a port for the server to call back on
     *=====================================================================*/
    memset(&newaddr, 0, sizeof(newaddr));
    newaddr.sin_family = AF_INET;
    newaddr.sin_port = INADDR_ANY;
    newaddr.sin_addr.s_addr = INADDR_ANY;
    if (bind(newfd, (struct sockaddr *)&newaddr, sizeof(newaddr)) < 0) {
	close(newfd);
	java_syserrmsg("Bind error for callback port");
	returnrc(-1);
    }

    /*======================================================================
     * Listen for the callback
     *=====================================================================*/
    if (listen(newfd,1) < 0) {
	close(newfd);
	java_syserrmsg("listen() failed for callback");
	returnrc(-1);
    }

    /*======================================================================
     * Get the bound port
     *=====================================================================*/
    len = sizeof(newaddr);
    if (getsockname(newfd, (struct sockaddr *)&newaddr,
					(socklen_t *)&len) < 0) {
	close(newfd);
	java_syserrmsg("Cannot get sock name");
	returnrc(-1);
    }

    /*======================================================================
     * Duplicate the encryption keys
     *=====================================================================*/
    if ((tmpcsched = (int)malloc(sizeof(des_key_schedule))) == 0) {
	close(newfd);
	java_syserrmsg("Cannot malloc.");
	returnrc(-1);
    }
    if ((tmpdsched = (int)malloc(sizeof(des_key_schedule))) == 0) {
	close(newfd);
	java_syserrmsg("Cannot malloc.");
	returnrc(-1);
    }
    memcpy((char *)tmpcsched, (char *)ckeyschedp, sizeof(des_key_schedule));
    memcpy((char *)tmpdsched, (char *)dkeyschedp, sizeof(des_key_schedule));
    SET_CGLUE_INT(passed_authID, auth);
    SET_CGLUE_INT(ckeyschedpID, tmpcsched);
    SET_CGLUE_INT(dkeyschedpID, tmpdsched);

    /*======================================================================
     * Send a requset to the server to callback so another socket can be opened
     * between the server and the client
     *=====================================================================*/
    sprintf(pkt.buf,"%d",ntohs(newaddr.sin_port));
    pkt.req.cmd = CALLBACK;
    pkt.req.component = CLOGIN;
    pkt.req.msglen = strlen(pkt.buf) + 1;
    order_bytes_in_req(&(pkt.req));

    /*======================================================================
     * Send the request
     *=====================================================================*/
    memset(&urg,0,sizeof(urg));
    if ((rc = send(fd, urg, 1, MSG_OOB)) < 0) {
	close(newfd);
	java_errmsg("putdata() failed to send urg byte for CALLBACK request");
	returnrc(-1);
    }

    if ((rc = putdata(fd, (char *)&pkt, sizeof(CSREQ) + ntohl(pkt.req.msglen),
					DESKEY(dkeyschedpID),1)) < 0) {
	close(newfd);
	java_errmsg("putdata() failed to send CALLBACK request");
	returnrc(-1);
    }

    /*======================================================================
     * See if server took the request
     *=====================================================================*/
    if ((rc = getdata(fd, (char *)&resp, sizeof(SCRESP),
				DESKEY(dkeyschedpID),1)) < 0) {
	close(newfd);
	close(fd);
	java_errmsg("getdata() failed to read CALLBACK response");
	returnrc(-1);
    }
    if (rc == 0) {
	close(newfd);
	java_errmsg("Server has closed connection");
	returnrc(-1);
    }
    order_bytes_in_resp(&resp);
    if (resp.rc != OK) {
	close(newfd);
	java_errmsg("CALLBACK request rejected by server");
	returnrc(-1);
    }

    /*======================================================================
     * Hold here for the server's connect back to our port
     *=====================================================================*/
    if ((rc = accept(newfd,0,0)) < 0) {
	close(newfd);
	java_syserrmsg("accept() failed for callback");
	returnrc(-1);
    }

    /*======================================================================
     * Close the file descriptor that is no longer needed and then save
     * the file descriptor returned from accept()
     * Don't close the original fd (unhand(handle)->fd) here since the
     * caller is still using it
     *=====================================================================*/
    close(newfd);
    SET_CGLUE_INT(fdID, rc);

    returnrc(0);
}

/*===========================================================================
 * This function is intended to be called from java to close the socket
 * represented by the static fd variable.
 *
 * Input:
 *
 * Output:
 *==========================================================================*/
JNIEXPORT void JNICALL Java_cglue_closefd(JNIEnv *env, jobject jo) {
    close(CGLUE_INT(fdID));
    SET_CGLUE_INT(fdID, -1);
    if (CGLUE_INT(ckeyschedpID) != 0)
	free((void *)CGLUE_INT(ckeyschedpID));
    SET_CGLUE_INT(ckeyschedpID, 0);
    if (CGLUE_INT(dkeyschedpID) != 0)
	free((void *)CGLUE_INT(dkeyschedpID));
    SET_CGLUE_INT(dkeyschedpID, 0);
    SET_CGLUE_INT(passed_authID, 0);
    return;
}

/*===========================================================================
 * This function is intended to be called from java to flush a socket
 * This is necessary when, e.g., the client side cancelled out of a monitor
 * function.  The server may have sent more data before receiving the cancel
 *
 * Input:
 *
 * Output:
 *==========================================================================*/
JNIEXPORT void JNICALL Java_cglue_flush(JNIEnv *env, jobject jo) {
    char buf[BUFSIZ];
    fd_set fs;
    struct timeval ts;

    ts.tv_sec = 0;
    ts.tv_usec = 0;

    FD_ZERO(&fs);
    while ( 1 ) {
	FD_SET(CGLUE_INT(fdID),&fs);
	if (select(CGLUE_INT(fdID)+1, &fs, (fd_set *)0,(fd_set *)0,&ts) <= 0)
	    return;
	read(CGLUE_INT(fdID),buf,sizeof(buf));
    }
}

/*===========================================================================
 * This function is only needed on a Windows system.
 * It is called by java right before exit to cleanup the socket system.
 *
 * Input:
 *
 * Output:
 *==========================================================================*/
JNIEXPORT void JNICALL Java_cglue_wsacleanup(JNIEnv *env, jobject jo) {
#ifdef MSDOS
    WSACleanup();
#endif
    return;
}

/*===========================================================================
 * This function is intended to be called from java to send a string to the
 * padmin server.
 *
 * Input:
 *	component	The code representing the target component (see
 *			cglue.java and cs.h).
 *	command		The command for the target component
 *	s		java string to be sent (It can be an empty string)
 *
 * Output:
 *	0	successful
 *	-1	fatal error
 *	-2	connection is closed
 *	-3	nonfatal error
 *		if this function returns negative, the errmsg member contains
 *		the error message
 * Note that the return value is a reference to a pktinfo instance.  In order to
 * access the pktinfo class from c, NewObject() will have to be used to obtain
 * a new instance of a class pktinfo object.  pktinfo contains: 
 * int rc	- is the integer return code
 * String str	- java string used by recvpkt to hold the data packet received
 *		  or used by getconf to hold the name of the temp file created
 * String errmsg- java string which hold an error message if rc is negative
 * These members can be accessed using the static field IDs initialized by
 * openfd.
 *
 *==========================================================================*/
JNIEXPORT jobject JNICALL Java_cglue_sendpkt(JNIEnv *env, jobject jo,
			jint component,
			jint command,
			jstring s,
			jint flag) {
    int rc;
    struct {
	struct cs_req req;
	char buf[BUFSIZ];
    } pkt;
    jobject pktrc;

    /*======================================================================
     * Establish addressability to the pktrc structure
     * Have to exit on error because we don't have the return structure to
     * return with.
     *=====================================================================*/
    pktrc = (*env)->NewObject(env, pktinfoClass, pktinfoID);
    if (pktrc == NULL) {
	fprintf(stderr,"pktinfo constructor failed");
	exit(-1);
    }

    /*======================================================================
     * If fd is still not opened, return right away.
     * This situation occurs at the login screen since as soon as the login
     * button is clicked, the 1st action is to send a KEEPALIVE
     *=====================================================================*/
    if (CGLUE_INT(fdID) < 0)
	returnrc(0);
    if ((pkt.req.msglen = (*env)->GetStringUTFLength(env, s)) >
						sizeof(pkt.buf)) {
	java_errmsg("String longer than 4096");
	returnrc(NONFATAL);
    }

    if (pkt.req.msglen > 0) {
	/* javaString2CString(s, pkt.buf, sizeof(pkt.buf)); */
	jboolean isCopy;
	const char *utfchar;
	utfchar = (*env)->GetStringUTFChars(env, s, &isCopy);
	strcpy(pkt.buf, utfchar);
	(*env)->ReleaseStringUTFChars(env, s, utfchar);
	/* Reset msglen just to make sure it contains the terminating null */
	pkt.req.msglen = strlen(pkt.buf) + 1;
    }

    pkt.req.component = component;
    pkt.req.cmd = command;
    order_bytes_in_req(&(pkt.req));

    if (CGLUE_INT(passed_authID)) {
	if ((rc = putdata(CGLUE_INT(fdID),
		(char *)&pkt, sizeof(pkt.req)+ntohl(pkt.req.msglen),
		DESKEY(ckeyschedpID),flag)) < 0) {
	    SET_CGLUE_INT(ErrnoID, errno);
	    java_errmsg("putdata() failed to send request");
	    returnrc(-1);
	}
    }
    else {
	if ((rc = putdata(CGLUE_INT(fdID),
		(char *)&pkt, sizeof(pkt.req)+ntohl(pkt.req.msglen),
		DESKEY(dkeyschedpID),flag)) < 0) {
	    SET_CGLUE_INT(ErrnoID, errno);
	    java_errmsg("putdata() failed to send request");
	    returnrc(-1);
	}
    }

    returnrc(0);
}

/*===========================================================================
 * This function is intended to be called from java to receive a response
 * from the padmin server.
 * The data received will be in the form of a String in fw.c.pktstr
 *
 * Input:
 *
 * Output:
 *	0	successful
 *	-1	fatal error
 *	-2	connection is closed
 *	-3	nonfatal error
 *		if this function returns negative, the errmsg member contains
 *		the error message
 * Note that the return value is a reference to a pktinfo instance.  In order to
 * access the pktinfo class from c, NewObject() will have to be used to obtain
 * a new instance of a class pktinfo object.  pktinfo contains: 
 * int rc	- is the integer return code
 * String str	- java string used by recvpkt to hold the data packet received
 *		  or used by getconf to hold the name of the temp file created
 * String errmsg- java string which hold an error message if rc is negative
 * These members can be accessed using the static field IDs initialized by
 * openfd.
 *
 *==========================================================================*/
JNIEXPORT jobject JNICALL Java_cglue_recvpkt(JNIEnv *env, jobject jo,
			jint flag) {
    int rc;
    SCRESP resp;
    char msg[BUFSIZ];
    struct des_ks_struct *dk;
    jobject pktrc;

    /*======================================================================
     * Establish addressability to the pktrc structure
     * Have to exit on error because we don't have the return structure to
     * return with.
     *=====================================================================*/
    pktrc = (*env)->NewObject(env, pktinfoClass, pktinfoID);
    if (pktrc == NULL) {
	fprintf(stderr,"pktinfo constructor failed");
	exit(-1);
    }

    if (CGLUE_INT(passed_authID))
	dk = DESKEY(ckeyschedpID);
    else
	dk = DESKEY(dkeyschedpID);

    if ((rc = getdata(CGLUE_INT(fdID),(char *)&resp,sizeof(resp),dk,flag))
						< 0) {
	SET_CGLUE_INT(ErrnoID, errno);
	java_syserrmsg("getdata() failed");
	returnrc(-1);
    }

    if (rc == CONTINUE)
	returnrc(CONTINUE);

    if (rc == 0) {
	java_errmsg("Server has closed connection");
	returnrc(-2);
    }

    order_bytes_in_resp(&resp);

    if (resp.msglen > 0) {
	if (resp.msglen < sizeof(msg)) {
	    if ((rc = getdata(CGLUE_INT(fdID),msg,resp.msglen,dk,flag))
						< 0) {
		SET_CGLUE_INT(ErrnoID, errno);
		java_syserrmsg("getdata() failed");
		returnrc(-1);
	    }
	    if (rc == 0) {
		java_errmsg("Server has closed connection");
		returnrc(-2);
	    }
	}
	else {
	    java_errmsg("Response from server longer than 4096");
	    returnrc(-1);
	}
    }
    else
	strcpy(msg,"");

    /*=======================================================================
     * If server returns a failure rc, stuff msg into errmsg because the
     * caller expects the error message in errmsg
     *======================================================================*/
    if (resp.rc < 0)
	java_errmsg(msg);

    (*env)->SetObjectField(env,pktrc,pktstrID,(*env)->NewStringUTF(env,msg));

    returnrc(resp.rc);
}

/*===========================================================================
 * This function is intended to be called from java to get a config file
 * from the padmin server.
 *
 * Input:
 *	component is the component this getconf request is intended for
 *
 * Output:
 *	returns 2 if getconf() loaded a locally saved file rather than from
 *			the server
 *	returns 0 if OK
 *	returns -1 if fatal error
 *	returns -2 if getdata returns an EOF condition
 *	returns -3 if non fatal error
 *	If this function returns a negative rc, the error message is in
 *	the errmsg member
 * Note that the return value is a reference to a pktinfo instance.  In order to
 * access the pktinfo class from c, NewObject() will have to be used to obtain
 * a new instance of a class pktinfo object.  pktinfo contains: 
 * int rc	- is the integer return code
 * String str	- java string used by recvpkt to hold the data packet received
 *		  or used by getconf to hold the name of the temp file created
 * String errmsg- java string which hold an error message if rc is negative
 * These members can be accessed using the static field IDs initialized by
 * openfd.
 *		
 *==========================================================================*/
JNIEXPORT jobject JNICALL Java_cglue_getconf(JNIEnv *env, jobject jo,
			jint component) {
    int rc;
    int len;
    int ofd;
    int nread;
    int total;
    SCRESP resp;
    char ext[4];
    char buf[BUFSIZ];
    char *fn;
    CSREQ req;
    struct sockaddr_in addr;
    struct des_ks_struct *dk;	/* key schedule for socket	*/
    struct des_ks_struct *ldk;	/* key schedule for local files	*/
    jobject pktrc;
    CBCBLK cbcblk;

    /*======================================================================
     * Establish addressability to the pktrc structure
     * Have to exit on error because we don't have the return structure to
     * return with.
     *=====================================================================*/
    pktrc = (*env)->NewObject(env, pktinfoClass, pktinfoID);
    if (pktrc == NULL) {
	fprintf(stderr,"pktinfo constructor failed");
	exit(-1);
    }

    /*======================================================================
     * Get key schedule for comminications with the server
     *=====================================================================*/
    if (CGLUE_INT(passed_authID))
	dk = DESKEY(ckeyschedpID);
    else
	dk = DESKEY(dkeyschedpID);

    /*======================================================================
     * Get key schedule for encrypting/decrypting local files
     *=====================================================================*/
    switch (CGLUE_INT(passed_authID)) {
	case 0:
	case 2:
	    ldk = DESKEY(dkeyschedpID);
	    break;
	case 1:
	    ldk = DESKEY(ckeyschedpID);
	    break;
    }

    /*=======================================================================
     * Generate the temporary local output filename
     *======================================================================*/
#ifdef MSDOS
    fn = tempnam(DOSTMPDIR,"_padm");
#else
    fn = tempnam(getenv("HOME"),"_padm");
#endif
    (*env)->SetObjectField(env, pktrc, pktstrID, (*env)->NewStringUTF(env, fn));

    /*======================================================================
     * Build the local save filename and save it in fw.c
     *=====================================================================*/
#ifdef MSDOS
    strcpy(buf,DOSTMPDIR);
    strcat(buf,"\\");
#else
    strcpy(buf,getenv("HOME"));
    strcat(buf,"/");
#endif

    /*======================================================================
     * Get the remote host's ip address
     *=====================================================================*/
    len = sizeof(addr);
    if (getpeername(CGLUE_INT(fdID), (struct sockaddr *)&addr,
						(socklen_t *)&len) < 0) {
	java_syserrmsg("Could not get info on remote host");
	returnrc(-1);
    }

    /*======================================================================
     * Get the file extension to use
     *=====================================================================*/
    if (fwext(component, ext) < 0) {
	java_syserrmsg("Component config file is not managed by padmin");
	returnrc(-1);
    }

    /*======================================================================
     * Save the save filename
     *=====================================================================*/
    sprintf(&buf[strlen(buf)],"%08x.%3s",addr.sin_addr.s_addr,ext);
    (*env)->SetObjectField(env, jo, localfnID, (*env)->NewStringUTF(env,buf));

    /*======================================================================
     * If a local save file exists, use it instead of downloading from the
     * server
     *=====================================================================*/
    if ((ofd = open(buf,O_RDONLY,0)) >= 0) {
	char tmpmsg[BUFSIZ];
	close(ofd);
	if (rename(buf,fn) < 0) {
	    sprintf(tmpmsg,"Cannot rename saved file to temp file %s  %s",
			fn,sys_errlist[errno]);
	    returnrc(NONFATAL);
	}
	returnrc(LOCALFILE);
    }

    /*=======================================================================
     * Open the local output file
     *======================================================================*/
#ifdef MSDOS
    ofd = open(fn, O_WRONLY|O_CREAT|O_BINARY, 0700);
#else
    while ((ofd = open(fn, O_WRONLY|O_CREAT|O_EXCL, 0700)) < 0 &&
			errno == EEXIST)
	unlink(fn);
#endif
    if (ofd < 0) {
	char tmpmsg[BUFSIZ];
	sprintf(tmpmsg,"open() failed for %s  %s",fn,sys_errlist[errno]);
	(*env)->SetObjectField(env, pktrc, pktemsgID,
		(*env)->NewStringUTF(env, tmpmsg));
	unlink(fn);
	free(fn);
	returnrc(NONFATAL);
    }

    req.component = component;
    req.cmd = GETCONF;
    req.msglen = 0;
    order_bytes_in_req(&req);

    if ((rc = putdata(CGLUE_INT(fdID), (char *)&req, sizeof(req), dk,1)) < 0) {
	java_syserrmsg("putdata failed");
	unlink(fn);
	free(fn);
	returnrc(-1);
    }

    if ((rc = getdata(CGLUE_INT(fdID), (char *)&resp, sizeof(resp),
			dk,1)) < 0) {
	java_syserrmsg("getdata() failed");
	unlink(fn);
	free(fn);
	returnrc(-1);
    }
    if (rc == 0) {
	java_errmsg("Server has closed connection");
	unlink(fn);
	free(fn);
	returnrc(-2);
    }
    order_bytes_in_resp(&resp);

    if (resp.rc == CONTINUE) {
	/*===============================================================
	 * rc indicates that someone else may be editing the same file
	 * Return to the caller with a CONTINUE rc
	 *==============================================================*/
	if ((rc = getdata(CGLUE_INT(fdID), buf, resp.msglen, dk,1)) < 0) {
	    java_syserrmsg("getdata() of error message failed");
	    unlink(fn);
	    free(fn);
	    returnrc(-1);
	}
	java_errmsg(buf);
	returnrc(CONTINUE);
    }

    if (resp.rc != OK) {
	if ((rc = getdata(CGLUE_INT(fdID), buf, resp.msglen, dk,1)) < 0) {
	    java_syserrmsg("getdata() of error message failed");
	    unlink(fn);
	    free(fn);
	    returnrc(-1);
	}
	else {
	    java_errmsg(buf);
	    unlink(fn);
	    free(fn);
	    returnrc(resp.rc);
	}
    }

    /*===================================================================
     * Initialize for cbc encryption/decryption
     *==================================================================*/
    memset((void *)&cbcblk, 0, sizeof(CBCBLK));
    memcpy(cbcblk.ivec, IVEC, sizeof(des_cblock));

    /*=======================================================================
     * Get the contents of the file over
     *======================================================================*/
    for (total=0; total < resp.msglen; total += nread) {
	int n;

	/*===================================================================
	 * Get a full buffer or whatever is left
	 *==================================================================*/
	if (resp.msglen - total > sizeof(buf))
	    n = sizeof(buf);
	else
	    n = resp.msglen - total;
	if ((nread = getsomedata(CGLUE_INT(fdID), buf, n, dk)) < 0) {
	    java_syserrmsg("Read error during file transfer");
	    unlink(fn);
	    free(fn);
	    returnrc(-1);
	}
	if (nread == 0) {
	    java_syserrmsg("Server has closed connection");
	    unlink(fn);
	    free(fn);
	    returnrc(-2);
	}

	/*===================================================================
	 * Write the bytes read
	 * Multiple writes are done to make sure a newline is always the
	 * last char in a block.  cbcreadLine() reads a block (8 bytes) at
	 * a time and decrypt it and when it encounters a newline char, it
	 * assumes there is no more input in the block.
	 *==================================================================*/
/*
	if (write(ofd, buf, nread) < nread) { 
*/
	for (n = 0; n < nread; n += len) {
	    char *p;

	    /*===============================================================
	     * If a newline is found, compute len so that we only write
	     * up to the newline.  The rest will be written separately.
	     *==============================================================*/
	    if (p = memchr(&(buf[n]), '\n', nread - n))
		len = p - &(buf[n]) + 1;
	    else
		len = nread - n;

	    if ((rc = cbc_write(ofd, (unsigned char *)&(buf[n]), len, ldk,
						&cbcblk)) < 0) {
		java_syserrmsg("write to temp file failed during file "
				"transfer");
		unlink(fn);
		free(fn);
		returnrc(-1);
	    }
	}
    }

    close(ofd);
    free(fn);
    returnrc(0);
}

/*===========================================================================
 * This function is intended to be called from java to put a config file
 * onto the padmin server.
 *
 * Input:
 *	component is the component this putconf request is intended for
 *
 * Output:
 *	returns 0 if OK
 *	returns -1 if fatal error
 *	returns -2 if putdata returns an EOF condition
 *	returns -3 if non fatal error
 *	If this function returns a negative rc, the error message is in
 *	the errmsg member
 * Note that the return value is a reference to a pktinfo instance.  In order to
 * access the pktinfo class from c, NewObject() will have to be used to obtain
 * a new instance of a class pktinfo object.  pktinfo contains: 
 * int rc	- is the integer return code
 * String str	- java string used by recvpkt to hold the data packet received
 *		  or used by getconf to hold the name of the temp file created
 * String errmsg- java string which hold an error message if rc is negative
 * These members can be accessed using the static field IDs initialized by
 * openfd.
 *		
 *==========================================================================*/
JNIEXPORT jobject JNICALL Java_cglue_putconf(JNIEnv *env, jobject jo,
			jint component, jstring jfn) {
    int rc;
    int ifd;
    int nread;
    int total;
    struct stat stat;
    SCRESP resp;
    char localfn[BUFSIZ];
    char buf[BUFSIZ];
    char fn[BUFSIZ];
    CSREQ req;
    struct des_ks_struct *dk;	/* key schedule for socket	*/
    struct des_ks_struct *ldk;	/* key schedule for local files	*/
    jstring jstr;
    const char *utfchar;
    jobject pktrc;
    jboolean isCopy;
    CBCBLK cbcblk;

    /*======================================================================
     * Establish addressability to the pktrc structure
     * Have to exit on error because we don't have the return structure to
     * return with.
     *=====================================================================*/
    pktrc = (*env)->NewObject(env, pktinfoClass, pktinfoID);
    if (pktrc == NULL) {
	fprintf(stderr,"pktinfo constructor failed");
	exit(-1);
    }

    /*======================================================================
     * Get key schedule for commications with the server
     *=====================================================================*/
    if (CGLUE_INT(passed_authID))
	dk = DESKEY(ckeyschedpID);
    else
	dk = DESKEY(dkeyschedpID);

    /*======================================================================
     * Get key schedule for encrypting/decrypting local files
     *=====================================================================*/
    switch (CGLUE_INT(passed_authID)) {
	case 0:
	case 2:
	    ldk = DESKEY(dkeyschedpID);
	    break;
	case 1:
	    ldk = DESKEY(ckeyschedpID);
	    break;
    }

    /*=======================================================================
     * Make sure the filename fits into our buffer
     *======================================================================*/
    if ((*env)->GetStringUTFLength(env, jfn) > sizeof(fn)) {
	java_errmsg("Temp file name too long");
	returnrc(-1);
    }
    /*=======================================================================
     * Convert to c string, open and stat the file to get the length
     *======================================================================*/
    jstr = (*env)->GetObjectField(env,jo,localfnID);
    utfchar = (*env)->GetStringUTFChars(env, jstr, &isCopy);
    strcpy(localfn, utfchar);
    (*env)->ReleaseStringUTFChars(env, jstr, utfchar);

    utfchar = (*env)->GetStringUTFChars(env, jfn, &isCopy);
    strcpy(fn, utfchar);
    (*env)->ReleaseStringUTFChars(env, jfn, utfchar);

    if (strlen(fn) == 0) {
	java_errmsg("Inetranl error: could not find local file for upload");
	returnrc(NONFATAL);
    }

#ifdef MSDOS
    if ((ifd = open(fn, O_RDONLY|O_BINARY, 0)) < 0) {
#else
    if ((ifd = open(fn, O_RDONLY, 0)) < 0) {
#endif
	java_syserrmsg("Cannot open local file for upload");
	returnrc(NONFATAL);
    }

    if (fstat(ifd, &stat) < 0) {
	java_syserrmsg("Cannot stat local file for upload");
	close(ifd);
	returnrc(NONFATAL);
    }

    /*=======================================================================
     * Make sure fd is still valid
     *======================================================================*/
    if (CGLUE_INT(fdID) < 0) {
	if (rename(fn, localfn) < 0) {
	    java_errmsg("Not connected to server.  "
		"Changes cannot be saved locally");
	}
	else {
	    java_errmsg("Not connected to server.  Changes saved locally");
	}
	returnrc(NONFATAL);
    }

    /*======================================================================
     * The putconf protocol is unusual in that the client first sends a
     * putconf request packet with msglen=0.  This makes sure that the
     * server is OK for the upload.  The server sends a response packet
     * indicating it is OK.  The client then send a second request packet,
     * this time with msglen set to the length of the upload.
     *=====================================================================*/
    req.component = component;
    req.cmd = PUTCONF;
    req.msglen = 0;
    order_bytes_in_req(&req);

    if ((rc = putdata(CGLUE_INT(fdID), (char *)&req, sizeof(req), dk,1)) < 0) {
	close(ifd);
	if (rename(fn, localfn) < 0) {
	    java_syserrmsg2("Remote write failed",
		"Changes cannot be saved locally");
	}
	else
	    java_syserrmsg2("Remote write failed",
		"Changes saved locally");
	returnrc(NONFATAL);
    }

    if ((rc = getdata(CGLUE_INT(fdID), (char *)&resp, sizeof(resp),
			dk,1)) <= 0) {
	close(ifd);
	if (rename(fn, localfn) < 0) {
	    java_syserrmsg2("Request to write remote file failed",
		"Changes cannot be saved locally");
	}
	else
	    java_syserrmsg2("Request to write remote file failed",
		"Changes saved locally");
	if (rc == 0) {
	    java_errmsg("Server has closed connection");
	    returnrc(-2);
	}
	else
	    returnrc(-1);
    }
    order_bytes_in_resp(&resp);

    if (resp.rc != OK) {
	close(ifd);
	if ((rc = getdata(CGLUE_INT(fdID), buf, resp.msglen, dk,1)) < 0) {
	    if (rename(fn, localfn) < 0) {
		java_syserrmsg2("Remote write failed",
			"Changes cannot be saved locally");
	    }
	    else
		java_syserrmsg2("Remote write failed",
			"Changes saved locally");
		returnrc(-1);
	}
	else {
	    char tmp[BUFSIZ];
	    strcpy(tmp,buf);
	    if (rename(fn,localfn) < 0) {
		strcat(tmp,"  Changes cannot be saved locally");
	    }
	    else
		strcat(tmp,"  Changes saved locally");
	    java_errmsg(tmp);
	    returnrc(resp.rc);
	}
    }

    /*======================================================================
     * This is the second request packet
     *=====================================================================*/
    req.msglen = htonl(stat.st_size);
    if ((rc = putdata(CGLUE_INT(fdID), (char *)&req, sizeof(req), dk,1)) < 0) {
	if (rename(fn,localfn) < 0) {
	    java_syserrmsg2("Second remote write request failed",
			"Changes cannot be saved locally");
	}
	else
	    java_syserrmsg2("Second remote write request failed",
			"Changes saved locally");
	close(ifd);
	returnrc(-1);
    }

    /*===================================================================
     * Initialize for cbc encryption/decryption
     *==================================================================*/
    memset((void *)&cbcblk, 0, sizeof(CBCBLK));
    memcpy(cbcblk.ivec, IVEC, sizeof(des_cblock));

    for (total=0; total < stat.st_size; total += nread) {
	fd_set rdset;
	struct timeval ts;

	/*===============================================================
	 * First find out if the server wants to stop the transfer.
	 *==============================================================*/
	FD_ZERO(&rdset);
	FD_SET(CGLUE_INT(fdID),&rdset);
	ts.tv_sec = 0;
	ts.tv_usec = 0;
	if ((rc = select(CGLUE_INT(fdID)+1, &rdset, (fd_set *)0,
					(fd_set *)0, &ts)) < 0) {
	    java_syserrmsg2("select() error polling server response",
			"Attempting to continue");
	}
	else if (rc > 0) {
	    if ((rc = getdata(CGLUE_INT(fdID), (char *)&resp, sizeof(resp),
			dk,1)) < 0) {
		if (rename(fn,localfn) < 0) {
		    java_syserrmsg2("Remote write failed",
			"Changes cannot be saved locally");
		}
		else
		    java_syserrmsg2("Remote write failed",
			"Changes saved locally");
		close(ifd);
		returnrc(-1);
	    }
	    if (rc == 0) {
		java_errmsg("Server has closed connection");
		close(ifd);
		returnrc(-2);
	    }
	    order_bytes_in_resp(&resp);

	    if (resp.rc != OK) {
		if ((rc = getdata(CGLUE_INT(fdID), buf, resp.msglen,
			dk,1)) < 0) {
		    if (rename(fn,localfn) < 0) {
			java_syserrmsg2("Remote write failed",
				"Changes cannot be saved locally");
		    }
		    else
			java_syserrmsg2("Remote write failed",
				"Changes saved locally");
		    close(ifd);
		    returnrc(-1);
		}
		else {
		    char tmp[BUFSIZ];
		    strcpy(tmp,buf);
		    if (rename(fn,localfn) < 0) {
			strcat(tmp,"  Changes cannot be saved locally");
		    }
		    else
			strcat(tmp,"  Changes saved locally");
		    java_errmsg(tmp);
		    close(ifd);
		    returnrc(resp.rc);
		}
	    }
	}
	
	if ((nread = cbc_read(ifd, (unsigned char *)buf, sizeof(buf), ldk,
							&cbcblk)) < 0) {
	    if (rename(fn,localfn) < 0) {
		java_syserrmsg2("Read error reading local file for upload",
			"Changes cannot be saved locally");
	    }
	    else
		java_syserrmsg2("Read error reading local file for upload",
			"Changes saved locally");
	    close(ifd);
	    returnrc(NONFATAL);
	}

	/*===============================================================
	 * Since an encrypted file can (and is likely) be longer than its
	 * unencrypted version, and we have already told the server we
	 * are sending over stat.st_size bytes, we may have to send filler
	 * bytes of nulls.
	 *==============================================================*/
	if (nread == 0) {
	    nread = stat.st_size - total;
	    if (nread > sizeof(buf))
		nread = sizeof(buf);
	    memset(buf, 0, nread);
	}
	if ((rc = putdata(CGLUE_INT(fdID), buf, nread, dk,1)) < nread) {
	    if (rename(fn,localfn) < 0) {
		java_syserrmsg2("Error writing file to server",
			"Changes cannot be saved locally");
	    }
	    else
		java_syserrmsg2("Error writing file to server",
			"Changes saved locally");
	    close(ifd);
	    returnrc(-1);
	}
    }

    if ((rc = getdata(CGLUE_INT(fdID), (char *)&resp, sizeof(resp),
			dk,1)) < 0) {
	if (rename(fn,localfn) < 0) {
	    java_syserrmsg2("Remote write failed",
			"Changes cannot be saved locally");
	}
	else
	    java_syserrmsg2("Remote write failed",
			"Changes saved locally");
	close(ifd);
	returnrc(-1);
    }
    if (rc == 0) {
	java_errmsg("Server has closed connection");
	close(ifd);
	returnrc(-2);
    }
    order_bytes_in_resp(&resp);

    if (resp.rc != OK) {
	if ((rc = getdata(CGLUE_INT(fdID), buf, resp.msglen,
			dk,1)) < 0) {
	    if (rename(fn,localfn) < 0) {
		java_syserrmsg2("Remote write failed",
			"Changes cannot be saved locally");
	    }
	    else
		java_syserrmsg2("Remote write failed",
			"Changes saved locally");
	    close(ifd);
	    returnrc(-1);
	}
	else {
	    char tmp[BUFSIZ];
	    strcpy(tmp,buf);
	    if (rename(fn,localfn) < 0) {
		strcat(tmp,"  Changes cannot be saved locally");
	    }
	    else
		strcat(tmp,"  Changes saved locally");
	    java_errmsg(tmp);
	    close(ifd);
	    returnrc(resp.rc);
	}
    }

    close(ifd);
    unlink(fn);
    returnrc(0);
}

/*===========================================================================
 * This function is intended to be called from java to setup the user
 * encryption key
 *
 * Input:
 *	jusername username in java String format
 *	jpid	  server child process pid for local connection shared area
 *
 * Output:
 *	returns 0 if OK
 *	returns -1 if fatal error
 *	returns -2 if putdata returns an EOF condition
 *	returns -3 if non fatal error
 *	If this function returns a negative rc, the error message is in
 *	the errmsg member
 * Note that the return value is a reference to a pktinfo instance.  In order to
 * access the pktinfo class from c, NewObject() will have to be used to obtain
 * a new instance of a class pktinfo object.  pktinfo contains: 
 * int rc	- is the integer return code
 * String str	- java string used by recvpkt to hold the data packet received
 *		  or used by getconf to hold the name of the temp file created
 * String errmsg- java string which hold an error message if rc is negative
 * These members can be accessed using the static field IDs initialized by
 * openfd.
 *		
 *==========================================================================*/
JNIEXPORT jobject JNICALL Java_cglue_setkey(JNIEnv *env, jobject jo,
			jstring jusername, jstring jpid) {
    char username[UNMSIZ];
    char pidname[9];
    int pid;
    jobject pktrc;
    const char *utfchar;
    jboolean isCopy;

    /*======================================================================
     * Establish addressability to the pktrc structure
     * Have to exit on error because we don't have the return structure to
     * return with.
     *=====================================================================*/
    pktrc = (*env)->NewObject(env, pktinfoClass, pktinfoID);
    if (pktrc == NULL) {
	fprintf(stderr,"pktinfo constructor failed");
	exit(-1);
    }

    /*======================================================================
     * Check to see if peer is the loopback address
     *=====================================================================*/

    /*===============================================================
     * Get the username into c string
     *==============================================================*/
    if ((*env)->GetStringUTFLength(env, jusername) > sizeof(username)) {
	java_errmsg("User name too long");
	returnrc(-1);
    }

    /* javaString2CString(jusername, username, sizeof(username)); */
    utfchar = (*env)->GetStringUTFChars(env, jusername, &isCopy);
    strcpy(username, utfchar);
    (*env)->ReleaseStringUTFChars(env, jusername, utfchar);

    /*===============================================================
     * Get the pid into c string
     *==============================================================*/
    if ((*env)->GetStringUTFLength(env, jpid) > sizeof(pidname)) {
	java_errmsg("Local connection key too long");
	returnrc(-1);
    }

    utfchar = (*env)->GetStringUTFChars(env, jpid, &isCopy);
    strcpy(pidname, utfchar);
    (*env)->ReleaseStringUTFChars(env, jpid, utfchar);

    pid = atoi(pidname);

    if (strlen(username) == 0) {
#ifndef MSDOS
	int shmid;
	void *shmptr;
	/*===============================================================
	 * Special case local connection, pick up the key in shared memory
	 *==============================================================*/
	if ((shmid = shmget(pid,sizeof(des_key_schedule),0)) < 0) {
	    java_syserrmsg2("Cannot complete local connection", "Make sure "
				"you are logged in as root");
	    returnrc(-1);
	}

	if ((shmptr = shmat(shmid,(void *)0,0)) == (void *)-1) {
	    java_syserrmsg2("Cannot complete local connection", "Make sure "
				"you are logged in as root");
	    shmctl(shmid, IPC_RMID, 0);
	    returnrc(-1);
	}

	memcpy((char *)CGLUE_INT(ckeyschedpID),(char *)shmptr,
			sizeof(des_key_schedule));

	shmdt(shmptr);
	if (shmctl(shmid, IPC_RMID, 0) < 0) {
	    java_syserrmsg("Cannot complete local connection");
	    returnrc(-1);
	}

	returnrc(0);
#else
	java_syserrmsg("Host 127.0.0.1 cannot be used from Windows OS");
	returnrc(-1);
#endif
    }
    else {
	des_cblock userkey;

	if (gettnkey(username, &userkey) < 0) {
	    java_errmsg("Could not get user encryption key");
	    returnrc(-1);
	}

	des_set_key(&userkey,DESKEY(ckeyschedpID));
	returnrc(0);
    }
}

/*===========================================================================
 * This function is intended to be called from java to get a gwuser/gwgroup
 * record from the server
 *
 * Input:
 *	action	ADDUSER,CLONEUSER,MODUSER,ADDGROUP,CLONEGROUP,MODGROUP
 *	rec	a gwrecord instance where the gwuser/gwgroup info is to
 *		be returned
 *
 * Output:
 *	returns 0 if OK
 *	returns -1 if fatal error
 *	returns -2 if putdata returns an EOF condition
 *	returns -3 if non fatal error
 *	If this function returns a negative rc, the error message is in
 *	the errmsg member
 * Note that the return value is a reference to a pktinfo instance.  In order to
 * access the pktinfo class from c, NewObject() will have to be used to obtain
 * a new instance of a class pktinfo object.  pktinfo contains: 
 * int rc	- is the integer return code
 * String str	- java string used by recvpkt to hold the data packet received
 *		  or used by getconf to hold the name of the temp file created
 * String errmsg- java string which hold an error message if rc is negative
 * These members can be accessed using the static field IDs initialized by
 * openfd.
 *		
 *==========================================================================*/
JNIEXPORT jobject JNICALL Java_cglue_getgwuser(JNIEnv *env, jobject jo,
		jint action, jobject rec) {
    int rc;
    struct GWUSER gwu;
    struct {
	CSREQ req;
	char buf[BUFSIZ];
    } pkt;
    SCRESP resp;
    const char *tmpname;
    char name[UNMSIZ];
    jobject pktrc;
    jfieldID nameID;
    jfieldID realnameID;
    jfieldID phoneID;
    jfieldID groupID;
    jfieldID passwordID;
    jfieldID tokentypeID;
    jfieldID keyID;
    jfieldID tnaccessID;
    jfieldID ftpaccessID;
    jfieldID startdateID;
    jfieldID stopdateID;
    jfieldID accesstimesID;
    jfieldID lockedID;
    jclass gwClass;
    jstring jstr;
    jboolean isCopy;
    jbyteArray ba;
    jbyte *lba;
    int isgwuser;

    /*======================================================================
     * Establish addressability to the pktrc structure
     * Have to exit on error because we don't have the return structure to
     * return with.
     *=====================================================================*/
    pktrc = (*env)->NewObject(env, pktinfoClass, pktinfoID);
    if (pktrc == NULL) {
	fprintf(stderr,"pktinfo constructor failed");
	exit(-1);
    }

    /*=======================================================================
     * Obtain the field IDs for all the member variables in gwrecord
     *======================================================================*/
    gwClass = (*env)->GetObjectClass(env, rec);
    if (gwClass == NULL) {
	java_errmsg("Could not find class gwrecord");
	returnrc(-1);
    }

    nameID = (*env)->GetFieldID(env, gwClass, "name", "Ljava/lang/String;");
    realnameID = (*env)->GetFieldID(env, gwClass, "realname",
						"Ljava/lang/String;");
    phoneID = (*env)->GetFieldID(env, gwClass, "phone", "Ljava/lang/String;");
    groupID = (*env)->GetFieldID(env, gwClass, "group", "Ljava/lang/String;");
    passwordID = (*env)->GetFieldID(env, gwClass, "password",
						"Ljava/lang/String;");
    tokentypeID = (*env)->GetFieldID(env, gwClass, "tokentype", "B");
    keyID = (*env)->GetFieldID(env, gwClass, "key", "Ljava/lang/String;");
    tnaccessID = (*env)->GetFieldID(env, gwClass, "tnaccess", "B");
    ftpaccessID = (*env)->GetFieldID(env, gwClass, "ftpaccess", "I");
    startdateID = (*env)->GetFieldID(env, gwClass, "startdate", "I");
    stopdateID = (*env)->GetFieldID(env, gwClass, "stopdate", "I");
    accesstimesID = (*env)->GetFieldID(env, gwClass, "accesstimes", "[B");
    lockedID = (*env)->GetFieldID(env, gwClass, "locked", "Z");

    /*=======================================================================
     * Get the user/group name from the caller
     *======================================================================*/
    jstr = (*env)->GetObjectField(env, rec, nameID);
    if ((*env)->GetStringUTFLength(env, jstr) > UNMSIZ - 1) {
	java_errmsg("Name too long");
	returnrc(-1);
    }
    tmpname = (*env)->GetStringUTFChars(env, jstr, &isCopy);
    strcpy(name, tmpname);
    (*env)->ReleaseStringUTFChars(env, jstr, tmpname);

    /*=======================================================================
     * Send a request to the server
     *======================================================================*/
    pkt.req.cmd = GETGWUSER;
    if (action == ADDUSER || action == CLONEUSER || action == MODUSER) {
	pkt.req.component = CGWUSER;
	isgwuser = 1;
    }
    else {
	pkt.req.component = CGWGROUP;
	isgwuser = 0;
    }
    pkt.req.msglen = strlen(name) + 1;
    strcpy(pkt.buf, name);
    order_bytes_in_req(&(pkt.req));
    if ((rc = putdata(CGLUE_INT(fdID), (char *)&pkt,
			sizeof(CSREQ) + ntohl(pkt.req.msglen),
			DESKEY(ckeyschedpID),1)) < 0) {
	java_syserrmsg("putdata() failed");
	returnrc(-1);
    }

    /*=======================================================================
     * Get the response from the server
     *======================================================================*/
    if ((rc = getdata(CGLUE_INT(fdID), (char *)&resp, sizeof(resp),
			DESKEY(ckeyschedpID),1)) < 0) {
	java_syserrmsg("getdata() failed");
	returnrc(-1);
    }

    if (rc == 0) {
	java_errmsg("Server has closed connection");
	returnrc(CLOSED);
    }
    order_bytes_in_resp(&resp);

    if (resp.rc != OK) {
	if (resp.msglen > 0 && getdata(CGLUE_INT(fdID), pkt.buf,
			resp.msglen,DESKEY(ckeyschedpID),1) > 0) {
	    java_errmsg(pkt.buf);
	}
	else
	    java_errmsg("getdata() failed");

	returnrc(resp.rc);
    }

    /*=======================================================================
     * Get the gwuser/gwgroup info
     *======================================================================*/
    if (resp.msglen != sizeof(struct GWUSER)) {
	java_errmsg("gwuser/gwgroup info mismatch");
	returnrc(-1);
    }

    if ((rc = getdata(CGLUE_INT(fdID), (char *)&gwu, resp.msglen,
			DESKEY(ckeyschedpID),1)) < 0) {
	java_syserrmsg("getdata() failed");
	returnrc(-1);
    }

    if (rc == 0) {
	java_errmsg("Server has closed connection");
	returnrc(CLOSED);
    }

    /*=======================================================================
     * Make sure the record returned is for the name we asked for
     *======================================================================*/
    if (strcmp(name,gwu.username)) {
	char tmp[BUFSIZ];
	sprintf(tmp,"Record returned does not match requested name %s",name);
	java_errmsg(tmp);
	returnrc(NONFATAL);
    }

    /*=======================================================================
     * Transfer the info into java fields
     *======================================================================*/
    jstr = (*env)->NewStringUTF(env, gwu.realname);
    (*env)->SetObjectField(env, rec, realnameID, jstr);

    (*env)->SetByteField(env, rec, tokentypeID, gwu.tokentype);

    (*env)->SetByteField(env, rec, tnaccessID, gwu.tnaccess);

    (*env)->SetIntField(env, rec, ftpaccessID, gwu.ftpaccess);

    ba = (*env)->NewByteArray(env, sizeof(gwu.accesstimes));
    lba = (*env)->GetByteArrayElements(env, ba, NULL);
    memcpy(lba, gwu.accesstimes, sizeof(gwu.accesstimes));
    (*env)->ReleaseByteArrayElements(env, ba, lba, 0);
    (*env)->SetObjectField(env, rec, accesstimesID, ba);

    if (isgwuser) {
	char keystring[32];

	jstr = (*env)->NewStringUTF(env, gwu.phonenum);
	(*env)->SetObjectField(env, rec, phoneID, jstr);

	jstr = (*env)->NewStringUTF(env, gwu.groupname);
	(*env)->SetObjectField(env, rec, groupID, jstr);

	jstr = (*env)->NewStringUTF(env, (char *)gwu.password);
	(*env)->SetObjectField(env, rec, passwordID, jstr);

	/*===================================================================
	 * For CLONEUSER, do not use the token key from the user record,
	 * generate a random key instead.
	 *==================================================================*/
	if (action == CLONEUSER)
	    randomkey(gwu.userkey);
	sprintf(keystring,"%3.3o %3.3o %3.3o %3.3o %3.3o %3.3o %3.3o %3.3o",
	gwu.userkey[0],gwu.userkey[1],gwu.userkey[2],gwu.userkey[3],
	gwu.userkey[4],gwu.userkey[5],gwu.userkey[6],gwu.userkey[7]);
	jstr = (*env)->NewStringUTF(env, keystring);
	(*env)->SetObjectField(env, rec, keyID, jstr);

	(*env)->SetIntField(env, rec, startdateID, gwu.startdate);
	(*env)->SetIntField(env, rec, stopdateID, gwu.stopdate);
	if (gwu.status == LOCKED)
	    (*env)->SetBooleanField(env, rec, lockedID, JNI_TRUE);
	else
	    (*env)->SetBooleanField(env, rec, lockedID, JNI_FALSE);
    }

    returnrc(0);
}

/*===========================================================================
 * This function is intended to be called from java to get the gwuser/gwgroup
 * default values from the server's gwuser.conf file
 *
 * Input:
 *	action	ADDUSER,CLONEUSER,MODUSER,ADDGROUP,CLONEGROUP,MODGROUP
 *	rec	a gwrecord instance where the gwuser/gwgroup info is to
 *		be returned
 *
 * Output:
 *	returns 0 if OK
 *	returns -1 if fatal error
 *	returns -2 if putdata returns an EOF condition
 *	returns -3 if non fatal error
 *	If this function returns a negative rc, the error message is in
 *	the errmsg member
 * Note that the return value is a reference to a pktinfo instance.  In order to
 * access the pktinfo class from c, NewObject() will have to be used to obtain
 * a new instance of a class pktinfo object.  pktinfo contains: 
 * int rc	- is the integer return code
 * String str	- java string used by recvpkt to hold the data packet received
 *		  or used by getconf to hold the name of the temp file created
 * String errmsg- java string which hold an error message if rc is negative
 * These members can be accessed using the static field IDs initialized by
 * openfd.
 *		
 *==========================================================================*/
JNIEXPORT jobject JNICALL Java_cglue_getgwdefault(JNIEnv *env, jobject jo,
		jint action, jobject rec) {
    int rc;
    struct GWUSER gwu;
    CSREQ req;
    SCRESP resp;
    char buf[BUFSIZ];
    jobject pktrc;
    jfieldID groupID;
    jfieldID tokentypeID;
    jfieldID keyID;
    jfieldID tnaccessID;
    jfieldID ftpaccessID;
    jfieldID accesstimesID;
    jclass gwClass;
    jstring jstr;
    jbyteArray ba;
    jbyte *lba;

    /*======================================================================
     * Establish addressability to the pktrc structure
     * Have to exit on error because we don't have the return structure to
     * return with.
     *=====================================================================*/
    pktrc = (*env)->NewObject(env, pktinfoClass, pktinfoID);
    if (pktrc == NULL) {
	fprintf(stderr,"pktinfo constructor failed");
	exit(-1);
    }

    /*=======================================================================
     * Obtain the field IDs for some of the member variables in gwrecord
     *======================================================================*/
    gwClass = (*env)->GetObjectClass(env, rec);
    if (gwClass == NULL) {
	java_errmsg("Could not find class gwrecord");
	returnrc(-1);
    }

    groupID = (*env)->GetFieldID(env, gwClass, "group", "Ljava/lang/String;");
    tokentypeID = (*env)->GetFieldID(env, gwClass, "tokentype", "B");
    keyID = (*env)->GetFieldID(env, gwClass, "key", "Ljava/lang/String;");
    tnaccessID = (*env)->GetFieldID(env, gwClass, "tnaccess", "B");
    ftpaccessID = (*env)->GetFieldID(env, gwClass, "ftpaccess", "I");
    accesstimesID = (*env)->GetFieldID(env, gwClass, "accesstimes", "[B");

    /*=======================================================================
     * Send a request to the server
     *======================================================================*/
    req.cmd = GETDEFGWUSER;
    req.component = CGWUSER;
    req.msglen = 0;
    order_bytes_in_req(&req);
    if ((rc = putdata(CGLUE_INT(fdID), (char *)&req, sizeof(CSREQ),
			DESKEY(ckeyschedpID),1)) < 0) {
	java_syserrmsg("putdata() failed");
	returnrc(-1);
    }

    /*=======================================================================
     * Get the response from the server
     *======================================================================*/
    if ((rc = getdata(CGLUE_INT(fdID), (char *)&resp, sizeof(resp),
			DESKEY(ckeyschedpID),1)) < 0) {
	java_syserrmsg("getdata() failed");
	returnrc(-1);
    }

    if (rc == 0) {
	java_errmsg("Server has closed connection");
	returnrc(CLOSED);
    }
    order_bytes_in_resp(&resp);

    if (resp.rc != OK) {
	if (resp.msglen > 0 && getdata(CGLUE_INT(fdID), buf,
			resp.msglen,DESKEY(ckeyschedpID),1) > 0) {
	    java_errmsg(buf);
	}
	else
	    java_errmsg("getdata() failed");

	returnrc(resp.rc);
    }

    /*=======================================================================
     * Get the gwuser/gwgroup info
     *======================================================================*/
    if (resp.msglen != sizeof(struct GWUSER)) {
	java_errmsg("gwuser/gwgroup info mismatch");
	returnrc(-1);
    }

    if ((rc = getdata(CGLUE_INT(fdID), (char *)&gwu, resp.msglen,
			DESKEY(ckeyschedpID),1)) < 0) {
	java_syserrmsg("getdata() failed");
	returnrc(-1);
    }

    if (rc == 0) {
	java_errmsg("Server has closed connection");
	returnrc(CLOSED);
    }

    /*=======================================================================
     * Transfer the info into java fields based on whether the request
     * was for userr defaults or group defaults
     *======================================================================*/
    if (action == ADDUSER) {
	char keystring[32];

	jstr = (*env)->NewStringUTF(env, gwu.groupname);
	(*env)->SetObjectField(env, rec, groupID, jstr);

	/*===================================================================
	 * Generate a random key
	 *==================================================================*/
	randomkey(buf);
	sprintf(keystring,"%3.3o %3.3o %3.3o %3.3o %3.3o %3.3o %3.3o %3.3o",
		buf[0],buf[1],buf[2],buf[3],buf[4],buf[5],buf[6],buf[7]);
	jstr = (*env)->NewStringUTF(env, keystring);
	(*env)->SetObjectField(env, rec, keyID, jstr);
    }
    else {
	(*env)->SetByteField(env, rec, tokentypeID, gwu.tokentype);

	(*env)->SetByteField(env, rec, tnaccessID, gwu.tnaccess);

	(*env)->SetIntField(env, rec, ftpaccessID, gwu.ftpaccess);

	ba = (*env)->NewByteArray(env, sizeof(gwu.accesstimes));
	lba = (*env)->GetByteArrayElements(env, ba, NULL);
	memcpy(lba, gwu.accesstimes, sizeof(gwu.accesstimes));
	(*env)->ReleaseByteArrayElements(env, ba, lba, 0);
	(*env)->SetObjectField(env, rec, accesstimesID, ba);
    }

    returnrc(0);
}

/*===========================================================================
 * This function is intended to be called from java to update or add a
 * gwuser/gwgroup record on the server
 *
 * Input:
 *	action	ADDUSER,CLONEUSER,MODUSER,ADDGROUP,CLONEGROUP,MODGROUP
 *	rec	a gwrecord instance which contains the gwuser/gwgroup info to be
 *		uploaded
 *
 * Output:
 *	returns 0 if OK
 *	returns -1 if fatal error
 *	returns -2 if putdata returns an EOF condition
 *	returns -3 if non fatal error
 *	If this function returns a negative rc, the error message is in
 *	the errmsg member
 * Note that the return value is a reference to a pktinfo instance.  In order to
 * access the pktinfo class from c, NewObject() will have to be used to obtain
 * a new instance of a class pktinfo object.  pktinfo contains: 
 * int rc	- is the integer return code
 * String str	- java string used by recvpkt to hold the data packet received
 *		  or used by getconf to hold the name of the temp file created
 * String errmsg- java string which hold an error message if rc is negative
 * These members can be accessed using the static field IDs initialized by
 * openfd.
 *		
 *==========================================================================*/
JNIEXPORT jobject JNICALL Java_cglue_putgwuser(JNIEnv *env, jobject jo,
		jint action, jobject rec) {
    int rc;
    struct {
	CSREQ req;
	struct GWUSER gwu;
    } pkt;
    SCRESP resp;
    const char *cp;
    jobject pktrc;
    jbyteArray jobj;
    jfieldID nameID;
    jfieldID realnameID;
    jfieldID phoneID;
    jfieldID groupID;
    jfieldID passwordID;
    jfieldID tokentypeID;
    jfieldID keyID;
    jfieldID tnaccessID;
    jfieldID ftpaccessID;
    jfieldID startdateID;
    jfieldID stopdateID;
    jfieldID accesstimesID;
    jfieldID lockedID;
    jclass gwClass;
    jstring jstr;
    jboolean isCopy;
    jbyte *lba;
    int isgwuser;

    /*======================================================================
     * Establish addressability to the pktrc structure
     * Have to exit on error because we don't have the return structure to
     * return with.
     *=====================================================================*/
    pktrc = (*env)->NewObject(env, pktinfoClass, pktinfoID);
    if (pktrc == NULL) {
	fprintf(stderr,"pktinfo constructor failed");
	exit(-1);
    }

    /*=======================================================================
     * Obtain the field IDs for all the member variables in gwrecord
     *======================================================================*/
    gwClass = (*env)->GetObjectClass(env, rec);
    if (gwClass == NULL) {
	java_errmsg("Could not find class gwrecord");
	returnrc(NONFATAL);
    }

    nameID = (*env)->GetFieldID(env, gwClass, "name", "Ljava/lang/String;");
    realnameID = (*env)->GetFieldID(env, gwClass, "realname",
						"Ljava/lang/String;");
    phoneID = (*env)->GetFieldID(env, gwClass, "phone", "Ljava/lang/String;");
    groupID = (*env)->GetFieldID(env, gwClass, "group", "Ljava/lang/String;");
    passwordID = (*env)->GetFieldID(env, gwClass, "password",
						"Ljava/lang/String;");
    tokentypeID = (*env)->GetFieldID(env, gwClass, "tokentype", "B");
    keyID = (*env)->GetFieldID(env, gwClass, "key", "Ljava/lang/String;");
    tnaccessID = (*env)->GetFieldID(env, gwClass, "tnaccess", "B");
    ftpaccessID = (*env)->GetFieldID(env, gwClass, "ftpaccess", "I");
    startdateID = (*env)->GetFieldID(env, gwClass, "startdate", "I");
    stopdateID = (*env)->GetFieldID(env, gwClass, "stopdate", "I");
    accesstimesID = (*env)->GetFieldID(env, gwClass, "accesstimes", "[B");
    lockedID = (*env)->GetFieldID(env, gwClass, "locked", "Z");

    /*=======================================================================
     * Construct the pkt request header
     *======================================================================*/
    switch (action) {
	case ADDUSER:
	case CLONEUSER:
	    pkt.req.cmd = ADDGWUSER;
	    pkt.req.component = CGWUSER;
	    isgwuser = 1;
	    break;
	case MODUSER:
	    pkt.req.cmd = MODGWUSER;
	    pkt.req.component = CGWUSER;
	    isgwuser = 1;
	    break;
	case ADDGROUP:
	case CLONEGROUP:
	    pkt.req.cmd = ADDGWUSER;
	    pkt.req.component = CGWGROUP;
	    isgwuser = 0;
	    break;
	case MODGROUP:
	    pkt.req.cmd = MODGWUSER;
	    pkt.req.component = CGWGROUP;
	    isgwuser = 0;
	    break;
    }

    /*=======================================================================
     * Initialize the gwuser structure
     *======================================================================*/
    memset((char *)&pkt.gwu, 0, sizeof(pkt.gwu));

    /*=======================================================================
     * Convert java username to c string
     *======================================================================*/
    jstr = (*env)->GetObjectField(env, rec, nameID);
    if ((*env)->GetStringUTFLength(env, jstr) > UNMSIZ - 1) {
	java_errmsg("Name too long");
	returnrc(NONFATAL);
    }
    cp = (*env)->GetStringUTFChars(env, jstr, &isCopy);
    strcpy(pkt.gwu.username, cp);
    (*env)->ReleaseStringUTFChars(env, jstr, cp);

    /*===================================================================
     * Convert java realname to c string
     *==================================================================*/
    jstr = (*env)->GetObjectField(env, rec, realnameID);
    if ((*env)->GetStringUTFLength(env, jstr) > UNMSIZ - 1) {
	java_errmsg("Real Name too long");
	returnrc(NONFATAL);
    }
    cp = (*env)->GetStringUTFChars(env, jstr, &isCopy);
    strcpy(pkt.gwu.realname, cp);
    (*env)->ReleaseStringUTFChars(env, jstr, cp);

    /*=======================================================================
     * Convert java byte tokentype to c
     *======================================================================*/
    pkt.gwu.tokentype = (*env)->GetByteField(env, rec, tokentypeID);

    /*=======================================================================
     * Convert java byte tnaccess to c
     *======================================================================*/
    pkt.gwu.tnaccess = (*env)->GetByteField(env, rec, tnaccessID);

    /*=======================================================================
     * Convert java int ftpaccess to c
     *======================================================================*/
    pkt.gwu.ftpaccess = (*env)->GetIntField(env, rec, ftpaccessID);

    /*=======================================================================
     * Convert java byte array to c
     *======================================================================*/
    jobj = (*env)->GetObjectField(env, rec, accesstimesID);
    lba = (*env)->GetByteArrayElements(env, jobj, NULL);
    memcpy(pkt.gwu.accesstimes, lba, sizeof(pkt.gwu.accesstimes));
    (*env)->ReleaseByteArrayElements(env, jobj, lba, 0);

    /*=======================================================================
     * Process gwuser only (not gwgroup) fields
     *======================================================================*/
    if (isgwuser) {
	int i;
	char *tmp[8];

	/*===================================================================
	 * Convert java phone to c string
	 *==================================================================*/
	jstr = (*env)->GetObjectField(env, rec, phoneID);
	if ((*env)->GetStringUTFLength(env, jstr) > PHONESIZ - 1) {
	    java_errmsg("Phone Number too long");
	    returnrc(NONFATAL);
	}
	cp = (*env)->GetStringUTFChars(env, jstr, &isCopy);
	strcpy(pkt.gwu.phonenum, cp);
	(*env)->ReleaseStringUTFChars(env, jstr, cp);

	/*===================================================================
	 * Convert java groupname to c string
	 *==================================================================*/
	jstr = (*env)->GetObjectField(env, rec, groupID);
	if ((*env)->GetStringUTFLength(env, jstr) > GRPSIZ - 1) {
	    java_errmsg("Group Name too long");
	    returnrc(NONFATAL);
	}
	cp = (*env)->GetStringUTFChars(env, jstr, &isCopy);
	strcpy(pkt.gwu.groupname, cp);
	(*env)->ReleaseStringUTFChars(env, jstr, cp);

	/*===================================================================
	 * Convert java password to c string
	 *==================================================================*/
	jstr = (*env)->GetObjectField(env, rec, passwordID);
	if ((*env)->GetStringUTFLength(env, jstr) > PWDSIZ - 1) {
	    java_errmsg("Password too long");
	    returnrc(NONFATAL);
	}
	cp = (*env)->GetStringUTFChars(env, jstr, &isCopy);
	strcpy((char *)&(pkt.gwu.password), cp);
	(*env)->ReleaseStringUTFChars(env, jstr, cp);

	/*===================================================================
	 * Convert java key to c string
	 * If tokentype is 0 (None), we don't care whether a tokenkey is
	 * specified or not
	 *==================================================================*/
	jstr = (*env)->GetObjectField(env, rec, keyID);
	cp = (*env)->GetStringUTFChars(env, jstr, &isCopy);
	if (strlen(cp) > 0 || pkt.gwu.tokentype != 0) {
	    for (i=0; i<8; i++) {
		int x;
#ifndef MSDOS
		char *sp;
		if ((tmp[i] = strtok_r((char *)cp, " \t", &sp)) == NULL) {
#else
                if ((tmp[i] = strtok((char *)cp, " \t")) == NULL) {
#endif
		    java_errmsg("Token Key must have exactly 8 octets");
		    returnrc(NONFATAL);
		}
		if (i && !strcmp(tmp[i],tmp[i-1])) {
		    java_errmsg("Token Key must not have repeated values");
		    returnrc(NONFATAL);
		}
		sscanf(tmp[i], "%o", &x);
		pkt.gwu.userkey[i] = x;
		cp = NULL;

		if (!strcmp(tmp[i], "000") || strlen(tmp[i]) != 3 ||
		    *(tmp[i]) < '0' || *(tmp[i]) > '1' ||
		    *(tmp[i]+1) < '0' || *(tmp[i]+1) > '7' ||
		    *(tmp[i]+2) < '0' || *(tmp[i]+2) > '7') {
		    java_errmsg("Token key octets must be 3 digit octal "
			"numbers between 001 to 177.");
		    returnrc(NONFATAL);
		}
	    }
	    (*env)->ReleaseStringUTFChars(env, jstr, cp);
	    pkt.gwu.userkey[sizeof(pkt.gwu.userkey) - 1] = '\0';
	}

	/*===================================================================
	 * Convert java startdate to c
	 *==================================================================*/
	pkt.gwu.startdate = (*env)->GetIntField(env, rec, startdateID);

	/*===================================================================
	 * Convert java stopdate to c
	 *==================================================================*/
	pkt.gwu.stopdate = (*env)->GetIntField(env, rec, stopdateID);

	/*===================================================================
	 * Convert java boolean status to c
	 *==================================================================*/
	if ((*env)->GetBooleanField(env, rec, lockedID) == JNI_TRUE)
	    pkt.gwu.status = LOCKED;
	else
	    pkt.gwu.status = UNLOCKED;
    }

    /*=======================================================================
     * Send the request to the server
     *======================================================================*/
    pkt.req.msglen = sizeof(pkt.gwu);
    order_bytes_in_req(&(pkt.req));
    if (putdata(CGLUE_INT(fdID), (char *)&pkt,
			sizeof(CSREQ) + ntohl(pkt.req.msglen),
			DESKEY(ckeyschedpID),1) < 0) {
	java_syserrmsg("putdata() failed");
	returnrc(NONFATAL);
    }

    /*=======================================================================
     * Get the response from the server
     *======================================================================*/
    if ((rc = getdata(CGLUE_INT(fdID), (char *)&resp, sizeof(resp),
			DESKEY(ckeyschedpID),1)) < 0) {
	java_syserrmsg("getdata() failed");
	returnrc(NONFATAL);
    }
    if (rc == 0) {
	java_errmsg("Server has closed connection");
	returnrc(CLOSED);
    }
    order_bytes_in_resp(&resp);

    if (resp.rc != OK) {
	char buf[BUFSIZ];
	if (resp.msglen > 0 && getdata(CGLUE_INT(fdID), buf,
			resp.msglen,DESKEY(ckeyschedpID),1) > 0) {
	    java_errmsg(buf);
	}
	else
	    java_errmsg("getdata() failed");

	returnrc(resp.rc);
    }

    /*=======================================================================
     * Put success message in pktrc.str
     *======================================================================*/
    if (resp.msglen > 0) {
	char buf[BUFSIZ];
	if (getdata(CGLUE_INT(fdID), buf, resp.msglen,
			DESKEY(ckeyschedpID),1) < 0) {
	    java_errmsg("getdata() failed");
	    returnrc(NONFATAL);
	}
	jstr = (*env)->NewStringUTF(env, buf);
	(*env)->SetObjectField(env,pktrc, pktstrID, jstr);
    }

    returnrc(OK);
}

/*===========================================================================
 * This function is intended to be called from java to open a local file.
 *
 * Input:
 *	jfn is the file name in Java String
 *	bin is 0 if the file to be opened is an ascii file
 *	    is 1 if the file to be opened is a binary file
 *
 * Output:
 *	cglue.cbcblkp is set to point to the malloc'ed space for the structure
 *	returns 0 if OK
 *	returns -1 if fatal error
 *	returns -2 if EOF
 *	returns -3 if non fatal error
 *	If this function returns a negative rc, the error message is in
 *	the errmsg member
 * Note that the return value is a reference to a pktinfo instance.  In order to
 * access the pktinfo class from c, NewObject() will have to be used to obtain
 * a new instance of a class pktinfo object.  pktinfo contains: 
 * int rc	- is the integer return code
 * String str	- java string used by recvpkt to hold the data packet received
 *		  or used by getconf to hold the name of the temp file created
 * String errmsg- java string which hold an error message if rc is negative
 * These members can be accessed using the static field IDs initialized by
 * openfd.
 *		
 *==========================================================================*/
JNIEXPORT jobject JNICALL Java_cglue_opencbcfile(JNIEnv *env, jobject jo,
			jstring jfn, jint binflag) {
    int rc;
    int ifd;
    int nread;
    char fn[BUFSIZ];
    jstring jstr;
    const char *utfchar;
    jobject pktrc;
    jboolean isCopy;
    CBCBLK *cbcblkp;
    int O_mode = O_RDWR|O_CREAT;

    /*======================================================================
     * Establish addressability to the pktrc structure
     * Have to exit on error because we don't have the return structure to
     * return with.
     *=====================================================================*/
    pktrc = (*env)->NewObject(env, pktinfoClass, pktinfoID);
    if (pktrc == NULL) {
	fprintf(stderr,"pktinfo constructor failed");
	exit(-1);
    }

    /*=======================================================================
     * Malloc a CBCBLK structure
     *======================================================================*/
    if ((cbcblkp = malloc(sizeof(CBCBLK))) == NULL) {
	java_syserrmsg("malloc for cbc structure failed");
	returnrc(FAILED);
    }

    /*=======================================================================
     * Initialize the ivec for encryption/decryption
     *======================================================================*/
    memset(cbcblkp, 0, sizeof(CBCBLK));
    memcpy(cbcblkp->ivec, IVEC, sizeof(des_cblock));
    if (binflag) {
	cbcblkp->flag |= CBC_BIN;
    }
    /*=======================================================================
     * On Windows, O_BINARY has to be set whether the file is a binary file
     * or not.  This is the only way to bypass the silly Windows (DOS)
     * behavior which treats a x'0a' as an EOF.
     *======================================================================*/
#ifdef MSDOS
    O_mode |= O_BINARY;
#endif

    /*=======================================================================
     * Set the pointer value in the cglue class
     *======================================================================*/
    SET_CGLUE_INT(cbcblkpID, (int)cbcblkp);

    /*=======================================================================
     * Convert to c string
     *======================================================================*/
    utfchar = (*env)->GetStringUTFChars(env, jfn, &isCopy);
    strcpy(fn, utfchar);
    (*env)->ReleaseStringUTFChars(env, jfn, utfchar);

    if (strlen(fn) == 0) {
	java_errmsg("Inetranl error: empty filename for open");
	returnrc(NONFATAL);
    }

    if ((ifd = open(fn, O_mode, 0600)) < 0) {
	java_syserrmsg("Cannot open local file");
	returnrc(NONFATAL);
    }

    returnrc(ifd);
}

/*===========================================================================
 * This function is intended to be called from java to close a file
 * descriptor.
 *
 * Input:
 *	fd file descriptor to close
 *
 * Output:
 *==========================================================================*/
JNIEXPORT void JNICALL Java_cglue_closefile(JNIEnv *env, jobject jo, jint fd) {
    close(fd);
    if (CGLUE_INT(cbcblkpID) != 0)
	free((void *)CGLUE_INT(cbcblkpID));
    SET_CGLUE_INT(cbcblkpID, 0);
    return;
}

/*===========================================================================
 * This function is intended to be called from java to read and decrypt a
 * local file.  It returns the entire input file.
 *
 * Input:
 *	fd is the file descriptor of the file to be read
 *
 * Output:
 *	returns 0 if OK
 *	returns -1 if fatal error
 *	returns -2 if EOF
 *	returns -3 if non fatal error
 *	If this function returns a negative rc, the error message is in
 *	the errmsg member
 * Note that the return value is a reference to a pktinfo instance.  In order to
 * access the pktinfo class from c, NewObject() will have to be used to obtain
 * a new instance of a class pktinfo object.  pktinfo contains: 
 * int rc	- is the integer return code
 * String str	- java string used by recvpkt to hold the data packet received
 *		  or used by getconf to hold the name of the temp file created
 * String errmsg- java string which hold an error message if rc is negative
 * These members can be accessed using the static field IDs initialized by
 * openfd.
 *		
 *==========================================================================*/
JNIEXPORT jobject JNICALL Java_cglue_cbcreadFile(JNIEnv *env, jobject jo,
			jint fd) {
    int rc;
    int nread;
    int n = 0;
    int lineend = 0;
    struct des_ks_struct *dk;
    struct stat stat;
    unsigned char *bp;
    jstring jstr;
    jobject pktrc;

    /*======================================================================
     * Establish addressability to the pktrc structure
     * Have to exit on error because we don't have the return structure to
     * return with.
     *=====================================================================*/
    pktrc = (*env)->NewObject(env, pktinfoClass, pktinfoID);
    if (pktrc == NULL) {
	fprintf(stderr,"pktinfo constructor failed");
	exit(-1);
    }

    switch (CGLUE_INT(passed_authID)) {
	case 0:
	case 2:
	    dk = DESKEY(dkeyschedpID);
	    break;
	case 1:
	    dk = DESKEY(ckeyschedpID);
	    break;
    }

    if (fstat(fd, &stat) < 0) {
	java_syserrmsg("Cannot stat local file for upload");
	close(fd);
	returnrc(NONFATAL);
    }

    if ((bp = malloc(stat.st_size)) == NULL) {
	java_syserrmsg("Unable to malloc for editing a file");
	close(fd);
	returnrc(NONFATAL);
    }

    /*======================================================================
     * Loop to read in the entire file
     *=====================================================================*/
    while (n < stat.st_size) {
	char *nl;

	if ((nread = cbc_read(fd, &(bp[n]), sizeof(des_cblock), dk,
				(CBCBLK *)CGLUE_INT(cbcblkpID))) < 0) {
	    java_syserrmsg("decrypt/read error");
	    free(bp);
	    returnrc(NONFATAL);
	}

	/*===================================================================
	 * Stop at EOF in addition to stopping at EOL
	 *==================================================================*/
	if (nread == 0) {
	    bp[n] = '\0';
	    break;
	}

	/*===================================================================
	 * Increment length
	 *==================================================================*/
	n += nread;
    }

    /*=======================================================================
     * If n == 0 and nread = 0, then we just read the EOF and there is nothing
     * in the buffer - so we return an EOF.
     *======================================================================*/
    if (n == 0 && nread == 0) {
	free(bp);
	returnrc(CLOSED);
    }
    else {
	/*===================================================================
	 * Convert the line just read into Java for return
	 *==================================================================*/
	jstr = (*env)->NewStringUTF(env, (char *)bp);
	(*env)->SetObjectField(env, pktrc, pktstrID, jstr);
	free(bp);
	returnrc(OK);
    }
}

/*===========================================================================
 * This function is intended to be called from java to read and decrypt a
 * local file.  It returns a line of input each time, with the terminating
 * newline (or EOF) character stripped off.
 *
 * Input:
 *	fd is the file descriptor of the file to be read
 *
 * Output:
 *	returns 0 if OK
 *	returns -1 if fatal error
 *	returns -2 if EOF
 *	returns -3 if non fatal error
 *	If this function returns a negative rc, the error message is in
 *	the errmsg member
 * Note that the return value is a reference to a pktinfo instance.  In order to
 * access the pktinfo class from c, NewObject() will have to be used to obtain
 * a new instance of a class pktinfo object.  pktinfo contains: 
 * int rc	- is the integer return code
 * String str	- java string used by recvpkt to hold the data packet received
 *		  or used by getconf to hold the name of the temp file created
 * String errmsg- java string which hold an error message if rc is negative
 * These members can be accessed using the static field IDs initialized by
 * openfd.
 *		
 *==========================================================================*/
JNIEXPORT jobject JNICALL Java_cglue_cbcreadLine(JNIEnv *env, jobject jo,
			jint fd) {
    int rc;
    int nread;
    int n = 0;
    int lineend = 0;
    struct des_ks_struct *dk;
    unsigned char buf[BUFSIZ];
    jstring jstr;
    jobject pktrc;

    /*======================================================================
     * Establish addressability to the pktrc structure
     * Have to exit on error because we don't have the return structure to
     * return with.
     *=====================================================================*/
    pktrc = (*env)->NewObject(env, pktinfoClass, pktinfoID);
    if (pktrc == NULL) {
	fprintf(stderr,"pktinfo constructor failed");
	exit(-1);
    }

    switch (CGLUE_INT(passed_authID)) {
	case 0:
	case 2:
	    dk = DESKEY(dkeyschedpID);
	    break;
	case 1:
	    dk = DESKEY(ckeyschedpID);
	    break;
    }

    /*======================================================================
     * Loop to read one block at a time until end of line or end of file
     *=====================================================================*/
    while (n + sizeof(des_cblock) < sizeof(buf)) {
	char *nl;

	if ((nread = cbc_read(fd, &(buf[n]), sizeof(des_cblock), dk,
				(CBCBLK *)CGLUE_INT(cbcblkpID))) < 0) {
	    java_syserrmsg("decrypt/read error");
	    returnrc(NONFATAL);
	}

	/*===================================================================
	 * Stop at EOF in addition to stopping at EOL
	 *==================================================================*/
	if (nread == 0) {
	    buf[n] = '\0';
	    break;
	}

	/*===================================================================
	 * Stop at EOL
	 *==================================================================*/
	if ((nl = memchr(&buf[n], '\n', nread)) != NULL) {
	    *nl = '\0';
	    n += nread - 1;
	    break;
	}

	/*===================================================================
	 * Increment length
	 *==================================================================*/
	n += nread;
    }

    /*=======================================================================
     * If n == 0 and nread = 0, then we just read the EOF and there is nothing
     * in the buffer - so we return an EOF.
     *======================================================================*/
    if (n == 0 && nread == 0) {
	returnrc(CLOSED);
    }
    else {
	/*===================================================================
	 * Convert the line just read into Java for return
	 *==================================================================*/
	jstr = (*env)->NewStringUTF(env, (char *)buf);
	(*env)->SetObjectField(env, pktrc, pktstrID, jstr);
	returnrc(OK);
    }
}

/*===========================================================================
 * This function is intended to be called from java to encrypt and then
 * write a line to the file
 *
 * Input:
 *	fd is the file descriptor of the file to write to
 *	line in the java String to be encrypted/written
 *
 * Output:
 *	returns 0 if OK
 *	returns -1 if fatal error
 *	returns -2 if EOF
 *	returns -3 if non fatal error
 *	If this function returns a negative rc, the error message is in
 *	the errmsg member
 * Note that the return value is a reference to a pktinfo instance.  In order to
 * access the pktinfo class from c, NewObject() will have to be used to obtain
 * a new instance of a class pktinfo object.  pktinfo contains: 
 * int rc	- is the integer return code
 * String str	- java string used by recvpkt to hold the data packet received
 *		  or used by getconf to hold the name of the temp file created
 * String errmsg- java string which hold an error message if rc is negative
 * These members can be accessed using the static field IDs initialized by
 * openfd.
 *		
 *==========================================================================*/
JNIEXPORT jobject JNICALL Java_cglue_cbcwriteLine(JNIEnv *env, jobject jo,
			jint fd, jstring jline) {
    int nwrite;
    unsigned char line[BUFSIZ];
    const char *utfchar;
    jobject pktrc;
    struct des_ks_struct *dk;
    jboolean isCopy;

    /*======================================================================
     * Establish addressability to the pktrc structure
     * Have to exit on error because we don't have the return structure to
     * return with.
     *=====================================================================*/
    pktrc = (*env)->NewObject(env, pktinfoClass, pktinfoID);
    if (pktrc == NULL) {
	fprintf(stderr,"pktinfo constructor failed");
	exit(-1);
    }

    switch (CGLUE_INT(passed_authID)) {
	case 0:
	case 2:
	    dk = DESKEY(dkeyschedpID);
	    break;
	case 1:
	    dk = DESKEY(ckeyschedpID);
	    break;
    }

    /*=======================================================================
     * Make sure the line fits in our buffer
     *======================================================================*/
    if ((*env)->GetStringUTFLength(env, jline) > sizeof(line)) {
	java_errmsg("Line too long");
	returnrc(NONFATAL);
    }

    /*=======================================================================
     * Convert to c string
     *======================================================================*/
    utfchar = (*env)->GetStringUTFChars(env, jline, &isCopy);
    strcpy((char *)line, utfchar);
    (*env)->ReleaseStringUTFChars(env, jline, utfchar);

    /*=======================================================================
     * Encrypt/write the line
     *======================================================================*/
    nwrite = strlen((char *)line);
    if (cbc_write(fd, line, nwrite, dk, (CBCBLK *)CGLUE_INT(cbcblkpID))
						!= nwrite) {
	java_syserrmsg("Encrypt/write failed");
	returnrc(NONFATAL);
    }
    returnrc(OK);
}

/*===========================================================================
 * This function is intended to be called from java to get the value of an
 * environment variable
 *
 * Input:
 *	name is the name of the variable to be returned
 *
 * Output:
 *	returns 0 if OK
 *	returns -1 if fatal error
 *	returns -2 if EOF
 *	returns -3 if non fatal error
 *	If this function returns a negative rc, the error message is in
 *	the errmsg member
 * Note that the return value is a reference to a pktinfo instance.  In order to
 * access the pktinfo class from c, NewObject() will have to be used to obtain
 * a new instance of a class pktinfo object.  pktinfo contains: 
 * int rc	- is the integer return code
 * String str	- java string used by recvpkt to hold the data packet received
 *		  or used by getconf to hold the name of the temp file created
 * String errmsg- java string which hold an error message if rc is negative
 * These members can be accessed using the static field IDs initialized by
 * openfd.
 *		
 *==========================================================================*/
JNIEXPORT jobject JNICALL Java_cglue_getvar(JNIEnv *env, jobject jo,
		jstring jname) {
    char name[BUFSIZ];
    const char *utfchar;
    char *value;
    jobject pktrc;
    jboolean isCopy;
    jstring jstr;

    /*======================================================================
     * Establish addressability to the pktrc structure
     * Have to exit on error because we don't have the return structure to
     * return with.
     *=====================================================================*/
    pktrc = (*env)->NewObject(env, pktinfoClass, pktinfoID);
    if (pktrc == NULL) {
	fprintf(stderr,"pktinfo constructor failed");
	exit(-1);
    }

    /*=======================================================================
     * Make sure the line fits in our buffer
     *======================================================================*/
    if ((*env)->GetStringUTFLength(env, jname) > sizeof(name)) {
	java_errmsg("Env variable name too long");
	returnrc(NONFATAL);
    }

    /*=======================================================================
     * Convert to c string
     *======================================================================*/
    utfchar = (*env)->GetStringUTFChars(env, jname, &isCopy);
    strcpy((char *)name, utfchar);
    (*env)->ReleaseStringUTFChars(env, jname, utfchar);

#if defined(MSDOS)
    /*=======================================================================
     * Windows does not have env var HOME
     *======================================================================*/
    if (strcmp(name, "HOME") == 0) {
	value = "\\firewall";

	/*===================================================================
	 * Convert the value into Java for return
	 *==================================================================*/
	jstr = (*env)->NewStringUTF(env, value);
	(*env)->SetObjectField(env, pktrc, pktstrID, jstr);

	returnrc(OK);
    }
#endif

    /*=======================================================================
     * Get the environment variable
     *======================================================================*/
    value = getenv(name);

    /*===================================================================
     * Convert the value into Java for return
     *==================================================================*/
    jstr = (*env)->NewStringUTF(env, value);
    (*env)->SetObjectField(env, pktrc, pktstrID, jstr);

    returnrc(OK);
}

/*===========================================================================
 * This function is intended to be called from java to request the server to
 * send a tar backup of the firewall.
 *
 * Input:
 *	name is the name of the local file to stor ethe tar into
 *	flag indicates which backup list(s) to use
 *
 * Output:
 *	returns 0 if OK
 *	returns -1 if fatal error
 *	returns -2 if EOF
 *	returns -3 if non fatal error
 *	If this function returns a negative rc, the error message is in
 *	the errmsg member
 * Note that the return value is a reference to a pktinfo instance.  In order to
 * access the pktinfo class from c, NewObject() will have to be used to obtain
 * a new instance of a class pktinfo object.  pktinfo contains: 
 * int rc	- is the integer return code
 * String str	- java string used by recvpkt to hold the data packet received
 *		  or used by getconf to hold the name of the temp file created
 * String errmsg- java string which hold an error message if rc is negative
 * These members can be accessed using the static field IDs initialized by
 * openfd.
 *		
 *==========================================================================*/
JNIEXPORT jobject JNICALL Java_cglue_downloadtar(JNIEnv *env, jobject jo,
		jstring jfn, jint flag) {
    int fd;
    int rc;
    int total;
    int nread;
    int nwrite;
    int truesize;
    char msg[10];
    char fn[BUFSIZ];
    char buf[BUFSIZ];
    const char *utfchar;
    jobject pktrc;
    jboolean isCopy;
    jstring jstr;
    CBCBLK cbcblk;
    CBCBLK rcbcblk;
    SCRESP resp;
    struct {
	CSREQ req;
	char msg[10];
    } reqpkt;
    struct des_ks_struct *dk;	/* key schedule for socket	*/
    struct des_ks_struct *ldk;	/* key schedule for local files	*/
    int O_mode = O_WRONLY|O_CREAT;

#ifdef MSDOS
    O_mode |= O_BINARY;
#endif

    /*======================================================================
     * Establish addressability to the pktrc structure
     * Have to exit on error because we don't have the return structure to
     * return with.
     *=====================================================================*/
    pktrc = (*env)->NewObject(env, pktinfoClass, pktinfoID);
    if (pktrc == NULL) {
	fprintf(stderr,"pktinfo constructor failed");
	exit(-1);
    }

    /*======================================================================
     * Get key schedule for comminications with the server
     *=====================================================================*/
    if (CGLUE_INT(passed_authID))
	dk = DESKEY(ckeyschedpID);
    else
	dk = DESKEY(dkeyschedpID);

    /*======================================================================
     * Get key schedule for encrypting/decrypting local files
     *=====================================================================*/
    switch (CGLUE_INT(passed_authID)) {
	case 0:
	case 2:
	    ldk = DESKEY(dkeyschedpID);
	    break;
	case 1:
	    ldk = DESKEY(ckeyschedpID);
	    break;
    }

    /*=======================================================================
     * Make sure the name fits in our buffer
     *======================================================================*/
    if ((*env)->GetStringUTFLength(env, jfn) > sizeof(fn)) {
	java_errmsg("File name too long");
	returnrc(NONFATAL);
    }

    /*=======================================================================
     * Convert to c string
     *======================================================================*/
    utfchar = (*env)->GetStringUTFChars(env, jfn, &isCopy);
    strcpy((char *)fn, utfchar);
    (*env)->ReleaseStringUTFChars(env, jfn, utfchar);

    if (strlen(fn) == 0) {
	java_errmsg("Inetranl error: empty filename for open");
	returnrc(NONFATAL);
    }

    if ((fd = open(fn, O_mode, 0600)) < 0) {
	java_syserrmsg("Cannot open local file");
	returnrc(NONFATAL);
    }

    /*=======================================================================
     * Initialize the ivec for encryption/decryption
     *======================================================================*/
    memset((char *)&cbcblk, 0, sizeof(CBCBLK));
    memcpy(&cbcblk.ivec, IVEC, sizeof(des_cblock));
    cbcblk.flag = CBC_BIN;
    memset((char *)&rcbcblk, 0, sizeof(CBCBLK));
    memcpy(&rcbcblk.ivec, IVEC, sizeof(des_cblock));
    rcbcblk.flag = CBC_BIN;

    /*=======================================================================
     * Send request to server
     *======================================================================*/
    sprintf(reqpkt.msg, "%d", flag);
    reqpkt.req.component = CALL;
    reqpkt.req.cmd = BACKUP;
    reqpkt.req.msglen = strlen(reqpkt.msg);
    order_bytes_in_req(&reqpkt.req);

    if ((rc = putdata(CGLUE_INT(fdID),(char *)&reqpkt,
			sizeof(CSREQ) + strlen(reqpkt.msg),dk,1)) < 0) {
	java_syserrmsg("Failed to send BACKUP request to server");
	unlink(fn);
	returnrc(-1);
    }

    if ((rc = getdata(CGLUE_INT(fdID),(char *)&resp,sizeof(resp),dk,1)) < 0) {
	java_syserrmsg("Failed getting response for BACKUP request");
	unlink(fn);
	returnrc(-1);
    }
    if (rc == 0) {
	java_errmsg("Server has closed connection");
	unlink(fn);
	returnrc(-2);
    }
    order_bytes_in_resp(&resp);

    if (resp.rc != OK) {
	if ((rc = getdata(CGLUE_INT(fdID), buf, resp.msglen, dk,1)) < 0) {
	    java_syserrmsg("getdata() of error message failed");
	    unlink(fn);
	    returnrc(-1);
	}
	else {
	    java_errmsg(buf);
	    unlink(fn);
	    returnrc(resp.rc);
	}
    }

    /*=======================================================================
     * Get the contents of the file over
     * truesize if the msglen rounded up to the next 8-byte multiple
     *======================================================================*/
    truesize = ((resp.msglen + sizeof(des_cblock) - 1) / sizeof(des_cblock)) *
				sizeof(des_cblock);
    for (total=0; total < truesize; total += nread) {
	int n;
	int nshort = 0;
	char shorty[sizeof(des_cblock)];

	/*===================================================================
	 * Get a full buffer or whatever is left
	 *==================================================================*/
	if (truesize - total > sizeof(buf) - nshort)
	    n = sizeof(buf) - nshort;
	else
	    n = truesize - total - nshort;

	if (n != 0) {
/*
	    if ((nread = getsomedata(CGLUE_INT(fdID),&buf[nshort],n,dk)) < 0) {
*/
	    if ((nread = cbc_read(CGLUE_INT(fdID),
			(unsigned char *)&buf[nshort], n, dk, &rcbcblk)) < 0) {
		java_syserrmsg("Read error during file download");
		unlink(fn);
		returnrc(NONFATAL);
	    }
	    if (nread == 0) {
		java_syserrmsg("Server has closed connection");
		unlink(fn);
		returnrc(-2);
	    }
	}

	nread += nshort;
	nshort = 0;

	/*===================================================================
	 * Write max number of complete (8-byte) blocks
	 * except when we just read the last block, then we write all bytes.
	 *==================================================================*/
	if (total + nread == truesize)
	    nwrite = resp.msglen - total;	/* use real len of file */
	else
	    nwrite = (nread/sizeof(des_cblock)) * sizeof(des_cblock);

	if (nwrite < nread) {
	    nshort = nread - nwrite;
	    memcpy(shorty, &buf[nwrite], nshort);
	}

	if ((rc = cbc_write(fd, (unsigned char *)buf, nwrite, ldk, &cbcblk))
				!= nwrite) {
	    java_syserrmsg("Write to local file failed during file "
				"transfer");
	    unlink(fn);
	    returnrc(NONFATAL);
	}
    }
    close(fd);
    returnrc(0);
}

/*===========================================================================
 * This function is intended to be called from java to request the server to
 * restore files from a tar file which is to be uploaded from the client.
 *
 * Input:
 *	name is the name of the local file to be uploaded
 *	args is a string containing names of files (within the tar) to be
 *		restored.
 *
 * Output:
 *	returns 0 if OK
 *	returns -1 if fatal error
 *	returns -2 if EOF
 *	returns -3 if non fatal error
 *	If this function returns a negative rc, the error message is in
 *	the errmsg member
 * Note that the return value is a reference to a pktinfo instance.  In order to
 * access the pktinfo class from c, NewObject() will have to be used to obtain
 * a new instance of a class pktinfo object.  pktinfo contains: 
 * int rc	- is the integer return code
 * String str	- java string used by recvpkt to hold the data packet received
 *		  or used by getconf to hold the name of the temp file created
 * String errmsg- java string which hold an error message if rc is negative
 * These members can be accessed using the static field IDs initialized by
 * openfd.
 *		
 *==========================================================================*/
JNIEXPORT jobject JNICALL Java_cglue_uploadtar(JNIEnv *env, jobject jo,
		jstring jfn, jstring jargs) {
    int fd;
    int rc;
    int total;
    int nread;
    int nwrite;
    int arglength;
    char *args;
    char fn[BUFSIZ];
    char buf[BUFSIZ];
    const char *utfchar;
    jobject pktrc;
    jboolean isCopy;
    jstring jstr;
    CBCBLK cbcblk;
    CBCBLK rcbcblk;
    struct stat stat;
    CSREQ req;
    SCRESP resp;
    struct des_ks_struct *dk;	/* key schedule for socket	*/
    struct des_ks_struct *ldk;	/* key schedule for local files	*/
    int O_mode = O_RDONLY;
#ifdef MSDOS
    O_mode |= O_BINARY;
#endif

    /*======================================================================
     * Establish addressability to the pktrc structure
     * Have to exit on error because we don't have the return structure to
     * return with.
     *=====================================================================*/
    pktrc = (*env)->NewObject(env, pktinfoClass, pktinfoID);
    if (pktrc == NULL) {
	fprintf(stderr,"pktinfo constructor failed");
	exit(-1);
    }

    /*======================================================================
     * Get key schedule for comminications with the server
     *=====================================================================*/
    if (CGLUE_INT(passed_authID))
	dk = DESKEY(ckeyschedpID);
    else
	dk = DESKEY(dkeyschedpID);

    /*======================================================================
     * Get key schedule for encrypting/decrypting local files
     *=====================================================================*/
    switch (CGLUE_INT(passed_authID)) {
	case 0:
	case 2:
	    ldk = DESKEY(dkeyschedpID);
	    break;
	case 1:
	    ldk = DESKEY(ckeyschedpID);
	    break;
    }

    /*=======================================================================
     * Make sure the name fits in our buffer
     *======================================================================*/
    if ((*env)->GetStringUTFLength(env, jfn) > sizeof(fn)) {
	java_errmsg("Fiel name too long");
	returnrc(NONFATAL);
    }

    /*=======================================================================
     * Convert to c string
     *======================================================================*/
    utfchar = (*env)->GetStringUTFChars(env, jfn, &isCopy);
    strcpy((char *)fn, utfchar);
    (*env)->ReleaseStringUTFChars(env, jfn, utfchar);

    if (strlen(fn) == 0) {
	java_errmsg("Inetranl error: empty filename for open");
	returnrc(NONFATAL);
    }

    if ((fd = open(fn, O_mode, 0)) < 0) {
	java_syserrmsg("Cannot open local file");
	returnrc(NONFATAL);
    }

    if (fstat(fd, &stat) < 0) {
	java_syserrmsg("Cannot stat local file for upload");
	close(fd);
	returnrc(NONFATAL);
    }

    /*=======================================================================
     * Allocate a buffer for the args
     *======================================================================*/
    arglength = (*env)->GetStringUTFLength(env, jargs) + 1;
    if ((args = malloc(arglength)) == NULL) {
	java_syserrmsg("malloc for RESTORE args failed");
	returnrc(NONFATAL);
    }

    /*=======================================================================
     * Convert args to c string
     *======================================================================*/
    utfchar = (*env)->GetStringUTFChars(env, jargs, &isCopy);
    strcpy(args, utfchar);
    (*env)->ReleaseStringUTFChars(env, jargs, utfchar);

    /*=======================================================================
     * Initialize the ivec for encryption/decryption
     *======================================================================*/
    memset((char *)&cbcblk, 0, sizeof(CBCBLK));
    memcpy(&cbcblk.ivec, IVEC, sizeof(des_cblock));
    cbcblk.flag = CBC_BIN;
    memset((char *)&rcbcblk, 0, sizeof(CBCBLK));
    memcpy(&rcbcblk.ivec, IVEC, sizeof(des_cblock));
    rcbcblk.flag = CBC_BIN;

    /*======================================================================
     * The restore protocol is unusual in that the client first sends a
     * putconf request packet with msglen=0.  This makes sure that the
     * server is OK for the upload.  The server sends a response packet
     * indicating it is OK.  The client then send a second request packet,
     * this time with msglen set to the length of the upload.
     *=====================================================================*/
    req.component = CALL;
    req.cmd = RESTORE;
    req.msglen = arglength;
    order_bytes_in_req(&req);

    if ((rc = putdata(CGLUE_INT(fdID), (char *)&req, sizeof(req), dk,1)) < 0) {
	close(fd);
	free(args);
	java_syserrmsg("archive upload failed");
	returnrc(NONFATAL);
    }

    /*=======================================================================
     * Send the args to server
     *======================================================================*/
    if (arglength > 0) {
	if ((rc = putdata(CGLUE_INT(fdID), args, arglength, dk,1)) < 0) {
	    close(fd);
	    free(args);
	    java_syserrmsg("archive upload failed");
	    returnrc(NONFATAL);
	}
    }

    free(args);
    if ((rc = getdata(CGLUE_INT(fdID), (char *)&resp, sizeof(resp),
			dk,1)) < 0) {
	close(fd);
	java_syserrmsg("Request to upload archive failed");
	returnrc(NONFATAL);
    }

    if (rc == 0) {
	java_errmsg("Server has closed connection");
	close(fd);
	returnrc(-2);
    }
    order_bytes_in_resp(&resp);

    if (resp.rc != OK) {
	close(fd);
	if ((rc = getdata(CGLUE_INT(fdID), buf, resp.msglen, dk,1)) < 0) {
	    java_syserrmsg("Restore request failed");
	    returnrc(resp.rc);
	}
	else {
	    java_errmsg(buf);
	    returnrc(resp.rc);
	}
    }

    /*======================================================================
     * This is the second request packet
     * Note that we don't need to worry about short blocks here (not like
     * in downloadtar()) since we are reading from a local file and not
     * from the socket.
     *=====================================================================*/
    req.msglen = htonl(stat.st_size);
    if ((rc = putdata(CGLUE_INT(fdID), (char *)&req, sizeof(req), dk,1)) < 0) {
	java_syserrmsg("Second restore request failed");
	close(fd);
	returnrc(NONFATAL);
    }

    for (total = 0; total < stat.st_size; total += nread) {
	if ((nread = cbc_read(fd,(unsigned char *)buf,sizeof(buf),ldk,
							&rcbcblk)) < 0) {
	    java_syserrmsg("Read error during file upload");
	    close(fd);
	    returnrc(NONFATAL);
	}
	if (nread == 0) {
	    java_syserrmsg("Unexpected EOF from local archive file");
	    close(fd);
	    returnrc(NONFATAL);
	}

	if ((rc = cbc_write(CGLUE_INT(fdID), (unsigned char *)buf, nread, dk,
				&rcbcblk)) != nread) {
	    java_syserrmsg("Remote write failed failed during file upload");
	    close(fd);
	    returnrc(NONFATAL);
	}
    }
    close(fd);
    returnrc(OK);
}

#define freeretinfo()	{for (retlast = rettop; retlast;) { \
				rettmp = retlast->next; \
				free(retlast); \
				retlast = rettmp; \
	    		}}
#define returntar(a)	{	(*env)->SetIntField(env,pktrc,tarrcID,a); \
				return(pktrc); \
			}

/*===========================================================================
 * This function is intended to be called from java to read an encrypted tar
 * file which is stored on the local system.
 *
 * Input:
 *	name is the name of the local file to be read
 *
 * Output:
 *	returns 0 if OK
 *	returns -1 if fatal error
 *	returns -2 if EOF
 *	returns -3 if non fatal error
 *	If this function returns a negative rc, the error message is in
 *	the errmsg member
 * Note that the return value is a reference to a pktinfo instance.  In order to
 * access the pktinfo class from c, NewObject() will have to be used to obtain
 * a new instance of a class pktinfo object.  pktinfo contains: 
 * int rc	- is the integer return code
 * String str	- java string used by recvpkt to hold the data packet received
 *		  or used by getconf to hold the name of the temp file created
 * String errmsg- java string which hold an error message if rc is negative
 * These members can be accessed using the static field IDs initialized by
 * openfd.
 *		
 *==========================================================================*/
JNIEXPORT jobject JNICALL Java_cglue_readtar(JNIEnv *env, jobject jo,
		jstring jfn) {
    int i;
    int fd;
    int rc;
    int size;
    int total;
    int count = 0;
    int nread;
    int zero = 0;
    int arglength;
    char *args;
    char fn[BUFSIZ];
    const char *utfchar;
    jobject pktrc;
    jint *localSizes;
    jint *localDates;
    jobject jstrobj;
    jboolean isCopy;
    CBCBLK cbcblk;
    struct stat stat;
    SCRESP resp;
    struct des_ks_struct *ldk;	/* key schedule for local files	*/
    typedef struct retinfo RETINFO;
    struct retinfo {
	int	size;
	time_t	moddate;
	RETINFO	*next;
	char	name[1];	/* place holder for variable length string */
    };
    RETINFO *rettop = NULL;
    RETINFO *rettmp = NULL;
    RETINFO *retlast = NULL;
    struct TAR {
	char	name[100];
	char	mode[8];
	char	uid[8];
	char	gid[8];
	char	size[12];
	char	mtime[12];
	char	cksum[8];
	char	typeflag[1];
	char	linkname[100];
	char	magic[TMAGLEN];
	char	version[2];
	char	uname[32];
	char	gname[32];
	char	devmajor[8];
	char	devminor[8];
	char	prefix[155];
	char	filler[12];
    };
    typedef	struct TAR TAR;
    TAR	tarhdr;
    jclass tarinfoClass;
    jmethodID tarinfoID;
    jfieldID tarrcID;
    jfieldID tarnamesID;
    jfieldID tarsizesID;
    jfieldID tardatesID;
    jfieldID taremsgID;
    jobjectArray jnames;
    jintArray jsizes;
    jintArray jdates;
    int O_mode = O_RDONLY;

#ifdef MSDOS
    O_mode |= O_BINARY;
#endif

    tarinfoClass = (*env)->FindClass(env, "tarinfo");
    if (tarinfoClass == NULL) {
	fprintf(stderr,"Could not find class tarinfo");
	exit(-1);
    }

    tarrcID = (*env)->GetFieldID(env, tarinfoClass, "rc", "I");
    tarnamesID = (*env)->GetFieldID(env, tarinfoClass, "names",
					"[Ljava/lang/String;");
    tarsizesID = (*env)->GetFieldID(env, tarinfoClass, "sizes", "[I");
    tardatesID = (*env)->GetFieldID(env, tarinfoClass, "dates", "[I");
    taremsgID = (*env)->GetFieldID(env, tarinfoClass, "errmsg",
					"Ljava/lang/String;");

    tarinfoID = (*env)->GetMethodID(env, tarinfoClass, "<init>","()V");
    if (tarinfoID == NULL) {
	fprintf(stderr,"Could not find constructor for tarinfo");
	exit(-1);
    }
    /*======================================================================
     * Establish addressability to the tarrc structure
     * Have to exit on error because we don't have the return structure to
     * return with.
     *=====================================================================*/
    pktrc = (*env)->NewObject(env, tarinfoClass, tarinfoID);
    if (pktrc == NULL) {
	fprintf(stderr,"tarinfo constructor failed");
	exit(-1);
    }

    /*======================================================================
     * Get key schedule for encrypting/decrypting local files
     *=====================================================================*/
    switch (CGLUE_INT(passed_authID)) {
	case 0:
	case 2:
	    ldk = DESKEY(dkeyschedpID);
	    break;
	case 1:
	    ldk = DESKEY(ckeyschedpID);
	    break;
    }

    /*=======================================================================
     * Make sure the name fits in our buffer
     *======================================================================*/
    if ((*env)->GetStringUTFLength(env, jfn) > sizeof(fn)) {
	java_errmsg("File name too long");
	returntar(NONFATAL);
    }

    /*=======================================================================
     * Convert to c string
     *======================================================================*/
    utfchar = (*env)->GetStringUTFChars(env, jfn, &isCopy);
    strcpy((char *)fn, utfchar);
    (*env)->ReleaseStringUTFChars(env, jfn, utfchar);

    if (strlen(fn) == 0) {
	java_errmsg("Inetranl error: empty filename for open");
	returntar(NONFATAL);
    }

    if ((fd = open(fn, O_mode, 0)) < 0) {
	java_syserrmsg("Cannot open local file");
	returntar(NONFATAL);
    }

    if (fstat(fd, &stat) < 0) {
	java_syserrmsg("Cannot stat local file for upload");
	close(fd);
	returntar(NONFATAL);
    }

    /*=======================================================================
     * Initialize the ivec for encryption/decryption
     *======================================================================*/
    memset((char *)&cbcblk, 0, sizeof(CBCBLK));
    memcpy(&cbcblk.ivec, IVEC, sizeof(des_cblock));
    cbcblk.flag = CBC_BIN;

    /*=======================================================================
     * Each iteration through this loop reads one tar'ed file within the tar
     *======================================================================*/
    while (1) {
	int i;
	int allzeros;
	char *p = (char *)&tarhdr;
	/*===================================================================
	 * Read the next header
	 *==================================================================*/
	if ((nread = cbc_read(fd,(unsigned char *)&tarhdr,sizeof(tarhdr),ldk,
							&cbcblk)) < 0) {
	    java_syserrmsg("Read error reading local archive file");
	    freeretinfo();
	    close(fd);
	    returntar(NONFATAL);
	}

	/*===================================================================
	 * See if this is an all zero block
	 *==================================================================*/
	allzeros = 1;
	for (i = 0; i < sizeof(tarhdr); i += sizeof(zero)) {
	    if (memcmp(&zero, p + i, sizeof(zero)) != 0) {
		allzeros = 0;
		break;
	    }
	}
	if (allzeros)
	    break;

	/*===================================================================
	 * Make sure we read the right stuff and the file is not corrupted
	 *==================================================================*/
	if (strcmp(tarhdr.magic, TMAGIC) != 0) {
	    java_errmsg("Archive file corrupted");
	    freeretinfo();
	    close(fd);
	    returntar(NONFATAL);
	}

	/*===================================================================
	 * Store relevant header info into temp c linked list
	 *==================================================================*/
	if ((rettmp = malloc(sizeof(RETINFO) + strlen(tarhdr.name))) == NULL) {
	    java_syserrmsg("malloc failed to get storage for archive header");
	    freeretinfo();
	    returntar(FAILED);
	}

	if (rettop == NULL) {
	    rettop = rettmp;
	    retlast = rettmp;
	}
	else {
	    retlast->next = rettmp;
	    retlast = rettmp;
	}

	count++;
	retlast->size = strtoul(tarhdr.size, NULL, 8);
	retlast->moddate = strtoul(tarhdr.mtime, NULL, 8);
	retlast->next = NULL;
	strcpy(retlast->name, tarhdr.name);

	size = strtoul(tarhdr.size, NULL, 8);

	/*===================================================================
	 * Skip the file contents to get to the next header
	 *==================================================================*/
	for (total = 0; total < size; total += nread) {
	    char buf[512];	/* size of one tar block */
	    if ((nread = cbc_read(fd,(unsigned char *)buf,sizeof(buf),ldk,
							&cbcblk)) < 0) {
		java_syserrmsg("Read error reading local archive file");
		freeretinfo();
		close(fd);
		returntar(NONFATAL);
	    }

	    if (nread == 0) {
		java_syserrmsg("Unexpected EOF from local archive file");
		freeretinfo();
		close(fd);
		returntar(NONFATAL);
	    }
	}
    }
    close(fd);

    /*=======================================================================
     * Now we know the length, we can create the java arrays to hold the
     * names and dates
     *======================================================================*/
    jnames = (*env)->NewObjectArray(env, count,
			(*env)->FindClass(env, "java/lang/String"), NULL);
    jdates = (*env)->NewIntArray(env, count);
    jsizes = (*env)->NewIntArray(env, count);
    localSizes = (*env)->GetIntArrayElements(env, jsizes, NULL);
    localDates = (*env)->GetIntArrayElements(env, jdates, NULL);

    /*=======================================================================
     * Transfer the info to the java arrays
     *======================================================================*/
    retlast = rettop;
    for (i = 0; i < count; i++) {
	jstrobj = (*env)->NewStringUTF(env, retlast->name);
	(*env)->SetObjectArrayElement(env, jnames, i, jstrobj);
	localSizes[i] = retlast->size;
	localDates[i] = retlast->moddate;
	rettmp = retlast;
	retlast = retlast->next;
	free(rettmp);
    }
    (*env)->ReleaseIntArrayElements(env, jsizes, localSizes, 0);
    (*env)->ReleaseIntArrayElements(env, jdates, localDates, 0);

    /*=======================================================================
     * Setup the java arrays for return to java
     *======================================================================*/
    (*env)->SetObjectField(env,pktrc,tarnamesID,jnames);
    (*env)->SetObjectField(env,pktrc,tarsizesID,jsizes);
    (*env)->SetObjectField(env,pktrc,tardatesID,jdates);

    returntar(OK);
}
