/*============================================================================
 * filename:		win32/ptelnet/ansiprsr.cpp
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
///////////////////////////////////////////////////////////////////////////////
//
// Contents:	ANSI parser base class
//
// Product:		telnet
//
// Revisions:	13.Jul.1995	igor.milavec@uni-lj.si
//					Original code
//
///////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <ansiprsr.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "tn.h"
// #include <winsock.h>
#include "telnet.h"
#include "firewall.h"
#include "des.h"

extern "C" {
int gettnkey(char *name,des_cblock *key);
int enc_write(int _fd, unsigned char *_buf, int _len, des_key_schedule _keysched);
};
int secrets=0;
des_key_schedule	keysched;

char* TANSIParser::GetTerminalId()
{
	return "\033[?1;2c";
}

unsigned char* TANSIParser::ParseIAC(unsigned char* pszBuffer, unsigned char* pszBufferEnd, SOCKET Socket)
{
   register int skip=1;      // skip at least 1 character
   unsigned char xuf[UNMSIZ+1];
   int i,nchars;
   des_cblock key;

   nchars = pszBufferEnd - pszBuffer;
	if (nchars > 1) {  //  At least 2 chars?
      switch (pszBuffer[1]) {
        case WILL:
        case WONT:
        	if (nchars >= 3)
         	skip = 3;
         else
         	skip = 2;
         break;
        case DONT:
        // Only send back WONT if it's not for TTYPE
        	if (nchars >= 3)
         {
         	if (pszBuffer[2] != TELOPT_TTYPE)
            {
            	xuf[0] = IAC;
            	xuf[1] = WONT;
            	xuf[2] = pszBuffer[2];
            	if (secrets)
            		enc_write(Socket, (char *)xuf, 3, keysched);
            	else
            		send(Socket, (char *)xuf, 3, 0);
        		}
         	skip = 3;
         }
         else
         	skip = 2;
         break;
		  case DO:
         if (nchars >= 3) {	//
            switch (pszBuffer[2]) {
             case TELOPT_TTYPE:
              xuf[0] = IAC;
              xuf[1] = WILL;
              xuf[2] = TELOPT_TTYPE;
              if (secrets)
              	enc_write(Socket, (char *)xuf, 3, keysched);
              else
              	send(Socket, (char *)xuf, 3, 0);
              skip = 3;
              break;
             case TELOPT_ENCRYPT:
              xuf[0] = IAC;
              xuf[1] = WILL;
              xuf[2] = TELOPT_ENCRYPT;
              send(Socket, (char *)xuf, 3, 0);
              skip = 3;
              break;
             case TELOPT_NAWS:
				  {
              		COORD Coord;
             		unsigned char xuf[] = {IAC,WILL,TELOPT_NAWS,IAC,SB,TELOPT_NAWS,
                  								0,0,0,0,IAC,SE};

                  ConGetdwSize(&Coord);
                  xuf[7] = Coord.X;
                  xuf[9] = Coord.Y;

              		skip = 3;
               	if (secrets)
                  	enc_write(Socket, (char *)xuf, sizeof(xuf), keysched);
                  else
                  	send(Socket, (char *)xuf, sizeof(xuf), 0);
                  break;
              }
             default:
              skip = 3;
              break;
           }
         }
         else
           skip = 2;		// found IAC,(DO,DONT,WILL,WONT) but not anything else
			break;

        case SB:
         if (nchars > 3) {  // At least 5 chars (IAC,SB,TELOPT_ENCRYPT,
         						 //  TELOPT_ENC_KEYID + 1)?
           switch (pszBuffer[2]) {
             case TELOPT_ENCRYPT:
              if (pszBuffer[3] == ENCRYPT_ENC_KEYID)
              {
                unsigned char enc_ok[7]={IAC,SB,TELOPT_ENCRYPT,ENCRYPT_ENC_KEYID,
                								  0,IAC,SE};
                for (i=0;i<UNMSIZ;i++)
                  if (nchars > i+4 && pszBuffer[4+i] != IAC)
                    xuf[i] = pszBuffer[4+i];
                  else
                    break;
                xuf[i] = '\0';		// terminate the username
                if (gettnkey((char *)xuf,&key))
                  exit(1);
                secrets=1;
                des_set_key(&key,keysched);
                send(Socket,(char *)enc_ok,7,0);
                skip = 4 + i;
              }
              break;

             case TELOPT_TTYPE:
              if (pszBuffer[3] == TELQUAL_SEND)
              {
                unsigned char ttype[]={IAC,SB,TELOPT_TTYPE,TELQUAL_IS,
                								'A','N','S','I',IAC,SE};
                if (secrets)
                	enc_write(Socket, (char *)ttype, sizeof(ttype), keysched);
                else
                	send(Socket, (char *)ttype, sizeof(ttype), 0);
                skip = 4;
              }
              else
                skip = 3;
              break;
             default:
              i = 0;
              while (nchars > i+2 && pszBuffer[i+2] != IAC)
                i++;
              skip = i + 2;
           }
         }
         else
           skip = nchars;
         break;
        case SE:
         if (nchars > 1)
           skip = 2;		// skip IAC SE
         break;

    	}
	}
   if (pszBuffer + skip < pszBufferEnd)
     pszBuffer += skip;
   else
     pszBuffer = pszBufferEnd;

	return pszBuffer;
}

#pragma argsused
unsigned char* TANSIParser::ParseEscapeANSI(unsigned char* pszBuffer, unsigned char* pszBufferEnd)
{
	int bExtMode = 0;

	const int nParam = 10;	// Maximum number of parameters
	int iParam[nParam] = {0, 0, 0, 0, 0};
	int iCurrentParam = 0;

	// Get parameters from escape sequence.
	while (!isalpha(*pszBuffer)) {

		// Check for parameter delimiter.
		if (*pszBuffer == ';') {
			pszBuffer++;
			continue;
		}
		if (*pszBuffer == '?') {
			bExtMode = 1;
			pszBuffer++;
			continue;
		}

		// pszHead should point to digit now. Otherwise we got bad escape
		// sequence, so we just get out of here!
		if (!isdigit(*pszBuffer))
			return pszBuffer;

		iParam[iCurrentParam] = strtoul((char *)pszBuffer, (char **)&pszBuffer, 10);

		if (iCurrentParam < nParam)
			iCurrentParam++;
	}

	switch (*pszBuffer++) {

	  // Set cursor position.
	  case 'H':
	  case 'f':
		if (iCurrentParam < 2)
			iParam[1] = 1;
		if (iCurrentParam < 1)
			iParam[0] = 1;
		ConSetCursorPosition(iParam[1] - 1,	iParam[0] - 1);
		break;

	  // Move cursor up.
	  case 'A':
		if (iCurrentParam < 1)
			iParam[0] = 1;
		ConMoveCursorPosition(0, -iParam[0]);
		break;

	  // Move cursor down.
	  case 'B':
		if (iCurrentParam < 1)
			iParam[0] = 1;
		ConMoveCursorPosition(0, iParam[0]);
		break;

	  // Move cursor right.
	  case 'C':
		if (iCurrentParam < 1)
			iParam[0] = 1;
		ConMoveCursorPosition(iParam[0], 0);
		break;

	  // Move cursor left.
	  case 'D':
		if (iCurrentParam < 1)
			iParam[0] = 1;
		ConMoveCursorPosition(-iParam[0], 0);
		break;

	  case 'J': {	// Clear screen
		if (iCurrentParam < 1)
			iParam[0] = 0;
		switch (iParam[0]) {
		  case 0:
			ConClearEOScreen();
			break;
		  case 1:
			ConClearBOScreen();
			break;
		  case 2:
			ConClearScreen();
			break;
		}
		break;
	  }

	  case 'K':	{	// Erase to End of Line
      ConClearEOLine();
      break;
     }

     case 'L':	{	// Insert line (used by vi reverse scroll)
		if (iCurrentParam < 1)
			iParam[0] = 1;
      ConInsertLine(iParam[0]);
      break;
     }

	  case 'M':	{	// Delete line
		if (iCurrentParam < 1)
			iParam[0] = 1;
		ConDeleteLine(iParam[0]);
		break;
	  }

     case 'P': {	// Delete the character at cursor
      ConDeleteChar();
      break;
     }

	  case 'c': {
		char* szTerminalId = GetTerminalId();
		NetWriteString(szTerminalId, strlen(szTerminalId));
		break;
	  }

	  case 'h': {
		if (bExtMode)
			for (int i = 0; i < iCurrentParam; i++)
				ConSetExtendedMode(iParam[i], 1);
		break;
	  }

	  case 'l': {
		if (bExtMode)
			for (int i = 0; i < iCurrentParam; i++)
				ConSetExtendedMode(iParam[i], 0);
		break;
	  }

	}

	return pszBuffer;
}

unsigned char* TANSIParser::ParseEscape(unsigned char* pszBuffer, unsigned char* pszBufferEnd)
{
/*	ConWriteString("\033", 1);
	return ++pszBuffer; //*/

	// Check if we have enough characters in buffer.
	if ((pszBufferEnd - pszBuffer) < 2)
		return pszBuffer;

	// Decode the command...
	pszBuffer++;
	switch (*pszBuffer++) {
	  case '[': {

		// Check if we have whole escape sequence in buffer.
		//unsigned char* pszChar = pszBuffer + 1;
      unsigned char* pszChar = pszBuffer;
		while ((pszChar < pszBufferEnd) && (!isalpha(*pszChar)))
			pszChar++;
		if (pszChar == pszBufferEnd)
			pszBuffer -= 2;
		else
			pszBuffer = ParseEscapeANSI(pszBuffer, pszBufferEnd);
		break;
	  }

	  case ')':
		pszBuffer++;
		break;

	  case 'Z': {
		char* szTerminalId = GetTerminalId();
		NetWriteString(szTerminalId, strlen(szTerminalId));
		break;
	  }
	}

	return pszBuffer;
}

unsigned char* TANSIParser::ParseBuffer(unsigned char* pszBuffer, unsigned char* pszBufferEnd, SOCKET Socket)
{
	switch (*pszBuffer) {

	  case 0:
		pszBuffer++;
		break;

	  case 12:
		pszBuffer++;
		ConClearScreen();
		ConSetCursorPosition(0, 0);
		break;

	  case 27:
		pszBuffer = ParseEscape(pszBuffer, pszBufferEnd);
		break;

	  case 255:
		pszBuffer = ParseIAC(pszBuffer, pszBufferEnd, Socket);
		break;

     case '\n':
     case '\r':
     	pszBuffer = ConCRLF(pszBuffer,pszBufferEnd);
      break;

     default: {
		unsigned char* pszCurrent = pszBuffer + 1;
		while ((pszCurrent < pszBufferEnd) && !iscntrl(*pszCurrent) &&
				(*pszCurrent != (unsigned char)255))
			pszCurrent++;

		pszBuffer += ConWriteString((char *)pszBuffer, (int)pszCurrent - (int)pszBuffer);
		break;
	  }
	}

	return pszBuffer;
}


