/*============================================================================
 * filename:		libsrc/enc_io.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*====================================================================== 
 *
 * (C) COPYRIGHT R. James Livermore 1996-2000.  All rights reserved.
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1996-2000.
 * All rights reserved.   
 *
 * Date written:	July      16, 1996
 * Last modified:	February   8, 2000 - added LPL
 *	
 * functions:
 *		enc_read()
 *		enc_write()
 *		gettnkey()
 * 
 *=====================================================================*/  
static	char	cprght[] = 
  		"(C) COPYRIGHT Freemont Avenue Software, Inc. 1996-2000 enc_io,"
  		" V 1.0 00/01/21"; 

#ifndef MSDOS
# ifndef __hppa
#  include <sys/select.h>
# else
#  include <netinet/in.h>
# endif
# include <syslog.h>
#else
# include <dos.h>
# include <io.h>
# include <winsock.h>
#endif
#include <errno.h>
#include <limits.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "firewall.h"
#include "tnkey.h"
#include "des.h"
#ifndef MSDOS
# include "global.h"
# include "md5.h"
#endif

#ifndef	linux
extern char *sys_errlist[];
#else
extern const char *const sys_errlist[];
#endif

#ifdef TREX_ENC
#ifdef MSDOS
#undef BUFSIZ
#define BUFSIZ 4096
#undef LINE_MAX
#define LINE_MAX 1024
#endif /* MSDOS */
#define		LEN_BYTE_MASK		0x80
extern		int			simple_des;
int		cleft_over=0;
int		sleft_over=0;

/*=======================================================================
 * enc_read()
 *
 * Read encrypted data
 *
 *		This function keeps 2 static buffers to hold left 
 *
 * input:
 *	fd	file descriptor to read from
 *	buf	pointer to data buffer
 *	len	number of bytes to read
 *	keysched des_key_schedule to use
 *
 * output:
 *	returns the number of bytes (clear text) that were put into the
 *		buf buffer.
 *	returns -1 if failed.
 *======================================================================*/
int enc_read(int fd, unsigned char *buf, int len, des_key_schedule keysched)
{
    static unsigned char	cprevbuf[BUFSIZ];
    static unsigned char	sprevbuf[BUFSIZ];
    static unsigned char	cshortblk[sizeof(des_cblock)];
    static unsigned char	sshortblk[sizeof(des_cblock)];
    static int			cshort=0;
    static int			sshort=0;

    unsigned char		ebuf[BUFSIZ];
    unsigned char		*tb,tbuf[sizeof(des_cblock)];
    int				n=0,nread=0,n8,readlen,done_read;
    unsigned char		block_len;
    unsigned char		*prev_buf;
    unsigned char		*shortblk;
    int				*left_over;
    int				*nshort;

    /*=======================================================================
     * Make sure we can handle the data length.
     *======================================================================*/
    if (len > sizeof(ebuf))
    {
	printf("Buffer limit exceeded trying to read %d bytes.\n",len);
#ifndef MSDOS
	syslog(LOG_ERR,"Error: Buffer limit exceeded trying to read %d bytes.\n",len);
#endif
	return(-1);
    }

    /*=======================================================================
     * Determine whether we should use the client or server buffers
     *======================================================================*/
    if (fd == 0)
    {
	prev_buf  = cprevbuf;
	left_over = &cleft_over;
	shortblk  = cshortblk;
	nshort    = &cshort;
    }
    else
    {
	prev_buf  = sprevbuf;
	left_over = &sleft_over;
	shortblk  = sshortblk;
	nshort    = &sshort;
    }

    /*=======================================================================
     * If there are left over bytes to hand back to caller, use them first.
     *======================================================================*/
    if (*left_over > 0)
    {
	if (*left_over < len)
	{
	    memcpy(buf,prev_buf,*left_over);
	    n = *left_over;
	    *left_over = 0;
	}
	else
	{
	    /*===============================================================
	     * We have more left_over bytes than are asked for, just take
	     * out what we need and return immediately.
	     *==============================================================*/
	    memcpy(buf,prev_buf,len);
	    memmove(prev_buf,prev_buf+len,*left_over-len);
	    *left_over -= len;
	    return(len);
	}
    }

    /*=======================================================================
     * If the last read contained a a block that's shorter than
     * sizeof(des_cblock) long, use that short block.
     *======================================================================*/
    if (*nshort > 0)
    {
	memcpy(ebuf,shortblk,*nshort);
    }

    /*=======================================================================
     * Read in enough to satisfy the read request
     *======================================================================*/
    readlen = ( (len - n + sizeof(des_cblock) - 1) / sizeof(des_cblock) )
			* sizeof(des_cblock) - *nshort;
    done_read = 0;
    do
    {
#ifdef MSDOS
   if ((nread=recv(fd,&ebuf[*nshort+done_read],readlen-done_read,0)) < 0
   	&& errno != EAGAIN)
#else
	if ((nread=read(fd,&ebuf[*nshort+done_read],readlen-done_read)) < 0
		&& errno != EAGAIN)
#endif /* MSDOS */
	{
	    printf("Encrypted read from fd %d failed, %s.\n",
			fd,sys_errlist[errno]);
#ifndef MSDOS
	    syslog(LOG_ERR,
		"Error: Encrypted read from fd %d failed, %s.\n",
		fd,sys_errlist[errno]);
#endif
	    return(-1);
	}
	if (nread > 0)
	   done_read += nread;
    }
    while (nread != 0 && *nshort+done_read < sizeof(des_cblock));

    nread = done_read;

    /*=======================================================================
     * Pretend that the short block from the last round was read in just now.
     *======================================================================*/
    nread = nread + *nshort;
    *nshort = 0;

    if (nread == 0)
	return(0);

    /*=======================================================================
     * Decrypt the buffer.
     *======================================================================*/
    n8 = 0;
    while (n8 < nread)
    {

	/*===================================================================
	 * Make sure data is sizeof(des_cblock)-byte multiples.
	 * If it isn't, save this short block and process it next time.
	 *==================================================================*/
	if (nread-n8 < sizeof(des_cblock))
	{
	    *nshort = nread-n8;
	    memcpy(shortblk,&ebuf[n8],*nshort);
	    break;
	}

	des_ecb_encrypt((des_cblock *)&ebuf[n8],(des_cblock *)&tbuf,
			keysched,DES_DECRYPT);
	n8 += sizeof(des_cblock);

	/*===================================================================
	 * See if the first byte is a length byte
	 *==================================================================*/
	if (tbuf[0] > LEN_BYTE_MASK && tbuf[0] <= (LEN_BYTE_MASK+0x07))
	{
	    block_len = tbuf[0] & ~LEN_BYTE_MASK; /* Remove len byte ind */
	    tb = &tbuf[1];
	}
	else
	{
	    block_len = sizeof(des_cblock);
	    tb = tbuf;
	}

	/*===================================================================
	 * Make sure we do not overflow cbuf.  If it is already full, start
	 * putting clear text in prev_buf.
	 *==================================================================*/
	if (n+block_len > len)
	{
	    memcpy(prev_buf+*left_over,tb+len-n,block_len-(len-n));
	    *left_over += block_len-(len-n);

	    if (n < len)
	    {
		memcpy(buf+n,tb,len-n);
		n = len;
	    }
	}
	else
	{
	    memcpy(buf+n,tb,block_len);
	    n += block_len;
	}

    }

    return(n);

}

/*=======================================================================
 * enc_write()
 *
 * Encrypt and write data
 *
 * input:
 *	fd	file descriptor to write to
 *	buf	pointer to data
 *	len	number of bytes to write
 *	keysched des_key_schedule to use
 *
 * output:
 *	returns the number of bytes processed.  So if everything is OK, it
 *		always returns len.
 *	returns -1 if failed.
 *======================================================================*/
int enc_write(int fd, unsigned char *buf, int len, des_key_schedule keysched)
{
    des_cblock		random_bits;
    unsigned char	ebuf[2*BUFSIZ];
    unsigned char	tbuf[sizeof(des_cblock)];
    int			i,nwrite;
    int			bytes_left=0;
    int			written=0;
#ifndef MSDOS
    fd_set		rdy;
    struct timeval	*to=(struct timeval *)0;
#endif

    /*=======================================================================
     * Make sure we can handle the data length.
     *======================================================================*/
    if (len > sizeof(ebuf))
    {
	printf("Buffer limit exceeded trying to write %d bytes.\n",len);
#ifndef MSDOS
	syslog(LOG_ERR,"Error: Buffer limit exceeded trying to write %d bytes.\n",len);
#endif
	return(-1);
    }

    /*=======================================================================
     * Make sure data is sizeof(des_cblock)-byte multiples.
     *======================================================================*/
    for (i=0; i<(len+sizeof(des_cblock)-1)/sizeof(des_cblock); i++)
    {
	/*====================================================================
	 * See if this is a block that is less than sizeof(des_cblock) bytes
	 * long
	 *===================================================================*/
	if (i*sizeof(des_cblock)+8 > len)
	{
	    unsigned char block_len;

	    /*================================================================
	     * Put some random bits into the buffer
	     *===============================================================*/
	    des_random_key(tbuf);
	    /*================================================================
	     * Put the length byte in the buffer.  Set MSB to 1 to indicate
	     * it is a length byte.
	     *===============================================================*/
	    block_len = len - i*sizeof(des_cblock);
	    tbuf[0] = block_len | LEN_BYTE_MASK;
	    memcpy(&tbuf[1],&buf[i*sizeof(des_cblock)],block_len);
	}
	else
	    memcpy(tbuf,&buf[i*sizeof(des_cblock)],sizeof(des_cblock));

	des_ecb_encrypt((des_cblock *)tbuf,
			(des_cblock *)&ebuf[i*sizeof(des_cblock)],
			keysched,DES_ENCRYPT);
    }

    /*=======================================================================
     * Write the encrypted data.
     * Make sure fd is ready to be written to before writing.
     *======================================================================*/
    errno = 0;
    bytes_left = i*sizeof(des_cblock);
    while (bytes_left > 0)
    {
int rc;
#ifndef MSDOS
	FD_ZERO(&rdy);
	FD_SET(fd,&rdy);
#ifndef __hppa
	if((rc = select(fd + 1,(fd_set *)0,&rdy,(fd_set *)0,to)) < 0)
#else
	if((rc = select(fd + 1,0,(int *)&rdy,0,to)) < 0)
#endif
	{
	    syslog(LOG_ERR,"Error: Select failed for encrypted write, "
			"fd=%d, %s.",
			fd,sys_errlist[errno]);
	    return(-1);
	}

	if ((nwrite=write(fd,&ebuf[written],bytes_left)) != bytes_left
		&& errno != EAGAIN && errno != EPIPE && errno !=0)
#else
	if ((nwrite=send(fd,&ebuf[written],bytes_left,0)) != bytes_left
		&& errno != EAGAIN && errno != EPIPE && errno !=0)
#endif /* MSDOS */
	{
	    printf("Encryptyed write to fd %d failed, nwrite=%d, %s.\n",
			fd,nwrite,sys_errlist[errno]);
#ifndef MSDOS
	    syslog(LOG_ERR,"Error: Encryptyed write to fd %d failed, "
			"nwrite=%d, %s.\n",
			fd,nwrite,sys_errlist[errno]);
#endif
	    return(-1);
	}

	if (nwrite < 0)
	    return(-1);

	if (nwrite > 0)
	{
	    bytes_left -= nwrite;
	    written += nwrite;
	}
    }

    /*=======================================================================
     * If everything is OK, return the length of the clear text, not the
     * length of the cypher text that waas actually written.
     *======================================================================*/
    return (len);
}

/*=======================================================================
 * gettnkey()
 *
 * Get the encryption key for a given username
 *
 * input:
 *	username	char * of the username
 *
 * output:
 *	returns 0 if username is found and the key is returned in the
 *		argument key.
 *	returns -1 if failed.
 *
 *
 * The key file TNKEYFILE contains entries that are 48 bytes each.  Each
 * entry is the ascii representation of the 24 bytes of user data.  The
 * user data consists of 16 bytes of username and 8 bytes of encryption key.
 *======================================================================*/
int gettnkey(char *username,des_cblock *key)
{
    /*=====================================================================
     * The old style key is 16 bytes username and 8 bytes encryption key
     * The new style key is 7 random bytes 17 bytes username
     * and 8 bytes encryption key
     *====================================================================*/
    unsigned long eentry[NEWTNKEYLEN/sizeof(long)];
    unsigned char e2entry[LINE_MAX];
    unsigned char centry[NEWTNKEYLEN];
    unsigned char *pound,*kstr;
    unsigned char tusername[UNMSIZ];
    unsigned char *unm,*k;
    des_key_schedule key_sched;
    unsigned long tkey[2];
    int lno=0;
    char path1[LINE_MAX],path2[LINE_MAX];
    FILE *kf;

    e2entry[2*sizeof(eentry)+1] = '\0';
#ifdef MSDOS
    strcpy(path1,"\\trex\\");
    strcat(path1,TNKEYFILE);
    if ((kf = fopen(path1,"r")) == 0)
#else
    /*================================================================
     * Get $HOME to build path for key file.
     *===============================================================*/
    strcpy(path1,getenv("HOME"));
    strcat(path1,"/");
    strcat(path1,TNKEYFILE);

    strcpy(path2,"/etc");
    strcat(path2,"/");
    strcat(path2,TNKEYFILE);

    /*================================================================
     * Try $HOME/TNKEYFILE and then /etc/TNKEYFILE
     *===============================================================*/
    if (((kf = fopen(path1,"r")) == 0) &&
	((kf = fopen(path2,"r")) == 0))
#endif
    {
	printf("Open of encryption key file failed, %s.\n\r",sys_errlist[errno]);
	sleep(10);
	return(-1);
    }
    tkey[0] = htonl(TNKEY1);
    tkey[1] = htonl(TNKEY2);
    des_set_key((des_cblock *)tkey,key_sched);
    memset(tkey,0,sizeof(tkey));

    /*================================================================
     * Read in the keyfile entries
     *===============================================================*/
    while (!feof(kf))
    {
	lno++;
	if (fgets((char *)e2entry,sizeof(e2entry),kf) == NULL)
	{
	    if (feof(kf))
		break;
	    printf("Error: failure in reading line %d of tnkey file %s.  %s"
				"\n\r",
				lno,TNKEYFILE,sys_errlist[errno]);
	    return(-1);
	}
	pound = (unsigned char *)strchr((char *)e2entry,'#');
	if (pound)
	    *pound = '\0';			/* ignore everything after # */
	kstr = (unsigned char *)strtok((char *)e2entry," \t\n");
	if (kstr == NULL)		/* ignore blank lines */
	    continue;
	/*================================================================
	 * If the remaining string (delimited by blanks and tabs) is not
	 * exactly 48 characters or if there is a second string after the
	 * first one, we report the line as bad
	 *===============================================================*/
	if ((strlen((char *)kstr) != 2*OLDTNKEYLEN &&
	     strlen((char *)kstr) != 2*NEWTNKEYLEN) ||
	    strtok(NULL," \t"))
	{
	    printf("Invalid record on line %d of tnkey file %s.\n",
			lno,TNKEYFILE);
	    continue;
	}

	/*================================================================
	 * Determine whether it's an old style key or new style by the
	 * length of the key.  old = 48 bytes, new = 64 bytes.
	 *===============================================================*/
	if (strlen((char *)kstr) == 2*OLDTNKEYLEN)
	{
	    /*============================================================
	     * Convert old style key to binary
	     *===========================================================*/
	    sscanf((const char *)kstr,"%8x%8x%8x%8x%8x%8x",&eentry[0],
				&eentry[1],
				&eentry[2],
				&eentry[3],
				&eentry[4],
				&eentry[5]);
	    /*============================================================
	     * Since these longs will be used as chars, we need to make sure
	     * they are big endians.
	     *===========================================================*/
	    eentry[0] = htonl(eentry[0]);
	    eentry[1] = htonl(eentry[1]);
	    eentry[2] = htonl(eentry[2]);
	    eentry[3] = htonl(eentry[3]);
	    eentry[4] = htonl(eentry[4]);
	    eentry[5] = htonl(eentry[5]);

	    /*============================================================
	     * Decrypt the entry
	     *===========================================================*/
	    des_ecb_encrypt((des_cblock *)eentry,(des_cblock *)centry,
			key_sched,DES_DECRYPT);
	    des_ecb_encrypt((des_cblock *)&eentry[2],(des_cblock *)&centry[8],
			key_sched,DES_DECRYPT);
	    des_ecb_encrypt((des_cblock *)&eentry[4],(des_cblock *)&centry[16],
			key_sched,DES_DECRYPT);

	    unm = &centry[OUNM];
	    k = &centry[OKEY];
	}
	else
	{
	    /*============================================================
	     * Convert new style key to binary
	     *===========================================================*/
	    sscanf((const char *)kstr,"%8x%8x%8x%8x%8x%8x%8x%8x",&eentry[0],
				&eentry[1],
				&eentry[2],
				&eentry[3],
				&eentry[4],
				&eentry[5],
				&eentry[6],
				&eentry[7]);
	    /*============================================================
	     * Since these longs will be used as chars, we need to make sure
	     * they are big endians.
	     *===========================================================*/
	    eentry[0] = htonl(eentry[0]);
	    eentry[1] = htonl(eentry[1]);
	    eentry[2] = htonl(eentry[2]);
	    eentry[3] = htonl(eentry[3]);
	    eentry[4] = htonl(eentry[4]);
	    eentry[5] = htonl(eentry[5]);
	    eentry[6] = htonl(eentry[6]);
	    eentry[7] = htonl(eentry[7]);

	    /*============================================================
	     * Decrypt the entry
	     *===========================================================*/
	    des_cbc_encrypt((des_cblock *)eentry,(des_cblock *)centry,
			32,key_sched,IVEC,DES_DECRYPT);

	    unm = &centry[NUNM];
	    k = &centry[NKEY];
	}

	/*================================================================
	 * See if this is the entry for the given username
	 *===============================================================*/
	memcpy(tusername,unm,UNMSIZ);
	if (!strcmp((const char *)username,(const char *)tusername))
	{
	    memcpy((char *)key,k,sizeof(des_cblock));
	    return(0);
	}
    }

    memset(key_sched,0,sizeof(key_sched));
    printf("username %s not found in encryption key file.\n\r",username);
    return(-1);
}

#ifndef MSDOS
/*============================================================================
 * md5key()
 *
 *	Use MD5 to transform the user authentication key to an encryption key.
 *	We do this because the encryption key (encrypted by a fixed key, see
 *	gettnkey() above), will be sitting on a client machine, perhaps with
 *	little or no physical security.  If someone somehow obtained and was
 *	able to decrypt the key file on a client and decipher the encryption
 *	key, this md5 transform will make it very, very difficult, if not
 *	impossible, to derive the authentication key from it.
 *
 * input:
 *	string		authentication key
 *
 * output:
 *	returns pointer to the transformed key 
 *==========================================================================*/
unsigned char *md5key(char *string)
{
   MD5_CTX context;
   static unsigned char digest[16];
   unsigned int len = strlen (string);

   MD5Init (&context);
   MD5Update (&context, string, len);
   MD5Final (digest, &context);

   return(digest);
}
#endif /* MSDOS */
#endif  /*TREX_ENC */
