/*============================================================================
 * filename:		win32/ptelnet/tnmsg.h
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*==========================================================================
 *
 * R. James Livermore grants Freemont Avenue Software, Inc. a nonexclusive
 * license to use, modify, and market this program using their tradename
 * T.Rex.	
 *
 *
 * (C) COPYRIGHT R. James Livermore 1998-2000.  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1998-2000.
 * All rights reserved. 
 *           
 *		
 * Date written:	September 26, 1998
 * Last modified:	February   8, 2000
 *
 *	                         
 * function:
 *
 *============================================================================*/
//
//  Values are 32 bit values layed out as follows:
//
//   3 3 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1
//   1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
//  +---+-+-+-----------------------+-------------------------------+
//  |Sev|C|R|     Facility          |               Code            |
//  +---+-+-+-----------------------+-------------------------------+
//
//  where
//
//      Sev - is the severity code
//
//          00 - Success
//          01 - Informational
//          10 - Warning
//          11 - Error
//
//      C - is the Customer code flag
//
//      R - is a reserved bit
//
//      Facility - is the facility code
//
//      Code - is the facility's status code
//
//
// Define the facility codes
//


//
// Define the severity codes
//
#define STATUS_SEVERITY_WARNING          0x2
#define STATUS_SEVERITY_SUCCESS          0x0
#define STATUS_SEVERITY_INFORMATIONAL    0x1
#define STATUS_SEVERITY_ERROR            0x3


//
// MessageId: MSG_COPYRIGHT
//
// MessageText:
//
//  
//
#define MSG_COPYRIGHT                    0x40000001L

//
// MessageId: MSG_USAGE
//
// MessageText:
//
//  Usage: %1 <switches> <parameters>
//  
//  Switches:
//
#define MSG_USAGE                        0x40000002L

//
// MessageId: MSG_ERROR
//
// MessageText:
//
//  %1 failed.
//
#define MSG_ERROR                        0xC0000003L

//
// MessageId: MSG_INFO
//
// MessageText:
//
//  %1
//
#define MSG_INFO                         0x40000004L

//
// MessageId: MSG_WARNING
//
// MessageText:
//
//  %1
//
#define MSG_WARNING                      0x80000005L

//
// MessageId: MSG_TRYING
//
// MessageText:
//
//  Trying... %0
//
#define MSG_TRYING                       0x40000006L

//
// MessageId: MSG_CONNECTED
//
// MessageText:
//
//  Connected to %1.
//
#define MSG_CONNECTED                    0x40000007L

//
// MessageId: MSG_WSAEINTR
//
// MessageText:
//
//  WSAEINTR
//
#define MSG_WSAEINTR                     0xC0002714L

//
// MessageId: MSG_WSAEBADF
//
// MessageText:
//
//  WSAEBADF
//
#define MSG_WSAEBADF                     0xC0002719L

//
// MessageId: MSG_WSAEACCESS
//
// MessageText:
//
//  WSAEACCESS
//
#define MSG_WSAEACCESS                   0xC000271DL

//
// MessageId: MSG_WSAEDEFAULT
//
// MessageText:
//
//  WSAEDEFAULT
//
#define MSG_WSAEDEFAULT                  0xC000271EL

//
// MessageId: MSG_WSAEINVAL
//
// MessageText:
//
//  WSAEINVAL
//
#define MSG_WSAEINVAL                    0xC0002726L

//
// MessageId: MSG_WSAEMFILE
//
// MessageText:
//
//  WSAEMFILE
//
#define MSG_WSAEMFILE                    0xC0002728L

//
// MessageId: MSG_WSAEWOULDBLOCK
//
// MessageText:
//
//  WSAEWOULDBLOCK
//
#define MSG_WSAEWOULDBLOCK               0xC0002733L

//
// MessageId: MSG_WSAEINPROGRESS
//
// MessageText:
//
//  WSAEINPROGRESS
//
#define MSG_WSAEINPROGRESS               0xC0002734L

//
// MessageId: MSG_WSAEALREADY
//
// MessageText:
//
//  WSAEALREADY
//
#define MSG_WSAEALREADY                  0xC0002735L

//
// MessageId: MSG_WSAENOTSOCK
//
// MessageText:
//
//  WSAENOTSOCK
//
#define MSG_WSAENOTSOCK                  0xC0002736L

//
// MessageId: MSG_WSAEDESTADDRREQ
//
// MessageText:
//
//  WSAEDESTADDRREQ
//
#define MSG_WSAEDESTADDRREQ              0xC0002737L

//
// MessageId: MSG_WSAEMSGSIZE
//
// MessageText:
//
//  WSAEMSGSIZE
//
#define MSG_WSAEMSGSIZE                  0xC0002738L

//
// MessageId: MSG_WSAEPROTOTYPE
//
// MessageText:
//
//  WSAEPROTOTYPE
//
#define MSG_WSAEPROTOTYPE                0xC0002739L

//
// MessageId: MSG_WSAENOPROTOOPT
//
// MessageText:
//
//  WSAENOPROTOOPT
//
#define MSG_WSAENOPROTOOPT               0xC000273AL

//
// MessageId: MSG_WSAEPROTONOTSUPPORT
//
// MessageText:
//
//  WSAEPROTONOTSUPPORT
//
#define MSG_WSAEPROTONOTSUPPORT          0xC000273BL

//
// MessageId: MSG_WSAESOCKNOTSUPPORT
//
// MessageText:
//
//  WSAESOCKNOTSUPPORT
//
#define MSG_WSAESOCKNOTSUPPORT           0xC000273CL

//
// MessageId: MSG_WSAEOPNOTSUPP
//
// MessageText:
//
//  WSAEOPNOTSUPP
//
#define MSG_WSAEOPNOTSUPP                0xC000273DL

//
// MessageId: MSG_WSAEPFNOTSUPPORT
//
// MessageText:
//
//  WSAEPFNOTSUPPORT
//
#define MSG_WSAEPFNOTSUPPORT             0xC000273EL

//
// MessageId: MSG_WSAEAFNOTSUPPORT
//
// MessageText:
//
//  WSAEAFNOTSUPPORT
//
#define MSG_WSAEAFNOTSUPPORT             0xC000273FL

//
// MessageId: MSG_WSAEADDRINUSE
//
// MessageText:
//
//  Address is in use.
//
#define MSG_WSAEADDRINUSE                0xC0002740L

//
// MessageId: MSG_WSAEADDRNOTAVAIL
//
// MessageText:
//
//  WSAEADDRNOTAVAIL
//
#define MSG_WSAEADDRNOTAVAIL             0xC0002741L

//
// MessageId: MSG_WSAENETDOWN
//
// MessageText:
//
//  Network is down.
//
#define MSG_WSAENETDOWN                  0xC0002742L

//
// MessageId: MSG_WSAENETUNREACH
//
// MessageText:
//
//  Network is unreachable.
//
#define MSG_WSAENETUNREACH               0xC0002743L

//
// MessageId: MSG_WSAENETRESET
//
// MessageText:
//
//  WSAENETRESET
//
#define MSG_WSAENETRESET                 0xC0002744L

//
// MessageId: MSG_WSAECONNABORTED
//
// MessageText:
//
//  WSAECONNABORTED
//
#define MSG_WSAECONNABORTED              0xC0002745L

//
// MessageId: MSG_WSAECONNRESET
//
// MessageText:
//
//  WSAECONNRESET
//
#define MSG_WSAECONNRESET                0xC0002746L

//
// MessageId: MSG_WSAENOBUFS
//
// MessageText:
//
//  WSAENOBUFS
//
#define MSG_WSAENOBUFS                   0xC0002747L

//
// MessageId: MSG_WSAEISCONN
//
// MessageText:
//
//  WSAEISCONN
//
#define MSG_WSAEISCONN                   0xC0002748L

//
// MessageId: MSG_WSAENOTCONN
//
// MessageText:
//
//  WSAENOTCONN
//
#define MSG_WSAENOTCONN                  0xC0002749L

//
// MessageId: MSG_WSAESHUTDOWN
//
// MessageText:
//
//  WSAESHUTDOWN
//
#define MSG_WSAESHUTDOWN                 0xC000274AL

//
// MessageId: MSG_WSAETOOMANYREFS
//
// MessageText:
//
//  WSAETOOMANYREFS
//
#define MSG_WSAETOOMANYREFS              0xC000274BL

//
// MessageId: MSG_WSAETIMEDOUT
//
// MessageText:
//
//  The request has timed out.
//
#define MSG_WSAETIMEDOUT                 0xC000274CL

//
// MessageId: MSG_WSAECONNREFUSED
//
// MessageText:
//
//  Connection refused.
//
#define MSG_WSAECONNREFUSED              0xC000274DL

//
// MessageId: MSG_WSAELOOP
//
// MessageText:
//
//  WSAELOOP
//
#define MSG_WSAELOOP                     0xC000274EL

//
// MessageId: MSG_WSAENAMETOOLONG
//
// MessageText:
//
//  WSAENAMETOOLONG
//
#define MSG_WSAENAMETOOLONG              0xC000274FL

//
// MessageId: MSG_WSAEHOSTDOWN
//
// MessageText:
//
//  WSAEHOSTDOWN
//
#define MSG_WSAEHOSTDOWN                 0xC0002750L

//
// MessageId: MSG_WSAEHOSTUNREACH
//
// MessageText:
//
//  Host is unreachable.
//
#define MSG_WSAEHOSTUNREACH              0xC0002751L

//
// MessageId: MSG_WSAESYSNOTREADY
//
// MessageText:
//
//  WSAESYSNOTREADY
//
#define MSG_WSAESYSNOTREADY              0xC000276BL

//
// MessageId: MSG_WSAVERNOTSUPPORTED
//
// MessageText:
//
//  Version not supported.
//
#define MSG_WSAVERNOTSUPPORTED           0xC000276CL

//
// MessageId: MSG_WSANOTINITIALISED
//
// MessageText:
//
//  Not initialised.
//
#define MSG_WSANOTINITIALISED            0xC000276DL

//
// MessageId: MSG_WSAHOST_NOT_FOUND
//
// MessageText:
//
//  Host not found.
//
#define MSG_WSAHOST_NOT_FOUND            0xC0002AF9L

//
// MessageId: MSG_WSATRY_AGAIN
//
// MessageText:
//
//  Try again.
//
#define MSG_WSATRY_AGAIN                 0xC0002AFAL

//
// MessageId: MSG_WSANO_RECOVERY
//
// MessageText:
//
//  No recovery.
//
#define MSG_WSANO_RECOVERY               0xC0002AFBL

//
// MessageId: MSG_WSANO_DATA
//
// MessageText:
//
//  No data.
//
#define MSG_WSANO_DATA                   0xC0002AFCL

