/*============================================================================
 * filename:		win32/ptelnet/tnnet.cpp
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
///////////////////////////////////////////////////////////////////////////////
//
// Contents:	telnet network processing
//
// Product:		telnet
//
// Revisions:	02.Apr.1995	igor.milavec@uni-lj.si
//					Original code
//
///////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <ansiprsr.h>
#include <dos.h>
#include "tnnet.h"
#include "des.h"

extern secrets;
extern des_key_schedule keysched;

extern "C" {
int enc_read(int _fd, unsigned char *_buf, int _len, des_key_schedule _keysched);
int enc_write(int _fd, unsigned char *_buf, int _len, des_key_schedule _keysched);
};

class TTelnetParser: public TANSIParser {
	HANDLE hConsole;
	SOCKET Socket;
protected:

	virtual int ConWriteString(char* pszString, int cbString);
	virtual void ConClearScreen();
	virtual void ConClearEOScreen();
	virtual void ConClearBOScreen();
	virtual void ConClearLine();
	virtual void ConClearEOLine();
	virtual void ConClearBOLine();
	virtual void ConSetCursorPosition(int x, int y);
	virtual void ConMoveCursorPosition(int x, int y);
	virtual void ConSetExtendedMode(int iFunction, BOOL bEnable);
   virtual void ConInsertLine(int n);
   virtual unsigned char *ConCRLF(unsigned char *String, unsigned char *End);
   virtual void ConDeleteChar();
	virtual void ConDeleteLine(int n);
	virtual int NetWriteString(char* pszString, int cbString);

   virtual void ConGetdwSize(COORD *Coord);

	int ConGetCursorX();
	int ConGetCursorY();
public:
	TTelnetParser(HANDLE hConsole, SOCKET Socket);
};

// Class constructor
TTelnetParser::TTelnetParser(HANDLE hConsole, SOCKET Socket):
	hConsole(hConsole), Socket(Socket)
{
   return;
}

int TTelnetParser::ConWriteString(char* pszString, int cbString)
{
	DWORD Result;
	WriteConsole(hConsole, pszString, cbString, &Result, 0);
	return Result;
}

// Clear the entire screen
void TTelnetParser::ConClearScreen()
{
	DWORD dwWritten;
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	if (!GetConsoleScreenBufferInfo(hConsole, &ConsoleInfo))
		return;

	COORD Coord = {ConsoleInfo.srWindow.Left, ConsoleInfo.srWindow.Top};
	FillConsoleOutputCharacter(hConsole, ' ',
		(DWORD)(ConsoleInfo.dwSize.X * ConsoleInfo.dwSize.Y),
		Coord, &dwWritten);
   return;
}

// Clear from cursor thru End of Screen
void TTelnetParser::ConClearEOScreen()
{
	DWORD dwWritten;
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	if (!GetConsoleScreenBufferInfo(hConsole, &ConsoleInfo))
		return;

	COORD Coord = {ConsoleInfo.srWindow.Left, ConsoleInfo.dwCursorPosition.Y + 1};
	FillConsoleOutputCharacter(hConsole, ' ',
		(DWORD)(ConsoleInfo.srWindow.Right - ConsoleInfo.srWindow.Left + 1)*
		(DWORD)(ConsoleInfo.srWindow.Bottom - ConsoleInfo.dwCursorPosition.Y),
		Coord, &dwWritten);

	ConClearEOLine();
   return;
}

// Clear from cursor to Beginning of Screen
void TTelnetParser::ConClearBOScreen()
{
	DWORD dwWritten;
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	if (!GetConsoleScreenBufferInfo(hConsole, &ConsoleInfo))
		return;

	COORD Coord = {ConsoleInfo.srWindow.Left, ConsoleInfo.srWindow.Top};
	FillConsoleOutputCharacter(hConsole, ' ',
		(DWORD)(ConsoleInfo.srWindow.Right - ConsoleInfo.srWindow.Left + 1)*
		(DWORD)(ConsoleInfo.dwCursorPosition.Y - ConsoleInfo.srWindow.Top),
		Coord, &dwWritten);

	ConClearBOLine();
   return;
}

// Blank the cursor'ed line
void TTelnetParser::ConClearLine()
{
	DWORD dwWritten;
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	if (!GetConsoleScreenBufferInfo(hConsole, &ConsoleInfo))
		return;

	COORD Coord = {
		ConsoleInfo.srWindow.Left,
		ConsoleInfo.dwCursorPosition.Y};
	FillConsoleOutputCharacter(hConsole, ' ',
		(DWORD)(ConsoleInfo.srWindow.Right - ConsoleInfo.srWindow.Left + 1),
		Coord, &dwWritten);
   return;
}

// Blank from cursor to End of Line
void TTelnetParser::ConClearEOLine()
{
	DWORD dwWritten;
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	if (!GetConsoleScreenBufferInfo(hConsole, &ConsoleInfo))
		return;

	COORD Coord = {
	 	(short)(ConsoleInfo.dwCursorPosition.X),
	 	(short)(ConsoleInfo.dwCursorPosition.Y)};
	FillConsoleOutputCharacter(hConsole, ' ',
	 	(DWORD)(ConsoleInfo.dwSize.X - ConsoleInfo.dwCursorPosition.X),
   	Coord, &dwWritten);
   return;
}

// Blank from cursor to Beginning of Line
void TTelnetParser::ConClearBOLine()
{
	DWORD dwWritten;
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	if (!GetConsoleScreenBufferInfo(hConsole, &ConsoleInfo))
		return;

	COORD Coord = {
		ConsoleInfo.srWindow.Left,
		ConsoleInfo.dwCursorPosition.Y};
	FillConsoleOutputCharacter(hConsole, ' ',
		ConsoleInfo.dwCursorPosition.X+1, Coord, &dwWritten);
	return;
}

void TTelnetParser::ConSetCursorPosition(int x, int y)
{
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	if (!GetConsoleScreenBufferInfo(hConsole, &ConsoleInfo))
		return;

	COORD Coord = {
		(short)(ConsoleInfo.srWindow.Left + x),
		(short)(ConsoleInfo.srWindow.Top + y)};

	SetConsoleCursorPosition(hConsole, Coord);
   return;
}

int TTelnetParser::ConGetCursorX()
{
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	if (!GetConsoleScreenBufferInfo(hConsole, &ConsoleInfo))
		return 0;

	return ConsoleInfo.dwCursorPosition.X - ConsoleInfo.srWindow.Left;
}

int TTelnetParser::ConGetCursorY()
{
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	if (!GetConsoleScreenBufferInfo(hConsole, &ConsoleInfo))
		return 0;

	return ConsoleInfo.dwCursorPosition.Y - ConsoleInfo.srWindow.Top;
}

void TTelnetParser::ConMoveCursorPosition(int x, int y)
{
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	if (!GetConsoleScreenBufferInfo(hConsole, &ConsoleInfo))
		return;

	COORD Coord = {
		(short)(ConsoleInfo.dwCursorPosition.X + x),
		(short)(ConsoleInfo.dwCursorPosition.Y + y)};

	SetConsoleCursorPosition(hConsole, Coord);
   return;
}

void TTelnetParser::ConSetExtendedMode(int iFunction, BOOL bEnable)
{
	// Probably should do something here...
}

// Insert blank lines at the cursor, scrolling the lines below cursor down.
void TTelnetParser::ConInsertLine(int n)
{
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
   SMALL_RECT scrollRect;
   CHAR_INFO consoleFill;
   COORD Coord;

   // get current console size
   GetConsoleScreenBufferInfo(hConsole,&ConsoleInfo);

   // define the rectangle to scroll
   scrollRect.Top = ConsoleInfo.dwCursorPosition.Y;
   scrollRect.Left = ConsoleInfo.srWindow.Left;
   scrollRect.Bottom = ConsoleInfo.dwSize.Y - n;
   scrollRect.Right = ConsoleInfo.srWindow.Right;

   // define destination of scrolled rectangle
   Coord.X = 0;
   Coord.Y = ConsoleInfo.dwCursorPosition.Y + n;

   // define how to fill blank line
   consoleFill.Attributes = ConsoleInfo.wAttributes;
   consoleFill.Char.AsciiChar = ' ';

   // perform the scroll (to insert n lines at the top)
   ScrollConsoleScreenBuffer(hConsole,&scrollRect,0,Coord,&consoleFill);

   return;
}

// Process CR and/or LF - Win32 does not emit the correct sequence resulting
// in a nl doing both CR and LF, i.e. the cursor after a nl character is always
// at the leftmost column.
// This function is to work around that problem by first saving the X position
// and then, after all the CR and LF, repositioning the cursor.
unsigned char *TTelnetParser::ConCRLF(unsigned char *pszString,
													unsigned char *pszBufferEnd)
{
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
   SMALL_RECT scrollRect;
   CHAR_INFO consoleFill;
   COORD Coord;
   unsigned char newlines[25] = {'\n','\n','\n','\n','\n',
   										'\n','\n','\n','\n','\n',
                                 '\n','\n','\n','\n','\n',
                                 '\n','\n','\n','\n','\n',
                                 '\n','\n','\n','\n','\n'};
   int cr = 0;
   int nl = 0;
   int tnl = 0;
   unsigned char *p = pszString;

   // save current console X position
   GetConsoleScreenBufferInfo(hConsole,&ConsoleInfo);
   Coord.X = ConsoleInfo.dwCursorPosition.X;

   // Tally the Cr and LF characters
   while (p < pszBufferEnd && (*p == '\n' || *p == '\r'))
   {
      if (*p == '\r')
      	cr++;
      else
      	nl++;
		p++;
   }

   // Process all the LF characters at the same time.
   if (nl > 0)
   {
   	tnl = nl;
      while (tnl > 25)
      {
   		ConWriteString(newlines,25);
      	tnl -= 25;
      }
      if (tnl)
      	ConWriteString(newlines,tnl);
   }

   // Only need to emit 1 CR if any CR character is present
   if (cr > 0)
   	ConWriteString("\r",1);
   else
   {
   	// get current console Y position
   	GetConsoleScreenBufferInfo(hConsole,&ConsoleInfo);
      Coord.Y = ConsoleInfo.dwCursorPosition.Y;

      // Reposition the cursor to the saved X position
   	SetConsoleCursorPosition(hConsole,Coord);
   }
   return(p);

}

// Delete the character at the cursor, shifting the rest of the line to the
// left.
void TTelnetParser::ConDeleteChar()
{
   CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	SMALL_RECT scrollRect;
   CHAR_INFO consoleFill;
   COORD Coord;

   // get current screen info
   GetConsoleScreenBufferInfo(hConsole,&ConsoleInfo);

   // define the rectangle to scroll
	scrollRect.Top = ConsoleInfo.dwCursorPosition.Y;
   scrollRect.Left = ConsoleInfo.dwCursorPosition.X + 1;
   scrollRect.Bottom = ConsoleInfo.dwCursorPosition.Y;
   scrollRect.Right = ConsoleInfo.dwSize.X - 1;

   // define destination of scrolled rectangle
   Coord.X = ConsoleInfo.dwCursorPosition.X;
   Coord.Y = ConsoleInfo.dwCursorPosition.Y;

   // define how to fill blank line
   consoleFill.Attributes = ConsoleInfo.wAttributes;
   consoleFill.Char.AsciiChar = ' ';

   // perform the scroll (to shift the line to the left)
   ScrollConsoleScreenBuffer(hConsole,&scrollRect,0,Coord,&consoleFill);

   return;
   }

// Delete n lines including the line the cursor is on and scroll the lines
// below up.
void TTelnetParser::ConDeleteLine(int n)
{
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
   SMALL_RECT scrollRect;
   CHAR_INFO consoleFill;
   COORD Coord;

   // get current console size
   GetConsoleScreenBufferInfo(hConsole,&ConsoleInfo);

   // if n is beyond the screen size, delete only up to the screen size
   if (n > (ConsoleInfo.dwSize.Y - 1) - ConsoleInfo.dwCursorPosition.Y)
		n = (ConsoleInfo.dwSize.Y - 1) - ConsoleInfo.dwCursorPosition.Y;

   // define the rectangle to scroll
	scrollRect.Top = ConsoleInfo.dwCursorPosition.Y + n;
   scrollRect.Left = ConsoleInfo.srWindow.Left;
   scrollRect.Bottom = ConsoleInfo.dwSize.Y - 1;
   scrollRect.Right = ConsoleInfo.dwSize.X - 1;

   // define destination of scrolled rectangle
   Coord.X = 0;
   Coord.Y = ConsoleInfo.dwCursorPosition.Y;

   // define how to fill blank line
   consoleFill.Attributes = ConsoleInfo.wAttributes;
   consoleFill.Char.AsciiChar = ' ';

   // perform the scroll (to write over n lines)
   ScrollConsoleScreenBuffer(hConsole,&scrollRect,0,Coord,&consoleFill);

   return;
}
int TTelnetParser::NetWriteString(char* pszString, int cbString)
{
	if (secrets)
   	enc_write(Socket, pszString, cbString, keysched);
   else
		send(Socket, pszString, cbString, 0);
	return cbString;
}

void TTelnetParser::ConGetdwSize(COORD *Coord)
{
   CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;

   // get current screen info
   GetConsoleScreenBufferInfo(hConsole,&ConsoleInfo);

   *Coord = ConsoleInfo.dwSize;
   return;
}


DWORD telProcessNetwork(LPVOID pvParams)
{
	TelThreadParams* pParams = (TelThreadParams*)pvParams;
	SOCKET Socket = pParams->Socket;
	TTelnetParser Parser(GetStdHandle(STD_OUTPUT_HANDLE), Socket);

	const DWORD dwBuffer = 1024;
	unsigned char szBuffer[dwBuffer];
	unsigned char* pszHead = szBuffer;
	unsigned char* pszTail = szBuffer;

	while (1) {

		int Result;
      if (secrets)
      	Result = enc_read(Socket,pszTail,
         							(szBuffer + dwBuffer) - pszTail,
         							keysched);
		else
      	Result = recv(Socket, (char *)pszTail,
         					 (szBuffer + dwBuffer) - pszTail, 0);
		if ((Result <= 0) || ( Result > dwBuffer))
			break;

		pszTail += Result;

		unsigned char* pszNewHead = pszHead;
		do {
			pszHead = pszNewHead;
			pszNewHead = Parser.ParseBuffer(pszHead, pszTail, Socket);
		} while ((pszNewHead != pszHead) && (pszNewHead < pszTail));
		pszHead = pszNewHead;

		// When we reach the end of the buffer, move contents to the
		// beginning of the buffer to get free space at the end.
		if (pszTail == (szBuffer + dwBuffer)) {
			MoveMemory(szBuffer, pszHead, pszTail - pszHead);
			pszTail = szBuffer + (pszTail - pszHead);
			pszHead = szBuffer;
		}
	}

	printf("\nConnection terminated by remote.\n");
   sleep(3);
	SetEvent(pParams->hExit);
	return 0;
}

