/*============================================================================
 * filename:		xforward/socket.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * (C) COPYRIGHT R. James Livermore  1995-2000. All rights reserved.
 * (C) COPYRIGHT Freemont Avenue Softwarem, Incorporated 1995-2000.  All rights
 * reserved.    
 *
 * Date written:	August	  20, 1995
 * Last modified:	February   8, 2000
 *	
 *
 * function:		Provide X Forwarding services. X forwarding is initiated
 *			by a authenticated tnproxy user.	
 *       
 *
 *===========================================================================*/
static	char	cprght[] = 
  		"(C) COPYRIGHT Freemont Avenue Software, Inc. 1995-2000 socket,"
  		" V 1.0 00/01/21";
#include 	<errno.h> 
#include 	<syslog.h> 
#include	"xforward.h" 
#include	<sys/types.h> 
#ifndef __hppa
#include 	<sys/select.h>		/* to support fd_set 		*/
#endif
#include	<netinet/in.h>
#include	<sys/socket.h>
#include	<sys/time.h>
#include	<stdio.h>
#include 	<sys/param.h>
#include	<netdb.h>
#include        <sys/un.h>

#ifndef linux
extern char sys_errlist[];
#endif

extern  struct   hostent	*gethostname();
/*============================================================================ 
 * function prototypes
 *============================================================================*/  
int bindport( int sd, int family, int fromport, int toport, 
	      struct sockaddr *addr, int try_only_once);
/*============================================================================ 
 *
 *============================================================================*/
void clearCloseFd( int fd, fd_set *rset, fd_set *wset) 
{
   FD_CLR(fd,rset);
   FD_CLR(fd,wset);
   close(fd);
}
/*============================================================================
 * function:	get local host name
 * return: -1: error to get the local host
 *          0: OK
 * localhname: Must have length at least MAXHOSTNAMELEN
 *===========================================================================*/
int getLocalHostName( char localhname[] )
{
   struct   hostent *host;
   char              buf[MAXHOSTLN+42];

   gethostname(localhname, MAXHOSTNAMELEN);
   if( !(host = gethostbyname(localhname)) ) 
   {  
      sprintf( buf,"hostname lookup up of firewall failed. %s.\n",
			localhname );
      if(uselog)
	    syslog(LOG_ERR,"Error: %s",buf);
	 else
	    fprintf(stderr,"Error: %s",buf);
      return -1;
   }
   sprintf(localhname,"%s\0",host->h_name);
   return 0;
}
/*============================================================================
 * function: serv_select
 * inputs:
 *
 *
 * outputs:
 * 	return: timeout -99.
 *         	error   -errno.
 *         	>=0     selected.
 *===========================================================================*/
int servSelect( int fds, fd_set *readable, fd_set *writable, int timemax, int idlemax )
{
   int     nready;
   
   char	buf[64];

   struct timeval    	timeout;
   static time_t	startime = 0; 	/* initial to zero the first time */
	
   timeout.tv_usec = 0;
   timeout.tv_sec  = timemax;

#ifdef __hppa
   nready = select(fds, (readable)?(int *)readable:NULL, 
			     (writable)?(int *)writable:NULL,
			     0, &timeout);
#else
   nready = select(fds, (readable)?readable:NULL, 
			     (writable)?writable:NULL,
			     0, &timeout);
#endif
   if( nready == -1) 
   {  
      /* If errno is NOT for an interrupted system call */
      if (errno != EINTR)
      {
         if(uselog)
	    syslog(LOG_ERR,"Error: select() returned -1. %s",
	                   sys_errlist[errno]);
	 else
	    fprintf(stderr,"Error: select() returned -1. %s",
	                   sys_errlist[errno]);
      } 
      return -errno;
   }
   if(!startime) 
      time(&startime);

   if (nready == 0) 
   {
      time_t	now;
      time(&now);
      if( (now-startime) > idlemax ) 
      {
         sprintf(buf,"connections timed out after %d seconds.\n",idlemax);
	 if(uselog)
	    syslog(LOG_ERR,"Error: %s",buf);
	 else
	    fprintf(stderr,"Error: %s",buf);
	 return -99;
      }
   } 
   else 
   {
	    /* reset timeout counter if there is some activity */
      time(&startime);
   }
   return nready;
}
/*===========================================================================
 * function: 
 * inputs:
 *
 * outputs:
 *
 *===========================================================================*/
int saddrInit(char *host, int family, int port, struct sockaddr *addr)
{
   struct sockaddr_in *in_name;   	/* AF_INET 	*/
   struct sockaddr_un *un_name;   	/* AF_UNIX 	*/
   struct hostent     *hptr;
   char 	       buf[MAXHOSTLN+32];
   char               *p;
   unsigned long       f;

   bzero(addr,sizeof(struct sockaddr));
   if( (p=host) ) 
   {  
      /*=================================================================
       * while( *p != '\0' && (*p == '.' || isdigit(*p)) ) p++;
       *===============================================================*/
      while( *p != '\0' && *p == '.' ) 
      {
         p++;
      }
      if( *p ) 		/* not all digits or dots */ 
      {
         if( (hptr=gethostbyname(host)) == (struct hostent *)0 ) 
         {
	    sprintf(buf,"Error: hostname unknown: %s\n",host);
	    if(uselog)
	       syslog(LOG_ERR,"%s",buf);
	    else
	       fprintf(stderr,"%s",buf);
	    return -1;
         }
      } 
      else 
      {
         if( (f=inet_addr(host)) == -1L ) 
         {
	    if(uselog)
	       syslog(LOG_ERR,"Error: hostname unparsable.");
	    else
	       fprintf(stderr,"Error: hostname unparsable.");
	    return(-1);
         }
      }
   }

   switch( family )
   {
	case AF_INET :  /* inet socket init */
		in_name            	= ( struct sockaddr_in * )addr;
		in_name->sin_family	= family;
		in_name->sin_port       = htons(port);

#ifdef NOT_BSD
		in_name->sin_addr.s_addr = (host && *host) ?
					   rhost(&host):INADDR_ANY;
#else
		if( !host || !*host ) 
		   in_name->sin_addr.s_addr=INADDR_ANY;
		else 
                {
		   if( *p )
		      memcpy( &in_name->sin_addr.s_addr,
		    	     hptr->h_addr,
				     hptr->h_length );
                   else if( f != -1L )
		      (void)bcopy( (char *)&f,
					 (char *)in_name->sin_addr.s_addr,
					 sizeof f );
		 }
#endif
	       break;

	case AF_UNIX: 	/* unix domain socket init */
		un_name = ( struct sockaddr_un * ) addr;
		memset( un_name, 0, sizeof(un_name) );

		un_name ->sun_family = family;
		if( !host ) 
		   break;
		if( *p )
		   strncpy( un_name->sun_path,
			       hptr->h_addr,
			       hptr->h_length ); 
                else  if( f != -1L )
		   (void)bcopy( (char *)&f,
				  (char *)un_name->sun_path,
				  sizeof f );
	        break;

	default:
		sprintf( buf,"The family(%d) not supported yet :-(\n",family );
		if(uselog)
	           syslog(LOG_ERR,"Error: %s",buf);
	        else
	           fprintf(stderr,"Error: %s",buf); 
		return -1;
	       break;
       }
       return 0;
}
/*============================================================================
 * function: sock init
 * inputs:              
 *	ptr to hostname
 *	family
 *	port
 *	type
 *      socket address
 *
 * outputs:
 *	return socket file descriptor if OK.
 *===========================================================================*/
int sockInit(char *host, int family, int port, int type, struct sockaddr *addr)  
{
   int	sockfd;

   if( ( sockfd = socket( family,type,0 ) ) < 0 ) 
   {
      if(uselog)
	    syslog(LOG_ERR,"Error: socket() returned < 0 in sock init. %s",
	                   sys_errlist[errno]);
	 else
	    fprintf(stderr,"Error: socket() returned < 0 in sock init. %s",
	                   sys_errlist[errno]);
      return sockfd;
   }
   if( saddrInit( host,family,port,addr ) < 0 ) 
      return -1;
       
   return sockfd;
}
/*===========================================================================
 * function:
 * inputs:
 *
 * outputs:
 *
 *==========================================================================*/
int saddrlen(int family, struct sockaddr *addr)
{
   int	len = 0;
 
   struct sockaddr_in *in_name;   	/* AF_INET 	*/
   struct sockaddr_un *un_name;   	/* AF_UNIX 	*/

   if( addr ) 
   {
      switch( family ) 
      {
	case AF_INET:	
		in_name = (struct sockaddr_in *)addr;
		len = sizeof(*in_name);
	  	break;

	case AF_UNIX:	 
		un_name = (struct sockaddr_un *)addr;
		len = sizeof(*un_name);
           	break;
      }
   }
   return len;
}
/*===========================================================================
 * function:   bind port
 *
 * inputs:
 *
 * outputs:
 * 	return -1 on error
 *	from port if ok
 *===========================================================================*/
int bindport( int sd, int family, int fromport, int toport, 
	      struct sockaddr *addr, int try_only_once)
{
   int    len = saddrlen( family,addr );
   struct sockaddr_in *in_name = (struct sockaddr_in *)addr;   /* AF_INET */
   struct sockaddr_un *un_name = (struct sockaddr_un *)addr;   /* AF_UNIX */
		       
   if( fromport > toport )
      toport = fromport;

   while( bind( sd, (struct sockaddr *)addr, len ) < 0 ) 
   {
      if( errno == EADDRINUSE && !try_only_once ) 
      {
         fromport++;
         if( fromport > toport ) 
         {
	    if(uselog)
	       syslog(LOG_ERR,"Error: all ports are in use. %s",
	                   sys_errlist[errno]);
	    else
	       fprintf(stderr,"Error: all ports are in use. %s",
	                   sys_errlist[errno]);
	    return -1;
	 }
	 switch( family ) 
         {
	    defaults :
	    case AF_INET : in_name->sin_port = htons( fromport );
		  break;

	    case AF_UNIX : 
		  return -1;
                  break;
	 }
	 len = saddrlen( family, addr );
	 continue;
      } 
      else 
      {
         if(uselog)
	    syslog(LOG_ERR,"Error: bind failed. %s",
	                   sys_errlist[errno]);
	 else
	    fprintf(stderr,"Error: bind failed. %s",
	                   sys_errlist[errno]);
	 return -1;
       }
   }/*while*/
   return fromport;
}
/*===========================================================================
 * function: serv listen                         
 * inputs: 
 * 	ptr to name
 *      family
 *      socket addr  		- none null for returning socket addr
 *      ptr to from port        - return actual port number the server 
 *				  is listening to
 *	ptr to to port          - maybe parse a null for no port selection
			    	  when failed      
 * outputs:
 *	return socket file descriptor is OK.
 *	return -1 if sock init failed.
 *	return -2 if set socket options failed.
 *	return -3 if bind o a specific port failed.
 *	return -4 if listen failed.
 *
 *==========================================================================*/
int servListen(char *name, int family, struct sockaddr *saddr, int *fromport,
                int *toport, int try_only_once)
{
   int	sockfd;
   int 	one = 1;
   int  port;
   struct sockaddr addr;

   if( (sockfd = sockInit( name, family, *fromport, SOCK_STREAM, 
				(saddr)?saddr:&addr) ) < 0 )
   {
      return sockfd;
   }
      switch( family ) 
      {
	case AF_INET : 
		if( setsockopt( sockfd, SOL_SOCKET, SO_REUSEADDR, (char *) &one,
		                sizeof(one) ) < 0 ) 
		{
		   if(uselog)
	              syslog(LOG_ERR,"Error: setsockopt() returned < 0. %s",
	                             sys_errlist[errno]);
	           else
	              fprintf(stderr,"Error: setsockopt() returned < 0. %s",
	                             sys_errlist[errno]);
		   return -2;
                }
                port= (fromport)? *fromport:4;
		break;

            defaults:
            case AF_UNIX : port = -1;
		break;
      }
      /*=============================================================
       * If toport is not NULL use it, else use NULL.    
       *
       *=============================================================*/
      if( ( port = bindport( sockfd, family, port, *toport,
			          (saddr)?saddr:&addr, try_only_once)
	     ) < 0 )
      {
         sockfd = -3;
      }
      else 
      {
         if( fromport ) 
	    *fromport = port;
	 if( listen( sockfd, SOMAXCONN ) < 0 ) 
         {
	    if(uselog)
	       syslog(LOG_ERR,"Error: listen() returned < 0. %s",
	                   sys_errlist[errno]);
	    else
	       fprintf(stderr,"Error: listen() returned < 0. %s",
	                   sys_errlist[errno]);
	    sockfd = -4;
	 }
      }
      return sockfd;
}

/*============================================================================
 * function: connect
 *		Create socket and connect
 *
 * inputs:                                                                          
 *	socket address structure
 * outputs:
 *	return new socket file descriptor
 *	return -1 on error. 
 *===========================================================================*/ 
int connect2IP(struct sockaddr_in  saddr) 
{
   int	sockfd;

   if( (sockfd = socket(PF_INET,SOCK_STREAM,0)) < 0 ) 
   {
      if(uselog)
	    syslog(LOG_ERR,"Error: socket() returned < 0. %s",
	                   sys_errlist[errno]);
	 else
	    fprintf(stderr,"Error: socket() returned < 0. %s",
	                   sys_errlist[errno]);
      return -1;
   }
   if( connect(sockfd,(struct sockaddr*)&saddr,sizeof(saddr)) < 0 ) 
   {
      if(uselog)
	    syslog(LOG_ERR,"Error: connect() returned < 0. %s",
	                   sys_errlist[errno]);
	 else
	    fprintf(stderr,"Error: connect() returned < 0. %s",
	                   sys_errlist[errno]);
      close(sockfd);
      return -1;
   }
   return sockfd;
}
