/*============================================================================
 * filename:		xforward/xforward.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*=========================================================================== 	
 *
 *
 * (C) COPYRIGHT R. James Livermore 1995-2000  All rights reserved.   
 *
 * (C) COPYRIGHT Freemont Avenue Software, Incorporated 1995-2000.  All rights
 * reserved.   
 *
 *  
 *
 * Date written:	August	  20, 1995
 * Last modified:	February   8, 2000
 *	
 *
 * function:		Provide X Forwarding services. X forwarding is initiated
 *			by a authenticated tnproxy user.	
 *      
 *
 *	
 *	
 * 
 *
 *===========================================================================*/
static  char	ident[] = "@(#) xforward, V 1.0, 00/01/21"
 			  " (C) Freemont Avenue Software, Inc."
			  " 1995-2000";
/*============================================================================
 * Include Header Files 
 *===========================================================================*/
#include 	<unistd.h>
#include	<stdio.h>  
#include 	<sys/wait.h>
#include 	<sys/time.h>
#include 	<sys/param.h>
#include 	<netdb.h>
#include 	<signal.h>
#include 	<syslog.h> 
#include 	"xforward.h"
/*=============================================================================
 * External variables    
 *===========================================================================*/
 extern	char	*getarg();
 extern char    *setarg();
 extern	int	errno;
#ifndef linux
 extern sys_errlist[];
#endif
/*============================================================================
 * Private Data
 *===========================================================================*/

#define BASE_PORT	6000
#define TIME_GRAIN	3
#define MAXIDLE		3600	/* 6 hours */	
static	time_t   	ontime;   
/*============================================================================
 * Public Data
 *==========================================================================*/
#ifdef DEBUG
int        uselog = 0;
#else
int        uselog = 1;
#endif 
/*============================================================================
 * Function Prototypes
 *==========================================================================*/
void exitMsg( time_t starttime, char *msg);   
char *getenv(char *name);		 
int handleSigPid(pid_t sigpid );
list_t *xfwdLoop(int lsock, struct sockaddr_in fwd, int timemax, int idlemax,
	         int (*conn_query_cb)(), void *userdata,
	         int (*app_cb)(), int (*cldexitcb)(),
	         void *app_data, pid_t dependchild,
		 char *noprompt );

extern  char   **addstr2list(char **list, char *s, int *len);

int pmatch(char *pattern, char *string);
int mtmatch(char *pattern, char *string);     

 
/*============================================================================ 
 * function:
 * inputs:
 *
 *
 * outputs:
 *
 *===========================================================================*/
int exitXfwdCb(int lsd, struct sockaddr_in fadd, list_t *pids, fd_set fdset,
                 void *user_data)   
{ 
   if( handleSigPid((pid_t)user_data) < 0 )
   { 
      return 0;
   }
   clearList(pids,(void *)kill,(void *)SIGKILL);
   exitMsg(ontime,"Control");
   exit(1);
} 
 
#define MAXPORT 300
/*============================================================================
 * main function for xforward
 * inputs:
 * 
 *        
 * outputs:
 *
 *============================================================================*/
int main(int argc, char *argv[]) 
{  
   struct sockaddr_in	fwd;

   char 	buf[SMBUFSIZE]; 
   char		*cptr1;
   char		*cptr2;
   char		*dpy;
   char		*portarg;
   char		noprompt[16];
   char		host[MAXHOSTNAMELEN]; 
   char 	*p;
   int      	port=BASE_PORT+10; /* x port starts from 6101 */
   char		lipaddr[20]; 		/* local IP address	*/ 						    
   int		toport = MAXPORT+BASE_PORT;
   int		try_only_once = 0;	/* Do not use another port	*/
   int		len;
   int		lsd;
   int		idlemax; 
   pid_t    	pid;
   sws_l	*switches = NULL;

   /*===============================================
    * Open syslog 	
    *==============================================*/
   openlog("xforward",LOG_PID|LOG_NDELAY,LOG_DAEMON); 

   uselog = 0; /*  errors go to stderr at setup 	*/

#ifdef  BINDDEBUG
        debugbind();
#endif
#ifdef FORK
	becomeChild();
#endif
   time(&ontime);
   /*==========================================
    * Get host name of firewall
    *==========================================*/
   if( getLocalHostName(host) ) 
   {  
      syslog(LOG_ERR,"Error: get local hostname failed for xforward."); 
      goto out;  
   }
   /* get switches */
   if(!(switches=(sws_l*)rd_sws(argc,argv,"from:user:disp:time:port:noprompt")) 
	   && argc > 1) 
   {
      usage(); 
   }
   if( !(dpy = getarg("disp",switches)) && /* Get dpy from -disp switch */
       !(dpy = getarg("from",switches)) && /* Get if from fromhost      */
       !(dpy = getenv("DISPLAY")) )        /* Get it from environment   */
   {
      sprintf( buf,"%s:0.0",host ); 	/* use proxy hostname        */
      dpy = setarg("disp",buf,switches);
   } 
   else 
      sprintf(buf,"%s",dpy);         	/* save dpy in the buf       */

   if( (len = strcspn(dpy,":"))>0 && !strstr(dpy,":") ) 
   {
      sprintf(buf,"%s:0.0\0",dpy);
      dpy = setarg("disp",buf,switches);
   } 
   buf[len] = '\0'; 			/* the display hostname */
   if( (pid = atoi(dpy+len+1)) < 0 || pid > (toport-BASE_PORT) ) 
   {
      sprintf(buf,"Invalid display = %s. Must be 0-%d\n",dpy+len+1, MAXPORT);
      if(uselog)
         syslog(LOG_ERR,"Error: %s",buf);
      else
         fprintf(stderr,buf); 
      goto out;
   }

   if( !XOpenDisplay(dpy) )  		/* check the display */
   {
      fprintf( stderr,"Unable to open display = %s\n",dpy );
      goto out;
   }

   /*============================================================== 
    * Get the port argument if specified
    *==============================================================*/
   if ((portarg = getarg("port",switches)) != NULL)
   {
      int tmp;
      if ((tmp = atoi(portarg)) < 0 || tmp > 600)
      {
         sprintf(buf,"Invalid port = %s.  Must be 0-600\n", portarg);
	 if(uselog)
	    syslog(LOG_ERR,"Error: %s",buf);
	 else
	    fprintf(stderr,buf);
         goto out;
      }
      try_only_once = 1;
      port = BASE_PORT + tmp;
   }

   /*============================================================== 
    * Get the noprompt argument if specified
    *==============================================================*/
   memset (&noprompt, 0, sizeof(noprompt));
   if ((portarg = getarg("noprompt",switches)) != NULL)
   {
         strcpy(noprompt, portarg);
   }

   /*============================================================== 
    * listen to an available port of the proxy 
    *==============================================================*/
   if( (lsd = servListen(NULL,AF_INET,NULL,&port,&toport,try_only_once)) < 0 ) 
   {
      goto out;
   }
   /*==============================================================
    * init fwd destination sd 
    *===============================================================*/
   if( saddrInit(buf,AF_INET,pid+BASE_PORT,&fwd) < 0 ) 
   {
      goto out;
   }

   /*==============================================================
    * Re-enable uselog here to log display port message and to
    * prevent Solaris getpeername error 
    *=============================================================*/
   uselog = 1;

   sprintf(buf,"display port = %s:%d\n",host,port-BASE_PORT);
   syslog(LOG_NOTICE,buf);
   /*================================================================ 
    * clean up to conserve descriptors. use syslog for error outputs 
    *===============================================================*/
   close(0); 
   close(1);
#ifndef DEBUG
   close(2); 
#endif
   /*======================================================================= 
    * Setup is ok now. Popup and display msgs are on the control window. 
    *=====================================================================*/ 
   if( (cptr1 = getarg("from",switches)) == NULL)
      cptr1 = "?";  
   if( (cptr2 = getarg("user",switches)) == NULL)
      cptr2 = "?";  
   sprintf( buf, "started from %s by %s to display = %s\n",
  		 cptr1, cptr2, dpy );
   if(uselog)
      syslog(LOG_NOTICE,buf);
   else
      fprintf(stderr,buf); 


   setenv( "DISPLAY",dpy,1 );
   sprintf( buf,"Display port is %s:%d\nClick the button to exit xforward.\0",
		 host, port-BASE_PORT );
   pid = childDialog( "xforward",dpy,NULL,buf,NULL,0,"EXIT",OK,0 );

   /*======================================================================
    * The main loop for x-forward. 
    *=====================================================================*/
   sigExit(ControlExit);
   idlemax = ( p = getarg("time",switches))?atoi(p):MAXIDLE;

      xfwdLoop( lsd,            fwd, 
	  	TIME_GRAIN,	idlemax, 
		QueryConnect,  dpy, 
		exitXfwdCb,   NULL,
		(void*)pid,	pid,
		noprompt );

out:	
   fflush(stdout);
#ifndef DEBUG
   closelog();
#endif
   exit(1);
}
/*============================================================================
 * function:	Query Connect
 * inputs:
 *
 *
 * outputs:
 * 	return:	child pid if the connection request is not matched.
 *        	-1 if error.
 *===========================================================================*/
int QueryConnect( int querysock, int outgoing, fd_set *rest, fd_set *wset,
		    struct sockaddr_in saddr, void *userdata,
		    char *noprompt ) 
{
   Widget   	dialog;
   char		PeerHost[MAXHOSTLN];
   char		PeerIP[MAXIP];
   char 	buf[SMBUFSIZE];
   int      	argc = 3;

   /*==================================================================== 
    * Initialize X and create pop-up on destination: (char*)userdata 
    *===================================================================*/
   static   char *argv[] =
		{ "xforward","-display",NULL,NULL };

   argv[2] =	(char*) userdata;
   xInit( argv,(char*)userdata,argc,argv );

   /*====================================================================
    * get host name and addr from PeerHost 
    *====================================================================*/
   sprintf(PeerHost,"%s",(char*)&saddr.sin_addr);
   if( peername(0,PeerHost,PeerIP,sizeof(PeerIP),0) ) 
   {
      sprintf(buf,"Error: xforward can not get peer name.\n",0);
      if(uselog)
         syslog(LOG_ERR,buf);
      else
         fprintf(stderr,buf); 
      return -1;
   } 
   if (strncmp(noprompt, PeerIP, strlen(noprompt)) != 0) 
   {
      sprintf(buf,"Allow X connection from %s (%s)?\0",PeerHost,PeerIP);
      if(loopDialog(NULL,buf,"OK",OK,"CANCEL",CANCEL,0)!=OK)
      {
         return -1;
      }
   }
   sprintf(buf,"permit host = %s(%s) x-connection\n",PeerHost,PeerIP);
   if(uselog)
         syslog(LOG_NOTICE,buf);
      else
         fprintf(stderr,buf); 
 
   return 0;
}

/*============================================================================
 * function:  get the next token
 * inputs:
 *	ptr to string  
 *	ptr to list of token separators.
 * outputs:
 * 	return pointer to next token.
 *===========================================================================*/
char *getNextToken(char *s, char *separator)   
{
   while( *s == '\t' || *s == *separator ) 
   { s++; 
   }
   return( s );
}

/*============================================================================
 * function: get end of token
 * inputs:  
 *	ptr to string
 * 	ptr to list of token separators.
 * outputs:
 *	return ptr to first character past the token.
 *===========================================================================*/
char *getEndofToken(char *s, char *separator)  
{
   while( *s != '\t' && *s != *separator && *s != '\0' ) 
   { 
      s++; 
   }
   return( s );
}

/*============================================================================
 * function:	get tokens
 *
 * 		Extract all the tokens from the input string and
 *		create a list of ptrs to each token. 
 * inputs:
 *	ptr to input string
 *	ptr to list of token separators.
 *	number of elements
 * outputs: 
 * 	Return the address of the token ptr list. 
 *      Set the input variable n equal to the number of tokens found.  
 *===========================================================================*/
char **getTokens(char *string, char *separator, int *n) 
{
   int	cnt = 0;
   char *begin;
   char	*end;
   char	**list = (char**) 0;

   if( !string || !*string ) 
      return list;
   /* Initialize the ptrs then process string until we run out of tokens */
   begin = end = string;
   while( *begin && *end ) 
   {
      begin = getNextToken( end,separator );
      end   = getEndofToken ( begin,separator );
      cnt   = end - begin;
      list  = addstr2list( list, begin, &cnt );
    }
    if( n ) *n = cnt;
 
    return list;
}
/*============================================================================
 *function: Exit after the control is killed
 *
 *===========================================================================*/
void ControlExit()
{
   exitMsg( ontime, "Control process killed." );
   exit(1);
} 
/*==========================================================================
 * function:	handle signals
 *
 * inputs:
 *
 * outputs:
 *        
 * return: -1: WIFSIGNALED
 *         -2: no status
 *         -3: alive   other wise it is dead
 *         >=0: WEXITSTATUS
 * when >= -1 the child died
 *========================================================================*/
int handleSigPid(pid_t sigpid )
{
   pid_t        temppid;
   int          wstatus;
   int		exitstat;
   int		wopts = WNOHANG;

   temppid = waitpid(sigpid,&wstatus,wopts);

   if ( !temppid )  /* there's no status */
   {
      exitstat= -2;
      goto out;
   }

   if (WIFSIGNALED(wstatus) || WIFSTOPPED(wstatus) ) 
   {
      if(uselog)
         syslog(LOG_ERR,"Error: child either died abnormally or stopped."
	                " wstatus = %d",wstatus);
      else
         fprintf(stderr,"Error: child either died abnormally or stopped."
	                " wstatus = %d",wstatus); 
      exitstat = -1;
      goto out;
   }

   if( !WIFEXITED(wstatus) ) 
   {
      exitstat = -3;
      goto out;
   }

   exitstat = WEXITSTATUS(wstatus);
out:
#ifdef DEBUG
   fprintf(stderr,"handleSigPid pid=%d: exit_status= %d\n",
                sigpid, exitstat );
#endif  
   return exitstat; /* must >= 0 */
}

/*=============================================================================
 * function: childsigs 
 *
 * list_t *pidlist must store the id as the child pid previously.
 * Check (wait) through the list of the pidlist of children. 
 * Callback when the child is exit or signaled.
 * If the child exits delete the pid on the list. 
 *
 * inputs:
 *
 * outputs:
 *
 *==========================================================================*/
list_t *childsigs(list_t *pidlist, int (*cb)(), void *data)
{
   int	rtn;
   list_t	*p = pidlist;

   while(p) 
   { 
      if( (rtn = handleSigPid(p->id) ) >= -1 && cb) 
	 rtn = cb(p->id,data,rtn);
      /* If the return is > =-1 then delete the pid from the list */ 
      if( rtn >= -1) 
         pidlist = (list_t*)deleteListItem(pidlist,p->id);
      p = p->next;
   } 
   return pidlist;
}
/*===========================================================================
 *
 *===========================================================================*/
void sigExit( void (*funct)() )

{
   signal(SIGINT, funct); 	/* control c	*/
   signal(SIGQUIT,funct);	/* control /	*/
   signal(SIGTERM,funct);	/* kill(1)	*/
   /* Should be a way to do this on HP.  Does anyone know how? */
#ifndef __hppa
   signal(SIGXCPU,funct);	/* exceeds cpu 	*/
#endif
}
/*============================================================================
 * function: exit message
 * inputs:
 *    start time
 *   	ptr to message      
 * outputs:
 *
 *===========================================================================*/
void exitMsg( time_t starttime, char *msg)
{
   char    buf[SMBUFSIZE];
   time_t  endtime;

   time(&endtime);
   if( msg )
   { 
     sprintf( buf,"%s exit duration = %d seconds.\n", msg, endtime-starttime );
   }
   else
   {
      sprintf( buf,"exit duration = %d seconds.\n", endtime-starttime );
   }
   if(uselog)
      syslog(LOG_NOTICE,"%s",buf); 
   else
      fprintf(stderr,"%s",buf);

   if( uselog ) 
      closelog();
}   

/*============================================================================
 * 
 *===========================================================================*/ 
pid_t becomeChild()
{
   pid_t	pid;
   
   if( (pid = fork() )< 0)  
   {  
      if(uselog)
         syslog(LOG_ERR,"Error: fork failed. %s",sys_errlist[errno]);
      else
         fprintf(stderr,"Error: fork failed.  %s",sys_errlist[errno]); 
      return -1;
   }
   /*=======================================================================
    * The child returns and the parent goes away. Kind of sad.
    *=====================================================================*/
   if( !pid ) 
   { 
      return;
   }
   exit(1);
}
/*============================================================================
 *
 *==========================================================================*/
int  dup_stdio( int rfd, int wfd) 
{
   if( (rfd != STDIN_FILENO) ) 
   {
      if( dup2(rfd,STDIN_FILENO) != STDIN_FILENO ) 
      {
         if(uselog)
            syslog(LOG_ERR,"Error: dup2 failed for stdin. %s",
	                   sys_errlist[errno]);
         else
            fprintf(stderr,"Error: dup2 failed for stdin. %s",
	                   sys_errlist[errno]); 
	 return(errno);
      }
      close(rfd);
   }
   if( wfd != STDOUT_FILENO ) 
   {
      if( dup2(wfd,STDOUT_FILENO) != STDOUT_FILENO ) 
      {
         if(uselog)
            syslog(LOG_ERR,"Error: dup2 failed for stdout. %s",
	                   sys_errlist[errno]);
         else
            fprintf(stderr,"Error: dup2 failed for stdout. %s",
	                   sys_errlist[errno]); 
	 return(errno);
      }
      close(wfd);
   }
   return 0;
}
/*============================================================================= 
 * function;	 freesws 
 * 	free sws_l
 * inputs:
 *
 *===========================================================================*/
void freesws(sws_l *swsl)  
{
   sws_t	*psws;

   if(!swsl || !(psws=swsl->sws)) 
      return;
   while( psws->key ) 
   {
      free(psws->key);
      if(psws->arg) 
         free(psws->arg);
      free(psws++);
   }
   free(swsl);
}

/*===========================================================================
 * function:	rd_sws 
 * Read the switch in argv. If an argv is match to the switch defined in
 * sws->key, argv++ is to be pointed by sws->arg.
 *
 * inputs:
 *
 * outputs:
 * 	return pswsl->cnt:   0 if no args or number of input args.
 * 	return pswsl: NULL there is some thing wrong or unmatch to the inputs.
 *===========================================================================*/
sws_l *rd_sws( int argc, char *argv[], char *sws) 
{
   int	i;
   int	n;
   int	z = sizeof(sws_t);
   sws_l	*pswsl;
   sws_t	*psws;
   char	**plist;

   /*=========================================================
    * get sws as token pointered by plist 
    *=========================================================*/
   if(!( plist = (char**)getTokens(sws,":",&n) ) || !n) 
      return NULL;

   /*========================================================= 
    * malloc pswsl and init key 
    *=========================================================*/
   if((pswsl = (sws_l*) malloc(sizeof(sws_l))) == (sws_l*)NULL)
   {
      syslog(LOG_ERR,"Error: rd_sws malloc(%d) failed.",sizeof(sws_l)); 
      return 0;  
   }
   bzero((psws = pswsl->sws = (sws_t*)calloc(n+1,z)),(n+1)*z);
   while(n--)
   { 
      pswsl->sws[n].key=plist[n]; 
   }
   /* get inputs matched to sws and save in psws->arg */
   n = argc-1; 
   pswsl->cnt = 0;
   for(i=1;i<argc;i++) 
   { 
      while(psws->key) 
      {
         if(argv[i][0]=='-' && i<n &&
			  !strncmp(&argv[i][1],psws->key,strlen(psws->key)) )
         {
	    i++; 
            pswsl->cnt++;   /* got matched inputs */
	    psws->arg = strSave(argv[i],strlen(argv[i]));
	    goto next;
         } 
         psws++;
      }
      freesws(pswsl); 
      return NULL; 	/* Ooops, input not match */
next: 
      psws=pswsl->sws;
   }
   return pswsl;	/* pswsl->cnt has the # of matched inputs */
}
/*============================================================================
 * function: 	getarg 
 * 	get the arg point saved by rd_sws with key match to sws->key.
 * inputs:
 *
 * outputs:
 * 	return NULL if unmatch or no key entry. 
 * 	return the point if found in sws table.        
 *==========================================================================*/
char *getarg(char *key, sws_l *swsl) 
{
   int	n = (!key || !*key)?0:strlen(key);
   sws_t	*sws;

   if(!n || !swsl || swsl->cnt<0 || !(sws=swsl->sws)) 
      return NULL;
   while(sws->key) 
   {
      if(!strncmp(sws->key,key,n)) 
         return sws->arg;
      sws++;
   }
   return NULL;
}
/*============================================================================
 * function:	setarg 
 * 	set an arg in the sws. 
 * inputs:
 *
 *
 * outputs:
 *	return arg if successful
 *	return NULL on failure
 *===========================================================================*/
char *setarg( char *key, char *arg, sws_l *swsl)  
{
   int     n;
   sws_t   *sws;

   if(!swsl || !key || !*key) 
      return NULL;

   sws = swsl->sws; 
   n = strlen(key);
   while(sws->key) 
   {
      if(!strncmp(sws->key,key,n)) 
      {
         if(sws->arg) free(sws->arg);
	 sws->arg = (arg && *arg)?strSave(arg,strlen(arg)):NULL;
	 return sws->arg;
      } 
   sws++;
   }
   return NULL;
}

/*==========================================================================
 *function:
 *
 * inputs:
 *
 * outputs:
 *
 *=========================================================================*/
list_t *createList(int id,void *item)
{
   list_t	*plist;
   int	 	sz = sizeof(list_t);

   if( (plist = (list_t*) malloc(sz) ) == NULL )
   {
      syslog(LOG_ERR,"Error: malloc failed for plist"); 
   }
   bzero(plist,sz);
   plist->id  	= id;
   plist->item 	= item;

   return plist;
}  
/*============================================================================ 
 * function: findListItem  
 * inputs:
 *     
 *
 * outputs:
 * 	return NULL if not find
 *===========================================================================*/
list_t *findListItem( list_t *plist, int id)  
{
   list_t  *p = plist;

   if(!p)  
      return NULL;
   while( id != p->id && p->next ) 
      p=p->next;
   if(id!=p->id) 
      return NULL;

   return p;
} 
/*============================================================================
 * add_list
 * inputs: 
 *	ptr to first list
 *      any number
 *	any structure
 *
 * outputs:
 *	return: a point to the list_t: if id is matched or
 *         added item to the list if(item). 
 *         NULL: neither match the id nor add to the list.
 *==========================================================================*/ 
list_t *setList( list_t *plist, int id, void *item)
{
   list_t	*p= plist;

   if(!p)
   {	
      return( createList(id,item) );
   }
   if( p = findListItem( plist,id ) ) 
   {
      return p;
   }
   if(!item) 
      return (list_t*)NULL; 		/* nothing to add */

   p = createList(id,item);		/* not found the same item and    */
   p->next = plist;		        /* add it to the top of the list */

   return p;
}
/*=============================================================================
 * function: deleteListItem
 * inputs:     
 *
 *	
 * delete it from the linked list if id is matched 
 *===========================================================================*/
list_t	*deleteListItem( list_t *plist, int id)  
{
   list_t  *p   = plist;
   list_t  *pre = NULL;

   if(!p)
   {
      return plist;   
   }
   while( id!=p->id && p->next ) 
   {
      pre = p;
      p = p->next;
   }
   /* If id matches then remove entry from list and delete it */
   if( id == p->id ) 
   {
      if(pre) 
         pre->next = p->next;
      else	
         plist = plist->next;
      free(p);
   }
   return plist;
}
/*============================================================================ 
 * function: 	clearList 
 * 		Do a callback before clearing the list.
 * inputs:
 *		ptr to list
 *		command to be executed
 * outputs:
 *
 *===========================================================================*/
void	clearList(list_t *plist, void (*cb)(), void *data) 
{
   list_t  *lp = plist;
   list_t  *nxtp;
   while(lp) 
   {
      nxtp = lp->next;
      if(cb)
      { 
         cb(lp->id,data,lp->item);
      }
      free(lp);
      lp = nxtp;
   }
}
/*=============================================================================
 * function: asncat	Auto string n cat
 *			This function automatically acquires the storage to
 *			concatenate two binary strings of specifed length.
 * inputs:
 *	ptr to first binary string
 *	length of first binary string
 *	ptr to second binary string
 *	length of second binary string
 * outputs:	
 * return 	pointer to newly allocated binary string if OK.
 *		NULL if request failed
 *============================================================================*/
char *asncat(char *s, int sbytes, char *add, int addbytes) 
{
   int	siz = sbytes + addbytes + 1;

   if( !add || addbytes<=0 ) 
      return s;

   if(!sbytes) 
      if((s = (char*) malloc(siz)) == NULL)
      {
         syslog(LOG_ERR,"asncat malloc(%d) failed.",siz); 
      }
   else 
      if((s = (char*) realloc(s,siz)) == NULL)
      {
         syslog(LOG_ERR,"asncat realloc(%d) failed.",siz);      
      }
   if(s)   
      bcopy((void*)add,(void*)&s[sbytes],(size_t)addbytes);

   return s;
} 
/*============================================================================
 * function: 	strSave
 * 		Allocate the memory to save the string, copy it and return
 *		the address.
 * inputs:
 *		ptr to string to be saved 
 *      	length of string
 * outputs:
 *		Address of saved string
 *============================================================================*/
char * strSave(char *s, int len)
{
   char  *ns;

   if( len <= 0 ) 
      len = strlen( s );

   if((ns = (char*)malloc( len+1 )) == NULL)
   {
      syslog(LOG_ERR,"strSave malloc(%d) failed.",len+1);
      return 0;
   }  
   bzero( ns, len+1 );
   strncpy( ns,s,len );
   return ns;
}
/*===========================================================================
 * function:	addstr2list
 *		Add new string to list of strings
 * inputs:
 *		ptr to list of ch ptrs
 *		ptr of new string to add to the bottom of list
 *		ptr length of new string
 *
 * outputs:
 * 		return ptr to updated list of ptrs
 *		If the input length is not NULL retun number of elements in
 *		ptr list. 
 *
 *==========================================================================*/
char **addstr2list(char **list, char *s, int *len)
{
   int	i;
   int	n = 0;
   char **olist = list;
   /*====================================================================
    * Find the existing number of string ptrs. Allocate space for a new 
    * ptr list which includes ptr to new string and null ptr.  Then
    * copy old list to new, free old list and add new str ptr. 
    *====================================================================*/
   if( list ) 
      while( list[++n] ); 
   list = (char **)calloc( n+2, sizeof(char*) );
   for( i = 0; i < n; i++ )
   { 
      list[i] = olist[i];  
   }
   free( olist );

   list[n++] = (char*)strSave( s, (len && *len>0)? *len: 0 );
   if( len ) 
      *len = n;
   list[n] = (char*)0;

   return list;
 }

/*===========================================================================
 * function: pattern matching
 * inputs:
 *
 * outputs:
 *
 *===========================================================================*/
int pmatch(char *pattern, char *string)  
{
   register char c;
   while(1)
   {
      switch(c = *pattern++) 
      {
	case '\0': 
		return(*string == '\0');
	case '*': 
		c = *pattern;
		while(c == '*') 
		   c = *++pattern;
                if(c == '\0') 
		   return(1);
                /*=================================
		 * general case, use recursion 
		 *================================*/
                while(*string != '\0') 
		{
                   if(pmatch(pattern,string)) 
		      return(1);
                   ++string;
                }
		return(0);

	default: 
		if(c != *string++) 
		   return(0);
                break;
      }
   }
}
/*============================================================================
 *
 * inputs:
 *	ptr to pattern
 *	ptr to string
 * outputs:
 *
 *===========================================================================*/
int ptmatch(char *pattern, char *string)  
{
   int  j = 0;
   int  n = 0;
   int	cnt;			/* number of tokens found	*/
   char **list;

   list = getTokens( pattern," ",&cnt );
   if( !list ) 
      return -1;

   while( cnt-- >0 ) 
   {
      if( (pmatch( list[cnt], string ))>0 ) 
      {
         n = cnt + 1;
	 break;
      }
   }
   /*=============================
    * Free the list of strings 
    *============================*/ 
   while( list[j] )
      free( (char*)list[j++] );
   free( (char**)list );
   
   return( n );
}
/*============================================================================
 * usage 
 *==========================================================================*/
void usage()
{
   fprintf(stderr,"usage: xforward [-disp display/hostname] [-from hostname]\
 [-user username] [-port display_port] [-noprompt ip_address]\r\n" );
   exit(1);
}


