/*============================================================================
 * filename:		xforward/xfwdio.c
 *
 * This software is distributed under the terms of the Livermore Public  
 * License version 1.0.   
 *	      
 *       The Livermore PUBLIC LICENSE (LPL) Version 1.0 
 *                              
 * Copyright (C) 1999-2000 Freemont Avenue Software, Inc, Houston, Texas, USA  
 *
 * Everyone is permitted to copy and distribute this license document.  The 
 * intent of this license is to establish freedom to share and change the 
 * software regulated by this license under the open source model.
 *
 * This license applies to any software containing a notice placed by the 
 * copyright holder saying that it may be distributed under the terms of the 
 * Livermore Public  License version 1.0. Such software is herein referred to as
 * the Software. This license covers modification and distribution of the 
 * Software, use of third-party application programs based on the Software, and
 * development of free software which uses the Software.
 *
 * Granted Rights
 *
 * 1. You are granted the non-exclusive rights set forth in this license
 *    provided you agree to and comply with any and all conditions in this
 *    license. Whole or partial distribution of the Software, or software items
 *    that link with the Software, in any form signifies acceptance of this 
 *    license.
 *
 * 2. You may copy and distribute the Software in unmodified form provided that
 *    the entire package, including, but not restricted to, copyright, trademark
 *    notices and disclaimers, as released by the initial developer of the
 *    Software, is distributed.
 *
 * 3. You may make modifications to the Software and distribute your
 *    modifications, in a form that is separate from the Software, such as
 *    patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in the
 *       Software.
 *
 *    b. When modifications to the Software are released under this license, a
 *       non-exclusive royalty-free right is granted to the initial developer of
 *       the Software to distribute your modification in future versions of the
 *       Software provided such versions remain available under these terms in
 *       addition to any other license(s) of the initial developer.
 *
 * 4. You may distribute machine-executable forms of the Software or 
 *    machine-executable forms of modified versions of the Software, provided
 *    that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable forms are
 *       also able to receive the complete machine-readable source code to the
 *       distributed Software, including all modifications, without any charge 
 *       beyond the costs of data transfer, and place prominent notices in the
 *       distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the machine-
 *       -executable forms are available under the terms of this license.
 *
 * 5. You may use the original or modified versions of the Software to compile,
 *    link and run application programs legally developed by you or by others. 
 *      
 * 6. You may develop application programs, reusable components and other 
 *    software items that link with the original or modified versions of the
 *    Software. These items, when distributed, are subject to the following
 *    requirements:
 *                   
 *    a. You must ensure that all recipients of machine-executable forms of
 *       these items are also able to receive and use the complete 
 *       machine-readable source code to the items without any charge beyond
 *       the costs of data transfer.
 *           
 *    b. You must explicitly license all recipients of your items to use and 
 *       re-distribute original and modified versions of the items in both 
 *       machine-executable and source code forms. The recipients must  be able
 *       to do so without any charges whatsoever, and they must be able to 
 *       re-distribute to anyone they choose.
 * 
 *    c. If the items are not available to the general public, and the initial
 *       developer of the Software requests a copy of the items, then you must
 *       supply one.
 *	 
 * Limitations of Liability
 *
 * In no event shall the initial developers or copyright holders be liable for 
 * any damages whatsoever, including - but not restricted to - lost revenue or 
 * profits or other direct, indirect, special, incidental or consequential 
 * damages, even if they have been advised of the possibility of such damages, 
 * except to the extent invariable law, if any, provides otherwise.
 *
 * THE SOFTWARE IS PROVIDED BY FREEMONT AVENUE SOFTWARE  ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FREEMONT AVENUE SOFTWARE OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * This license is governed by the Laws of the State of Texas and any disputes
 * shall be decided by mediation.
 *
 * T.REX, T.REX free edition and other T.REX names are trademarks of Freemont 
 * Avenue Software, Inc. 
 *============================================================================*/
/*============================================================================
 *
 * (C) COPYRITHT R. James Livermore 1995-2000. All rights reserved.
 * (C) COPYRIGHT Freemont Avenue Softwarem, Incorporated 1995-2000.  All rights
 * reserved.    
 *
 * Date written:	August	  20, 1995
 * Last modified:	February   8, 2000
 *	
 *
 * function:		Provide X Forwarding I/O services. 
 *			X forwarding is initiated by a authenticated tnproxy
 *			user.	
 *	
 *	
 * 
 *============================================================================*/
static	char	cprght[] = 
  		"(C) COPYRIGHT Freemont Avenue Software, Inc. 1996-2000 xfwdio,"
  		" V 1.0 00/01/21";
		
#include <stdio.h>
#include <errno.h> 
#include "list.h"

#include <sys/types.h> 
#ifndef __hppa
#include 	<sys/select.h>		/* to support fd_set 		*/
#endif
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netdb.h>
#include <signal.h>
#include <syslog.h> 
#include "firewall.h"
#include "xforward.h"
#ifndef linux
extern char *sys_errlist[];
#endif
/*=========================================================================
 * function prototypes
 *==========================================================================*/
extern char *asncat();     
int servSelect( int fds, fd_set *readable, fd_set *writable, int timemax, 
                 int idlemax );

/*========================================================================== 
 * function: read FD
 * inputs:
 *
 *
 * outputs:
 *	return  -1 on error
 *		number of bytes read if ok.                       
 *
 *==========================================================================*/
int readfd(int *readfd, fd_set *rset, fd_set *wset, char **pbuf,int bsize)
{
   char     buf[BUFSIZE+5];
   int	 cnt;

   switch( cnt = read(*readfd,buf,BUFSIZE) ) 
   {
	case -1: 
		printf(buf,"read fd:%d %s",readfd); 
		if(uselog)
                   syslog(LOG_ERR,"Error: %s %s",buf,sys_errlist[errno]);
                else
                   fprintf(stderr,"%s %s",buf,sys_errlist[errno]);  
	case  0: 
		clearCloseFd(*readfd,rset,wset);
		return -1;
	default:
		/*=========================================================
                 * If buffer is filled then reset fd.  Then call asncat to
                 * acquire more storage and do a binary concatenation of the
		 * previous data with the new buffer.  
		 *========================================================*/   
		if(cnt == BUFSIZE) 
		   FD_SET(*readfd,rset);
		*pbuf = asncat(*pbuf,bsize,buf,cnt);
		cnt += bsize;
		break;
   }
   return cnt;
}  
/*===========================================================================
 * function: write FD
 * inputs:
 *
 *
 * outputs:
 *==========================================================================*/
int writefd(int *writefd, fd_set *rset, fd_set *wset, char **pbuf, int bsize)
{
   char buf[BUFSIZE+5];
   char	*p = NULL;
   int	cnt;

   if( !*pbuf ) 
      return 0;
   switch( (cnt = write(*writefd,*pbuf,bsize)) ) 
   {
   	case -1: 
		printf(buf,"write fd:%d",readfd);
		if(uselog)
                   syslog(LOG_ERR,"Error: %s %s",buf,sys_errlist[errno]);
                else
                   fprintf(stderr,buf,sys_errlist[errno]); 
	case  0: 
		clearCloseFd(*writefd,rset,wset);
		free(*pbuf);
		return -1;
	default: 
		if(cnt < bsize) 
		{
		   p = asncat(NULL,0,&pbuf[cnt],bsize-cnt);
		   FD_SET(*writefd,wset);
                }
		free(*pbuf);
		*pbuf= p;
		break;
   }
   return bsize-cnt;
}  
/*============================================================================ 
 * function:	pipe2sockets 
 * 	connect two sockets. 
 * 	This is a blocking loop handling/forwarding data
 * 	between two sockets so that at both of other
 * 	ends of the sockets looks just connected. 
 * inputs:
 *
 *
 *
 * outputs:
 *
 *==========================================================================*/
void pipe2sockets( int dest, int from, fd_set *rset, fd_set *wset,
		   int timemax, int idlemax, pid_t ppid )
{
   fd_set	readable;
   fd_set	writable;
   char		*pbuffrom = NULL; 	/* buf from dest to forward to from */
   char		*pbufdest = NULL; 	/* buf from from to forward to dest */
   int		szfrom    = 0;
   int		szdest    = 0; 		/* buf size of pbufxxx 	*/
   int	 	ret;
   int 		maxfd;

   maxfd = max(dest,from)+1;
   /*==============================================
    * Yet another Livermore loop. 
    *==============================================*/
   for(;;) 
   {
      memcpy(&readable, rset, sizeof(readable));

      ret = servSelect( maxfd,&readable,NULL,timemax,idlemax );
      if( ret < 0 ) 
      {
         if( -ret == EINTR ) 
            continue;
	 break;
      }
      else if( !ret ) 
         continue;

      if( FD_ISSET(from,&readable)) 
      {
         if ((szfrom = readfd(&from,rset,wset,&pbuffrom,szfrom)) < 0)
            break;
         if (pbuffrom && szfrom > 0 &&
		(szfrom = writefd(&dest,rset,wset,&pbuffrom,szfrom)) < 0)
            break;
      }
      if(FD_ISSET(dest,&readable)) 
      {
         if ((szdest = readfd(&dest,rset,wset,&pbufdest,szdest)) < 0)
            break;;
         if (pbufdest && szdest > 0 &&
		(szdest = writefd(&from,rset,wset,&pbufdest,szdest)) < 0)
            break;
      }

      if( ppid >= 0 && ppid != getppid() ) 
         break;
   } 	/* end of for loop	*/   
   /*=================================
    * When done free the buffers.
    *================================*/
   if( pbuffrom ) 
      free(pbuffrom);
   if( pbufdest ) 
      free(pbufdest);
}
/*=========================================================================== 
 * function: 
 * inputs:
 *	file descriptor
 *	socket addr
 *	fd_set
 *	fd_set
 *	ptr to call back routine when connection requested 
 *	noprompt indicator to determine whether "Allow connection" dialog is
 *		to be displayed
 *
 *
 *
 * output:
 *
 *===========================================================================*/
static pid_t query_conn( int fd, struct sockaddr_in fwd, fd_set *rset, 
			 fd_set *wset, int (*conn_query_cb)(), void *userdata,
			 int timemax, int idlemax, pid_t dependchild,
			 char *noprompt) 
{
   int	querysock;
   int 	outgoing,
   saddrlen = sizeof(struct sockaddr_in);
   pid_t               pid;
   struct sockaddr_in  saddr;

   /*==================================================================
    *  new connection 
    *==================================================================*/
   if( (querysock = accept(fd,(struct sockaddr *)&saddr,
				(addrlen_t *)&saddrlen)) < 0 ) 
   {
      if( errno != EINTR ) 
      {
         if(uselog)
            syslog(LOG_ERR,"Error: accept() failed. %s",sys_errlist[errno]);
         else
            fprintf(stderr,"Error: accept() failed. %s ",sys_errlist[errno]); 
      }
      return -1;
   }
   if( (outgoing = connect2IP(fwd)) < 0 ) 
   {
      close(querysock);
      return -1;
   }
   FD_SET(querysock,rset);	
   FD_SET(querysock,wset);
   FD_SET(outgoing, rset);	
   FD_SET(outgoing, wset);

#ifndef DEBUG
   if( (pid = fork()) < 0) 
   {
      if(uselog)
         syslog(LOG_ERR,"Error: fork failed for query connect. %s",
	                sys_errlist[errno]);
      else
         fprintf(stderr,"Error: fork failed for query connect. %s",
	                sys_errlist[errno]); 
      return -1;
   } 
   else if( pid ) 
   {
      close(querysock); 
      close(outgoing);
      return pid;
   }
#endif
   /*========================================================
    * if conn_query_cb return <0, there is something
    * error and to quit the query_conn
    *========================================================*/
   if( conn_query_cb( querysock,outgoing,rset,wset,saddr,userdata,noprompt )
						>= 0 ) 
   {
      time_t	ontime;
      time(&ontime);
      pipe2sockets(outgoing,querysock,rset,wset,timemax,idlemax,
         	   (dependchild)?getppid():-1);
      exitMsg( ontime, "Child" );
   } 
   close(querysock); 
   close(outgoing);
   exit(1);
}
/*============================================================================ 
 * function: x-forward loop
 * inputs:       
 *	local socket
 *	forward address
 *      max timeout  
 *	idle time
 *	ptr to call back routine
 *	ptr to user data 
 *	ptr to appl routines		- handle exit routines
 *					  return < 0 to stop looping    
 *	ptr to child exit callback	- handle children exit callback
 *					return < -1 to keep the pid in the list
 *	ptr to application data		- for both app_cb & cldexitcb       
 *  				 != 0: 	the children will be killed
 *       				when its parent die.
 *  				0: otherwise the children will keep alive.
 * outputs:
 *
 *===========================================================================*/
list_t *xfwdLoop(int lsock, struct sockaddr_in fwd, int timemax, int idlemax,
	         int (*conn_query_cb)(), void *userdata,
	         int (*app_cb)(), int (*cldexitcb)(),
	         void *app_data, pid_t dependchild,
		 char *noprompt )
	
{
   register	int       n;

   int		fds = lsock + 1;
   list_t	*pids = NULL;
   fd_set	allset;
   fd_set	rset;
   fd_set	wset;
   pid_t	pid;        /* may need to handle child exit */

   FD_ZERO(&allset); 
   FD_ZERO(&wset); 
   FD_ZERO(&rset); 
   FD_SET(lsock,&allset);
   /*====================================================================
    * Loop until done.
    *====================================================================*/    
   while (1) 
   {
      rset = allset;
      if( (n = servSelect( fds,&rset,NULL,timemax,idlemax )) < 0 ) 
      {
         if( -n == EINTR ) 
            continue;
	 else 
            exit(1);
      }
	    
      if(FD_ISSET(lsock,&rset)) 
      {
         pid = query_conn( lsock, fwd, 
				  &rset,	 &wset,
				  conn_query_cb, userdata,
				  timemax,	 idlemax,
				  dependchild,	 noprompt ); 
	       	if( pid<0 ) exit(1);
		pids = (list_t*) setList(pids,pid,(void*)pid);
	} 
	pids = (list_t*) childsigs(pids,cldexitcb,app_data);

        if( app_cb && app_cb( lsock,fwd,pids,allset,app_data ) < 0 ) 
	   break;
    } 
    return pids;
}

